Package: vectra
Title: Columnar Query Engine for Larger-than-RAM Data
Version: 0.11.8
Authors@R: person("Gilles", "Colling", email = "gilles.colling051@gmail.com",
    role = c("aut", "cre", "cph"),
    comment = c(ORCID = "0000-0003-3070-6066"))
Description: A minimal columnar query engine with lazy execution on datasets
    larger than RAM. Provides 'dplyr'-like verbs (filter(), select(), mutate(),
    group_by(), summarise(), joins, window functions) and common aggregations
    (n(), sum(), mean(), min(), max(), sd(), first(), last()) backed by a
    pure C11 pull-based execution engine and a custom on-disk format ('.vtr').
    Reads and writes 'GeoTIFF' (including tiled and 'BigTIFF' layouts) and a
    tiled raster format ('.vec') with overview pyramids and time cubes for
    larger-than-RAM raster data. Streams vector operations (spatial transforms,
    point-in-polygon and nearest-feature joins including a two-sided
    grid-partitioned join, select-by-location, clip, erase, dissolve,
    'rasterization', 'polygonization', and contouring) through 'sf', and runs
    raster operations (zonal statistics, focal windows, terrain derivatives,
    resample or 'reproject' warp, polygon masking, map algebra, and 'mosaicking')
    in native C or over the tiled '.vec' format, one batch or tile at a time for
    data larger than RAM.
License: MIT + file LICENSE
Depends: R (>= 4.1.0)
SystemRequirements: GNU make
Encoding: UTF-8
Imports: tidyselect, rlang, libgeos, parallel
LinkingTo: libgeos
Suggests: biglm, Biostrings, bit64, DBI, GenomicRanges, igraph,
        IRanges, knitr, openxlsx2, rmarkdown, RSQLite, S4Vectors, sf,
        ShortRead, stringdist, terra, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
URL: https://gillescolling.com/vectra/,
        https://github.com/gcol33/vectra
BugReports: https://github.com/gcol33/vectra/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-07-30 13:12:19 UTC; Gilles Colling
Author: Gilles Colling [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-3070-6066>)
Maintainer: Gilles Colling <gilles.colling051@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-31 00:00:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-07-31 01:54:17 UTC; unix
Archs: vectra.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[[",vectra_partition)
S3method(anti_join,vectra_node)
S3method(append_vtr,data.frame)
S3method(append_vtr,vectra_node)
S3method(arrange,vectra_node)
S3method(collect,data.frame)
S3method(collect,vectra_node)
S3method(collect,vectra_offload)
S3method(collect,vectra_shard)
S3method(collect_chunked,default)
S3method(collect_chunked,vectra_node)
S3method(collect_chunked,vectra_offload)
S3method(collect_chunked,vectra_partition)
S3method(collect_chunked,vectra_shard)
S3method(count,vectra_node)
S3method(cross_join,vectra_node)
S3method(dim,vectra_node)
S3method(distinct,vectra_node)
S3method(explain,vectra_node)
S3method(filter,vectra_node)
S3method(full_join,vectra_node)
S3method(fuzzy_join,vectra_node)
S3method(glimpse,vectra_node)
S3method(group_by,vectra_node)
S3method(group_map,vectra_partition)
S3method(group_modify,vectra_partition)
S3method(head,vectra_node)
S3method(inner_join,vectra_node)
S3method(interval_join,vectra_node)
S3method(kmer,vectra_node)
S3method(left_join,vectra_node)
S3method(lookup,vectra_schema)
S3method(materialize,vectra_node)
S3method(mutate,vectra_node)
S3method(print,offload_grade)
S3method(print,vectra_block)
S3method(print,vectra_grid)
S3method(print,vectra_network)
S3method(print,vectra_node)
S3method(print,vectra_partition)
S3method(print,vectra_raster)
S3method(print,vectra_schema)
S3method(print,vectra_shard)
S3method(pull,vectra_node)
S3method(reframe,vectra_node)
S3method(relocate,vectra_node)
S3method(rename,vectra_node)
S3method(resample,vectra_node)
S3method(right_join,vectra_node)
S3method(select,vectra_node)
S3method(semi_join,vectra_node)
S3method(sf::st_write,vectra_node)
S3method(slice,vectra_node)
S3method(slice_head,vectra_node)
S3method(slice_max,vectra_node)
S3method(slice_min,vectra_node)
S3method(slice_tail,vectra_node)
S3method(summarise,vectra_node)
S3method(tally,vectra_node)
S3method(transmute,vectra_node)
S3method(ungroup,vectra_node)
S3method(write_csv,data.frame)
S3method(write_csv,vectra_node)
S3method(write_sqlite,data.frame)
S3method(write_sqlite,vectra_node)
S3method(write_tiff,data.frame)
S3method(write_tiff,vectra_node)
S3method(write_vtr,data.frame)
S3method(write_vtr,vectra_node)
export(across)
export(anti_join)
export(append_vtr)
export(arrange)
export(as_embedding)
export(bind_cols)
export(bind_rows)
export(block_fuzzy_lookup)
export(block_lookup)
export(chunk_feeder)
export(collect)
export(collect_chunked)
export(collect_sf)
export(contours)
export(cosine)
export(count)
export(create_index)
export(cross_join)
export(delete_vtr)
export(desc)
export(diff_vtr)
export(distinct)
export(dot)
export(explain)
export(feature_knn)
export(filter)
export(floor_time)
export(focal)
export(full_join)
export(fuzzy_join)
export(glimpse)
export(grid)
export(group_by)
export(group_map)
export(group_modify)
export(has_index)
export(inner_join)
export(interval_join)
export(kmer)
export(l2)
export(left_join)
export(link)
export(lookup)
export(mask)
export(materialize)
export(mosaic)
export(mutate)
export(offload)
export(polygonize)
export(proximity)
export(pull)
export(rast_calc)
export(rast_feature_distance)
export(rasterize)
export(reframe)
export(relocate)
export(rename)
export(resample)
export(right_join)
export(roll_max)
export(roll_mean)
export(roll_min)
export(roll_n)
export(roll_sum)
export(select)
export(semi_join)
export(slice)
export(slice_head)
export(slice_max)
export(slice_min)
export(slice_tail)
export(spatial_centerline)
export(spatial_clip)
export(spatial_construct)
export(spatial_dissolve)
export(spatial_eliminate)
export(spatial_explode)
export(spatial_filter)
export(spatial_join)
export(spatial_knn)
export(spatial_line_merge)
export(spatial_locate)
export(spatial_map)
export(spatial_network)
export(spatial_overlay)
export(spatial_polygonize)
export(spatial_route)
export(spatial_service_area)
export(spatial_simplify)
export(spatial_smooth)
export(spatial_snap)
export(spatial_snap_grid)
export(spatial_split)
export(spatial_topology)
export(summarise)
export(summarize)
export(tally)
export(tbl)
export(tbl_bed)
export(tbl_csv)
export(tbl_fasta)
export(tbl_fastq)
export(tbl_sqlite)
export(tbl_tiff)
export(tbl_xlsx)
export(terrain)
export(tiff_band_names)
export(tiff_crs)
export(tiff_extract_points)
export(tiff_metadata)
export(transmute)
export(ungroup)
export(vec_build_overviews)
export(vec_close_raster)
export(vec_extract_points)
export(vec_open_raster)
export(vec_raster_layout)
export(vec_raster_times)
export(vec_read_pixel_series)
export(vec_read_time_slice)
export(vec_read_window)
export(vec_to_tiff)
export(vec_write_raster)
export(vec_write_time_cube)
export(vectra_mem)
export(vtr_schema)
export(warp)
export(write_csv)
export(write_sqlite)
export(write_tiff)
export(write_vtr)
export(zonal)
importFrom(libgeos,libgeos_version)
importFrom(rlang,as_function)
importFrom(rlang,enquo)
importFrom(rlang,expr)
importFrom(rlang,is_formula)
importFrom(stats,median)
importFrom(stats,sd)
importFrom(stats,var)
importFrom(utils,head)
useDynLib(vectra, .registration = TRUE)

# vectra

*querying data that won't fit in memory*

[![CRAN status](https://www.r-pkg.org/badges/version/vectra)](https://CRAN.R-project.org/package=vectra)
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/grand-total/vectra)](https://cran.r-project.org/package=vectra)
[![Monthly downloads](https://cranlogs.r-pkg.org/badges/vectra)](https://cran.r-project.org/package=vectra)
[![R-CMD-check](https://github.com/gcol33/vectra/actions/workflows/R-CMD-check.yml/badge.svg)](https://github.com/gcol33/vectra/actions/workflows/R-CMD-check.yml)
[![ASAN/UBSAN](https://github.com/gcol33/vectra/actions/workflows/sanitizers.yml/badge.svg)](https://github.com/gcol33/vectra/actions/workflows/sanitizers.yml)
[![Codecov test coverage](https://codecov.io/gh/gcol33/vectra/graph/badge.svg)](https://app.codecov.io/gh/gcol33/vectra)
[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)

**dplyr-style queries on larger-than-RAM data, backed by a pull-based columnar engine written in C11.**

Point vectra at a file too big to load and query it with the verbs you already use.
Data flows through the engine one row group at a time, so peak memory stays bounded
no matter how large the file gets. Arrow needs compiled binaries that match your
platform, DuckDB links a bundled library, Spark wants a JVM. vectra is a standard R
extension with no external dependencies: it compiles where R compiles.

```r
library(vectra)

# Lazy scan over a multi-GB CSV; nothing runs until collect()
tbl_csv("measurements.csv") |>
  filter(temperature > 30, year >= 2020) |>
  group_by(station) |>
  summarise(avg_temp = mean(temperature), n = n()) |>
  collect()
```

## One engine, several file formats

`.vtr` (vectra's own columnar format), CSV, SQLite, and GeoTIFF all open into the same
lazy query nodes, so the same pipeline runs against any of them:

```r
# GeoTIFF climate raster as tidy data
tbl_tiff("worldclim_bio1.tif") |>
  filter(band1 > 0) |>
  mutate(temp_c = band1 / 10) |>
  collect()

# SQLite without DBI
tbl_sqlite("survey.db", "responses") |>
  filter(year == 2025) |>
  left_join(tbl_sqlite("survey.db", "sites"), by = "site_id") |>
  collect()
```

Convert to `.vtr` once for repeated queries, then read it back fast:

```r
write_vtr(big_df, "data.vtr", batch_size = 100000)

tbl("data.vtr") |>
  filter(x > 0, region == "EU") |>
  group_by(region) |>
  summarise(total = sum(value), n = n()) |>
  collect()
```

## How the engine stays bounded

The optimizer rewrites the plan before any data moves, and execution streams the
result rather than materializing it:

- **Streaming execution**: one row group flows through at a time, never the whole file
- **Column pruning**: columns you never reference are skipped at the scan
- **Predicate pushdown**: per-rowgroup min/max statistics skip row groups that cannot match a filter
- **Hash joins**: build the right side, stream the left; join a fact table against a lookup without holding both in memory
- **External sort**: a memory budget with automatic spill to disk

`explain()` shows the optimized plan, including which columns and row groups were pruned:

```r
tbl("data.vtr") |>
  filter(x > 0) |>
  select(id, x) |>
  explain()
#> vectra execution plan
#>
#> ProjectNode [streaming]
#>   FilterNode [streaming]
#>     ScanNode [streaming, 2/5 cols (pruned), predicate pushdown, v3 stats]
```

## Fuzzy matching in the engine

String distance runs inside the C engine, with no round-trip to R per row:

```r
tbl("taxa.vtr") |>
  filter(levenshtein(species, "Quercus robur") <= 2) |>
  mutate(similarity = jaro_winkler(species, "Quercus robur")) |>
  arrange(desc(similarity)) |>
  collect()
```

`levenshtein()`, `dl_dist()` (Damerau-Levenshtein), and `jaro_winkler()` are available in
`filter()` and `mutate()`, alongside `nchar()`, `substr()`, `grepl()`, `gsub()` and the rest
of the string toolkit.

## Star schemas instead of flat-table column creep

Register dimension tables once, then pull columns from any of them; joins are built for
you, and unmatched keys are reported:

```r
s <- vtr_schema(
  fact    = tbl("observations.vtr"),
  species = link("sp_id", tbl("species.vtr")),
  site    = link("site_id", tbl("sites.vtr"))
)

lookup(s, count, species$name, site$habitat) |> collect()
#> species: all 500 keys matched
#> site: 3/500 unmatched keys (X1, X2, X3)
```

## Incremental updates

Append new rows as a new row group without rewriting the file, or take a key-based diff
between two snapshots:

```r
append_vtr(new_rows_df, "data.vtr")

d <- diff_vtr("snapshot_old.vtr", "snapshot_new.vtr", key_col = "id")
collect(d$added)   # rows present in new but not old
d$deleted          # key values present in old but not new
```

## Streaming results into a model fit

`collect()` brings the whole result into memory. `collect_chunked()` folds a
function over a query one batch at a time, holding a single batch plus the
accumulator, so a result larger than memory reduces to a small summary in one
pass: a running count, per-group sufficient statistics, or the cross-products
behind a linear fit.

```r
# Accumulate X'X and X'y for an exact OLS fit, one streaming batch at a time
acc <- tbl("survey.vtr") |>
  select(mpg, wt, hp) |>
  collect_chunked(
    function(acc, chunk) {
      X <- cbind(1, chunk$wt, chunk$hp)
      list(XtX = acc$XtX + crossprod(X),
           Xty = acc$Xty + crossprod(X, chunk$mpg))
    },
    .init = list(XtX = matrix(0, 3, 3), Xty = matrix(0, 3, 1))
  )
solve(acc$XtX, acc$Xty)
```

For models that re-read the data on every iteration, `chunk_feeder()` exposes a
query as a resettable generator that `biglm::bigglm()` drives directly, fitting
a GLM on data too large to hold in memory:

```r
src <- function() tbl("occurrences.vtr") |> select(presence, bio1, bio12)
biglm::bigglm(presence ~ bio1 + bio12, data = chunk_feeder(src),
              family = binomial())
```

`offload()` spills a prepared query to disk once and streams it back, so each
reweighted pass reads the prepared columns from a file rather than rebuilding
the pipeline. With a `by` key it splits the query into per-key shards on disk,
each small enough to pull into memory on its own. `group_map()` then runs your
function on each shard, passing the shard as an ordinary data.frame along with
its group key, and returns the results keyed by shard. That function is
arbitrary R: a model fit, a clustering call, a custom summary, or any
computation that needs the whole group in memory at once. `group_modify()`
recombines per-shard data.frames into one table.

```r
# Prepare once, then let bigglm re-read the spill on every pass
s <- offload(tbl("occurrences.vtr") |> select(presence, bio1, bio12))
biglm::bigglm(presence ~ bio1 + bio12, data = chunk_feeder(s),
              family = binomial())

# Per-region: group_map runs any function on each in-memory shard (here a GLM)
p <- offload(tbl("occurrences.vtr"), by = "region")
fits <- group_map(p, function(d, region)
  glm(presence ~ bio1 + bio12, data = d, family = binomial()))
```

## Verbs

vectra covers the dplyr surface most analysis pipelines use: `filter()`, `select()`,
`mutate()`, `group_by()` / `summarise()` with the common aggregations, all the joins
(`left_join()` through `cross_join()`, plus `fuzzy_join()`), `arrange()` and the
`slice_*()` family, and window functions (`row_number()`, `rank()`, `lag()`, `lead()`,
`cumsum()`, `ntile()`, and more). Date/time and string functions evaluate in the engine.
`select()`, `rename()`, `relocate()`, and `across()` accept the full tidyselect helper set
(`starts_with()`, `where()`, `all_of()`, and the rest).

The [Function Reference](https://gillescolling.com/vectra/reference/) lists every verb and
expression with examples.

## Installation

```r
install.packages("vectra")            # CRAN

install.packages("pak")               # development version
pak::pak("gcol33/vectra")
```

## Documentation

- [Getting Started](https://gillescolling.com/vectra/articles/quickstart.html)
- [Engine Reference](https://gillescolling.com/vectra/articles/engine.html)
- [Format Backends](https://gillescolling.com/vectra/articles/formats.html)
- [Joins](https://gillescolling.com/vectra/articles/joins.html)
- [Star Schemas](https://gillescolling.com/vectra/articles/schema.html)
- [String Operations](https://gillescolling.com/vectra/articles/string-ops.html)
- [Indexing and Optimization](https://gillescolling.com/vectra/articles/indexing.html)
- [Working with Large Data](https://gillescolling.com/vectra/articles/large-data.html)
- [Species Distribution Models](https://gillescolling.com/vectra/articles/sdm.html)

## Support

> "Software is like sex: it's better when it's free." — Linus Torvalds

I'm a PhD student who builds R packages in my free time because I believe good tools
should be free and open. I started these projects for my own work and figured others
might find them useful too.

If this package saved you some time, buying me a coffee is a nice way to say thanks.
It helps with my coffee addiction.

[![Buy Me A Coffee](https://img.shields.io/badge/-Buy%20me%20a%20coffee-FFDD00?logo=buymeacoffee&logoColor=black)](https://buymeacoffee.com/gcol33)

## License

MIT (see the LICENSE.md file)

## Citation

```bibtex
@software{vectra,
  author = {Colling, Gilles},
  title  = {vectra: Columnar Query Engine for Larger-Than-RAM Data},
  year   = {2026},
  url    = {https://github.com/gcol33/vectra}
}
```

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----va-engine, echo = FALSE, results = "asis", eval = TRUE-------------------
cat("<script>\n", paste(readLines("va-engine.js"), collapse = "\n"), "\n</script>\n", sep = "")

## ----streaming-memory-anim, echo = FALSE, results = "asis", eval = TRUE-------
body <- "
var s=VA.setup('sm-cv'); if(!s)return; var x=s.ctx,W=s.w,H=s.h,C=VA.C;
var PX=72,PY=58,PW=W-72-36,PH=H-58-66,RAM=0.74;
var PERIOD=12.0;
function draw(tc){VA.bg(x,W,H);
  VA.title(x,'Memory as rows are processed');
  x.strokeStyle=C.mut;x.lineWidth=1;x.beginPath();x.moveTo(PX,PY);x.lineTo(PX,PY+PH);x.lineTo(PX+PW,PY+PH);x.stroke();
  x.fillStyle=C.mut;x.font=VA.F(11);x.textAlign='right';x.fillText('memory',PX-6,PY+10);x.textAlign='center';x.fillText('rows processed',PX+PW/2,PY+PH+24);
  var ry=PY+PH*(1-RAM);x.strokeStyle=C.red;x.setLineDash([6,4]);x.beginPath();x.moveTo(PX,ry);x.lineTo(PX+PW,ry);x.stroke();x.setLineDash([]);x.fillStyle=C.red;x.font=VA.F(11,true);x.textAlign='left';x.fillText('RAM limit',PX+6,ry-6);
  var t=VA.ease(VA.clamp(tc/(PERIOD-1.2),0,1)),le=Math.floor(t*PW);
  var redMax=1.16;
  x.strokeStyle=C.red;x.lineWidth=2.5;x.beginPath();
  for(var i=0;i<=le;i++){var v=(i/PW)*redMax;x.lineTo(PX+i,PY+PH*(1-Math.min(v,1.02)));}x.stroke();
  var g=0.16;x.strokeStyle=C.green;x.lineWidth=2.5;x.beginPath();
  for(var i=0;i<=le;i++){x.lineTo(PX+i,PY+PH*(1-g-0.012*Math.sin(i/13)));}x.stroke();
  var LX=PX+Math.min(le,PW-96),SY=PY+PH*(1-g)-9;
  x.fillStyle=C.green;x.font=VA.F(12,true);x.textAlign='left';x.fillText('streaming',LX,SY);
  var redv=t*redMax,ly=PY+PH*(1-Math.min(redv,1.02));x.fillStyle=C.red;x.fillText('in memory',LX,Math.min(Math.max(PY+13,ly-9),SY-20));
  x.fillStyle=C.mut;x.font=VA.F(11);x.textAlign='center';x.fillText('streaming keeps the footprint flat as the file grows',W/2,H-14);
}
VA.run(draw,PERIOD,null,'sm','sm-cv');
"
cat(paste0(
  "<canvas id='sm-cv' width='760' height='340' style='max-width:100%'></canvas>\n",
  "<script>\n(function(){\n", body, "\n})();\n</script>\n"))

## ----ram-box-anim, echo = FALSE, results = "asis", eval = TRUE----------------
body <- "
var s=VA.setup('rb-cv'); if(!s)return; var x=s.ctx,W=s.w,H=s.h,C=VA.C;
var BX=150,BW=W-150-40,RAM=0.7,ramX=BX+BW*RAM,BH=52,PERIOD=11.0;
var rows=[
 {y:88,name:'in memory',over:true,segs:[['data',0.44,'cyan'],['working copy',0.34,'amber'],['R',0.15,'purple']]},
 {y:176,name:'streaming',over:false,segs:[['batch',0.09,'green'],['R',0.15,'purple']]}
];
function draw(tc){VA.bg(x,W,H);
  VA.title(x,'What has to fit in RAM');
  x.strokeStyle=C.red;x.setLineDash([6,4]);x.beginPath();x.moveTo(ramX,58);x.lineTo(ramX,H-46);x.stroke();x.setLineDash([]);x.fillStyle=C.red;x.font=VA.F(12,true);x.textAlign='left';x.fillText('RAM',ramX+8,70);
  var t=VA.ease(VA.clamp(tc/(PERIOD-1.6),0,1));
  for(var r=0;r<2;r++){var rw=rows[r],X=BX;
    x.fillStyle=C.ink;x.font=VA.F(13,true);x.textAlign='right';x.fillText(rw.name,BX-14,rw.y+BH/2+5);
    for(var k=0;k<rw.segs.length;k++){var seg=rw.segs[k],w=seg[1]*BW*t;
      x.fillStyle=C[seg[2]];x.globalAlpha=0.85;x.fillRect(X,rw.y,w,BH);x.globalAlpha=1;x.strokeStyle=C.bg;x.lineWidth=1.5;x.strokeRect(X,rw.y,w,BH);
      x.font=VA.F(11,true);if(w>x.measureText(seg[0]).width+12){x.fillStyle=C.bg;x.textAlign='center';x.fillText(seg[0],X+w/2,rw.y+BH/2+4);}
      X+=w;}
    if(rw.over&&X>ramX){var ox=Math.max(ramX,BX);x.save();x.beginPath();x.rect(ox,rw.y,X-ox,BH);x.clip();x.strokeStyle=VA.A(C.red,0.5);x.lineWidth=1.5;for(var hx=ox-BH;hx<X;hx+=10){x.beginPath();x.moveTo(hx,rw.y+BH);x.lineTo(hx+BH,rw.y);x.stroke();}x.restore();x.strokeStyle=C.red;x.lineWidth=2;x.strokeRect(ox,rw.y,X-ox,BH);x.lineWidth=1;}
    if(!rw.over){x.fillStyle=C.green;x.font=VA.F(12,true);x.textAlign='left';x.fillText('fits',X+12,rw.y+BH/2+5);}}
  x.fillStyle=C.mut;x.font=VA.F(11);x.textAlign='center';x.fillText('streaming holds one batch and R at a time',W/2,H-14);
}
VA.run(draw,PERIOD,null,'rb','rb-cv');
"
cat(paste0(
  "<canvas id='rb-cv' width='760' height='300' style='max-width:100%'></canvas>\n",
  "<script>\n(function(){\n", body, "\n})();\n</script>\n"))

## ----lazy-pipeline-anim, echo = FALSE, results = "asis", eval = TRUE----------
body <- "
var s=VA.setup('lp-cv'); if(!s)return; var x=s.ctx,W=s.w,H=s.h,C=VA.C;
var nodes=[{n:'scan',d:'read row group',c:'cyan'},{n:'filter',d:'cyl > 4',c:'amber'},{n:'project',d:'mpg, hp',c:'green'},{n:'collect',d:'data.frame',c:'purple'}];
var NN=nodes.length,BW=300,BX=W/2-BW/2,BH=44,GAPY=22,TOP=70;
var BUILD=0.9,buildEnd=NN*BUILD+0.3,PULL=5.0,PERIOD=buildEnd+2*PULL;
function ny(i){return TOP+i*(BH+GAPY);}
function box(i,active){var y=ny(i),nd=nodes[i],col=C[nd.c];x.fillStyle=active?VA.A(col,0.12):C.bg;x.strokeStyle=active?col:VA.A(C.mut,0.45);x.lineWidth=active?2:1;x.beginPath();x.rect(BX,y,BW,BH);x.fill();x.stroke();x.lineWidth=1;x.fillStyle=active?C.ink:C.mut;x.font=VA.F(13,true);x.textAlign='left';x.fillText(nd.n+'()',BX+14,y+19);x.fillStyle=col;x.font=VA.F(11);x.fillText(nd.d,BX+14,y+35);}
function arrow(i){var y0=ny(i)+BH,y1=ny(i+1);x.strokeStyle=VA.A(C.mut,0.6);x.lineWidth=1;x.beginPath();x.moveTo(BX+BW/2,y0);x.lineTo(BX+BW/2,y1);x.stroke();}
function token(tx,ty,col,lab){x.fillStyle=col;x.beginPath();x.arc(tx,ty,6,0,6.2832);x.fill();x.font=VA.F(10,true);x.textAlign='center';x.fillText(lab,tx,ty-12);}
function draw(tc){VA.bg(x,W,H);
  VA.title(x,'A lazy plan, then one batch pulled through');
  var built=tc<buildEnd?Math.min(NN,Math.floor(tc/BUILD)+1):NN;
  for(var i=0;i<built-1;i++)arrow(i);
  var pulling=tc>=buildEnd,reqPos=-1,batchPos=-1;
  if(pulling){var pt=((tc-buildEnd)%PULL)/PULL;if(pt<0.4)reqPos=(NN-1)-(pt/0.4)*(NN-1);else batchPos=((pt-0.4)/0.6)*(NN-1);}
  for(var i=0;i<built;i++){var lit=pulling?((reqPos>=0&&Math.abs(i-reqPos)<0.55)||(batchPos>=0&&Math.abs(i-batchPos)<0.55)):true;box(i,lit);}
  if(reqPos>=0)token(BX-26,ny(0)+BH/2+reqPos*(BH+GAPY),C.cyan,'pull');
  if(batchPos>=0)token(BX+BW+26,ny(0)+BH/2+batchPos*(BH+GAPY),C.green,'batch');
  x.fillStyle=C.mut;x.font=VA.F(11);x.textAlign='center';x.fillText(pulling?'collect() requests; one batch flows back up the tree':'building the plan, no data moves yet',W/2,H-16);
}
VA.run(draw,PERIOD,null,'lp','lp-cv');
"
cat(paste0(
  "<canvas id='lp-cv' width='760' height='348' style='max-width:100%'></canvas>\n",
  "<script>\n(function(){\n", body, "\n})();\n</script>\n"))

## ----write-read---------------------------------------------------------------
library(vectra)

f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

node <- tbl(f)
node

## ----collect------------------------------------------------------------------
tbl(f) |> collect() |> head()

## ----write-batch-size---------------------------------------------------------
f_batched <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f_batched, batch_size = 10)
tbl(f_batched) |> collect() |> nrow()

## ----filter-and---------------------------------------------------------------
tbl(f) |>
  filter(cyl == 6, mpg > 19) |>
  select(mpg, cyl, hp, wt) |>
  collect()

## ----filter-or----------------------------------------------------------------
tbl(f) |>
  filter(cyl == 4 | cyl == 8) |>
  select(mpg, cyl) |>
  collect() |>
  head()

## ----filter-in----------------------------------------------------------------
tbl(f) |>
  filter(cyl %in% c(4, 6)) |>
  select(mpg, cyl) |>
  collect() |>
  head()

## ----select-helpers-----------------------------------------------------------
tbl(f) |>
  select(starts_with("d"), mpg) |>
  collect() |>
  head()

## ----select-negate------------------------------------------------------------
tbl(f) |>
  select(-am, -vs, -gear, -carb) |>
  collect() |>
  head()

## ----explain-filter-----------------------------------------------------------
tbl(f) |>
  filter(cyl > 4) |>
  select(mpg, cyl, hp) |>
  explain()

## ----mutate-arith-------------------------------------------------------------
tbl(f) |>
  mutate(kpl = mpg * 0.425144, hp_per_wt = hp / wt) |>
  select(mpg, kpl, hp, wt, hp_per_wt) |>
  collect() |>
  head()

## ----mutate-math--------------------------------------------------------------
tbl(f) |>
  mutate(
    log_hp = log(hp),
    hp_floor = floor(hp / 10) * 10,
    bounded = pmin(pmax(mpg, 15), 25)
  ) |>
  select(hp, log_hp, hp_floor, mpg, bounded) |>
  collect() |>
  head()

## ----transmute----------------------------------------------------------------
tbl(f) |>
  transmute(
    efficiency = mpg / wt,
    power_ratio = hp / disp
  ) |>
  collect() |>
  head()

## ----mutate-cast--------------------------------------------------------------
tbl(f) |>
  mutate(cyl_str = as.character(cyl)) |>
  select(cyl, cyl_str) |>
  collect() |>
  head(3)

## ----mutate-control-----------------------------------------------------------
tbl(f) |>
  mutate(
    size = case_when(
      cyl == 4 ~ "small",
      cyl == 6 ~ "medium",
      cyl == 8 ~ "large"
    ),
    mpg_class = if_else(mpg > 20, "high", "low"),
    in_range = between(hp, 100, 200)
  ) |>
  select(cyl, size, mpg, mpg_class, hp, in_range) |>
  collect() |>
  head()

## ----mutate-coalesce----------------------------------------------------------
df_na <- data.frame(
  a = c(NA, 2, NA, 4),
  b = c(10, NA, NA, 40),
  stringsAsFactors = FALSE
)
f_na <- tempfile(fileext = ".vtr")
write_vtr(df_na, f_na)

tbl(f_na) |>
  mutate(filled = coalesce(a, b, 0)) |>
  collect()

## ----string-data--------------------------------------------------------------
people <- data.frame(
  name = c("  Alice  ", "Bob", "Charlie Brown", "Diana"),
  city = c("Amsterdam", "Berlin", "Chicago", "Dublin"),
  email = c("alice@example.com", "bob@test.org",
            "charlie.b@work.net", "diana@example.com"),
  stringsAsFactors = FALSE
)
fs <- tempfile(fileext = ".vtr")
write_vtr(people, fs)

## ----string-basic-------------------------------------------------------------
tbl(fs) |>
  mutate(
    name_trimmed = trimws(name),
    name_len = nchar(trimws(name)),
    city_prefix = substr(city, 1, 3)
  ) |>
  select(name_trimmed, name_len, city_prefix) |>
  collect()

## ----string-case--------------------------------------------------------------
tbl(fs) |>
  mutate(
    city_upper = toupper(city),
    is_example = endsWith(email, "example.com"),
    starts_a = startsWith(city, "A")
  ) |>
  select(city_upper, email, is_example, starts_a) |>
  collect()

## ----string-grepl-------------------------------------------------------------
tbl(fs) |>
  mutate(has_at = grepl("@example", email)) |>
  select(email, has_at) |>
  collect()

## ----string-gsub--------------------------------------------------------------
tbl(fs) |>
  mutate(domain = gsub(".*@", "", email, fixed = FALSE)) |>
  select(email, domain) |>
  collect()

## ----string-extract-----------------------------------------------------------
tbl(fs) |>
  mutate(user = str_extract(email, "^[^@]+")) |>
  select(email, user) |>
  collect()

## ----string-paste-------------------------------------------------------------
tbl(fs) |>
  mutate(
    greeting = paste0("Hello, ", trimws(name), "!"),
    label = paste(trimws(name), city, sep = " - ")
  ) |>
  select(greeting, label) |>
  collect()

## ----summarise-basic----------------------------------------------------------
tbl(f) |>
  group_by(cyl) |>
  summarise(
    count = n(),
    avg_mpg = mean(mpg),
    total_hp = sum(hp),
    best_mpg = max(mpg)
  ) |>
  collect()

## ----summarise-advanced-------------------------------------------------------
tbl(f) |>
  group_by(cyl) |>
  summarise(
    mpg_sd = sd(mpg),
    mpg_var = var(mpg),
    first_hp = first(hp),
    last_hp = last(hp)
  ) |>
  collect()

## ----summarise-median---------------------------------------------------------
tbl(f) |>
  group_by(cyl) |>
  summarise(
    med_mpg = median(mpg),
    unique_gears = n_distinct(gear)
  ) |>
  collect()

## ----count--------------------------------------------------------------------
tbl(f) |>
  count(cyl, sort = TRUE) |>
  collect()

## ----tally--------------------------------------------------------------------
tbl(f) |>
  group_by(gear) |>
  tally() |>
  collect()

## ----across-summarise---------------------------------------------------------
tbl(f) |>
  group_by(cyl) |>
  summarise(across(c(mpg, hp, wt), mean)) |>
  collect()

## ----across-multi-------------------------------------------------------------
tbl(f) |>
  group_by(cyl) |>
  summarise(across(
    c(mpg, hp),
    list(avg = mean, total = sum),
    .names = "{.col}_{.fn}"
  )) |>
  collect()

## ----ungroup------------------------------------------------------------------
tbl(f) |>
  group_by(cyl, gear) |>
  summarise(n = n(), .groups = "keep") |>
  ungroup() |>
  arrange(desc(n)) |>
  collect()

## ----arrange------------------------------------------------------------------
tbl(f) |>
  select(mpg, cyl, hp) |>
  arrange(cyl, desc(mpg)) |>
  collect() |>
  head(8)

## ----slice-head---------------------------------------------------------------
tbl(f) |>
  slice_head(n = 5) |>
  collect()

## ----slice-min----------------------------------------------------------------
tbl(f) |>
  select(mpg, cyl, hp) |>
  slice_min(order_by = mpg, n = 3) |>
  collect()

## ----slice-no-ties------------------------------------------------------------
tbl(f) |>
  select(mpg, cyl) |>
  slice_min(order_by = cyl, n = 3, with_ties = FALSE) |>
  collect()

## ----slice-max----------------------------------------------------------------
tbl(f) |>
  select(mpg, cyl, hp) |>
  slice_max(order_by = hp, n = 4, with_ties = FALSE) |>
  collect()

## ----join-setup---------------------------------------------------------------
cyl_info <- data.frame(
  cyl = c(4, 6, 8),
  engine_type = c("inline", "v-type", "v-type"),
  stringsAsFactors = FALSE
)
f_cyl <- tempfile(fileext = ".vtr")
write_vtr(cyl_info, f_cyl)

## ----left-join----------------------------------------------------------------
tbl(f) |>
  select(mpg, cyl, hp) |>
  left_join(tbl(f_cyl), by = "cyl") |>
  collect() |>
  head()

## ----semi-anti----------------------------------------------------------------
tbl(f) |>
  select(mpg, cyl) |>
  anti_join(
    tbl(f_cyl) |> filter(engine_type == "v-type"),
    by = "cyl"
  ) |>
  collect() |>
  head()

## ----join-named---------------------------------------------------------------
ratings <- data.frame(
  cylinders = c(4, 6, 8),
  rating = c("A", "B", "C"),
  stringsAsFactors = FALSE
)
f_rat <- tempfile(fileext = ".vtr")
write_vtr(ratings, f_rat)

tbl(f) |>
  select(mpg, cyl) |>
  inner_join(tbl(f_rat), by = c("cyl" = "cylinders")) |>
  collect() |>
  head()

## ----fuzzy-join---------------------------------------------------------------
ref_species <- data.frame(
  canonical = c("Quercus robur", "Quercus petraea",
                 "Fagus sylvatica"),
  code = c("QR", "QP", "FS"),
  stringsAsFactors = FALSE
)
query_species <- data.frame(
  name = c("Quercus robur", "Qurecus petraea",
           "Fagus sylvatca"),
  stringsAsFactors = FALSE
)
f_ref <- tempfile(fileext = ".vtr")
f_query <- tempfile(fileext = ".vtr")
write_vtr(ref_species, f_ref)
write_vtr(query_species, f_query)

tbl(f_query) |>
  fuzzy_join(
    tbl(f_ref),
    by = c("name" = "canonical"),
    method = "dl",
    max_dist = 0.15
  ) |>
  collect()

## ----window-rank--------------------------------------------------------------
tbl(f) |>
  select(mpg, cyl, hp) |>
  slice_head(n = 8) |>
  mutate(
    rn = row_number(),
    mpg_rank = rank(mpg),
    mpg_dense = dense_rank(mpg)
  ) |>
  collect()

## ----window-lag-lead----------------------------------------------------------
tbl(f) |>
  select(mpg, hp) |>
  slice_head(n = 6) |>
  mutate(
    prev_mpg = lag(mpg),
    next_mpg = lead(mpg),
    prev2_hp = lag(hp, n = 2, default = 0)
  ) |>
  collect()

## ----window-cum---------------------------------------------------------------
tbl(f) |>
  select(mpg, hp) |>
  slice_head(n = 6) |>
  mutate(
    running_hp = cumsum(hp),
    running_avg = cummean(mpg),
    running_min = cummin(mpg)
  ) |>
  collect()

## ----window-grouped-----------------------------------------------------------
tbl(f) |>
  select(mpg, cyl) |>
  group_by(cyl) |>
  mutate(rn = row_number(), pct = percent_rank(mpg)) |>
  slice_head(n = 10) |>
  collect()

## ----date-data----------------------------------------------------------------
events <- data.frame(
  event_date = as.Date(c("2020-03-15", "2020-07-01",
                          "2021-01-15", "2021-06-30")),
  event_time = as.POSIXct(c("2020-03-15 09:30:00",
                             "2020-07-01 14:00:00",
                             "2021-01-15 08:15:00",
                             "2021-06-30 17:45:00"),
                           tz = "UTC"),
  value = c(10, 20, 30, 40)
)
fd <- tempfile(fileext = ".vtr")
write_vtr(events, fd)

## ----date-extract-------------------------------------------------------------
tbl(fd) |>
  mutate(
    yr = year(event_date),
    mo = month(event_date),
    dy = day(event_date)
  ) |>
  group_by(yr) |>
  summarise(total = sum(value)) |>
  collect()

## ----time-extract-------------------------------------------------------------
tbl(fd) |>
  mutate(
    hr = hour(event_time),
    mn = minute(event_time)
  ) |>
  select(event_time, hr, mn) |>
  collect()

## ----date-filter--------------------------------------------------------------
tbl(fd) |>
  filter(event_date >= as.Date("2021-01-01")) |>
  collect()

## ----date-arith---------------------------------------------------------------
tbl(fd) |>
  mutate(plus_30 = event_date + 30) |>
  select(event_date, plus_30) |>
  collect()

## ----similarity-data----------------------------------------------------------
species <- data.frame(
  name = c("Quercus robur", "Quercus rubra",
           "Fagus sylvatica", "Acer platanoides",
           "Quercus petraea"),
  stringsAsFactors = FALSE
)
fs2 <- tempfile(fileext = ".vtr")
write_vtr(species, fs2)

## ----similarity-metrics-------------------------------------------------------
tbl(fs2) |>
  mutate(
    lev = levenshtein(name, "Quercus robur"),
    dl = dl_dist(name, "Quercus robur"),
    jw = jaro_winkler(name, "Quercus robur")
  ) |>
  filter(lev <= 5) |>
  arrange(lev) |>
  collect()

## ----similarity-norm----------------------------------------------------------
tbl(fs2) |>
  mutate(
    lev_norm = levenshtein_norm(name, "Quercus robur"),
    dl_norm = dl_dist_norm(name, "Quercus robur")
  ) |>
  collect()

## ----dl-transposition---------------------------------------------------------
tbl(fs2) |>
  mutate(
    lev = levenshtein(name, "Qurecus robur"),
    dl = dl_dist(name, "Qurecus robur")
  ) |>
  collect()

## ----resolve------------------------------------------------------------------
taxa <- data.frame(
  id        = c(1L, 2L, 3L, 4L),
  name      = c("Fagaceae", "Quercus", "Q. robur", "Q. petraea"),
  parent_id = c(NA, 1L, 2L, 2L),
  stringsAsFactors = FALSE
)
ft <- tempfile(fileext = ".vtr")
write_vtr(taxa, ft)

tbl(ft) |>
  mutate(parent_name = resolve(parent_id, id, name)) |>
  collect()

## ----propagate----------------------------------------------------------------
tbl(ft) |>
  mutate(family = propagate(
    parent_id, id,
    if_else(is.na(parent_id), name, NA_character_)
  )) |>
  collect()

## ----csv-roundtrip------------------------------------------------------------
csv_in <- tempfile(fileext = ".csv")
write.csv(mtcars, csv_in, row.names = FALSE)

tbl_csv(csv_in) |>
  filter(cyl == 6) |>
  select(mpg, cyl, hp) |>
  collect()

## ----sqlite-roundtrip---------------------------------------------------------
db <- tempfile(fileext = ".sqlite")
f_src <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f_src)
tbl(f_src) |> write_sqlite(db, "cars")

tbl_sqlite(db, "cars") |>
  filter(mpg > 25) |>
  collect()

## ----format-conversion--------------------------------------------------------
csv_file <- tempfile(fileext = ".csv")
vtr_file <- tempfile(fileext = ".vtr")
csv_out <- tempfile(fileext = ".csv")

write.csv(mtcars, csv_file, row.names = FALSE)
tbl_csv(csv_file) |> write_vtr(vtr_file)

tbl(vtr_file) |>
  filter(cyl == 6) |>
  write_csv(csv_out)

read.csv(csv_out) |> head()

## ----index-create-------------------------------------------------------------
f_idx <- tempfile(fileext = ".vtr")
write_vtr(
  data.frame(id = letters, val = 1:26, stringsAsFactors = FALSE),
  f_idx,
  batch_size = 5
)

has_index(f_idx, "id")  # FALSE
create_index(f_idx, "id")
has_index(f_idx, "id")  # TRUE

## ----index-query--------------------------------------------------------------
tbl(f_idx) |>
  filter(id == "m") |>
  collect()

## ----index-composite----------------------------------------------------------
f_comp <- tempfile(fileext = ".vtr")
write_vtr(
  data.frame(
    region = rep(c("north", "south"), each = 13),
    id = letters,
    val = 1:26,
    stringsAsFactors = FALSE
  ),
  f_comp,
  batch_size = 5
)
create_index(f_comp, c("region", "id"))

tbl(f_comp) |>
  filter(region == "north", id == "c") |>
  collect()

## ----append-------------------------------------------------------------------
fa <- tempfile(fileext = ".vtr")
write_vtr(mtcars[1:16, ], fa)
append_vtr(mtcars[17:32, ], fa)
tbl(fa) |> collect() |> nrow()

## ----delete-------------------------------------------------------------------
delete_vtr(fa, c(0, 1, 2))  # 0-based row indices
tbl(fa) |> collect() |> nrow()
unlink(c(fa, paste0(fa, ".del")))

## ----diff---------------------------------------------------------------------
fd1 <- tempfile(fileext = ".vtr")
fd2 <- tempfile(fileext = ".vtr")
old <- data.frame(id = 1:5, val = letters[1:5],
                  stringsAsFactors = FALSE)
new <- data.frame(id = c(3L, 4L, 5L, 6L, 7L),
                  val = c("C", "d", "e", "f", "g"),
                  stringsAsFactors = FALSE)
write_vtr(old, fd1)
write_vtr(new, fd2)

d <- diff_vtr(fd1, fd2, "id")
d$deleted
collect(d$added)
unlink(c(fd1, fd2))

## ----block-materialize--------------------------------------------------------
blk_data <- data.frame(
  taxonID = c("T1", "T2", "T3", "T4", "T5"),
  name = c("Quercus robur", "Pinus sylvestris",
           "Fagus sylvatica", "Acer campestre",
           "Betula pendula"),
  stringsAsFactors = FALSE
)
f_blk <- tempfile(fileext = ".vtr")
write_vtr(blk_data, f_blk)

blk <- materialize(tbl(f_blk))
blk

## ----block-lookup-------------------------------------------------------------
block_lookup(blk, "name", c("Quercus robur", "Betula pendula"))

## ----block-fuzzy--------------------------------------------------------------
block_fuzzy_lookup(
  blk, "name",
  c("Qurecus robur", "Pinus silvestris"),
  method = "dl",
  max_dist = 0.2
)

## ----explain-full-------------------------------------------------------------
tbl(f) |>
  filter(cyl > 4) |>
  select(mpg, cyl, hp) |>
  arrange(desc(mpg)) |>
  explain()

## ----glimpse------------------------------------------------------------------
tbl(f) |> glimpse()

## ----cleanup------------------------------------------------------------------
unlink(c(f, f_batched, f_na, fs, fs2, f_cyl, f_rat, f_ref, f_query, fd,
         ft, csv_in, csv_out, csv_file, vtr_file, db, f_src, f_idx,
         paste0(f_idx, ".id.vtri"), f_comp,
         paste0(f_comp, ".region_id.vtri"), f_blk))

## ----setup, include = FALSE---------------------------------------------------
has_sf <- requireNamespace("sf", quietly = TRUE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = has_sf,
  fig.width = 7,
  fig.height = 4.2,
  out.width = "100%",
  dpi = 96
)

## ----libs, message = FALSE----------------------------------------------------
library(vectra)
library(sf)

## ----load-nc------------------------------------------------------------------
nc <- st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
nc <- st_transform(nc, 32119)        # NAD83 / North Carolina, metres
crs_nc <- st_crs(nc)
nrow(nc)

## ----write-poly---------------------------------------------------------------
f_poly <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME  = nc$NAME,
  BIR74 = nc$BIR74,
  SID74 = nc$SID74,
  geometry = st_as_binary(st_geometry(nc), hex = TRUE)
), f_poly)

## ----map-buffer---------------------------------------------------------------
cent <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME = nc$NAME,
  geometry = st_as_binary(st_centroid(st_geometry(nc)), hex = TRUE)
), cent)

buffered <- tbl(cent) |>
  spatial_map(~ st_buffer(.x, 15000), crs = crs_nc)
buffered

## ----map-buffer-plot----------------------------------------------------------
b_sf <- collect_sf(buffered)
plot(st_geometry(nc), border = "grey80", col = NA,
     main = "County centroids buffered by 15 km")
plot(st_geometry(b_sf), border = "#3366cc", col = "#3366cc22", add = TRUE)

## ----map-area-----------------------------------------------------------------
areas <- tbl(f_poly) |>
  spatial_map(~ data.frame(NAME = .x$NAME,
                           area_km2 = as.numeric(st_area(.x)) / 1e6),
              crs = crs_nc)
head(collect(areas))

## ----smooth-------------------------------------------------------------------
zig <- st_linestring(rbind(c(0, 0), c(1, 1), c(2, 0), c(3, 1), c(4, 0)))
f_zig <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1L, geometry = st_as_binary(st_sfc(zig), hex = TRUE)), f_zig)

tbl(f_zig) |> spatial_smooth(iterations = 3) |> collect_sf()
unlink(f_zig)

## ----sample-points------------------------------------------------------------
set.seed(1)
pts <- st_coordinates(st_sample(st_union(nc), 500))
fp <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), fp)

## ----filter-region------------------------------------------------------------
region <- nc[nc$NAME %in% c("Ashe", "Alleghany", "Surry", "Wilkes", "Watauga"),
             "NAME"]

keep_xy <- tbl(fp) |>
  spatial_filter(region, coords = c("x", "y"), crs = crs_nc) |>
  collect()
nrow(keep_xy)

## ----filter-plot--------------------------------------------------------------
plot(st_geometry(nc), border = "grey85", col = NA, main = "Select by location")
plot(st_geometry(region), border = "#cc3344", col = "#cc334411", add = TRUE)
points(pts, pch = 16, cex = 0.5, col = "grey70")
points(keep_xy$x, keep_xy$y, pch = 16, cex = 0.6, col = "#cc3344")

## ----filter-distance----------------------------------------------------------
near <- tbl(fp) |>
  spatial_filter(region, predicate = st_is_within_distance,
                 coords = c("x", "y"), crs = crs_nc, dist = 30000)
nrow(collect(near))

## ----clip---------------------------------------------------------------------
mask_region <- st_union(st_geometry(region))

clipped <- tbl(f_poly) |> spatial_clip(mask_region, crs = crs_nc)
c_sf <- collect_sf(clipped)
nrow(c_sf)

## ----clip-plot----------------------------------------------------------------
plot(st_geometry(nc), border = "grey85", col = NA,
     main = "Counties clipped to the region")
plot(st_geometry(c_sf), border = "#2a9d5c", col = "#2a9d5c33", add = TRUE)

## ----split--------------------------------------------------------------------
square <- st_polygon(list(rbind(c(0, 0), c(4, 0), c(4, 4), c(0, 4), c(0, 0))))
blade  <- st_sfc(st_linestring(rbind(c(2, -1), c(2, 5))))

f_sq <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1L, geometry = st_as_binary(st_sfc(square), hex = TRUE)), f_sq)

tbl(f_sq) |> spatial_split(blade) |> collect_sf()
unlink(f_sq)

## ----join-tag-----------------------------------------------------------------
tagged <- tbl(fp) |>
  spatial_join(nc["NAME"], coords = c("x", "y"), crs = crs_nc)
tdf <- collect(tagged)
head(tdf[, c("id", "NAME")])

## ----join-plot----------------------------------------------------------------
tagged_sf <- st_as_sf(tdf, coords = c("x", "y"), crs = crs_nc)
plot(st_geometry(nc), border = "grey85", col = NA,
     main = "Points tagged with their county")
plot(tagged_sf["NAME"], pch = 16, cex = 0.5, add = TRUE)

## ----join-nearest-------------------------------------------------------------
ncent <- st_sf(NAME = nc$NAME, geometry = st_centroid(st_geometry(nc)))
nearest <- tbl(fp) |>
  spatial_join(ncent, join = st_nearest_feature,
               coords = c("x", "y"), crs = crs_nc)
nrow(collect(nearest))

## ----join-partition-----------------------------------------------------------
g_poly <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME = nc$NAME,
  geometry = st_as_binary(st_geometry(nc), hex = TRUE)
), g_poly)

tagged2 <- tbl(fp) |>
  spatial_join(tbl(g_poly), coords = c("x", "y"), crs = crs_nc,
               partition = grid(80000))
t2 <- collect(tagged2)
sum(!is.na(t2$NAME))

## ----join-equality------------------------------------------------------------
streamed <- collect(
  tbl(fp) |> spatial_join(nc["NAME"], coords = c("x", "y"), crs = crs_nc))

resident <- st_join(
  st_as_sf(collect(tbl(fp)), coords = c("x", "y"), crs = crs_nc, remove = FALSE),
  nc["NAME"], join = st_intersects)

all.equal(streamed$NAME[order(streamed$id)],
          resident$NAME[order(resident$id)])

## ----filter-scale-------------------------------------------------------------
set.seed(42)
bb <- st_bbox(nc)
n_big <- 2e5
big <- data.frame(id = seq_len(n_big),
                  x = runif(n_big, bb["xmin"], bb["xmax"]),
                  y = runif(n_big, bb["ymin"], bb["ymax"]))
fbig <- tempfile(fileext = ".vtr")
write_vtr(big, fbig)

kept <- tbl(fbig) |>
  spatial_filter(region, coords = c("x", "y"), crs = crs_nc) |>
  collect()
nrow(kept)

## ----cleanup-scale, include = FALSE-------------------------------------------
unlink(fbig)

## ----knn----------------------------------------------------------------------
towns <- suppressWarnings(st_centroid(st_geometry(nc)))[1:5]
towns <- st_sf(town = nc$NAME[1:5], geometry = towns)

set.seed(1)
pts <- suppressWarnings(st_coordinates(st_sample(nc, 100)))
f_pts <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), f_pts)

tbl(f_pts) |>
  spatial_knn(towns, k = 2, coords = c("x", "y"), crs = crs_nc, y_id = "town") |>
  collect() |> head()
unlink(f_pts)

## ----dissolve-----------------------------------------------------------------
nc$band <- ifelse(nc$SID74 > 5, "high", "low")
fb <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  band = nc$band, BIR74 = nc$BIR74,
  geometry = st_as_binary(st_geometry(nc), hex = TRUE)
), fb)

merged <- tbl(fb) |>
  spatial_dissolve(by = "band", crs = crs_nc,
                   .fun = list(births = function(d) sum(d$BIR74)))
m_sf <- collect_sf(merged)
m_sf

## ----dissolve-plot------------------------------------------------------------
plot(m_sf["band"], main = "Counties dissolved into two SIDS bands")

## ----overlay------------------------------------------------------------------
sq <- function(a, b) st_polygon(list(rbind(
  c(a, 0), c(b, 0), c(b, 1), c(a, 1), c(a, 0))))
polys <- st_sf(year = c(1990L, 2010L, 2000L),
               geometry = st_sfc(sq(0, 2), sq(1, 3), sq(1.5, 3.5)))

pieces <- collect_sf(spatial_overlay(polys))
nrow(pieces)
length(unique(pieces$piece_id))

## ----overlay-resolve----------------------------------------------------------
first <- spatial_overlay(polys) |>
  group_by(piece_id) |>
  slice_min(year, n = 1, with_ties = FALSE) |>
  collect_sf()
nrow(first)

plot(first["year"], main = "Overlay pieces, earliest year wins")

## ----overlay-two--------------------------------------------------------------
zones <- st_sf(zone = c("A", "B"),
               geometry = st_sfc(sq(0, 1.5), sq(1.5, 3)))

inter <- spatial_overlay(polys, zones, how = "intersection") |> collect_sf()
inter

## ----roundtrip----------------------------------------------------------------
out <- tempfile(fileext = ".vtr")
tbl(fp) |>
  spatial_filter(region, coords = c("x", "y"), crs = crs_nc) |>
  write_vtr(out)
nrow(collect(tbl(out)))

## ----cleanup, include = FALSE-------------------------------------------------
unlink(c(f_poly, cent, fp, g_poly, fb, out))

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
set.seed(1)
has_biglm <- requireNamespace("biglm", quietly = TRUE)
library(vectra)

## ----fold-setup---------------------------------------------------------------
f <- tempfile(fileext = ".vtr")
df <- data.frame(g = rep(c("a", "b", "c"), length.out = 6000),
                 wt = rnorm(6000, 3, 1), hp = rnorm(6000, 150, 40))
df$mpg <- 30 - 5 * df$wt - 0.05 * df$hp + rnorm(6000, 0, 1)
write_vtr(df, f)

## ----fold-count---------------------------------------------------------------
collect_chunked(tbl(f), function(acc, chunk) acc + nrow(chunk), .init = 0L)

## ----fold-ols-----------------------------------------------------------------
acc <- collect_chunked(
  tbl(f) |> select(mpg, wt, hp),
  function(acc, chunk) {
    X <- cbind(1, chunk$wt, chunk$hp); y <- chunk$mpg
    list(XtX = acc$XtX + crossprod(X), Xty = acc$Xty + crossprod(X, y))
  },
  .init = list(XtX = matrix(0, 3, 3), Xty = matrix(0, 3, 1))
)
drop(solve(acc$XtX, acc$Xty))
coef(lm(mpg ~ wt + hp, data = df))

## ----offload-identity---------------------------------------------------------
s <- offload(tbl(f) |> select(mpg, wt, hp))

identical(collect(s), collect(tbl(f) |> select(mpg, wt, hp)))
s

## ----offload-bigglm, eval = has_biglm-----------------------------------------
s <- offload(tbl(f) |> select(mpg, wt, hp))
fit <- biglm::bigglm(mpg ~ wt + hp, data = chunk_feeder(s),
                     family = gaussian())
coef(fit)

## ----offload-bigglm-note, echo = FALSE, results = "asis", eval = !has_biglm----
# cat("> The out-of-core GLM example needs the **biglm** package, which is not",
#     "installed, so it was not run here.\n")

## ----arrange------------------------------------------------------------------
tbl(f) |> arrange(desc(mpg)) |> head(3) |> collect()

## ----partition----------------------------------------------------------------
p <- offload(tbl(f), by = "g")
p

## ----partition-fits-----------------------------------------------------------
fits <- group_map(p, function(d, g) coef(lm(mpg ~ wt + hp, data = d)))
fits

## ----partition-modify---------------------------------------------------------
group_modify(p, function(d, g)
  data.frame(n = nrow(d), mean_mpg = mean(d$mpg)))

## ----partition-reduce---------------------------------------------------------
accumulate <- function(acc, chunk) {
  X <- cbind(1, chunk$wt, chunk$hp); y <- chunk$mpg
  list(XtX = acc$XtX + crossprod(X), Xty = acc$Xty + crossprod(X, y))
}
combine <- function(a, b) list(XtX = a$XtX + b$XtX, Xty = a$Xty + b$Xty)

acc <- collect_chunked(
  offload(tbl(f) |> select(g, mpg, wt, hp), by = "g"),
  accumulate,
  .init = list(XtX = matrix(0, 3, 3), Xty = matrix(0, 3, 1)),
  combine = combine, commutative = TRUE
)
drop(solve(acc$XtX, acc$Xty))
coef(lm(mpg ~ wt + hp, data = df))

## ----cleanup, include = FALSE-------------------------------------------------
unlink(f)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## ----va-engine, echo = FALSE, results = "asis", eval = TRUE-------------------
cat("<script>\n", paste(readLines("va-engine.js"), collapse = "\n"), "\n</script>\n", sep = "")

## -----------------------------------------------------------------------------
# library(vectra)

## ----raster-to-points-anim, echo = FALSE, results = "asis", eval = TRUE-------
body <- "
var s=VA.setup('m2p-cv'); if(!s)return; var x=s.ctx,W=s.w,H=s.h,C=VA.C;
var GX=26,GY=72,COLS=8,ROWS=5,TW=50,TH=50,GW=COLS*TW,GH=ROWS*TH;
var N=COLS*ROWS;
var LX=GX+GW+70, LY=86, RH=15;
var seed=20260625;function rnd(){seed=(seed*1103515245+12345)&0x7fffffff;return seed/0x7fffffff;}
var V=[];for(var i=0;i<N;i++){V.push(VA.clamp(0.18+0.62*((i%COLS)/(COLS-1))+0.18*(rnd()-0.5),0,1));}
var PER=0.42, PERIOD=N*PER+1.8;
function cell(i){var c=i%COLS,r=Math.floor(i/COLS);return {c:c,r:r,cx:GX+c*TW+TW/2,cy:GY+r*TH+TH/2};}
function draw(tc){VA.bg(x,W,H);
  VA.title(x,'A raster read as one row per cell');
  var done=Math.min(Math.floor(tc/PER),N);
  for(var i=0;i<N;i++){var p=cell(i);var lit=i<done;
    x.fillStyle=lit?VA.A(C.green,(0.22+0.62*V[i]).toFixed(3)):C.off;
    x.fillRect(p.cx-TW/2+2,p.cy-TH/2+2,TW-4,TH-4);
    x.strokeStyle=C.grid;x.strokeRect(p.cx-TW/2+2,p.cy-TH/2+2,TW-4,TH-4);}
  if(done<N){var pc=cell(done);var f=(tc/PER)-done;
    x.strokeStyle=C.amber;x.lineWidth=2;x.strokeRect(pc.cx-TW/2+2,pc.cy-TH/2+2,TW-4,TH-4);x.lineWidth=1;
    var ty=LY+Math.min(done,15)*RH;
    var tx=VA.lerp(pc.cx,LX-12,VA.easeOut(f)),tyy=VA.lerp(pc.cy,ty,VA.easeOut(f));
    x.fillStyle=C.green;x.beginPath();x.arc(tx,tyy,4,0,6.2832);x.fill();}
  x.fillStyle=C.cyan;x.font=VA.F(12,true);x.textAlign='left';x.fillText('x     y     value',LX,LY-16);
  x.strokeStyle=C.grid;x.beginPath();x.moveTo(LX,LY-9);x.lineTo(W-16,LY-9);x.stroke();
  var show=Math.min(done,15);
  for(var k=0;k<show;k++){var idx=done-show+k;if(idx<0)continue;var p2=cell(idx);
    x.fillStyle=C.ink;x.font=VA.F(11);x.textAlign='left';
    x.fillText(p2.c+'     '+(ROWS-1-p2.r)+'     '+V[idx].toFixed(2),LX,LY+6+k*RH);}
  x.fillStyle=C.mut;x.font=VA.F(11);x.textAlign='left';x.fillText(done+' / '+N+' cells read',LX,H-18);
}
VA.run(draw,PERIOD,null,'m2p','m2p-cv');
"
cat(paste0(
  "<canvas id='m2p-cv' width='760' height='372' style='max-width:100%'></canvas>\n",
  "<script>\n(function(){\n", body, "\n})();\n</script>\n"))

## -----------------------------------------------------------------------------
# clim <- vec_open_raster("worldclim_bio.vec")
# 
# occ <- read.csv("occurrences.csv")          # species, x, y
# occ <- cbind(occ, vec_extract_points(clim, occ$x, occ$y))
# head(occ)

## ----select-by-location-anim, echo = FALSE, results = "asis", eval = TRUE-----
body <- "
var s=VA.setup('sbl-cv'); if(!s)return; var x=s.ctx,W=s.w,H=s.h,C=VA.C;
var PER=9.5,HOLD=2.0,PERIOD=PER+HOLD;
var poly=[[330,128],[470,116],[548,214],[472,322],[332,332],[262,216]];
var seed=77;function rnd(){seed=(seed*1103515245+12345)&0x7fffffff;return seed/0x7fffffff;}
var P=[];for(var i=0;i<56;i++){P.push([60+rnd()*(W-120),64+rnd()*(H-118)]);}
P.sort(function(a,b){return a[0]-b[0];});
function inPoly(px,py){var inside=false;for(var i=0,j=poly.length-1;i<poly.length;j=i++){var xi=poly[i][0],yi=poly[i][1],xj=poly[j][0],yj=poly[j][1];var hit=((yi>py)!=(yj>py))&&(px<(xj-xi)*(py-yi)/(yj-yi)+xi);if(hit)inside=!inside;}return inside;}
function drawPoly(){x.save();x.strokeStyle=C.amber;x.lineWidth=2;x.beginPath();x.moveTo(poly[0][0],poly[0][1]);for(var i=1;i<poly.length;i++)x.lineTo(poly[i][0],poly[i][1]);x.closePath();x.stroke();x.fillStyle=VA.A(C.amber,0.09);x.fill();x.restore();}
function draw(tc){VA.bg(x,W,H);
  VA.title(x,'Select by location');
  drawPoly();
  x.fillStyle=C.amber;x.font=VA.F(12,true);x.textAlign='center';x.fillText('study region',405,222);
  var sweep=VA.clamp(tc/PER,0,1)*W;
  var kept=0,seen=0;
  for(var i=0;i<P.length;i++){var px=P[i][0],py=P[i][1];var reached=px<=sweep;var ins=inPoly(px,py);
    if(reached){seen++;if(ins)kept++;}
    var col,a;
    if(!reached){col=C.mut;a=0.5;} else if(ins){col=C.green;a=1;} else {col=C.red;a=0.26;}
    x.globalAlpha=a;x.fillStyle=col;x.beginPath();x.arc(px,py,4.2,0,6.2832);x.fill();x.globalAlpha=1;}
  if(tc<PER){x.strokeStyle=C.cyan;x.lineWidth=2;x.beginPath();x.moveTo(sweep,46);x.lineTo(sweep,H-34);x.stroke();x.lineWidth=1;}
  x.fillStyle=C.green;x.font=VA.F(13,true);x.textAlign='left';x.fillText('kept '+kept,16,H-20);
  x.fillStyle=C.red;x.fillText('dropped '+(seen-kept),110,H-20);
  x.fillStyle=C.mut;x.font=VA.F(11);x.textAlign='right';x.fillText('spatial_filter(points, region)',W-16,H-18);
}
VA.run(draw,PERIOD,null,'sbl','sbl-cv');
"
cat(paste0(
  "<canvas id='sbl-cv' width='760' height='400' style='max-width:100%'></canvas>\n",
  "<script>\n(function(){\n", body, "\n})();\n</script>\n"))

## -----------------------------------------------------------------------------
# region <- sf::st_read("study_area.gpkg", quiet = TRUE)
# 
# inside <- tbl("occurrences.vtr") |>
#   spatial_filter(region, coords = c("x", "y"), crs = 4326)

## -----------------------------------------------------------------------------
# clipped <- tbl("rivers.vtr") |> spatial_clip(region, crs = 4326)

## -----------------------------------------------------------------------------
# tagged <- tbl("gbif_points.vtr") |>
#   spatial_join(tbl("countries.vtr"), coords = c("x", "y"),
#                crs = 4326, partition = grid(1))

## ----rasterize-anim, echo = FALSE, results = "asis", eval = TRUE--------------
body <- "
var s=VA.setup('rz-cv'); if(!s)return; var x=s.ctx,W=s.w,H=s.h,C=VA.C;
var GX=44,GW=W-88,COLS=16,ROWS=8,TW=GW/COLS,GY=128,TH=27,GH=ROWS*TH;
var PERIOD=14.0,FALL=0.9;
var seed=531;function rnd(){seed=(seed*1103515245+12345)&0x7fffffff;return seed/0x7fffffff;}
var D=[];var cxC=10.5,cyC=4.5;
for(var i=0;i<150;i++){var c=Math.round(VA.clamp(cxC+(rnd()-0.5)*9+(rnd()-0.5)*5,0,COLS-1));var r=Math.round(VA.clamp(cyC+(rnd()-0.5)*5+(rnd()-0.5)*3,0,ROWS-1));D.push([c,r,rnd()]);}
function draw(tc){VA.bg(x,W,H);
  VA.title(x,'Rasterize: points to a density grid');
  var cnt=new Array(COLS*ROWS).fill(0),maxc=1;
  for(var i=0;i<D.length;i++){var rt=D[i][2]*PERIOD*0.82;if(tc>=rt+FALL){var k=D[i][1]*COLS+D[i][0];cnt[k]++;if(cnt[k]>maxc)maxc=cnt[k];}}
  for(var r=0;r<ROWS;r++)for(var c=0;c<COLS;c++){var k=r*COLS+c;var v=cnt[k]/maxc;
    x.fillStyle=v>0?VA.A(C.cyan,(0.14+0.78*v).toFixed(3)):C.off;
    x.fillRect(GX+c*TW+1,GY+r*TH+1,TW-2,TH-2);}
  x.strokeStyle=C.grid;for(var c2=0;c2<=COLS;c2++){x.beginPath();x.moveTo(GX+c2*TW,GY);x.lineTo(GX+c2*TW,GY+GH);x.stroke();}
  for(var r2=0;r2<=ROWS;r2++){x.beginPath();x.moveTo(GX,GY+r2*TH);x.lineTo(GX+GW,GY+r2*TH);x.stroke();}
  for(var i=0;i<D.length;i++){var rt=D[i][2]*PERIOD*0.82;var fall=tc-rt;
    if(fall>=0&&fall<FALL){var f=fall/FALL;var tcx=GX+(D[i][0]+0.5)*TW,tcy=GY+(D[i][1]+0.5)*TH;var py=VA.lerp(62,tcy,VA.ease(f));
      x.fillStyle=C.green;x.globalAlpha=1-0.25*f;x.beginPath();x.arc(tcx,py,3.4,0,6.2832);x.fill();x.globalAlpha=1;}}
  x.fillStyle=C.cyan;x.font=VA.F(12,true);x.textAlign='right';x.fillText('brighter = more points per cell',W-16,GY-12);
  x.fillStyle=C.mut;x.font=VA.F(11);x.textAlign='left';x.fillText('one batch streams past; the grid stays resident',16,H-18);
}
VA.run(draw,PERIOD,null,'rz','rz-cv');
"
cat(paste0(
  "<canvas id='rz-cv' width='760' height='378' style='max-width:100%'></canvas>\n",
  "<script>\n(function(){\n", body, "\n})();\n</script>\n"))

## -----------------------------------------------------------------------------
# # Point density on a continental grid, streamed from a billion-row file.
# density <- tbl("gbif_points.vtr") |>
#   rasterize(extent = c(-180, -90, 180, 90), res = 0.1,
#             fun = "count", path = "density.vec")

## -----------------------------------------------------------------------------
# admin <- sf::st_read("regions.gpkg", quiet = TRUE)
# zonal(clim, admin, fun = c("mean", "sd"), zone_field = "region_id")

## -----------------------------------------------------------------------------
# terrain("dem.vec", v = c("slope", "aspect", "hillshade"),
#         path = "dem_derivatives.vec")

## -----------------------------------------------------------------------------
# warp("dem.vec", list(crs = 3035, res = 25),
#      method = "bilinear", path = "dem_laea.vec")

## -----------------------------------------------------------------------------
# habitat <- polygonize("landcover.vec")       # one polygon per class

## -----------------------------------------------------------------------------
# isolines <- contours("dem.vec", levels = seq(0, 3000, by = 200))

## -----------------------------------------------------------------------------
# study <- mask("dem.vec", region, path = "dem_study.vec")

## -----------------------------------------------------------------------------
# ndvi <- rast_calc(list(nir = "nir.vec", red = "red.vec"),
#                   (nir - red) / (nir + red), path = "ndvi.vec")

## -----------------------------------------------------------------------------
# tile_merge <- mosaic(list("n50.vec", "n51.vec", "n52.vec"), fun = "mean")

## -----------------------------------------------------------------------------
# dist_to_road <- proximity("roads.vec", path = "road_distance.vec")
# sea_distance <- proximity("landcover.vec", target = 0)   # 0 = water

## ----cost-tiers-anim, echo = FALSE, results = "asis", eval = TRUE-------------
body <- "
var s=VA.setup('ct-cv'); if(!s)return; var x=s.ctx,W=s.w,H=s.h,C=VA.C;
var PERIOD=12.0;
var rows=[
 {name:'monoid fold',sub:'spatial_filter, rasterize, zonal',col:'green',f:function(t){return 0.15+0.02*Math.sin(t*9);}},
 {name:'sort / partition',sub:'focal, terrain, warp, partitioned join',col:'amber',f:function(t){return 0.16+0.2*(0.5-0.5*Math.cos(t*6.5));}},
 {name:'all-to-all',sub:'self-overlay, voronoi, neighbour graphs',col:'red',f:function(t){return VA.clamp(t,0,1)*0.9+0.05;}}
];
var PX=310,PW=W-310-92,BH=56,GAP=36,TOP=74;
function draw(tc){VA.bg(x,W,H);
  VA.title(x,'Cost-model tiers: memory over a run');
  var t=tc/PERIOD;
  for(var i=0;i<3;i++){var y=TOP+i*(BH+GAP);var rw=rows[i];
    x.strokeStyle=C.red;x.setLineDash([5,4]);x.beginPath();x.moveTo(PX,y-7);x.lineTo(PX+PW,y-7);x.stroke();x.setLineDash([]);
    x.fillStyle=C.off;x.fillRect(PX,y,PW,BH);x.strokeStyle=C.grid;x.strokeRect(PX,y,PW,BH);
    var cursor=VA.ease(VA.clamp(t/0.98,0,1)),steps=Math.floor(cursor*PW);
    x.fillStyle=C[rw.col];x.globalAlpha=0.82;
    for(var px=0;px<steps;px++){var hv=VA.clamp(rw.f(px/PW),0,1);x.fillRect(PX+px,y+BH-hv*BH,1.4,hv*BH);}
    x.globalAlpha=1;
    if(cursor<1){var cx2=PX+steps;x.strokeStyle=C.cyan;x.beginPath();x.moveTo(cx2,y);x.lineTo(cx2,y+BH);x.stroke();}
    x.fillStyle=C[rw.col];x.font=VA.F(14,true);x.textAlign='right';x.fillText(rw.name,PX-14,y+24);
    x.fillStyle=C.mut;x.font=VA.F(9);x.fillText(rw.sub,PX-14,y+40);
    x.fillStyle=C.red;x.font=VA.F(10);x.textAlign='left';x.fillText('RAM',PX+PW+8,y-3);}
  x.fillStyle=C.mut;x.font=VA.F(11);x.textAlign='center';x.fillText('fold and partition stay under the RAM line; all-to-all reaches it',W/2,H-16);
}
VA.run(draw,PERIOD,null,'ct','ct-cv');
"
cat(paste0(
  "<canvas id='ct-cv' width='760' height='400' style='max-width:100%'></canvas>\n",
  "<script>\n(function(){\n", body, "\n})();\n</script>\n"))

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
set.seed(1)
has_biglm <- requireNamespace("biglm", quietly = TRUE)
library(vectra)

## ----reproject, eval = FALSE--------------------------------------------------
# # Occurrences recorded in lon/lat (EPSG:4326), raster in a projected CRS.
# target <- tiff_crs("predictors.tif")$epsg          # e.g. 3035 (LAEA Europe)
# 
# pts <- sf::st_as_sf(occ, coords = c("lon", "lat"), crs = 4326)
# pts <- sf::st_transform(pts, target)
# xy  <- sf::st_coordinates(pts)
# 
# env <- tiff_extract_points("predictors.tif", x = xy[, 1], y = xy[, 2])

## ----raster-------------------------------------------------------------------
nx <- 80; ny <- 50
xmin <- 5; xmax <- 17; ymin <- 45; ymax <- 55
xres <- (xmax - xmin) / nx
yres <- (ymax - ymin) / ny

g <- expand.grid(col = 0:(nx - 1), row = 0:(ny - 1))
g$x <- xmin + (g$col + 0.5) * xres          # pixel centres
g$y <- ymax - (g$row + 0.5) * yres          # row 1 is the northern edge
g$band1 <- 14 - 0.6 * (g$y - 50) + 0.2 * (g$x - 11) + rnorm(nrow(g), 0, 0.3)
g$band2 <- 750 + 25 * (g$y - 50) - 8 * (g$x - 11) + rnorm(nrow(g), 0, 10)

tif <- tempfile(fileext = ".tif")
write_tiff(g[, c("x", "y", "band1", "band2")], tif, crs = 4326L)

tiff_crs(tif)

## ----extract------------------------------------------------------------------
N <- 1500
sites <- data.frame(
  x = runif(N, xmin + 0.3, xmax - 0.3),
  y = runif(N, ymin + 0.3, ymax - 0.3)
)

env <- tiff_extract_points(tif, sites)
head(env)

## ----outcome------------------------------------------------------------------
occ <- data.frame(
  bio1  = as.numeric(scale(env$band1)),   # standardized once, then stored
  bio12 = as.numeric(scale(env$band2))
)
eta <- -0.2 + 1.3 * occ$bio1 - 0.9 * occ$bio12
occ$presence <- rbinom(N, 1, plogis(eta))
table(occ$presence)

## ----glm----------------------------------------------------------------------
fit <- glm(presence ~ bio1 + bio12, data = occ, family = binomial())
round(coef(fit), 3)

## ----write-vtr----------------------------------------------------------------
vtr <- tempfile(fileext = ".vtr")
write_vtr(occ, vtr)

## ----collect-chunked----------------------------------------------------------
acc <- collect_chunked(
  tbl(vtr),
  function(acc, chunk) {
    p <- chunk$presence == 1
    acc$n      <- acc$n      + nrow(chunk)
    acc$n_pres <- acc$n_pres + sum(p)
    acc$sum1   <- acc$sum1   + tapply(chunk$bio1, p, sum)[c("FALSE", "TRUE")]
    acc
  },
  .init = list(n = 0L, n_pres = 0L, sum1 = c("FALSE" = 0, "TRUE" = 0))
)
acc$n_pres / acc$n                      # prevalence

## ----bigglm, eval = has_biglm-------------------------------------------------
src <- function() tbl(vtr) |> select(presence, bio1, bio12)

fit_stream <- biglm::bigglm(
  presence ~ bio1 + bio12,
  data = chunk_feeder(src),
  family = binomial()
)
round(coef(fit_stream), 3)

## ----bigglm-note, echo = FALSE, results = "asis", eval = !has_biglm-----------
# cat("> The out-of-core GLM example needs the **biglm** package, which is not",
#     "installed, so it was not run here. Install it with",
#     "`install.packages(\"biglm\")` to reproduce the streamed fit.\n")

## ----cleanup, include = FALSE-------------------------------------------------
unlink(c(tif, vtr))

## ----feature-knn--------------------------------------------------------------
calib <- occ[, c("bio1", "bio12")]          # the standardized training cloud

# Two sites inside the training conditions and one well outside them.
proj <- data.frame(
  bio1  = c(0.0, 0.5, 4.0),
  bio12 = c(0.0, -0.5, -4.0)
)
pv <- tempfile(fileext = ".vtr")
write_vtr(proj, pv)

tbl(pv) |>
  feature_knn(calib, percentage = 5) |>
  collect()
unlink(pv)

# -- internal helpers ---------------------------------------------------------

#' Convert a data.frame to a temporary vectra_node
#'
#' Writes `df` to a temporary `.vtr` file, opens it as a lazy node, and
#' registers cleanup of the temp file on the *caller's* frame so the file
#' lives until the caller returns.
#'
#' @param df A `data.frame`.
#' @param envir Environment in which to register the `on.exit` cleanup
#'   (typically `parent.frame()`).
#' @return A `vectra_node`.
#' @noRd
df_to_node <- function(df, envir = parent.frame()) {

  tmp <- tempfile(fileext = ".vtr")
  do.call(on.exit, list(substitute(unlink(tmp)), add = TRUE), envir = envir)
  write_vtr(df, tmp)
  tbl(tmp)
}

check_scalar_string <- function(x, name = deparse(substitute(x))) {
  if (!is.character(x) || length(x) != 1)
    stop(sprintf("%s must be a single character string", name))
}

# -- write_csv ----------------------------------------------------------------

#' Write query results or a data.frame to a CSV file
#'
#' For `vectra_node` inputs, data is streamed batch-by-batch to disk without
#' materializing the full result in memory. For `data.frame` inputs, the data
#' is written directly.
#'
#' @param x A `vectra_node` (lazy query) or a `data.frame`.
#' @param path File path for the output CSV file.
#' @param ... Reserved for future use.
#'
#' @return Invisible `NULL`.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars[1:5, ], f)
#' csv <- tempfile(fileext = ".csv")
#' tbl(f) |> write_csv(csv)
#' unlink(c(f, csv))
#'
#' @export
write_csv <- function(x, path, ...) {
  UseMethod("write_csv")
}

#' @export
write_csv.vectra_node <- function(x, path, ...) {
  check_scalar_string(path)
  path <- normalizePath(path, mustWork = FALSE)
  .Call(C_write_csv, x$.node, path)
  invisible(NULL)
}

#' @export
write_csv.data.frame <- function(x, path, ...) {
  write_csv.vectra_node(df_to_node(x), path, ...)
}

#' Write query results or a data.frame to a SQLite table
#'
#' For `vectra_node` inputs, data is streamed batch-by-batch to disk without
#' materializing the full result in memory. For `data.frame` inputs, the data
#' is written directly.
#'
#' @param x A `vectra_node` (lazy query) or a `data.frame`.
#' @param path File path for the SQLite database.
#' @param table Name of the table to create/write into.
#' @param ... Reserved for future use.
#'
#' @return Invisible `NULL`.
#'
#' @examples
#' db <- tempfile(fileext = ".sqlite")
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars[1:5, ], f)
#' tbl(f) |> write_sqlite(db, "cars")
#' unlink(c(f, db))
#'
#' @export
write_sqlite <- function(x, path, table, ...) {
  UseMethod("write_sqlite")
}

#' @export
write_sqlite.vectra_node <- function(x, path, table, ...) {
  check_scalar_string(path)
  check_scalar_string(table)
  path <- normalizePath(path, mustWork = FALSE)
  .Call(C_write_sqlite, x$.node, path, table)
  invisible(NULL)
}

#' @export
write_sqlite.data.frame <- function(x, path, table, ...) {
  write_sqlite.vectra_node(df_to_node(x), path, table, ...)
}

#' Write query results to a GeoTIFF file
#'
#' The data must contain `x` and `y` columns (pixel center coordinates) and
#' one or more numeric band columns. Grid dimensions and geotransform are
#' inferred from the x/y coordinate arrays. Missing pixels are written as NaN
#' (or the type-appropriate nodata value for integer pixel types).
#'
#' @param x A `vectra_node` (lazy query) or a `data.frame`.
#' @param path File path for the output GeoTIFF file.
#' @param compress Logical; use DEFLATE compression? Default `FALSE`.
#' @param pixel_type Character string specifying the output pixel type.
#'   One of `"float64"` (default), `"float32"`, `"int16"`, `"int32"`,
#'   `"uint8"`, or `"uint16"`.
#' @param metadata Optional character string of GDAL_METADATA XML to embed
#'   in the file (tag 42112). Use [tiff_metadata()] to read it back.
#' @param crs Optional CRS to embed as a GeoKey directory (TIFF tag 34735).
#'   Accepts an integer EPSG code, an `"EPSG:xxxx"` string, or a list with
#'   named fields `epsg`, `geographic` (`TRUE`/`FALSE`), and optionally
#'   `citation`. Codes that are not auto-classified as projected/geographic
#'   default to projected; pass `geographic = TRUE` to override.
#'   Use [tiff_crs()] to read it back.
#' @param tiled Logical; write a tiled GeoTIFF (TIFF tags 322/323/324/325)
#'   instead of strips. Default `FALSE`. Tiled layout enables random-access
#'   block reads and is required for Cloud-Optimized GeoTIFF (COG).
#' @param tile_size Integer; tile edge length in pixels. Must be a positive
#'   multiple of 16 (TIFF spec). Either a single value (square tiles) or a
#'   length-2 vector `c(width, height)`. Default `256`. Edge tiles at the
#'   right and bottom of the image are padded to full tile size with the
#'   NoData / NaN value.
#' @param bigtiff Controls BigTIFF dispatch. `"auto"` (default) emits BigTIFF
#'   when the expected raw payload would exceed the classic-TIFF 4 GB
#'   ceiling, otherwise emits classic TIFF. `TRUE` forces BigTIFF (magic
#'   `0x002B`, 64-bit offsets), useful for round-trip tests on small data.
#'   `FALSE` forces classic TIFF — beware that classic TIFF will silently
#'   corrupt outputs larger than 4 GB. Tiled BigTIFF is not yet supported.
#' @param ... Reserved for future use.
#'
#' @return Invisible `NULL`.
#'
#' @examples
#' \donttest{
#' # Write as int16 with DEFLATE compression and an EPSG:4326 GeoKey
#' df <- data.frame(x = 1:4, y = rep(1:2, each = 2), band1 = c(100, 200, 300, 400))
#' f <- tempfile(fileext = ".tif")
#' write_tiff(df, f, compress = TRUE, pixel_type = "int16", crs = 4326L)
#' tiff_crs(f)
#' unlink(f)
#' }
#'
#' @export
write_tiff <- function(x, path, compress = FALSE, pixel_type = "float64",
                       metadata = NULL, crs = NULL,
                       tiled = FALSE, tile_size = 256L,
                       bigtiff = "auto", ...) {
  UseMethod("write_tiff")
}

.pixel_type_code <- c(
  float64 = 0L, float32 = 1L, int16 = 2L, int32 = 3L,
  uint8 = 4L, uint16 = 5L
)

# Mirrors TIFF_BIGTIFF_AUTO / TIFF_BIGTIFF_OFF / TIFF_BIGTIFF_FORCE in
# src/tiff_format.h. Keep these in lockstep — they're shipped through
# C_write_tiff_typed as plain integers.
.BIGTIFF_AUTO  <- 0L
.BIGTIFF_OFF   <- 1L
.BIGTIFF_FORCE <- 2L

.parse_bigtiff <- function(bigtiff) {
  if (is.null(bigtiff)) return(.BIGTIFF_AUTO)
  if (is.character(bigtiff) && length(bigtiff) == 1) {
    if (identical(bigtiff, "auto")) return(.BIGTIFF_AUTO)
    stop("bigtiff string must be \"auto\" (got '", bigtiff, "')")
  }
  if (is.logical(bigtiff) && length(bigtiff) == 1 && !is.na(bigtiff)) {
    return(if (bigtiff) .BIGTIFF_FORCE else .BIGTIFF_OFF)
  }
  stop("bigtiff must be \"auto\", TRUE, or FALSE")
}

# Translate a user-supplied `crs` argument into (epsg_geographic,
# epsg_projected, citation). The C side requires us to commit each EPSG to
# exactly one slot. We classify with a small table covering the common
# geographic codes (4326 WGS84, 4269 NAD83, the EPSG:42xx historical block
# and EPSG:4xxx geographic block); everything else is treated as projected,
# matching the behavior most users want for UTM/Web Mercator/national grids.
.crs_is_geographic_code <- function(epsg) {
  if (is.na(epsg)) return(FALSE)
  # EPSG geographic 2D CRS block is roughly 4001..4999, with a few outliers
  # in 4xxxx that we don't try to enumerate.
  epsg >= 4001L && epsg <= 4999L
}

.parse_crs <- function(crs) {
  if (is.null(crs)) return(list(epsg_g = 0L, epsg_p = 0L, cit = NULL))

  cit <- NULL
  geographic <- NULL
  epsg <- NA_integer_

  if (is.list(crs)) {
    if (!is.null(crs$epsg))      epsg <- as.integer(crs$epsg)
    if (!is.null(crs$citation))  cit  <- as.character(crs$citation)
    if (!is.null(crs$geographic)) geographic <- isTRUE(crs$geographic)
  } else if (is.character(crs) && length(crs) == 1) {
    m <- regmatches(crs, regexec("^EPSG:(\\d+)$", crs, ignore.case = TRUE))[[1]]
    if (length(m) == 2) epsg <- as.integer(m[2])
    else stop("crs string must look like 'EPSG:xxxx', got '", crs, "'")
  } else if (is.numeric(crs) && length(crs) == 1) {
    epsg <- as.integer(crs)
  } else {
    stop("crs must be NULL, an integer EPSG, an 'EPSG:xxxx' string, ",
         "or a list with $epsg")
  }

  if (is.na(epsg) || epsg <= 0)
    stop("crs needs a positive integer EPSG code")

  if (is.null(geographic))
    geographic <- .crs_is_geographic_code(epsg)

  if (geographic)
    list(epsg_g = epsg, epsg_p = 0L, cit = cit)
  else
    list(epsg_g = 0L, epsg_p = epsg, cit = cit)
}

# Translate (tiled, tile_size) into integer (tile_w, tile_h) for the C side.
# tile_size accepts a single integer (square tiles) or length-2 c(w, h).
.parse_tile_size <- function(tiled, tile_size) {
  if (!isTRUE(tiled)) return(list(w = 0L, h = 0L))

  if (is.null(tile_size))
    stop("tile_size must be supplied when tiled = TRUE")
  if (!is.numeric(tile_size) || any(is.na(tile_size)))
    stop("tile_size must be a positive integer")
  if (length(tile_size) == 1L) {
    w <- as.integer(tile_size); h <- w
  } else if (length(tile_size) == 2L) {
    w <- as.integer(tile_size[1]); h <- as.integer(tile_size[2])
  } else {
    stop("tile_size must have length 1 or 2")
  }
  if (w <= 0 || h <= 0)
    stop("tile_size must be positive")
  if (w %% 16L != 0L || h %% 16L != 0L)
    stop("tile_size must be a multiple of 16 (TIFF spec)")
  list(w = w, h = h)
}

#' @export
write_tiff.vectra_node <- function(x, path, compress = FALSE,
                                   pixel_type = "float64",
                                   metadata = NULL, crs = NULL,
                                   tiled = FALSE, tile_size = 256L,
                                   bigtiff = "auto", ...) {
  check_scalar_string(path)
  path <- normalizePath(path, mustWork = FALSE)

  pt <- .pixel_type_code[pixel_type]
  if (is.na(pt))
    stop(sprintf("pixel_type must be one of [%s], got '%s'",
                 paste(names(.pixel_type_code), collapse = ", "), pixel_type))

  if (!is.null(metadata) && (!is.character(metadata) || length(metadata) != 1))
    stop("metadata must be a single character string or NULL")

  cs <- .parse_crs(crs)
  ts <- .parse_tile_size(tiled, tile_size)
  bt <- .parse_bigtiff(bigtiff)

  # Always use the typed path — it handles all pixel types including float64
  .Call(C_write_tiff_typed, x$.node, path, as.logical(compress), pt, metadata,
        cs$epsg_g, cs$epsg_p, cs$cit, ts$w, ts$h, bt)
  invisible(NULL)
}

#' @export
write_tiff.data.frame <- function(x, path, compress = FALSE,
                                  pixel_type = "float64",
                                  metadata = NULL, crs = NULL,
                                  tiled = FALSE, tile_size = 256L,
                                  bigtiff = "auto", ...) {
  write_tiff.vectra_node(df_to_node(x), path, compress, pixel_type,
                         metadata, crs, tiled = tiled,
                         tile_size = tile_size, bigtiff = bigtiff, ...)
}

#' Write data to a .vtr file
#'
#' For `vectra_node` inputs (lazy queries from any format: CSV, SQLite, TIFF,
#' or another .vtr), data is streamed batch-by-batch to disk without
#' materializing the full result in memory. Each batch becomes one row group.
#' The output file is written atomically (via temp file + rename) so readers
#' never see a partial file.
#'
#' For `data.frame` inputs, the data is written directly from memory.
#'
#' @param x A `vectra_node` (lazy query) or a `data.frame`.
#' @param path File path for the output .vtr file.
#' @param compress Compression level: `"fast"` (default, byte-shuffle + greedy
#'   LZ), `"small"` (per-block adaptive — tries greedy LZ, separated-streams
#'   LZ, and LZ + Huffman entropy coding, and writes whichever shrank the
#'   block the most; never worse than `"fast"` on any block, typically
#'   10-25 percent smaller files at the cost of slower encode), or `"none"`.
#' @param batch_size Target number of rows per row group in the output file.
#'   Defaults to 131072 for data.frames (1 MB per double column, cache-friendly
#'   for decompression). For nodes, defaults to `NULL` (one row group per
#'   upstream batch).
#' @param col_types Optional named character vector specifying narrow integer
#'   storage types. Names must match column names; values must be `"int8"`,
#'   `"int16"`, or `"int32"`. Only applies to integer columns.
#'   Example: `col_types = c(age = "int8", year = "int16")`.
#' @param quantize Optional named list for lossy quantization of `double`
#'   columns. Each element is named after a column and is itself a named list
#'   with `scale` (or `precision = 1/scale`), `type` (`"int8"`, `"int16"`,
#'   `"int32"`; default `"int16"`), and optionally `offset` (default 0).
#'   Example: `quantize = list(temp = list(precision = 0.001, type = "int16"))`.
#' @param spatial Optional list for 2D spatial predictor encoding. Either a
#'   global spec applied to all numeric columns (`list(nx = 2000, ny = 2000)`)
#'   or per-column specs (`list(temp = list(nx = 2000, ny = 2000))`).
#'   When provided, a spatial predictor removes smooth 2D trends before
#'   compression, dramatically improving compression of raster data.
#'   Combines with `quantize` for maximum effect.
#' @param ... Additional arguments passed to methods.
#'
#' @return Invisible `NULL`.
#'
#' @examples
#' # From a data.frame
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#'
#' # Streaming format conversion (CSV -> VTR)
#' csv <- tempfile(fileext = ".csv")
#' write.csv(mtcars, csv, row.names = FALSE)
#' f2 <- tempfile(fileext = ".vtr")
#' tbl_csv(csv) |> write_vtr(f2)
#'
#' unlink(c(f, f2, csv))
#'
#' @export
write_vtr <- function(x, path, compress = c("fast", "small", "none"), batch_size = NULL,
                      col_types = NULL, quantize = NULL, spatial = NULL,
                      ...) {
  UseMethod("write_vtr")
}

.check_compress <- function(compress) {
  match.arg(compress, c("fast", "small", "none"))
}

#' @export
write_vtr.vectra_node <- function(x, path, compress = c("fast", "small", "none"),
                                  batch_size = NULL, col_types = NULL,
                                  quantize = NULL, spatial = NULL, ...) {
  check_scalar_string(path)
  path <- normalizePath(path, mustWork = FALSE)
  compress <- .check_compress(compress)
  bs <- if (!is.null(batch_size)) as.double(batch_size) else NULL
  .Call(C_write_vtr_node, x$.node, path, bs, compress, col_types, quantize,
        spatial)
  invisible(NULL)
}

#' @export
write_vtr.data.frame <- function(x, path, compress = c("fast", "small", "none"),
                                 batch_size = 131072L, col_types = NULL,
                                 quantize = NULL, spatial = NULL, ...) {
  check_scalar_string(path)
  path <- normalizePath(path, mustWork = FALSE)
  compress <- .check_compress(compress)
  .Call(C_write_vtr, x, path, as.integer(batch_size), compress, col_types,
        quantize, spatial)
  invisible(NULL)
}

#' Append rows to an existing .vtr file
#'
#' Appends one or more new row groups to the end of an existing `.vtr` file
#' without touching or recompressing existing row groups. The schema of `x`
#' must exactly match the schema of the target file (same column names and
#' types, in the same order).
#'
#' The operation is not fully atomic: if the process is interrupted after
#' new row groups are written but before the header is patched, the file
#' will be in a corrupted state. Use `write_vtr()` for safety-critical
#' write-once workloads.
#'
#' @param x A `vectra_node` (lazy query) or a `data.frame`.
#' @param path File path of an existing `.vtr` file to append to.
#' @param ... Additional arguments passed to methods.
#'
#' @return Invisible `NULL`.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars[1:10, ], f)
#' append_vtr(mtcars[11:20, ], f)
#' result <- tbl(f) |> collect()
#' stopifnot(nrow(result) == 20L)
#' unlink(f)
#'
#' @export
append_vtr <- function(x, path, ...) {
  UseMethod("append_vtr")
}

#' @export
append_vtr.vectra_node <- function(x, path, ...) {
  check_scalar_string(path)
  path <- normalizePath(path, mustWork = TRUE)
  # Force GC so any lingering tbl()-owned file handles on `path` are closed
  # before we stream-rewrite the file. On Windows the rename at the end of
  # C_append_vtr otherwise fails with a sharing violation.
  gc(verbose = FALSE)
  .Call(C_append_vtr, x$.node, path)
  invisible(NULL)
}

#' @export
append_vtr.data.frame <- function(x, path, ...) {
  append_vtr.vectra_node(df_to_node(x), path, ...)
}

# --- write_csv tests ---

test_that("write_csv from node round-trips all types", {
  df <- data.frame(a = c(1L, 2L, 3L), b = c(1.5, 2.5, 3.5),
                   c = c(TRUE, FALSE, TRUE),
                   d = c("x", "y", "z"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  csv <- tempfile(fileext = ".csv")
  on.exit(unlink(c(f, csv)))
  write_vtr(df, f)
  tbl(f) |> write_csv(csv)
  result <- read.csv(csv, stringsAsFactors = FALSE)
  expect_equal(result$a, c(1, 2, 3))
  expect_equal(result$b, c(1.5, 2.5, 3.5))
  expect_equal(result$c, c(TRUE, FALSE, TRUE))
  expect_equal(result$d, c("x", "y", "z"))
})

test_that("write_csv handles NAs", {
  df <- data.frame(a = c(1L, NA, 3L), b = c(NA, 2.5, NA),
                   c = c(TRUE, NA, FALSE),
                   d = c(NA, "y", NA), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  csv <- tempfile(fileext = ".csv")
  on.exit(unlink(c(f, csv)))
  write_vtr(df, f)
  tbl(f) |> write_csv(csv)
  result <- read.csv(csv, stringsAsFactors = FALSE)
  expect_true(is.na(result$a[2]))
  expect_true(is.na(result$b[1]))
  expect_true(is.na(result$c[2]))
  expect_true(is.na(result$d[1]))
  expect_equal(result$a[1], 1)
  expect_equal(result$d[2], "y")
})

test_that("write_csv quotes strings with commas", {
  df <- data.frame(x = c("a,b", "c\"d", "normal"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  csv <- tempfile(fileext = ".csv")
  on.exit(unlink(c(f, csv)))
  write_vtr(df, f)
  tbl(f) |> write_csv(csv)
  result <- read.csv(csv, stringsAsFactors = FALSE)
  expect_equal(result$x, c("a,b", "c\"d", "normal"))
})

test_that("write_csv quotes strings with newlines", {
  df <- data.frame(x = c("line1\nline2", "ok"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  csv <- tempfile(fileext = ".csv")
  on.exit(unlink(c(f, csv)))
  write_vtr(df, f)
  tbl(f) |> write_csv(csv)
  result <- read.csv(csv, stringsAsFactors = FALSE)
  expect_equal(result$x, c("line1\nline2", "ok"))
})

test_that("write_csv from data.frame works", {
  df <- data.frame(a = c(1.0, 2.0), b = c("x", "y"), stringsAsFactors = FALSE)
  csv <- tempfile(fileext = ".csv")
  on.exit(unlink(csv))
  write_csv(df, csv)
  result <- read.csv(csv, stringsAsFactors = FALSE)
  expect_equal(result$a, c(1, 2))
  expect_equal(result$b, c("x", "y"))
})

test_that("write_csv streams after filter", {
  df <- data.frame(x = 1:10, y = letters[1:10], stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  csv <- tempfile(fileext = ".csv")
  on.exit(unlink(c(f, csv)))
  write_vtr(df, f)
  tbl(f) |> filter(x > 7) |> write_csv(csv)
  result <- read.csv(csv, stringsAsFactors = FALSE)
  expect_equal(nrow(result), 3)
  expect_equal(result$y, c("h", "i", "j"))
})

test_that("write_csv streams after mutate", {
  df <- data.frame(x = c(1.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  csv <- tempfile(fileext = ".csv")
  on.exit(unlink(c(f, csv)))
  write_vtr(df, f)
  tbl(f) |> mutate(y = x * 10) |> write_csv(csv)
  result <- read.csv(csv, stringsAsFactors = FALSE)
  expect_equal(result$y, c(10, 20, 30))
})

test_that("write_csv streams after summarise", {
  df <- data.frame(g = c("a", "a", "b"), v = c(1.0, 2.0, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  csv <- tempfile(fileext = ".csv")
  on.exit(unlink(c(f, csv)))
  write_vtr(df, f)
  tbl(f) |> group_by(g) |> summarise(s = sum(v)) |> write_csv(csv)
  result <- read.csv(csv, stringsAsFactors = FALSE)
  result <- result[order(result$g), ]
  expect_equal(result$g, c("a", "b"))
  expect_equal(result$s, c(3, 3))
})

test_that("write_csv empty result", {
  df <- data.frame(x = c(1.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  csv <- tempfile(fileext = ".csv")
  on.exit(unlink(c(f, csv)))
  write_vtr(df, f)
  tbl(f) |> filter(x > 100) |> write_csv(csv)
  lines <- readLines(csv)
  expect_equal(length(lines), 1)  # header only
  expect_equal(lines[1], "x")
})

test_that("write_csv column names with special chars are quoted", {
  df <- data.frame(1.0, 2.0)
  names(df) <- c("a,b", "c d")
  f <- tempfile(fileext = ".vtr")
  csv <- tempfile(fileext = ".csv")
  on.exit(unlink(c(f, csv)))
  write_vtr(df, f)
  tbl(f) |> write_csv(csv)
  result <- read.csv(csv, stringsAsFactors = FALSE, check.names = FALSE)
  expect_true("a,b" %in% names(result))
  expect_true("c d" %in% names(result))
})

# --- lag ---

test_that("lag shifts values down by 1", {
  df <- data.frame(x = c(10.0, 20.0, 30.0, 40.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(prev = lag(x)) |> collect()
  expect_true(is.na(result$prev[1]))
  expect_equal(result$prev[2:4], c(10, 20, 30))
})

test_that("lag with n = 2", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 4.0, 5.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(prev2 = lag(x, 2)) |> collect()
  expect_true(is.na(result$prev2[1]))
  expect_true(is.na(result$prev2[2]))
  expect_equal(result$prev2[3:5], c(1, 2, 3))
})

test_that("lag with default value", {
  df <- data.frame(x = c(10.0, 20.0, 30.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(prev = lag(x, 1, default = 0)) |> collect()
  expect_equal(result$prev, c(0, 10, 20))
})

# --- lead ---

test_that("lead shifts values up by 1", {
  df <- data.frame(x = c(10.0, 20.0, 30.0, 40.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(nxt = lead(x)) |> collect()
  expect_equal(result$nxt[1:3], c(20, 30, 40))
  expect_true(is.na(result$nxt[4]))
})

test_that("lead with default value", {
  df <- data.frame(x = c(10.0, 20.0, 30.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(nxt = lead(x, 1, default = 99)) |> collect()
  expect_equal(result$nxt, c(20, 30, 99))
})

# --- row_number ---

test_that("row_number assigns sequential integers", {
  df <- data.frame(x = c(5.0, 3.0, 1.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(rn = row_number()) |> collect()
  expect_equal(result$rn, c(1, 2, 3))
})

# --- cumsum ---

test_that("cumsum computes cumulative sum", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(cs = cumsum(x)) |> collect()
  expect_equal(result$cs, c(1, 3, 6, 10))
})

test_that("cumsum propagates NA", {
  df <- data.frame(x = c(1.0, NA, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(cs = cumsum(x)) |> collect()
  expect_equal(result$cs[1], 1)
  expect_true(is.na(result$cs[2]))
  expect_true(is.na(result$cs[3]))
})

# --- cummean ---

test_that("cummean computes cumulative mean", {
  df <- data.frame(x = c(2.0, 4.0, 6.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(cm = cummean(x)) |> collect()
  expect_equal(result$cm, c(2, 3, 4))
})

# --- cummin / cummax ---

test_that("cummin computes running minimum", {
  df <- data.frame(x = c(5.0, 3.0, 4.0, 1.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(cmi = cummin(x)) |> collect()
  expect_equal(result$cmi, c(5, 3, 3, 1))
})

test_that("cummax computes running maximum", {
  df <- data.frame(x = c(1.0, 4.0, 2.0, 5.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(cmx = cummax(x)) |> collect()
  expect_equal(result$cmx, c(1, 4, 4, 5))
})

# --- grouped windows ---

test_that("lag works within groups", {
  df <- data.frame(
    g = c("a", "a", "a", "b", "b"),
    x = c(1.0, 2.0, 3.0, 10.0, 20.0),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> mutate(prev = lag(x)) |> collect()
  # Group a: NA, 1, 2; Group b: NA, 10
  expect_true(is.na(result$prev[1]))
  expect_equal(result$prev[2], 1)
  expect_equal(result$prev[3], 2)
  expect_true(is.na(result$prev[4]))
  expect_equal(result$prev[5], 10)
})

test_that("row_number resets within groups", {
  df <- data.frame(
    g = c("a", "a", "b", "b", "b"),
    x = c(1.0, 2.0, 3.0, 4.0, 5.0),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> mutate(rn = row_number()) |> collect()
  expect_equal(result$rn, c(1, 2, 1, 2, 3))
})

test_that("cumsum resets within groups", {
  df <- data.frame(
    g = c("a", "a", "b", "b"),
    x = c(1.0, 2.0, 10.0, 20.0),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> mutate(cs = cumsum(x)) |> collect()
  expect_equal(result$cs, c(1, 3, 10, 30))
})

# --- mixed window + regular mutate ---

test_that("window and regular mutate in same call", {
  df <- data.frame(x = c(1.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(cs = cumsum(x), doubled = x * 2) |> collect()
  expect_equal(result$cs, c(1, 3, 6))
  expect_equal(result$doubled, c(2, 4, 6))
})

# --- rank ---

test_that("rank assigns min rank with gaps for ties", {
  df <- data.frame(x = c(3.0, 1.0, 3.0, 2.0, 1.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(r = rank(x)) |> collect()
  expect_equal(result$r, c(4, 1, 4, 3, 1))
})

test_that("rank works with groups", {
  df <- data.frame(g = c("a", "a", "a", "b", "b"),
                   x = c(3.0, 1.0, 3.0, 2.0, 1.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> mutate(r = rank(x)) |> collect()
  expect_equal(result$r[result$g == "a"], c(2, 1, 2))
  expect_equal(result$r[result$g == "b"], c(2, 1))
})

# --- dense_rank ---

test_that("dense_rank assigns consecutive ranks without gaps", {
  df <- data.frame(x = c(3.0, 1.0, 3.0, 2.0, 1.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(dr = dense_rank(x)) |> collect()
  expect_equal(result$dr, c(3, 1, 3, 2, 1))
})

test_that("dense_rank works with groups", {
  df <- data.frame(g = c("a", "a", "a", "b", "b"),
                   x = c(3.0, 1.0, 3.0, 2.0, 1.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> mutate(dr = dense_rank(x)) |> collect()
  expect_equal(result$dr[result$g == "a"], c(2, 1, 2))
  expect_equal(result$dr[result$g == "b"], c(2, 1))
})

test_that("binary search narrows range on sorted int column", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  # Write sorted data with small batch size to create multiple row groups
  df <- data.frame(id = 1:1000, val = runif(1000))
  write_vtr(df, f, batch_size = 100L)

  # Point lookup should find the correct row
  result <- tbl(f) |> filter(id == 500) |> collect()
  expect_equal(nrow(result), 1L)
  expect_equal(result$id, 500)

  # Range lookup
  result <- tbl(f) |> filter(id >= 900) |> collect()
  expect_equal(nrow(result), 101L)
  expect_true(all(result$id >= 900))

  result <- tbl(f) |> filter(id <= 50) |> collect()
  expect_equal(nrow(result), 50L)
  expect_true(all(result$id <= 50))

  # Combined range (AND)
  result <- tbl(f) |> filter(id >= 200, id <= 300) |> collect()
  expect_equal(nrow(result), 101L)
  expect_true(all(result$id >= 200 & result$id <= 300))
})

test_that("binary search works on sorted string column", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  # Create sorted string data
  df <- data.frame(
    name = sort(paste0("name_", sprintf("%04d", 1:500))),
    val = seq_len(500),
    stringsAsFactors = FALSE
  )
  write_vtr(df, f, batch_size = 50L)

  result <- tbl(f) |> filter(name == "name_0250") |> collect()
  expect_equal(nrow(result), 1L)
  expect_equal(result$name, "name_0250")
})

test_that("binary search handles unsorted data correctly", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  # Write unsorted data — binary search should not engage
  df <- data.frame(id = sample(1:500), val = runif(500))
  write_vtr(df, f, batch_size = 50L)

  result <- tbl(f) |> filter(id == 100) |> collect()
  expect_equal(nrow(result), 1L)
  expect_equal(result$id, 100)
})

test_that("binary search handles single row group", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  df <- data.frame(id = 1:100, val = runif(100))
  write_vtr(df, f)

  result <- tbl(f) |> filter(id == 50) |> collect()
  expect_equal(nrow(result), 1L)
  expect_equal(result$id, 50)
})

test_that("binary search handles empty result", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  df <- data.frame(id = 1:1000, val = runif(1000))
  write_vtr(df, f, batch_size = 100L)

  result <- tbl(f) |> filter(id == 9999) |> collect()
  expect_equal(nrow(result), 0L)
})

test_that("binary search on double column", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  df <- data.frame(x = seq(0, 10, by = 0.01))
  write_vtr(df, f, batch_size = 100L)

  result <- tbl(f) |> filter(x == 5.0) |> collect()
  expect_equal(nrow(result), 1L)
  expect_equal(result$x, 5.0)

  result <- tbl(f) |> filter(x > 9.5) |> collect()
  expect_true(nrow(result) > 0)
  expect_true(all(result$x > 9.5))
})

test_that("round-trip preserves double columns", {
  df <- data.frame(x = c(1.5, 2.7, NA, 4.0), y = c(NA, 10.0, 20.0, 30.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  expect_equal(result$x, df$x)
  expect_equal(result$y, df$y)
})

test_that("round-trip preserves integer columns as double", {
  df <- data.frame(a = c(1L, 2L, NA, 4L), b = c(10L, NA, 30L, 40L))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  # int64 stored internally, returned as double by default
  expect_equal(result$a, c(1, 2, NA, 4))
  expect_equal(result$b, c(10, NA, 30, 40))
})

test_that("round-trip preserves logical columns", {
  df <- data.frame(flag = c(TRUE, FALSE, NA, TRUE))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  expect_equal(result$flag, df$flag)
})

test_that("round-trip preserves string columns", {
  df <- data.frame(s = c("hello", "world", NA, ""), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  expect_equal(result$s, df$s)
})

test_that("round-trip with all types together", {
  df <- data.frame(
    i = c(1L, 2L, NA),
    d = c(1.1, NA, 3.3),
    b = c(TRUE, NA, FALSE),
    s = c("a", NA, "c"),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  expect_equal(result$d, df$d)
  expect_equal(result$b, df$b)
  expect_equal(result$s, df$s)
  expect_equal(result$i, as.double(df$i))
})

test_that("multiple row groups round-trip correctly", {
  df <- data.frame(x = 1:100, y = as.double(101:200))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, batch_size = 30)
  result <- tbl(f) |> collect()
  expect_equal(result$x, as.double(df$x))
  expect_equal(result$y, df$y)
})

test_that("empty-string column round-trips", {
  df <- data.frame(s = c("", "", ""), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  expect_equal(result$s, df$s)
})

test_that("single-row data.frame round-trips", {
  df <- data.frame(x = 42L, y = "test", stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  expect_equal(result$x, 42)
  expect_equal(result$y, "test")
})

test_that("print method works", {
  df <- data.frame(x = 1:3, y = c(1.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  node <- tbl(f)
  expect_output(print(node), "vectra query node")
  expect_output(print(node), "int64")
  expect_output(print(node), "double")
})

test_that("tiff_extract_points extracts correct values at pixel centers", {
  nx <- 3; ny <- 2
  df <- expand.grid(col = 0:(nx - 1), row = 0:(ny - 1))
  df$x <- 0.5 + df$col
  df$y <- (ny - 0.5) - df$row
  df$band1 <- c(10, 20, 30, 40, 50, 60)
  df <- df[, c("x", "y", "band1")]

  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))
  write_tiff(df, tmp)

  # Query exact pixel centers
  pts <- data.frame(x = c(0.5, 1.5, 2.5), y = c(1.5, 1.5, 0.5))
  result <- tiff_extract_points(tmp, pts)

  expect_equal(nrow(result), 3)
  expect_equal(names(result), c("x", "y", "band1"))
  expect_equal(result$band1, c(10, 20, 60))
})

test_that("tiff_extract_points returns NA for out-of-bounds points", {
  df <- data.frame(x = c(0.5, 1.5), y = c(0.5, 0.5), band1 = c(1, 2))
  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))
  write_tiff(df, tmp)

  result <- tiff_extract_points(tmp,
    x = c(0.5, 99.0, -1.0),
    y = c(0.5, 0.5,   0.5))

  expect_equal(nrow(result), 3)
  expect_equal(result$band1[1], 1)
  expect_true(is.na(result$band1[2]))
  expect_true(is.na(result$band1[3]))
})

test_that("tiff_extract_points works with multi-band rasters", {
  df <- data.frame(
    x = c(0.5, 1.5, 0.5, 1.5),
    y = c(1.5, 1.5, 0.5, 0.5),
    band1 = c(1, 2, 3, 4),
    band2 = c(10, 20, 30, 40)
  )
  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))
  write_tiff(df, tmp)

  result <- tiff_extract_points(tmp, x = c(0.5, 1.5), y = c(1.5, 0.5))

  expect_equal(names(result), c("x", "y", "band1", "band2"))
  expect_equal(result$band1, c(1, 4))
  expect_equal(result$band2, c(10, 40))
})

test_that("tiff_extract_points accepts separate x, y vectors", {
  df <- data.frame(x = c(0.5, 1.5), y = c(0.5, 0.5), band1 = c(5, 6))
  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))
  write_tiff(df, tmp)

  result <- tiff_extract_points(tmp, x = 0.5, y = 0.5)
  expect_equal(nrow(result), 1)
  expect_equal(result$band1, 5)
})

test_that("tiff_extract_points validates inputs", {
  expect_error(tiff_extract_points(123, x = 1, y = 1),
               "path must be a single character string")

  tmp <- tempfile(fileext = ".tif")
  df <- data.frame(x = 0.5, y = 0.5, band1 = 1)
  write_tiff(df, tmp)
  on.exit(unlink(tmp))

  expect_error(tiff_extract_points(tmp, x = 1),
               "y is required")

  expect_error(tiff_extract_points(tmp, x = c(1, 2), y = 1),
               "same length")

  expect_error(tiff_extract_points(tmp, data.frame(a = 1, b = 2)),
               "must have columns")
})

test_that("tiff_extract_points with near-center offsets snaps to correct pixel", {
  df <- data.frame(
    x = c(0.5, 1.5, 2.5),
    y = c(0.5, 0.5, 0.5),
    band1 = c(100, 200, 300)
  )
  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))
  write_tiff(df, tmp)

  # Slightly off-center — should still snap to same pixel
  result <- tiff_extract_points(tmp, x = c(0.6, 1.4, 2.5), y = c(0.5, 0.5, 0.5))
  expect_equal(result$band1, c(100, 200, 300))
})

# ifelse() across branch types: the result column type must match the value the
# evaluator produces (common type of the two branches), or int64/double bit
# patterns get reinterpreted. Regression for ifelse(int64_col, <double/NA>).

make_f <- function(df) { f <- tempfile(fileext = ".vtr"); write_vtr(df, f); f }

test_that("ifelse(int64, NA) keeps the int64 values and yields real NAs", {
  f <- make_f(data.frame(x = c(2000L, 0L, NA, 1985L))); on.exit(unlink(f))
  # NA-condition rows are NA (R semantics); the kept values are intact, not the
  # ~4.6e18 garbage from an int64/double type mismatch.
  r <- tbl(f) |> mutate(y = ifelse(x > 0, x, NA_integer_)) |> collect()
  expect_equal(r$y, c(2000, NA, NA, 1985))
  expect_warning(
    tbl(f) |> mutate(y = ifelse(x > 0, x, NA_integer_)) |> collect(),
    regexp = NA)   # no "int64 exceeds 2^53" warning
})

test_that("ifelse(int64, double) promotes to double without corruption", {
  f <- make_f(data.frame(x = c(2000L, 0L, NA, 1985L))); on.exit(unlink(f))
  r <- tbl(f) |> mutate(y = ifelse(x > 0, x, 1.5)) |> collect()
  expect_equal(r$y, c(2000, 1.5, NA, 1985))
})

test_that("ifelse(int64, int) stays int64-valued and exact", {
  f <- make_f(data.frame(x = c(2000L, 0L, NA, 1985L))); on.exit(unlink(f))
  r <- tbl(f) |> mutate(y = ifelse(x > 0, x, 99999L)) |> collect()
  expect_equal(r$y, c(2000, 99999, NA, 1985))
})

test_that("ifelse with string branches is unaffected", {
  f <- make_f(data.frame(x = c(2000L, 0L, NA, 1985L))); on.exit(unlink(f))
  r <- tbl(f) |> mutate(y = ifelse(x > 0, "yes", "no")) |> collect()
  expect_equal(r$y, c("yes", "no", NA, "yes"))
})

test_that("earliest-known-year pattern: unknown maps to NA and sorts last", {
  # the demo idiom now that ifelse(int64, NA) is correct
  df <- data.frame(piece_id = c(1L, 1L, 2L, 2L, 3L, 3L),
                   STATUS_YR = c(2000L, 1990L, 0L, 0L, NA, 1985L))
  f <- make_f(df); on.exit(unlink(f))
  res <- tbl(f) |>
    mutate(year = ifelse(STATUS_YR > 0, STATUS_YR, NA_integer_)) |>
    group_by(piece_id) |>
    slice_min(year, n = 1, with_ties = FALSE) |>
    collect()
  res <- res[order(res$piece_id), ]
  expect_equal(res$STATUS_YR, c(1990, 0, 1985))   # piece 2 stays unknown (0)
})

## Tests for append_vtr, delete_vtr, and diff_vtr

# ── append_vtr ────────────────────────────────────────────────────────────────

test_that("append_vtr adds rows after existing rows", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  df1 <- data.frame(x = 1:5, y = letters[1:5], stringsAsFactors = FALSE)
  df2 <- data.frame(x = 6:10, y = letters[6:10], stringsAsFactors = FALSE)

  write_vtr(df1, f)
  append_vtr(df2, f)

  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 10L)
  expect_equal(result$x, as.double(1:10))
  expect_equal(result$y, letters[1:10])
})

test_that("append_vtr works with data.frame input", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  write_vtr(mtcars[1:10, ], f)
  append_vtr(mtcars[11:20, ], f)

  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 20L)
  expect_equal(result$mpg, mtcars$mpg[1:20])
})

test_that("append_vtr works with vectra_node input", {
  f  <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, f2)))

  df1 <- data.frame(v = as.double(1:4))
  df2 <- data.frame(v = as.double(5:8))

  write_vtr(df1, f)
  write_vtr(df2, f2)

  append_vtr(tbl(f2), f)

  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 8L)
  expect_equal(result$v, as.double(1:8))
})

test_that("append_vtr rejects schema mismatch", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  write_vtr(data.frame(a = 1:3), f)
  expect_error(append_vtr(data.frame(b = 4:6), f), "mismatch")
})

test_that("append_vtr preserves multiple row groups", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  df <- data.frame(x = as.double(1:100))
  write_vtr(df, f, batch_size = 30L)   # 4 row groups
  append_vtr(data.frame(x = as.double(101:110)), f)  # 1 more row group

  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 110L)
  expect_equal(result$x, as.double(1:110))
})

# ── delete_vtr ────────────────────────────────────────────────────────────────

test_that("delete_vtr removes specified rows", {
  f <- tempfile(fileext = ".vtr")
  del <- paste0(f, ".del")
  on.exit(unlink(c(f, del)))

  df <- data.frame(id = 1:5, stringsAsFactors = FALSE)
  write_vtr(df, f)

  # 0-based: index 0 = id 1, index 2 = id 3, index 4 = id 5
  delete_vtr(f, c(0, 2, 4))

  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 2L)
  expect_equal(result$id, c(2, 4))
})

test_that("delete_vtr tombstone file is created", {
  f <- tempfile(fileext = ".vtr")
  del <- paste0(f, ".del")
  on.exit(unlink(c(f, del)))

  write_vtr(data.frame(x = 1:3), f)
  delete_vtr(f, 1L)

  expect_true(file.exists(del))
})

test_that("delete_vtr is cumulative across calls", {
  f <- tempfile(fileext = ".vtr")
  del <- paste0(f, ".del")
  on.exit(unlink(c(f, del)))

  df <- data.frame(id = 1:6)
  write_vtr(df, f)

  delete_vtr(f, 0L)   # delete row 1
  delete_vtr(f, 2L)   # delete row 3

  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 4L)
  expect_equal(result$id, c(2, 4, 5, 6))
})

test_that("delete_vtr deduplicates tombstone entries", {
  f <- tempfile(fileext = ".vtr")
  del <- paste0(f, ".del")
  on.exit(unlink(c(f, del)))

  write_vtr(data.frame(x = 1:5), f)
  delete_vtr(f, c(1, 1, 1))   # same row multiple times

  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 4L)
})

test_that("delete_vtr works across multiple row groups", {
  f <- tempfile(fileext = ".vtr")
  del <- paste0(f, ".del")
  on.exit(unlink(c(f, del)))

  df <- data.frame(id = 1:9)
  write_vtr(df, f, batch_size = 3L)   # 3 row groups of 3 rows

  # Delete last row of group 0 (idx 2), first of group 1 (idx 3)
  delete_vtr(f, c(2, 3))

  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 7L)
  expect_equal(result$id, c(1, 2, 5, 6, 7, 8, 9))
})

test_that("delete all rows in a group skips it cleanly", {
  f <- tempfile(fileext = ".vtr")
  del <- paste0(f, ".del")
  on.exit(unlink(c(f, del)))

  df <- data.frame(id = 1:6)
  write_vtr(df, f, batch_size = 3L)   # 2 row groups: [0,1,2] and [3,4,5]

  delete_vtr(f, c(0, 1, 2))   # delete entire first row group

  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 3L)
  expect_equal(result$id, c(4, 5, 6))
})

test_that("tbl with no .del file works as before", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  df <- data.frame(x = 1:10)
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 10L)
})

# ── diff_vtr ──────────────────────────────────────────────────────────────────

test_that("diff_vtr detects added and deleted rows", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))

  df1 <- data.frame(id = 1:5, val = letters[1:5], stringsAsFactors = FALSE)
  df2 <- data.frame(id = c(3L, 4L, 5L, 6L, 7L),
                    val = c("c", "d", "e", "f", "g"),
                    stringsAsFactors = FALSE)
  write_vtr(df1, f1)
  write_vtr(df2, f2)

  d <- diff_vtr(f1, f2, "id")

  expect_true(is.list(d))
  expect_true(inherits(d$added, "vectra_node"))
  expect_true(is.vector(d$deleted))

  # IDs 1 and 2 were deleted
  expect_equal(sort(d$deleted), c(1, 2))
  # IDs 6 and 7 were added
  added_df <- collect(d$added)
  expect_equal(sort(added_df$id), c(6, 7))
  # All columns from B are present
  expect_true(all(c("id", "val") %in% names(added_df)))
})

test_that("diff_vtr returns empty added/deleted when files are identical", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))

  df <- data.frame(id = 1:5)
  write_vtr(df, f1)
  write_vtr(df, f2)

  d <- diff_vtr(f1, f2, "id")
  expect_equal(nrow(collect(d$added)), 0L)
  expect_equal(length(d$deleted), 0L)
})

test_that("diff_vtr errors on missing key column", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))

  write_vtr(data.frame(a = 1:3), f1)
  write_vtr(data.frame(a = 2:4), f2)

  expect_error(diff_vtr(f1, f2, "no_such_col"), "key_col")
})

test_that("diff_vtr handles all rows added (old is subset)", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))

  write_vtr(data.frame(id = 1:3), f1)
  write_vtr(data.frame(id = 1:6), f2)

  d <- diff_vtr(f1, f2, "id")
  expect_equal(nrow(collect(d$added)), 3L)
  expect_equal(length(d$deleted), 0L)
})

test_that("diff_vtr handles all rows deleted (new is empty subset)", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))

  write_vtr(data.frame(id = 1:5), f1)
  write_vtr(data.frame(id = integer(0)), f2)

  d <- diff_vtr(f1, f2, "id")
  expect_equal(length(d$deleted), 5L)
  expect_equal(nrow(collect(d$added)), 0L)
})

test_that("select picks specified columns", {
  df <- data.frame(a = 1:3, b = 4:6, c = 7:9)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(a, c) |> collect()
  expect_equal(names(result), c("a", "c"))
  expect_equal(result$a, c(1, 2, 3))
  expect_equal(result$c, c(7, 8, 9))
})

test_that("select preserves column order", {
  df <- data.frame(x = 1:3, y = 4:6, z = 7:9)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(z, x) |> collect()
  expect_equal(names(result), c("z", "x"))
})

# --- resolve() tests ---

test_that("resolve() looks up string values via integer FK", {
  df <- data.frame(
    id = c(1L, 2L, 3L, 4L),
    name = c("Alpha", "Beta", "Gamma", "Delta"),
    parent_id = c(NA, 1L, 1L, 2L),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |>
    mutate(parent_name = resolve(parent_id, id, name)) |>
    collect()

  expect_equal(result$parent_name, c(NA, "Alpha", "Alpha", "Beta"))
})

test_that("resolve() returns NA for missing FK values", {
  df <- data.frame(
    id = c(1L, 2L, 3L),
    name = c("A", "B", "C"),
    fk = c(2L, 99L, 1L),  # 99 doesn't exist
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |>
    mutate(resolved = resolve(fk, id, name)) |>
    collect()

  expect_equal(result$resolved, c("B", NA, "A"))
})

test_that("resolve() returns NA when FK is NA", {
  df <- data.frame(
    id = c(1L, 2L, 3L),
    value = c(10.0, 20.0, 30.0),
    fk = c(NA, 1L, 2L)
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |>
    mutate(resolved = resolve(fk, id, value)) |>
    collect()

  expect_equal(result$resolved, c(NA, 10.0, 20.0))
})

test_that("resolve() works with numeric columns", {
  df <- data.frame(
    id = c(1L, 2L, 3L),
    score = c(100.0, 200.0, 300.0),
    ref = c(3L, 1L, 2L)
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |>
    mutate(ref_score = resolve(ref, id, score)) |>
    collect()

  expect_equal(result$ref_score, c(300.0, 100.0, 200.0))
})

test_that("resolve() works with string FK and PK", {
  df <- data.frame(
    code = c("A", "B", "C"),
    label = c("Alpha", "Beta", "Gamma"),
    parent_code = c(NA, "A", "B"),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |>
    mutate(parent_label = resolve(parent_code, code, label)) |>
    collect()

  expect_equal(result$parent_label, c(NA, "Alpha", "Beta"))
})

# --- propagate() tests ---

test_that("propagate() fills children from parent values", {
  # Tree: 1(root, rank=FAMILY) -> 2(child) -> 3(grandchild)
  df <- data.frame(
    id = c(1L, 2L, 3L),
    parent_id = c(NA, 1L, 2L),
    rank = c("FAMILY", "GENUS", "SPECIES"),
    name = c("Fagaceae", "Quercus", "Quercus robur"),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |>
    mutate(family = propagate(
      parent_id, id,
      ifelse(rank == "FAMILY", name, NA)
    )) |>
    collect()

  expect_equal(result$family, c("Fagaceae", "Fagaceae", "Fagaceae"))
})

test_that("propagate() handles disconnected trees", {
  df <- data.frame(
    id = c(1L, 2L, 3L, 4L, 5L),
    parent_id = c(NA, 1L, 2L, NA, 4L),
    rank = c("FAMILY", "GENUS", "SPECIES", "FAMILY", "GENUS"),
    name = c("Fagaceae", "Quercus", "Q. robur", "Rosaceae", "Rosa"),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |>
    mutate(family = propagate(
      parent_id, id,
      ifelse(rank == "FAMILY", name, NA)
    )) |>
    collect()

  expect_equal(result$family, c("Fagaceae", "Fagaceae", "Fagaceae",
                                 "Rosaceae", "Rosaceae"))
})

test_that("propagate() returns NA for orphan nodes", {
  # Node 3 has parent 99 which doesn't exist
  df <- data.frame(
    id = c(1L, 2L, 3L),
    parent_id = c(NA, 1L, 99L),
    rank = c("FAMILY", "GENUS", "SPECIES"),
    name = c("Fagaceae", "Quercus", "Unknown"),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |>
    mutate(family = propagate(
      parent_id, id,
      ifelse(rank == "FAMILY", name, NA)
    )) |>
    collect()

  expect_equal(result$family, c("Fagaceae", "Fagaceae", NA))
})

test_that("propagate() works with numeric seed values", {
  df <- data.frame(
    id = c(1L, 2L, 3L, 4L),
    parent_id = c(NA, 1L, 2L, 1L),
    is_root = c(TRUE, FALSE, FALSE, FALSE),
    weight = c(42.0, NA, NA, NA)
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |>
    mutate(root_weight = propagate(
      parent_id, id,
      ifelse(is_root, weight, NA)
    )) |>
    collect()

  expect_equal(result$root_weight, c(42.0, 42.0, 42.0, 42.0))
})

test_that("propagate() converges for deep trees", {
  # Chain: 1 -> 2 -> 3 -> 4 -> 5 -> 6 -> 7 -> 8 -> 9 -> 10
  n <- 10L
  df <- data.frame(
    id = seq_len(n),
    parent_id = c(NA, seq_len(n - 1L)),
    rank = c("ROOT", rep("CHILD", n - 1L)),
    name = c("TopValue", paste0("node", 2:n)),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |>
    mutate(root_name = propagate(
      parent_id, id,
      ifelse(rank == "ROOT", name, NA)
    )) |>
    collect()

  expect_equal(result$root_name, rep("TopValue", n))
})

test_that("tiff round-trip: write then read back", {
  # Create a small test raster as a data.frame with x, y, band1
  nx <- 4; ny <- 3
  xres <- 0.5; yres <- 0.5
  xmin <- 10; ymax <- 50

  df <- expand.grid(
    col = seq_len(nx) - 1,
    row = seq_len(ny) - 1
  )
  df$x <- xmin + (df$col + 0.5) * xres
  df$y <- ymax - (df$row + 0.5) * yres
  df$band1 <- as.double(seq_len(nrow(df)))
  df <- df[, c("x", "y", "band1")]

  # Write via write_tiff
  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))

  vtr_tmp <- tempfile(fileext = ".vtr")
  on.exit(unlink(vtr_tmp), add = TRUE)
  write_vtr(df, vtr_tmp)
  tbl(vtr_tmp) |> write_tiff(tmp)

  # Read back via tbl_tiff
  node <- tbl_tiff(tmp)
  result <- node |> collect()

  expect_equal(ncol(result), 3)
  expect_true("x" %in% names(result))
  expect_true("y" %in% names(result))
  expect_true("band1" %in% names(result))
  expect_equal(nrow(result), nx * ny)
})

test_that("tiff round-trip preserves values", {
  nx <- 3; ny <- 2
  xres <- 1; yres <- 1

  df <- expand.grid(col = 0:(nx-1), row = 0:(ny-1))
  df$x <- 0.5 + df$col * xres
  df$y <- (ny - 0.5) - df$row * yres
  df$band1 <- c(1.5, 2.5, 3.5, 4.5, 5.5, 6.5)
  df <- df[, c("x", "y", "band1")]

  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))

  vtr_tmp <- tempfile(fileext = ".vtr")
  on.exit(unlink(vtr_tmp), add = TRUE)
  write_vtr(df, vtr_tmp)
  tbl(vtr_tmp) |> write_tiff(tmp)

  result <- tbl_tiff(tmp) |> collect()

  # Sort both by x then y for comparison
  df_sorted <- df[order(df$y, df$x), ]
  res_sorted <- result[order(result$y, result$x), ]

  expect_equal(res_sorted$band1, df_sorted$band1, tolerance = 1e-10)
})

test_that("tiff filter works (extent crop)", {
  nx <- 4; ny <- 4
  df <- expand.grid(col = 0:(nx-1), row = 0:(ny-1))
  df$x <- 0.5 + df$col
  df$y <- (ny - 0.5) - df$row
  df$band1 <- as.double(seq_len(nrow(df)))
  df <- df[, c("x", "y", "band1")]

  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))

  vtr_tmp <- tempfile(fileext = ".vtr")
  on.exit(unlink(vtr_tmp), add = TRUE)
  write_vtr(df, vtr_tmp)
  tbl(vtr_tmp) |> write_tiff(tmp)

  # Crop to x >= 1.5 (should get cols 1,2,3 = 3 cols)
  result <- tbl_tiff(tmp) |>
    filter(x >= 1.5) |>
    collect()

  expect_true(all(result$x >= 1.5))
  expect_equal(nrow(result), 3 * ny)
})

test_that("tiff multi-band round-trip", {
  nx <- 3; ny <- 2
  df <- expand.grid(col = 0:(nx-1), row = 0:(ny-1))
  df$x <- 0.5 + df$col
  df$y <- (ny - 0.5) - df$row
  df$band1 <- as.double(1:6)
  df$band2 <- as.double(7:12)
  df <- df[, c("x", "y", "band1", "band2")]

  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))
  write_tiff(df, tmp)

  result <- tbl_tiff(tmp) |> collect()

  expect_equal(ncol(result), 4)
  expect_true("band1" %in% names(result))
  expect_true("band2" %in% names(result))
  expect_equal(nrow(result), 6)
})

test_that("tiff explain shows TiffScanNode", {
  nx <- 2; ny <- 2
  df <- data.frame(x = c(0.5, 1.5, 0.5, 1.5),
                   y = c(1.5, 1.5, 0.5, 0.5),
                   band1 = c(1, 2, 3, 4))

  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))
  write_tiff(df, tmp)

  node <- tbl_tiff(tmp)
  out <- capture.output(explain(node))
  expect_true(any(grepl("TiffScanNode", out)))
})

test_that("tbl_tiff returns metadata", {
  df <- data.frame(x = c(0.5, 1.5, 0.5, 1.5),
                   y = c(1.5, 1.5, 0.5, 0.5),
                   band1 = c(10, 20, 30, 40))

  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))
  write_tiff(df, tmp)

  node <- tbl_tiff(tmp)
  expect_true(!is.null(node$.tiff_meta))
  expect_equal(node$.tiff_meta$width, 2)
  expect_equal(node$.tiff_meta$height, 2)
  expect_equal(node$.tiff_meta$nbands, 1L)
})

test_that("write_tiff with deflate compression", {
  df <- data.frame(x = c(0.5, 1.5, 0.5, 1.5),
                   y = c(1.5, 1.5, 0.5, 0.5),
                   band1 = c(1, 2, 3, 4))

  tmp <- tempfile(fileext = ".tif")
  on.exit(unlink(tmp))
  write_tiff(df, tmp, compress = TRUE)

  result <- tbl_tiff(tmp) |> collect()
  expect_equal(nrow(result), 4)

  res_sorted <- result[order(result$y, result$x), ]
  df_sorted <- df[order(df$y, df$x), ]
  expect_equal(res_sorted$band1, df_sorted$band1)
})

# Join key parsing: handles by = "col", by = c("a" = "b"), or NULL (natural)
parse_join_keys <- function(x, y, by) {
  x_schema <- .Call(C_node_schema, x$.node)
  y_schema <- .Call(C_node_schema, y$.node)

  if (is.null(by)) {
    # Natural join: common column names
    common <- intersect(x_schema$name, y_schema$name)
    if (length(common) == 0)
      stop("no common columns found for natural join; specify 'by'")
    message(sprintf("Joining by: %s", paste(common, collapse = ", ")))
    return(list(left = common, right = common))
  }

  if (is.null(names(by))) {
    # by = c("a", "b") -> same names on both sides
    for (nm in by) {
      if (!nm %in% x_schema$name) stop(sprintf("column '%s' not found in x", nm))
      if (!nm %in% y_schema$name) stop(sprintf("column '%s' not found in y", nm))
    }
    return(list(left = by, right = by))
  }

  # by = c("a" = "b") -> left_key = "a", right_key = "b"
  left_keys <- names(by)
  right_keys <- unname(by)
  # Handle unnamed entries (same name on both sides)
  for (i in seq_along(left_keys)) {
    if (left_keys[i] == "") left_keys[i] <- right_keys[i]
  }
  for (nm in left_keys) {
    if (!nm %in% x_schema$name) stop(sprintf("column '%s' not found in x", nm))
  }
  for (nm in right_keys) {
    if (!nm %in% y_schema$name) stop(sprintf("column '%s' not found in y", nm))
  }
  list(left = left_keys, right = right_keys)
}

#' Join two vectra tables
#'
#' @param x A `vectra_node` object (left table).
#' @param y A `vectra_node` object (right table).
#' @param by A character vector of column names to join by, or a named vector
#'   like `c("a" = "b")`. `NULL` for natural join (common columns).
#' @param suffix A character vector of length 2 for disambiguating non-key
#'   columns with the same name (default `c(".x", ".y")`).
#' @param ... Ignored.
#'
#' @return A `vectra_node` with the joined result.
#'
#' @details
#' All joins use a build-right, probe-left hash join. The entire right-side
#' table is materialized into a hash table; left-side batches stream through.
#' Memory cost is proportional to the right-side table size.
#'
#' NA keys never match (SQL NULL semantics). Key types are auto-coerced
#' following the `bool < int64 < double` hierarchy. Joining string against
#' numeric keys is an error.
#'
#' @examples
#' f1 <- tempfile(fileext = ".vtr")
#' f2 <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(id = c(1, 2, 3), x = c(10, 20, 30)), f1)
#' write_vtr(data.frame(id = c(1, 2, 4), y = c(100, 200, 400)), f2)
#' left_join(tbl(f1), tbl(f2), by = "id") |> collect()
#' unlink(c(f1, f2))
#'
#' @export
left_join <- function(x, y, by = NULL, suffix = c(".x", ".y"), ...) {
  UseMethod("left_join")
}

# Internal: shared implementation for standard joins (left, inner, full)
join_impl <- function(x, y, by, suffix, type) {
  keys <- parse_join_keys(x, y, by)
  new_xptr <- .Call(C_join_node, x$.node, y$.node,
                    type, keys$left, keys$right, suffix[1], suffix[2])
  structure(list(.node = new_xptr, .path = NULL), class = "vectra_node")
}

# Internal: shared implementation for filtering joins (semi, anti)
filter_join_impl <- function(x, y, by, type) {
  keys <- parse_join_keys(x, y, by)
  new_xptr <- .Call(C_join_node, x$.node, y$.node,
                    type, keys$left, keys$right, ".x", ".y")
  structure(list(.node = new_xptr, .path = x$.path), class = "vectra_node")
}

#' @export
left_join.vectra_node <- function(x, y, by = NULL, suffix = c(".x", ".y"), ...) {
  if (!inherits(y, "vectra_node"))
    stop(sprintf("y must be a vectra_node, got %s", class(y)[1]))
  if (!is.character(suffix) || length(suffix) != 2)
    stop(sprintf("suffix must be character(2), got %s of length %d", class(suffix)[1], length(suffix)))
  join_impl(x, y, by, suffix, "left")
}

#' @rdname left_join
#' @export
inner_join <- function(x, y, by = NULL, suffix = c(".x", ".y"), ...) {
  UseMethod("inner_join")
}

#' @export
inner_join.vectra_node <- function(x, y, by = NULL, suffix = c(".x", ".y"), ...) {
  if (!inherits(y, "vectra_node"))
    stop(sprintf("y must be a vectra_node, got %s", class(y)[1]))
  if (!is.character(suffix) || length(suffix) != 2)
    stop(sprintf("suffix must be character(2), got %s of length %d", class(suffix)[1], length(suffix)))
  join_impl(x, y, by, suffix, "inner")
}

#' @rdname left_join
#' @export
right_join <- function(x, y, by = NULL, suffix = c(".x", ".y"), ...) {
  UseMethod("right_join")
}

#' @export
right_join.vectra_node <- function(x, y, by = NULL, suffix = c(".x", ".y"), ...) {
  if (!inherits(y, "vectra_node"))
    stop(sprintf("y must be a vectra_node, got %s", class(y)[1]))
  if (!is.character(suffix) || length(suffix) != 2)
    stop(sprintf("suffix must be character(2), got %s of length %d", class(suffix)[1], length(suffix)))
  # right_join(x, y) = left_join(y, x) with swapped keys and reordered columns
  keys <- parse_join_keys(x, y, by)

  # Get schemas before C_join_node clears the external pointers
  x_schema <- .Call(C_node_schema, x$.node)
  y_schema <- .Call(C_node_schema, y$.node)

  # Swap: build on left (x), probe with right (y)
  new_xptr <- .Call(C_join_node, y$.node, x$.node,
                    "left", keys$right, keys$left, suffix[2], suffix[1])
  result_node <- structure(list(.node = new_xptr, .path = NULL),
                           class = "vectra_node")
  schema <- .Call(C_node_schema, result_node$.node)

  # Current output: y_cols + x_non_key_cols
  # Desired output: x_cols + y_non_key_cols (with key values from y for matched)
  # This is complex to reorder at the node level. Use a project node.
  cur_names <- schema$name

  # x columns come after y columns in current output
  # Build the desired column order
  x_names <- x_schema$name
  y_names <- y_schema$name
  y_key_set <- keys$right

  # Desired: all x columns, then y non-key columns
  desired <- character(0)
  expr_lists <- list()

  # x key columns: use values from y side (key col in current output)
  for (xn in x_names) {
    ki <- match(xn, keys$left)
    if (!is.na(ki)) {
      # This x key maps to y key: find the y key col name in current output
      yk <- keys$right[ki]
      desired <- c(desired, xn)
      if (xn != yk) {
        expr_lists <- c(expr_lists, list(list(kind = "col_ref", name = yk)))
      } else {
        expr_lists <- c(expr_lists, list(NULL))
      }
    } else {
      # Non-key x column: exists in current output (possibly suffixed)
      match_name <- xn
      if (!match_name %in% cur_names) {
        # Try with suffix
        match_name <- paste0(xn, suffix[1])
      }
      desired <- c(desired, xn)
      if (match_name != xn) {
        expr_lists <- c(expr_lists, list(list(kind = "col_ref", name = match_name)))
      } else {
        expr_lists <- c(expr_lists, list(NULL))
      }
    }
  }
  # y non-key columns
  for (yn in y_names) {
    if (yn %in% y_key_set) next
    match_name <- yn
    if (!match_name %in% cur_names) {
      match_name <- paste0(yn, suffix[2])
    }
    desired <- c(desired, yn)
    if (match_name != yn) {
      expr_lists <- c(expr_lists, list(list(kind = "col_ref", name = match_name)))
    } else {
      expr_lists <- c(expr_lists, list(NULL))
    }
  }

  new_xptr2 <- .Call(C_project_node, result_node$.node, desired, expr_lists)
  structure(list(.node = new_xptr2, .path = NULL), class = "vectra_node")
}

#' @rdname left_join
#' @export
full_join <- function(x, y, by = NULL, suffix = c(".x", ".y"), ...) {
  UseMethod("full_join")
}

#' @export
full_join.vectra_node <- function(x, y, by = NULL, suffix = c(".x", ".y"), ...) {
  if (!inherits(y, "vectra_node"))
    stop(sprintf("y must be a vectra_node, got %s", class(y)[1]))
  if (!is.character(suffix) || length(suffix) != 2)
    stop(sprintf("suffix must be character(2), got %s of length %d", class(suffix)[1], length(suffix)))
  join_impl(x, y, by, suffix, "full")
}

#' @rdname left_join
#' @export
semi_join <- function(x, y, by = NULL, ...) {
  UseMethod("semi_join")
}

#' @export
semi_join.vectra_node <- function(x, y, by = NULL, ...) {
  if (!inherits(y, "vectra_node"))
    stop(sprintf("y must be a vectra_node, got %s", class(y)[1]))
  filter_join_impl(x, y, by, "semi")
}

#' @rdname left_join
#' @export
anti_join <- function(x, y, by = NULL, ...) {
  UseMethod("anti_join")
}

#' @export
anti_join.vectra_node <- function(x, y, by = NULL, ...) {
  if (!inherits(y, "vectra_node"))
    stop(sprintf("y must be a vectra_node, got %s", class(y)[1]))
  filter_join_impl(x, y, by, "anti")
}

#' Cross join two vectra tables
#'
#' Returns every combination of rows from `x` and `y` (Cartesian product).
#' Both tables are collected before joining.
#'
#' @param x A `vectra_node` object or data.frame.
#' @param y A `vectra_node` object or data.frame.
#' @param suffix Suffixes for disambiguating column names (default `c(".x", ".y")`).
#' @param ... Ignored.
#'
#' @return A data.frame with `nrow(x) * nrow(y)` rows.
#'
#' @examples
#' f1 <- tempfile(fileext = ".vtr")
#' f2 <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(a = 1:2), f1)
#' write_vtr(data.frame(b = c("x", "y", "z"), stringsAsFactors = FALSE), f2)
#' cross_join(tbl(f1), tbl(f2))
#' unlink(c(f1, f2))
#'
#' @export
cross_join <- function(x, y, suffix = c(".x", ".y"), ...) {
  UseMethod("cross_join")
}

#' @export
cross_join.vectra_node <- function(x, y, suffix = c(".x", ".y"), ...) {
  if (!inherits(y, "vectra_node") && !is.data.frame(y))
    stop(sprintf("y must be a vectra_node or data.frame, got %s", class(y)[1]))
  if (!is.character(suffix) || length(suffix) != 2)
    stop(sprintf("suffix must be character(2), got %s of length %d", class(suffix)[1], length(suffix)))
  df_x <- if (inherits(x, "vectra_node")) collect(x) else x
  df_y <- if (inherits(y, "vectra_node")) collect(y) else y

  nx <- nrow(df_x)
  ny <- nrow(df_y)

  # Expand: repeat each x row ny times, tile y nx times
  idx_x <- rep(seq_len(nx), each = ny)
  idx_y <- rep(seq_len(ny), times = nx)

  result_x <- df_x[idx_x, , drop = FALSE]
  result_y <- df_y[idx_y, , drop = FALSE]

  # Handle name collisions
  common <- intersect(names(result_x), names(result_y))
  if (length(common) > 0) {
    for (nm in common) {
      names(result_x)[names(result_x) == nm] <- paste0(nm, suffix[1])
      names(result_y)[names(result_y) == nm] <- paste0(nm, suffix[2])
    }
  }

  result <- cbind(result_x, result_y)
  rownames(result) <- NULL
  result
}


#' Fuzzy join two vectra tables by string distance
#'
#' Joins two tables using approximate string matching on key columns.
#' Optionally blocks by a second column (e.g., genus) for performance —
#' only rows sharing the same blocking key are compared.
#'
#' @param x A `vectra_node` object (probe / query side).
#' @param y A `vectra_node` object (build / reference side).
#' @param by A named character vector of length 1: `c("probe_col" = "build_col")`.
#'   The columns to compute string distance on.
#' @param method Character. Distance algorithm: `"dl"` (Damerau-Levenshtein,
#'   default), `"levenshtein"`, or `"jw"` (Jaro-Winkler).
#' @param max_dist Numeric. Maximum normalized distance (0-1) to keep a match.
#'   Default `0.2`.
#' @param block_by Optional named character vector of length 1:
#'   `c("probe_col" = "build_col")`. Rows must match exactly on these columns
#'   before distance is computed. Dramatically reduces comparisons.
#' @param n_threads Integer. Number of OpenMP threads for parallel distance
#'   computation over partitions. Default `4L`.
#' @param suffix Character. Suffix appended to build-side column names that
#'   collide with probe-side names. Default `".y"`.
#'
#' @return A `vectra_node` with all probe columns, all build columns (suffixed
#'   on collision), and a `fuzzy_dist` column (double).
#'
#' @export
fuzzy_join <- function(x, y, by, method = "dl", max_dist = 0.2,
                       block_by = NULL, n_threads = 4L,
                       suffix = ".y") {
  UseMethod("fuzzy_join")
}

#' @export
fuzzy_join.vectra_node <- function(x, y, by, method = "dl", max_dist = 0.2,
                                   block_by = NULL, n_threads = 4L,
                                   suffix = ".y") {
  if (!inherits(y, "vectra_node"))
    stop(sprintf("y must be a vectra_node, got %s", class(y)[1]))
  if (!is.character(by) || length(by) != 1)
    stop(sprintf("by must be a named character vector of length 1, got length %d", length(by)))
  if (!is.numeric(max_dist) || length(max_dist) != 1 || max_dist < 0 || max_dist > 1)
    stop(sprintf("max_dist must be a number between 0 and 1, got %s", deparse(max_dist)))
  if (!is.numeric(n_threads) || length(n_threads) != 1 || n_threads < 1)
    stop(sprintf("n_threads must be a positive integer, got %s", deparse(n_threads)))
  method_int <- match(method, c("dl", "levenshtein", "jw")) - 1L
  if (is.na(method_int)) stop("method must be 'dl', 'levenshtein', or 'jw'")

  by_probe <- names(by)
  by_build <- unname(by)
  if (is.null(by_probe) || by_probe == "") by_probe <- by_build

  block_probe <- if (!is.null(block_by)) names(block_by) else NULL
  block_build <- if (!is.null(block_by)) unname(block_by) else NULL
  if (!is.null(block_probe) && block_probe == "") block_probe <- block_build

  new_xptr <- .Call(C_fuzzy_join_node,
                    x$.node, y$.node,
                    by_probe, by_build,
                    block_probe, block_build,
                    as.integer(method_int),
                    as.double(max_dist),
                    as.integer(n_threads),
                    suffix)
  structure(list(.node = new_xptr, .path = NULL), class = "vectra_node")
}

# Helper to create two tables for join tests
make_join_tables <- function() {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(id = c(1, 2, 3), x = c(10, 20, 30)), f1)
  write_vtr(data.frame(id = c(1, 2, 4), y = c(100, 200, 400)), f2)
  list(f1 = f1, f2 = f2)
}

# --- inner_join ---

test_that("inner_join keeps only matching rows", {
  ft <- make_join_tables()
  on.exit(unlink(c(ft$f1, ft$f2)))
  result <- inner_join(tbl(ft$f1), tbl(ft$f2), by = "id") |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(sort(result$id), c(1, 2))
  expect_equal(result$x[result$id == 1], 10)
  expect_equal(result$y[result$id == 1], 100)
})

# --- left_join ---

test_that("left_join keeps all left rows, NAs for unmatched", {
  ft <- make_join_tables()
  on.exit(unlink(c(ft$f1, ft$f2)))
  result <- left_join(tbl(ft$f1), tbl(ft$f2), by = "id") |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(result$y[result$id == 1], 100)
  expect_equal(result$y[result$id == 2], 200)
  expect_true(is.na(result$y[result$id == 3]))
})

# --- semi_join ---

test_that("semi_join keeps left rows with matches, no right cols", {
  ft <- make_join_tables()
  on.exit(unlink(c(ft$f1, ft$f2)))
  result <- semi_join(tbl(ft$f1), tbl(ft$f2), by = "id") |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(sort(result$id), c(1, 2))
  expect_equal(names(result), c("id", "x"))
})

# --- anti_join ---

test_that("anti_join keeps left rows without matches", {
  ft <- make_join_tables()
  on.exit(unlink(c(ft$f1, ft$f2)))
  result <- anti_join(tbl(ft$f1), tbl(ft$f2), by = "id") |> collect()
  expect_equal(nrow(result), 1)
  expect_equal(result$id, 3)
  expect_equal(result$x, 30)
})

# --- named by (different column names) ---

test_that("join with named by works", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(a = c(1, 2, 3), x = c(10, 20, 30)), f1)
  write_vtr(data.frame(b = c(1, 2, 4), y = c(100, 200, 400)), f2)
  result <- inner_join(tbl(f1), tbl(f2), by = c("a" = "b")) |> collect()
  expect_equal(nrow(result), 2)
  expect_true("x" %in% names(result))
  expect_true("y" %in% names(result))
})

# --- natural join (no by) ---

test_that("natural join finds common columns", {
  ft <- make_join_tables()
  on.exit(unlink(c(ft$f1, ft$f2)))
  expect_message(
    result <- inner_join(tbl(ft$f1), tbl(ft$f2)) |> collect(),
    "Joining by"
  )
  expect_equal(nrow(result), 2)
})

# --- duplicate handling ---

test_that("inner_join with duplicates on right produces multiple rows", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1, 2), x = c(10, 20)), f1)
  write_vtr(data.frame(id = c(1, 1, 2), y = c(100, 101, 200)), f2)
  result <- inner_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(sum(result$id == 1), 2)
})

# --- NA key handling ---

test_that("NAs in join keys do not match", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1, NA, 3), x = c(10, 20, 30)), f1)
  write_vtr(data.frame(id = c(1, NA, 4), y = c(100, 200, 400)), f2)
  result <- inner_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 1)
  expect_equal(result$id, 1)
})

test_that("left_join with NA keys: unmatched NAs get NA right cols", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1, NA), x = c(10, 20)), f1)
  write_vtr(data.frame(id = c(1, NA), y = c(100, 200)), f2)
  result <- left_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(result$y[which(result$id == 1)], 100)
  expect_true(is.na(result$y[which(is.na(result$id))]))
})

# --- suffix for name collisions ---

test_that("suffix disambiguates columns with same name", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1, 2), val = c(10, 20)), f1)
  write_vtr(data.frame(id = c(1, 2), val = c(100, 200)), f2)
  result <- left_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_true("val.x" %in% names(result))
  expect_true("val.y" %in% names(result))
})

# --- string keys ---

test_that("join works with string keys", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(name = c("alice", "bob"), x = c(1, 2),
                       stringsAsFactors = FALSE), f1)
  write_vtr(data.frame(name = c("bob", "carol"), y = c(20, 30),
                       stringsAsFactors = FALSE), f2)
  result <- inner_join(tbl(f1), tbl(f2), by = "name") |> collect()
  expect_equal(nrow(result), 1)
  expect_equal(result$name, "bob")
  expect_equal(result$y, 20)
})

# --- composite keys ---

test_that("join works with composite keys", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(a = c(1, 1, 2), b = c("x", "y", "x"),
                       v1 = c(10, 20, 30), stringsAsFactors = FALSE), f1)
  write_vtr(data.frame(a = c(1, 2), b = c("x", "x"),
                       v2 = c(100, 300), stringsAsFactors = FALSE), f2)
  result <- inner_join(tbl(f1), tbl(f2), by = c("a", "b")) |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(sort(result$v1), c(10, 30))
})

# --- join after filter ---

test_that("string-key join survives many unique keys", {
  # 200 unique string keys forces hash table to work hard.
  # Regression test for join key lifetime / string data integrity.
  set.seed(42)
  n_keys <- 200
  keys <- paste0("key_", seq_len(n_keys))
  left_df <- data.frame(
    id = rep(keys, each = 3),
    x = rnorm(n_keys * 3),
    stringsAsFactors = FALSE
  )
  right_df <- data.frame(
    id = keys,
    y = rnorm(n_keys),
    stringsAsFactors = FALSE
  )
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(left_df, f1)
  write_vtr(right_df, f2)

  # left_join: all left rows should have a match
  result <- tbl(f1) |> left_join(tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), nrow(left_df))
  expect_false(any(is.na(result$y)))

  # full_join with partial overlap
  partial_right <- data.frame(
    id = keys[1:100],
    z = rnorm(100),
    stringsAsFactors = FALSE
  )
  f3 <- tempfile(fileext = ".vtr")
  on.exit(unlink(f3), add = TRUE)
  write_vtr(partial_right, f3)
  result2 <- tbl(f1) |> full_join(tbl(f3), by = "id") |> collect()
  # All left rows present, no extra right-only rows (right keys are subset)
  expect_equal(nrow(result2), nrow(left_df))
  # Keys 101-200 have NA z
  unmatched <- result2[result2$id %in% keys[101:200], ]
  expect_true(all(is.na(unmatched$z)))

  # Chain a downstream op on the result to verify string integrity
  f4 <- tempfile(fileext = ".vtr")
  on.exit(unlink(f4), add = TRUE)
  write_vtr(result, f4)
  roundtrip <- tbl(f4) |> filter(y > 0) |> collect()
  expect_true(all(roundtrip$y > 0))
  expect_true(nrow(roundtrip) > 0)
})

# --- key type mismatch ---

test_that("join auto-coerces compatible key types (int64 vs double)", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  # integer keys on left, double keys on right -> should auto-coerce
  write_vtr(data.frame(id = c(1L, 2L, 3L), x = c(10, 20, 30)), f1)
  write_vtr(data.frame(id = c(1, 2, 4), y = c(100, 200, 400)), f2)
  result <- inner_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(sort(result$id), c(1, 2))
})

test_that("join errors on key type mismatch (string vs double)", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c("a", "b"), x = c(10, 20),
                       stringsAsFactors = FALSE), f1)
  write_vtr(data.frame(id = c(1, 2), y = c(100, 200)), f2)
  expect_error(
    left_join(tbl(f1), tbl(f2), by = "id") |> collect(),
    "left_join key type mismatch: x\\.id \\(string\\) vs y\\.id \\(double\\)"
  )
})

test_that("composite key auto-coerces mixed numeric types", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  # First key (a) matches (both double), second key (b) int vs double
  write_vtr(data.frame(a = c(1, 2), b = c(10L, 20L), x = c(1, 2)), f1)
  write_vtr(data.frame(a = c(1, 2), b = c(10, 20), y = c(3, 4)), f2)
  result <- inner_join(tbl(f1), tbl(f2), by = c("a", "b")) |> collect()
  expect_equal(nrow(result), 2)
})

test_that("join works when both sides have same numeric type", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  # Both double
  write_vtr(data.frame(id = c(1, 2, 3), x = c(10, 20, 30)), f1)
  write_vtr(data.frame(id = c(1, 2, 4), y = c(100, 200, 400)), f2)
  result <- inner_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 2)
  # Both integer
  write_vtr(data.frame(id = c(1L, 2L, 3L), x = c(10, 20, 30)), f1)
  write_vtr(data.frame(id = c(1L, 2L, 4L), y = c(100, 200, 400)), f2)
  result2 <- inner_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result2), 2)
})

# --- long string keys ---

test_that("join handles very long string keys", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  # Keys with 100k+ bytes each
  long_a <- paste(rep("a", 100000), collapse = "")
  long_b <- paste(rep("b", 100000), collapse = "")
  long_c <- paste(rep("c", 100000), collapse = "")
  write_vtr(data.frame(key = c(long_a, long_b, long_c),
                       x = c(1, 2, 3), stringsAsFactors = FALSE), f1)
  write_vtr(data.frame(key = c(long_b, long_c, "short"),
                       y = c(20, 30, 99), stringsAsFactors = FALSE), f2)
  result <- inner_join(tbl(f1), tbl(f2), by = "key") |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(result$x[result$key == long_b], 2)
  expect_equal(result$y[result$key == long_c], 30)
})

test_that("join handles many distinct long string keys", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  n <- 500
  # Each key is ~1000 chars with a unique suffix
  base <- paste(rep("x", 1000), collapse = "")
  keys <- paste0(base, "_", seq_len(n))
  write_vtr(data.frame(key = keys, x = seq_len(n),
                       stringsAsFactors = FALSE), f1)
  write_vtr(data.frame(key = keys[1:250], y = seq_len(250),
                       stringsAsFactors = FALSE), f2)
  result <- left_join(tbl(f1), tbl(f2), by = "key") |> collect()
  expect_equal(nrow(result), n)
  expect_equal(sum(!is.na(result$y)), 250)
  expect_equal(sum(is.na(result$y)), 250)
})

# --- join after filter ---

# --- streaming probe tests (multi-batch left side) ---

test_that("streaming inner_join across multiple probe batches", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  # 100 left rows in 10 batches of 10 rows each (all integer)
  left_df <- data.frame(id = rep(1L:10L, each = 10L),
                        x = seq_len(100L))
  right_df <- data.frame(id = c(1L, 5L, 10L), y = c(10L, 50L, 100L))
  write_vtr(left_df, f1, batch_size = 10L)
  write_vtr(right_df, f2)
  result <- inner_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 30)  # 3 matching ids * 10 rows each
  expect_equal(sort(unique(result$id)), c(1, 5, 10))
})

test_that("streaming left_join with unmatched rows across batches", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  left_df <- data.frame(id = 1L:50L, x = seq_len(50L))
  right_df <- data.frame(id = c(1L, 25L, 50L), y = c(100L, 250L, 500L))
  write_vtr(left_df, f1, batch_size = 10L)  # 5 batches
  write_vtr(right_df, f2)
  result <- left_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 50)
  expect_equal(sum(!is.na(result$y)), 3)
  expect_equal(sum(is.na(result$y)), 47)
  expect_equal(result$y[result$id == 25], 250)
})

test_that("streaming semi_join across multiple batches", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  left_df <- data.frame(id = 1L:30L, x = seq_len(30L))
  right_df <- data.frame(id = c(5L, 15L, 25L), y = c(1L, 2L, 3L))
  write_vtr(left_df, f1, batch_size = 10L)  # 3 batches
  write_vtr(right_df, f2)
  result <- semi_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(sort(result$id), c(5, 15, 25))
  expect_equal(names(result), c("id", "x"))
})

test_that("streaming anti_join across batches where some batches produce no output", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  # Batch 1 (ids 1-10): all matched -> 0 output rows
  # Batch 2 (ids 11-20): none matched -> 10 output rows
  left_df <- data.frame(id = 1L:20L, x = seq_len(20L))
  right_df <- data.frame(id = 1L:10L, y = seq_len(10L))
  write_vtr(left_df, f1, batch_size = 10L)
  write_vtr(right_df, f2)
  result <- anti_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 10)
  expect_equal(sort(result$id), 11:20)
})

test_that("streaming full_join emits unmatched build rows after probe", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  left_df <- data.frame(id = 1L:20L, x = seq_len(20L))
  right_df <- data.frame(id = c(5L, 15L, 99L, 100L),
                         y = c(50L, 150L, 990L, 1000L))
  write_vtr(left_df, f1, batch_size = 10L)
  write_vtr(right_df, f2)
  result <- full_join(tbl(f1), tbl(f2), by = "id") |> collect()
  # 20 left rows + 2 unmatched right rows (99, 100)
  expect_equal(nrow(result), 22)
  # Unmatched right rows have NA x
  unmatched <- result[result$id %in% c(99, 100), ]
  expect_equal(nrow(unmatched), 2)
  expect_true(all(is.na(unmatched$x)))
  expect_equal(sort(unmatched$y), c(990, 1000))
})

test_that("streaming many-to-many join works across batches", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  # 30 left rows, 3 batches of 10 (all integer)
  # Right side has 5 rows per key for keys 1-3
  left_df <- data.frame(id = rep(1L:3L, each = 10L),
                        x = seq_len(30L))
  right_df <- data.frame(id = rep(1L:3L, each = 5L),
                         y = seq_len(15L))
  write_vtr(left_df, f1, batch_size = 10L)
  write_vtr(right_df, f2)
  result <- inner_join(tbl(f1), tbl(f2), by = "id") |> collect()
  # Each of 30 left rows matches 5 right rows -> 150
  expect_equal(nrow(result), 150)
})

# --- join after filter ---

test_that("join composes with filter", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1, 2, 3), x = c(10, 20, 30)), f1)
  write_vtr(data.frame(id = c(1, 2, 3), y = c(100, 200, 300)), f2)
  result <- tbl(f1) |>
    filter(x > 10) |>
    inner_join(tbl(f2), by = "id") |>
    collect()
  expect_equal(nrow(result), 2)
  expect_true(all(result$x > 10))
})

# --- empty filter results ---

test_that("filter producing zero rows returns empty data.frame", {
  df <- data.frame(x = c(1.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(x > 100) |> collect()
  expect_equal(nrow(result), 0)
  expect_equal(names(result), "x")
})

# --- group_by on zero-row input ---

test_that("summarise on filtered-to-empty input works", {
  df <- data.frame(g = c("a", "b"), x = c(1.0, 2.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(x > 100) |> group_by(g) |> summarise(s = sum(x)) |> collect()
  expect_equal(nrow(result), 0)
})

# --- join with empty right table ---

test_that("left_join with empty right table preserves all left rows", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1, 2, 3), x = c(10, 20, 30)), f1)
  write_vtr(data.frame(id = c(99.0), y = c(999.0)), f2)
  result <- left_join(tbl(f1), tbl(f2) |> filter(id > 1000), by = "id") |> collect()
  expect_equal(nrow(result), 3)
  expect_true(all(is.na(result$y)))
})

test_that("inner_join with empty right table returns zero rows", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1, 2), x = c(10, 20)), f1)
  write_vtr(data.frame(id = c(99.0), y = c(999.0)), f2)
  result <- inner_join(tbl(f1), tbl(f2) |> filter(id > 1000), by = "id") |> collect()
  expect_equal(nrow(result), 0)
})

# --- distinct with .keep_all fallback ---

test_that("distinct with .keep_all = TRUE and column subset falls back to R", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, 2.0, 3.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  expect_message(
    result <- tbl(f) |> distinct(g, .keep_all = TRUE),
    "falling back"
  )
  expect_equal(nrow(result), 2)
  expect_true("x" %in% names(result))
})

# --- NA handling in window functions ---

test_that("lead with NA values propagates NAs correctly", {
  df <- data.frame(x = c(1.0, NA, 3.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(nxt = lead(x)) |> collect()
  expect_true(is.na(result$nxt[1]))   # lead shifts up: value at position 2 is NA
  expect_equal(result$nxt[2], 3)      # value at position 3
  expect_equal(result$nxt[3], 4)      # value at position 4
  expect_true(is.na(result$nxt[4]))   # last row has no next value
})

# --- bind_rows with different schemas ---

test_that("bind_rows with different columns fills missing with NA", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(a = c(1.0, 2.0), b = c(3.0, 4.0)), f1)
  write_vtr(data.frame(b = c(5.0, 6.0), c = c(7.0, 8.0)), f2)
  result <- bind_rows(tbl(f1), tbl(f2))
  expect_equal(nrow(result), 4)
  expect_true(all(c("a", "b", "c") %in% names(result)))
  expect_true(is.na(result$a[3]))
  expect_true(is.na(result$c[1]))
})

# --- empty input to bind_rows ---

test_that("bind_rows with single input returns that input", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(data.frame(x = c(1.0, 2.0)), f)
  result <- bind_rows(tbl(f))
  expect_equal(nrow(result), 2)
})

# --- reframe ---

test_that("reframe returns variable-length output", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, 2.0, 3.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> reframe(range_x = range(x))
  # group a: range(1,2) = c(1,2) -> 2 rows; group b: range(3) = c(3,3) -> 2 rows
  expect_equal(nrow(result), 4)
})

# --- across in summarise ---

test_that("across in summarise applies function to multiple columns", {
  df <- data.frame(x = c(1.0, 2.0, 3.0), y = c(10.0, 20.0, 30.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> summarise(across(c(x, y), mean)) |> collect()
  expect_equal(result$x, 2)
  expect_equal(result$y, 20)
})

# --- string edge cases in filter ---

test_that("filter with empty string comparison works", {
  df <- data.frame(s = c("", "a", "b"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(s == "") |> collect()
  expect_equal(nrow(result), 1)
  expect_equal(result$s, "")
})

test_that("grepl with fixed string match works", {
  df <- data.frame(s = c("hello world", "foo", "hello there"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(grepl("hello", s)) |> collect()
  expect_equal(nrow(result), 2)
})

# --- right_join ---

test_that("right_join keeps all right rows", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1.0, 2.0, 3.0), x = c(10.0, 20.0, 30.0)), f1)
  write_vtr(data.frame(id = c(2.0, 3.0, 4.0), y = c(200.0, 300.0, 400.0)), f2)
  result <- right_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 3)
  expect_true(all(c(2, 3, 4) %in% result$id))
  # id=4 should have NA for x
  expect_true(is.na(result$x[result$id == 4]))
  expect_equal(result$y[result$id == 2], 200)
})

test_that("right_join with named by", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(a = c(1.0, 2.0), x = c(10.0, 20.0)), f1)
  write_vtr(data.frame(b = c(2.0, 3.0), y = c(200.0, 300.0)), f2)
  result <- right_join(tbl(f1), tbl(f2), by = c("a" = "b")) |> collect()
  expect_equal(nrow(result), 2)
  expect_true("a" %in% names(result))
  expect_true("y" %in% names(result))
})

# --- full_join ---

test_that("full_join keeps all rows from both sides", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1.0, 2.0, 3.0), x = c(10.0, 20.0, 30.0)), f1)
  write_vtr(data.frame(id = c(2.0, 3.0, 4.0), y = c(200.0, 300.0, 400.0)), f2)
  result <- full_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 4)
  expect_true(all(c(1, 2, 3, 4) %in% result$id))
  # id=1 has NA y, id=4 has NA x
  expect_true(is.na(result$y[result$id == 1]))
  expect_true(is.na(result$x[result$id == 4]))
  # matching rows
  expect_equal(result$x[result$id == 2], 20)
  expect_equal(result$y[result$id == 2], 200)
})

test_that("full_join with no overlap", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1.0, 2.0), x = c(10.0, 20.0)), f1)
  write_vtr(data.frame(id = c(3.0, 4.0), y = c(30.0, 40.0)), f2)
  result <- full_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 4)
  expect_true(all(is.na(result$y[result$id %in% c(1, 2)])))
  expect_true(all(is.na(result$x[result$id %in% c(3, 4)])))
})

test_that("full_join with complete overlap", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(id = c(1.0, 2.0), x = c(10.0, 20.0)), f1)
  write_vtr(data.frame(id = c(1.0, 2.0), y = c(100.0, 200.0)), f2)
  result <- full_join(tbl(f1), tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 2)
  expect_false(any(is.na(result$x)))
  expect_false(any(is.na(result$y)))
})

# --- bind_rows ---

test_that("bind_rows combines same-schema tables (streaming)", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(x = c(1.0, 2.0)), f1)
  write_vtr(data.frame(x = c(3.0, 4.0)), f2)
  node <- bind_rows(tbl(f1), tbl(f2))
  expect_s3_class(node, "vectra_node")
  result <- collect(node)
  expect_equal(nrow(result), 4)
  expect_equal(result$x, c(1, 2, 3, 4))
})

test_that("bind_rows fills missing columns with NA", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(x = c(1.0, 2.0), y = c(10.0, 20.0)), f1)
  write_vtr(data.frame(x = c(3.0, 4.0)), f2)
  result <- bind_rows(tbl(f1), tbl(f2))
  expect_equal(nrow(result), 4)
  expect_true(all(is.na(result$y[3:4])))
})

test_that("bind_rows with .id", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(x = c(1.0, 2.0)), f1)
  write_vtr(data.frame(x = c(3.0, 4.0)), f2)
  result <- bind_rows(tbl(f1), tbl(f2), .id = "source")
  expect_true("source" %in% names(result))
  expect_equal(result$source, c(1, 1, 2, 2))
})

test_that("streaming bind_rows composes with filter", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(x = c(1.0, 2.0, 3.0)), f1)
  write_vtr(data.frame(x = c(4.0, 5.0, 6.0)), f2)
  result <- bind_rows(tbl(f1), tbl(f2)) |> filter(x > 3) |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(result$x, c(4, 5, 6))
})

test_that("streaming bind_rows explain shows ConcatNode", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(x = c(1.0, 2.0)), f1)
  write_vtr(data.frame(x = c(3.0, 4.0)), f2)
  output <- capture.output(bind_rows(tbl(f1), tbl(f2)) |> explain())
  expect_true(any(grepl("ConcatNode", output)))
  expect_true(any(grepl("streaming", output)))
})

# --- bind_cols ---

test_that("bind_cols combines columns", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(x = c(1.0, 2.0)), f1)
  write_vtr(data.frame(y = c(10.0, 20.0)), f2)
  result <- bind_cols(tbl(f1), tbl(f2))
  expect_equal(names(result), c("x", "y"))
  expect_equal(nrow(result), 2)
})

test_that("bind_cols errors on row count mismatch", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(x = c(1.0, 2.0)), f1)
  write_vtr(data.frame(y = c(1.0, 2.0, 3.0)), f2)
  expect_error(bind_cols(tbl(f1), tbl(f2)))
})

# --- reframe ---

test_that("reframe allows variable-length output", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(data.frame(g = c("a", "a", "b", "b", "b"),
                       x = c(1.0, 2.0, 3.0, 4.0, 5.0),
                       stringsAsFactors = FALSE), f)
  result <- tbl(f) |> group_by(g) |> reframe(vals = sort(x))
  expect_equal(nrow(result), 5)
  expect_true(all(c("g", "vals") %in% names(result)))
})

# --- explain ---

test_that("explain shows plan with node types", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(data.frame(x = c(1.0, 2.0, 3.0)), f)
  output <- capture.output(tbl(f) |> filter(x > 1) |> explain())
  expect_true(any(grepl("ScanNode", output)))
  expect_true(any(grepl("FilterNode", output)))
  expect_true(any(grepl("streaming", output)))
})

# Basic dplyr verbs: arrange, desc, filter, select, mutate
# (including window function detection logic inside mutate)

#' Sort rows by column values
#'
#' @param .data A `vectra_node` object.
#' @param ... Column names (unquoted). Wrap in [desc()] for descending order.
#'
#' @return A new `vectra_node` with sorted rows.
#'
#' @details
#' Uses an external merge sort with a 1 GB memory budget. When data exceeds
#' this limit, sorted runs are spilled to temporary `.vtr` files and merged
#' via a k-way min-heap. NAs sort last in ascending order.
#'
#' This is a materializing operation.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> arrange(desc(mpg)) |> collect() |> head()
#' unlink(f)
#'
#' @export
arrange <- function(.data, ...) {
  UseMethod("arrange")
}

#' @export
arrange.vectra_node <- function(.data, ...) {
  dots <- eval(substitute(alist(...)))
  if (length(dots) == 0) return(.data)

  col_names <- character(length(dots))
  desc_flags <- logical(length(dots))
  for (i in seq_along(dots)) {
    expr <- dots[[i]]
    if (is.call(expr) && identical(expr[[1]], as.name("desc"))) {
      col_names[i] <- as.character(expr[[2]])
      desc_flags[i] <- TRUE
    } else {
      col_names[i] <- as.character(expr)
      desc_flags[i] <- FALSE
    }
  }

  new_xptr <- .Call(C_sort_node, .data$.node, col_names, desc_flags)
  structure(list(.node = new_xptr, .path = .data$.path,
                 .groups = .data$.groups), class = "vectra_node")
}

#' Mark a column for descending sort order
#'
#' Used inside [arrange()] to sort a column in descending order.
#'
#' @param x A column name.
#'
#' @return A marker used by [arrange()].
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> arrange(desc(mpg)) |> collect() |> head()
#' unlink(f)
#'
#' @export
desc <- function(x) {
  structure(x, desc = TRUE)
}

#' Filter rows of a vectra query
#'
#' @param .data A `vectra_node` object.
#' @param ... Filter expressions (combined with `&`).
#'
#' @return A new `vectra_node` with the filter applied.
#'
#' @details
#' Filter uses zero-copy selection vectors: matching rows are indexed without
#' copying data. Multiple conditions are combined with `&`. Supported
#' expression types: arithmetic (`+`, `-`, `*`, `/`, `%%`), comparison
#' (`==`, `!=`, `<`, `<=`, `>`, `>=`), boolean (`&`, `|`, `!`), `is.na()`,
#' and string functions (`nchar()`, `substr()`, `grepl()` with fixed patterns).
#'
#' NA comparisons return NA (SQL semantics). Use `is.na()` to filter NAs
#' explicitly.
#'
#' This is a streaming operation (constant memory per batch).
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> filter(cyl > 4) |> collect() |> head()
#' unlink(f)
#'
#' @export
filter <- function(.data, ...) {
  UseMethod("filter")
}

#' @export
filter.vectra_node <- function(.data, ...) {
  exprs <- eval(substitute(alist(...)))
  if (length(exprs) == 0) return(.data)
  schema <- .Call(C_node_schema, .data$.node)
  pred <- combine_predicates(exprs, parent.frame(), schema$name)
  new_xptr <- .Call(C_filter_node, .data$.node, pred)
  structure(list(.node = new_xptr, .path = .data$.path,
                 .groups = .data$.groups), class = "vectra_node")
}

#' Select columns from a vectra query
#'
#' @param .data A `vectra_node` object.
#' @param ... Column names (unquoted).
#'
#' @return A new `vectra_node` with only the selected columns.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> select(mpg, cyl) |> collect() |> head()
#' unlink(f)
#'
#' @export
select <- function(.data, ...) {
  UseMethod("select")
}

#' @export
select.vectra_node <- function(.data, ...) {
  schema <- .Call(C_node_schema, .data$.node)
  proxy <- schema_proxy(schema)

  sel <- tidyselect::eval_select(rlang::expr(c(...)), data = proxy)
  col_names <- schema$name
  out_names <- names(sel)

  n <- length(out_names)
  expr_lists <- vector("list", n)
  # If renamed (name differs from original), use col_ref
  orig_names <- unname(col_names[sel])
  for (i in seq_len(n)) {
    if (out_names[i] != orig_names[i]) {
      expr_lists[[i]] <- list(kind = "col_ref", name = orig_names[i])
    }
  }

  new_xptr <- .Call(C_project_node, .data$.node, out_names, expr_lists)
  # Drop group columns that were removed by select
  grps <- .data$.groups
  if (!is.null(grps)) {
    grps <- intersect(grps, out_names)
    if (length(grps) == 0) grps <- NULL
  }
  structure(list(.node = new_xptr, .path = .data$.path,
                 .groups = grps), class = "vectra_node")
}

#' Add or transform columns
#'
#' @param .data A `vectra_node` object.
#' @param ... Named expressions for new or transformed columns.
#'
#' @return A new `vectra_node` with mutated columns.
#'
#' @details
#' Supported expression types: arithmetic (`+`, `-`, `*`, `/`, `%%`),
#' comparison, boolean, `is.na()`, `nchar()`, `substr()`, `grepl()` (fixed
#' match only). Window functions (`row_number()`, `rank()`, `dense_rank()`,
#' `lag()`, `lead()`, `cumsum()`, `cummean()`, `cummin()`, `cummax()`) are
#' detected automatically and routed to a dedicated window node.
#'
#' When grouped, window functions respect partition boundaries.
#'
#' This is a streaming operation for regular expressions; window functions
#' materialize all rows within each partition.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> mutate(kpl = mpg * 0.425144) |> collect() |> head()
#' unlink(f)
#'
#' @export
mutate <- function(.data, ...) {
  UseMethod("mutate")
}

#' @export
mutate.vectra_node <- function(.data, ...) {
  dots <- eval(substitute(alist(...)))
  # Expand across() calls
  schema <- .Call(C_node_schema, .data$.node)
  proxy <- schema_proxy(schema)
  dots <- expand_across(dots, schema$name, parent.frame(), proxy)
  dot_names <- names(dots)
  if (is.null(dot_names) || any(dot_names == ""))
    stop("all mutate expressions must be named")

  # Split into window functions and regular expressions
  split <- split_window_exprs(dots)
  node <- .data

  # Apply window functions first (if any)
  if (length(split$win_specs) > 0) {
    node <- create_window_node(node, split$win_specs)
  }

  # Apply regular expressions (if any)
  if (length(split$regular_dots) > 0) {
    schema <- .Call(C_node_schema, node$.node)
    existing_names <- schema$name

    out_names <- character(0)
    out_exprs <- list()

    for (nm in existing_names) {
      out_names <- c(out_names, nm)
      out_exprs <- c(out_exprs, list(NULL))
    }

    for (i in seq_along(split$regular_dots)) {
      nm <- split$regular_names[i]
      expr_ser <- serialize_expr(split$regular_dots[[i]], parent.frame(), existing_names)
      idx <- match(nm, out_names)
      if (!is.na(idx)) {
        out_exprs[[idx]] <- expr_ser
      } else {
        out_names <- c(out_names, nm)
        out_exprs <- c(out_exprs, list(expr_ser))
      }
    }

    new_xptr <- .Call(C_project_node, node$.node, out_names, out_exprs)
    node <- structure(list(.node = new_xptr, .path = node$.path,
                           .groups = node$.groups), class = "vectra_node")
  }

  # If window node added columns that should not have been pass-through
  # but no regular exprs, just return the window node
  node
}

# -- Phase 1: compression round-trips ------------------------------------------

test_that("compress='fast' round-trips doubles", {
  df <- data.frame(x = rnorm(500), y = runif(500))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast")
  result <- tbl(f) |> collect()
  expect_equal(result$x, df$x)
  expect_equal(result$y, df$y)
})

test_that("compress='small' round-trips doubles", {
  df <- data.frame(x = rnorm(500), y = runif(500))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "small")
  result <- tbl(f) |> collect()
  expect_equal(result$x, df$x)
  expect_equal(result$y, df$y)
})

test_that("compress='small' round-trips integers", {
  df <- data.frame(a = 1:1000, b = sample(1e6, 1000))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "small")
  result <- tbl(f) |> collect()
  expect_equal(result$a, as.double(df$a))
  expect_equal(result$b, as.double(df$b))
})

test_that("compress='small' round-trips strings", {
  df <- data.frame(s = sample(letters, 500, replace = TRUE),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "small")
  result <- tbl(f) |> collect()
  expect_equal(result$s, df$s)
})

test_that("compress='small' round-trips booleans with NA", {
  df <- data.frame(flag = c(TRUE, FALSE, NA, TRUE, FALSE, rep(TRUE, 495)))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "small")
  result <- tbl(f) |> collect()
  expect_equal(result$flag, df$flag)
})

test_that("compress='small' round-trips all types together", {
  df <- data.frame(
    i = c(1L, 2L, NA, 4L, seq(5L, 500L)),
    d = c(1.1, NA, 3.3, 4.4, rnorm(496)),
    b = c(TRUE, NA, FALSE, TRUE, sample(c(TRUE, FALSE), 496, replace = TRUE)),
    s = c("a", NA, "c", "d", sample(letters, 496, replace = TRUE)),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "small")
  result <- tbl(f) |> collect()
  expect_equal(result$d, df$d)
  expect_equal(result$b, df$b)
  expect_equal(result$s, df$s)
  expect_equal(result$i, as.double(df$i))
})

test_that("small files are no larger than fast files on structured data", {
  df <- data.frame(x = rep(1:100, 50), y = rep(rnorm(100), 50))
  f_fast  <- tempfile(fileext = ".vtr")
  f_ratio <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f_fast, f_ratio)))
  write_vtr(df, f_fast,  compress = "fast")
  write_vtr(df, f_ratio, compress = "small")
  expect_lte(file.size(f_ratio), file.size(f_fast))
})

test_that("compress='small' with multiple row groups", {
  df <- data.frame(x = rnorm(1000), y = 1:1000)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "small", batch_size = 200)
  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 1000L)
  expect_equal(result$x, df$x)
  expect_equal(result$y, as.double(df$y))
})

test_that("small with quantize round-trips", {
  set.seed(42)
  df <- data.frame(temp = rnorm(1000, 15, 5))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "small",
            quantize = list(temp = list(precision = 0.01, type = "int16")))
  result <- tbl(f) |> collect()
  max_err <- max(abs(result$temp - df$temp))
  expect_lt(max_err, 0.01 + 1e-10)
})

test_that("small with spatial round-trips", {
  nx <- 50; ny <- 50
  set.seed(1)
  temp <- 10 + 0.05 * rep(1:nx, ny) + 0.03 * rep(1:ny, each = nx) +
          rnorm(nx * ny, 0, 0.5)
  df <- data.frame(temp = temp)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "small",
            quantize = list(temp = list(precision = 0.001, type = "int16")),
            spatial = list(nx = nx, ny = ny))
  result <- tbl(f) |> collect()
  max_err <- max(abs(result$temp - df$temp))
  expect_lt(max_err, 0.001 + 1e-10)
})

test_that("small with narrow int round-trips", {
  df <- data.frame(a = sample(-100L:100L, 500, replace = TRUE),
                   b = sample(-30000L:30000L, 500, replace = TRUE))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "small",
            col_types = c(a = "int8", b = "int16"))
  result <- tbl(f) |> collect()
  expect_equal(result$a, as.double(df$a))
  expect_equal(result$b, as.double(df$b))
})

test_that("compress='none' round-trips doubles", {
  df <- data.frame(x = rnorm(500), y = runif(500))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "none")
  result <- tbl(f) |> collect()
  expect_equal(result$x, df$x)
  expect_equal(result$y, df$y)
})

test_that("compress='fast' round-trips integers", {
  df <- data.frame(a = 1:1000, b = sample(1e6, 1000))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast")
  result <- tbl(f) |> collect()
  expect_equal(result$a, as.double(df$a))
  expect_equal(result$b, as.double(df$b))
})

test_that("compress='fast' round-trips strings", {
  df <- data.frame(s = sample(letters, 500, replace = TRUE),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast")
  result <- tbl(f) |> collect()
  expect_equal(result$s, df$s)
})

test_that("compress='fast' round-trips booleans with NA", {
  df <- data.frame(flag = c(TRUE, FALSE, NA, TRUE, FALSE, rep(TRUE, 495)))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast")
  result <- tbl(f) |> collect()
  expect_equal(result$flag, df$flag)
})

test_that("compress='fast' round-trips all types together", {
  df <- data.frame(
    i = c(1L, 2L, NA, 4L, seq(5L, 500L)),
    d = c(1.1, NA, 3.3, 4.4, rnorm(496)),
    b = c(TRUE, NA, FALSE, TRUE, sample(c(TRUE, FALSE), 496, replace = TRUE)),
    s = c("a", NA, "c", "d", sample(letters, 496, replace = TRUE)),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast")
  result <- tbl(f) |> collect()
  expect_equal(result$d, df$d)
  expect_equal(result$b, df$b)
  expect_equal(result$s, df$s)
  expect_equal(result$i, as.double(df$i))
})

test_that("compressed files are smaller than uncompressed", {
  df <- data.frame(x = rep(1:100, 50), y = rep(rnorm(100), 50))
  f_none <- tempfile(fileext = ".vtr")
  f_fast <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f_none, f_fast)))

  write_vtr(df, f_none, compress = "none")
  write_vtr(df, f_fast, compress = "fast")

  expect_lt(file.size(f_fast), file.size(f_none))
})

test_that("compress='fast' with multiple row groups", {
  df <- data.frame(x = rnorm(1000), y = 1:1000)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast", batch_size = 200)
  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 1000L)
  expect_equal(result$x, df$x)
  expect_equal(result$y, as.double(df$y))
})

test_that("uncompressed file reads after rewrite with compression", {
  df <- data.frame(x = 1:100, y = rnorm(100))
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))

  write_vtr(df, f1, compress = "none")
  write_vtr(df, f2, compress = "fast")

  r1 <- tbl(f1) |> collect()
  r2 <- tbl(f2) |> collect()
  expect_equal(r1$x, r2$x)
  expect_equal(r1$y, r2$y)
})

# -- Phase 2a: narrow integer types -------------------------------------------

test_that("col_types int8 round-trips", {
  df <- data.frame(val = c(-100L, 0L, 50L, 127L, NA))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, col_types = c(val = "int8"))
  result <- tbl(f) |> collect()
  expect_equal(result$val, as.double(df$val))
})

test_that("col_types int16 round-trips", {
  df <- data.frame(val = c(-30000L, 0L, 15000L, 32767L, NA))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, col_types = c(val = "int16"))
  result <- tbl(f) |> collect()
  expect_equal(result$val, as.double(df$val))
})

test_that("col_types int32 round-trips", {
  df <- data.frame(val = c(-2000000L, 0L, 1000000L, NA))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, col_types = c(val = "int32"))
  result <- tbl(f) |> collect()
  expect_equal(result$val, as.double(df$val))
})

test_that("narrow int with compression round-trips", {
  df <- data.frame(a = sample(-100L:100L, 500, replace = TRUE),
                   b = sample(-30000L:30000L, 500, replace = TRUE))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast",
            col_types = c(a = "int8", b = "int16"))
  result <- tbl(f) |> collect()
  expect_equal(result$a, as.double(df$a))
  expect_equal(result$b, as.double(df$b))
})

test_that("narrow int files are smaller", {
  df <- data.frame(val = sample(-100L:100L, 1000, replace = TRUE))
  f64 <- tempfile(fileext = ".vtr")
  f8  <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f64, f8)))
  write_vtr(df, f64, compress = "none")
  write_vtr(df, f8, compress = "none", col_types = c(val = "int8"))
  expect_lt(file.size(f8), file.size(f64))
})

test_that("filter works on narrow int columns", {
  df <- data.frame(id = 1:200, val = sample(-100L:100L, 200, replace = TRUE))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, col_types = c(val = "int8"))
  result <- tbl(f) |> filter(val > 0) |> collect()
  expected <- df[df$val > 0, ]
  expect_equal(nrow(result), nrow(expected))
  expect_equal(result$val, as.double(expected$val))
})

test_that("mutate works on narrow int columns", {
  df <- data.frame(val = c(10L, 20L, 30L))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, col_types = c(val = "int8"))
  result <- tbl(f) |> mutate(doubled = val * 2) |> collect()
  expect_equal(result$doubled, c(20, 40, 60))
})

# -- Phase 2b: lossy quantization ---------------------------------------------

test_that("quantize round-trips within precision", {
  set.seed(42)
  df <- data.frame(temp = runif(500, -20, 20))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, quantize = list(temp = list(precision = 0.001,
                                                type = "int16")))
  result <- tbl(f) |> collect()
  expect_equal(length(result$temp), 500L)
  max_err <- max(abs(result$temp - df$temp), na.rm = TRUE)
  expect_lt(max_err, 0.001 + 1e-10)
})

test_that("quantize with scale parameter", {
  df <- data.frame(val = c(0.0, 0.5, 1.0, -1.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, quantize = list(val = list(scale = 1000, type = "int16")))
  result <- tbl(f) |> collect()
  max_err <- max(abs(result$val - df$val))
  expect_lt(max_err, 0.001 + 1e-10)
})

test_that("quantize preserves NA", {
  df <- data.frame(temp = c(1.5, NA, -3.2, NA, 10.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, quantize = list(temp = list(precision = 0.01,
                                                type = "int16")))
  result <- tbl(f) |> collect()
  expect_true(is.na(result$temp[2]))
  expect_true(is.na(result$temp[4]))
  non_na <- c(1, 3, 5)
  max_err <- max(abs(result$temp[non_na] - df$temp[non_na]))
  expect_lt(max_err, 0.01 + 1e-10)
})

test_that("quantize with compression round-trips", {
  set.seed(123)
  df <- data.frame(temp = rnorm(1000, 15, 5))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast",
            quantize = list(temp = list(precision = 0.01, type = "int16")))
  result <- tbl(f) |> collect()
  max_err <- max(abs(result$temp - df$temp))
  expect_lt(max_err, 0.01 + 1e-10)
})

test_that("quantize to int8", {
  df <- data.frame(val = c(-5.0, 0.0, 5.0, 12.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, quantize = list(val = list(precision = 0.1,
                                               type = "int8")))
  result <- tbl(f) |> collect()
  max_err <- max(abs(result$val - df$val))
  expect_lt(max_err, 0.1 + 1e-10)
})

test_that("quantize produces smaller files", {
  set.seed(99)
  df <- data.frame(temp = rnorm(5000, 15, 5))
  f_raw <- tempfile(fileext = ".vtr")
  f_q   <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f_raw, f_q)))
  write_vtr(df, f_raw, compress = "fast")
  write_vtr(df, f_q, compress = "fast",
            quantize = list(temp = list(precision = 0.01, type = "int16")))
  expect_lt(file.size(f_q), file.size(f_raw))
})

# -- Phase 2c: spatial encoding -----------------------------------------------

test_that("spatial round-trips on smooth raster data", {
  nx <- 50; ny <- 50
  x_coord <- rep(1:nx, ny)
  y_coord <- rep(1:ny, each = nx)
  # smooth temperature field: gradient + small noise
  set.seed(1)
  temp <- 10 + 0.05 * x_coord + 0.03 * y_coord + rnorm(nx * ny, 0, 0.5)
  df <- data.frame(temp = temp)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast",
            quantize = list(temp = list(precision = 0.001, type = "int16")),
            spatial = list(nx = nx, ny = ny))
  result <- tbl(f) |> collect()
  max_err <- max(abs(result$temp - df$temp))
  expect_lt(max_err, 0.001 + 1e-10)
})

test_that("spatial without quantize round-trips integers", {
  nx <- 20; ny <- 20
  vals <- as.integer(rep(1:20, each = 20) + rep(1:20, 20))
  df <- data.frame(val = vals)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast",
            spatial = list(nx = nx, ny = ny))
  result <- tbl(f) |> collect()
  expect_equal(result$val, as.double(df$val))
})

test_that("spatial produces smaller files on smooth data", {
  nx <- 100; ny <- 100
  set.seed(2)
  temp <- 10 + 0.05 * rep(1:nx, ny) + 0.03 * rep(1:ny, each = nx)
  df <- data.frame(temp = temp)

  f_plain <- tempfile(fileext = ".vtr")
  f_spatial <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f_plain, f_spatial)))

  write_vtr(df, f_plain, compress = "fast",
            quantize = list(temp = list(precision = 0.001, type = "int16")))
  write_vtr(df, f_spatial, compress = "fast",
            quantize = list(temp = list(precision = 0.001, type = "int16")),
            spatial = list(nx = nx, ny = ny))

  expect_lt(file.size(f_spatial), file.size(f_plain))
})

test_that("spatial + quantize + fast compression round-trips", {
  nx <- 30; ny <- 30
  set.seed(3)
  temp <- 15 + 0.1 * rep(1:nx, ny) + rnorm(nx * ny, 0, 0.2)
  df <- data.frame(temp = temp)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast",
            quantize = list(temp = list(precision = 0.001, type = "int16")),
            spatial = list(nx = nx, ny = ny))
  result <- tbl(f) |> collect()
  max_err <- max(abs(result$temp - df$temp))
  expect_lt(max_err, 0.001 + 1e-10)
})

test_that("spatial requires nx*ny == n_rows", {
  nx <- 40; ny <- 40
  df <- data.frame(temp = rnorm(100))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  expect_error(
    write_vtr(df, f, compress = "fast",
              spatial = list(nx = nx, ny = ny)),
    "nx\\*ny"
  )
})

# -- DIFF encoding (auto-selected for slowly varying data) --------------------

test_that("slowly varying integer data round-trips (triggers DIFF encoding)", {
  # Monotonically increasing with small increments - should trigger DIFF
  df <- data.frame(ts = cumsum(sample(1:5, 1000, replace = TRUE)))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast")
  result <- tbl(f) |> collect()
  expect_equal(result$ts, as.double(df$ts))
})

test_that("slowly varying doubles round-trip (may trigger DIFF)", {
  set.seed(5)
  df <- data.frame(val = cumsum(rnorm(1000, 0, 0.01)))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast")
  result <- tbl(f) |> collect()
  expect_equal(result$val, df$val)
})

# -- Combined pipeline: narrow + compress + filter ----------------------------

test_that("full pipeline: write compressed narrow int, then filter + collect", {
  df <- data.frame(
    id = 1:1000,
    category = sample(c(-1L, 0L, 1L), 1000, replace = TRUE),
    value = rnorm(1000)
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast",
            col_types = c(category = "int8"))
  result <- tbl(f) |> filter(category == 1) |> collect()
  expected <- df[df$category == 1L, ]
  expect_equal(nrow(result), nrow(expected))
  expect_equal(result$value, expected$value)
})

test_that("full pipeline: quantize + compress + summarise", {
  set.seed(10)
  df <- data.frame(
    grp = rep(c("A", "B"), each = 500),
    val = c(rnorm(500, 10, 2), rnorm(500, 20, 2))
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, compress = "fast",
            quantize = list(val = list(precision = 0.01, type = "int16")))

  result <- tbl(f) |>
    group_by(grp) |>
    summarise(avg = mean(val)) |>
    collect()

  # Means should be close to 10 and 20 (within quantization + sampling error)
  result <- result[order(result$grp), ]
  expect_equal(nrow(result), 2L)
  expect_lt(abs(result$avg[1] - 10), 0.5)
  expect_lt(abs(result$avg[2] - 20), 0.5)
})

test_that("mutate adds a new column", {
  df <- data.frame(x = c(1.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(y = x * 2) |> collect()
  expect_equal(result$x, c(1, 2, 3))
  expect_equal(result$y, c(2, 4, 6))
})

test_that("mutate with arithmetic expressions", {
  df <- data.frame(a = c(10.0, 20.0), b = c(3.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(c = a + b, d = a - b) |> collect()
  expect_equal(result$c, c(13, 24))
  expect_equal(result$d, c(7, 16))
})

test_that("mutate sequential: later expr references earlier", {
  df <- data.frame(x = c(1.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(y = x + 1, z = y * 2) |> collect()
  expect_equal(result$y, c(2, 3, 4))
  expect_equal(result$z, c(4, 6, 8))
})

test_that("mutate replaces existing column", {
  df <- data.frame(x = c(1.0, 2.0, 3.0), y = c(10.0, 20.0, 30.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(x = x * 100) |> collect()
  expect_equal(result$x, c(100, 200, 300))
  expect_equal(result$y, c(10, 20, 30))
})

test_that("mutate with NA propagation", {
  df <- data.frame(x = c(1.0, NA, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(y = x + 1) |> collect()
  expect_equal(result$y, c(2, NA, 4))
})

test_that("filter then select pipeline", {
  df <- data.frame(x = 1:10, y = as.double(11:20),
                   z = c(rep("a", 5), rep("b", 5)),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(x > 5) |> select(x, z) |> collect()
  expect_equal(names(result), c("x", "z"))
  expect_equal(nrow(result), 5)
  expect_equal(result$z, rep("b", 5))
})

test_that("tbl_xlsx reads basic xlsx and returns correct data.frame", {
  skip_if_not_installed("openxlsx2")

  f <- tempfile(fileext = ".xlsx")
  on.exit(unlink(f))
  df <- data.frame(x = 1:5, y = c(10.1, 20.2, 30.3, 40.4, 50.5),
                   name = c("a", "b", "c", "d", "e"),
                   stringsAsFactors = FALSE)
  openxlsx2::write_xlsx(df, f)

  result <- tbl_xlsx(f) |> collect()
  expect_equal(nrow(result), 5)
  expect_equal(ncol(result), 3)
  expect_equal(names(result), c("x", "y", "name"))
  expect_equal(result$name, c("a", "b", "c", "d", "e"))
  expect_equal(result$y, c(10.1, 20.2, 30.3, 40.4, 50.5))
})

test_that("tbl_xlsx selects sheet by index", {
  skip_if_not_installed("openxlsx2")

  f <- tempfile(fileext = ".xlsx")
  on.exit(unlink(f))
  wb <- openxlsx2::wb_workbook()
  wb$add_worksheet("First")
  wb$add_data(sheet = "First", x = data.frame(a = 1:3))
  wb$add_worksheet("Second")
  wb$add_data(sheet = "Second", x = data.frame(b = 4:6))
  wb$save(f)

  result1 <- tbl_xlsx(f, sheet = 1L) |> collect()
  result2 <- tbl_xlsx(f, sheet = 2L) |> collect()

  expect_equal(names(result1), "a")
  expect_equal(names(result2), "b")
  expect_equal(result2$b, c(4, 5, 6))
})

test_that("tbl_xlsx selects sheet by name", {
  skip_if_not_installed("openxlsx2")

  f <- tempfile(fileext = ".xlsx")
  on.exit(unlink(f))
  wb <- openxlsx2::wb_workbook()
  wb$add_worksheet("Alpha")
  wb$add_data(sheet = "Alpha", x = data.frame(val = 10:12))
  wb$add_worksheet("Beta")
  wb$add_data(sheet = "Beta", x = data.frame(val = 20:22))
  wb$save(f)

  result <- tbl_xlsx(f, sheet = "Beta") |> collect()
  expect_equal(result$val, c(20, 21, 22))
})

test_that("tbl_xlsx errors on missing file", {
  skip_if_not_installed("openxlsx2")
  expect_error(tbl_xlsx(tempfile(fileext = ".xlsx")))
})

test_that("tbl_xlsx errors when openxlsx2 not available", {
  # Mock requireNamespace to return FALSE
  mockr <- function(...) FALSE
  with_mocked_bindings(
    {
      f <- tempfile(fileext = ".xlsx")
      expect_error(tbl_xlsx(f), "openxlsx2")
    },
    requireNamespace = mockr,
    .package = "base"
  )
})

test_that("tbl_xlsx validates path argument", {
  skip_if_not_installed("openxlsx2")
  expect_error(tbl_xlsx(123), "path must be a single character string")
  expect_error(tbl_xlsx(c("a.xlsx", "b.xlsx")),
               "path must be a single character string")
})

test_that("tbl_xlsx validates sheet argument", {
  skip_if_not_installed("openxlsx2")

  f <- tempfile(fileext = ".xlsx")
  on.exit(unlink(f))
  openxlsx2::write_xlsx(data.frame(x = 1), f)

  expect_error(tbl_xlsx(f, sheet = TRUE),
               "sheet must be a single character string or integer")
  expect_error(tbl_xlsx(f, sheet = c(1L, 2L)),
               "sheet must be a single character string or integer")
})

test_that("tbl_xlsx works with filter pipeline", {
  skip_if_not_installed("openxlsx2")

  f <- tempfile(fileext = ".xlsx")
  on.exit(unlink(f))
  df <- data.frame(x = 1:10, y = c(rep("a", 5), rep("b", 5)),
                   stringsAsFactors = FALSE)
  openxlsx2::write_xlsx(df, f)

  result <- tbl_xlsx(f) |>
    filter(x > 7) |>
    collect()

  expect_equal(nrow(result), 3)
  expect_equal(result$x, c(8, 9, 10))
})

test_that("tbl_csv reads basic CSV and returns correct data.frame", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  write.csv(mtcars[1:10, ], f, row.names = FALSE)

  result <- tbl_csv(f) |> collect()
  expect_equal(nrow(result), 10)
  expect_equal(ncol(result), ncol(mtcars))
  expect_equal(names(result), names(mtcars))
  expect_equal(result$mpg, mtcars$mpg[1:10])
})

test_that("tbl_csv infers integer columns correctly", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  df <- data.frame(x = 1:5, y = c(1.5, 2.5, 3.5, 4.5, 5.5))
  write.csv(df, f, row.names = FALSE)

  result <- tbl_csv(f) |> collect()
  # x should be int64 (returned as double by default)
  expect_type(result$x, "double")
  expect_equal(result$x, c(1, 2, 3, 4, 5))
  expect_equal(result$y, df$y)
})

test_that("tbl_csv handles string columns", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  df <- data.frame(name = c("Alice", "Bob", "Charlie"), score = c(90, 85, 95),
                   stringsAsFactors = FALSE)
  write.csv(df, f, row.names = FALSE)

  result <- tbl_csv(f) |> collect()
  expect_equal(result$name, c("Alice", "Bob", "Charlie"))
  expect_equal(result$score, c(90, 85, 95))
})

test_that("tbl_csv handles NA values", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  writeLines("x,y,z\n1,hello,TRUE\nNA,NA,NA\n3,world,FALSE", f)

  result <- tbl_csv(f) |> collect()
  expect_equal(result$x[1], 1)
  expect_true(is.na(result$x[2]))
  expect_equal(result$x[3], 3)
  expect_true(is.na(result$y[2]))
  expect_true(is.na(result$z[2]))
})

test_that("tbl_csv handles logical columns", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  writeLines("a,b\nTRUE,1\nFALSE,2\nTRUE,3", f)

  result <- tbl_csv(f) |> collect()
  expect_equal(result$a, c(TRUE, FALSE, TRUE))
})

test_that("tbl_csv handles quoted fields with commas", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  writeLines('name,city\n"Smith, John","New York"\n"Doe, Jane",Boston', f)

  result <- tbl_csv(f) |> collect()
  expect_equal(result$name, c("Smith, John", "Doe, Jane"))
  expect_equal(result$city, c("New York", "Boston"))
})

test_that("tbl_csv works with filter pipeline", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  write.csv(mtcars, f, row.names = FALSE)

  result <- tbl_csv(f) |>
    filter(cyl == 6) |>
    select(mpg, cyl, hp) |>
    collect()

  expected <- mtcars[mtcars$cyl == 6, c("mpg", "cyl", "hp")]
  expect_equal(nrow(result), nrow(expected))
  expect_equal(result$mpg, expected$mpg)
})

test_that("tbl_csv works with group_by + summarise", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  write.csv(mtcars, f, row.names = FALSE)

  result <- tbl_csv(f) |>
    group_by(cyl) |>
    summarise(mean_mpg = mean(mpg)) |>
    arrange(cyl) |>
    collect()

  expect_equal(nrow(result), 3)
  expect_equal(result$cyl, c(4, 6, 8))
})

test_that("tbl_csv respects batch_size parameter", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  write.csv(mtcars, f, row.names = FALSE)

  # Small batch size should still give correct results
  result <- tbl_csv(f, batch_size = 5L) |> collect()
  expect_equal(nrow(result), nrow(mtcars))
  expect_equal(result$mpg, mtcars$mpg)
})

test_that("tbl_csv handles empty fields as NA", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  writeLines("x,y\n1,hello\n2,\n3,world", f)

  result <- tbl_csv(f) |> collect()
  expect_true(is.na(result$y[2]))
  expect_equal(result$y[c(1, 3)], c("hello", "world"))
})

test_that("tbl_csv explain shows CsvScanNode", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  write.csv(mtcars[1:5, ], f, row.names = FALSE)

  node <- tbl_csv(f)
  plan <- capture.output(explain(node))
  expect_true(any(grepl("CsvScanNode", plan)))
})

test_that("tbl_csv pipe to write_csv round-trips", {
  f_in <- tempfile(fileext = ".csv")
  f_out <- tempfile(fileext = ".csv")
  on.exit(unlink(c(f_in, f_out)))

  df <- data.frame(a = 1:3, b = c("x", "y", "z"), stringsAsFactors = FALSE)
  write.csv(df, f_in, row.names = FALSE)

  tbl_csv(f_in) |> write_csv(f_out)

  result <- tbl_csv(f_out) |> collect()
  expect_equal(result$a, c(1, 2, 3))
  expect_equal(result$b, c("x", "y", "z"))
})

test_that("tbl_csv reads gzip-compressed CSV via the miniz path", {
  f <- tempfile(fileext = ".csv.gz")
  on.exit(unlink(f))

  df <- data.frame(
    a = 1:5,
    b = c(1.5, 2.5, 3.5, 4.5, 5.5),
    c = c("foo", "bar", "baz", "qux", "quux"),
    stringsAsFactors = FALSE
  )
  con <- gzfile(f, "w")
  write.csv(df, con, row.names = FALSE)
  close(con)

  result <- tbl_csv(f) |> collect()
  expect_equal(nrow(result), 5)
  expect_equal(names(result), c("a", "b", "c"))
  expect_equal(result$a, c(1, 2, 3, 4, 5))
  expect_equal(result$b, df$b)
  expect_equal(result$c, df$c)
})

test_that("tbl_csv on a gz CSV exercises the type-inference rewind path", {
  # csv_scan infers types from the first 1000 rows, then seeks back to the
  # data start. For the gz path that means seeking inside the in-memory
  # decompressed buffer. Use >1000 rows so the rewind actually fires.
  f <- tempfile(fileext = ".csv.gz")
  on.exit(unlink(f))

  n <- 1500
  df <- data.frame(
    id = seq_len(n),
    val = as.double(seq_len(n)) / 2,
    tag = sprintf("row%04d", seq_len(n)),
    stringsAsFactors = FALSE
  )
  con <- gzfile(f, "w")
  write.csv(df, con, row.names = FALSE)
  close(con)

  result <- tbl_csv(f) |> collect()
  expect_equal(nrow(result), n)
  expect_equal(result$id[1], 1)
  expect_equal(result$id[n], n)
  expect_equal(result$val[n], n / 2)
  expect_equal(result$tag[1], "row0001")
  expect_equal(result$tag[n], sprintf("row%04d", n))
})

# Streaming consumption: collect_chunked() fold and chunk_feeder() generator.
# These verbs must agree exactly with the materializing path (collect()) and
# with in-memory model fits, while only ever holding one batch at a time.

test_that("collect_chunked row count matches collect()", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(mtcars, f)

  n <- collect_chunked(tbl(f), function(acc, chunk) acc + nrow(chunk),
                       .init = 0L)
  expect_equal(n, nrow(mtcars))
})

test_that("collect_chunked actually streams in multiple batches", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  df <- data.frame(x = as.double(1:100), g = rep(1:4, 25))
  write.csv(df, f, row.names = FALSE)

  # Small batch_size forces several chunks; the cursor must keep its place
  # across separate C_node_next_batch calls.
  n_batches <- collect_chunked(tbl_csv(f, batch_size = 10),
                               function(acc, chunk) acc + 1L, .init = 0L)
  expect_gt(n_batches, 1L)

  total <- collect_chunked(tbl_csv(f, batch_size = 10),
                           function(acc, chunk) acc + sum(chunk$x), .init = 0)
  expect_equal(total, sum(df$x))
})

test_that("collect_chunked column sums match collect() across batches", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  set.seed(1)
  df <- data.frame(a = rnorm(250), b = rnorm(250))
  write.csv(df, f, row.names = FALSE)

  streamed <- collect_chunked(
    tbl_csv(f, batch_size = 32),
    function(acc, chunk) {
      acc$a <- acc$a + sum(chunk$a)
      acc$b <- acc$b + sum(chunk$b)
      acc
    },
    .init = list(a = 0, b = 0)
  )
  full <- tbl_csv(f) |> collect()
  expect_equal(streamed$a, sum(full$a))
  expect_equal(streamed$b, sum(full$b))
})

test_that("collect_chunked handles filtered (selection-vector) batches", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  df <- data.frame(x = as.double(1:200))
  write.csv(df, f, row.names = FALSE)

  kept <- collect_chunked(
    tbl_csv(f, batch_size = 16) |> filter(x > 150),
    function(acc, chunk) acc + nrow(chunk), .init = 0L
  )
  expect_equal(kept, sum(df$x > 150))
})

test_that("collect_chunked returns .init unchanged when no rows match", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(mtcars, f)

  out <- collect_chunked(
    tbl(f) |> filter(mpg > 1e6),
    function(acc, chunk) acc + nrow(chunk), .init = 0L
  )
  expect_identical(out, 0L)
})

test_that("collect_chunked sufficient statistics reproduce lm() exactly", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  set.seed(42)
  df <- data.frame(wt = rnorm(500, 3, 1), hp = rnorm(500, 150, 40))
  df$mpg <- 30 - 5 * df$wt - 0.05 * df$hp + rnorm(500, 0, 1)
  write.csv(df, f, row.names = FALSE)

  acc <- collect_chunked(
    tbl_csv(f, batch_size = 64) |> select(mpg, wt, hp),
    function(acc, chunk) {
      X <- cbind(1, chunk$wt, chunk$hp)
      y <- chunk$mpg
      list(XtX = acc$XtX + crossprod(X), Xty = acc$Xty + crossprod(X, y))
    },
    .init = list(XtX = matrix(0, 3, 3), Xty = matrix(0, 3, 1))
  )
  beta_stream <- drop(solve(acc$XtX, acc$Xty))
  beta_lm <- unname(coef(lm(mpg ~ wt + hp, data = df)))
  expect_equal(beta_stream, beta_lm, tolerance = 1e-8)
})

test_that("collect_chunked validates its arguments", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(mtcars, f)

  expect_error(collect_chunked(mtcars, function(a, c) a), "vectra_node")
  expect_error(collect_chunked(tbl(f), "not a function"), "function")
})

test_that("collect_chunked preserves string, factor, and Date columns", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  df <- data.frame(
    s = rep(c("alpha", "beta", "gamma"), length.out = 120),
    g = factor(rep(c("lo", "hi"), length.out = 120), levels = c("lo", "hi")),
    d = as.Date("2020-01-01") + 0:119,
    v = as.double(1:120),
    stringsAsFactors = FALSE
  )
  write_vtr(df, f, batch_size = 25)

  # Rebuild each column by concatenating the per-batch chunks, then compare to
  # the materializing path. Exercises batch_to_dataframe's string fill and the
  # annotation (factor/Date) reconstruction one batch at a time.
  parts <- collect_chunked(tbl(f), function(acc, chunk) c(acc, list(chunk)),
                           .init = list())
  expect_gt(length(parts), 1L)
  rebuilt <- do.call(rbind, parts)
  full <- tbl(f) |> collect()

  expect_equal(rebuilt$s, full$s)
  expect_identical(levels(rebuilt$g), levels(full$g))
  expect_equal(as.character(rebuilt$g), as.character(full$g))
  expect_s3_class(rebuilt$d, "Date")
  expect_equal(rebuilt$d, full$d)
  expect_equal(rebuilt$v, full$v)
})

test_that("chunk_feeder replays the full stream on every reset", {
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  df <- data.frame(x = as.double(1:120))
  write.csv(df, f, row.names = FALSE)

  src <- function() tbl_csv(f, batch_size = 25)
  feed <- chunk_feeder(src)

  drain <- function() {
    feed(reset = TRUE)
    rows <- 0L
    while (!is.null(chunk <- feed(reset = FALSE))) rows <- rows + nrow(chunk)
    rows
  }

  # Two independent passes must each see every row.
  expect_equal(drain(), nrow(df))
  expect_equal(drain(), nrow(df))
})

test_that("chunk_feeder validates its argument", {
  expect_error(chunk_feeder(42), "function")
})

test_that("chunk_feeder drives an out-of-core Gaussian GLM equal to lm()", {
  skip_if_not_installed("biglm")
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  set.seed(7)
  df <- data.frame(wt = rnorm(800, 3, 1), hp = rnorm(800, 150, 40))
  df$mpg <- 30 - 5 * df$wt - 0.05 * df$hp + rnorm(800, 0, 1)
  write.csv(df, f, row.names = FALSE)

  src <- function() tbl_csv(f, batch_size = 128) |> select(mpg, wt, hp)
  fit <- biglm::bigglm(mpg ~ wt + hp, data = chunk_feeder(src),
                       family = gaussian())
  expect_equal(unname(coef(fit)),
               unname(coef(lm(mpg ~ wt + hp, data = df))),
               tolerance = 1e-6)
})

test_that("chunk_feeder drives an out-of-core logistic GLM equal to glm()", {
  skip_if_not_installed("biglm")
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  set.seed(11)
  n <- 2000
  bio1 <- rnorm(n)
  bio12 <- rnorm(n)
  eta <- -0.3 + 1.2 * bio1 - 0.8 * bio12
  presence <- rbinom(n, 1, plogis(eta))
  df <- data.frame(presence = as.double(presence), bio1 = bio1, bio12 = bio12)
  write.csv(df, f, row.names = FALSE)

  src <- function() tbl_csv(f, batch_size = 256) |>
    select(presence, bio1, bio12)
  fit <- biglm::bigglm(presence ~ bio1 + bio12, data = chunk_feeder(src),
                       family = binomial(), maxit = 25)

  ref <- glm(presence ~ bio1 + bio12, data = df, family = binomial())
  expect_equal(unname(coef(fit)), unname(coef(ref)), tolerance = 1e-3)
})

test_that("filter with simple comparison", {
  df <- data.frame(x = c(1L, 2L, 3L, 4L, 5L), y = c(10.0, 20.0, 30.0, 40.0, 50.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(x > 3) |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(result$x, c(4, 5))
  expect_equal(result$y, c(40, 50))
})

test_that("filter with equality", {
  df <- data.frame(x = c(1L, 2L, 3L), s = c("a", "b", "c"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(x == 2) |> collect()
  expect_equal(nrow(result), 1)
  expect_equal(result$s, "b")
})

test_that("filter with AND (multiple args)", {
  df <- data.frame(x = 1:10, y = rep(c(TRUE, FALSE), 5))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(x > 3, x < 8) |> collect()
  expect_equal(result$x, c(4, 5, 6, 7))
})

test_that("filter with boolean operators", {
  df <- data.frame(x = 1:6)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(x < 3 | x > 4) |> collect()
  expect_equal(result$x, c(1, 2, 5, 6))
})

test_that("filter with NOT", {
  df <- data.frame(x = c(TRUE, FALSE, TRUE, FALSE))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(!x) |> collect()
  expect_equal(result$x, c(FALSE, FALSE))
})

test_that("filter with is.na", {
  df <- data.frame(x = c(1L, NA, 3L, NA, 5L))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(!is.na(x)) |> collect()
  expect_equal(result$x, c(1, 3, 5))
})

test_that("filter with NAs in predicate (treated as FALSE)", {
  df <- data.frame(x = c(1L, NA, 3L), y = c(10.0, 20.0, 30.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(x > 1) |> collect()
  # NA > 1 is NA, treated as FALSE
  expect_equal(nrow(result), 1)
  expect_equal(result$x, 3)
})

# ---- Selection vector tests ----

test_that("filter -> collect with selection vector", {
  df <- data.frame(
    x = c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
    y = c("a", "b", "c", "d", "e", "f", "g", "h", "i", "j"),
    z = c(TRUE, FALSE, TRUE, FALSE, TRUE, FALSE, TRUE, FALSE, TRUE, FALSE),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(x > 5) |> collect()
  expect_equal(nrow(result), 5)
  expect_equal(result$x, c(6, 7, 8, 9, 10))
  expect_equal(result$y, c("f", "g", "h", "i", "j"))
  expect_equal(result$z, c(FALSE, TRUE, FALSE, TRUE, FALSE))
})

test_that("double filter composes selection vectors", {
  df <- data.frame(x = c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(x > 3) |> filter(x < 8) |> collect()
  expect_equal(result$x, c(4, 5, 6, 7))
})

test_that("filter -> select with selection vector", {
  df <- data.frame(
    x = c(1, 2, 3, 4, 5),
    y = c(10, 20, 30, 40, 50),
    z = c("a", "b", "c", "d", "e"),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(x >= 3) |> select(x, z) |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(result$x, c(3, 4, 5))
  expect_equal(result$z, c("c", "d", "e"))
  expect_equal(ncol(result), 2)
})

test_that("filter -> mutate with selection vector", {
  df <- data.frame(x = c(1, 2, 3, 4, 5), y = c(10, 20, 30, 40, 50))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(x > 2) |> mutate(z = x + y) |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(result$x, c(3, 4, 5))
  expect_equal(result$z, c(33, 44, 55))
})

test_that("filter -> inner_join with selection vector on probe side", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))

  write_vtr(data.frame(id = c(1, 2, 3, 4, 5), x = c(10, 20, 30, 40, 50)), f1)
  write_vtr(data.frame(id = c(2, 4, 6), y = c(200, 400, 600)), f2)

  result <- tbl(f1) |> filter(id > 1) |>
    inner_join(tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(result$id, c(2, 4))
  expect_equal(result$x, c(20, 40))
  expect_equal(result$y, c(200, 400))
})

test_that("filter -> left_join with selection vector", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))

  write_vtr(data.frame(id = c(1, 2, 3, 4, 5), x = c(10, 20, 30, 40, 50)), f1)
  write_vtr(data.frame(id = c(2, 4), y = c(200, 400)), f2)

  result <- tbl(f1) |> filter(id >= 2, id <= 4) |>
    left_join(tbl(f2), by = "id") |> collect()
  expect_equal(nrow(result), 3)
  # Sort by id for comparison (hash join order: matched first, then unmatched)
  result <- result[order(result$id), ]
  expect_equal(result$id, c(2, 3, 4))
  expect_equal(result$x, c(20, 30, 40))
  expect_equal(result$y, c(200, NA, 400))
})

test_that("empty filter result propagates correctly", {
  df <- data.frame(x = c(1, 2, 3))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(x > 100) |> collect()
  expect_equal(nrow(result), 0)
})

# ---- Selection vector v1.5 tests (ProjectNode gather) ----

test_that("filter -> select on wide table gathers only needed cols", {
  # Wide table: 20 columns, selective filter, narrow select
  cols <- lapply(1:20, function(i) seq(1, 100))
  names(cols) <- paste0("col", 1:20)
  df <- as.data.frame(cols)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(col1 > 95) |> select(col1, col20) |> collect()
  expect_equal(nrow(result), 5)
  expect_equal(ncol(result), 2)
  expect_equal(result$col1, 96:100)
  expect_equal(result$col20, 96:100)
})

test_that("filter -> mutate with expr referencing one col", {
  df <- data.frame(a = c(1, 2, 3, 4, 5), b = c(10, 20, 30, 40, 50))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(a > 2) |> mutate(c = a * 2) |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(result$a, c(3, 4, 5))
  expect_equal(result$b, c(30, 40, 50))
  expect_equal(result$c, c(6, 8, 10))
})

test_that("sequential mutate with selection: a2 = a+1, a3 = a2+1", {
  df <- data.frame(a = c(1, 2, 3, 4, 5))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(a > 2) |>
    mutate(a2 = a + 1, a3 = a2 + 1) |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(result$a, c(3, 4, 5))
  expect_equal(result$a2, c(4, 5, 6))
  expect_equal(result$a3, c(5, 6, 7))
})

test_that("double filter then mutate", {
  df <- data.frame(x = c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(x > 3) |> filter(x < 8) |>
    mutate(y = x * 10) |> collect()
  expect_equal(result$x, c(4, 5, 6, 7))
  expect_equal(result$y, c(40, 50, 60, 70))
})

test_that("filter -> mutate -> left_join correctness", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))

  write_vtr(data.frame(id = c(1, 2, 3, 4, 5), x = c(10, 20, 30, 40, 50)), f1)
  write_vtr(data.frame(id = c(2, 4, 6), y = c(200, 400, 600)), f2)

  result <- tbl(f1) |> filter(id > 1) |> mutate(x2 = x * 2) |>
    left_join(tbl(f2), by = "id") |> collect()
  result <- result[order(result$id), ]
  expect_equal(result$id, c(2, 3, 4, 5))
  expect_equal(result$x, c(20, 30, 40, 50))
  expect_equal(result$x2, c(40, 60, 80, 100))
  expect_equal(result$y, c(200, NA, 400, NA))
})

test_that("filter -> group_by -> summarise with selection vector (no compact)", {
  df <- data.frame(
    g = c("a", "a", "b", "b", "c", "c"),
    x = c(1, 2, 3, 4, 5, 6),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(x > 2) |>
    group_by(g) |> summarise(total = sum(x)) |> collect()
  # After filter: b(3,4), c(5,6)
  expect_equal(nrow(result), 2)
  expect_true("b" %in% result$g)
  expect_true("c" %in% result$g)
  expect_equal(result$total[result$g == "b"], 7)
  expect_equal(result$total[result$g == "c"], 11)
})

test_that("filter -> group_by with multiple aggs and sel", {
  df <- data.frame(
    g = c("x", "x", "x", "y", "y", "y", "z", "z", "z"),
    val = c(1, 2, 3, 4, 5, 6, 7, 8, 9)
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(val > 1, val < 9) |>
    group_by(g) |>
    summarise(n = n(), s = sum(val), m = mean(val)) |> collect()
  result <- result[order(result$g), ]
  expect_equal(result$g, c("x", "y", "z"))
  expect_equal(result$n, c(2, 3, 2))
  expect_equal(result$s, c(5, 15, 15))
  expect_equal(result$m, c(2.5, 5, 7.5))
})

test_that("filter -> arrange with selection vector (no compact)", {
  df <- data.frame(x = c(5, 3, 1, 4, 2, 8, 7, 6))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(x > 3) |> arrange(x) |> collect()
  expect_equal(result$x, c(4, 5, 6, 7, 8))
})

test_that("filter -> slice_head with selection vector (no compact)", {
  df <- data.frame(x = c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(x > 3) |> slice_head(n = 3) |> collect()
  expect_equal(nrow(result), 3)
  # After filter: 4,5,6,7,8,9,10 -> first 3 = 4,5,6
  expect_equal(result$x, c(4, 5, 6))
})

test_that("filter -> window function with selection vector", {
  df <- data.frame(
    g = c("a", "a", "a", "b", "b", "b"),
    x = c(10, 20, 30, 40, 50, 60)
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(x > 10) |>
    group_by(g) |> mutate(rn = row_number()) |> collect()
  result <- result[order(result$g, result$rn), ]
  expect_equal(result$x, c(20, 30, 40, 50, 60))
  expect_equal(result$rn, c(1, 2, 1, 2, 3))
})

# --- year/month/day extraction from Date ---

test_that("year extracts year from Date column", {
  df <- data.frame(d = as.Date(c("2020-03-15", "2021-12-01", "1999-06-30")))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(y = year(d)) |> collect()
  expect_equal(result$y, c(2020, 2021, 1999))
})

test_that("month extracts month from Date column", {
  df <- data.frame(d = as.Date(c("2020-03-15", "2021-12-01")))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(m = month(d)) |> collect()
  expect_equal(result$m, c(3, 12))
})

test_that("day extracts day from Date column", {
  df <- data.frame(d = as.Date(c("2020-03-15", "2021-12-01")))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(dy = day(d)) |> collect()
  expect_equal(result$dy, c(15, 1))
})

# --- date filtering with as.Date literal ---

test_that("filter with as.Date literal works", {
  df <- data.frame(d = as.Date(c("2019-01-01", "2020-06-15", "2021-12-31")),
                   x = c(1.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(d >= as.Date("2020-01-01")) |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(result$x, c(2, 3))
})

# --- as.Date from string column ---

test_that("as.Date converts string column to date numeric", {
  df <- data.frame(s = c("2020-01-15", "2021-06-30"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(d = as.Date(s)) |> collect()
  expected <- as.double(as.Date(c("2020-01-15", "2021-06-30")))
  expect_equal(result$d, expected)
})

# --- NA handling in date functions ---

test_that("year handles NA dates", {
  df <- data.frame(d = as.Date(c("2020-03-15", NA)))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(y = year(d)) |> collect()
  expect_equal(result$y[1], 2020)
  expect_true(is.na(result$y[2]))
})

# --- POSIXct extraction ---

test_that("hour/minute/second extract from POSIXct", {
  df <- data.frame(t = as.POSIXct(c("2020-03-15 14:30:45", "2021-12-01 08:15:00"), tz = "UTC"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(h = hour(t), mi = minute(t), s = second(t)) |> collect()
  expect_equal(result$h, c(14, 8))
  expect_equal(result$mi, c(30, 15))
  expect_equal(result$s, c(45, 0))
})

# --- date arithmetic ---

test_that("date arithmetic works (add days)", {
  df <- data.frame(d = as.Date(c("2020-01-01", "2020-06-15")))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(d2 = d + 30) |> collect()
  expected <- as.double(as.Date(c("2020-01-01", "2020-06-15")) + 30)
  expect_equal(result$d2, expected)
})

# --- group_by with year ---

test_that("group_by year works with Date column", {
  df <- data.frame(
    d = as.Date(c("2020-03-15", "2020-07-01", "2021-01-15", "2021-06-30")),
    x = c(1.0, 2.0, 3.0, 4.0)
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(yr = year(d)) |>
    group_by(yr) |>
    summarise(total = sum(x)) |>
    collect()
  expect_equal(result$total[result$yr == 2020], 3)
  expect_equal(result$total[result$yr == 2021], 7)
})

# --- as.POSIXct literal in filter ---

test_that("filter with as.POSIXct literal works", {
  df <- data.frame(
    t = as.POSIXct(c("2020-01-01 00:00:00", "2020-06-15 12:00:00", "2021-01-01 00:00:00"), tz = "UTC"),
    x = c(1.0, 2.0, 3.0)
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(t >= as.POSIXct("2020-06-01", tz = "UTC")) |> collect()
  expect_equal(nrow(result), 2)
})

# --- levenshtein distance ---

test_that("levenshtein exact match is 0", {
  df <- data.frame(name = c("rubra", "alba", "rubra"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(d = levenshtein(name, "rubra")) |> collect()
  expect_equal(result$d, c(0, 3, 0))
})

test_that("levenshtein handles NA", {
  df <- data.frame(name = c("rubra", NA, "alba"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(d = levenshtein(name, "rubra")) |> collect()
  expect_equal(result$d, c(0, NA, 3))
})

test_that("levenshtein known distances", {
  df <- data.frame(name = c("kitten", "saturday", ""))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(d = levenshtein(name, "sitting")) |> collect()
  expect_equal(result$d, c(3, 6, 7))
})

test_that("levenshtein column vs column", {
  df <- data.frame(a = c("rubra", "alba"), b = c("rubrum", "alba"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(d = levenshtein(a, b)) |> collect()
  expect_equal(result$d, c(2, 0))
})

test_that("levenshtein with max_dist early termination", {
  df <- data.frame(name = c("rubra", "pratensis", "rubrum"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(d = levenshtein(name, "rubra", max_dist = 2)) |> collect()
  expect_equal(result$d[1], 0)  # exact match
  expect_equal(result$d[2], 3)  # capped at max_dist + 1
  expect_equal(result$d[3], 2)  # within bound
})

test_that("levenshtein works in filter", {
  df <- data.frame(name = c("rubra", "pratensis", "rubrum", "alba"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(levenshtein(name, "rubra") <= 2) |> collect()
  expect_equal(nrow(result), 2)
  expect_true(all(result$name %in% c("rubra", "rubrum")))
})

# --- levenshtein_norm ---

test_that("levenshtein_norm in 0-1 range", {
  df <- data.frame(name = c("rubra", "rubrum", "pratensis"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(d = levenshtein_norm(name, "rubra")) |> collect()
  expect_true(all(result$d >= 0 & result$d <= 1))
  expect_equal(result$d[1], 0)  # exact match
})

test_that("levenshtein_norm both empty is 0", {
  df <- data.frame(name = c("", "abc"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(d = levenshtein_norm(name, "")) |> collect()
  expect_equal(result$d[1], 0)   # both empty -> 0
  expect_equal(result$d[2], 1)   # "abc" vs "" -> 3/3 = 1.0
})

test_that("levenshtein_norm handles NA", {
  df <- data.frame(name = c("rubra", NA))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(d = levenshtein_norm(name, "rubra")) |> collect()
  expect_equal(result$d[1], 0)
  expect_true(is.na(result$d[2]))
})

# --- Damerau-Levenshtein distance ---

test_that("dl_dist counts transpositions as 1", {
  # "ab" -> "ba" is 1 transposition (DL) vs 2 ops (plain Levenshtein)
  df <- data.frame(name = c("ab", "abc", "pratesnsis"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  res_dl <- tbl(f) |> mutate(d = dl_dist(name, "ba")) |> collect()
  res_lv <- tbl(f) |> mutate(d = levenshtein(name, "ba")) |> collect()
  expect_equal(res_dl$d[1], 1)  # transposition: cost 1
  expect_equal(res_lv$d[1], 2)  # no transposition op: cost 2
})

test_that("dl_dist exact match is 0", {
  df <- data.frame(name = c("rubra", "alba"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(d = dl_dist(name, "rubra")) |> collect()
  expect_equal(result$d[1], 0)
})

test_that("dl_dist handles NA", {
  df <- data.frame(name = c("rubra", NA))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(d = dl_dist(name, "rubra")) |> collect()
  expect_equal(result$d[1], 0)
  expect_true(is.na(result$d[2]))
})

test_that("dl_dist column vs column", {
  df <- data.frame(a = c("abc", "alba"), b = c("bac", "alba"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(d = dl_dist(a, b)) |> collect()
  expect_equal(result$d, c(1, 0))  # "abc"->"bac" = 1 transposition
})

test_that("dl_dist with max_dist early termination", {
  df <- data.frame(name = c("rubra", "pratensis", "rubrum"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(d = dl_dist(name, "rubra", max_dist = 2)) |> collect()
  expect_equal(result$d[1], 0)
  expect_equal(result$d[2], 3)  # capped
})

test_that("dl_dist_norm in 0-1 range", {
  df <- data.frame(name = c("rubra", "rubrum", "ba"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(d = dl_dist_norm(name, "rubra")) |> collect()
  expect_true(all(result$d >= 0 & result$d <= 1))
  expect_equal(result$d[1], 0)
})

test_that("dl_dist works in filter", {
  df <- data.frame(name = c("rubra", "rубра", "pratensis", "rubmra"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(dl_dist(name, "rubra") <= 1) |> collect()
  expect_true("rubra" %in% result$name)
})

# --- Jaro-Winkler similarity ---

test_that("jaro_winkler identical strings are 1.0", {
  df <- data.frame(name = c("rubra", "pratensis"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(jw = jaro_winkler(name, "rubra")) |> collect()
  expect_equal(result$jw[1], 1.0)
  expect_true(result$jw[2] < 1.0)
})

test_that("jaro_winkler in 0-1 range", {
  df <- data.frame(name = c("rubra", "rubrum", "alba", "pratensis", ""))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(jw = jaro_winkler(name, "rubra")) |> collect()
  expect_true(all(result$jw >= 0 & result$jw <= 1))
})

test_that("jaro_winkler prefix bonus rewards shared prefix", {
  # "rubrum" shares 4-char prefix "rubr" with "rubra" -> high JW
  # "xubra" shares 0-char prefix -> lower JW even though edit distance is 1
  df <- data.frame(name = c("rubrum", "xubra"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(jw = jaro_winkler(name, "rubra")) |> collect()
  expect_true(result$jw[1] > result$jw[2])
})

test_that("jaro_winkler handles NA", {
  df <- data.frame(name = c("rubra", NA))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(jw = jaro_winkler(name, "rubra")) |> collect()
  expect_equal(result$jw[1], 1.0)
  expect_true(is.na(result$jw[2]))
})

test_that("jaro_winkler both empty is 1.0", {
  df <- data.frame(name = c("", "abc"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    mutate(jw = jaro_winkler(name, "")) |> collect()
  expect_equal(result$jw[1], 1.0)   # both empty -> identical
  expect_equal(result$jw[2], 0.0)   # "abc" vs "" -> no matches
})

test_that("jaro_winkler column vs column", {
  df <- data.frame(a = c("rubra", "alba"), b = c("rubra", "pratensis"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(jw = jaro_winkler(a, b)) |> collect()
  expect_equal(result$jw[1], 1.0)
  expect_true(result$jw[2] < 1.0)
})

test_that("jaro_winkler works in filter", {
  df <- data.frame(name = c("rubra", "rubrum", "pratensis", "alba"))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    filter(jaro_winkler(name, "rubra") >= 0.85) |> collect()
  expect_true("rubra" %in% result$name)
  expect_true("rubrum" %in% result$name)
})

# Native libgeos scalar geometry expressions in mutate()/filter()/summarise().
# Geometry rides through the engine as hex-WKB in a string column; each result
# is recovery-tested against sf's own computation (the ground truth). CRS is
# dropped on both sides so the comparison is purely planar.

skip_if_not_installed("sf")

vtr_from <- function(sfobj) {
  f <- tempfile(fileext = ".vtr")
  df <- as.data.frame(sf::st_drop_geometry(sfobj))
  if (ncol(df) == 0) df <- data.frame(id = seq_along(sf::st_geometry(sfobj)))
  df$geometry <- sf::st_as_binary(sf::st_geometry(sfobj), hex = TRUE)
  write_vtr(df, f)
  f
}

planar <- function(g) sf::st_set_crs(sf::st_geometry(g), NA)

# A handful of unit squares at increasing x offsets, plus one point.
squares <- function() {
  mk <- function(ox) sf::st_polygon(list(rbind(
    c(ox, 0), c(ox + 1, 0), c(ox + 1, 1), c(ox, 1), c(ox, 0))))
  sf::st_sf(id = 1:3, geometry = sf::st_sfc(mk(0), mk(2), mk(4)))
}

# ---- measures -------------------------------------------------------------

test_that("st_area / st_length recover sf's measures", {
  x <- squares(); f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> mutate(a = st_area(geometry), p = st_length(geometry)) |> collect()
  expect_equal(d$a, as.numeric(sf::st_area(planar(x))), tolerance = 1e-9)
  # st_length of a polygon is its perimeter (boundary length)
  expect_equal(d$p, as.numeric(sf::st_length(
    sf::st_cast(planar(x), "MULTILINESTRING"))), tolerance = 1e-9)
})

test_that("st_npoints and st_geometry_type recover sf", {
  x <- squares(); f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> mutate(n = st_npoints(geometry),
                        t = st_geometry_type(geometry)) |> collect()
  expect_equal(d$n, rep(5, 3))                 # 5 coords per closed square ring
  expect_true(all(toupper(d$t) == "POLYGON"))
})

test_that("st_x / st_y recover point coordinates and NA on non-points", {
  pts <- sf::st_sf(geometry = sf::st_sfc(
    sf::st_point(c(1, 2)), sf::st_point(c(3, 4))))
  f <- vtr_from(pts); on.exit(unlink(f))
  d <- tbl(f) |> mutate(px = st_x(geometry), py = st_y(geometry)) |> collect()
  expect_equal(d$px, c(1, 3)); expect_equal(d$py, c(2, 4))
  # polygons are not points -> NA
  fp <- vtr_from(squares()); on.exit(unlink(fp), add = TRUE)
  dp <- tbl(fp) |> mutate(px = st_x(geometry)) |> collect()
  expect_true(all(is.na(dp$px)))
})

# ---- predicates -----------------------------------------------------------

test_that("unary predicates recover is_valid / is_empty", {
  x <- squares(); f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> mutate(v = st_is_valid(geometry),
                        e = st_is_empty(geometry)) |> collect()
  expect_true(all(d$v)); expect_true(all(!d$e))
})

test_that("st_intersects against a constant AOI filters like sf", {
  x <- squares(); f <- vtr_from(x); on.exit(unlink(f))
  aoi <- sf::st_as_sfc(sf::st_bbox(c(xmin = 1.5, ymin = 0, xmax = 4.5, ymax = 1)))
  got <- tbl(f) |> filter(st_intersects(geometry, aoi)) |> collect()
  ref <- which(lengths(sf::st_intersects(planar(x), planar(aoi))) > 0)
  expect_setequal(got$id, ref)
})

test_that("st_within / st_contains distinguish direction", {
  x <- squares(); f <- vtr_from(x); on.exit(unlink(f))
  big <- sf::st_as_sfc(sf::st_bbox(c(xmin = -1, ymin = -1, xmax = 2, ymax = 2)))
  d <- tbl(f) |> mutate(w = st_within(geometry, big)) |> collect()
  ref <- as.logical(lengths(sf::st_within(planar(x), planar(big))) > 0)
  expect_equal(d$w, ref)
})

test_that("predicate between two geometry columns (self) holds", {
  x <- squares(); f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> mutate(eq = st_equals(geometry, geometry),
                        ix = st_intersects(geometry, geometry)) |> collect()
  expect_true(all(d$eq)); expect_true(all(d$ix))
})

# ---- distance -------------------------------------------------------------

test_that("st_distance recovers sf, to a constant and to self", {
  x <- squares(); f <- vtr_from(x); on.exit(unlink(f))
  pt <- sf::st_sfc(sf::st_point(c(0, 0)))
  d <- tbl(f) |> mutate(dp = st_distance(geometry, pt),
                        d0 = st_distance(geometry, geometry)) |> collect()
  expect_equal(d$dp, as.numeric(sf::st_distance(planar(x), planar(pt))),
               tolerance = 1e-9)
  expect_equal(d$d0, rep(0, 3), tolerance = 1e-12)
})

# ---- transforms -----------------------------------------------------------

test_that("st_centroid recovers sf centroid coordinates", {
  x <- squares(); f <- vtr_from(x); on.exit(unlink(f))
  cent <- tbl(f) |> mutate(geometry = st_centroid(geometry)) |>
    select(id, geometry) |> collect_sf()
  ref <- suppressWarnings(sf::st_centroid(planar(x)))
  expect_equal(sf::st_coordinates(cent), sf::st_coordinates(ref),
               tolerance = 1e-9, ignore_attr = TRUE)
})

test_that("st_buffer area matches sf and st_convex_hull is a polygon", {
  x <- squares(); f <- vtr_from(x); on.exit(unlink(f))
  buf <- tbl(f) |> mutate(geometry = st_buffer(geometry, 0.25)) |>
    select(id, geometry) |> collect_sf()
  expect_equal(as.numeric(sf::st_area(buf)),
               as.numeric(sf::st_area(sf::st_buffer(planar(x), 0.25))),
               tolerance = 1e-3)
  hull <- tbl(f) |> mutate(geometry = st_convex_hull(geometry)) |>
    select(id, geometry) |> collect_sf()
  expect_true(all(sf::st_geometry_type(hull) == "POLYGON"))
})

test_that("st_envelope of a square is the square itself", {
  x <- squares(); f <- vtr_from(x); on.exit(unlink(f))
  env <- tbl(f) |> mutate(geometry = st_envelope(geometry)) |>
    select(id, geometry) |> collect_sf()
  expect_equal(as.numeric(sf::st_area(env)), rep(1, 3), tolerance = 1e-9)
})

test_that("st_make_valid repairs a self-intersecting bowtie", {
  bow <- sf::st_sfc(sf::st_polygon(list(rbind(
    c(0, 0), c(1, 1), c(1, 0), c(0, 1), c(0, 0)))))
  x <- sf::st_sf(id = 1L, geometry = bow)
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |>
    mutate(before = st_is_valid(geometry),
           geometry = st_make_valid(geometry)) |>
    mutate(after = st_is_valid(geometry)) |>
    collect_sf()
  expect_false(d$before[1]); expect_true(d$after[1])
})

# ---- NA handling ----------------------------------------------------------

test_that("a missing geometry yields NA, not an error", {
  x <- squares()
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  df <- data.frame(id = 1:4)
  df$geometry <- c(sf::st_as_binary(sf::st_geometry(x), hex = TRUE), NA)
  write_vtr(df, f)
  d <- tbl(f) |> mutate(a = st_area(geometry),
                        c_is_na = is.na(st_centroid(geometry))) |> collect()
  expect_true(is.na(d$a[4])); expect_false(any(is.na(d$a[1:3])))
  expect_true(d$c_is_na[4])
})

# --- TopN heap-based optimization tests ---

test_that("slice_min basic correctness", {
  df <- data.frame(x = c(5.0, 1.0, 3.0, 2.0, 4.0),
                   y = c("e", "a", "c", "b", "d"),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 3) |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(result$x, c(1, 2, 3))
  expect_equal(result$y, c("a", "b", "c"))
})

test_that("slice_max basic correctness", {
  df <- data.frame(x = c(5.0, 1.0, 3.0, 2.0, 4.0),
                   y = c("e", "a", "c", "b", "d"),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_max(order_by = x, n = 3) |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(result$x, c(5, 4, 3))
  expect_equal(result$y, c("e", "d", "c"))
})

test_that("topn with n > nrow returns all rows sorted", {
  df <- data.frame(x = c(3.0, 1.0, 2.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 10) |> collect()
  expect_equal(nrow(result), 3)
  expect_equal(result$x, c(1, 2, 3))
})

test_that("topn with n = 1", {
  df <- data.frame(x = c(5.0, 1.0, 3.0, 2.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result_min <- tbl(f) |> slice_min(order_by = x, n = 1) |> collect()
  expect_equal(result_min$x, 1)
  result_max <- tbl(f) |> slice_max(order_by = x, n = 1) |> collect()
  expect_equal(result_max$x, 5)
})

test_that("topn with n = nrow returns all rows sorted", {
  df <- data.frame(x = c(3.0, 1.0, 2.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 3) |> collect()
  expect_equal(result$x, c(1, 2, 3))
})

test_that("topn with NAs sorts NAs last", {
  df <- data.frame(x = c(5.0, NA, 3.0, NA, 1.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 2) |> collect()
  expect_equal(result$x, c(1, 3))
})

test_that("topn with all NAs returns NAs", {
  df <- data.frame(x = c(NA_real_, NA_real_, NA_real_))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 2) |> collect()
  expect_equal(nrow(result), 2)
  expect_true(all(is.na(result$x)))
})

test_that("topn on integer column", {
  df <- data.frame(x = c(50L, 10L, 30L, 20L, 40L))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 3) |> collect()
  expect_equal(result$x, c(10, 20, 30))
})

test_that("topn on string column", {
  df <- data.frame(x = c("cherry", "apple", "banana", "date", "elderberry"),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 3) |> collect()
  expect_equal(result$x, c("apple", "banana", "cherry"))
})

test_that("topn on boolean column", {
  df <- data.frame(x = c(TRUE, FALSE, TRUE, FALSE, TRUE),
                   y = c(1.0, 2.0, 3.0, 4.0, 5.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 2) |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(result$x, c(FALSE, FALSE))
})

test_that("topn preserves all columns", {
  df <- data.frame(a = c(3.0, 1.0, 2.0), b = c("x", "y", "z"),
                   c = c(TRUE, FALSE, TRUE), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = a, n = 2) |> collect()
  expect_equal(result$a, c(1, 2))
  expect_equal(result$b, c("y", "z"))
  expect_equal(result$c, c(FALSE, TRUE))
})

test_that("topn after filter", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(x > 3) |> slice_min(order_by = x, n = 2) |> collect()
  expect_equal(result$x, c(4, 5))
})

test_that("topn after mutate", {
  df <- data.frame(x = c(3.0, 1.0, 2.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(y = x * 2) |> slice_min(order_by = y, n = 2) |> collect()
  expect_equal(result$x, c(1, 2))
  expect_equal(result$y, c(2, 4))
})

test_that("topn with duplicate values includes ties by default", {
  df <- data.frame(x = c(1.0, 2.0, 2.0, 3.0, 3.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  # with_ties = TRUE (default): boundary is 3.0 (4th smallest), keep all <= 3
  result <- tbl(f) |> slice_min(order_by = x, n = 4) |> collect()
  expect_equal(nrow(result), 6)
  expect_equal(result$x, c(1, 2, 2, 3, 3, 3))
})

test_that("topn with duplicate values without ties returns exactly n", {
  df <- data.frame(x = c(1.0, 2.0, 2.0, 3.0, 3.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 4, with_ties = FALSE) |> collect()
  expect_equal(nrow(result), 4)
})

test_that("topn single row dataset", {
  df <- data.frame(x = 42.0)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 1) |> collect()
  expect_equal(result$x, 42)
})

test_that("topn large dataset correctness", {
  set.seed(42)
  n <- 10000
  df <- data.frame(x = runif(n), y = sample(letters, n, replace = TRUE),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  k <- 50
  result <- tbl(f) |> slice_min(order_by = x, n = k) |> collect()
  expected <- head(df[order(df$x), ], k)
  expect_equal(result$x, expected$x)
})

# --- Column pruning ---

test_that("column pruning only reads needed columns", {
  df <- data.frame(g = c("a", "a", "b"), x = 1:3, y = 4:6,
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  # select only x -> scan should prune g and y

  result <- tbl(f) |> select(x) |> collect()
  expect_equal(result$x, c(1, 2, 3))
  expect_equal(ncol(result), 1)
})

test_that("column pruning works with filter", {
  df <- data.frame(a = c(1, 2, 3), b = c(10, 20, 30))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(a > 1) |> select(b) |> collect()
  expect_equal(result$b, c(20, 30))
})

test_that("column pruning with group_by summarise", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1, 2, 3), y = c(4, 5, 6),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> group_by(g) |> summarise(s = sum(x)) |> collect()
  expect_equal(nrow(result), 2)
  # y should have been pruned from the scan
})

# --- Predicate pushdown ---

test_that("predicate pushdown skips row groups", {
  # Create a multi-rowgroup file where one group can be skipped
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  fp <- file(f, "wb")
  close(fp)

  # Write via R: two row groups, first with x in [1,10], second with x in [100,200]
  df1 <- data.frame(x = 1:10)
  df2 <- data.frame(x = 100:200)
  write_vtr(rbind(df1, df2), f)

  # filter x > 50 should match only second group conceptually
  result <- tbl(f) |> filter(x > 50) |> collect()
  expect_true(all(result$x > 50))
})

test_that("predicate pushdown works with comparison operators", {
  df <- data.frame(x = c(1, 5, 10, 15, 20))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  expect_equal(nrow(tbl(f) |> filter(x >= 10) |> collect()), 3)
  expect_equal(nrow(tbl(f) |> filter(x < 10) |> collect()), 2)
  expect_equal(nrow(tbl(f) |> filter(x == 5) |> collect()), 1)
  expect_equal(nrow(tbl(f) |> filter(x != 5) |> collect()), 4)
})

# --- Nested aggregation expressions ---

test_that("nested expression in summarise works", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1, 2, 3), y = c(10, 20, 30),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> group_by(g) |> summarise(s = sum(x + y)) |> collect()
  result <- result[order(result$g), ]
  expect_equal(result$s[result$g == "a"], 11 + 22)
  expect_equal(result$s[result$g == "b"], 33)
})

test_that("nested expression with mean works", {
  df <- data.frame(g = c("a", "a"), x = c(2, 4), y = c(1, 1))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> group_by(g) |> summarise(m = mean(x * y)) |> collect()
  expect_equal(result$m, mean(c(2, 4)))
})

# --- String operations ---

test_that("nchar works on string column", {
  df <- data.frame(s = c("abc", "de", "f", "hello"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> mutate(n = nchar(s)) |> collect()
  expect_equal(result$n, c(3, 2, 1, 5))
})

test_that("nchar handles NA", {
  df <- data.frame(s = c("abc", NA, "de"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> mutate(n = nchar(s)) |> collect()
  expect_equal(result$n, c(3, NA, 2))
})

test_that("substr extracts substring", {
  df <- data.frame(s = c("hello", "world"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> mutate(sub = substr(s, 1, 3)) |> collect()
  expect_equal(result$sub, c("hel", "wor"))
})

test_that("grepl matches fixed pattern", {
  df <- data.frame(s = c("apple", "banana", "grape", "pineapple"),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(grepl("app", s)) |> collect()
  expect_equal(nrow(result), 2)
  expect_true(all(grepl("app", result$s)))
})

test_that("grepl with filter on NA", {
  df <- data.frame(s = c("abc", NA, "abcdef"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> filter(grepl("abc", s)) |> collect()
  expect_equal(nrow(result), 2)
})

# --- rank/dense_rank correctness ---

test_that("rank produces correct results", {
  df <- data.frame(x = c(3, 1, 2, 1, 3))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> mutate(r = rank(x)) |> collect()
  expect_equal(result$r, c(4, 1, 3, 1, 4))
})

test_that("dense_rank produces correct results", {
  df <- data.frame(x = c(3, 1, 2, 1, 3))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  result <- tbl(f) |> mutate(r = dense_rank(x)) |> collect()
  expect_equal(result$r, c(3, 1, 2, 1, 3))
})

# Group-aware slice_min / slice_max and ordered row_number().
# Known-truth recovery: earliest/latest row per group, whole winning row kept.

make_tbl <- function() {
  df <- data.frame(
    piece_id  = c(1L,1L,1L, 2L,2L, 3L,3L,3L),
    STATUS_YR = c(2010L,1995L,2020L, 0L,2001L, 1980L,1980L,1999L),
    src       = c(11L,12L,13L, 21L,22L, 31L,32L,33L),
    geom_wkb  = c("A1","A2","A3","B1","B2","C1","C2","C3"),
    stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  write_vtr(df, f)
  f
}

test_that("grouped slice_min keeps the earliest row per group (with_ties = FALSE)", {
  f <- make_tbl(); on.exit(unlink(f))
  g <- tbl(f) |> group_by(piece_id) |>
    slice_min(STATUS_YR, n = 1, with_ties = FALSE) |> collect()
  g <- g[order(g$piece_id), ]

  expect_equal(nrow(g), 3L)
  expect_equal(as.numeric(g$STATUS_YR), c(1995, 0, 1980))
  # The whole winning row is preserved, including the string geometry column.
  expect_equal(g$geom_wkb, c("A2", "B1", "C1"))
  expect_equal(g$src, c(12L, 21L, 31L))
})

test_that("grouped slice_max keeps the latest row per group (with_ties = FALSE)", {
  f <- make_tbl(); on.exit(unlink(f))
  g <- tbl(f) |> group_by(piece_id) |>
    slice_max(STATUS_YR, n = 1, with_ties = FALSE) |> collect()
  g <- g[order(g$piece_id), ]

  expect_equal(nrow(g), 3L)
  expect_equal(as.numeric(g$STATUS_YR), c(2020, 2001, 1999))
  expect_equal(g$geom_wkb, c("A3", "B2", "C3"))
})

test_that("grouped slice_min with_ties = TRUE keeps boundary ties", {
  f <- make_tbl(); on.exit(unlink(f))
  g <- tbl(f) |> group_by(piece_id) |>
    slice_min(STATUS_YR, n = 1, with_ties = TRUE) |> collect()
  g <- g[order(g$piece_id, g$src), ]

  # piece 3 has two sources at 1980 -> both survive.
  expect_equal(nrow(g), 4L)
  expect_equal(g$src[g$piece_id == 3], c(31L, 32L))
})

test_that("grouped slice keeps exactly n per group when n > 1", {
  f <- make_tbl(); on.exit(unlink(f))
  g <- tbl(f) |> group_by(piece_id) |>
    slice_min(STATUS_YR, n = 2, with_ties = FALSE) |> collect()

  expect_equal(sum(g$piece_id == 1), 2L)   # 1995, 2010
  expect_equal(sum(g$piece_id == 2), 2L)   # 0, 2001
  expect_equal(sort(g$STATUS_YR[g$piece_id == 1]), c(1995, 2010))
})

test_that("NA in the order column sorts last so a known value wins", {
  df <- data.frame(
    g    = c(1L, 1L, 1L),
    yr   = c(NA_integer_, 2000L, 2010L),
    tag  = c("na", "win", "late"),
    stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f)

  g <- tbl(f) |> group_by(g) |>
    slice_min(yr, n = 1, with_ties = FALSE) |> collect()
  expect_equal(nrow(g), 1L)
  expect_equal(g$tag, "win")
})

test_that("ordered row_number() ranks by the column within each group", {
  f <- make_tbl(); on.exit(unlink(f))
  rn <- tbl(f) |> group_by(piece_id) |>
    mutate(rk = row_number(STATUS_YR)) |> collect()

  p1 <- rn[rn$piece_id == 1, ]
  p1 <- p1[order(p1$rk), ]
  expect_equal(p1$rk, c(1, 2, 3))
  expect_equal(as.numeric(p1$STATUS_YR), c(1995, 2010, 2020))
})

test_that("row_number(desc()) ranks largest first within each group", {
  f <- make_tbl(); on.exit(unlink(f))
  rn <- tbl(f) |> group_by(piece_id) |>
    mutate(rk = row_number(desc(STATUS_YR))) |> collect()

  p1 <- rn[rn$piece_id == 1, ]
  p1 <- p1[order(p1$rk), ]
  expect_equal(as.numeric(p1$STATUS_YR), c(2020, 2010, 1995))
})

test_that("bare row_number() is unchanged (input order within group)", {
  f <- make_tbl(); on.exit(unlink(f))
  rn <- tbl(f) |> group_by(piece_id) |>
    mutate(rk = row_number()) |> collect()
  p1 <- rn[rn$piece_id == 1, ]
  expect_equal(p1$rk[match(c("A1","A2","A3"), p1$geom_wkb)], c(1, 2, 3))
})

test_that("ungrouped slice_min/slice_max remain global", {
  f <- make_tbl(); on.exit(unlink(f))
  u <- tbl(f) |> slice_min(STATUS_YR, n = 2, with_ties = FALSE) |> collect()
  expect_equal(nrow(u), 2L)
  expect_true(all(u$STATUS_YR <= 1980))   # two globally smallest (0, 1980)
})

# Base-R reference: the winning row per group is the earliest known order value
# (NA sorts last), ties broken by first appearance -- matching the streaming
# grouped-top-n operator that n == 1, with_ties = FALSE routes through.
ref_winner <- function(df, group_cols, ord_col, desc = FALSE) {
  key <- do.call(paste, c(df[group_cols], sep = "\r"))
  ord <- df[[ord_col]]
  rn  <- seq_len(nrow(df))
  win <- vapply(split(rn, factor(key, levels = unique(key))), function(idx) {
    v <- ord[idx]; valid <- !is.na(v)
    if (!any(valid)) return(idx[1L])
    iv <- idx[valid]; vv <- v[valid]
    iv[if (desc) which.max(vv) else which.min(vv)]
  }, integer(1))
  df[sort(win), , drop = FALSE]
}

test_that("grouped slice recovers the per-group winner across types", {
  set.seed(7)
  n <- 4000L
  d <- data.frame(
    k1   = sample(1:200, n, replace = TRUE),
    k2   = sample(letters[1:6], n, replace = TRUE),
    ord  = sample.int(500L, n, replace = TRUE),
    dbl  = rnorm(n),
    flag = sample(c(TRUE, FALSE), n, replace = TRUE),
    blob = paste0("WKB", sprintf("%07d", seq_len(n)), strrep("z", 30)),
    stringsAsFactors = FALSE)
  d$ord[sample(n, n %/% 8)] <- NA_integer_   # NA order values sort last
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f)); write_vtr(d, f)

  sort_rows <- function(x) x[order(x$k1, x$k2), , drop = FALSE]

  for (desc in c(FALSE, TRUE)) {
    fn  <- if (desc) slice_max else slice_min
    got <- tbl(f) |> group_by(k1, k2) |>
      fn(ord, n = 1, with_ties = FALSE) |> collect()
    ref <- ref_winner(d, c("k1", "k2"), "ord", desc = desc)
    g <- sort_rows(got); r <- sort_rows(ref)
    expect_equal(nrow(g), nrow(r))
    expect_equal(g$ord, r$ord)
    expect_equal(g$dbl, r$dbl)             # double passthrough
    expect_equal(g$flag, r$flag)           # bool passthrough
    expect_equal(g$blob, r$blob)           # whole geometry-sized string kept
  }
})

test_that("grouped slice emits its winners in batches (crosses the emit boundary)", {
  # More than one output batch (the operator emits 131072 winners per batch), so
  # this exercises the row-range boundary in the chunked emit path.
  N <- 140000L
  d <- data.frame(g = 1:N, ord = sample.int(9999L, N, replace = TRUE),
                  tag = sprintf("t%06d", 1:N), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f)); write_vtr(d, f)

  out <- tbl(f) |> group_by(g) |> slice_min(ord, n = 1, with_ties = FALSE) |> collect()
  out <- out[order(out$g), ]
  expect_equal(nrow(out), N)                       # one winner per group
  expect_identical(as.integer(out$g), 1:N)         # contiguous across the boundary
  expect_identical(as.integer(out$ord), as.integer(d$ord))
  expect_identical(out$tag, d$tag)                 # string column intact across batches
})

test_that("write_sqlite + tbl_sqlite round-trips a data.frame", {
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(db))

  write_sqlite(mtcars[1:10, ], db, "cars")
  result <- tbl_sqlite(db, "cars") |> collect()

  expect_equal(nrow(result), 10)
  expect_equal(ncol(result), ncol(mtcars))
  expect_equal(result$mpg, mtcars$mpg[1:10])
  expect_equal(result$cyl, mtcars$cyl[1:10])
})

test_that("tbl_sqlite + filter gives same result as R subset", {
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(db))

  write_sqlite(mtcars, db, "cars")
  result <- tbl_sqlite(db, "cars") |>
    filter(cyl == 6) |>
    select(mpg, cyl) |>
    collect()

  expected <- mtcars[mtcars$cyl == 6, c("mpg", "cyl")]
  expect_equal(nrow(result), nrow(expected))
  expect_equal(result$mpg, expected$mpg)
})

test_that("tbl_sqlite handles string columns", {
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(db))

  df <- data.frame(name = c("Alice", "Bob", "Charlie"),
                   score = c(90, 85, 95),
                   stringsAsFactors = FALSE)
  write_sqlite(df, db, "people")
  result <- tbl_sqlite(db, "people") |> collect()

  expect_equal(result$name, c("Alice", "Bob", "Charlie"))
  expect_equal(result$score, c(90, 85, 95))
})

test_that("tbl_sqlite handles NA values", {
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(db))

  df <- data.frame(x = c(1, NA, 3), y = c("a", NA, "c"),
                   stringsAsFactors = FALSE)
  write_sqlite(df, db, "nulls")
  result <- tbl_sqlite(db, "nulls") |> collect()

  expect_equal(result$x[1], 1)
  expect_true(is.na(result$x[2]))
  expect_equal(result$x[3], 3)
  expect_true(is.na(result$y[2]))
})

test_that("tbl_sqlite works with filter pipeline", {
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(db))

  write_sqlite(mtcars, db, "cars")
  result <- tbl_sqlite(db, "cars") |>
    filter(cyl == 6) |>
    select(mpg, cyl, hp) |>
    collect()

  expected <- mtcars[mtcars$cyl == 6, c("mpg", "cyl", "hp")]
  expect_equal(nrow(result), nrow(expected))
  expect_equal(result$mpg, expected$mpg)
})

test_that("tbl_sqlite works with group_by + summarise", {
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(db))

  write_sqlite(mtcars, db, "cars")
  result <- tbl_sqlite(db, "cars") |>
    group_by(cyl) |>
    summarise(mean_mpg = mean(mpg)) |>
    arrange(cyl) |>
    collect()

  expect_equal(nrow(result), 3)
  expect_equal(result$cyl, c(4, 6, 8))
})

test_that("tbl_sqlite explain shows SqlScanNode", {
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(db))

  write_sqlite(mtcars[1:5, ], db, "cars")
  node <- tbl_sqlite(db, "cars")
  plan <- capture.output(explain(node))
  expect_true(any(grepl("SqlScanNode", plan)))
})

test_that("write_sqlite streams from vectra_node", {
  f <- tempfile(fileext = ".vtr")
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(c(f, db)))

  write_vtr(mtcars, f)
  tbl(f) |>
    filter(cyl == 4) |>
    write_sqlite(db, "four_cyl")

  result <- tbl_sqlite(db, "four_cyl") |> collect()
  expected <- mtcars[mtcars$cyl == 4, ]
  expect_equal(nrow(result), nrow(expected))
  expect_equal(result$mpg, expected$mpg)
})

test_that("tbl_sqlite handles logical columns via INTEGER", {
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(db))

  df <- data.frame(flag = c(TRUE, FALSE, TRUE), val = c(1, 2, 3))
  write_sqlite(df, db, "bools")
  result <- tbl_sqlite(db, "bools") |> collect()

  # Bools stored as INTEGER in SQLite, read back as int64 → double
  expect_equal(result$val, c(1, 2, 3))
})

test_that("tbl_sqlite respects batch_size", {
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(db))

  write_sqlite(mtcars, db, "cars")
  result <- tbl_sqlite(db, "cars", batch_size = 5L) |> collect()
  expect_equal(nrow(result), nrow(mtcars))
  expect_equal(result$mpg, mtcars$mpg)
})

test_that("csv -> sqlite -> collect round-trip works", {
  csv_f <- tempfile(fileext = ".csv")
  db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(c(csv_f, db)))

  write.csv(mtcars[1:10, ], csv_f, row.names = FALSE)
  tbl_csv(csv_f) |> write_sqlite(db, "from_csv")

  result <- tbl_sqlite(db, "from_csv") |> collect()
  expect_equal(nrow(result), 10)
  expect_equal(result$mpg, mtcars$mpg[1:10])
})

# --- math functions ---

test_that("abs works in mutate", {
  df <- data.frame(x = c(-1.0, 2.0, -3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(a = abs(x)) |> collect()
  expect_equal(result$a, c(1, 2, 3))
})

test_that("sqrt works in mutate", {
  df <- data.frame(x = c(4.0, 9.0, 16.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(s = sqrt(x)) |> collect()
  expect_equal(result$s, c(2, 3, 4))
})

test_that("log and exp are inverses", {
  df <- data.frame(x = c(1.0, 2.0, 10.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(lx = log(x), elx = exp(log(x))) |> collect()
  expect_equal(result$elx, result$x, tolerance = 1e-10)
})

test_that("floor and ceiling work", {
  df <- data.frame(x = c(1.2, 2.8, -1.5))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(fl = floor(x), ce = ceiling(x)) |> collect()
  expect_equal(result$fl, c(1, 2, -2))
  expect_equal(result$ce, c(2, 3, -1))
})

test_that("round works", {
  df <- data.frame(x = c(1.3, 2.7, 3.4))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(r = round(x)) |> collect()
  expect_equal(result$r, c(1, 3, 3))
})

test_that("math functions propagate NA", {
  df <- data.frame(x = c(4.0, NA, 9.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(s = sqrt(x)) |> collect()
  expect_equal(result$s[1], 2)
  expect_true(is.na(result$s[2]))
  expect_equal(result$s[3], 3)
})

# --- if_else ---

test_that("if_else works in mutate", {
  df <- data.frame(x = c(1.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(label = if_else(x > 2, 10.0, 0.0)) |> collect()
  expect_equal(result$label, c(0, 0, 10))
})

test_that("if_else with NA condition returns NA", {
  df <- data.frame(x = c(1.0, NA, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(label = if_else(x > 2, 1.0, 0.0)) |> collect()
  expect_equal(result$label[1], 0)
  expect_true(is.na(result$label[2]))
  expect_equal(result$label[3], 1)
})

# --- %in% ---

test_that("%in% works in filter", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 4.0, 5.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(x %in% c(2, 4)) |> collect()
  expect_equal(result$x, c(2, 4))
})

test_that("%in% works with strings", {
  df <- data.frame(s = c("a", "b", "c", "d"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(s %in% c("a", "c")) |> collect()
  expect_equal(result$s, c("a", "c"))
})

# --- between ---

test_that("between works in filter", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 4.0, 5.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(between(x, 2, 4)) |> collect()
  expect_equal(result$x, c(2, 3, 4))
})

# --- type casting ---

test_that("as.numeric works in mutate", {
  df <- data.frame(x = c(1L, 2L, 3L))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(y = as.numeric(x)) |> collect()
  expect_true(is.double(result$y))
})

# --- string functions ---

test_that("tolower and toupper work", {
  df <- data.frame(s = c("Hello", "WORLD"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(lo = tolower(s), up = toupper(s)) |> collect()
  expect_equal(result$lo, c("hello", "world"))
  expect_equal(result$up, c("HELLO", "WORLD"))
})

test_that("trimws removes whitespace", {
  df <- data.frame(s = c("  hello  ", "\tfoo\n"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(t = trimws(s)) |> collect()
  expect_equal(result$t, c("hello", "foo"))
})

# --- additional math functions ---

test_that("log2 and log10 work", {
  df <- data.frame(x = c(8.0, 100.0, 1.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(l2 = log2(x), l10 = log10(x)) |> collect()
  expect_equal(result$l2, log2(c(8, 100, 1)))
  expect_equal(result$l10, log10(c(8, 100, 1)))
})

test_that("sign works", {
  df <- data.frame(x = c(-5.0, 0.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(s = sign(x)) |> collect()
  expect_equal(result$s, c(-1, 0, 1))
})

test_that("trunc works", {
  df <- data.frame(x = c(1.7, -2.3, 0.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(t = trunc(x)) |> collect()
  expect_equal(result$t, c(1, -2, 0))
})

# --- pmin / pmax ---

test_that("pmin returns element-wise minimum", {
  df <- data.frame(x = c(1.0, 5.0, 3.0), y = c(4.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(m = pmin(x, y)) |> collect()
  expect_equal(result$m, c(1, 2, 3))
})

test_that("pmax returns element-wise maximum", {
  df <- data.frame(x = c(1.0, 5.0, 3.0), y = c(4.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(m = pmax(x, y)) |> collect()
  expect_equal(result$m, c(4, 5, 3))
})

# --- paste0 ---

test_that("paste0 concatenates two string columns", {
  df <- data.frame(a = c("hello", "foo"), b = c(" world", "bar"),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(c = paste0(a, b)) |> collect()
  expect_equal(result$c, c("hello world", "foobar"))
})

test_that("paste0 with NA returns NA", {
  df <- data.frame(a = c("hello", NA), b = c(" world", "bar"),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(c = paste0(a, b)) |> collect()
  expect_equal(result$c[1], "hello world")
  expect_true(is.na(result$c[2]))
})

# --- startsWith / endsWith ---

test_that("startsWith works in filter", {
  df <- data.frame(s = c("apple", "banana", "avocado"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(startsWith(s, "a")) |> collect()
  expect_equal(result$s, c("apple", "avocado"))
})

test_that("endsWith works in filter", {
  df <- data.frame(s = c("cat", "dog", "rat"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> filter(endsWith(s, "at")) |> collect()
  expect_equal(result$s, c("cat", "rat"))
})

# --- gsub / sub ---

test_that("gsub replaces all occurrences", {
  df <- data.frame(s = c("aXbXc", "XdX"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(r = gsub("X", "_", s)) |> collect()
  expect_equal(result$r, c("a_b_c", "_d_"))
})

test_that("sub replaces only first occurrence", {
  df <- data.frame(s = c("aXbXc", "XdX"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(r = sub("X", "_", s)) |> collect()
  expect_equal(result$r, c("a_bXc", "_dX"))
})

# Core benchmarks for vectra regression testing
#
# Usage:
#   Rscript inst/benchmarks/bench_core.R              # run and print
#   Rscript inst/benchmarks/bench_core.R --save       # run, print, save baseline
#   Rscript inst/benchmarks/bench_core.R --check      # run and compare to baseline
#
# Measures runtime for key operations. Not comparative (no dplyr/arrow),
# just a baseline for tracking regressions as features grow.
#
# Exit code 1 if any benchmark exceeds its slowdown threshold (--check mode).

library(vectra)

# --- Configuration ---
SLOWDOWN_THRESHOLD <- 1.5  # flag if > 50% slower than baseline
# Locate baseline file: installed package or dev source tree
BASELINE_FILE <- tryCatch({
  d <- system.file("benchmarks", package = "vectra", mustWork = TRUE)
  file.path(d, "baseline.rds")
}, error = function(e) {
  # Dev mode: look relative to script or working directory
  candidates <- c(
    if (!is.null(sys.frame(1)$ofile)) file.path(dirname(sys.frame(1)$ofile), "baseline.rds"),
    file.path("inst", "benchmarks", "baseline.rds")
  )
  for (p in candidates) if (file.exists(p)) return(p)
  candidates[length(candidates)]  # default even if doesn't exist yet
})

args <- commandArgs(trailingOnly = TRUE)
save_baseline <- "--save" %in% args
check_baseline <- "--check" %in% args

# --- Timing helper ---
bench <- function(name, expr) {
  gc(FALSE)
  t0 <- proc.time()
  result <- eval(expr)
  elapsed <- (proc.time() - t0)[3]
  nr <- if (is.data.frame(result)) nrow(result) else NA
  list(name = name, elapsed = elapsed, nrow = nr)
}

# --- Setup ---
set.seed(42)
n <- 1e6
n_groups <- 1000
n_groups_high <- 50000

cat(sprintf("vectra benchmark suite | %d rows, %d groups\n", n, n_groups))
cat(sprintf("R %s | %s | %s\n", R.version.string,
            .Platform$OS.type, Sys.info()["machine"]))
cat("---\n")

wide <- data.frame(
  id = seq_len(n) + 0.0,
  g = sample(paste0("g", seq_len(n_groups)), n, replace = TRUE),
  x = rnorm(n),
  y = rnorm(n),
  z = rnorm(n),
  a = rnorm(n),
  b = rnorm(n),
  c = rnorm(n),
  d = rnorm(n),
  e = rnorm(n),
  stringsAsFactors = FALSE
)

# Long string keys for join stress test
long_keys <- data.frame(
  key = paste0("very_long_key_name_for_hashing_stress_", seq_len(n)),
  val = rnorm(n),
  stringsAsFactors = FALSE
)

# High-cardinality group data
hc <- data.frame(
  g = sample(paste0("hc", seq_len(n_groups_high)), n, replace = TRUE),
  x = rnorm(n),
  stringsAsFactors = FALSE
)

# Many-to-many join data
m2m_left <- data.frame(
  key = sample(seq_len(100), 10000, replace = TRUE) + 0.0,
  lx = rnorm(10000),
  stringsAsFactors = FALSE
)
m2m_right <- data.frame(
  key = sample(seq_len(100), 10000, replace = TRUE) + 0.0,
  ry = rnorm(10000),
  stringsAsFactors = FALSE
)

f <- tempfile(fileext = ".vtr")
f2 <- tempfile(fileext = ".vtr")
f3 <- tempfile(fileext = ".vtr")
f4 <- tempfile(fileext = ".vtr")
f5 <- tempfile(fileext = ".vtr")
f6 <- tempfile(fileext = ".vtr")
f7 <- tempfile(fileext = ".vtr")
on.exit(unlink(c(f, f2, f3, f4, f5, f6, f7)))

results <- list()

# --- B00: write_vtr ---
r <- bench("write_vtr", quote({
  write_vtr(wide, f)
  data.frame(size = file.size(f))
}))
cat(sprintf("%-30s %.3fs\n", r$name, r$elapsed))
results <- c(results, list(r))

# --- B01: filter + select (selective) ---
r <- bench("filter_select", quote(
  tbl(f) |> filter(x > 1.5) |> select(id, x, y) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B02: filter + select (broad) ---
r <- bench("filter_select_broad", quote(
  tbl(f) |> filter(x > 0) |> select(id, x, y) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B03: wide select (3 of 10 columns) ---
r <- bench("wide_select", quote(
  tbl(f) |> select(id, x, e) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B04: group_by + summarise (1K groups) ---
r <- bench("group_summarise_1k", quote(
  tbl(f) |> group_by(g) |>
    summarise(sx = sum(x), mx = mean(y), n = n()) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d groups)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B05: group_by + summarise (50K groups) ---
write_vtr(hc, f3)
r <- bench("group_summarise_50k", quote(
  tbl(f3) |> group_by(g) |>
    summarise(sx = sum(x), n = n()) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d groups)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B06: filter -> group_by -> summarise ---
r <- bench("filter_group_summarise", quote(
  tbl(f) |> filter(x > 0) |> group_by(g) |>
    summarise(sx = sum(x), n = n()) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d groups)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B07: arrange (full sort) ---
r <- bench("arrange_full", quote(
  tbl(f) |> arrange(x) |> slice_head(n = 10) |> collect()
))
cat(sprintf("%-30s %.3fs\n", r$name, r$elapsed))
results <- c(results, list(r))

# --- B08: slice_min (top-N, no full sort) ---
r <- bench("slice_min_100", quote(
  tbl(f) |> slice_min(order_by = x, n = 100) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B09: left_join (1M x 1K lookup) ---
lookup <- data.frame(
  g = paste0("g", seq_len(n_groups)),
  label = paste0("label_", seq_len(n_groups)),
  stringsAsFactors = FALSE
)
write_vtr(lookup, f2)
r <- bench("left_join_lookup", quote(
  tbl(f) |> left_join(tbl(f2), by = "g") |>
    select(id, g, x, label) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B10: filter -> left_join ---
r <- bench("filter_join", quote(
  tbl(f) |> filter(x > 1) |> left_join(tbl(f2), by = "g") |>
    select(id, g, x, label) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B11: many-to-many inner_join ---
write_vtr(m2m_left, f4)
write_vtr(m2m_right, f5)
r <- bench("inner_join_m2m", quote(
  inner_join(tbl(f4), tbl(f5), by = "key") |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B12: long string key join ---
write_vtr(long_keys, f6)
lk_right <- data.frame(
  key = sample(long_keys$key, 10000),
  extra = rnorm(10000),
  stringsAsFactors = FALSE
)
write_vtr(lk_right, f7)
r <- bench("join_long_string_key", quote(
  inner_join(tbl(f6), tbl(f7), by = "key") |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B13: window functions (grouped) ---
r <- bench("window_grouped", quote(
  tbl(f) |> group_by(g) |>
    mutate(rn = row_number(), cs = cumsum(x)) |>
    select(id, g, x, rn, cs) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B14: window mutate (ungrouped lag/lead) ---
r <- bench("window_lag_lead", quote(
  tbl(f) |> mutate(x_lag = lag(x, 1), x_lead = lead(x, 1)) |>
    select(id, x, x_lag, x_lead) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B15: streaming bind_rows + filter ---
r <- bench("bind_rows_filter", quote(
  bind_rows(tbl(f), tbl(f)) |> filter(x > 1.5) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B16: mutate (arithmetic) ---
r <- bench("mutate_arithmetic", quote(
  tbl(f) |> mutate(w = (x + y) * z - a / (b + 1)) |>
    select(id, w) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B17: nested aggregation expression ---
r <- bench("nested_agg_expr", quote(
  tbl(f) |> group_by(g) |>
    summarise(s = sum(x + y), m = mean(x * y)) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d groups)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B18: string operations ---
str_df <- data.frame(
  s = paste0("item_", sample(letters, n, replace = TRUE),
             sample(1000:9999, n, replace = TRUE)),
  x = rnorm(n),
  stringsAsFactors = FALSE
)
f_str <- tempfile(fileext = ".vtr")
write_vtr(str_df, f_str)

r <- bench("string_nchar_substr", quote(
  tbl(f_str) |>
    mutate(len = nchar(s), prefix = substr(s, 1, 6)) |>
    select(len, prefix) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B19: grepl filter ---
r <- bench("string_grepl_filter", quote(
  tbl(f_str) |> filter(grepl("item_a", s)) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

# --- B20: rank + dense_rank ---
r <- bench("rank_dense_rank", quote(
  tbl(f) |> mutate(r = rank(x), dr = dense_rank(x)) |>
    select(id, r, dr) |> collect()
))
cat(sprintf("%-30s %.3fs  (%d rows)\n", r$name, r$elapsed, r$nrow))
results <- c(results, list(r))

unlink(f_str)

# --- Build results table ---
cat("\n=== Summary ===\n")
res_df <- data.frame(
  benchmark = vapply(results, `[[`, character(1), "name"),
  seconds = vapply(results, `[[`, double(1), "elapsed"),
  stringsAsFactors = FALSE
)

# --- Save baseline ---
if (save_baseline) {
  saveRDS(res_df, BASELINE_FILE)
  cat(sprintf("Baseline saved to %s\n", BASELINE_FILE))
}

# --- Check against baseline ---
if (check_baseline) {
  if (!file.exists(BASELINE_FILE)) {
    cat("No baseline found. Run with --save first.\n")
    quit(status = 1)
  }
  baseline <- readRDS(BASELINE_FILE)
  cat("\n%-30s %8s %8s %8s  %s\n" |> sprintf("benchmark", "base", "now", "ratio", "status"))
  cat(strrep("-", 72), "\n")
  any_fail <- FALSE
  for (i in seq_len(nrow(res_df))) {
    bm <- res_df$benchmark[i]
    bi <- match(bm, baseline$benchmark)
    if (is.na(bi)) {
      cat(sprintf("%-30s %8s %8.3f %8s  NEW\n", bm, "---", res_df$seconds[i], "---"))
      next
    }
    base_t <- baseline$seconds[bi]
    now_t <- res_df$seconds[i]
    ratio <- now_t / base_t
    status <- if (ratio > SLOWDOWN_THRESHOLD) {
      any_fail <- TRUE
      "SLOW"
    } else if (ratio < 1 / SLOWDOWN_THRESHOLD) {
      "FAST"
    } else {
      "ok"
    }
    cat(sprintf("%-30s %8.3f %8.3f %7.2fx  %s\n", bm, base_t, now_t, ratio, status))
  }
  if (any_fail) {
    cat("\nFAILED: one or more benchmarks exceeded the slowdown threshold.\n")
    quit(status = 1)
  } else {
    cat("\nPASSED: all benchmarks within threshold.\n")
  }
}

#' Apply a function across multiple columns
#'
#' Used inside [mutate()] or [summarise()] to apply a function to multiple
#' columns selected with tidyselect. Returns a named list of expressions.
#'
#' @param .cols Column selection (tidyselect).
#' @param .fns A function, formula, or named list of functions.
#' @param ... Additional arguments passed to `.fns`.
#' @param .names A glue-style naming pattern. Uses `{.col}` and `{.fn}`.
#'   Default: `"{.col}"` if `.fns` is a single function,
#'   `"{.col}_{.fn}"` if `.fns` is a named list.
#'
#' @return A named list used internally by mutate/summarise.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' # In summarise (conceptual; across is expanded to individual expressions)
#' unlink(f)
#'
#' @export
across <- function(.cols, .fns, ..., .names = NULL) {
  stop("across() can only be used inside mutate() or summarise()")
}

# Internal: expand across() calls in mutate/summarise dots
# Returns a named list of expressions
expand_across <- function(dots, schema_names, env, proxy = NULL) {
  result_names <- character(0)
  result_exprs <- list()

  for (i in seq_along(dots)) {
    expr <- dots[[i]]
    nm <- names(dots)[i]

    # Check if this is an across() call
    if (is.call(expr) && identical(expr[[1]], as.name("across"))) {
      expanded <- do_expand_across(expr, schema_names, env, proxy)
      result_names <- c(result_names, names(expanded))
      result_exprs <- c(result_exprs, expanded)
    } else {
      result_names <- c(result_names, nm)
      result_exprs <- c(result_exprs, list(expr))
    }
  }

  names(result_exprs) <- result_names
  result_exprs
}

# Resolve a function to its name string (e.g., sum -> "sum")
resolve_fn_str <- function(fn) {
  # Check if it's a primitive or builtin
  if (is.primitive(fn)) {
    # Extract name from the deparse
    d <- deparse(fn)[1]
    m <- regmatches(d, regexpr('"[^"]*"', d))
    if (length(m) > 0) return(gsub('"', '', m))
  }
  # Check if it's a named function in an environment
  env <- environment(fn)
  if (!is.null(env)) {
    for (nm in ls(env)) {
      if (identical(get(nm, envir = env), fn)) return(nm)
    }
  }
  # Try matching known functions
  for (nm in c("sum", "mean", "min", "max", "n", "sd", "var", "median")) {
    if (identical(fn, get(nm, envir = baseenv(), inherits = TRUE)))
      return(nm)
  }
  # Fallback: deparse
  deparse(fn)[1]
}

do_expand_across <- function(expr, schema_names, env, proxy = NULL) {
  # Parse across(cols, fns, ..., .names = pattern)
  args <- as.list(expr)[-1]  # drop "across"
  arg_names <- names(args)

  cols_expr <- args[[1]]
  fns_expr <- args[[2]]

  # Get .names pattern if present
  names_pattern <- NULL
  if (!is.null(arg_names)) {
    nm_idx <- match(".names", arg_names)
    if (!is.na(nm_idx)) names_pattern <- eval(args[[nm_idx]], env)
  }

  # Resolve column selection via typed proxy (enables where())
  if (is.null(proxy)) {
    named_cols <- schema_names
    names(named_cols) <- schema_names
    proxy <- named_cols
  }
  sel <- tidyselect::eval_select(cols_expr, data = proxy)
  selected_cols <- unname(schema_names[sel])

  # Evaluate fns
  fns <- eval(fns_expr, env)

  # Handle different fn formats
  if (is.function(fns)) {
    fn_list <- list(fns)
    fn_names <- NULL
  } else if (rlang::is_formula(fns)) {
    fn_list <- list(rlang::as_function(fns))
    fn_names <- NULL
  } else if (is.list(fns)) {
    fn_list <- lapply(fns, function(f) {
      if (rlang::is_formula(f)) rlang::as_function(f) else f
    })
    fn_names <- names(fns)
  } else {
    stop("across .fns must be a function, formula, or named list")
  }

  # Generate expressions
  result <- list()
  for (fi in seq_along(fn_list)) {
    for (col in selected_cols) {
      fn_name <- if (!is.null(fn_names)) fn_names[fi] else NULL

      # Determine output name
      if (!is.null(names_pattern)) {
        out_name <- names_pattern
        out_name <- gsub("{.col}", col, out_name, fixed = TRUE)
        if (!is.null(fn_name))
          out_name <- gsub("{.fn}", fn_name, out_name, fixed = TRUE)
      } else if (!is.null(fn_names)) {
        out_name <- paste0(col, "_", fn_names[fi])
      } else {
        out_name <- col
      }

      # Build the call expression: fn(col_name)
      # Resolve the actual function name from the function object
      fn_obj <- fn_list[[fi]]
      fn_str <- resolve_fn_str(fn_obj)

      call_expr <- call(fn_str, as.name(col))
      result[[out_name]] <- call_expr
    }
  }

  result
}

#' Create a hash index on a .vtr file column
#'
#' Builds a persistent hash index stored as a `.vtri` sidecar file alongside
#' the `.vtr` file. The index maps key hashes to row group indices, enabling
#' O(1) row group identification for equality predicates (`filter(col == value)`).
#'
#' For composite indexes on multiple columns, pass a character vector.
#' Composite indexes accelerate AND-combined equality predicates
#' (e.g., `filter(col1 == "a", col2 == "b")`).
#'
#' The index is automatically loaded by [tbl()] when present. It composes with
#' zone-map pruning and binary search on sorted columns.
#'
#' @param path Path to a `.vtr` file.
#' @param column Character vector. Name(s) of column(s) to index.
#' @param ci Logical. Build a case-insensitive index? Default `FALSE`.
#'
#' @return Invisible `NULL`. The index is written as a `.vtri` sidecar file.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(id = letters, val = 1:26, stringsAsFactors = FALSE), f)
#' create_index(f, "id")
#' tbl(f) |> filter(id == "m") |> collect()
#' unlink(c(f, paste0(f, ".id.vtri")))
#'
#' @export
create_index <- function(path, column, ci = FALSE) {
  check_scalar_string(path)
  if (!is.character(column) || length(column) < 1)
    stop("column must be a character vector of length >= 1")
  path <- normalizePath(path, mustWork = TRUE)
  .Call(C_create_index, path, column, as.logical(ci))
  invisible(NULL)
}

#' Check if a hash index exists for a .vtr column
#'
#' @param path Path to a `.vtr` file.
#' @param column Character vector. Name(s) of column(s).
#'
#' @return Logical scalar: `TRUE` if a `.vtri` index file exists.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(id = letters, val = 1:26, stringsAsFactors = FALSE), f)
#' has_index(f, "id")   # FALSE
#' create_index(f, "id")
#' has_index(f, "id")   # TRUE
#' unlink(c(f, paste0(f, ".id.vtri")))
#'
#' @export
has_index <- function(path, column) {
  check_scalar_string(path)
  if (!is.character(column) || length(column) < 1)
    stop("column must be a character vector of length >= 1")
  path <- normalizePath(path, mustWork = TRUE)
  .Call(C_has_index, path, column)
}

test_that("create_index and has_index work", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, paste0(f, ".name.vtri"))))

  df <- data.frame(
    name = c("alice", "bob", "charlie", "diana", "eve"),
    val = 1:5,
    stringsAsFactors = FALSE
  )
  write_vtr(df, f)

  expect_false(has_index(f, "name"))
  create_index(f, "name")
  expect_true(has_index(f, "name"))
})

test_that("hash index accelerates equality lookups on strings", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, paste0(f, ".genus.vtri"))))

  df <- data.frame(
    genus = rep(c("Quercus", "Pinus", "Fagus", "Betula", "Acer"), each = 100),
    val = seq_len(500),
    stringsAsFactors = FALSE
  )
  write_vtr(df, f, batch_size = 100L)
  create_index(f, "genus")

  result <- tbl(f) |> filter(genus == "Pinus") |> collect()
  expect_equal(nrow(result), 100L)
  expect_true(all(result$genus == "Pinus"))
})

test_that("hash index works on integer columns", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, paste0(f, ".id.vtri"))))

  df <- data.frame(id = 1:500, val = runif(500))
  write_vtr(df, f, batch_size = 50L)
  create_index(f, "id")

  result <- tbl(f) |> filter(id == 250) |> collect()
  expect_equal(nrow(result), 1L)
  expect_equal(result$id, 250)
})

test_that("hash index with case-insensitive flag", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, paste0(f, ".name.vtri"))))

  df <- data.frame(
    name = c("Alice", "BOB", "Charlie"),
    val = 1:3,
    stringsAsFactors = FALSE
  )
  write_vtr(df, f)
  create_index(f, "name", ci = TRUE)

  # Lookup should match case-insensitively via the hash index
  # (the index provides the row groups, then filter does exact matching)
  # Since the index is CI, "alice" hashes to the same bucket as "Alice"
  expect_true(has_index(f, "name"))
})

test_that("hash index handles empty result", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, paste0(f, ".name.vtri"))))

  df <- data.frame(
    name = c("a", "b", "c"),
    val = 1:3,
    stringsAsFactors = FALSE
  )
  write_vtr(df, f)
  create_index(f, "name")

  result <- tbl(f) |> filter(name == "zzz") |> collect()
  expect_equal(nrow(result), 0L)
})

test_that("hash index with multiple row groups returns correct results", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, paste0(f, ".key.vtri"))))

  # Key values are spread across row groups
  df <- data.frame(
    key = rep(letters[1:10], times = 50),
    val = seq_len(500),
    stringsAsFactors = FALSE
  )
  write_vtr(df, f, batch_size = 100L)
  create_index(f, "key")

  result <- tbl(f) |> filter(key == "e") |> collect()
  expect_equal(nrow(result), 50L)
  expect_true(all(result$key == "e"))
})

test_that("create_index errors on non-existent column", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))

  write_vtr(data.frame(x = 1:10), f)
  expect_error(create_index(f, "nonexistent"), "not found")
})

test_that("index survives re-creation after data change", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, paste0(f, ".name.vtri"))))

  df1 <- data.frame(name = c("a", "b"), val = 1:2, stringsAsFactors = FALSE)
  write_vtr(df1, f)
  create_index(f, "name")

  # Overwrite with new data
  df2 <- data.frame(name = c("x", "y", "z"), val = 1:3, stringsAsFactors = FALSE)
  write_vtr(df2, f)

  # Old index is stale — re-create
  create_index(f, "name")

  result <- tbl(f) |> filter(name == "y") |> collect()
  expect_equal(nrow(result), 1L)
  expect_equal(result$name, "y")
})

#' Define a link between a fact table and a dimension table
#'
#' Creates a link descriptor that specifies how to join a dimension table
#' to a fact table via one or more key columns.
#'
#' @param key A character vector or named character vector specifying join keys.
#'   Unnamed: same column name in both tables. Named: `c("fact_col" = "dim_col")`.
#' @param node A `vectra_node` object (the dimension table). Must be file-backed
#'   (created via [tbl()], [tbl_csv()], or [tbl_sqlite()]).
#'
#' @return A `vectra_link` object.
#'
#' @examples
#' \donttest{
#' f_obs <- tempfile(fileext = ".vtr")
#' f_sp  <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(sp_id = 1:3, value = c(10, 20, 30)), f_obs)
#' write_vtr(data.frame(sp_id = 1:3, name = c("A", "B", "C")), f_sp)
#' lnk <- link("sp_id", tbl(f_sp))
#' unlink(c(f_obs, f_sp))
#' }
#'
#' @export
link <- function(key, node) {
  if (!is.character(key) || length(key) == 0)
    stop("key must be a non-empty character vector")
  if (!inherits(node, "vectra_node"))
    stop(sprintf("node must be a vectra_node, got %s", class(node)[1]))
  if (is.null(node$.path))
    stop("schema links require file-backed nodes (created via tbl, tbl_csv, etc.)")
  structure(list(key = key, node = node), class = "vectra_link")
}

#' Create a star schema over linked vectra tables
#'
#' Registers a fact table with named dimension links. The schema enables
#' [lookup()] to resolve columns from dimension tables without writing
#' explicit joins.
#'
#' @param fact A `vectra_node` object (the central fact table). Must be
#'   file-backed (created via [tbl()], [tbl_csv()], or [tbl_sqlite()]).
#' @param ... Named `vectra_link` objects created by [link()]. Names become
#'   the dimension aliases used in [lookup()] (e.g., `species$name`).
#'
#' @return A `vectra_schema` object.
#'
#' @examples
#' \donttest{
#' f_obs <- tempfile(fileext = ".vtr")
#' f_sp  <- tempfile(fileext = ".vtr")
#' f_ct  <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(sp_id = 1:3, ct_code = c("AT", "DE", "FR"),
#'                       value = 10:12), f_obs)
#' write_vtr(data.frame(sp_id = 1:3,
#'                       name = c("Oak", "Beech", "Pine")), f_sp)
#' write_vtr(data.frame(ct_code = c("AT", "DE", "FR"),
#'                       gdp = c(400, 3800, 2700)), f_ct)
#'
#' s <- vtr_schema(
#'   fact    = tbl(f_obs),
#'   species = link("sp_id", tbl(f_sp)),
#'   country = link("ct_code", tbl(f_ct))
#' )
#' print(s)
#' unlink(c(f_obs, f_sp, f_ct))
#' }
#'
#' @export
vtr_schema <- function(fact, ...) {
  if (!inherits(fact, "vectra_node"))
    stop(sprintf("fact must be a vectra_node, got %s", class(fact)[1]))
  if (is.null(fact$.path))
    stop("fact must be a file-backed node (created via tbl, tbl_csv, etc.)")

  dims <- list(...)
  if (length(dims) == 0)
    stop("at least one dimension link is required")

  dim_names <- names(dims)
  if (is.null(dim_names) || any(dim_names == ""))
    stop("all dimension links must be named (e.g., species = link(...))")

  for (nm in dim_names) {
    if (!inherits(dims[[nm]], "vectra_link"))
      stop(sprintf("'%s' must be a vectra_link created by link(), got %s",
                    nm, class(dims[[nm]])[1]))
  }

  structure(
    list(fact = fact, dims = dims),
    class = "vectra_schema"
  )
}

#' @export
print.vectra_schema <- function(x, ...) {
  fact_schema <- .Call(C_node_schema, x$fact$.node)
  cat("vectra schema\n")
  cat(sprintf("Fact table: %d columns\n", length(fact_schema$name)))

  for (nm in names(x$dims)) {
    lnk <- x$dims[[nm]]
    dim_schema <- .Call(C_node_schema, lnk$node$.node)
    key_str <- format_key_str(lnk$key)
    cat(sprintf("  %s: %d columns (key: %s)\n",
                nm, length(dim_schema$name), key_str))
  }
  invisible(x)
}

# Format a join key spec for display
format_key_str <- function(key) {
  if (is.null(names(key))) return(paste(key, collapse = ", "))
  parts <- character(length(key))
  for (i in seq_along(key)) {
    k <- names(key)[i]
    if (k == "") k <- key[i]
    parts[i] <- if (k == key[i]) k else paste0(k, " = ", key[i])
  }
  paste(parts, collapse = ", ")
}

# Re-open a node from its stored file path (fresh external pointer)
reopen_node <- function(node) {
  path <- node$.path
  ext <- tolower(tools::file_ext(path))
  if (ext == "gz" && grepl("\\.csv\\.gz$", tolower(path))) ext <- "csv"
  switch(ext,
    vtr = tbl(path),
    csv = tbl_csv(path),
    sqlite = , db = tbl_sqlite(path, node$.table),
    stop(sprintf("cannot re-open node from .%s file", ext))
  )
}

#' Look up columns from linked dimension tables
#'
#' Resolves columns from dimension tables registered in a [vtr_schema()],
#' automatically building the necessary join tree. Reports unmatched keys
#' as a diagnostic message.
#'
#' Column references use `dimension$column` syntax (e.g., `species$name`).
#' Columns from the fact table can be referenced by name directly.
#'
#' @param .schema A `vectra_schema` object.
#' @param ... Column references: bare names for fact columns, or
#'   `dimension$column` for dimension columns.
#' @param .join Join type: `"left"` (default, keeps all fact rows) or
#'   `"inner"` (drops unmatched fact rows).
#' @param .report Logical. If `TRUE` (default), print a message with
#'   the number of unmatched keys per dimension.
#'
#' @return A `vectra_node` with the selected columns.
#'
#' @details
#' When `.report = TRUE`, each needed dimension is checked for unmatched keys
#' by opening fresh scans of the fact and dimension tables. This adds one
#' extra read pass per dimension but does not affect the lazy result node.
#'
#' Only dimensions referenced in `...` are joined. Unreferenced dimensions
#' are never scanned.
#'
#' @examples
#' \donttest{
#' f_obs <- tempfile(fileext = ".vtr")
#' f_sp  <- tempfile(fileext = ".vtr")
#' f_ct  <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(sp_id = 1:4, ct_code = c("AT", "DE", "FR", "XX"),
#'                       value = 10:13), f_obs)
#' write_vtr(data.frame(sp_id = 1:3,
#'                       name = c("Oak", "Beech", "Pine")), f_sp)
#' write_vtr(data.frame(ct_code = c("AT", "DE", "FR"),
#'                       gdp = c(400, 3800, 2700)), f_ct)
#'
#' s <- vtr_schema(
#'   fact    = tbl(f_obs),
#'   species = link("sp_id", tbl(f_sp)),
#'   country = link("ct_code", tbl(f_ct))
#' )
#'
#' # Pull columns from any linked dimension
#' result <- lookup(s, value, species$name, country$gdp)
#' collect(result)
#'
#' unlink(c(f_obs, f_sp, f_ct))
#' }
#'
#' @export
lookup <- function(.schema, ..., .join = "left", .report = TRUE) {
  UseMethod("lookup")
}

#' @export
lookup.vectra_schema <- function(.schema, ..., .join = "left", .report = TRUE) {
  if (!.join %in% c("left", "inner"))
    stop(sprintf(".join must be 'left' or 'inner', got '%s'", .join))

  exprs <- eval(substitute(alist(...)))
  if (length(exprs) == 0)
    stop("at least one column reference is required")

  # Parse each expression into {dim, col} pairs
  requests <- parse_lookup_refs(exprs, names(.schema$dims))

  # Determine which dimensions are needed
  needed_dims <- unique(vapply(
    Filter(function(r) !is.null(r$dim), requests),
    function(r) r$dim, character(1)
  ))

  # Report unmatched keys (uses fresh nodes, does not affect the join tree)
  if (.report && length(needed_dims) > 0) {
    report_unmatched(.schema, needed_dims)
  }

  # Build join tree from fresh nodes
  node <- reopen_node(.schema$fact)
  for (dim_name in needed_dims) {
    lnk <- .schema$dims[[dim_name]]
    dim_node <- reopen_node(lnk$node)
    keys <- parse_join_keys(node, dim_node, lnk$key)
    new_xptr <- .Call(C_join_node, node$.node, dim_node$.node,
                      .join, keys$left, keys$right, ".x", ".y")
    node <- structure(list(.node = new_xptr, .path = NULL),
                      class = "vectra_node")
  }

  # Select only the requested columns
  out_names <- vapply(requests, function(r) r$col, character(1))
  expr_lists <- vector("list", length(out_names))
  new_xptr <- .Call(C_project_node, node$.node, out_names, expr_lists)
  structure(list(.node = new_xptr, .path = NULL), class = "vectra_node")
}

# Parse lookup column references: bare names or dim$col expressions
parse_lookup_refs <- function(exprs, dim_names) {
  requests <- vector("list", length(exprs))
  for (i in seq_along(exprs)) {
    e <- exprs[[i]]
    if (is.call(e) && identical(e[[1]], as.name("$"))) {
      dim_name <- as.character(e[[2]])
      col_name <- as.character(e[[3]])
      if (!dim_name %in% dim_names)
        stop(sprintf("dimension '%s' not found in schema (available: %s)",
                      dim_name, paste(dim_names, collapse = ", ")))
      requests[[i]] <- list(dim = dim_name, col = col_name)
    } else {
      requests[[i]] <- list(dim = NULL, col = as.character(e))
    }
  }
  requests
}

# Report unmatched keys for each dimension (opens fresh nodes)
report_unmatched <- function(schema, needed_dims) {
  for (dim_name in needed_dims) {
    lnk <- schema$dims[[dim_name]]

    # Fresh nodes for the anti_join
    fact_node <- reopen_node(schema$fact)
    dim_node <- reopen_node(lnk$node)
    keys <- parse_join_keys(fact_node, dim_node, lnk$key)

    anti_xptr <- .Call(C_join_node, fact_node$.node, dim_node$.node,
                       "anti", keys$left, keys$right, ".x", ".y")
    anti_node <- structure(list(.node = anti_xptr, .path = NULL),
                           class = "vectra_node")
    anti_df <- .Call(C_collect, anti_node$.node)
    n_unmatched <- nrow(anti_df)

    # Count total unique keys in fact table
    fact_node2 <- reopen_node(schema$fact)
    fact_df <- .Call(C_collect, fact_node2$.node)
    n_total <- nrow(fact_df)

    if (n_unmatched > 0) {
      unmatched_keys <- unique(anti_df[[keys$left[1]]])
      preview <- if (length(unmatched_keys) <= 5) {
        paste(format(unmatched_keys, trim = TRUE), collapse = ", ")
      } else {
        paste(c(format(head(unmatched_keys, 5), trim = TRUE), "..."),
              collapse = ", ")
      }
      message(sprintf("%s: %d/%d unmatched keys (%s)",
                      dim_name, n_unmatched, n_total, preview))
    } else {
      message(sprintf("%s: all %d keys matched", dim_name, n_total))
    }
  }
}

# Helper to create a star-schema test setup
make_star_schema <- function() {
  f_obs <- tempfile(fileext = ".vtr")
  f_sp  <- tempfile(fileext = ".vtr")
  f_ct  <- tempfile(fileext = ".vtr")

  write_vtr(data.frame(
    sp_id   = c(1, 2, 3, 4),
    ct_code = c("AT", "DE", "FR", "XX"),
    value   = c(10, 20, 30, 40)
  ), f_obs)

  write_vtr(data.frame(
    sp_id = c(1, 2, 3),
    name  = c("Oak", "Beech", "Pine")
  ), f_sp)

  write_vtr(data.frame(
    ct_code = c("AT", "DE", "FR"),
    gdp     = c(400, 3800, 2700)
  ), f_ct)

  list(f_obs = f_obs, f_sp = f_sp, f_ct = f_ct)
}

# --- link() ---

test_that("link() validates inputs", {
  f <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(x = 1), f)
  on.exit(unlink(f))

  expect_error(link(123, tbl(f)), "key must be")
  expect_error(link(character(0), tbl(f)), "key must be")
  expect_error(link("x", data.frame()), "must be a vectra_node")
})

test_that("link() creates a vectra_link object", {
  f <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(x = 1), f)
  on.exit(unlink(f))

  lnk <- link("x", tbl(f))
  expect_s3_class(lnk, "vectra_link")
  expect_equal(lnk$key, "x")
})

# --- vtr_schema() ---

test_that("vtr_schema() validates inputs", {
  f <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(x = 1), f)
  on.exit(unlink(f))

  expect_error(vtr_schema(fact = data.frame()), "must be a vectra_node")
  expect_error(vtr_schema(fact = tbl(f)), "at least one dimension")
  expect_error(vtr_schema(fact = tbl(f), link("x", tbl(f))),
               "must be named")
  expect_error(vtr_schema(fact = tbl(f), dim1 = "not_a_link"),
               "must be a vectra_link")
})

test_that("vtr_schema() creates a vectra_schema object", {
  ft <- make_star_schema()
  on.exit(unlink(c(ft$f_obs, ft$f_sp, ft$f_ct)))

  s <- vtr_schema(
    fact    = tbl(ft$f_obs),
    species = link("sp_id", tbl(ft$f_sp)),
    country = link("ct_code", tbl(ft$f_ct))
  )
  expect_s3_class(s, "vectra_schema")
  expect_equal(names(s$dims), c("species", "country"))
})

test_that("print.vectra_schema produces output", {
  ft <- make_star_schema()
  on.exit(unlink(c(ft$f_obs, ft$f_sp, ft$f_ct)))

  s <- vtr_schema(
    fact    = tbl(ft$f_obs),
    species = link("sp_id", tbl(ft$f_sp))
  )
  out <- capture.output(print(s))
  expect_true(any(grepl("vectra schema", out)))
  expect_true(any(grepl("species", out)))
})

# --- lookup() ---

test_that("lookup() resolves fact columns by bare name", {
  ft <- make_star_schema()
  on.exit(unlink(c(ft$f_obs, ft$f_sp, ft$f_ct)))

  s <- vtr_schema(
    fact    = tbl(ft$f_obs),
    species = link("sp_id", tbl(ft$f_sp))
  )
  result <- lookup(s, value, .report = FALSE) |> collect()
  expect_equal(result$value, c(10, 20, 30, 40))
})

test_that("lookup() resolves dimension columns via dim$col syntax", {
  ft <- make_star_schema()
  on.exit(unlink(c(ft$f_obs, ft$f_sp, ft$f_ct)))

  s <- vtr_schema(
    fact    = tbl(ft$f_obs),
    species = link("sp_id", tbl(ft$f_sp)),
    country = link("ct_code", tbl(ft$f_ct))
  )
  result <- lookup(s, value, species$name, country$gdp, .report = FALSE) |>
    collect()
  expect_equal(ncol(result), 3)
  expect_equal(names(result), c("value", "name", "gdp"))
  # Row 4 has sp_id=4 which doesn't match species (left join -> NA)
  expect_equal(result$name[1:3], c("Oak", "Beech", "Pine"))
  expect_true(is.na(result$name[4]))
  # Row 4 has ct_code="XX" which doesn't match country
  expect_equal(result$gdp[1:3], c(400, 3800, 2700))
  expect_true(is.na(result$gdp[4]))
})

test_that("lookup() with inner join drops unmatched rows", {
  ft <- make_star_schema()
  on.exit(unlink(c(ft$f_obs, ft$f_sp, ft$f_ct)))

  s <- vtr_schema(
    fact    = tbl(ft$f_obs),
    species = link("sp_id", tbl(ft$f_sp))
  )
  result <- lookup(s, value, species$name, .join = "inner",
                   .report = FALSE) |> collect()
  expect_equal(nrow(result), 3)
  expect_false(any(is.na(result$name)))
})

test_that("lookup() only joins needed dimensions", {
  ft <- make_star_schema()
  on.exit(unlink(c(ft$f_obs, ft$f_sp, ft$f_ct)))

  s <- vtr_schema(
    fact    = tbl(ft$f_obs),
    species = link("sp_id", tbl(ft$f_sp)),
    country = link("ct_code", tbl(ft$f_ct))
  )
  # Only request species columns - country should not be joined
  result <- lookup(s, value, species$name, .report = FALSE) |> collect()
  expect_equal(names(result), c("value", "name"))
})

test_that("lookup() reports unmatched keys", {
  ft <- make_star_schema()
  on.exit(unlink(c(ft$f_obs, ft$f_sp, ft$f_ct)))

  s <- vtr_schema(
    fact    = tbl(ft$f_obs),
    species = link("sp_id", tbl(ft$f_sp)),
    country = link("ct_code", tbl(ft$f_ct))
  )
  msgs <- capture.output(
    result <- lookup(s, value, species$name, country$gdp, .report = TRUE) |>
      collect(),
    type = "message"
  )
  # species: sp_id=4 is unmatched
  expect_true(any(grepl("species.*1/4.*unmatched", msgs)))
  # country: ct_code="XX" is unmatched
  expect_true(any(grepl("country.*1/4.*unmatched", msgs)))
})

test_that("lookup() reports all matched when no mismatches", {
  f_obs <- tempfile(fileext = ".vtr")
  f_sp  <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(sp_id = c(1, 2), value = c(10, 20)), f_obs)
  write_vtr(data.frame(sp_id = c(1, 2), name = c("A", "B")), f_sp)
  on.exit(unlink(c(f_obs, f_sp)))

  s <- vtr_schema(
    fact    = tbl(f_obs),
    species = link("sp_id", tbl(f_sp))
  )
  msgs <- capture.output(
    result <- lookup(s, value, species$name) |> collect(),
    type = "message"
  )
  expect_true(any(grepl("all 2 keys matched", msgs)))
})

test_that("lookup() errors on unknown dimension", {
  ft <- make_star_schema()
  on.exit(unlink(c(ft$f_obs, ft$f_sp, ft$f_ct)))

  s <- vtr_schema(
    fact    = tbl(ft$f_obs),
    species = link("sp_id", tbl(ft$f_sp))
  )
  expect_error(lookup(s, bogus$col, .report = FALSE), "dimension 'bogus'")
})

test_that("lookup() errors with no column references", {
  ft <- make_star_schema()
  on.exit(unlink(c(ft$f_obs, ft$f_sp, ft$f_ct)))

  s <- vtr_schema(
    fact    = tbl(ft$f_obs),
    species = link("sp_id", tbl(ft$f_sp))
  )
  expect_error(lookup(s, .report = FALSE), "at least one column")
})

test_that("lookup() works with named keys (different column names)", {
  f_obs <- tempfile(fileext = ".vtr")
  f_sp  <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(species_id = c(1, 2), value = c(10, 20)), f_obs)
  write_vtr(data.frame(sp_id = c(1, 2), name = c("A", "B")), f_sp)
  on.exit(unlink(c(f_obs, f_sp)))

  s <- vtr_schema(
    fact    = tbl(f_obs),
    species = link(c("species_id" = "sp_id"), tbl(f_sp))
  )
  result <- lookup(s, value, species$name, .report = FALSE) |> collect()
  expect_equal(result$name, c("A", "B"))
})

test_that("schema can be reused for multiple lookups", {
  ft <- make_star_schema()
  on.exit(unlink(c(ft$f_obs, ft$f_sp, ft$f_ct)))

  s <- vtr_schema(
    fact    = tbl(ft$f_obs),
    species = link("sp_id", tbl(ft$f_sp)),
    country = link("ct_code", tbl(ft$f_ct))
  )

  r1 <- lookup(s, value, species$name, .report = FALSE) |> collect()
  r2 <- lookup(s, value, country$gdp, .report = FALSE) |> collect()
  r3 <- lookup(s, species$name, country$gdp, .report = FALSE) |> collect()

  expect_equal(nrow(r1), 4)
  expect_equal(nrow(r2), 4)
  expect_equal(nrow(r3), 4)
  expect_equal(names(r1), c("value", "name"))
  expect_equal(names(r2), c("value", "gdp"))
  expect_equal(names(r3), c("name", "gdp"))
})

#' Materialize a vectra node into a reusable in-memory block
#'
#' Consumes a vectra node (pulling all batches) and stores the result as a
#' persistent columnar block in memory.  Unlike nodes, blocks can be probed
#' repeatedly via [block_lookup()] without re-scanning.
#'
#' @param .data A `vectra_node` (consumed; cannot be used after this call).
#' @return A `vectra_block` object (external pointer to C-level ColumnBlock).
#'
#' @examples
#' \donttest{
#' f <- tempfile(fileext = ".vtr")
#' df <- data.frame(taxonID = 1:3,
#'                  canonicalName = c("Quercus robur", "Pinus sylvestris",
#'                                    "Fagus sylvatica"))
#' write_vtr(df, f)
#' blk <- materialize(tbl(f) |> select(taxonID, canonicalName))
#' hits <- block_lookup(blk, "canonicalName",
#'                      c("Quercus robur", "Pinus sylvestris"))
#' unlink(f)
#' }
#'
#' @export
materialize <- function(.data) {
  UseMethod("materialize")
}

#' @export
materialize.vectra_node <- function(.data) {
  xptr <- .Call(C_block_materialize, .data$.node)
  schema <- tryCatch(.Call(C_node_schema, .data$.node), error = function(e) NULL)
  structure(list(.block = xptr, .path = .data$.path),
            class = "vectra_block")
}

#' @export
print.vectra_block <- function(x, ...) {
  cat("vectra_block [materialized in memory]\n")
  invisible(x)
}


#' Probe a materialized block by column value
#'
#' Performs a hash lookup on a string column of a materialized block.
#' Returns all rows where the column value matches one of the query keys.
#' Hash indices are built lazily on first use and cached for subsequent calls.
#'
#' @param block A `vectra_block` from [materialize()].
#' @param column Character scalar. Name of the string column to match against.
#' @param keys Character vector. Query values to look up.
#' @param ci Logical. Case-insensitive matching (default `FALSE`).
#' @return A data.frame with column `query_idx` (1-based position in `keys`)
#'   plus all columns from the block, for each (query, block_row) match pair.
#'
#' @examples
#' \donttest{
#' f <- tempfile(fileext = ".vtr")
#' df <- data.frame(taxonID = 1:2,
#'                  canonicalName = c("Quercus robur", "Pinus sylvestris"))
#' write_vtr(df, f)
#' blk <- materialize(tbl(f))
#' hits <- block_lookup(blk, "canonicalName", c("Quercus robur"))
#' ci_hits <- block_lookup(blk, "canonicalName", c("quercus robur"), ci = TRUE)
#' unlink(f)
#' }
#'
#' @export
block_lookup <- function(block, column, keys, ci = FALSE) {
  if (!inherits(block, "vectra_block"))
    stop("block must be a vectra_block from materialize()", call. = FALSE)
  .Call(C_block_lookup, block$.block, column, keys, ci)
}


#' Fuzzy-match query keys against a materialized block
#'
#' Computes string distances between query keys and a string column in a
#' materialized block. Optionally uses exact-match blocking on a second column
#' (e.g., genus) to reduce the search space.
#'
#' @param block A `vectra_block` from [materialize()].
#' @param column Character scalar. Name of the string column to fuzzy-match against.
#' @param keys Character vector. Query strings to match.
#' @param method Character. Distance method: `"dl"` (Damerau-Levenshtein, default),
#'   `"levenshtein"`, or `"jw"` (Jaro-Winkler).
#' @param max_dist Numeric. Maximum normalized distance (default 0.2).
#' @param block_col Optional character scalar. Column name for exact-match blocking
#'   (e.g., genus). When provided, only rows where `block_col` matches the
#'   corresponding `block_keys` value are compared.
#' @param block_keys Optional character vector (same length as `keys`). Exact-match
#'   values for blocking. Required when `block_col` is provided.
#' @param n_threads Integer. Number of OpenMP threads (default 4L).
#' @return A data.frame with columns `query_idx` (1-based position in `keys`),
#'   `fuzzy_dist` (normalized distance), plus all columns from the block.
#'
#' @export
block_fuzzy_lookup <- function(block, column, keys, method = "dl",
                               max_dist = 0.2, block_col = NULL,
                               block_keys = NULL, n_threads = 4L) {
  if (!inherits(block, "vectra_block"))
    stop("block must be a vectra_block from materialize()", call. = FALSE)
  method_int <- match(method, c("dl", "levenshtein", "jw")) - 1L
  if (is.na(method_int))
    stop("method must be 'dl', 'levenshtein', or 'jw'", call. = FALSE)
  .Call(C_block_fuzzy_lookup, block$.block, column, keys,
        as.integer(method_int), as.double(max_dist),
        block_col, block_keys, as.integer(n_threads))
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----basic-data---------------------------------------------------------------
library(vectra)

trees <- data.frame(
  species  = c("Quercus robur  ", "  fagus sylvatica", "ACER platanoides",
               "Betula pendula", "  Pinus SYLVESTRIS  "),
  site     = c("plot_A01", "plot_B12", "plot_A01", "plot_C03", "plot_B12"),
  observer = c("J. Smith", "A. Mueller", "J. Smith",
               "B. Novak", "A. Mueller"),
  dbh_cm   = c(45.2, 38.1, 22.7, 31.0, 55.3),
  stringsAsFactors = FALSE
)

f <- tempfile(fileext = ".vtr")
write_vtr(trees, f)

## ----trimws-------------------------------------------------------------------
tbl(f) |>
  mutate(species_clean = trimws(species)) |>
  select(species, species_clean) |>
  collect()

## ----case---------------------------------------------------------------------
tbl(f) |>
  mutate(species_lower = tolower(trimws(species)),
         species_upper = toupper(trimws(species))) |>
  select(species_lower, species_upper) |>
  collect()

## ----nchar--------------------------------------------------------------------
tbl(f) |>
  mutate(name_len = nchar(trimws(species))) |>
  select(species, name_len) |>
  collect()

## ----substr-------------------------------------------------------------------
tbl(f) |>
  mutate(plot_letter = substr(site, 6, 6),
         plot_number = substr(site, 7, 8)) |>
  select(site, plot_letter, plot_number) |>
  collect()

## ----starts-ends--------------------------------------------------------------
tbl(f) |>
  filter(startsWith(site, "plot_A")) |>
  select(species, site) |>
  collect()

## ----endswith-----------------------------------------------------------------
tbl(f) |>
  mutate(is_mueller = endsWith(observer, "Mueller")) |>
  select(observer, is_mueller) |>
  collect()

## ----paste0-------------------------------------------------------------------
tbl(f) |>
  mutate(record_id = paste0(site, "_", trimws(species))) |>
  select(record_id) |>
  collect()

## ----paste-sep----------------------------------------------------------------
tbl(f) |>
  mutate(label = paste(observer, site, sep = " @ ")) |>
  select(label) |>
  collect()

## ----paste-multi--------------------------------------------------------------
tbl(f) |>
  mutate(full_label = paste(observer, site,
                            trimws(species), sep = " | ")) |>
  select(full_label) |>
  collect()

## ----grepl-fixed--------------------------------------------------------------
tbl(f) |>
  filter(grepl("Smith", observer, fixed = TRUE)) |>
  collect()

## ----gsub-fixed---------------------------------------------------------------
messy_sites <- data.frame(
  site = c("plot_A01", "plot-B12", "plot_A01", "plot-C03", "plot_B12"),
  stringsAsFactors = FALSE
)
f2 <- tempfile(fileext = ".vtr")
write_vtr(messy_sites, f2)

tbl(f2) |>
  mutate(site_clean = gsub("-", "_", site, fixed = TRUE)) |>
  collect()

## ----sub-fixed----------------------------------------------------------------
notes <- data.frame(
  note = c("tree dead, bark loose, dead branches",
           "alive, healthy canopy",
           "dead standing, no bark"),
  stringsAsFactors = FALSE
)
f3 <- tempfile(fileext = ".vtr")
write_vtr(notes, f3)

tbl(f3) |>
  mutate(note_edited = sub("dead", "DEAD", note, fixed = TRUE)) |>
  select(note, note_edited) |>
  collect()

## ----grepl-regex--------------------------------------------------------------
species_list <- data.frame(
  name = c("Quercus robur", "fagus sylvatica", "ACER PLATANOIDES",
           "Betula pendula", "Pinus  sylvestris", "Tilia cordata"),
  stringsAsFactors = FALSE
)
f4 <- tempfile(fileext = ".vtr")
write_vtr(species_list, f4)

tbl(f4) |>
  mutate(valid_format = grepl("^[A-Z][a-z]+ [a-z]+$",
                              name, fixed = FALSE)) |>
  collect()

## ----gsub-regex---------------------------------------------------------------
tbl(f4) |>
  mutate(short_name = gsub("^([A-Za-z])[a-z]+ ([a-z]+)$",
                           "\\1. \\2",
                           name, fixed = FALSE)) |>
  select(name, short_name) |>
  collect()

## ----sub-regex----------------------------------------------------------------
records <- data.frame(
  code = c("ID:001-Quercus", "ID:042-Fagus",
           "ID:007-Betula", "ID:113-Pinus"),
  stringsAsFactors = FALSE
)
f5 <- tempfile(fileext = ".vtr")
write_vtr(records, f5)

tbl(f5) |>
  mutate(genus = sub("^ID:[0-9]+-", "", code, fixed = FALSE)) |>
  select(code, genus) |>
  collect()

## ----str-extract--------------------------------------------------------------
sites <- data.frame(
  site_code = c("Forest_Plot_042", "Meadow_Transect_007",
                "Forest_Plot_113", "Wetland_Quad_019"),
  stringsAsFactors = FALSE
)
f6 <- tempfile(fileext = ".vtr")
write_vtr(sites, f6)

tbl(f6) |>
  mutate(plot_num = str_extract(site_code, "([0-9]+)")) |>
  select(site_code, plot_num) |>
  collect()

## ----str-extract-habitat------------------------------------------------------
tbl(f6) |>
  mutate(habitat = str_extract(site_code, "^([A-Za-z]+)_")) |>
  select(site_code, habitat) |>
  collect()

## ----levenshtein-data---------------------------------------------------------
typos <- data.frame(
  field_name = c("Qurecus robur", "Fagus silvatica",
                 "Acer platanodes", "Betula pendula",
                 "Pinus sylvestrs"),
  ref_name   = c("Quercus robur", "Fagus sylvatica",
                 "Acer platanoides", "Betula pendula",
                 "Pinus sylvestris"),
  stringsAsFactors = FALSE
)
f7 <- tempfile(fileext = ".vtr")
write_vtr(typos, f7)

tbl(f7) |>
  mutate(lev = levenshtein(field_name, ref_name),
         lev_norm = levenshtein_norm(field_name, ref_name)) |>
  collect()

## ----dl-dist------------------------------------------------------------------
tbl(f7) |>
  mutate(dl = dl_dist(field_name, ref_name),
         dl_norm = dl_dist_norm(field_name, ref_name)) |>
  collect()

## ----jaro-winkler-------------------------------------------------------------
tbl(f7) |>
  mutate(jw = jaro_winkler(field_name, ref_name)) |>
  collect()

## ----filter-fuzzy-------------------------------------------------------------
field_data <- data.frame(
  species = c("Qurecus robur", "Quercus robor", "Fagus sylvatica",
              "Quercus robur", "Quercis rubur", "Betula pendula"),
  plot    = c("A1", "A2", "B1", "A3", "B2", "C1"),
  stringsAsFactors = FALSE
)
f8 <- tempfile(fileext = ".vtr")
write_vtr(field_data, f8)

tbl(f8) |>
  mutate(dist = dl_dist_norm(species, "Quercus robur")) |>
  filter(dist < 0.15) |>
  collect()

## ----rank-fuzzy---------------------------------------------------------------
tbl(f8) |>
  mutate(sim = jaro_winkler(species, "Quercus robur")) |>
  arrange(desc(sim)) |>
  collect()

## ----fuzzy-join-basic---------------------------------------------------------
ref <- data.frame(
  canonical = c("Quercus robur", "Fagus sylvatica",
                "Acer platanoides", "Betula pendula",
                "Pinus sylvestris", "Tilia cordata"),
  family    = c("Fagaceae", "Fagaceae", "Sapindaceae",
                "Betulaceae", "Pinaceae", "Malvaceae"),
  stringsAsFactors = FALSE
)
f_ref <- tempfile(fileext = ".vtr")
write_vtr(ref, f_ref)

messy <- data.frame(
  field_species = c("Qurecus robur", "Fagus silvatica",
                    "Acer platanodes", "Betla pendula",
                    "Pinis sylvestris"),
  count         = c(12L, 7L, 3L, 15L, 9L),
  stringsAsFactors = FALSE
)
f_messy <- tempfile(fileext = ".vtr")
write_vtr(messy, f_messy)

fuzzy_join(
  tbl(f_messy), tbl(f_ref),
  by = c("field_species" = "canonical"),
  method = "dl",
  max_dist = 0.25
) |> collect()

## ----fuzzy-join-jw------------------------------------------------------------
fuzzy_join(
  tbl(f_messy), tbl(f_ref),
  by = c("field_species" = "canonical"),
  method = "jw",
  max_dist = 0.15
) |> collect()

## ----fuzzy-join-blocked-------------------------------------------------------
ref_blocked <- data.frame(
  genus     = c("Quercus", "Fagus", "Acer",
                "Betula", "Pinus", "Tilia"),
  canonical = c("Quercus robur", "Fagus sylvatica",
                "Acer platanoides", "Betula pendula",
                "Pinus sylvestris", "Tilia cordata"),
  family    = c("Fagaceae", "Fagaceae", "Sapindaceae",
                "Betulaceae", "Pinaceae", "Malvaceae"),
  stringsAsFactors = FALSE
)
f_ref2 <- tempfile(fileext = ".vtr")
write_vtr(ref_blocked, f_ref2)

messy_blocked <- data.frame(
  genus_field   = c("Quercus", "Fagus", "Acer",
                    "Betula", "Pinus"),
  field_species = c("Qurecus robur", "Fagus silvatica",
                    "Acer platanodes", "Betla pendula",
                    "Pinis sylvestris"),
  count         = c(12L, 7L, 3L, 15L, 9L),
  stringsAsFactors = FALSE
)
f_messy2 <- tempfile(fileext = ".vtr")
write_vtr(messy_blocked, f_messy2)

fuzzy_join(
  tbl(f_messy2), tbl(f_ref2),
  by = c("field_species" = "canonical"),
  method = "dl",
  max_dist = 0.25,
  block_by = c("genus_field" = "genus")
) |> collect()

## ----fuzzy-join-threads, eval = FALSE-----------------------------------------
# fuzzy_join(
#   tbl(f_messy), tbl(f_ref),
#   by = c("field_species" = "canonical"),
#   method = "dl",
#   max_dist = 0.25,
#   n_threads = 8L
# ) |> collect()

## ----block-lookup-------------------------------------------------------------
blk <- tbl(f_ref) |>
  select(canonical, family) |>
  materialize()

block_lookup(blk, "canonical",
             c("Quercus robur", "Betula pendula"))

## ----block-lookup-ci----------------------------------------------------------
block_lookup(blk, "canonical",
             c("quercus robur", "BETULA PENDULA"),
             ci = TRUE)

## ----block-fuzzy--------------------------------------------------------------
blk2 <- tbl(f_ref2) |>
  select(genus, canonical, family) |>
  materialize()

block_fuzzy_lookup(
  blk2, "canonical",
  keys = c("Qurecus robur", "Fagus silvatica"),
  method = "dl",
  max_dist = 0.2,
  block_col = "genus",
  block_keys = c("Quercus", "Fagus"),
  n_threads = 2L
)

## ----cleaning-pattern---------------------------------------------------------
raw_names <- data.frame(
  species = c("  Quercus ROBUR ", "fagus sylvatica.",
              "Acer platanoides (L.)", "BETULA   pendula"),
  stringsAsFactors = FALSE
)
f9 <- tempfile(fileext = ".vtr")
write_vtr(raw_names, f9)

tbl(f9) |>
  mutate(clean = tolower(trimws(species))) |>
  mutate(clean = gsub(".", "", clean, fixed = TRUE)) |>
  mutate(clean = gsub(" +", " ", clean, fixed = FALSE)) |>
  select(species, clean) |>
  collect()

## ----layered-matching---------------------------------------------------------
ref_clean <- data.frame(
  canonical = c("quercus robur", "fagus sylvatica",
                "acer platanoides", "betula pendula"),
  status    = c("accepted", "accepted", "accepted", "accepted"),
  stringsAsFactors = FALSE
)
f_refc <- tempfile(fileext = ".vtr")
write_vtr(ref_clean, f_refc)

cleaned <- tbl(f9) |>
  mutate(clean = tolower(trimws(species))) |>
  mutate(clean = gsub(".", "", clean, fixed = TRUE)) |>
  mutate(clean = gsub(" +", " ", clean, fixed = FALSE))

# Step 1: exact join on cleaned names
exact <- left_join(
  cleaned, tbl(f_refc),
  by = c("clean" = "canonical")
) |> collect()

exact

#' Bind rows or columns from multiple vectra tables
#'
#' @param ... `vectra_node` objects or data.frames to combine.
#' @param .id Optional column name for a source identifier.
#'
#' @return A `vectra_node` (streaming) when all inputs are `vectra_node` with
#'   identical schemas and `.id` is NULL. Otherwise a data.frame.
#'
#' @details
#' When all inputs are `vectra_node` objects with identical column names and
#' types and no `.id` is requested, `bind_rows` creates a streaming
#' `ConcatNode` that iterates children sequentially without materializing.
#'
#' Otherwise, inputs are collected and combined in R. Missing columns are
#' filled with NA.
#'
#' `bind_cols` requires the same number of rows in each input.
#'
#' @examples
#' f1 <- tempfile(fileext = ".vtr")
#' f2 <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(x = 1:3, y = 4:6), f1)
#' write_vtr(data.frame(x = 7:9, y = 10:12), f2)
#' bind_rows(tbl(f1), tbl(f2)) |> collect()
#' bind_cols(tbl(f1), tbl(f2))
#' unlink(c(f1, f2))
#'
#' @export
bind_rows <- function(..., .id = NULL) {
  dots <- list(...)

  # Check if we can use streaming ConcatNode
  all_nodes <- all(vapply(dots, inherits, logical(1), "vectra_node"))
  if (all_nodes && is.null(.id) && length(dots) >= 2) {
    schemas <- lapply(dots, function(x) .Call(C_node_schema, x$.node))
    ref <- schemas[[1]]
    # Require same column names; types can differ (C engine coerces)
    names_match <- all(vapply(schemas[-1], function(s) {
      identical(s$name, ref$name)
    }, logical(1)))

    if (names_match) {
      xptrs <- lapply(dots, function(x) x$.node)
      new_xptr <- .Call(C_concat_node, xptrs)
      return(structure(list(.node = new_xptr, .path = NULL),
                       class = "vectra_node"))
    }
  }

  # Fallback: collect and combine in R
  dfs <- lapply(dots, function(x) {
    if (inherits(x, "vectra_node")) collect(x) else x
  })

  all_names <- unique(unlist(lapply(dfs, names)))

  aligned <- lapply(dfs, function(df) {
    missing <- setdiff(all_names, names(df))
    for (nm in missing) df[[nm]] <- NA
    df[all_names]
  })

  result <- do.call(rbind, aligned)
  rownames(result) <- NULL

  if (!is.null(.id)) {
    src <- rep(seq_along(aligned), vapply(aligned, nrow, integer(1)))
    result <- cbind(stats::setNames(data.frame(src), .id), result)
  }

  result
}

#' @rdname bind_rows
#' @export
bind_cols <- function(...) {
  dots <- list(...)
  dfs <- lapply(dots, function(x) {
    if (inherits(x, "vectra_node")) collect(x) else x
  })

  nrows <- vapply(dfs, nrow, integer(1))
  if (length(unique(nrows)) > 1)
    stop(sprintf("all inputs must have the same number of rows (got %s)",
                 paste(unique(nrows), collapse = ", ")))

  do.call(cbind, dfs)
}

#' @useDynLib vectra, .registration = TRUE
#' @importFrom rlang expr enquo is_formula as_function
#' @importFrom stats sd var median
#' @importFrom libgeos libgeos_version
NULL

# -- package-level constants --------------------------------------------------

.DEFAULT_BATCH_SIZE <- 65536L
.TIFF_BATCH_SIZE <- 256L
.TYPE_INFERENCE_ROWS <- 1000L

# Build a 0-row data.frame proxy from schema, enabling where() etc.
# Maps vectra types to R column types so tidyselect predicates work.
# Annotations override the base type (Date, POSIXct -> double with class;
# factor -> integer with levels).
schema_proxy <- function(schema) {
  n <- length(schema$name)
  ann <- if (!is.null(schema$annotation)) schema$annotation else rep(NA_character_, n)
  cols <- vector("list", n)
  for (i in seq_len(n)) {
    a <- ann[i]
    if (!is.na(a) && startsWith(a, "factor")) {
      cols[[i]] <- factor(character(0))
    } else if (!is.na(a) && a == "Date") {
      cols[[i]] <- structure(double(0), class = "Date")
    } else if (!is.na(a) && startsWith(a, "POSIXct")) {
      cols[[i]] <- structure(double(0), class = c("POSIXct", "POSIXt"))
    } else {
      cols[[i]] <- switch(schema$type[i],
        int64  = integer(0),
        double = double(0),
        bool   = logical(0),
        string = character(0),
        double(0)
      )
    }
  }
  names(cols) <- schema$name
  structure(cols, class = "data.frame", row.names = integer(0))
}

.onLoad <- function(libname, pkgname) {
  # Nothing needed; .Call registration is done in init.c
}

.onUnload <- function(libpath) {
  library.dynam.unload("vectra", libpath)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----synthetic-data-----------------------------------------------------------
library(vectra)

set.seed(42)

n <- 50000
species_pool <- c(
  "Quercus robur", "Fagus sylvatica", "Pinus sylvestris",
  "Betula pendula", "Acer pseudoplatanus", "Fraxinus excelsior",
  "Picea abies", "Tilia cordata", "Carpinus betulus",
  "Alnus glutinosa", "Sorbus aucuparia", "Ulmus glabra"
)
sites <- paste0("SITE_", sprintf("%03d", 1:50))
years <- 2015:2024

obs <- data.frame(
  obs_id    = seq_len(n),
  site      = sample(sites, n, replace = TRUE),
  species   = sample(species_pool, n, replace = TRUE),
  year      = sample(years, n, replace = TRUE),
  abundance = rpois(n, lambda = 15),
  cover_pct = round(runif(n, 0.1, 95.0), 1),
  quality   = sample(c("good", "moderate", "poor"), n,
                     replace = TRUE, prob = c(0.6, 0.3, 0.1)),
  stringsAsFactors = FALSE
)

## ----write-sources------------------------------------------------------------
csv_path <- tempfile(fileext = ".csv")
write.csv(obs, csv_path, row.names = FALSE)

vtr_path <- tempfile(fileext = ".vtr")
write_vtr(obs, vtr_path)

## ----streaming-basic----------------------------------------------------------
clean_path <- tempfile(fileext = ".vtr")

tbl_csv(csv_path) |>
  filter(quality == "good") |>
  mutate(log_abundance = log(abundance + 1)) |>
  write_vtr(clean_path)

## ----verify-streaming---------------------------------------------------------
tbl(clean_path) |>
  select(obs_id, site, species, abundance, log_abundance) |>
  slice_head(n = 5) |>
  collect()

## ----csv-to-sqlite------------------------------------------------------------
db_path <- tempfile(fileext = ".sqlite")

tbl_csv(csv_path) |>
  filter(year >= 2020) |>
  write_sqlite(db_path, "recent_obs")

## ----vtr-to-csv---------------------------------------------------------------
export_csv <- tempfile(fileext = ".csv")

tbl(vtr_path) |>
  filter(species == "Fagus sylvatica") |>
  select(obs_id, site, year, abundance) |>
  write_csv(export_csv)

## ----collect-agg--------------------------------------------------------------
tbl(vtr_path) |>
  group_by(species, year) |>
  summarise(
    n_obs     = n(),
    mean_abun = mean(abundance),
    mean_cov  = mean(cover_pct)
  ) |>
  arrange(species, year) |>
  collect() |>
  head(10)

## ----batch-size-write---------------------------------------------------------
small_groups <- tempfile(fileext = ".vtr")
large_groups <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  arrange(year) |>
  write_vtr(small_groups, batch_size = 5000)

tbl(vtr_path) |>
  arrange(year) |>
  write_vtr(large_groups, batch_size = 50000)

## ----batch-size-compare-------------------------------------------------------
tbl(small_groups) |>
  filter(year == 2023) |>
  explain()

## ----csv-batch-size-----------------------------------------------------------
tbl_csv(csv_path, batch_size = 10000) |>
  filter(quality == "good") |>
  slice_head(n = 3) |>
  collect()

## ----append-init--------------------------------------------------------------
archive <- tempfile(fileext = ".vtr")
first_year <- obs[obs$year == 2015, ]
write_vtr(first_year, archive)

nrow(tbl(archive) |> collect())

## ----append-year--------------------------------------------------------------
year_2016 <- obs[obs$year == 2016, ]
append_vtr(year_2016, archive)

nrow(tbl(archive) |> collect())

## ----append-streaming---------------------------------------------------------
csv_2017 <- tempfile(fileext = ".csv")
write.csv(obs[obs$year == 2017, ], csv_2017, row.names = FALSE)

tbl_csv(csv_2017) |> append_vtr(archive)

nrow(tbl(archive) |> collect())

## ----compact------------------------------------------------------------------
compacted <- tempfile(fileext = ".vtr")
tbl(archive) |> write_vtr(compacted, batch_size = 50000)

## ----append-schema------------------------------------------------------------
new_data_node <- tbl_csv(csv_2017) |>
  mutate(observer = NA_character_)

## ----delete-basic-------------------------------------------------------------
del_demo <- tempfile(fileext = ".vtr")
write_vtr(obs[1:100, ], del_demo)

# Delete rows 0, 1, and 99 (first two and last)
delete_vtr(del_demo, c(0, 1, 99))

tbl(del_demo) |> collect() |> nrow()

## ----delete-cumulative--------------------------------------------------------
delete_vtr(del_demo, c(10, 11, 12))

tbl(del_demo) |> collect() |> nrow()

## ----delete-undo--------------------------------------------------------------
unlink(paste0(del_demo, ".del"))
tbl(del_demo) |> collect() |> nrow()

## ----delete-compact-----------------------------------------------------------
delete_vtr(del_demo, 0:49)
compacted_del <- tempfile(fileext = ".vtr")
tbl(del_demo) |> write_vtr(compacted_del)

nrow(tbl(compacted_del) |> collect())

## ----diff-setup---------------------------------------------------------------
snap_v1 <- tempfile(fileext = ".vtr")
snap_v2 <- tempfile(fileext = ".vtr")

# Version 1: observations 1-100
write_vtr(obs[1:100, ], snap_v1)

# Version 2: observations 51-150 (rows 1-50 removed, 101-150 added)
write_vtr(obs[51:150, ], snap_v2)

## ----diff-compute-------------------------------------------------------------
changes <- diff_vtr(snap_v1, snap_v2, "obs_id")

# Keys that disappeared
head(changes$deleted)
length(changes$deleted)

## ----diff-added---------------------------------------------------------------
collect(changes$added) |> head()

## ----diff-append--------------------------------------------------------------
archive_diff <- tempfile(fileext = ".vtr")
write_vtr(obs[1:100, ], archive_diff)

changes$added |> append_vtr(archive_diff)

nrow(tbl(archive_diff) |> collect())

## ----external-sort------------------------------------------------------------
sorted_path <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  arrange(species, desc(abundance)) |>
  write_vtr(sorted_path)

tbl(sorted_path) |>
  select(species, abundance, site) |>
  slice_head(n = 8) |>
  collect()

## ----join-setup---------------------------------------------------------------
# Small reference table: site metadata
site_meta <- data.frame(
  site      = sites,
  region    = sample(c("North", "South", "East", "West"),
                     length(sites), replace = TRUE),
  elevation = round(runif(length(sites), 100, 2500)),
  stringsAsFactors = FALSE
)
site_path <- tempfile(fileext = ".vtr")
write_vtr(site_meta, site_path)

## ----join-streaming-----------------------------------------------------------
enriched <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  left_join(tbl(site_path), by = "site") |>
  write_vtr(enriched)

tbl(enriched) |>
  select(obs_id, site, region, elevation, species) |>
  slice_head(n = 5) |>
  collect()

## ----join-filter-right--------------------------------------------------------
joined_path <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  filter(year >= 2022) |>
  inner_join(
    tbl(site_path) |> filter(region == "North"),
    by = "site"
  ) |>
  write_vtr(joined_path)

nrow(tbl(joined_path) |> collect())

## ----join-pre-agg-------------------------------------------------------------
# Right side: per-site-year summary from a second dataset
summary_path <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  group_by(site, year) |>
  summarise(site_year_avg = mean(abundance)) |>
  write_vtr(summary_path)

# Join the summary back to the detail table
tbl(vtr_path) |>
  left_join(tbl(summary_path), by = c("site", "year")) |>
  select(obs_id, site, year, abundance, site_year_avg) |>
  slice_head(n = 5) |>
  collect()

## ----multifile-setup----------------------------------------------------------
monthly_paths <- character(6)
for (m in 1:6) {
  monthly_paths[m] <- tempfile(fileext = ".vtr")
  idx <- which(obs$year == 2024 & ((obs$obs_id %% 6) + 1) == m)
  month_data <- obs[idx[seq_len(min(200, length(idx)))], ]
  write_vtr(month_data, monthly_paths[m])
}

## ----multifile-combine--------------------------------------------------------
nodes <- lapply(monthly_paths, tbl)
combined <- do.call(bind_rows, nodes)

combined |>
  group_by(species) |>
  summarise(
    total_obs  = n(),
    mean_abun  = mean(abundance)
  ) |>
  arrange(desc(total_obs)) |>
  slice_head(n = 5) |>
  collect()

## ----multifile-consolidate----------------------------------------------------
consolidated <- tempfile(fileext = ".vtr")
nodes2 <- lapply(monthly_paths, tbl)
do.call(bind_rows, nodes2) |> write_vtr(consolidated)

nrow(tbl(consolidated) |> collect())

## ----multifile-append---------------------------------------------------------
running_archive <- tempfile(fileext = ".vtr")
initial_nodes <- lapply(monthly_paths[1:3], tbl)
do.call(bind_rows, initial_nodes) |> write_vtr(running_archive)

# Next month's files arrive
new_nodes <- lapply(monthly_paths[4:6], tbl)
do.call(bind_rows, new_nodes) |> append_vtr(running_archive)

nrow(tbl(running_archive) |> collect())

## ----etl-basic----------------------------------------------------------------
vtr_archive <- tempfile(fileext = ".vtr")
tbl_csv(csv_path) |> write_vtr(vtr_archive)

## ----etl-clean----------------------------------------------------------------
clean_archive <- tempfile(fileext = ".vtr")

tbl_csv(csv_path) |>
  filter(quality != "poor") |>
  mutate(
    abundance_log = log(abundance + 1),
    cover_frac    = cover_pct / 100
  ) |>
  select(-quality) |>
  arrange(site, year) |>
  write_vtr(clean_archive, batch_size = 10000)

## ----etl-sqlite---------------------------------------------------------------
sqlite_export <- tempfile(fileext = ".sqlite")

tbl(clean_archive) |>
  filter(year >= 2020) |>
  write_sqlite(sqlite_export, "observations")

## ----etl-summary-csv----------------------------------------------------------
summary_csv <- tempfile(fileext = ".csv")

tbl(clean_archive) |>
  group_by(site, year) |>
  summarise(
    n_species = n_distinct(species),
    total_abundance = sum(abundance)
  ) |>
  write_csv(summary_csv)

read.csv(summary_csv) |> head()

## ----etl-multi----------------------------------------------------------------
csv_north <- tempfile(fileext = ".csv")
csv_south <- tempfile(fileext = ".csv")
write.csv(obs[1:25000, ], csv_north, row.names = FALSE)
write.csv(obs[25001:50000, ], csv_south, row.names = FALSE)

merged_vtr <- tempfile(fileext = ".vtr")

bind_rows(
  tbl_csv(csv_north),
  tbl_csv(csv_south)
) |>
  filter(abundance > 0) |>
  write_vtr(merged_vtr, batch_size = 25000)

nrow(tbl(merged_vtr) |> collect())

## ----memory-estimate----------------------------------------------------------
# Example pipeline:
# tbl(huge.vtr) |>
#   filter(year == 2023) |>        -> streaming, ~5 MB per batch
#   left_join(sites, by = "site")  -> build side = 50 sites, ~1 KB
#   group_by(species) |>           -> 12 groups, ~1 KB
#   summarise(total = sum(abun))   -> 12 accumulators, ~1 KB
#   arrange(desc(total))           -> 12 rows, in-memory
#
# Total peak: ~5 MB (one batch) + negligible join/agg overhead

# Compare to:
# tbl(huge.vtr) |>
#   arrange(species)               -> external sort, up to 1 GB
#   left_join(big_ref, by = "id")  -> build side = big_ref size
#
# Total peak: 1 GB (sort) + big_ref size

## ----memory-pipeline----------------------------------------------------------
final_path <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  filter(year >= 2020, quality != "poor") |>
  left_join(tbl(site_path), by = "site") |>
  group_by(region, species) |>
  summarise(
    n_obs     = n(),
    mean_cov  = mean(cover_pct)
  ) |>
  arrange(region, desc(n_obs)) |>
  write_vtr(final_path)

tbl(final_path) |> collect() |> head(10)

## ----cleanup------------------------------------------------------------------
all_files <- c(
  csv_path, vtr_path, clean_path, db_path, export_csv,
  del_demo, paste0(del_demo, ".del"),
  snap_v1, snap_v2,
  sorted_path, site_path, enriched, joined_path,
  summary_path, monthly_paths, consolidated, running_archive,
  vtr_archive, clean_archive, sqlite_export, summary_csv,
  csv_north, csv_south, merged_vtr,
  small_groups, large_groups,
  archive, csv_2017, archive_diff, compacted, compacted_del,
  final_path
)
unlink(all_files[file.exists(all_files)])

#' Compute the logical diff between two .vtr files
#'
#' Streams both files and computes a set-level diff keyed on `key_col`.
#' Returns a list with two elements:
#'
#' - `added`: a `vectra_node` (lazy [tbl()]) of rows present in `new_path`
#'   but not `old_path` (matched on `key_col`).  Call [collect()] to
#'   materialise.  The underlying temp file is deleted when the node is
#'   garbage-collected **or** when the calling R session ends via
#'   `on.exit()`.
#' - `deleted`: a vector of key values present in `old_path` but not
#'   `new_path`.
#'
#' This is a **logical diff** (key-based set difference), not a binary file
#' diff.  Rows with the same key that have changed values are not reported
#' as modified — use `added` and `deleted` together to detect updates (a key
#' that appears in both means a row was replaced).
#'
#' @param old_path Path to the older `.vtr` file.
#' @param new_path Path to the newer `.vtr` file.
#' @param key_col  Name of the column to use as the row key (must exist in
#'   both files with the same type).
#'
#' @return A named list with elements `added` (a `vectra_node`) and `deleted`
#'   (a vector of key values).
#'
#' @examples
#' f1 <- tempfile(fileext = ".vtr")
#' f2 <- tempfile(fileext = ".vtr")
#' df1 <- data.frame(id = 1:5, val = letters[1:5], stringsAsFactors = FALSE)
#' df2 <- data.frame(id = c(3L, 4L, 5L, 6L, 7L),
#'                   val = c("C", "d", "e", "f", "g"),
#'                   stringsAsFactors = FALSE)
#' write_vtr(df1, f1)
#' write_vtr(df2, f2)
#'
#' d <- diff_vtr(f1, f2, "id")
#' # Rows 1 and 2 deleted; rows 6 and 7 added
#' stopifnot(all(d$deleted %in% c(1, 2)))
#' stopifnot(all(collect(d$added)$id %in% c(6, 7)))
#'
#' unlink(c(f1, f2))
#'
#' @export
diff_vtr <- function(old_path, new_path, key_col) {
  if (!is.character(old_path) || length(old_path) != 1)
    stop("old_path must be a single character string")
  if (!is.character(new_path) || length(new_path) != 1)
    stop("new_path must be a single character string")
  if (!is.character(key_col) || length(key_col) != 1)
    stop("key_col must be a single character string")

  old_path <- normalizePath(old_path, mustWork = TRUE)
  new_path <- normalizePath(new_path, mustWork = TRUE)

  # Delegate key-set diff to C.
  # Pass 1: stream A key column -> build hash set.
  # Pass 2: stream ALL B columns; write added rows to a temp .vtr file.
  # Returns list(added_path = <string>, deleted_keys = <vector>).
  raw          <- .Call(C_diff_vtr, old_path, new_path, key_col)
  added_path   <- raw$added_path
  deleted_keys <- raw$deleted_keys

  # Wrap the temp file as a lazy vectra_node.
  # Register cleanup so the temp file is removed when this frame exits or
  # when the caller discards the result (whichever comes first via reg.finalizer
  # on the node's external pointer is handled by the C side; here we also
  # schedule an on.exit in the *caller's* frame via sys.on.exit if available,
  # but the simplest and correct approach is to return the path alongside the
  # node so the caller can unlink() it when done.  The node itself does NOT
  # own the file — only the R session's temp directory ownership applies).
  added_node <- tbl(added_path)

  # Attach the temp path to the node so callers can clean up explicitly if
  # desired, and schedule removal when the node is GC'd via a weak reference.
  attr(added_node, ".tmp_path") <- added_path
  reg.finalizer(added_node$.node,
                function(e) unlink(added_path),
                onexit = TRUE)

  list(added = added_node, deleted = deleted_keys)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----intro-demo---------------------------------------------------------------
library(vectra)

# Write mtcars to .vtr, then read it back lazily
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
node <- tbl(f)
node

## ----vtr-roundtrip------------------------------------------------------------
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

tbl(f) |>
  filter(cyl == 6) |>
  select(mpg, cyl, hp) |>
  collect()

## ----vtr-batch-size-----------------------------------------------------------
f <- tempfile(fileext = ".vtr")
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)

# Convert CSV to .vtr with 10-row row groups
tbl_csv(csv) |> write_vtr(f, batch_size = 10)

# The file now has multiple row groups
tbl(f) |> collect() |> nrow()

## ----csv-read-----------------------------------------------------------------
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)

tbl_csv(csv) |>
  filter(hp > 200) |>
  select(mpg, hp, wt) |>
  collect()

## ----csv-write----------------------------------------------------------------
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

out_csv <- tempfile(fileext = ".csv")
tbl(f) |>
  filter(cyl == 4) |>
  write_csv(out_csv)

# Verify the output
read.csv(out_csv) |> head()

## ----sqlite-read--------------------------------------------------------------
db <- tempfile(fileext = ".sqlite")
write_sqlite(mtcars, db, "cars")

tbl_sqlite(db, "cars") |>
  filter(mpg > 25) |>
  select(mpg, cyl, wt) |>
  collect()

## ----sqlite-write-------------------------------------------------------------
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

db <- tempfile(fileext = ".sqlite")
tbl(f) |>
  filter(cyl == 8) |>
  write_sqlite(db, "v8_cars")

# Read it back through vectra
tbl_sqlite(db, "v8_cars") |> collect()

## ----convert-csv-vtr----------------------------------------------------------
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)

vtr <- tempfile(fileext = ".vtr")
tbl_csv(csv) |> write_vtr(vtr)

tbl(vtr) |> collect() |> head()

## ----convert-filtered---------------------------------------------------------
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)

vtr <- tempfile(fileext = ".vtr")
tbl_csv(csv) |>
  filter(mpg > 20) |>
  mutate(kpl = mpg * 0.425144) |>
  write_vtr(vtr)

tbl(vtr) |> collect()

## ----etl-pipeline-------------------------------------------------------------
# CSV -> filter + transform -> SQLite
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)

db <- tempfile(fileext = ".sqlite")
tbl_csv(csv) |>
  filter(cyl >= 6) |>
  select(mpg, cyl, hp, wt) |>
  mutate(power_weight = hp / wt) |>
  write_sqlite(db, "powerful_cars")

# SQLite -> VTR
vtr <- tempfile(fileext = ".vtr")
tbl_sqlite(db, "powerful_cars") |> write_vtr(vtr)

tbl(vtr) |> collect()

## ----join-across-formats------------------------------------------------------
f1 <- tempfile(fileext = ".vtr")
f2 <- tempfile(fileext = ".csv")

cars_main <- mtcars[, c("mpg", "cyl", "hp")]
cars_extra <- data.frame(cyl = c(4, 6, 8), label = c("small", "mid", "big"))

write_vtr(cars_main, f1)
write.csv(cars_extra, f2, row.names = FALSE)

tbl(f1) |>
  left_join(tbl_csv(f2), by = "cyl") |>
  collect() |>
  head()

## ----batch-size-effect--------------------------------------------------------
csv <- tempfile(fileext = ".csv")
big <- data.frame(
  id = seq_len(1000),
  value = rnorm(1000)
)
write.csv(big, csv, row.names = FALSE)

# Small row groups: more granular zone maps
f_small <- tempfile(fileext = ".vtr")
tbl_csv(csv) |> write_vtr(f_small, batch_size = 100)

# Default: single row group for 1000 rows
f_default <- tempfile(fileext = ".vtr")
tbl_csv(csv) |> write_vtr(f_default)

cat("Small batches:", file.size(f_small), "bytes\n")
cat("Default:      ", file.size(f_default), "bytes\n")

# spatial_more.R - raster <-> vector toolbox ops on the streamed spine.
#
# These complete the vector<->raster round trip on top of the machinery in
# spatial.R: the strip sink (.raster_sink), the pixel-centre helper
# (.zonal_strip_xy), the run-file accumulator (.run_accumulator), and
# spatial_dissolve() for the by-value merge. No new C and no new VecType.
#
#   mask        raster x polygon -> raster (clip pixels to a layer)
#   rast_calc   aligned rasters  -> raster (cellwise map algebra)
#   mosaic      aligned rasters  -> raster (merge onto a common grid)
#   polygonize  raster           -> vector polygons (one per value)
#   contours    raster           -> vector iso-lines

# Build an sfc of axis-aligned rectangles from corner vectors, one polygon per
# entry. Used to turn raster cells / runs into polygon geometry.
.rects_sfc <- function(xmin, xmax, ymin, ymax, crs) {
  polys <- lapply(seq_along(xmin), function(i)
    sf::st_polygon(list(rbind(
      c(xmin[i], ymin[i]), c(xmax[i], ymin[i]), c(xmax[i], ymax[i]),
      c(xmin[i], ymax[i]), c(xmin[i], ymin[i])))))
  sf::st_sfc(polys, crs = crs)
}

# Build an sfc of two-point line segments from endpoint vectors.
.segments_sfc <- function(x1, y1, x2, y2, crs) {
  lines <- lapply(seq_along(x1), function(i)
    sf::st_linestring(rbind(c(x1[i], y1[i]), c(x2[i], y2[i]))))
  sf::st_sfc(lines, crs = crs)
}

# -- mask (clip a raster to a polygon layer) ----------------------------------

#' Mask a streamed raster to a polygon layer
#'
#' Keeps the pixels of a `.vec` raster whose cell centre falls inside a resident
#' polygon layer and sets the rest to `background`, reading the raster one
#' tile-row strip at a time so the whole grid is never resident. It is the raster
#' counterpart of [spatial_clip()]: the streamed side is the (large) raster and
#' the small `mask` layer stays in memory. With `inverse = TRUE` the inside is
#' cleared and the outside kept.
#'
#' This is the *monoid fold* tier of the spatial toolbox: bounded to one strip
#' plus the resident mask, a single streaming pass, no spill. A pixel is tested
#' against the mask only when its centre falls in the mask bounding box, so the
#' point-in-polygon work stays proportional to the overlap rather than the whole
#' grid. Point-in-polygon is delegated to \pkg{sf} (an optional dependency);
#' topology and CRS handling are \pkg{sf}'s.
#'
#' @param x A `vectra_raster` (from [vec_open_raster()]) or a path to a `.vec`
#'   raster.
#' @param mask An `sf` or `sfc` polygon layer to clip against. When it carries no
#'   CRS it inherits the raster's EPSG.
#' @param inverse If `FALSE` (default) keep pixels inside `mask`; if `TRUE` keep
#'   the pixels outside it.
#' @param band Band(s) to mask (1-based). Default `NULL` masks every band.
#' @param background Value written to cleared pixels. Default `NA_real_`.
#' @param path Optional output `.vec` path. When given the result is streamed to
#'   disk and the opened [vec_open_raster()] handle is returned invisibly; when
#'   `NULL` the result is returned in memory (a matrix for one band, a list of
#'   matrices for several).
#' @param dtype Storage dtype for `.vec` output (see [vec_write_raster()]).
#'   Default `"f32"`.
#' @param compression Compression effort for `.vec` output. Default `"fast"`.
#'
#' @return When `path` is `NULL`, a numeric matrix (one band) or a list of
#'   matrices (several), each carrying `gt`, `extent`, and `crs` attributes
#'   (row 1 northmost). When `path` is given, the written `vectra_raster` handle
#'   (invisibly).
#'
#' @seealso [spatial_clip()] for the vector analogue, [zonal()] for per-zone
#'   summaries over the same pixel-in-polygon assignment.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' vals <- matrix(1:100, 10, 10, byrow = TRUE)
#' f <- tempfile(fileext = ".vec")
#' vec_write_raster(vals, f, dtype = "f64", extent = c(0, 0, 10, 10))
#'
#' disc <- sf::st_buffer(sf::st_sfc(sf::st_point(c(5, 5))), 3)
#' inside <- mask(f, disc)
#' sum(!is.na(inside))
#' unlink(f)
#'
#' @export
mask <- function(x, mask, inverse = FALSE, band = NULL,
                 background = NA_real_, path = NULL, dtype = "f32",
                 compression = c("fast", "balanced", "max")) {
  .check_sf()
  if (!inherits(mask, "sf") && !inherits(mask, "sfc"))
    stop("`mask` must be an sf or sfc polygon layer")
  comp_code <- switch(match.arg(compression), fast = 0L, balanced = 1L, max = 2L)

  vh <- .zonal_open(x, "x"); r <- vh$r
  if (vh$close) on.exit(vec_close_raster(r), add = TRUE)
  W <- as.integer(r$width); H <- as.integer(r$height); gt <- r$gt
  TS <- max(1L, as.integer(r$tile_size))
  epsg <- if (!is.null(r$epsg)) as.integer(r$epsg) else 0L
  bands <- if (is.null(band)) seq_len(r$n_bands) else as.integer(band)
  nout <- length(bands)

  mcrs <- sf::st_crs(mask)
  if (is.na(mcrs) && epsg > 0L) mcrs <- sf::st_crs(epsg)
  mg <- sf::st_geometry(mask)
  if (is.na(sf::st_crs(mg)) && !is.na(mcrs)) mg <- sf::st_set_crs(mg, mcrs)
  mg <- sf::st_union(mg)
  bb <- sf::st_bbox(mg)

  bn <- if (!is.null(r$band_names)) r$band_names[bands] else NULL
  sink <- .raster_sink(W, H, nout, gt, epsg, TS, path, dtype, bn, comp_code)

  tiles_y <- (H + TS - 1L) %/% TS
  for (ty in seq_len(tiles_y) - 1L) {
    r0 <- ty * TS + 1L; r1 <- min(r0 + TS - 1L, H); out_h <- r1 - r0 + 1L
    xy <- .zonal_strip_xy(gt, r0, r1, W)
    inbb <- xy$x >= bb[["xmin"]] & xy$x <= bb[["xmax"]] &
            xy$y >= bb[["ymin"]] & xy$y <= bb[["ymax"]]
    covered <- logical(length(inbb))
    if (any(inbb)) {
      pts <- sf::st_as_sf(data.frame(x = xy$x[inbb], y = xy$y[inbb]),
                          coords = c("x", "y"), crs = mcrs)
      covered[inbb] <- lengths(sf::st_intersects(pts, mg)) > 0L
    }
    clear <- if (inverse) covered else !covered
    os_list <- vector("list", nout)
    for (k in seq_len(nout)) {
      v <- as.vector(vec_read_window(r, band = bands[k], cols = c(1L, W),
                                     rows = c(r0, r1)))
      v[clear] <- background
      os_list[[k]] <- matrix(v, out_h, W)
    }
    sink$write(ty, r0, r1, do.call(cbind, os_list))
  }
  out <- sink$finish()
  if (!is.null(path)) return(out)
  if (nout == 1L) out[[1L]] else out
}

# -- rast_calc (cellwise map algebra over aligned rasters) --------------------

#' Cellwise calculation over aligned rasters (map algebra)
#'
#' Evaluates an expression cell by cell across one or more `.vec` rasters that
#' share a grid, reading every input one tile-row strip at a time so no whole
#' band is ever resident. Inside `expr` each name in `rasters` refers to that
#' raster's strip as a numeric vector, so ordinary vectorised R expresses the
#' calculation: a band index `(nir - red) / (nir + red)`, a reclassification
#' `cut(dem, breaks)`, a threshold `ifelse(slope > 30, 1L, 0L)`, or arithmetic
#' across layers. The result is written one strip at a time to a single-band
#' output.
#'
#' This is the *monoid fold* tier of the spatial toolbox: bounded to one strip
#' per input, a single streaming pass, no spill. The rasters must share
#' dimensions and geotransform; [warp()] them onto a common grid first if they
#' do not. No \pkg{sf} is needed.
#'
#' @param rasters A named list of `vectra_raster` handles or `.vec` paths sharing
#'   a grid. The names are the variables available inside `expr`.
#' @param expr An expression in those names producing one value per cell (or a
#'   scalar, recycled). Evaluated against each strip with the caller's
#'   environment as the enclosing scope.
#' @param band Band read from every input (1-based). Default 1.
#' @param path Optional output `.vec` path. When given the result is streamed to
#'   disk and the opened [vec_open_raster()] handle is returned invisibly; when
#'   `NULL` the result is returned as an in-memory matrix.
#' @param dtype Storage dtype for `.vec` output (see [vec_write_raster()]).
#'   Default `"f32"`.
#' @param compression Compression effort for `.vec` output. Default `"fast"`.
#'
#' @return When `path` is `NULL`, a numeric matrix (row 1 northmost) carrying
#'   `gt`, `extent`, and `crs` attributes. When `path` is given, the written
#'   `vectra_raster` handle (invisibly).
#'
#' @seealso [warp()] to align rasters onto a shared grid first, [focal()] for
#'   neighbourhood rather than cellwise calculation.
#'
#' @examples
#' nir <- matrix(c(40, 50, 60, 70), 2, 2)
#' red <- matrix(c(10, 20, 30, 40), 2, 2)
#' fn <- tempfile(fileext = ".vec"); fr <- tempfile(fileext = ".vec")
#' vec_write_raster(nir, fn, dtype = "f64", extent = c(0, 0, 2, 2))
#' vec_write_raster(red, fr, dtype = "f64", extent = c(0, 0, 2, 2))
#'
#' ndvi <- rast_calc(list(nir = fn, red = fr), (nir - red) / (nir + red))
#' round(ndvi, 3)
#' unlink(c(fn, fr))
#'
#' @export
rast_calc <- function(rasters, expr, band = 1L, path = NULL, dtype = "f32",
                      compression = c("fast", "balanced", "max")) {
  if (!is.list(rasters) || !length(rasters) ||
      is.null(names(rasters)) || any(names(rasters) == ""))
    stop("`rasters` must be a named list of vectra_raster handles or .vec paths")
  ex <- substitute(expr)
  comp_code <- switch(match.arg(compression), fast = 0L, balanced = 1L, max = 2L)

  handles <- lapply(rasters, function(z) .zonal_open(z, "rasters"))
  on.exit(for (h in handles) if (h$close) try(vec_close_raster(h$r), silent = TRUE),
          add = TRUE)
  rs <- lapply(handles, `[[`, "r")
  r1r <- rs[[1L]]
  W <- as.integer(r1r$width); H <- as.integer(r1r$height); gt <- as.numeric(r1r$gt)
  for (z in rs[-1L])
    if (z$width != W || z$height != H ||
        !isTRUE(all.equal(as.numeric(z$gt), gt)))
      stop("all rasters must share dimensions and geotransform; warp() them first")
  TS <- max(1L, as.integer(r1r$tile_size))
  epsg <- if (!is.null(r1r$epsg)) as.integer(r1r$epsg) else 0L
  band <- as.integer(band)
  pe <- parent.frame()

  sink <- .raster_sink(W, H, 1L, gt, epsg, TS, path, dtype, NULL, comp_code)
  tiles_y <- (H + TS - 1L) %/% TS
  for (ty in seq_len(tiles_y) - 1L) {
    r0 <- ty * TS + 1L; r1 <- min(r0 + TS - 1L, H); out_h <- r1 - r0 + 1L
    env <- new.env(parent = pe)
    for (nm in names(rs))
      assign(nm, as.vector(vec_read_window(rs[[nm]], band = band,
                                           cols = c(1L, W), rows = c(r0, r1))),
             envir = env)
    v <- eval(ex, env)
    if (length(v) == 1L) v <- rep(v, out_h * W)
    if (length(v) != out_h * W)
      stop("`expr` must return one value per cell (or a scalar)")
    sink$write(ty, r0, r1, matrix(as.numeric(v), out_h, W))
  }
  out <- sink$finish()
  if (!is.null(path)) return(out)
  out[[1L]]
}

# -- mosaic (merge aligned rasters onto a common grid) ------------------------

# Resolve the union grid for a set of rasters that must share resolution and sit
# on a common cell grid. Returns the target W/H/gt and each input's integer
# col/row offset into it.
.mosaic_grid <- function(rs) {
  g1 <- as.numeric(rs[[1L]]$gt)
  xres <- g1[2L]; yres <- g1[6L]
  if (g1[3L] != 0 || g1[5L] != 0)
    stop("mosaic needs north-up (unrotated) rasters")
  n <- length(rs)
  xmins <- xmaxs <- ymins <- ymaxs <- numeric(n)
  for (i in seq_len(n)) {
    g <- as.numeric(rs[[i]]$gt); Wi <- rs[[i]]$width; Hi <- rs[[i]]$height
    if (abs(g[2L] - xres) > 1e-9 * abs(xres) ||
        abs(g[6L] - yres) > 1e-9 * abs(yres) || g[3L] != 0 || g[5L] != 0)
      stop("all rasters must share resolution and be north-up; warp() them first")
    xmins[i] <- g[1L]; ymaxs[i] <- g[4L]
    xmaxs[i] <- g[1L] + Wi * g[2L]; ymins[i] <- g[4L] + Hi * g[6L]
  }
  xmin <- min(xmins); xmax <- max(xmaxs); ymin <- min(ymins); ymax <- max(ymaxs)
  W <- as.integer(round((xmax - xmin) / xres))
  H <- as.integer(round((ymax - ymin) / (-yres)))
  col_off <- (xmins - xmin) / xres
  row_off <- (ymax - ymaxs) / (-yres)
  if (any(abs(col_off - round(col_off)) > 1e-6) ||
      any(abs(row_off - round(row_off)) > 1e-6))
    stop("rasters are not on a common cell grid; warp() them to a shared grid first")
  list(W = W, H = H, gt = c(xmin, xres, 0, ymax, 0, yres),
       col_off = as.integer(round(col_off)), row_off = as.integer(round(row_off)))
}

#' Merge aligned rasters onto a common grid
#'
#' Combines several `.vec` rasters that share a resolution and cell grid into one
#' raster spanning their union, resolving overlap with `fun`. The output is
#' walked one tile-row strip at a time and each input contributes only the window
#' overlapping the current strip, so neither the inputs nor the output are held
#' whole in memory.
#'
#' This is the *monoid fold* tier of the spatial toolbox: each output strip folds
#' the overlapping input windows, bounded memory, no spill. The inputs must share
#' resolution and lie on a common cell grid; [warp()] them onto a shared grid
#' first if they do not. No \pkg{sf} is needed.
#'
#' @param rasters A list of `vectra_raster` handles or `.vec` paths sharing
#'   resolution and grid alignment.
#' @param fun Overlap rule where inputs cover the same cell: `"first"` (the
#'   earliest input in `rasters`, default), `"last"`, `"mean"`, `"sum"`, `"min"`,
#'   or `"max"`. Cells covered by no input come back `NA`.
#' @param band Band read from every input (1-based). Default 1.
#' @param path Optional output `.vec` path. When given the result is streamed to
#'   disk and the opened [vec_open_raster()] handle is returned invisibly; when
#'   `NULL` the result is returned as an in-memory matrix.
#' @param dtype Storage dtype for `.vec` output (see [vec_write_raster()]).
#'   Default `"f32"`.
#' @param compression Compression effort for `.vec` output. Default `"fast"`.
#'
#' @return When `path` is `NULL`, a numeric matrix on the union grid (row 1
#'   northmost) carrying `gt`, `extent`, and `crs` attributes. When `path` is
#'   given, the written `vectra_raster` handle (invisibly).
#'
#' @seealso [warp()] to bring rasters onto a shared grid, [rast_calc()] for
#'   cellwise combination of already-aligned rasters.
#'
#' @examples
#' a <- matrix(1, 4, 4); b <- matrix(2, 4, 4)
#' fa <- tempfile(fileext = ".vec"); fb <- tempfile(fileext = ".vec")
#' vec_write_raster(a, fa, dtype = "f64", extent = c(0, 0, 4, 4))
#' vec_write_raster(b, fb, dtype = "f64", extent = c(2, 2, 6, 6))
#'
#' m <- mosaic(list(fa, fb), fun = "mean")
#' dim(m)
#' unlink(c(fa, fb))
#'
#' @export
mosaic <- function(rasters, fun = c("first", "last", "mean", "sum", "min", "max"),
                   band = 1L, path = NULL, dtype = "f32",
                   compression = c("fast", "balanced", "max")) {
  fun <- match.arg(fun)
  comp_code <- switch(match.arg(compression), fast = 0L, balanced = 1L, max = 2L)
  if (!is.list(rasters) || !length(rasters))
    stop("`rasters` must be a list of vectra_raster handles or .vec paths")

  handles <- lapply(rasters, function(z) .zonal_open(z, "rasters"))
  on.exit(for (h in handles) if (h$close) try(vec_close_raster(h$r), silent = TRUE),
          add = TRUE)
  rs <- lapply(handles, `[[`, "r")
  band <- as.integer(band)
  mg <- .mosaic_grid(rs)
  W <- mg$W; H <- mg$H; gt <- mg$gt
  epsg <- if (!is.null(rs[[1L]]$epsg)) as.integer(rs[[1L]]$epsg) else 0L
  TS <- max(1L, as.integer(rs[[1L]]$tile_size))
  sink <- .raster_sink(W, H, 1L, gt, epsg, TS, path, dtype, NULL, comp_code)

  # "first" wants the earliest input to win a shared cell; process inputs in
  # reverse so the earliest is written last under the last-write-wins fill.
  order_in <- if (fun == "first") rev(seq_along(rs)) else seq_along(rs)

  tiles_y <- (H + TS - 1L) %/% TS
  for (ty in seq_len(tiles_y) - 1L) {
    r0 <- ty * TS + 1L; r1 <- min(r0 + TS - 1L, H); out_h <- r1 - r0 + 1L
    sm  <- matrix(0,        out_h, W); cnt <- matrix(0L,        out_h, W)
    mn  <- matrix(Inf,      out_h, W); mx  <- matrix(-Inf,      out_h, W)
    val <- matrix(NA_real_, out_h, W)
    for (i in order_in) {
      ro <- mg$row_off[i]; co <- mg$col_off[i]
      iW <- rs[[i]]$width; iH <- rs[[i]]$height
      sr0 <- max(1L, r0 - ro); sr1 <- min(iH, r1 - ro)
      if (sr1 < sr0) next
      win <- vec_read_window(rs[[i]], band = band, cols = c(1L, iW),
                             rows = c(sr0, sr1))
      lr0 <- (sr0 + ro) - r0 + 1L; lr1 <- (sr1 + ro) - r0 + 1L
      br <- lr0:lr1; bc <- (co + 1L):(co + iW)
      ok <- !is.na(win)
      w0 <- win; w0[!ok] <- 0
      sm[br, bc]  <- sm[br, bc] + w0
      cnt[br, bc] <- cnt[br, bc] + ok
      sub_mn <- mn[br, bc]; sub_mn[ok] <- pmin(sub_mn[ok], win[ok]); mn[br, bc] <- sub_mn
      sub_mx <- mx[br, bc]; sub_mx[ok] <- pmax(sub_mx[ok], win[ok]); mx[br, bc] <- sub_mx
      sub_v  <- val[br, bc]; sub_v[ok]  <- win[ok];                  val[br, bc] <- sub_v
    }
    out <- switch(fun,
      first = val, last = val,
      sum  = ifelse(cnt > 0L, sm, NA_real_),
      mean = ifelse(cnt > 0L, sm / cnt, NA_real_),
      min  = ifelse(is.finite(mn), mn, NA_real_),
      max  = ifelse(is.finite(mx), mx, NA_real_))
    sink$write(ty, r0, r1, out)
  }
  res <- sink$finish()
  if (!is.null(path)) return(res)
  res[[1L]]
}

# -- proximity (Euclidean distance to the nearest feature) --------------------

# Sentinel cost for a cell that holds no feature. The squared distance between
# any two cells of a realistic grid stays far below this, so it acts as
# "infinity" while keeping the parabola intersections in finite arithmetic.
.FH_INF <- 1e20

# One streamed row pass of the distance transform. Reads `rsrc` one tile-row
# strip at a time, optionally maps each strip to feature/sentinel costs through
# `prep`, transforms every row with spacing `scale` in C, and writes to `sink`.
.proximity_row_pass <- function(rsrc, band, W, H, TS, scale, sink, prep) {
  tiles_y <- (H + TS - 1L) %/% TS
  for (ty in seq_len(tiles_y) - 1L) {
    r0 <- ty * TS + 1L; r1 <- min(r0 + TS - 1L, H); out_h <- r1 - r0 + 1L
    vm <- vec_read_window(rsrc, band = band, cols = c(1L, W), rows = c(r0, r1))
    f <- if (is.null(prep)) vm else prep(vm)
    storage.mode(f) <- "double"
    g <- .Call(C_edt_strip, f, c(out_h, W), as.numeric(scale))
    sink$write(ty, r0, r1, g)
  }
}

# Streamed out-of-core transpose. The destination raster has width `srcH` and
# height `srcW`; its tile-row strip [r0, r1] is the source columns [r0, r1] over
# every source row, transposed. `finalize` optionally maps the values before
# they are written (used on the final pass to take the square root and mark
# no-feature cells).
.proximity_transpose <- function(rsrc, srcW, srcH, TS, sink, finalize) {
  tiles_y <- (srcW + TS - 1L) %/% TS
  for (ty in seq_len(tiles_y) - 1L) {
    r0 <- ty * TS + 1L; r1 <- min(r0 + TS - 1L, srcW)
    block <- vec_read_window(rsrc, band = 1L, cols = c(r0, r1), rows = c(1L, srcH))
    tb <- t(block)
    if (!is.null(finalize)) tb <- finalize(tb)
    sink$write(ty, r0, r1, tb)
  }
}

#' Euclidean distance to the nearest feature (proximity)
#'
#' Computes, for every cell of a `.vec` raster, the straight-line Euclidean
#' distance to the nearest feature cell, in CRS units. Feature cells are the
#' non-NA cells by default, or the cells whose value is in `target`. This is the
#' raster proximity / Euclidean-distance staple, the distance companion to
#' [rasterize()].
#'
#' The exact Euclidean distance transform is separable (Felzenszwalb and
#' Huttenlocher 2012): a one-dimensional lower-envelope-of-parabolas transform
#' along the rows, then the same transform along the columns, each linear in the
#' line length and each line independent. vectra runs it as four streamed passes
#' over tile-row strips, with an out-of-core transpose between the row pass and
#' the column pass, so the whole grid is never resident. The row pass scales
#' squared distances by the x resolution and the column pass by the y
#' resolution, so the result is exact on anisotropic (non-square) cells. This
#' places proximity on the sort / partition tier of the spatial toolbox.
#'
#' Distances are straight-line Euclidean in the raster CRS units. Cost-distance,
#' which accumulates a per-cell friction along the path, is a global
#' shortest-path problem and stays resident: [collect()] the raster and run a
#' resident solver for that.
#'
#' @param x A `vectra_raster` (from [vec_open_raster()]) or a path to a `.vec`
#'   raster.
#' @param target Optional numeric vector of feature values. When `NULL`
#'   (default) every non-NA cell is a feature; otherwise a cell is a feature
#'   when its value is in `target`.
#' @param band Band to read (1-based). Default 1.
#' @param path Optional output `.vec` path. When given the result is streamed to
#'   disk and the opened [vec_open_raster()] handle is returned invisibly; when
#'   `NULL` the result is returned as an in-memory matrix.
#' @param dtype Storage dtype for `.vec` output (see [vec_write_raster()]).
#'   Default `"f32"`.
#' @param compression Compression effort for `.vec` output. Default `"fast"`.
#'
#' @return When `path` is `NULL`, a numeric matrix (row 1 northmost) carrying
#'   `gt`, `extent`, and `crs` attributes, with distance in CRS units and `NA`
#'   where the raster holds no feature anywhere. When `path` is given, the
#'   written `vectra_raster` handle (invisibly).
#'
#' @seealso [rasterize()] to build a raster from streamed points, [mask()] to
#'   clip a raster to a polygon layer.
#'
#' @examples
#' m <- matrix(NA_real_, 12, 12)
#' m[3, 4] <- 1; m[9, 10] <- 1
#' f <- tempfile(fileext = ".vec")
#' vec_write_raster(m, f, dtype = "f64", extent = c(0, 0, 12, 12))
#'
#' d <- proximity(f)
#' round(d[1:3, 1:3], 2)
#' unlink(f)
#'
#' @export
proximity <- function(x, target = NULL, band = 1L, path = NULL, dtype = "f32",
                      compression = c("fast", "balanced", "max")) {
  comp_code <- switch(match.arg(compression), fast = 0L, balanced = 1L, max = 2L)
  if (!is.null(target) && !is.numeric(target))
    stop("`target` must be a numeric vector of feature values, or NULL")

  vh <- .zonal_open(x, "x"); r <- vh$r
  if (vh$close) on.exit(vec_close_raster(r), add = TRUE)
  W <- as.integer(r$width); H <- as.integer(r$height); gt <- r$gt
  TS <- max(1L, as.integer(r$tile_size))
  epsg <- if (!is.null(r$epsg)) as.integer(r$epsg) else 0L
  band <- as.integer(band)
  xres <- abs(gt[2L]); yres <- abs(gt[6L])
  dgt <- c(0, 1, 0, 0, 0, -1)   # placeholder grid for the transposed temporaries

  t1  <- tempfile(fileext = ".vec")
  t1t <- tempfile(fileext = ".vec")
  t2t <- tempfile(fileext = ".vec")
  on.exit(unlink(c(t1, t1t, t2t)), add = TRUE)

  # Pass A: row transform of the source (spacing xres), features -> 0 cost and
  # every other cell -> the sentinel. Output t1 holds the partial transform.
  prep <- function(vm) {
    v <- as.numeric(vm)
    feat <- if (is.null(target)) !is.na(v) else !is.na(v) & v %in% target
    f <- rep(.FH_INF, length(v)); f[feat] <- 0
    matrix(f, nrow(vm), ncol(vm))
  }
  s1 <- .raster_sink(W, H, 1L, gt, epsg, TS, t1, "f64", NULL, 0L)
  .proximity_row_pass(r, band, W, H, TS, xres, s1, prep)
  h1 <- s1$finish()

  # Transpose t1 (W x H) -> t1t (H x W): the source columns become rows.
  s1t <- .raster_sink(H, W, 1L, dgt, 0L, TS, t1t, "f64", NULL, 0L)
  .proximity_transpose(h1, W, H, TS, s1t, NULL)
  h1t <- s1t$finish(); vec_close_raster(h1)

  # Pass B: row transform of the transpose (spacing yres) is the column
  # transform of the source, completing the exact squared distance.
  s2 <- .raster_sink(H, W, 1L, dgt, 0L, TS, t2t, "f64", NULL, 0L)
  .proximity_row_pass(h1t, 1L, H, W, TS, yres, s2, NULL)
  h2t <- s2$finish(); vec_close_raster(h1t)

  # Transpose back (H x W -> W x H), take the square root, and map cells with no
  # feature anywhere to NA.
  bn <- if (!is.null(r$band_names)) r$band_names[band] else NULL
  out_sink <- .raster_sink(W, H, 1L, gt, epsg, TS, path, dtype, bn, comp_code)
  finalize <- function(tb) {
    d <- sqrt(tb); d[tb >= 1e19] <- NA_real_; d
  }
  .proximity_transpose(h2t, H, W, TS, out_sink, finalize)
  out <- out_sink$finish(); vec_close_raster(h2t)

  if (!is.null(path)) return(out)
  out[[1L]]
}

# -- polygonize (raster -> vector polygons) -----------------------------------

# Turn one strip of pixels into rectangle corners tagged by value. With
# per_cell = FALSE, equal-valued runs along a row collapse to one rectangle (the
# by-value dissolve then merges across rows and strips); with per_cell = TRUE,
# every cell becomes its own square.
.polygonize_strip <- function(vm, gt, r0, W, na_rm, per_cell) {
  out_h <- nrow(vm)
  xres <- gt[2L]; yres <- gt[6L]; x0 <- gt[1L]; y0 <- gt[4L]
  xmin <- xmax <- ymin <- ymax <- val <- numeric(0)
  for (a in seq_len(out_h)) {
    row <- vm[a, ]
    ytop <- y0 + (r0 - 1L + a - 1L) * yres
    ybot <- y0 + (r0 - 1L + a) * yres
    yt <- max(ytop, ybot); yb <- min(ytop, ybot)
    if (per_cell) {
      cols <- if (na_rm) which(!is.na(row)) else seq_len(W)
      if (!length(cols)) next
      xmin <- c(xmin, x0 + (cols - 1L) * xres); xmax <- c(xmax, x0 + cols * xres)
      ymin <- c(ymin, rep(yb, length(cols)));   ymax <- c(ymax, rep(yt, length(cols)))
      val  <- c(val, row[cols])
    } else {
      rr <- rle(row)
      ends <- cumsum(rr$lengths); starts <- ends - rr$lengths + 1L
      for (k in seq_along(rr$values)) {
        vv <- rr$values[k]
        if (na_rm && is.na(vv)) next
        xmin <- c(xmin, x0 + (starts[k] - 1L) * xres)
        xmax <- c(xmax, x0 + ends[k] * xres)
        ymin <- c(ymin, yb); ymax <- c(ymax, yt); val <- c(val, vv)
      }
    }
  }
  list(xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax, val = val)
}

#' Vectorise a raster into polygons
#'
#' Converts a `.vec` raster into polygon features, the inverse of [rasterize()].
#' The raster is read one tile-row strip at a time; within each strip equal-valued
#' cells along a row collapse to a rectangle, and (with `dissolve = TRUE`) the
#' rectangles are then merged by value through [spatial_dissolve()] so each
#' distinct value becomes a single polygon spanning the whole raster. The result
#' is a lazy `vectra_node` carrying a value column and hex-WKB geometry.
#'
#' Extraction is the *monoid fold* tier (one strip at a time); the by-value
#' dissolve rides the *sort / partition* tier of [spatial_dissolve()], localising
#' each value to a shard before the union. Geometry assembly is delegated to
#' \pkg{sf} (an optional dependency).
#'
#' @param x A `vectra_raster` (from [vec_open_raster()]) or a path to a `.vec`
#'   raster.
#' @param band Band to vectorise (1-based). Default 1.
#' @param dissolve If `TRUE` (default) merge cells of equal value into one polygon
#'   per value; if `FALSE` emit one square polygon per cell.
#' @param na_rm Drop nodata cells (`TRUE`, default) or vectorise them as a value.
#' @param values Name of the output value column. Default `"value"`.
#' @param crs Coordinate reference system recorded on the node. Defaults to the
#'   raster's EPSG, else unknown.
#' @param flush_rows Rows buffered before a spill flush. Defaults to
#'   `getOption("vectra.spatial_flush", 5e5)`.
#'
#' @return A `vectra_node` with the value column and a hex-WKB `geometry` column,
#'   materialise it with [collect_sf()].
#'
#' @seealso [rasterize()] for the inverse, [contours()] for iso-lines,
#'   [collect_sf()] to materialise as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' m <- matrix(c(1, 1, 2, 2, 1, 1, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3), 4, 4,
#'             byrow = TRUE)
#' f <- tempfile(fileext = ".vec")
#' vec_write_raster(m, f, dtype = "f64", extent = c(0, 0, 4, 4))
#'
#' polys <- polygonize(f)
#' collect_sf(polys)
#' unlink(f)
#'
#' @export
polygonize <- function(x, band = 1L, dissolve = TRUE, na_rm = TRUE,
                       values = "value", crs = NA, flush_rows = NULL) {
  .check_sf()
  vh <- .zonal_open(x, "x"); r <- vh$r
  if (vh$close) on.exit(vec_close_raster(r), add = TRUE)
  W <- as.integer(r$width); H <- as.integer(r$height); gt <- r$gt
  TS <- max(1L, as.integer(r$tile_size))
  if (identical(crs, NA) && !is.null(r$epsg) && r$epsg > 0L) crs <- r$epsg
  rcrs <- .as_crs(crs)
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  acc <- .run_accumulator(fr)

  tiles_y <- (H + TS - 1L) %/% TS
  for (ty in seq_len(tiles_y) - 1L) {
    r0 <- ty * TS + 1L; r1 <- min(r0 + TS - 1L, H)
    vm <- vec_read_window(r, band = band, cols = c(1L, W), rows = c(r0, r1))
    p <- .polygonize_strip(vm, gt, r0, W, na_rm, per_cell = !dissolve)
    if (!length(p$val)) next
    g <- .rects_sfc(p$xmin, p$xmax, p$ymin, p$ymax, rcrs)
    df <- stats::setNames(data.frame(p$val), values)
    acc$push(.sf_encode_result(sf::st_sf(df, geometry = g), "geometry"))
  }
  node <- acc$finish(crs = rcrs, empty_geom = "geometry")
  if (dissolve) node <- spatial_dissolve(node, by = values, crs = rcrs)
  node
}

# -- contours (raster -> vector iso-lines) ------------------------------------

# Marching-squares iso-line segments for one window of pixel centres. `vm` is an
# in_h x W matrix whose rows span target rows r0 .. r0+in_h-1 (with one bottom
# halo row so every cell-row with its top row in the strip is covered). Returns
# segment endpoint and level vectors for all `levels`.
.contour_segments <- function(vm, gt, r0, W, levels) {
  inh <- nrow(vm)
  if (inh < 2L || W < 2L) return(NULL)
  xres <- gt[2L]; yres <- gt[6L]; x0 <- gt[1L]; y0 <- gt[4L]
  na <- inh - 1L
  tl <- vm[1:na, 1:(W - 1L), drop = FALSE]; tr <- vm[1:na, 2:W, drop = FALSE]
  bl <- vm[2:inh, 1:(W - 1L), drop = FALSE]; br <- vm[2:inh, 2:W, drop = FALSE]
  ctop <- matrix(rep(1:(W - 1L), each = na), na, W - 1L)
  rtop <- matrix(rep(r0:(r0 + na - 1L), times = W - 1L), na, W - 1L)
  xL <- x0 + (ctop - 0.5) * xres; xR <- x0 + (ctop + 0.5) * xres
  yT <- y0 + (rtop - 0.5) * yres; yB <- y0 + (rtop + 0.5) * yres
  ctr <- (tl + tr + bl + br) / 4

  x1 <- y1 <- x2 <- y2 <- lev <- numeric(0)
  # edge map: each case -> pairs of edge codes (T=1, R=2, B=3, L=4).
  segmap <- list(
    `1` = list(c(4, 3)), `2` = list(c(3, 2)), `3` = list(c(4, 2)),
    `4` = list(c(1, 2)), `6` = list(c(1, 3)), `7` = list(c(1, 4)),
    `8` = list(c(1, 4)), `9` = list(c(1, 3)), `11` = list(c(1, 2)),
    `12` = list(c(4, 2)), `13` = list(c(3, 2)), `14` = list(c(4, 3)))

  for (L in levels) {
    ok <- is.finite(tl) & is.finite(tr) & is.finite(bl) & is.finite(br)
    a <- tl > L; b <- tr > L; c <- br > L; d <- bl > L
    code <- 8L * a + 4L * b + 2L * c + 1L * d
    code[!ok] <- 0L
    # edge crossing coordinates (vectors over the cell grid)
    ex <- list(
      Tx = xL + (xR - xL) * (L - tl) / (tr - tl), Ty = yT,
      Rx = xR, Ry = yT + (yB - yT) * (L - tr) / (br - tr),
      Bx = xL + (xR - xL) * (L - bl) / (br - bl), By = yB,
      Lx = xL, Ly = yT + (yB - yT) * (L - tl) / (bl - tl))
    pt <- function(edge, idx) {
      if (edge == 1L) list(x = ex$Tx[idx], y = ex$Ty[idx])
      else if (edge == 2L) list(x = ex$Rx[idx], y = ex$Ry[idx])
      else if (edge == 3L) list(x = ex$Bx[idx], y = ex$By[idx])
      else list(x = ex$Lx[idx], y = ex$Ly[idx])
    }
    emit <- function(idx, e1, e2) {
      if (!length(idx)) return(invisible())
      p1 <- pt(e1, idx); p2 <- pt(e2, idx)
      x1 <<- c(x1, p1$x); y1 <<- c(y1, p1$y)
      x2 <<- c(x2, p2$x); y2 <<- c(y2, p2$y); lev <<- c(lev, rep(L, length(idx)))
    }
    for (cs in names(segmap)) {
      idx <- which(code == as.integer(cs))
      if (length(idx)) for (pair in segmap[[cs]]) emit(idx, pair[1L], pair[2L])
    }
    # saddles: connection depends on the cell-centre value.
    s5 <- which(code == 5L); s10 <- which(code == 10L)
    cf <- ctr > L
    if (length(s5)) {
      hi <- s5[cf[s5]]; lo <- s5[!cf[s5]]
      emit(hi, 1L, 4L); emit(hi, 3L, 2L)   # centre filled: T-L and B-R
      emit(lo, 1L, 2L); emit(lo, 3L, 4L)   # centre empty:  T-R and B-L
    }
    if (length(s10)) {
      hi <- s10[cf[s10]]; lo <- s10[!cf[s10]]
      emit(hi, 1L, 2L); emit(hi, 3L, 4L)   # centre filled: T-R and B-L
      emit(lo, 1L, 4L); emit(lo, 3L, 2L)   # centre empty:  T-L and B-R
    }
  }
  if (!length(lev)) return(NULL)
  list(x1 = x1, y1 = y1, x2 = x2, y2 = y2, level = lev)
}

#' Extract contour iso-lines from a streamed raster
#'
#' Traces contour lines at one or more levels from a `.vec` raster with marching
#' squares, reading the raster one tile-row strip at a time (each strip expanded
#' by one row so a cell straddling the strip boundary is traced once). Each strip
#' contributes line segments, which are accumulated into a lazy `vectra_node`
#' carrying a `level` column and hex-WKB geometry. With `merge = TRUE` the
#' segments of each level are joined into continuous lines.
#'
#' Extraction is the *sort / partition* tier of the spatial toolbox: bounded to
#' one haloed strip at a time. The optional final merge collects the segment set,
#' which is small relative to the raster, and joins it per level; this is the
#' small all-to-all step on the output, not on the grid. Geometry assembly and
#' the merge are delegated to \pkg{sf} (an optional dependency).
#'
#' @param x A `vectra_raster` (from [vec_open_raster()]) or a path to a `.vec`
#'   raster.
#' @param levels Numeric vector of contour levels to trace.
#' @param band Band to contour (1-based). Default 1.
#' @param merge If `TRUE` (default) join each level's segments into continuous
#'   lines with [sf::st_line_merge()]; if `FALSE` return the raw per-cell
#'   segments.
#' @param crs Coordinate reference system recorded on the node. Defaults to the
#'   raster's EPSG, else unknown.
#' @param flush_rows Rows buffered before a spill flush. Defaults to
#'   `getOption("vectra.spatial_flush", 5e5)`.
#'
#' @return A `vectra_node` with a `level` column and a hex-WKB `geometry` column,
#'   materialise it with [collect_sf()].
#'
#' @seealso [polygonize()] for area features, [terrain()] for the DEM
#'   derivatives contours often accompany, [collect_sf()] to materialise as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' z <- outer(1:20, 1:20, function(r, c) r + c)
#' f <- tempfile(fileext = ".vec")
#' vec_write_raster(z, f, dtype = "f64", extent = c(0, 0, 20, 20))
#'
#' iso <- contours(f, levels = c(15, 25, 35))
#' collect_sf(iso)
#' unlink(f)
#'
#' @export
contours <- function(x, levels, band = 1L, merge = TRUE, crs = NA,
                     flush_rows = NULL) {
  .check_sf()
  if (!is.numeric(levels) || !length(levels))
    stop("`levels` must be a numeric vector of contour levels")
  vh <- .zonal_open(x, "x"); r <- vh$r
  if (vh$close) on.exit(vec_close_raster(r), add = TRUE)
  W <- as.integer(r$width); H <- as.integer(r$height); gt <- r$gt
  TS <- max(1L, as.integer(r$tile_size))
  if (identical(crs, NA) && !is.null(r$epsg) && r$epsg > 0L) crs <- r$epsg
  rcrs <- .as_crs(crs)
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  acc <- .run_accumulator(fr)

  tiles_y <- (H + TS - 1L) %/% TS
  for (ty in seq_len(tiles_y) - 1L) {
    r0 <- ty * TS + 1L; r1 <- min(r0 + TS - 1L, H)
    in_r1 <- min(H, r1 + 1L)             # one bottom halo row for the cell below
    vm <- vec_read_window(r, band = band, cols = c(1L, W), rows = c(r0, in_r1))
    s <- .contour_segments(vm, gt, r0, W, levels)
    if (is.null(s)) next
    g <- .segments_sfc(s$x1, s$y1, s$x2, s$y2, rcrs)
    acc$push(.sf_encode_result(
      sf::st_sf(data.frame(level = s$level), geometry = g), "geometry"))
  }
  node <- acc$finish(crs = rcrs, empty_geom = "geometry")
  if (!merge) return(node)

  df <- collect(node)
  if (!nrow(df)) return(node)
  sb <- .sf_decode_chunk(df, "geometry", NULL, rcrs)
  geo <- sf::st_geometry(sb)
  parts <- split(seq_len(nrow(df)), df$level)
  out_lev <- as.numeric(names(parts))
  merged <- lapply(parts, function(ix)
    sf::st_line_merge(sf::st_union(geo[ix])))
  acc2 <- .run_accumulator(fr)
  for (j in seq_along(merged))
    acc2$push(.sf_encode_result(
      sf::st_sf(data.frame(level = out_lev[j]), geometry = merged[[j]]),
      "geometry"))
  acc2$finish(crs = rcrs, empty_geom = "geometry")
}

# Recovery and streaming-invariance tests for mosaic().
#
# Ground truth is a hand-assembled union grid and terra::mosaic where available.
# The streamed strip pass must reproduce it and stay equal across tile sizes.

wr <- function(m, ext, tile = 512L) {
  f <- tempfile(fileext = ".vec")
  vec_write_raster(m, f, dtype = "f64", extent = ext, tile_size = tile)
  f
}

test_that("mosaic builds the union grid and resolves overlap by fun", {
  a <- matrix(1, 4, 4); b <- matrix(2, 4, 4)
  fa <- wr(a, c(0, 0, 4, 4), tile = 2L); fb <- wr(b, c(2, 2, 6, 6), tile = 2L)
  on.exit(unlink(c(fa, fb)))

  mm <- mosaic(list(fa, fb), fun = "mean")
  expect_equal(dim(mm), c(6L, 6L))
  # fa covers x[0,4] y[0,4]; fb covers x[2,6] y[2,6]; union is x[0,6] y[0,6].
  # overlap is the 2x2 block of target rows/cols 3:4 -> mean(1, 2).
  expect_equal(mm[3, 3], 1.5)
  expect_equal(mm[1, 6], 2)    # top-right (x[5,6] y[5,6]) is b only
  expect_equal(mm[6, 1], 1)    # bottom-left (x[0,1] y[0,1]) is a only
  expect_true(is.na(mm[1, 1])) # top-left covered by neither
  expect_true(is.na(mm[6, 6])) # bottom-right covered by neither
})

test_that("first/last/min/max/sum resolve overlap as specified", {
  a <- matrix(1, 4, 4); b <- matrix(2, 4, 4)
  fa <- wr(a, c(0, 0, 4, 4), tile = 2L); fb <- wr(b, c(2, 2, 6, 6), tile = 2L)
  on.exit(unlink(c(fa, fb)))

  expect_equal(mosaic(list(fa, fb), "first")[3, 3], 1)
  expect_equal(mosaic(list(fa, fb), "last")[3, 3], 2)
  expect_equal(mosaic(list(fa, fb), "min")[3, 3], 1)
  expect_equal(mosaic(list(fa, fb), "max")[3, 3], 2)
  expect_equal(mosaic(list(fa, fb), "sum")[3, 3], 3)
})

test_that("mosaic matches terra::mosaic cell-for-cell", {
  skip_if_not_installed("terra")
  set.seed(7)
  a <- matrix(rnorm(8 * 8), 8, 8); b <- matrix(rnorm(8 * 8), 8, 8)
  fa <- wr(a, c(0, 0, 8, 8), tile = 4L); fb <- wr(b, c(4, 4, 12, 12), tile = 4L)
  on.exit(unlink(c(fa, fb)))

  for (fn in c("mean", "sum", "min", "max")) {
    mm <- mosaic(list(fa, fb), fun = fn)
    ta <- terra::rast(a, extent = terra::ext(0, 8, 0, 8))
    tb <- terra::rast(b, extent = terra::ext(4, 12, 4, 12))
    tmo <- terra::mosaic(ta, tb, fun = fn)
    tmv <- matrix(terra::values(tmo), terra::nrow(tmo), terra::ncol(tmo),
                  byrow = TRUE)
    expect_equal(dim(mm), dim(tmv))
    expect_equal(mm, tmv, tolerance = 1e-9, ignore_attr = TRUE,
                 label = paste("mosaic fun =", fn))
  }
})

test_that("mosaic is invariant to tile size (streaming invariance)", {
  set.seed(8)
  a <- matrix(rnorm(12 * 12), 12, 12); b <- matrix(rnorm(12 * 12), 12, 12)
  fa1 <- wr(a, c(0, 0, 12, 12), tile = 3L); fb1 <- wr(b, c(6, 6, 18, 18), tile = 3L)
  fa2 <- wr(a, c(0, 0, 12, 12), tile = 12L); fb2 <- wr(b, c(6, 6, 18, 18), tile = 12L)
  on.exit(unlink(c(fa1, fb1, fa2, fb2)))

  expect_equal(mosaic(list(fa1, fb1), "mean"), mosaic(list(fa2, fb2), "mean"))
})

test_that("mosaic rejects rasters off a common grid", {
  fa <- wr(matrix(1, 4, 4), c(0, 0, 4, 4))
  fb <- wr(matrix(2, 4, 4), c(0.3, 0, 4.3, 4))   # half-cell shift
  on.exit(unlink(c(fa, fb)))
  expect_error(mosaic(list(fa, fb)), "common cell grid")
})

# Recovery and streaming-invariance tests for polygonize().
#
# Ground truth is the hand-computed area each value covers and terra::as.polygons
# where available. The streamed strip pass must reproduce per-value area and
# stay equal across tile sizes.

wr <- function(m, ext = c(0, 0, ncol(m), nrow(m)), tile = 512L) {
  f <- tempfile(fileext = ".vec")
  vec_write_raster(m, f, dtype = "f64", extent = ext, tile_size = tile)
  f
}

area_by_value <- function(sf_polys, vcol = "value") {
  a <- as.numeric(sf::st_area(sf_polys))
  tapply(a, sf_polys[[vcol]], sum)
}

test_that("polygonize yields one dissolved polygon per value", {
  skip_if_not_installed("sf")
  m <- matrix(c(1, 1, 2, 2, 1, 1, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3), 4, 4, byrow = TRUE)
  f <- wr(m, tile = 2L)
  on.exit(unlink(f))

  pg <- collect_sf(polygonize(f))
  expect_equal(nrow(pg), 3L)
  expect_setequal(pg$value, c(1, 2, 3))
  ab <- area_by_value(pg)
  expect_equal(as.numeric(ab[order(as.numeric(names(ab)))]), c(4, 4, 8))
})

test_that("polygonize per-value area matches terra::as.polygons", {
  skip_if_not_installed("sf")
  skip_if_not_installed("terra")
  set.seed(11)
  m <- matrix(sample(1:4, 24 * 24, replace = TRUE), 24, 24)
  f <- wr(m, tile = 6L)
  on.exit(unlink(f))

  pg <- collect_sf(polygonize(f))
  va <- area_by_value(pg)

  tr <- terra::rast(m, extent = terra::ext(0, 24, 0, 24))
  tp <- sf::st_as_sf(terra::as.polygons(tr))
  ta <- tapply(as.numeric(sf::st_area(tp)), tp[[1]], sum)

  expect_equal(as.numeric(va[order(as.numeric(names(va)))]),
               as.numeric(ta[order(as.numeric(names(ta)))]), tolerance = 1e-9)
})

test_that("polygonize per-value area is invariant to tile size", {
  skip_if_not_installed("sf")
  set.seed(12)
  m <- matrix(sample(1:3, 18 * 18, replace = TRUE), 18, 18)
  f1 <- wr(m, tile = 3L); f2 <- wr(m, tile = 18L)
  on.exit(unlink(c(f1, f2)))

  a1 <- area_by_value(collect_sf(polygonize(f1)))
  a2 <- area_by_value(collect_sf(polygonize(f2)))
  expect_equal(a1[order(as.numeric(names(a1)))],
               a2[order(as.numeric(names(a2)))], tolerance = 1e-9)
})

test_that("dissolve = FALSE emits one square per cell", {
  skip_if_not_installed("sf")
  m <- matrix(c(1, 1, 2, 2), 2, 2, byrow = TRUE)
  f <- wr(m, tile = 1L)
  on.exit(unlink(f))

  pc <- collect_sf(polygonize(f, dissolve = FALSE))
  expect_equal(nrow(pc), 4L)
  expect_equal(sum(as.numeric(sf::st_area(pc))), 4)
})

test_that("na_rm drops nodata cells", {
  skip_if_not_installed("sf")
  m <- matrix(c(1, NA, NA, 2), 2, 2, byrow = TRUE)
  f <- wr(m, tile = 2L)
  on.exit(unlink(f))

  pg <- collect_sf(polygonize(f))
  expect_setequal(pg$value, c(1, 2))
  expect_equal(sum(as.numeric(sf::st_area(pg))), 2)
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----example-data-------------------------------------------------------------
library(vectra)

obs_path <- tempfile(fileext = ".vtr")
sites_path <- tempfile(fileext = ".vtr")
ref_path <- tempfile(fileext = ".vtr")

observations <- data.frame(
  site_id = c(1, 1, 2, 2, 3, 3, 4, 4),
  species = c("Quercus robur", "Pinus sylvestris",
              "Quercus robur", "Fagus sylvatica",
              "Pinus sylvestris", "Betula pendula",
              "Fagus sylvatica", "Acer pseudoplatanus"),
  date = c("2024-06-01", "2024-06-01", "2024-06-15",
           "2024-06-15", "2024-07-01", "2024-07-01",
           "2024-07-10", "2024-07-10"),
  count = c(12, 5, 8, 3, 7, 14, 6, 2),
  stringsAsFactors = FALSE
)

sites <- data.frame(
  site_id = c(1, 2, 3, 5),
  latitude = c(48.21, 47.07, 46.62, 48.85),
  longitude = c(16.37, 15.44, 14.31, 13.02),
  habitat = c("forest", "wetland", "grassland", "forest"),
  stringsAsFactors = FALSE
)

reference <- data.frame(
  species_name = c("Quercus robur", "Pinus sylvestris",
                   "Fagus sylvatica", "Betula pendula",
                   "Alnus glutinosa"),
  family = c("Fagaceae", "Pinaceae", "Fagaceae",
             "Betulaceae", "Betulaceae"),
  conservation_status = c("LC", "LC", "LC", "LC", "LC"),
  stringsAsFactors = FALSE
)

write_vtr(observations, obs_path)
write_vtr(sites, sites_path)
write_vtr(reference, ref_path)

## ----by-same-name-------------------------------------------------------------
left_join(tbl(obs_path), tbl(sites_path), by = "site_id") |>
  collect()

## ----by-named-----------------------------------------------------------------
inner_join(
  tbl(obs_path),
  tbl(ref_path),
  by = c("species" = "species_name")
) |> collect()

## ----by-natural---------------------------------------------------------------
left_join(tbl(obs_path), tbl(sites_path)) |>
  collect()

## ----multi-key-setup----------------------------------------------------------
survey1_path <- tempfile(fileext = ".vtr")
survey2_path <- tempfile(fileext = ".vtr")

write_vtr(data.frame(
  site_id = c(1, 1, 2, 2),
  year = c(2023, 2024, 2023, 2024),
  richness = c(5, 7, 3, 4)
), survey1_path)

write_vtr(data.frame(
  site_id = c(1, 2, 2),
  year = c(2024, 2023, 2024),
  temperature = c(18.2, 16.1, 17.5)
), survey2_path)

inner_join(
  tbl(survey1_path), tbl(survey2_path),
  by = c("site_id", "year")
) |> collect()

## ----left-join----------------------------------------------------------------
enriched <- left_join(
  tbl(obs_path), tbl(sites_path), by = "site_id"
) |> collect()
enriched

## ----left-join-suffix---------------------------------------------------------
extra_path <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  site_id = c(1, 2, 3),
  count = c(100, 200, 300)
), extra_path)

left_join(
  tbl(obs_path), tbl(extra_path),
  by = "site_id", suffix = c("_obs", "_site")
) |> collect()

## ----inner-join---------------------------------------------------------------
inner_join(
  tbl(obs_path), tbl(sites_path), by = "site_id"
) |> collect()

## ----right-join---------------------------------------------------------------
right_join(
  tbl(obs_path), tbl(sites_path), by = "site_id"
) |> collect()

## ----full-join----------------------------------------------------------------
full_join(
  tbl(obs_path), tbl(sites_path), by = "site_id"
) |> collect()

## ----semi-join-setup----------------------------------------------------------
targets_path <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  species = c("Quercus robur", "Betula pendula"),
  stringsAsFactors = FALSE
), targets_path)

## ----semi-join----------------------------------------------------------------
semi_join(
  tbl(obs_path), tbl(targets_path), by = "species"
) |> collect()

## ----anti-join----------------------------------------------------------------
anti_join(
  tbl(obs_path),
  tbl(ref_path),
  by = c("species" = "species_name")
) |> collect()

## ----anti-join-reverse--------------------------------------------------------
anti_join(
  tbl(ref_path),
  tbl(obs_path),
  by = c("species_name" = "species")
) |> collect()

## ----cross-join---------------------------------------------------------------
years_path <- tempfile(fileext = ".vtr")
write_vtr(data.frame(year = c(2022, 2023, 2024)), years_path)

grid <- cross_join(tbl(sites_path), tbl(years_path))
grid

## ----fuzzy-setup--------------------------------------------------------------
messy_path <- tempfile(fileext = ".vtr")
clean_path <- tempfile(fileext = ".vtr")

write_vtr(data.frame(
  obs_id = c(1, 2, 3, 4, 5),
  name = c("Quercus robar", "Pinus sylvestris L.",
           "Fagus silvatica", "Betula pendla",
           "Acer pseudoplatanus"),
  stringsAsFactors = FALSE
), messy_path)

write_vtr(data.frame(
  species_name = c("Quercus robur", "Pinus sylvestris",
                   "Fagus sylvatica", "Betula pendula",
                   "Acer pseudoplatanus"),
  family = c("Fagaceae", "Pinaceae", "Fagaceae",
             "Betulaceae", "Sapindaceae"),
  stringsAsFactors = FALSE
), clean_path)

## ----fuzzy-join---------------------------------------------------------------
fuzzy_join(
  tbl(messy_path),
  tbl(clean_path),
  by = c("name" = "species_name"),
  method = "dl",
  max_dist = 0.2
) |> collect()

## ----fuzzy-block-setup--------------------------------------------------------
messy2_path <- tempfile(fileext = ".vtr")
clean2_path <- tempfile(fileext = ".vtr")

write_vtr(data.frame(
  genus = c("Quercus", "Pinus", "Fagus", "Betula"),
  name = c("Quercus robar", "Pinus sylvestris L.",
           "Fagus silvatica", "Betula pendla"),
  stringsAsFactors = FALSE
), messy2_path)

write_vtr(data.frame(
  genus = c("Quercus", "Pinus", "Fagus",
            "Betula", "Alnus"),
  species_name = c("Quercus robur", "Pinus sylvestris",
                   "Fagus sylvatica", "Betula pendula",
                   "Alnus glutinosa"),
  stringsAsFactors = FALSE
), clean2_path)

## ----fuzzy-block--------------------------------------------------------------
fuzzy_join(
  tbl(messy2_path),
  tbl(clean2_path),
  by = c("name" = "species_name"),
  method = "dl",
  max_dist = 0.25,
  block_by = c("genus" = "genus")
) |> collect()

## ----multi-key-data-----------------------------------------------------------
counts_path <- tempfile(fileext = ".vtr")
effort_path <- tempfile(fileext = ".vtr")

write_vtr(data.frame(
  site_id = c(1, 1, 2, 2, 3),
  year = c(2023, 2024, 2023, 2024, 2024),
  n_species = c(12, 15, 8, 10, 20)
), counts_path)

write_vtr(data.frame(
  site_id = c(1, 1, 2, 3, 3),
  year = c(2023, 2024, 2024, 2023, 2024),
  hours = c(4.5, 5.0, 3.0, 6.0, 4.0)
), effort_path)

## ----multi-key-join-----------------------------------------------------------
left_join(
  tbl(counts_path), tbl(effort_path),
  by = c("site_id", "year")
) |> collect()

## ----multi-key-named----------------------------------------------------------
effort2_path <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  loc = c(1, 1, 2, 3, 3),
  survey_year = c(2023, 2024, 2024, 2023, 2024),
  hours = c(4.5, 5.0, 3.0, 6.0, 4.0)
), effort2_path)

left_join(
  tbl(counts_path), tbl(effort2_path),
  by = c("site_id" = "loc", "year" = "survey_year")
) |> collect()

## ----coercion-----------------------------------------------------------------
int_path <- tempfile(fileext = ".vtr")
dbl_path <- tempfile(fileext = ".vtr")

write_vtr(data.frame(id = c(1L, 2L, 3L), x = c(10, 20, 30)), int_path)
write_vtr(data.frame(id = c(1.0, 2.0, 4.0), y = c(100, 200, 400)), dbl_path)

inner_join(tbl(int_path), tbl(dbl_path), by = "id") |> collect()

## ----na-keys------------------------------------------------------------------
na_left <- tempfile(fileext = ".vtr")
na_right <- tempfile(fileext = ".vtr")

write_vtr(data.frame(
  id = c(1, NA, 3), x = c(10, 20, 30)
), na_left)
write_vtr(data.frame(
  id = c(1, NA, 4), y = c(100, 200, 400)
), na_right)

left_join(tbl(na_left), tbl(na_right), by = "id") |> collect()

## ----guidance-enrich----------------------------------------------------------
# Attach habitat type to every observation
left_join(tbl(obs_path), tbl(sites_path), by = "site_id") |>
  select(site_id, species, habitat) |>
  collect()

## ----guidance-preprocess------------------------------------------------------
# Preprocessing before an exact join
tbl(obs_path) |>
  mutate(species_clean = trimws(tolower(species))) |>
  select(site_id, species_clean, count) |>
  collect()

## ----cleanup, include = FALSE-------------------------------------------------
unlink(c(obs_path, sites_path, ref_path, targets_path,
         extra_path, survey1_path, survey2_path,
         years_path, messy_path, clean_path,
         messy2_path, clean2_path,
         counts_path, effort_path, effort2_path,
         int_path, dbl_path, na_left, na_right))

#' Print a vectra query node
#'
#' @param x A `vectra_node` object.
#' @param ... Ignored.
#'
#' @return Invisible `x`.
#'
#' @export
print.vectra_node <- function(x, ...) {
  schema <- .Call(C_node_schema, x$.node)
  cat("vectra query node\n")
  cat(sprintf("Columns (%d):\n", length(schema$name)))
  for (i in seq_along(schema$name)) {
    cat(sprintf("  %s <%s>\n", schema$name[i], schema$type[i]))
  }
  if (!is.null(x$.grade)) print(x$.grade)
  invisible(x)
}

#' Get a glimpse of a vectra table
#'
#' Shows column names, types, and a preview of the first few values without
#' collecting the full result.
#'
#' @param x A `vectra_node` object.
#' @param width Maximum number of preview rows to fetch (default 5).
#' @param ... Ignored.
#'
#' @return Invisible `x`.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> glimpse()
#' unlink(f)
#'
#' @export
glimpse <- function(x, width = 5L, ...) {
  UseMethod("glimpse")
}

#' @export
glimpse.vectra_node <- function(x, width = 5L, ...) {
  schema <- .Call(C_node_schema, x$.node)
  n_cols <- length(schema$name)

  type_map <- c("0" = "int64", "1" = "double", "2" = "bool", "3" = "string")
  types <- type_map[as.character(schema$type)]

  cat(sprintf("vectra lazy table [? x %d]\n", n_cols))

  # Fetch preview rows
  preview <- head(x, width)
  if (nrow(preview) == 0) {
    for (i in seq_len(n_cols)) {
      cat(sprintf("$ %-15s <%s>\n", schema$name[i], types[i]))
    }
  } else {
    term_width <- getOption("width", 80L)
    for (i in seq_len(n_cols)) {
      col <- preview[[schema$name[i]]]
      vals <- if (is.character(col)) {
        paste0('"', col, '"')
      } else {
        format(col, trim = TRUE)
      }
      vals_str <- paste(vals, collapse = ", ")
      prefix <- sprintf("$ %-15s <%s> ", schema$name[i], types[i])
      avail <- term_width - nchar(prefix)
      if (nchar(vals_str) > avail) {
        vals_str <- paste0(substr(vals_str, 1, avail - 3), "...")
      }
      cat(prefix, vals_str, "\n", sep = "")
    }
  }

  invisible(x)
}

# Utility/convenience verbs: rename, relocate, transmute, distinct, pull,
# head, tail, slice_head, slice_tail, slice_min, slice_max, slice

#' Rename columns
#'
#' @param .data A `vectra_node` object.
#' @param ... Rename pairs: `new_name = old_name`.
#'
#' @return A new `vectra_node` with renamed columns.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> rename(miles_per_gallon = mpg) |> collect() |> head()
#' unlink(f)
#'
#' @export
rename <- function(.data, ...) {
  UseMethod("rename")
}

#' @export
rename.vectra_node <- function(.data, ...) {
  schema <- .Call(C_node_schema, .data$.node)
  existing <- schema$name
  proxy <- schema_proxy(schema)

  sel <- tidyselect::eval_rename(rlang::expr(c(...)), data = proxy)
  new_names <- names(sel)
  old_names <- unname(existing[sel])

  # Build project: pass-through all columns, with col_ref exprs for renames
  out_names <- existing
  expr_lists <- vector("list", length(out_names))
  for (i in seq_along(old_names)) {
    idx <- match(old_names[i], out_names)
    out_names[idx] <- new_names[i]
    expr_lists[[idx]] <- list(kind = "col_ref", name = old_names[i])
  }
  new_xptr <- .Call(C_project_node, .data$.node, out_names, expr_lists)
  # Update group names if any were renamed
  grps <- .data$.groups
  if (!is.null(grps)) {
    for (i in seq_along(old_names)) {
      grps[grps == old_names[i]] <- new_names[i]
    }
  }
  structure(list(.node = new_xptr, .path = .data$.path,
                 .groups = grps), class = "vectra_node")
}

#' Relocate columns
#'
#' @param .data A `vectra_node` object.
#' @param ... Column names to move.
#' @param .before Column name to place before (unquoted).
#' @param .after Column name to place after (unquoted).
#'
#' @return A new `vectra_node` with reordered columns.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> relocate(hp, wt, .before = cyl) |> collect() |> head()
#' unlink(f)
#'
#' @export
relocate <- function(.data, ..., .before = NULL, .after = NULL) {
  UseMethod("relocate")
}

#' @export
relocate.vectra_node <- function(.data, ..., .before = NULL, .after = NULL) {
  schema <- .Call(C_node_schema, .data$.node)
  existing <- schema$name
  proxy <- schema_proxy(schema)

  sel <- tidyselect::eval_select(rlang::expr(c(...)), data = proxy)
  to_move <- unname(existing[sel])

  .before <- if (!missing(.before)) {
    bsel <- tidyselect::eval_select(rlang::enquo(.before), data = proxy)
    unname(existing[bsel])
  } else NULL
  .after <- if (!missing(.after)) {
    asel <- tidyselect::eval_select(rlang::enquo(.after), data = proxy)
    unname(existing[asel])
  } else NULL

  remaining <- setdiff(existing, to_move)

  if (!is.null(.before)) {
    pos <- match(.before[1], remaining)
    if (is.na(pos)) stop(sprintf(".before column not found: %s", .before[1]))
    if (pos > 1) {
      out_names <- c(remaining[seq_len(pos - 1)], to_move, remaining[pos:length(remaining)])
    } else {
      out_names <- c(to_move, remaining)
    }
  } else if (!is.null(.after)) {
    pos <- match(.after[1], remaining)
    if (is.na(pos)) stop(sprintf(".after column not found: %s", .after[1]))
    if (pos < length(remaining)) {
      out_names <- c(remaining[seq_len(pos)], to_move, remaining[(pos + 1):length(remaining)])
    } else {
      out_names <- c(remaining, to_move)
    }
  } else {
    out_names <- c(to_move, remaining)
  }

  expr_lists <- vector("list", length(out_names))
  new_xptr <- .Call(C_project_node, .data$.node, out_names, expr_lists)
  structure(list(.node = new_xptr, .path = .data$.path,
                 .groups = .data$.groups), class = "vectra_node")
}

#' Keep only columns from mutate expressions
#'
#' Like [mutate()] but drops all other columns.
#'
#' @param .data A `vectra_node` object.
#' @param ... Named expressions.
#'
#' @return A new `vectra_node` with only the computed columns.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> transmute(kpl = mpg * 0.425) |> collect() |> head()
#' unlink(f)
#'
#' @export
transmute <- function(.data, ...) {
  UseMethod("transmute")
}

#' @export
transmute.vectra_node <- function(.data, ...) {
  dots <- eval(substitute(alist(...)))
  # Expand across() calls
  schema <- .Call(C_node_schema, .data$.node)
  proxy <- schema_proxy(schema)
  dots <- expand_across(dots, schema$name, parent.frame(), proxy)
  dot_names <- names(dots)
  if (is.null(dot_names) || any(dot_names == ""))
    stop("all transmute expressions must be named")

  out_names <- character(length(dots))
  out_exprs <- vector("list", length(dots))
  for (i in seq_along(dots)) {
    out_names[i] <- dot_names[i]
    out_exprs[[i]] <- serialize_expr(dots[[i]], parent.frame(), schema$name)
  }

  new_xptr <- .Call(C_project_node, .data$.node, out_names, out_exprs)
  structure(list(.node = new_xptr, .path = .data$.path), class = "vectra_node")
}

#' Keep distinct/unique rows
#'
#' @param .data A `vectra_node` object.
#' @param ... Column names (unquoted). If empty, uses all columns.
#' @param .keep_all If `TRUE`, keep all columns (not just those in `...`).
#'
#' @return A `vectra_node` with unique rows.
#'
#' @details
#' Uses hash-based grouping with zero aggregations. When `.keep_all = TRUE`
#' with a column subset, falls back to R's `duplicated()` with a message.
#'
#' This is a materializing operation.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> distinct(cyl) |> collect()
#' unlink(f)
#'
#' @export
distinct <- function(.data, ..., .keep_all = FALSE) {
  UseMethod("distinct")
}

#' @export
distinct.vectra_node <- function(.data, ..., .keep_all = FALSE) {
  schema <- .Call(C_node_schema, .data$.node)
  proxy <- schema_proxy(schema)

  col_exprs <- eval(substitute(alist(...)))
  if (length(col_exprs) == 0) {
    key_names <- schema$name
  } else {
    sel <- tidyselect::eval_select(rlang::expr(c(...)), data = proxy)
    key_names <- unname(schema$name[sel])
  }

  if (.keep_all && length(col_exprs) > 0) {
    # .keep_all with subset of columns: fall back to collect + base R
    message("distinct(.keep_all = TRUE) with column subset: falling back to R")
    df <- collect(.data)
    return(df[!duplicated(df[, key_names, drop = FALSE]), , drop = FALSE])
  }

  # Use group_agg with zero aggregations to get unique key combos
  agg_specs <- list()
  new_xptr <- .Call(C_group_agg_node, .data$.node, key_names, agg_specs)
  structure(list(.node = new_xptr, .path = .data$.path), class = "vectra_node")
}

#' Extract a single column as a vector
#'
#' @param .data A `vectra_node` object.
#' @param var Column name (unquoted) or positive integer position.
#'
#' @return A vector.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> pull(mpg) |> head()
#' unlink(f)
#'
#' @export
pull <- function(.data, var = -1) {
  UseMethod("pull")
}

#' @export
pull.vectra_node <- function(.data, var = -1) {
  var_expr <- substitute(var)
  # Validate length 1 for literal values (not symbols that need schema lookup)
  if (!is.name(var_expr)) {
    val <- eval(var_expr, parent.frame())
    if (length(val) != 1)
      stop(sprintf("var must be length 1, got length %d", length(val)))
  }
  schema <- .Call(C_node_schema, .data$.node)

  if (is.name(var_expr)) {
    nm <- as.character(var_expr)
    if (nm %in% schema$name) {
      col_name <- nm
    } else {
      # Could be a variable in the caller's env
      val <- eval(var_expr, parent.frame())
      if (is.numeric(val)) {
        idx <- as.integer(val)
        if (idx < 0) idx <- length(schema$name) + idx + 1L
        if (idx < 1 || idx > length(schema$name))
          stop(sprintf("column index %d out of range (1:%d)", idx, length(schema$name)))
        col_name <- schema$name[idx]
      } else {
        col_name <- as.character(val)
      }
    }
  } else {
    val <- eval(var_expr, parent.frame())
    if (is.numeric(val)) {
      idx <- as.integer(val)
      if (idx < 0) idx <- length(schema$name) + idx + 1L
      if (idx < 1 || idx > length(schema$name))
        stop(sprintf("column index %d out of range (1:%d)", idx, length(schema$name)))
      col_name <- schema$name[idx]
    } else {
      col_name <- as.character(val)
    }
  }

  # Select just the one column, collect, extract
  expr_lists <- list(NULL)
  new_xptr <- .Call(C_project_node, .data$.node, col_name, expr_lists)
  result <- .Call(C_collect, new_xptr)
  result[[1]]
}

#' Limit results to first n rows
#'
#' @param x A `vectra_node` object.
#' @param n Number of rows to return.
#' @param ... Ignored.
#'
#' @return A data.frame with the first `n` rows.
#'
#' @importFrom utils head
#' @export
head.vectra_node <- function(x, n = 6L, ...) {
  new_xptr <- .Call(C_limit_node, x$.node, as.double(n))
  node <- structure(list(.node = new_xptr, .path = x$.path), class = "vectra_node")
  collect(node)
}

#' Select first or last rows
#'
#' @param .data A `vectra_node` object.
#' @param n Number of rows to select.
#' @param order_by Column to order by (for `slice_min`/`slice_max`).
#' @param with_ties If `TRUE` (default), includes all rows that tie with the
#'   nth value. If `FALSE`, returns exactly `n` rows.
#'
#' @details
#' When `slice_min()`/`slice_max()` follow [group_by()], the n smallest/largest
#' rows are taken within each group and the whole winning row is kept (every
#' column, including geometry carried as a string). `with_ties = FALSE` returns
#' exactly `n` rows per group; `with_ties = TRUE` keeps rows tied at the nth
#' value via min-rank. The `n = 1`, `with_ties = FALSE` case streams: it holds
#' only the running winner per group, so memory scales with the number of groups
#' (the result size), not the input. Other grouped cases buffer their input.
#'
#' @return A `vectra_node` for `slice_head()`, for grouped
#'   `slice_min()`/`slice_max()`, and for ungrouped `slice_min/max(...,
#'   with_ties = FALSE)`. A data.frame for `slice_tail()` and ungrouped
#'   `slice_min/max(..., with_ties = TRUE)` (the default), since these must
#'   materialize all rows.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> slice_head(n = 3) |> collect()
#' tbl(f) |> slice_min(order_by = mpg, n = 3) |> collect()
#' tbl(f) |> slice_max(order_by = mpg, n = 3) |> collect()
#' # earliest row per group, geometry/attrs preserved:
#' tbl(f) |> group_by(cyl) |> slice_min(mpg, n = 1, with_ties = FALSE) |> collect()
#' unlink(f)
#'
#' @export
slice_head <- function(.data, n = 1L) {
  UseMethod("slice_head")
}

#' @export
slice_head.vectra_node <- function(.data, n = 1L) {
  if (!is.numeric(n) || length(n) != 1 || is.na(n) || n < 1 || n != floor(n))
    stop(sprintf("n must be a positive integer, got %s", deparse(n)))
  new_xptr <- .Call(C_limit_node, .data$.node, as.double(n))
  structure(list(.node = new_xptr, .path = .data$.path), class = "vectra_node")
}

#' @rdname slice_head
#' @export
slice_tail <- function(.data, n = 1L) {
  UseMethod("slice_tail")
}

#' @export
slice_tail.vectra_node <- function(.data, n = 1L) {
  if (!is.numeric(n) || length(n) != 1 || is.na(n) || n < 1 || n != floor(n))
    stop(sprintf("n must be a positive integer, got %s", deparse(n)))
  # Must materialize to know total rows, then take last n
  df <- collect(.data)
  nr <- nrow(df)
  if (n >= nr) return(df)
  df[(nr - n + 1):nr, , drop = FALSE]
}

# Group-aware slice: keep the n rows with the smallest (desc = FALSE) or largest
# (desc = TRUE) order_col within each group. with_ties = FALSE uses an ordered
# row_number (exactly n per group, deterministic); with_ties = TRUE uses min-rank
# (includes rows tied at the nth value). Runs through the window node, so the
# winning whole row -- geometry and all attributes -- is preserved.
#
# n == 1 without ties is the common case (one winner per group). It routes to
# the streaming grouped-top-n node, which holds only the running champion per
# group -- O(#groups) memory -- instead of materializing every input row to
# rank it. That keeps the whole winning row, geometry included, and lets the
# resolution survive overlay outputs whose geometry column dwarfs RAM.
.grouped_slice_topn <- function(.data, order_col, n, with_ties, desc) {
  if (n == 1L && !with_ties) {
    keys <- .data$.groups
    new_xptr <- .Call(C_group_topn_node, .data$.node, keys, order_col, desc)
    return(structure(list(.node = new_xptr, .path = .data$.path,
                          .groups = .data$.groups), class = "vectra_node"))
  }
  schema <- .Call(C_node_schema, .data$.node)
  orig_names <- schema$name
  rank_col <- ".__vtr_slice_rank"
  kind <- if (with_ties) "rank" else "row_number"
  spec <- list(list(name = rank_col, kind = kind, col = order_col,
                    offset = 1L, default = NULL, desc = desc))
  win <- create_window_node(.data, spec)
  pred <- combine_predicates(list(bquote(.(as.name(rank_col)) <= .(n))),
                             environment(), c(orig_names, rank_col))
  filt <- .Call(C_filter_node, win$.node, pred)
  proj <- .Call(C_project_node, filt, orig_names,
                vector("list", length(orig_names)))
  structure(list(.node = proj, .path = .data$.path,
                 .groups = .data$.groups), class = "vectra_node")
}

#' @rdname slice_head
#' @export
slice_min <- function(.data, order_by, n = 1L, with_ties = TRUE) {
  UseMethod("slice_min")
}

#' @export
slice_min.vectra_node <- function(.data, order_by, n = 1L, with_ties = TRUE) {
  if (!is.numeric(n) || length(n) != 1 || is.na(n) || n < 1 || n != floor(n))
    stop(sprintf("n must be a positive integer, got %s", deparse(n)))
  if (!is.logical(with_ties) || length(with_ties) != 1 || is.na(with_ties))
    stop(sprintf("with_ties must be TRUE or FALSE, got %s", deparse(with_ties)))
  order_col <- as.character(substitute(order_by))
  if (!is.null(.data$.groups) && length(.data$.groups) > 0)
    return(.grouped_slice_topn(.data, order_col, n, with_ties, desc = FALSE))
  if (!with_ties) {
    new_xptr <- .Call(C_topn_node, .data$.node, order_col, FALSE,
                      as.double(n))
    return(structure(list(.node = new_xptr, .path = .data$.path),
                     class = "vectra_node"))
  }
  # with_ties = TRUE: collect all data, sort, find the nth value, keep all
  # rows that tie with it. Must collect first because C nodes are single-use.
  df <- collect(.data)
  if (nrow(df) == 0) return(df)
  vals <- df[[order_col]]
  ord <- order(vals, na.last = TRUE)
  # Take at most n non-NA values; if fewer than n non-NA, include NAs up to n
  n_nonNA <- sum(!is.na(vals))
  take <- min(n, nrow(df))
  selected <- ord[seq_len(take)]
  result <- df[selected, , drop = FALSE]
  # Check for ties: if there are more rows beyond n with same boundary value
  if (take <= n_nonNA && take < nrow(df)) {
    boundary <- vals[ord[take]]
    extra <- which(!is.na(vals) & vals == boundary)
    all_keep <- union(selected, extra)
    result <- df[sort(all_keep), , drop = FALSE]
  }
  result[order(result[[order_col]], na.last = TRUE), , drop = FALSE]
}

#' @rdname slice_head
#' @export
slice_max <- function(.data, order_by, n = 1L, with_ties = TRUE) {
  UseMethod("slice_max")
}

#' @export
slice_max.vectra_node <- function(.data, order_by, n = 1L, with_ties = TRUE) {
  if (!is.numeric(n) || length(n) != 1 || is.na(n) || n < 1 || n != floor(n))
    stop(sprintf("n must be a positive integer, got %s", deparse(n)))
  if (!is.logical(with_ties) || length(with_ties) != 1 || is.na(with_ties))
    stop(sprintf("with_ties must be TRUE or FALSE, got %s", deparse(with_ties)))
  order_col <- as.character(substitute(order_by))
  if (!is.null(.data$.groups) && length(.data$.groups) > 0)
    return(.grouped_slice_topn(.data, order_col, n, with_ties, desc = TRUE))
  if (!with_ties) {
    new_xptr <- .Call(C_topn_node, .data$.node, order_col, TRUE,
                      as.double(n))
    return(structure(list(.node = new_xptr, .path = .data$.path),
                     class = "vectra_node"))
  }
  df <- collect(.data)
  if (nrow(df) == 0) return(df)
  vals <- df[[order_col]]
  ord <- order(vals, decreasing = TRUE, na.last = TRUE)
  n_nonNA <- sum(!is.na(vals))
  take <- min(n, nrow(df))
  selected <- ord[seq_len(take)]
  result <- df[selected, , drop = FALSE]
  if (take <= n_nonNA && take < nrow(df)) {
    boundary <- vals[ord[take]]
    extra <- which(!is.na(vals) & vals == boundary)
    all_keep <- union(selected, extra)
    result <- df[sort(all_keep), , drop = FALSE]
  }
  result[order(result[[order_col]], decreasing = TRUE, na.last = TRUE), , drop = FALSE]
}

#' Select rows by position
#'
#' @param .data A `vectra_node` object.
#' @param ... Integer row indices (positive or negative).
#'
#' @return A data.frame with the selected rows.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> slice(1, 3, 5)
#' unlink(f)
#'
#' @export
slice <- function(.data, ...) {
  UseMethod("slice")
}

#' @export
slice.vectra_node <- function(.data, ...) {
  indices <- c(...)
  df <- collect(.data)
  if (all(indices > 0)) {
    indices <- indices[indices <= nrow(df)]
    df[indices, , drop = FALSE]
  } else if (all(indices < 0)) {
    df[indices, , drop = FALSE]
  } else {
    stop("slice indices must be all positive or all negative")
  }
}

# --- sd ---

test_that("sd computes sample standard deviation", {
  df <- data.frame(g = c("a", "a", "a"), x = c(2.0, 4.0, 6.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(s = sd(x)) |> collect()
  expect_equal(result$s, sd(c(2, 4, 6)))
})

test_that("sd of single value returns NA", {
  df <- data.frame(g = c("a"), x = c(5.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(s = sd(x)) |> collect()
  expect_true(is.na(result$s))
})

test_that("sd with na.rm = TRUE skips NAs", {
  df <- data.frame(g = c("a", "a", "a"), x = c(2.0, NA, 6.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(s = sd(x, na.rm = TRUE)) |> collect()
  expect_equal(result$s, sd(c(2, 6)))
})

test_that("sd without na.rm returns NA when group has NAs", {
  df <- data.frame(g = c("a", "a", "a"), x = c(2.0, NA, 6.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(s = sd(x)) |> collect()
  expect_true(is.na(result$s))
})

# --- var ---

test_that("var computes sample variance", {
  df <- data.frame(g = c("a", "a", "a"), x = c(2.0, 4.0, 6.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(v = var(x)) |> collect()
  expect_equal(result$v, var(c(2, 4, 6)))
})

test_that("var of single value returns NA", {
  df <- data.frame(g = c("a"), x = c(5.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(v = var(x)) |> collect()
  expect_true(is.na(result$v))
})

# --- first ---

test_that("first returns the first value per group", {
  df <- data.frame(g = c("a", "a", "b", "b"),
                   x = c(10.0, 20.0, 30.0, 40.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(f = first(x)) |> collect()
  expect_equal(result$f[result$g == "a"], 10)
  expect_equal(result$f[result$g == "b"], 30)
})

test_that("first with na.rm = TRUE skips leading NAs", {
  df <- data.frame(g = c("a", "a", "a"), x = c(NA, 2.0, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(f = first(x, na.rm = TRUE)) |> collect()
  expect_equal(result$f, 2)
})

test_that("first without na.rm returns NA when first value is NA", {
  df <- data.frame(g = c("a", "a"), x = c(NA, 2.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(f = first(x)) |> collect()
  expect_true(is.na(result$f))
})

# --- last ---

test_that("last returns the last value per group", {
  df <- data.frame(g = c("a", "a", "b", "b"),
                   x = c(10.0, 20.0, 30.0, 40.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(l = last(x)) |> collect()
  expect_equal(result$l[result$g == "a"], 20)
  expect_equal(result$l[result$g == "b"], 40)
})

test_that("last with na.rm = TRUE skips trailing NAs", {
  df <- data.frame(g = c("a", "a", "a"), x = c(1.0, 2.0, NA),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(l = last(x, na.rm = TRUE)) |> collect()
  expect_equal(result$l, 2)
})

# --- sd/var with multiple groups ---

test_that("sd works across multiple groups", {
  df <- data.frame(g = c("a", "a", "a", "b", "b", "b"),
                   x = c(1.0, 2.0, 3.0, 10.0, 20.0, 30.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(s = sd(x)) |> collect()
  expect_equal(result$s[result$g == "a"], sd(c(1, 2, 3)))
  expect_equal(result$s[result$g == "b"], sd(c(10, 20, 30)))
})

# --- slice_min/max with_ties ---

test_that("slice_min with_ties = TRUE includes tied rows", {
  df <- data.frame(x = c(1.0, 2.0, 2.0, 3.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 2, with_ties = TRUE)
  # n=2 -> boundary is 2.0, so include 1, 2, 2 = 3 rows
  expect_equal(nrow(result), 3)
  expect_true(all(result$x <= 2))
})

test_that("slice_min with_ties = FALSE returns exactly n rows", {
  df <- data.frame(x = c(1.0, 2.0, 2.0, 3.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 2, with_ties = FALSE) |> collect()
  expect_equal(nrow(result), 2)
})

test_that("slice_max with_ties = TRUE includes tied rows", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 3.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_max(order_by = x, n = 2, with_ties = TRUE)
  # n=2 -> boundary is 3.0, so include 3, 3, 4 = 3 rows
  expect_equal(nrow(result), 3)
  expect_true(all(result$x >= 3))
})

test_that("slice_max with_ties = FALSE returns exactly n rows", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 3.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_max(order_by = x, n = 2, with_ties = FALSE) |> collect()
  expect_equal(nrow(result), 2)
})

test_that("slice_min with no ties behaves same either way", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 4.0, 5.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result_ties <- tbl(f) |> slice_min(order_by = x, n = 2, with_ties = TRUE)
  result_no <- tbl(f) |> slice_min(order_by = x, n = 2, with_ties = FALSE) |> collect()
  expect_equal(nrow(result_ties), 2)
  expect_equal(nrow(result_no), 2)
})

# --- first/last with all NAs ---

test_that("first with na.rm = TRUE on all-NA group returns NA", {
  df <- data.frame(g = c("a", "a"), x = c(NA_real_, NA_real_),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(f = first(x, na.rm = TRUE)) |> collect()
  expect_true(is.na(result$f))
})

test_that("last with na.rm = TRUE on all-NA group returns NA", {
  df <- data.frame(g = c("a", "a"), x = c(NA_real_, NA_real_),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(l = last(x, na.rm = TRUE)) |> collect()
  expect_true(is.na(result$l))
})

# --- transmute with across ---

test_that("transmute supports across() in summarise context", {
  # across() with aggregation functions works in summarise, not transmute
  # Test that across() expands correctly in summarise
  df <- data.frame(x = c(1.0, 2.0, 3.0), y = c(10.0, 20.0, 30.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> summarise(across(c(x, y), mean)) |> collect()
  expect_equal(names(result), c("x", "y"))
  expect_equal(result$x, 2)
  expect_equal(result$y, 20)
})

# --- rename ---

test_that("rename changes column names", {
  df <- data.frame(a = 1:3, b = 4:6)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> rename(alpha = a, beta = b) |> collect()
  expect_equal(names(result), c("alpha", "beta"))
  expect_equal(result$alpha, c(1, 2, 3))
  expect_equal(result$beta, c(4, 5, 6))
})

test_that("rename errors on missing column", {
  df <- data.frame(a = 1:3)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  expect_error(tbl(f) |> rename(x = nonexistent))
})

# --- relocate ---

test_that("relocate moves columns to front by default", {
  df <- data.frame(a = 1, b = 2, c = 3)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> relocate(c) |> collect()
  expect_equal(names(result), c("c", "a", "b"))
})

test_that("relocate with .before", {
  df <- data.frame(a = 1, b = 2, c = 3, d = 4)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> relocate(d, .before = b) |> collect()
  expect_equal(names(result), c("a", "d", "b", "c"))
})

test_that("relocate with .after", {
  df <- data.frame(a = 1, b = 2, c = 3, d = 4)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> relocate(a, .after = c) |> collect()
  expect_equal(names(result), c("b", "c", "a", "d"))
})

test_that("relocate with .after last column", {
  df <- data.frame(a = 1, b = 2, c = 3)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> relocate(a, .after = c) |> collect()
  expect_equal(names(result), c("b", "c", "a"))
})

# --- transmute ---

test_that("transmute keeps only computed columns", {
  df <- data.frame(x = c(1.0, 2.0, 3.0), y = c(10.0, 20.0, 30.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> transmute(total = x + y) |> collect()
  expect_equal(names(result), "total")
  expect_equal(result$total, c(11, 22, 33))
})

# --- distinct ---

test_that("distinct removes duplicate rows", {
  df <- data.frame(a = c("x", "x", "y", "y"), b = c(1.0, 1.0, 2.0, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> distinct() |> collect()
  expect_equal(nrow(result), 3)
})

test_that("distinct on specific columns", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, 2.0, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> distinct(g) |> collect()
  expect_equal(nrow(result), 2)
  expect_equal(sort(result$g), c("a", "b"))
})

# --- ungroup ---

test_that("ungroup removes grouping", {
  df <- data.frame(g = c("a", "b"), x = 1:2)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  node <- tbl(f) |> group_by(g)
  expect_equal(node$.groups, "g")
  node2 <- ungroup(node)
  expect_null(node2$.groups)
})

# --- group preservation ---

test_that("filter preserves groups", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, 2.0, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  node <- tbl(f) |> group_by(g) |> filter(x > 1)
  expect_equal(node$.groups, "g")
})

test_that("select preserves groups", {
  df <- data.frame(g = c("a", "b"), x = 1:2, y = 3:4)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  node <- tbl(f) |> group_by(g) |> select(g, x)
  expect_equal(node$.groups, "g")
})

test_that("select drops groups for removed group columns", {
  df <- data.frame(g = c("a", "b"), x = 1:2, y = 3:4,
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  # Selecting without the group column drops the group
  node <- tbl(f) |> group_by(g) |> select(x, y)
  expect_null(node$.groups)
})

test_that("select with multi-group drops only removed groups", {
  df <- data.frame(a = c("x", "x", "y"), b = c("p", "q", "p"),
                   v = c(1.0, 2.0, 3.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  node <- tbl(f) |> group_by(a, b) |> select(a, v)
  expect_equal(node$.groups, "a")
})

test_that("rename preserves and updates groups", {
  df <- data.frame(g = c("a", "b"), x = 1:2)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  node <- tbl(f) |> group_by(g) |> rename(group = g)
  expect_equal(node$.groups, "group")
})

test_that("relocate preserves groups", {
  df <- data.frame(g = c("a", "b"), x = 1:2, y = 3:4)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  node <- tbl(f) |> group_by(g) |> relocate(y, .before = g)
  expect_equal(node$.groups, "g")
})

test_that("arrange preserves groups", {
  df <- data.frame(g = c("b", "a"), x = c(2.0, 1.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  node <- tbl(f) |> group_by(g) |> arrange(x)
  expect_equal(node$.groups, "g")
})

# --- count ---

test_that("count counts by group", {
  df <- data.frame(g = c("a", "a", "b"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> count(g) |> collect()
  expect_equal(result$n[result$g == "a"], 2)
  expect_equal(result$n[result$g == "b"], 1)
})

test_that("count with custom name", {
  df <- data.frame(g = c("a", "a", "b"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> count(g, name = "cnt") |> collect()
  expect_true("cnt" %in% names(result))
})

# --- tally ---

test_that("tally counts rows in grouped data", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, 2.0, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> tally() |> collect()
  expect_equal(result$n[result$g == "a"], 2)
  expect_equal(result$n[result$g == "b"], 1)
})

# --- pull ---

test_that("pull extracts a column as vector", {
  df <- data.frame(x = c(1.0, 2.0, 3.0), y = c(4.0, 5.0, 6.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> pull(x)
  expect_equal(result, c(1, 2, 3))
})

test_that("pull with negative index gets last column", {
  df <- data.frame(x = c(1.0, 2.0), y = c(3.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> pull(-1)
  expect_equal(result, c(3, 4))
})

# --- head ---

test_that("head returns first n rows", {
  df <- data.frame(x = 1:10 + 0.0)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- head(tbl(f), 3)
  expect_equal(nrow(result), 3)
  expect_equal(result$x, c(1, 2, 3))
})

# --- arrange ---

test_that("arrange sorts ascending by default", {
  df <- data.frame(x = c(3.0, 1.0, 2.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> arrange(x) |> collect()
  expect_equal(result$x, c(1, 2, 3))
})

test_that("arrange desc() sorts descending", {
  df <- data.frame(x = c(3.0, 1.0, 2.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> arrange(desc(x)) |> collect()
  expect_equal(result$x, c(3, 2, 1))
})

test_that("arrange by multiple columns", {
  df <- data.frame(g = c("a", "b", "a", "b"),
                   x = c(2.0, 1.0, 1.0, 2.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> arrange(g, x) |> collect()
  expect_equal(result$g, c("a", "a", "b", "b"))
  expect_equal(result$x, c(1, 2, 1, 2))
})

test_that("arrange puts NA last in ascending order", {
  df <- data.frame(x = c(2.0, NA, 1.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> arrange(x) |> collect()
  expect_equal(result$x[1:3], c(1, 2, 3))
  expect_true(is.na(result$x[4]))
})

test_that("arrange with strings", {
  df <- data.frame(s = c("banana", "apple", "cherry"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> arrange(s) |> collect()
  expect_equal(result$s, c("apple", "banana", "cherry"))
})

# --- slice_head / slice_tail ---

test_that("slice_head returns first n rows", {
  df <- data.frame(x = c(10.0, 20.0, 30.0, 40.0, 50.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_head(n = 3) |> collect()
  expect_equal(result$x, c(10, 20, 30))
})

test_that("slice_tail returns last n rows", {
  df <- data.frame(x = c(10.0, 20.0, 30.0, 40.0, 50.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_tail(n = 2)
  expect_equal(result$x, c(40, 50))
})

# --- slice_min / slice_max ---

test_that("slice_min returns rows with smallest values", {
  df <- data.frame(x = c(5.0, 1.0, 3.0, 2.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_min(order_by = x, n = 2) |> collect()
  expect_equal(result$x, c(1, 2))
})

test_that("slice_max returns rows with largest values", {
  df <- data.frame(x = c(5.0, 1.0, 3.0, 2.0, 4.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice_max(order_by = x, n = 2) |> collect()
  expect_equal(result$x, c(5, 4))
})

# --- .groups argument in summarise ---

test_that("summarise with .groups = 'drop' drops all groups", {
  df <- data.frame(g1 = c("a", "a", "b"), g2 = c("x", "y", "x"),
                   v = c(1.0, 2.0, 3.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g1, g2) |>
    summarise(s = sum(v), .groups = "drop")
  expect_null(result$.groups)
})

test_that("summarise with .groups = 'keep' preserves groups", {
  df <- data.frame(g1 = c("a", "a", "b"), g2 = c("x", "y", "x"),
                   v = c(1.0, 2.0, 3.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g1, g2) |>
    summarise(s = sum(v), .groups = "keep")
  expect_equal(result$.groups, c("g1", "g2"))
})

test_that("summarise default .groups = 'drop_last' drops last group", {
  df <- data.frame(g1 = c("a", "a", "b"), g2 = c("x", "y", "x"),
                   v = c(1.0, 2.0, 3.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g1, g2) |> summarise(s = sum(v))
  expect_equal(result$.groups, "g1")
})

# spatial_overlay(): self-overlay into disjoint pieces, composed with grouped
# slice_min for earliest-wins. Planar (CRS NA / 3857) so areas are exact.

skip_if_not_installed("sf")

mk <- function(x0, x1, y0 = 0, y1 = 1) sf::st_polygon(list(rbind(
  c(x0, y0), c(x1, y0), c(x1, y1), c(x0, y1), c(x0, y0))))

test_that("two overlapping polygons split into 3 pieces, overlap duplicated", {
  polys <- sf::st_sf(year = c(1990L, 2010L), geometry = sf::st_sfc(mk(0, 2), mk(1, 3)))
  df <- spatial_overlay(polys) |> collect()

  expect_equal(length(unique(df$piece_id)), 3L)   # 3 disjoint pieces
  expect_equal(nrow(df), 4L)                       # overlap carries both years
  tab <- table(df$piece_id)
  overlap <- as.integer(names(tab)[tab == 2])
  expect_equal(sort(df$year[df$piece_id == overlap]), c(1990L, 2010L))
  expect_true("geometry" %in% names(df))
})

test_that("earliest-year-wins via grouped slice_min keeps the whole piece", {
  polys <- sf::st_sf(year = c(1990L, 2010L), geometry = sf::st_sfc(mk(0, 2), mk(1, 3)))
  first <- spatial_overlay(polys) |>
    group_by(piece_id) |>
    slice_min(year, n = 1, with_ties = FALSE) |>
    collect_sf()

  expect_s3_class(first, "sf")
  expect_equal(nrow(first), 3L)
  # tolerance covers the ~1e-7 coordinate snap of the fixed-precision overlay.
  expect_equal(sort(as.numeric(sf::st_area(first))), c(1, 1, 1), tolerance = 1e-4)
  # overlap piece resolves to the earliest year (1990).
  expect_equal(sort(first$year), c(1990L, 1990L, 2010L))
})

test_that("partition area reconstructs the union (precision-robust, no slivers)", {
  geoms <- sf::st_sfc(mk(0, 3), mk(2, 5), mk(4, 7), mk(1, 2, 0, 3), mk(6, 8))
  polys <- sf::st_sf(id = seq_along(geoms), geometry = geoms)
  pieces <- spatial_overlay(polys) |>
    group_by(piece_id) |>
    slice_min(id, n = 1, with_ties = FALSE) |>
    collect_sf()

  a_union <- as.numeric(sf::st_area(sf::st_union(sf::st_geometry(polys))))
  a_part  <- sum(as.numeric(sf::st_area(pieces)))
  # tolerance covers the fixed-precision grid the overlay snaps to.
  expect_equal(a_part, a_union, tolerance = 1e-6)
})

test_that("disjoint overlap clusters are tiled independently", {
  # two separate overlapping pairs (union area 3 each) + a lone square (area 1)
  geoms <- sf::st_sfc(mk(0, 2), mk(1, 3), mk(10, 12), mk(11, 13), mk(20, 21))
  polys <- sf::st_sf(year = 1:5, geometry = geoms)
  pieces <- spatial_overlay(polys) |>
    group_by(piece_id) |>
    slice_min(year, n = 1, with_ties = FALSE) |>
    collect_sf()

  expect_equal(nrow(pieces), 7L)                  # 3 + 3 + 1 disjoint pieces
  expect_equal(sum(as.numeric(sf::st_area(pieces))), 7, tolerance = 1e-4)
})

test_that("vars selects carried attributes; bad input errors", {
  polys <- sf::st_sf(year = c(1L, 2L), keep = c("a", "b"),
                     geometry = sf::st_sfc(mk(0, 2), mk(1, 3)))
  df <- spatial_overlay(polys, vars = "year") |> collect()
  expect_true("year" %in% names(df))
  expect_false("keep" %in% names(df))

  expect_error(spatial_overlay(42), "must be an sf object")
  expect_error(spatial_overlay(polys, piece = "year"), "already exists")
  expect_error(spatial_overlay(polys, vars = "nope"), "not found")
})

test_that("CRS is carried onto the overlay node", {
  polys <- sf::st_sf(year = c(1L, 2L),
                     geometry = sf::st_sfc(mk(0, 2), mk(1, 3), crs = 3857))
  ov <- spatial_overlay(polys)
  expect_equal(sf::st_crs(collect_sf(ov)), sf::st_crs(3857))
})

test_that("coverage invariant: piece areas sum to each input's area", {
  # nested + offset squares -> high coverage multiplicity in the centre
  geoms <- sf::st_sfc(mk(0, 6, 0, 6), mk(1, 5, 1, 5), mk(2, 4, 2, 4),
                      mk(3, 7, 3, 7), mk(-1, 3, -1, 3))
  polys <- sf::st_sf(id = seq_along(geoms), geometry = geoms)
  df <- spatial_overlay(polys, vars = "id") |> collect()

  g     <- sf::st_as_sfc(structure(df$geometry, class = "WKB"), EWKB = FALSE)
  parea <- as.numeric(sf::st_area(g))
  cov   <- tapply(parea, df$id, sum)                 # area covered per source input
  cov   <- as.numeric(cov[order(as.integer(names(cov)))])
  truth <- as.numeric(sf::st_area(sf::st_geometry(polys)))
  expect_equal(cov, truth, tolerance = 1e-6)
})

test_that("invalid input polygons are repaired before overlay", {
  # self-intersecting bowtie repaired to two triangles, overlapped by a square
  bowtie <- sf::st_polygon(list(rbind(c(0, 0), c(2, 2), c(2, 0), c(0, 2), c(0, 0))))
  polys  <- sf::st_sf(id = 1:2, geometry = sf::st_sfc(bowtie, mk(1, 3, 0, 2)))
  df <- spatial_overlay(polys, vars = "id") |> collect()
  expect_gt(nrow(df), 0L)
  g <- sf::st_as_sfc(structure(df$geometry, class = "WKB"), EWKB = FALSE)
  expect_true(all(sf::st_is_valid(g)))
})

test_that("reading from a GeoPackage in batches matches the in-memory overlay", {
  skip_if_not_installed("sf")
  sq <- function(x, y, s) sf::st_polygon(list(rbind(
    c(x, y), c(x+s, y), c(x+s, y+s), c(x, y+s), c(x, y))))
  g <- vector("list", 30L); yr <- integer(30L); k <- 0L
  for (i in 0:5) for (j in 0:4) { k <- k+1L; g[[k]] <- sq(i*600, j*600, 1000); yr[k] <- 1900L+k }
  x  <- sf::st_sf(year = yr, geometry = sf::st_sfc(g), crs = 3857)
  gp <- tempfile(fileext = ".gpkg"); on.exit(unlink(gp))
  sf::st_write(x, gp, "lyr", quiet = TRUE)

  pull <- function(o) {
    d <- as.data.frame(collect(o))
    d[order(d$piece_id, d$year), c("piece_id", "year", "geometry")]
  }
  ref  <- pull(spatial_overlay(x, vars = "year"))                                  # in-memory
  file <- pull(spatial_overlay(gp, layer = "lyr", vars = "year"))                  # file, one batch
  chnk <- pull(spatial_overlay(gp, layer = "lyr", vars = "year", read_chunk = 7L)) # file, many batches

  expect_equal(nrow(file), nrow(ref))
  expect_identical(file$piece_id, ref$piece_id)
  expect_identical(file$year, ref$year)
  expect_identical(file$geometry, ref$geometry)   # geometry byte-identical
  expect_identical(chnk$geometry, ref$geometry)   # batch count does not change output
})

test_that("file overlay needs layer or query, and query without grid asks for grid", {
  skip_if_not_installed("sf")
  x  <- sf::st_sf(year = 1L, geometry = sf::st_sfc(mk(0, 1)), crs = 3857)
  gp <- tempfile(fileext = ".gpkg"); on.exit(unlink(gp))
  sf::st_write(x, gp, "lyr", quiet = TRUE)
  expect_error(spatial_overlay(gp), "layer.*query")
  expect_error(spatial_overlay(gp, query = "SELECT * FROM lyr"), "grid")
})

test_that("st_write streams a resolved overlay and matches collect_sf", {
  skip_if_not_installed("sf")
  sq <- function(x, y, s) sf::st_polygon(list(rbind(
    c(x, y), c(x+s, y), c(x+s, y+s), c(x, y+s), c(x, y))))
  g <- vector("list", 30L); yr <- integer(30L); k <- 0L
  for (i in 0:5) for (j in 0:4) { k <- k+1L; g[[k]] <- sq(i*600, j*600, 1000); yr[k] <- 1900L+k }
  x <- sf::st_sf(year = yr, geometry = sf::st_sfc(g), crs = 3857)

  resolve <- function() spatial_overlay(x, vars = "year") |>
    group_by(piece_id) |> slice_min(year, n = 1, with_ties = FALSE)
  out <- tempfile(fileext = ".gpkg"); on.exit(unlink(out))
  sf::st_write(resolve(), out, crs = sf::st_crs(x), delete_dsn = TRUE, quiet = TRUE)
  B <- sf::st_read(out, quiet = TRUE)
  A <- collect_sf(resolve(), crs = sf::st_crs(x))

  key <- function(s) { o <- order(s$piece_id, s$year, round(as.numeric(sf::st_area(s)), 6)); s[o, ] }
  A <- key(A); B <- key(B)
  expect_equal(nrow(A), nrow(B))
  expect_identical(A$piece_id, B$piece_id)
  expect_identical(A$year, B$year)
  expect_equal(as.numeric(sf::st_area(A)), as.numeric(sf::st_area(B)), tolerance = 1e-9)
  expect_true(sf::st_crs(B) == sf::st_crs(x))   # same CRS (label may differ)
})

#' Execute a lazy query and return a data.frame
#'
#' Pulls all batches from the execution plan and materializes the result as an
#' R data.frame.
#'
#' @param x A `vectra_node` object.
#' @param ... Ignored.
#'
#' @return A data.frame with the query results.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' result <- tbl(f) |> collect()
#' head(result)
#' unlink(f)
#'
#' @export
collect <- function(x, ...) {
  UseMethod("collect")
}

#' @export
collect.vectra_node <- function(x, ...) {
  .Call(C_collect, x$.node)
}

#' @export
collect.data.frame <- function(x, ...) {
  x
}

#' Create a lazy table reference from a .vtr file
#'
#' Opens a vectra1 file and returns a lazy query node. No data is read until
#' [collect()] is called.
#'
#' @param path Path to a `.vtr` file.
#'
#' @return A `vectra_node` object representing a lazy scan of the file.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' node <- tbl(f)
#' print(node)
#' unlink(f)
#'
#' @export
tbl <- function(path) {
  if (!is.character(path) || length(path) != 1)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = TRUE)
  xptr <- .Call(C_scan_node, path)
  structure(list(.node = xptr, .path = path), class = "vectra_node")
}

#' Create a lazy table reference from a CSV file
#'
#' Opens a CSV file for lazy, streaming query execution. Column types are
#' inferred from the first 1000 rows. No data is read until [collect()] is
#' called. Gzip-compressed files (`.csv.gz`) are supported transparently.
#'
#' @param path Path to a `.csv` or `.csv.gz` file.
#' @param batch_size Number of rows per batch (default 65536).
#'
#' @return A `vectra_node` object representing a lazy scan of the CSV file.
#'
#' @examples
#' f <- tempfile(fileext = ".csv")
#' write.csv(mtcars, f, row.names = FALSE)
#' node <- tbl_csv(f)
#' print(node)
#' unlink(f)
#'
#' @export
tbl_csv <- function(path, batch_size = .DEFAULT_BATCH_SIZE) {
  if (!is.character(path) || length(path) != 1)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = TRUE)
  xptr <- .Call(C_csv_scan_node, path, as.double(batch_size))
  structure(list(.node = xptr, .path = path), class = "vectra_node")
}

#' Create a lazy table reference from a SQLite database
#'
#' Opens a SQLite database and lazily scans a table. Column types are inferred
#' from declared types in the CREATE TABLE statement. All filtering, grouping,
#' and aggregation is handled by vectra's C engine --- no SQL parsing needed.
#' No data is read until [collect()] is called.
#'
#' @param path Path to a SQLite database file.
#' @param table Name of the table to scan.
#' @param batch_size Number of rows per batch (default 65536).
#'
#' @return A `vectra_node` object representing a lazy scan of the table.
#'
#' @examples
#' \donttest{
#' f <- tempfile(fileext = ".sqlite")
#' write_sqlite(mtcars, f, "cars")
#' node <- tbl_sqlite(f, "cars")
#' node |> filter(cyl == 6) |> collect()
#' unlink(f)
#' }
#'
#' @export
tbl_sqlite <- function(path, table, batch_size = .DEFAULT_BATCH_SIZE) {
  if (!is.character(path) || length(path) != 1)
    stop("path must be a single character string")
  if (!is.character(table) || length(table) != 1)
    stop("table must be a single character string")

  path <- normalizePath(path, mustWork = TRUE)
  xptr <- .Call(C_sql_scan_node, path, table, as.double(batch_size))
  structure(list(.node = xptr, .path = path, .table = table),
            class = "vectra_node")
}

#' Create a lazy table reference from an Excel (.xlsx) file
#'
#' Reads a sheet from an Excel workbook into a vectra node for lazy query
#' execution. The sheet is read into memory via
#' \code{\link[openxlsx2:read_xlsx]{openxlsx2::read_xlsx()}} and then converted
#' to vectra's internal format. Requires the \pkg{openxlsx2} package.
#'
#' @param path Path to an `.xlsx` file.
#' @param sheet Sheet to read: either a name (character) or 1-based index
#'   (integer). Default `1L` (first sheet).
#' @param batch_size Number of rows per batch (default 65536).
#'
#' @return A `vectra_node` object representing a lazy scan of the sheet.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("openxlsx2", quietly = TRUE)) {
#'   f <- tempfile(fileext = ".xlsx")
#'   openxlsx2::write_xlsx(mtcars, f)
#'   node <- tbl_xlsx(f)
#'   node |> filter(cyl == 6) |> collect()
#'   unlink(f)
#' }
#' }
#'
#' @export
tbl_xlsx <- function(path, sheet = 1L, batch_size = .DEFAULT_BATCH_SIZE) {
  if (!is.character(path) || length(path) != 1)
    stop("path must be a single character string")
  if (!(is.character(sheet) || is.numeric(sheet)) || length(sheet) != 1)
    stop("sheet must be a single character string or integer")

  if (!requireNamespace("openxlsx2", quietly = TRUE))
    stop("Package 'openxlsx2' is required for tbl_xlsx(). ",
         "Install it with: install.packages('openxlsx2')")

  path <- normalizePath(path, mustWork = TRUE)
  df <- openxlsx2::read_xlsx(path, sheet = sheet)
  df_to_node(df)
}

#' Create a lazy table reference from a GeoTIFF raster
#'
#' Opens a GeoTIFF file and returns a lazy query node. Each pixel becomes a row
#' with columns `x`, `y`, `band1`, `band2`, etc. Coordinates are pixel centers
#' derived from the affine geotransform. NoData values become `NA`.
#'
#' Use `filter(x >= ..., y <= ...)` for extent-based cropping and
#' `filter(band1 > ...)` for value-based cropping. Results can be converted
#' back to a raster with `terra::rast(df, type = "xyz")`.
#'
#' @param path Path to a GeoTIFF file.
#' @param batch_size Number of raster rows per batch (default 256).
#'
#' @return A `vectra_node` object representing a lazy scan of the raster.
#'
#' @examples
#' \donttest{
#' f <- tempfile(fileext = ".tif")
#' df <- data.frame(x = as.double(rep(1:4, 3)),
#'                  y = as.double(rep(1:3, each = 4)),
#'                  band1 = as.double(1:12))
#' write_tiff(df, f)
#' node <- tbl_tiff(f)
#' node |> filter(band1 > 6) |> collect()
#' unlink(f)
#' }
#'
#' @export
tbl_tiff <- function(path, batch_size = .TIFF_BATCH_SIZE) {
  if (!is.character(path) || length(path) != 1)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = TRUE)
  xptr <- .Call(C_tiff_scan_node, path, as.double(batch_size))
  meta <- .Call(C_tiff_scan_meta, xptr)
  structure(list(.node = xptr, .path = path, .tiff_meta = meta),
            class = "vectra_node")
}

#' Extract raster values at point coordinates
#'
#' Samples band values from a GeoTIFF at specific (x, y) locations using the
#' file's affine geotransform. Only the strips containing query points are read,
#' making this efficient for sparse point sets on large rasters.
#'
#' Points that fall outside the raster extent return `NA` for all bands.
#' Pixel assignment uses nearest-pixel rounding (i.e., the point is assigned to
#' the pixel whose center is closest).
#'
#' @param path Path to a GeoTIFF file.
#' @param x Numeric vector of x coordinates, or a data.frame / matrix with
#'   columns named `x` and `y`.
#' @param y Numeric vector of y coordinates (ignored if `x` is a data.frame).
#'
#' @return A data.frame with columns `x`, `y`, `band1`, `band2`, etc.
#'   One row per input point, in the same order as the input.
#'
#' @examples
#' \donttest{
#' f <- tempfile(fileext = ".tif")
#' df <- data.frame(x = as.double(rep(1:4, 3)),
#'                  y = as.double(rep(1:3, each = 4)),
#'                  band1 = as.double(1:12))
#' write_tiff(df, f)
#'
#' # Sample at specific locations via data.frame
#' pts <- data.frame(x = c(2, 3), y = c(1, 2))
#' tiff_extract_points(f, pts)
#'
#' # Or pass x and y separately
#' tiff_extract_points(f, x = c(2, 3), y = c(1, 2))
#' unlink(f)
#' }
#'
#' @export
tiff_extract_points <- function(path, x, y = NULL) {
  if (!is.character(path) || length(path) != 1)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = TRUE)

  if (is.data.frame(x) || is.matrix(x)) {
    if (!all(c("x", "y") %in% colnames(x)))
      stop("when x is a data.frame/matrix it must have columns 'x' and 'y'")
    y <- as.double(x[, "y"])
    x <- as.double(x[, "x"])
  } else {
    if (is.null(y))
      stop("y is required when x is a numeric vector")
    x <- as.double(x)
    y <- as.double(y)
  }

  if (length(x) != length(y))
    stop(sprintf("x and y must have the same length (got %d and %d)",
                 length(x), length(y)))

  .Call(C_tiff_extract_points, path, x, y)
}

#' Read GDAL_METADATA from a GeoTIFF
#'
#' Returns the GDAL_METADATA XML string (TIFF tag 42112) embedded in a
#' GeoTIFF file. Returns `NA` if the tag is not present.
#'
#' @param path Path to a GeoTIFF file.
#' @return A single character string containing the XML, or `NA_character_`.
#'
#' @examples
#' \donttest{
#' f <- tempfile(fileext = ".tif")
#' df <- data.frame(x = 1:4, y = rep(1:2, each = 2), band1 = as.double(1:4))
#' write_tiff(df, f, metadata = "<GDALMetadata></GDALMetadata>")
#' tiff_metadata(f)
#' unlink(f)
#' }
#'
#' @export
tiff_metadata <- function(path) {
  if (!is.character(path) || length(path) != 1)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = TRUE)
  .Call(C_tiff_read_metadata, path)
}

#' Read per-band names from a GeoTIFF
#'
#' Returns the band names embedded in the file's GDAL_METADATA XML
#' (TIFF tag 42112). GDAL writes per-band names as
#' `<Item name="DESCRIPTION" sample="N" role="description">...</Item>` entries,
#' where `sample` is the 0-based band index. Bands without a name in the XML
#' are reported as `NA`. Files with no GDAL_METADATA tag at all return a
#' length-`nbands` vector of `NA_character_`.
#'
#' This is a small, dependency-free scanner intended for the common case
#' (`terra::names(r) <- ...` and similar). For arbitrary XML, parse the raw
#' string from [tiff_metadata()] yourself.
#'
#' @param path Path to a GeoTIFF file.
#' @return A character vector of length `nbands`. Element `i` is the name
#'   of band `i` (or `NA_character_` if the file does not name it).
#'
#' @examples
#' \donttest{
#' f <- tempfile(fileext = ".tif")
#' df <- data.frame(x = rep(1:2, 2), y = rep(1:2, each = 2),
#'                  band1 = as.double(1:4), band2 = as.double(5:8))
#' xml <- paste0(
#'   "<GDALMetadata>",
#'   "<Item name=\"DESCRIPTION\" sample=\"0\" role=\"description\">temperature</Item>",
#'   "<Item name=\"DESCRIPTION\" sample=\"1\" role=\"description\">humidity</Item>",
#'   "</GDALMetadata>")
#' write_tiff(df, f, metadata = xml)
#' tiff_band_names(f)
#' unlink(f)
#' }
#'
#' @export
tiff_band_names <- function(path) {
  if (!is.character(path) || length(path) != 1)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = TRUE)

  # We need nbands to size the output. Open via tbl_tiff() to reuse the
  # existing C scan node.
  node <- tbl_tiff(path)
  nb <- node$.tiff_meta$nbands
  out <- rep(NA_character_, nb)

  meta <- tiff_metadata(path)
  if (is.na(meta) || !nzchar(meta)) return(out)

  items <- .parse_gdal_metadata_items(meta)
  if (nrow(items) == 0) return(out)

  desc <- items[items$name == "DESCRIPTION" & !is.na(items$sample), ,
                drop = FALSE]
  if (nrow(desc) == 0) return(out)

  # sample is 0-based; clamp out-of-range entries silently.
  ok <- desc$sample >= 0 & desc$sample < nb
  for (i in which(ok)) out[desc$sample[i] + 1L] <- desc$value[i]
  out
}

# Internal: scan a GDAL_METADATA XML blob for <Item ...>value</Item> entries.
# Returns a data.frame with columns name (chr), sample (int, NA if absent),
# role (chr, NA if absent), value (chr). Built for the well-constrained
# format GDAL emits — not a general XML parser.
.parse_gdal_metadata_items <- function(xml) {
  m <- gregexpr("<Item\\b([^>]*)>(.*?)</Item>", xml, perl = TRUE)[[1]]
  if (length(m) == 1 && m == -1)
    return(data.frame(name = character(0), sample = integer(0),
                      role = character(0), value = character(0),
                      stringsAsFactors = FALSE))

  starts <- as.integer(m)
  lens   <- attr(m, "match.length")
  cap_starts <- attr(m, "capture.start")
  cap_lens   <- attr(m, "capture.length")

  n <- length(starts)
  attrs_chr  <- substring(xml, cap_starts[, 1], cap_starts[, 1] + cap_lens[, 1] - 1L)
  values_chr <- substring(xml, cap_starts[, 2], cap_starts[, 2] + cap_lens[, 2] - 1L)

  get_attr <- function(s, key) {
    pat <- paste0(key, "\\s*=\\s*\"([^\"]*)\"")
    m2  <- regmatches(s, regexec(pat, s, perl = TRUE))[[1]]
    if (length(m2) == 2) m2[2] else NA_character_
  }

  name   <- vapply(attrs_chr, get_attr, character(1), key = "name")
  sample <- vapply(attrs_chr, function(s) {
    v <- get_attr(s, "sample")
    if (is.na(v)) NA_integer_ else suppressWarnings(as.integer(v))
  }, integer(1))
  role   <- vapply(attrs_chr, get_attr, character(1), key = "role")

  data.frame(name   = unname(name),
             sample = unname(sample),
             role   = unname(role),
             value  = unname(.unescape_xml(values_chr)),
             stringsAsFactors = FALSE)
}

.unescape_xml <- function(x) {
  x <- gsub("&lt;",   "<",  x, fixed = TRUE)
  x <- gsub("&gt;",   ">",  x, fixed = TRUE)
  x <- gsub("&quot;", "\"", x, fixed = TRUE)
  x <- gsub("&apos;", "'",  x, fixed = TRUE)
  gsub("&amp;",  "&",  x, fixed = TRUE)
}

#' Read CRS metadata from a GeoTIFF
#'
#' Returns the spatial reference system embedded in a GeoTIFF, parsed from
#' the GeoKey directory (TIFF tag 34735). The projected CRS EPSG
#' (`PCSTypeGeoKey` 3072) is preferred over the geographic CRS EPSG
#' (`GeographicTypeGeoKey` 2048). Citation strings are read from
#' `GeoAsciiParams` (tag 34737) with priority PCS > GeoTIFF > geographic.
#'
#' Files written without a GeoKey directory return `NA` for both fields.
#'
#' @param path Path to a GeoTIFF file.
#' @return A list with elements `epsg` (integer or `NA_integer_`) and
#'   `citation` (character or `NA_character_`).
#'
#' @examples
#' \donttest{
#' f <- tempfile(fileext = ".tif")
#' df <- data.frame(x = 1:4, y = rep(1:2, each = 2), band1 = as.double(1:4))
#' write_tiff(df, f)
#' tiff_crs(f)  # epsg = NA, citation = NA — vectra writer omits GeoKeys
#' unlink(f)
#' }
#'
#' @export
tiff_crs <- function(path) {
  if (!is.character(path) || length(path) != 1)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = TRUE)
  .Call(C_tiff_read_crs, path)
}

## GeoKey writer / reader, tiled-TIFF reader, and GDAL_METADATA band names.
## terra is only used as an external producer/consumer to validate spec
## compliance — tests that need it are skipped when it isn't installed.

skip_if_no_terra <- function() {
  testthat::skip_if_not_installed("terra")
}

test_that("CRS round-trips: integer EPSG (geographic)", {
  df <- data.frame(x = as.double(rep(1:4, 3)),
                   y = as.double(rep(1:3, each = 4)),
                   band1 = as.double(1:12))
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  write_tiff(df, f, crs = 4326L)

  crs <- tiff_crs(f)
  expect_identical(crs$epsg, 4326L)
  expect_true(is.na(crs$citation))
})

test_that("CRS round-trips: 'EPSG:xxxx' string (projected)", {
  df <- data.frame(x = as.double(rep(1:4, 3)),
                   y = as.double(rep(1:3, each = 4)),
                   band1 = as.double(1:12))
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  write_tiff(df, f, crs = "EPSG:3857")

  expect_identical(tiff_crs(f)$epsg, 3857L)
})

test_that("CRS round-trips: list with citation", {
  df <- data.frame(x = as.double(rep(1:4, 3)),
                   y = as.double(rep(1:3, each = 4)),
                   band1 = as.double(1:12))
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  write_tiff(df, f, crs = list(epsg = 31287L, citation = "MGI / Austria Lambert"))

  crs <- tiff_crs(f)
  expect_identical(crs$epsg, 31287L)
  expect_identical(crs$citation, "MGI / Austria Lambert")
})

test_that("no crs argument writes no GeoKey directory", {
  df <- data.frame(x = as.double(rep(1:4, 3)),
                   y = as.double(rep(1:3, each = 4)),
                   band1 = as.double(1:12))
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  write_tiff(df, f)

  crs <- tiff_crs(f)
  expect_true(is.na(crs$epsg))
  expect_true(is.na(crs$citation))
})

test_that("write_tiff(crs = ...) is readable by terra", {
  skip_if_no_terra()
  df <- data.frame(x = as.double(rep(1:4, 3)),
                   y = as.double(rep(1:3, each = 4)),
                   band1 = as.double(1:12))

  for (epsg in c(4326L, 3857L, 31287L)) {
    f <- tempfile(fileext = ".tif")
    write_tiff(df, f, crs = epsg)
    info <- terra::crs(terra::rast(f), describe = TRUE)
    expect_equal(as.integer(info$code), epsg,
                 info = sprintf("epsg=%d", epsg))
    unlink(f)
  }
})

test_that("invalid crs argument is rejected", {
  df <- data.frame(x = c(0.5, 1.5), y = c(0.5, 0.5), band1 = c(1.0, 2.0))
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  expect_error(write_tiff(df, f, crs = "not an EPSG"),
               "crs string must look like")
  expect_error(write_tiff(df, f, crs = 0L),
               "needs a positive integer EPSG")
})

test_that("tiled (uncompressed) GeoTIFF reads correctly", {
  skip_if_no_terra()
  nrows <- 64; ncols <- 64
  r <- terra::rast(nrows = nrows, ncols = ncols,
                   xmin = 0, xmax = ncols, ymin = 0, ymax = nrows,
                   crs = "EPSG:4326")
  terra::values(r) <- as.double(seq_len(nrows * ncols))

  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  terra::writeRaster(r, f, overwrite = TRUE,
                     gdal = c("TILED=YES", "BLOCKXSIZE=16", "BLOCKYSIZE=16",
                              "COMPRESS=NONE"))

  node <- tbl_tiff(f)
  expect_equal(node$.tiff_meta$width, ncols)
  expect_equal(node$.tiff_meta$height, nrows)
  expect_equal(node$.tiff_meta$epsg, 4326L)

  df <- collect(node)
  expect_equal(nrow(df), nrows * ncols)
  expect_equal(sort(df$band1), sort(terra::values(r)[, 1]))
})

test_that("tiled DEFLATE GeoTIFF with edge padding reads correctly", {
  skip_if_no_terra()
  # 200x200 image with 64x64 tiles → 4x4 grid, last col/row tiles padded.
  nrows <- 200; ncols <- 200
  r <- terra::rast(nrows = nrows, ncols = ncols,
                   xmin = 0, xmax = ncols, ymin = 0, ymax = nrows,
                   crs = "EPSG:3857")
  terra::values(r) <- as.double(seq_len(nrows * ncols))

  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  terra::writeRaster(r, f, overwrite = TRUE,
                     gdal = c("TILED=YES", "BLOCKXSIZE=64", "BLOCKYSIZE=64",
                              "COMPRESS=DEFLATE"))

  df <- collect(tbl_tiff(f))
  expect_equal(nrow(df), nrows * ncols)
  expect_equal(sort(df$band1), sort(terra::values(r)[, 1]))
})

test_that("point extraction works on tiled TIFF", {
  skip_if_no_terra()
  nrows <- 64; ncols <- 64
  r <- terra::rast(nrows = nrows, ncols = ncols,
                   xmin = 0, xmax = ncols, ymin = 0, ymax = nrows,
                   crs = "EPSG:4326")
  terra::values(r) <- as.double(seq_len(nrows * ncols))

  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  terra::writeRaster(r, f, overwrite = TRUE,
                     gdal = c("TILED=YES", "BLOCKXSIZE=16", "BLOCKYSIZE=16",
                              "COMPRESS=NONE"))

  pts <- data.frame(x = c(0.5, 7.5, 31.5, 63.5),
                    y = c(63.5, 31.5, 7.5, 0.5))
  got <- tiff_extract_points(f, pts)
  expected <- terra::extract(r, pts)
  expect_equal(got$band1, expected[[2]])
})

test_that("tiff_band_names: explicit DESCRIPTION items round-trip", {
  df <- data.frame(x = rep(1:2, 2), y = rep(1:2, each = 2),
                   band1 = as.double(1:4), band2 = as.double(5:8))
  xml <- paste0(
    "<GDALMetadata>",
    "<Item name=\"DESCRIPTION\" sample=\"0\" role=\"description\">temperature</Item>",
    "<Item name=\"DESCRIPTION\" sample=\"1\" role=\"description\">humidity</Item>",
    "<Item name=\"scale\" sample=\"0\">0.01</Item>",
    "</GDALMetadata>")
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  write_tiff(df, f, metadata = xml)

  expect_identical(tiff_band_names(f), c("temperature", "humidity"))
})

test_that("tiff_band_names returns NA per band when no metadata", {
  df <- data.frame(x = rep(1:2, 2), y = rep(1:2, each = 2),
                   band1 = as.double(1:4), band2 = as.double(5:8))
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  write_tiff(df, f)

  expect_identical(tiff_band_names(f), rep(NA_character_, 2))
})

test_that("tiff_band_names reads names written by terra", {
  skip_if_no_terra()
  r <- terra::rast(nrows = 4, ncols = 4, nlyrs = 3,
                   xmin = 0, xmax = 4, ymin = 0, ymax = 4,
                   crs = "EPSG:4326")
  terra::values(r) <- as.double(rep(1:16, 3))
  names(r) <- c("alpha", "beta", "gamma")

  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  terra::writeRaster(r, f, overwrite = TRUE, gdal = "COMPRESS=NONE")

  expect_identical(tiff_band_names(f), c("alpha", "beta", "gamma"))
})

test_that("tiff_band_names decodes XML entity escapes", {
  df <- data.frame(x = c(0.5, 1.5), y = c(0.5, 0.5), band1 = c(1.0, 2.0))
  xml <- paste0(
    "<GDALMetadata>",
    "<Item name=\"DESCRIPTION\" sample=\"0\">a &amp; b &lt;c&gt;</Item>",
    "</GDALMetadata>")
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))
  write_tiff(df, f, metadata = xml)

  expect_identical(tiff_band_names(f), "a & b <c>")
})

# ---------------------------------------------------------------------------
# Phase 4c: tiled TIFF write
# ---------------------------------------------------------------------------

# Helper: build a regular grid as an x/y/band data.frame.
.make_grid_df <- function(ncols, nrows, fn = function(i, j) i + (j - 1) * ncols) {
  data.frame(
    x = rep(seq(0.5, ncols - 0.5, by = 1), nrows),
    y = rep(seq(nrows - 0.5, 0.5, by = -1), each = ncols),
    band1 = as.double(fn(rep(seq_len(ncols), nrows),
                         rep(seq_len(nrows), each = ncols)))
  )
}

test_that("tiled write: round-trip via tbl_tiff preserves values", {
  ncols <- 32; nrows <- 32
  df <- .make_grid_df(ncols, nrows)
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, tiled = TRUE, tile_size = 16L)
  result <- collect(tbl_tiff(f))

  expect_equal(nrow(result), ncols * nrows)
  expect_equal(sort(result$band1), sort(df$band1))
})

test_that("tiled write: emits TileWidth/TileLength matching tile_size", {
  ncols <- 32; nrows <- 48
  df <- .make_grid_df(ncols, nrows)
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, tiled = TRUE, tile_size = 16L)
  meta <- tbl_tiff(f)$.tiff_meta
  expect_equal(meta$width, ncols)
  expect_equal(meta$height, nrows)
})

test_that("tiled write with edge padding (image not multiple of tile)", {
  # 40x40 image with 16x16 tiles → 3x3 = 9 tiles; right/bottom edges padded.
  ncols <- 40; nrows <- 40
  df <- .make_grid_df(ncols, nrows)
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, tiled = TRUE, tile_size = 16L)
  result <- collect(tbl_tiff(f))

  # Reader returns only in-image pixels.
  expect_equal(nrow(result), ncols * nrows)
  expect_equal(sort(result$band1), sort(df$band1))
})

test_that("tiled DEFLATE write round-trips correctly", {
  ncols <- 64; nrows <- 64
  df <- .make_grid_df(ncols, nrows)
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, compress = TRUE, tiled = TRUE, tile_size = 32L)
  result <- collect(tbl_tiff(f))

  expect_equal(nrow(result), ncols * nrows)
  expect_equal(sort(result$band1), sort(df$band1))
})

test_that("tiled multi-band write round-trips", {
  ncols <- 32; nrows <- 32
  base <- .make_grid_df(ncols, nrows)
  # Reader assigns band1/band2 names regardless of writer-side column names.
  df <- data.frame(x = base$x, y = base$y,
                   b1 = base$band1, b2 = base$band1 * 2)
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, tiled = TRUE, tile_size = 16L)
  result <- collect(tbl_tiff(f))

  expect_equal(nrow(result), ncols * nrows)
  expect_equal(sort(result$band1), sort(df$b1))
  expect_equal(sort(result$band2), sort(df$b2))
})

test_that("tiled write rejects non-multiple-of-16 tile_size", {
  df <- data.frame(x = c(0.5, 1.5), y = c(0.5, 0.5), band1 = c(1.0, 2.0))
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  expect_error(write_tiff(df, f, tiled = TRUE, tile_size = 17L),
               "multiple of 16")
  expect_error(write_tiff(df, f, tiled = TRUE, tile_size = 0L),
               "positive")
})

test_that("tiled write: tile count = ceil(W/TW) * ceil(H/TH) (terra)", {
  skip_if_no_terra()
  ncols <- 50; nrows <- 30
  TW <- 16L; TH <- 16L
  df <- .make_grid_df(ncols, nrows)
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, tiled = TRUE, tile_size = c(TW, TH))

  # Round-trip through terra: output values should match input.
  r <- terra::rast(f)
  expect_equal(terra::nrow(r), nrows)
  expect_equal(terra::ncol(r), ncols)

  vals <- terra::values(r)[, 1]
  # terra fills nodata-padded edges if the writer didn't pad properly. We
  # expect a 50x30 raster with ncols*nrows valid cells, no NA.
  expect_equal(length(vals), ncols * nrows)
  expect_equal(sort(vals[!is.na(vals)]), sort(df$band1))

  # Confirm tile structure: the file's tile count must be 4 * 2 = 8.
  expected_n_tiles <- ceiling(ncols / TW) * ceiling(nrows / TH)
  expect_equal(expected_n_tiles, 4 * 2)
})

test_that("tiled write: terra round-trip with CRS and DEFLATE", {
  skip_if_no_terra()
  ncols <- 64; nrows <- 64
  df <- .make_grid_df(ncols, nrows)
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, compress = TRUE, tiled = TRUE, tile_size = 32L,
             crs = 4326L)

  r <- terra::rast(f)
  expect_equal(terra::ncol(r), ncols)
  expect_equal(terra::nrow(r), nrows)
  info <- terra::crs(r, describe = TRUE)
  expect_equal(as.integer(info$code), 4326L)
  vals <- terra::values(r)[, 1]
  expect_equal(sort(vals), sort(df$band1))
})

test_that("tiled write: rectangular tile_size c(w, h)", {
  ncols <- 32; nrows <- 48
  df <- .make_grid_df(ncols, nrows)
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, tiled = TRUE, tile_size = c(16L, 32L))
  result <- collect(tbl_tiff(f))

  expect_equal(nrow(result), ncols * nrows)
  expect_equal(sort(result$band1), sort(df$band1))
})

test_that("int16 round-trip preserves values", {
  df <- data.frame(
    x = rep(seq(0.5, 4.5, by = 1), 3),
    y = rep(seq(2.5, 0.5, by = -1), each = 5),
    band1 = as.double(1:15)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, compress = TRUE, pixel_type = "int16")
  result <- tbl_tiff(f) |> collect()

  expect_equal(nrow(result), 15)
  expect_equal(sort(result$band1), 1:15)
})

test_that("uint8 round-trip preserves 0-255 values", {
  df <- data.frame(
    x = rep(seq(0.5, 4.5, by = 1), 2),
    y = rep(c(1.5, 0.5), each = 5),
    band1 = as.double(c(0, 50, 100, 200, 255, 10, 20, 30, 40, 128))
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, pixel_type = "uint8")
  result <- tbl_tiff(f) |> collect()

  expect_equal(nrow(result), 10)
  # all values except nodata (255) should be present
  expect_true(all(c(0, 10, 20, 30, 40, 50, 100, 128, 200) %in% result$band1))
})

test_that("uint16 round-trip preserves values", {
  df <- data.frame(
    x = rep(seq(0.5, 2.5, by = 1), 2),
    y = rep(c(1.5, 0.5), each = 3),
    band1 = as.double(c(100, 1000, 30000, 200, 500, 60000))
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, pixel_type = "uint16")
  result <- tbl_tiff(f) |> collect()

  expect_equal(nrow(result), 6)
  expect_true(all(c(100, 1000, 30000, 200, 500) %in% result$band1))
})

test_that("float32 round-trip preserves approximate values", {
  df <- data.frame(
    x = rep(seq(0.5, 2.5, by = 1), 2),
    y = rep(c(1.5, 0.5), each = 3),
    band1 = c(1.5, 2.7, 3.14, -1.0, 0.0, 100.5)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, pixel_type = "float32")
  result <- tbl_tiff(f) |> collect()

  expect_equal(nrow(result), 6)
  expect_equal(sort(result$band1), sort(df$band1), tolerance = 1e-5)
})

test_that("int16 with NaN writes nodata and reads as NA", {
  df <- data.frame(
    x = rep(seq(0.5, 2.5, by = 1), 2),
    y = rep(c(1.5, 0.5), each = 3),
    band1 = c(10, NA, 30, 40, 50, NA)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, pixel_type = "int16")
  result <- tbl_tiff(f) |> collect()

  expect_equal(nrow(result), 6)
  # nodata pixels read as NA
  valid <- result[!is.na(result$band1), ]
  expect_true(all(c(10, 30, 40, 50) %in% valid$band1))
})

test_that("GDAL_METADATA write and read round-trip", {
  df <- data.frame(
    x = rep(seq(0.5, 2.5, by = 1), 2),
    y = rep(c(1.5, 0.5), each = 3),
    band1 = as.double(1:6)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  xml <- '<GDALMetadata>\n  <Item name="scale">0.01</Item>\n  <Item name="offset">-273.15</Item>\n</GDALMetadata>'
  write_tiff(df, f, pixel_type = "int16", metadata = xml)

  meta <- tiff_metadata(f)
  expect_false(is.na(meta))
  expect_true(grepl("scale", meta))
  expect_true(grepl("0.01", meta))
  expect_true(grepl("offset", meta))
  expect_true(grepl("-273.15", meta))
})

test_that("tiff_metadata returns NA for files without metadata", {
  df <- data.frame(
    x = c(0.5, 1.5), y = c(0.5, 0.5),
    band1 = c(1.0, 2.0)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f)
  meta <- tiff_metadata(f)
  expect_true(is.na(meta))
})

test_that("int16 point extraction returns correct values", {
  df <- data.frame(
    x = rep(seq(0.5, 4.5, by = 1), 3),
    y = rep(seq(2.5, 0.5, by = -1), each = 5),
    band1 = as.double(1:15)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, compress = TRUE, pixel_type = "int16")

  pts <- tiff_extract_points(f, x = c(0.5, 2.5, 4.5), y = c(2.5, 1.5, 0.5))
  expect_equal(nrow(pts), 3)
  expect_equal(pts$band1, c(1, 8, 15))
})

test_that("int32 round-trip with large values", {
  df <- data.frame(
    x = c(0.5, 1.5), y = c(0.5, 0.5),
    band1 = c(100000.0, -100000.0)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, pixel_type = "int32")
  result <- tbl_tiff(f) |> collect()
  expect_equal(sort(result$band1), c(-100000, 100000))
})

test_that("pixel_type validation rejects invalid types", {
  df <- data.frame(x = 0.5, y = 0.5, band1 = 1.0)
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  expect_error(write_tiff(df, f, pixel_type = "int8"),
               "pixel_type must be one of")
})

## BigTIFF write path (Phase 4d).
##
## We never actually generate a >4 GB file in tests — instead we
## force-enable BigTIFF on small data and verify:
##   1. The on-disk header bytes start with the BigTIFF magic 0x002B and
##      the 8-byte offset width.
##   2. Our own reader (which already supports BigTIFF) reproduces the
##      pixel values.
##   3. terra::rast() accepts the file (terra reads BigTIFF natively).
##   4. Auto mode emits classic TIFF for tiny rasters.
##   5. The auto-promotion threshold trips when the expected raw payload
##      would exceed the 4 GB ceiling.

skip_if_no_terra <- function() {
  testthat::skip_if_not_installed("terra")
}

# Read the first few header bytes and decode magic / offset width.
.read_tiff_header <- function(path) {
  con <- file(path, "rb")
  on.exit(close(con), add = TRUE)
  bom <- readBin(con, "integer", n = 2, size = 1, signed = FALSE)
  if (!(bom[1] == as.integer(charToRaw("I")[1]) &&
        bom[2] == as.integer(charToRaw("I")[1])))
    stop("not a little-endian TIFF (got byte order ", bom[1], " ", bom[2], ")")
  magic <- readBin(con, "integer", n = 1, size = 2, signed = FALSE,
                   endian = "little")
  out <- list(magic = magic)
  if (magic == 43L) {
    out$offset_size <- readBin(con, "integer", n = 1, size = 2, signed = FALSE,
                                endian = "little")
    out$reserved <- readBin(con, "integer", n = 1, size = 2, signed = FALSE,
                              endian = "little")
  }
  out
}

test_that("bigtiff = TRUE writes BigTIFF magic 0x002B", {
  df <- data.frame(
    x = rep(seq(0.5, 3.5, by = 1), 3),
    y = rep(seq(2.5, 0.5, by = -1), each = 4),
    band1 = as.double(1:12)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, bigtiff = TRUE)

  hdr <- .read_tiff_header(f)
  expect_identical(hdr$magic, 43L)         # 0x002B
  expect_identical(hdr$offset_size, 8L)
  expect_identical(hdr$reserved, 0L)
})

test_that("bigtiff = TRUE round-trips through tbl_tiff()", {
  df <- data.frame(
    x = rep(seq(0.5, 3.5, by = 1), 3),
    y = rep(seq(2.5, 0.5, by = -1), each = 4),
    band1 = as.double(1:12)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, bigtiff = TRUE)
  result <- tbl_tiff(f) |> collect()

  expect_equal(nrow(result), 12)
  expect_equal(sort(result$band1), 1:12)
})

test_that("bigtiff = TRUE round-trips through terra", {
  skip_if_no_terra()
  df <- data.frame(
    x = rep(seq(0.5, 3.5, by = 1), 3),
    y = rep(seq(2.5, 0.5, by = -1), each = 4),
    band1 = as.double(1:12)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, bigtiff = TRUE, crs = 4326L)
  r <- terra::rast(f)
  expect_equal(terra::nrow(r), 3)
  expect_equal(terra::ncol(r), 4)
  expect_equal(sort(terra::values(r)[, 1]), 1:12)
})

test_that("bigtiff = TRUE works with multi-band, deflate, nodata, CRS", {
  skip_if_no_terra()
  df <- data.frame(
    x = rep(seq(0.5, 3.5, by = 1), 3),
    y = rep(seq(2.5, 0.5, by = -1), each = 4),
    band1 = as.double(1:12),
    band2 = c(as.double(10:20), NA_real_)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, compress = TRUE, pixel_type = "int16",
             crs = list(epsg = 31287L, citation = "MGI / Austria Lambert"),
             bigtiff = TRUE)

  hdr <- .read_tiff_header(f)
  expect_identical(hdr$magic, 43L)

  r <- terra::rast(f)
  expect_equal(terra::nlyr(r), 2)
  vals <- terra::values(r)
  expect_equal(sort(vals[, 1]), 1:12)
  # band2 has one NA-as-nodata; everything else round-trips
  expect_setequal(setdiff(vals[, 2], NA), 10:20)

  expect_identical(tiff_crs(f)$epsg, 31287L)
})

test_that("bigtiff = FALSE produces classic TIFF (magic 42)", {
  df <- data.frame(
    x = rep(seq(0.5, 3.5, by = 1), 3),
    y = rep(seq(2.5, 0.5, by = -1), each = 4),
    band1 = as.double(1:12)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, bigtiff = FALSE)

  hdr <- .read_tiff_header(f)
  expect_identical(hdr$magic, 42L)
})

test_that("bigtiff = \"auto\" stays classic for small rasters", {
  df <- data.frame(
    x = rep(seq(0.5, 3.5, by = 1), 3),
    y = rep(seq(2.5, 0.5, by = -1), each = 4),
    band1 = as.double(1:12)
  )
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  write_tiff(df, f, bigtiff = "auto")
  expect_identical(.read_tiff_header(f)$magic, 42L)

  # Default for the writer is "auto"; same expectation.
  f2 <- tempfile(fileext = ".tif")
  on.exit(unlink(f2), add = TRUE)
  write_tiff(df, f2)
  expect_identical(.read_tiff_header(f2)$magic, 42L)
})

test_that("invalid bigtiff argument is rejected", {
  df <- data.frame(x = c(0.5, 1.5), y = c(0.5, 0.5), band1 = c(1.0, 2.0))
  f <- tempfile(fileext = ".tif")
  on.exit(unlink(f))

  expect_error(write_tiff(df, f, bigtiff = "yes"), "bigtiff string")
  expect_error(write_tiff(df, f, bigtiff = NA), "bigtiff must be")
  expect_error(write_tiff(df, f, bigtiff = 1), "bigtiff must be")
})

test_that("write_vtr streams VTR -> VTR", {
  f <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, f2)))
  write_vtr(mtcars, f)
  tbl(f) |> write_vtr(f2)
  result <- tbl(f2) |> collect()
  expect_equal(nrow(result), 32L)
  expect_equal(ncol(result), 11L)
  expect_equal(result$mpg, mtcars$mpg)
})

test_that("write_vtr streams CSV -> VTR", {
  csv <- tempfile(fileext = ".csv")
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(csv, f)))
  write.csv(mtcars, csv, row.names = FALSE)
  tbl_csv(csv) |> write_vtr(f)
  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 32L)
  expect_equal(result$mpg, mtcars$mpg)
})

test_that("write_vtr streams SQLite -> VTR", {
  f <- tempfile(fileext = ".vtr")
  db <- tempfile(fileext = ".sqlite")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, db, f2)))
  write_vtr(mtcars, f)
  tbl(f) |> write_sqlite(db, "cars")
  tbl_sqlite(db, "cars") |> write_vtr(f2)
  result <- tbl(f2) |> collect()
  expect_equal(nrow(result), 32L)
})

test_that("write_vtr streams filtered node", {
  f <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, f2)))
  write_vtr(mtcars, f)
  tbl(f) |> filter(mpg > 20) |> write_vtr(f2)
  result <- tbl(f2) |> collect()
  expect_equal(nrow(result), sum(mtcars$mpg > 20))
})

test_that("write_vtr streams mutated node", {
  f <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, f2)))
  write_vtr(mtcars, f)
  tbl(f) |> mutate(kpl = mpg * 0.425144) |> write_vtr(f2)
  result <- tbl(f2) |> collect()
  expect_equal(ncol(result), 12L)
  expect_true("kpl" %in% names(result))
})

test_that("write_vtr atomic write leaves no temp file", {
  f <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f, f2)))
  write_vtr(mtcars, f)
  tbl(f) |> write_vtr(f2)
  expect_false(file.exists(paste0(f2, ".~writing")))
  expect_true(file.exists(f2))
})

test_that("write_vtr.data.frame still works", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(iris, f)
  result <- tbl(f) |> collect()
  expect_equal(nrow(result), 150L)
  expect_equal(ncol(result), 5L)
})

#' Summarise with variable-length output per group
#'
#' Like [summarise()] but allows expressions that return more than one row
#' per group. Currently implemented via `collect()` fallback.
#'
#' @param .data A `vectra_node` object.
#' @param ... Named expressions.
#'
#' @return A data.frame (not a lazy node).
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(g = c("a", "a", "b"), x = c(1, 2, 3)), f)
#' tbl(f) |> group_by(g) |> reframe(range_x = range(x))
#' unlink(f)
#'
#' @export
reframe <- function(.data, ...) {
  UseMethod("reframe")
}

#' @export
reframe.vectra_node <- function(.data, ...) {
  df <- collect(.data)
  groups <- .data$.groups
  dots <- eval(substitute(alist(...)))
  # Expand across() calls
  dot_names_check <- names(dots)
  has_across <- any(vapply(dots, function(e) is.call(e) && identical(e[[1]], as.name("across")), logical(1)))
  if (has_across) {
    schema_names <- names(df)
    proxy <- df[0, , drop = FALSE]
    dots <- expand_across(dots, schema_names, parent.frame(), proxy)
  }
  dot_names <- names(dots)

  if (is.null(groups) || length(groups) == 0) {
    # Ungrouped: evaluate expressions on whole data
    env <- list2env(df, parent = parent.frame())
    results <- lapply(seq_along(dots), function(i) {
      eval(dots[[i]], envir = env)
    })
    names(results) <- dot_names
    as.data.frame(results, stringsAsFactors = FALSE)
  } else {
    # Grouped: split-apply-combine
    split_idx <- interaction(df[groups], drop = TRUE)
    pieces <- split(df, split_idx, drop = TRUE)

    result_list <- lapply(pieces, function(chunk) {
      env <- list2env(chunk, parent = parent.frame())
      results <- lapply(seq_along(dots), function(i) {
        eval(dots[[i]], envir = env)
      })
      names(results) <- dot_names
      n <- max(vapply(results, length, integer(1)))
      # Key columns
      key_df <- chunk[rep(1, n), groups, drop = FALSE]
      val_df <- as.data.frame(results, stringsAsFactors = FALSE)
      cbind(key_df, val_df)
    })

    result <- do.call(rbind, result_list)
    rownames(result) <- NULL
    result
  }
}

# Topology verbs that build new geometry from a whole set of features in scope:
# polygonize (lines -> faces), line_merge (segments -> maximal lines), and
# coverage-preserving simplify, all on the partition tier shared with dissolve
# and construct; plus linear referencing (locate points along a resident line
# layer), a resident-y streamed verb in the spatial_knn / spatial_split family.

# -- polygonize (build polygonal faces from a line network) -------------------

# One group's worth of lines -> the polygonal faces they enclose. The lines are
# unioned and noded so every crossing becomes a shared vertex, then GEOS forms
# the faces of the resulting arrangement; faces carry only the group's `by`
# values, as the faces are new geometry with no single source feature.
.polygonize_fn <- function(by, geom, crs) function(df) {
  sb    <- .sf_decode_chunk(df, geom, NULL, crs)
  noded <- sf::st_node(sf::st_union(sf::st_geometry(sb)))
  faces <- sf::st_collection_extract(sf::st_polygonize(noded), "POLYGON")
  faces <- faces[!sf::st_is_empty(faces)]
  if (!length(faces)) return(NULL)
  rowdf <- if (is.null(by)) data.frame(matrix(nrow = length(faces), ncol = 0))
           else df[rep(1L, length(faces)), by, drop = FALSE]
  rowdf[[geom]] <- sf::st_as_binary(faces, hex = TRUE)
  rownames(rowdf) <- NULL
  rowdf
}

#' Build polygonal faces from a line network
#'
#' Forms the polygons enclosed by a set of lines (the QGIS "Polygonize", GEOS
#' `Polygonize`): the inverse of taking polygon boundaries. The lines of each
#' group are unioned and noded so every crossing becomes a shared vertex, then
#' the faces of that planar arrangement are returned, one per row. A pile of
#' lines that does not close any area yields no faces. Like [spatial_dissolve()]
#' and [spatial_construct()] it rides the **partition tier**: `x` is spilled once
#' and routed into one disjoint shard per `by` group in a single bounded pass,
#' then each group's lines are polygonized together. Peak memory is the routing
#' budget during the pass, then one group's geometry while its faces are built --
#' partition on a key whose groups fit in memory. With no `by`, the whole layer
#' yields one set of faces.
#'
#' Each face is new geometry built from the whole group, so it carries the `by`
#' columns only, not the attributes of any single source line. Geometry travels
#' through the engine as hex-encoded WKB in a string column and the CRS is
#' carried on the returned node; the noding is \pkg{sf}/GEOS and expects
#' projected or unprojected planar data. The \pkg{sf} package is an optional
#' dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param by Character vector of attribute columns to polygonize within: one set
#'   of faces per distinct combination of their values. `NULL` (default)
#'   polygonizes the whole layer at once.
#'
#' @return A `vectra_node` of one row per face, carrying the `by` columns and the
#'   input CRS, backed by temporary `.vtr` spills removed when the node is
#'   garbage-collected.
#'
#' @seealso [spatial_split()] to cut existing polygons by a blade,
#'   [spatial_construct()] for hulls and tessellations, [spatial_dissolve()] to
#'   merge geometries by group, [collect_sf()] to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' grid <- sf::st_sfc(
#'   sf::st_linestring(rbind(c(0, 0), c(2, 0))),
#'   sf::st_linestring(rbind(c(0, 1), c(2, 1))),
#'   sf::st_linestring(rbind(c(0, 2), c(2, 2))),
#'   sf::st_linestring(rbind(c(0, 0), c(0, 2))),
#'   sf::st_linestring(rbind(c(1, 0), c(1, 2))),
#'   sf::st_linestring(rbind(c(2, 0), c(2, 2))))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   geometry = sf::st_as_binary(grid, hex = TRUE)
#' ), f)
#'
#' # The four unit cells enclosed by the grid of lines.
#' tbl(f) |> spatial_polygonize() |> collect_sf()
#' unlink(f)
#'
#' @export
spatial_polygonize <- function(x, by = NULL, geom = "geometry", crs = NA,
                               flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed line layer to polygonize)")
  crs <- .resolve_crs(x, crs)
  .partition_each(x, by, geom, crs, .polygonize_fn(by, geom, crs), flush_rows)
}

# -- line merge (sew contiguous segments into maximal lines) ------------------

# One group's worth of lines -> the maximal linestrings formed by sewing
# segments that meet end to end. The lines are unioned (which nodes them at
# crossings) and merged; the merged result is exploded so each maximal chain is
# one row, carrying the group's `by` values.
.line_merge_fn <- function(by, geom, crs) function(df) {
  sb   <- .sf_decode_chunk(df, geom, NULL, crs)
  m    <- sf::st_line_merge(sf::st_union(sf::st_geometry(sb)))
  segs <- sf::st_cast(m, "LINESTRING", warn = FALSE)
  segs <- segs[!sf::st_is_empty(segs)]
  if (!length(segs)) return(NULL)
  rowdf <- if (is.null(by)) data.frame(matrix(nrow = length(segs), ncol = 0))
           else df[rep(1L, length(segs)), by, drop = FALSE]
  rowdf[[geom]] <- sf::st_as_binary(segs, hex = TRUE)
  rownames(rowdf) <- NULL
  rowdf
}

#' Merge contiguous line segments into maximal lines
#'
#' Sews the line segments of each group into the longest possible linestrings
#' (`sf::st_line_merge`, the line counterpart of a dissolve): segments that meet
#' end to end become one chain, and each chain is emitted as its own row. Where a
#' plain union of lines returns a single multilinestring of all the parts, this
#' joins the parts through their shared endpoints; at a crossing where more than
#' two segments meet the merge is ambiguous and the segments stay separate. Like
#' [spatial_dissolve()] it rides the **partition tier**: `x` is spilled once and
#' routed into one disjoint shard per `by` group in a single bounded pass, then
#' each group's segments are merged together. Peak memory is the routing budget
#' during the pass, then one group's geometry while it is merged. With no `by`,
#' the whole layer is merged at once.
#'
#' Each merged line is new geometry built from the whole group, so it carries the
#' `by` columns only, not the attributes of any single source segment. Geometry
#' travels through the engine as hex-encoded WKB in a string column and the CRS
#' is carried on the returned node. The \pkg{sf} package is an optional
#' dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param by Character vector of attribute columns to merge within: one set of
#'   maximal lines per distinct combination of their values. `NULL` (default)
#'   merges the whole layer at once.
#'
#' @return A `vectra_node` of one row per maximal merged line, carrying the `by`
#'   columns and the input CRS, backed by temporary `.vtr` spills removed when
#'   the node is garbage-collected.
#'
#' @seealso [spatial_dissolve()] to union geometries by group,
#'   [spatial_explode()] for the opposite direction (multipart to single part),
#'   [collect_sf()] to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' seg <- sf::st_sfc(
#'   sf::st_linestring(rbind(c(0, 0), c(1, 0))),
#'   sf::st_linestring(rbind(c(1, 0), c(2, 0))),
#'   sf::st_linestring(rbind(c(2, 0), c(3, 0))))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   geometry = sf::st_as_binary(seg, hex = TRUE)
#' ), f)
#'
#' # The three end-to-end segments become one line.
#' tbl(f) |> spatial_line_merge() |> collect_sf()
#' unlink(f)
#'
#' @export
spatial_line_merge <- function(x, by = NULL, geom = "geometry", crs = NA,
                               flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed line layer to merge)")
  crs <- .resolve_crs(x, crs)
  .partition_each(x, by, geom, crs, .line_merge_fn(by, geom, crs), flush_rows)
}

# -- coverage-preserving simplification ---------------------------------------

# One group's worth of polygons -> the same polygons with simplified boundaries,
# shared edges kept coincident. The boundaries are unioned (so a border shared by
# two polygons is one line) and noded into arcs at every junction, each arc is
# simplified once with topology-preserving Douglas-Peucker (its junction
# endpoints pinned), and the simplified arcs are re-polygonized into faces. Each
# face is matched back to the source polygon containing its interior point, so it
# inherits that feature's full attribute row -- a face that falls in no source
# polygon (a hole between shapes) is dropped.
.simplify_fn <- function(geom, crs, tolerance) function(df) {
  sb    <- .sf_decode_chunk(df, geom, NULL, crs)
  g     <- sf::st_geometry(sb)
  bdry  <- sf::st_node(sf::st_union(sf::st_boundary(g)))
  bsimp <- sf::st_simplify(bdry, dTolerance = tolerance, preserveTopology = TRUE)
  faces <- sf::st_collection_extract(
             sf::st_polygonize(sf::st_union(bsimp)), "POLYGON")
  faces <- faces[!sf::st_is_empty(faces)]
  if (!length(faces)) return(NULL)
  ip  <- suppressWarnings(sf::st_point_on_surface(faces))
  hit <- sf::st_within(ip, g)
  idx <- vapply(hit, function(h) if (length(h)) h[[1L]] else NA_integer_,
                integer(1))
  keep <- !is.na(idx)
  if (!any(keep)) return(NULL)
  faces <- faces[keep]
  idx   <- idx[keep]
  rowdf <- df[idx, setdiff(names(df), geom), drop = FALSE]
  rowdf[[geom]] <- sf::st_as_binary(faces, hex = TRUE)
  rownames(rowdf) <- NULL
  rowdf
}

#' Simplify a polygon coverage without tearing shared edges
#'
#' Simplifies polygon boundaries while keeping a shared border between two
#' polygons identical on both sides, so adjacent polygons stay edge-matched with
#' no slivers or gaps -- the topology-preserving simplification that a per-feature
#' `spatial_map(~ sf::st_simplify(.x))` cannot give, because it simplifies each
#' polygon's copy of a shared border independently. The boundaries of each group
#' are unioned so a shared border is one line, noded into arcs at every junction,
#' each arc simplified once (its junction endpoints pinned), and the arcs
#' re-polygonized; each resulting face inherits the attributes of the source
#' polygon containing it. Like [spatial_dissolve()] it rides the **partition
#' tier**: `x` is spilled once and routed into one disjoint shard per `by` group
#' in a single bounded pass, and each group is simplified as an independent
#' coverage. Peak memory is the routing budget during the pass, then one group's
#' geometry while it is simplified -- partition on a key whose groups fit in
#' memory. With no `by`, the whole layer is one coverage.
#'
#' The simplification is topology-preserving Douglas-Peucker (`dTolerance =
#' tolerance`) on the noded boundary arcs. Geometry travels through the engine as
#' hex-encoded WKB in a string column and the CRS is carried on the returned
#' node; the noding is \pkg{sf}/GEOS and expects projected or unprojected planar
#' data. The \pkg{sf} package is an optional dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param tolerance Distance tolerance for the boundary simplification, in CRS
#'   units: vertices that deviate less than this from the simplified line are
#'   dropped. Larger values simplify more.
#' @param by Character vector of attribute columns whose groups are each
#'   simplified as an independent coverage. `NULL` (default) treats the whole
#'   layer as one coverage.
#'
#' @return A `vectra_node` of the simplified polygons, each carrying its source
#'   feature's attributes and the input CRS, backed by temporary `.vtr` spills
#'   removed when the node is garbage-collected.
#'
#' @seealso [spatial_map()] with `~ sf::st_simplify(.x)` for independent
#'   per-feature simplification, [spatial_smooth()] for Chaikin corner-rounding,
#'   [collect_sf()] to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' p1 <- sf::st_polygon(list(rbind(
#'   c(0, 0), c(1, 0), c(1, 0.5), c(1, 1), c(0, 1), c(0, 0))))
#' p2 <- sf::st_polygon(list(rbind(
#'   c(1, 0), c(2, 0), c(2, 1), c(1, 1), c(1, 0.5), c(1, 0))))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   id = c("a", "b"),
#'   geometry = sf::st_as_binary(sf::st_sfc(p1, p2), hex = TRUE)
#' ), f)
#'
#' # The shared edge is simplified once, so the two polygons stay edge-matched.
#' tbl(f) |> spatial_simplify(tolerance = 0.6) |> collect_sf()
#' unlink(f)
#'
#' @export
spatial_simplify <- function(x, tolerance, by = NULL, geom = "geometry",
                             crs = NA, flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed polygon coverage to simplify)")
  if (!is.numeric(tolerance) || length(tolerance) != 1L ||
      !is.finite(tolerance) || tolerance <= 0)
    stop("`tolerance` must be a single positive number (the simplify distance)")
  crs <- .resolve_crs(x, crs)
  .partition_each(x, by, geom, crs, .simplify_fn(geom, crs, tolerance),
                  flush_rows)
}

# -- linear referencing (locate points along a resident line layer) -----------

# For one decoded batch of points, find the nearest resident line and the
# position of each point along it. Returns the batch plus the line identifier,
# the measure (distance along the line from its start to the projected point),
# and the perpendicular distance to the line; with `snap = TRUE` the point
# geometry is replaced by its projection onto the line.
.locate_batch <- function(sb, yg, yid, id_col, measure_col, dist_col, snap) {
  if (nrow(sb) == 0L) {
    sb[[id_col]]      <- yid[integer(0)]
    sb[[measure_col]] <- numeric(0)
    sb[[dist_col]]    <- numeric(0)
    return(sb)
  }
  pts  <- sf::st_geometry(sb)
  nf   <- sf::st_nearest_feature(pts, yg)
  ln   <- yg[nf]
  meas <- as.numeric(sf::st_line_project(ln, pts))
  dper <- as.numeric(sf::st_distance(pts, ln, by_element = TRUE))
  if (snap)
    sf::st_geometry(sb) <- sf::st_line_interpolate(ln, meas)
  sb[[id_col]]      <- yid[nf]
  sb[[measure_col]] <- meas
  sb[[dist_col]]    <- dper
  rownames(sb) <- NULL
  sb
}

#' Locate streamed points along a resident line layer
#'
#' Streams a large point layer `x` through the engine and, for each point, finds
#' the nearest line of a small resident `line` layer and where the point falls
#' along it -- linear referencing (`sf::st_line_project`). Each point gets the
#' identifier of its nearest line, the **measure** (distance along that line from
#' its start to the point's projection), and the perpendicular distance to the
#' line. With `snap = TRUE` the point geometry is moved onto the line at that
#' measure. This is the two-layer companion to a per-feature
#' `sf::st_line_interpolate`, which goes the other way (a measure back to a
#' point); the billion-row point stream never materializes, while `line` (the
#' reference network) stays resident.
#'
#' Nearest line and distance are \pkg{sf}'s [sf::st_nearest_feature] and
#' [sf::st_distance]: planar (CRS units) on projected or unprojected planar data,
#' great-circle (metres) on geographic coordinates with spherical geometry on
#' (`sf::sf_use_s2()`). Points arrive either as a hex-WKB geometry column
#' (`geom`) or as two coordinate columns (`coords`). The \pkg{sf} package is an
#' optional dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param line An `sf` or `sfc` object of the reference lines (the resident
#'   layer).
#' @param y_id Optional name of a column in `line` whose value identifies the
#'   matched line in the output. Default `NULL` uses `line`'s 1-based row index.
#' @param id_col,measure_col,dist_col Names of the output columns holding the
#'   matched-line identifier, the measure along the line, and the perpendicular
#'   distance. Defaults `"line"`, `"measure"`, `"distance"`.
#' @param snap If `TRUE`, replace each point's geometry with its projection onto
#'   the nearest line. Default `FALSE` keeps the original points.
#'
#' @return A `vectra_node` of `x`'s rows -- geometry included (or snapped onto
#'   the line) -- plus the matched-line identifier, the measure, and the
#'   perpendicular distance, backed by temporary `.vtr` spills (removed when the
#'   node is garbage-collected) and carrying the input CRS.
#'
#' @seealso [spatial_knn()] for nearest neighbours with distances,
#'   [spatial_join()] for a nearest-feature attribute join, [spatial_map()] with
#'   `~ sf::st_line_interpolate(line, .x$m)` for the inverse, [collect_sf()] to
#'   materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' line <- sf::st_sfc(
#'   sf::st_linestring(rbind(c(0, 0), c(10, 0))),
#'   sf::st_linestring(rbind(c(0, 5), c(0, 15))))
#' line <- sf::st_sf(road = c("main", "side"), geometry = line)
#' pts  <- data.frame(id = 1:2, x = c(3, 1), y = c(1, 9))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(pts, f)
#'
#' # Each point's position along its nearest road.
#' tbl(f) |>
#'   spatial_locate(line, coords = c("x", "y"), y_id = "road") |>
#'   collect()
#' unlink(f)
#'
#' @export
spatial_locate <- function(x, line, geom = "geometry", coords = NULL, crs = NA,
                           y_id = NULL, id_col = "line", measure_col = "measure",
                           dist_col = "distance", snap = FALSE, out_geom = NULL,
                           flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed point layer)")
  if (!inherits(line, "sf") && !inherits(line, "sfc"))
    stop("`line` must be an sf or sfc object (the resident reference lines)")
  crs  <- .resolve_crs(x, crs)
  line <- .align_resident_crs(line, crs)
  yg   <- sf::st_geometry(line)
  yid  <- if (is.null(y_id)) seq_len(length(yg)) else {
    if (!inherits(line, "sf") || !y_id %in% names(line))
      stop(sprintf("`y_id` column '%s' not found in `line`", y_id))
    line[[y_id]]
  }
  if (is.null(out_geom)) out_geom <- if (is.null(coords)) geom else "geometry"
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  batch_fn <- function(sb)
    .locate_batch(sb, yg, yid, id_col, measure_col, dist_col, snap)
  .spatial_stream(x, batch_fn, geom, coords, crs, out_geom, fr)
}

# -- centerline (medial axis of a polygon) ------------------------------------

# Approximate the medial axis (centerline) of one polygon from the Voronoi
# diagram of its densified boundary: the Voronoi edges that lie inside the
# polygon trace the points equidistant from two boundary stretches, i.e. its
# skeleton. The boundary is densified to spacing `density` so the diagram is fine
# enough; interior edges are kept (their midpoint inside a slightly shrunk
# polygon), merged into maximal lines, and optionally pruned of branches shorter
# than `prune`. A non-polygon geometry, or one too small to sample, returns
# unchanged so a feature never vanishes.
.centerline_one <- function(g, density, prune, crs) {
  gs <- sf::st_sfc(g, crs = crs)
  if (!grepl("POLYGON", class(g)[2L])) return(gs)
  bb   <- sf::st_bbox(gs)
  diag <- sqrt((bb[["xmax"]] - bb[["xmin"]])^2 + (bb[["ymax"]] - bb[["ymin"]])^2)
  d <- if (!is.null(density) && density > 0) density else diag / 100
  if (!is.finite(d) || d <= 0) return(gs)
  b   <- sf::st_segmentize(sf::st_cast(gs, "MULTILINESTRING"), dfMaxLength = d)
  v   <- sf::st_cast(b, "POINT")
  vor <- sf::st_collection_extract(sf::st_voronoi(sf::st_union(v)), "POLYGON")
  edges <- sf::st_cast(sf::st_union(sf::st_boundary(vor)), "LINESTRING",
                       warn = FALSE)
  edges <- sf::st_cast(sf::st_node(sf::st_union(edges)), "LINESTRING",
                       warn = FALSE)
  if (!length(edges)) return(gs)
  mid    <- sf::st_line_sample(edges, sample = 0.5)
  inside <- lengths(sf::st_within(mid, sf::st_buffer(gs, -d * 0.01))) > 0
  ce <- edges[inside]
  if (!length(ce)) return(gs)
  m <- sf::st_cast(sf::st_line_merge(sf::st_union(ce)), "LINESTRING",
                   warn = FALSE)
  if (prune > 0) {
    keep <- as.numeric(sf::st_length(m)) >= prune
    if (any(keep)) m <- m[keep]
  }
  if (!length(m)) gs else m
}

.centerline_batch <- function(sb, density, prune, crs) {
  if (nrow(sb) == 0L) return(sb)
  crs <- .as_crs(crs)
  g   <- sf::st_geometry(sb)
  pieces <- lapply(seq_along(g),
                   function(i) .centerline_one(g[[i]], density, prune, crs))
  np  <- lengths(pieces)
  out <- sb[rep.int(seq_len(nrow(sb)), np), , drop = FALSE]
  sf::st_geometry(out) <- sf::st_sfc(unlist(pieces, recursive = FALSE), crs = crs)
  rownames(out) <- NULL
  out
}

#' Trace the centerline (medial axis) of streamed polygons
#'
#' Approximates the centerline of every polygon in a streamed layer, one batch
#' at a time -- the medial axis a per-feature transform such as a buffer cannot
#' produce. Each polygon's boundary is densified and its Voronoi diagram taken;
#' the Voronoi edges that fall inside the polygon trace the points equidistant
#' from two stretches of boundary, which is its skeleton, and they are merged
#' into maximal lines. This is the usual approximation for river or road
#' centerlines from a filled shape; `prune` drops the short branches that the
#' skeleton grows toward convex corners. A non-polygon geometry passes through
#' unchanged.
#'
#' The centerline is an approximation whose detail is set by `density` (the
#' boundary sampling spacing): finer sampling traces the axis more closely at
#' more cost. Geometry travels through the engine as hex-encoded WKB in a string
#' column and the CRS is carried on the returned node; the Voronoi construction
#' is \pkg{sf}/GEOS and expects projected or unprojected planar data. The
#' \pkg{sf} package is an optional dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param density Boundary sampling spacing in CRS units: the polygon outline is
#'   densified to at most this vertex spacing before the Voronoi diagram is
#'   built. `NULL` (default) uses one-hundredth of each polygon's bounding-box
#'   diagonal.
#' @param prune Drop centerline branches shorter than this length (CRS units),
#'   removing the short spurs the skeleton grows toward convex corners. `0`
#'   (default) keeps every branch.
#'
#' @return A `vectra_node` of the centerlines, each carrying its source polygon's
#'   attributes (replicated if the centerline is several lines) and the input
#'   CRS, backed by temporary `.vtr` spills removed when the node is
#'   garbage-collected.
#'
#' @seealso [spatial_construct()] with `kind = "pole"` for the single deepest
#'   interior point, [spatial_simplify()] to simplify a coverage, [collect_sf()]
#'   to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' road <- sf::st_polygon(list(rbind(
#'   c(0, 0), c(10, 0), c(10, 2), c(0, 2), c(0, 0))))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   geometry = sf::st_as_binary(sf::st_sfc(road), hex = TRUE)
#' ), f)
#'
#' # The centerline runs down the middle of the strip.
#' tbl(f) |> spatial_centerline(density = 0.25, prune = 0.5) |> collect_sf()
#' unlink(f)
#'
#' @export
spatial_centerline <- function(x, density = NULL, prune = 0,
                               geom = "geometry", crs = NA, out_geom = NULL,
                               flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed polygon layer)")
  if (!is.null(density) && (!is.numeric(density) || length(density) != 1L ||
                            !is.finite(density) || density <= 0))
    stop("`density` must be a single positive number, or NULL")
  if (!is.numeric(prune) || length(prune) != 1L || !is.finite(prune) ||
      prune < 0)
    stop("`prune` must be a single non-negative number")
  crs <- .resolve_crs(x, crs)
  if (is.null(out_geom)) out_geom <- geom
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  .spatial_stream(x, function(sb) .centerline_batch(sb, density, prune, crs),
                  geom, coords = NULL, crs = crs, out_geom = out_geom,
                  flush_rows = fr)
}

# -- planar topology (shared-edge arc table of a coverage) --------------------

# One group's worth of polygons -> the arcs of their shared-edge topology. The
# boundaries are unioned (so a border shared by two polygons is one line) and
# noded at every junction; each resulting arc is emitted once, tagged with the
# identifiers of the (up to two) polygons whose boundary covers it -- two for an
# internal shared edge, one for an outer edge (the other side `NA`). The arc
# identifiers come from `id` (a column of `x`) or the 1-based feature order.
.topology_fn <- function(id, by, geom, crs, face_cols) function(df) {
  sb   <- .sf_decode_chunk(df, geom, NULL, crs)
  g    <- sf::st_geometry(sb)
  bnd  <- sf::st_boundary(g)
  arcs <- sf::st_cast(sf::st_node(sf::st_union(bnd)), "LINESTRING", warn = FALSE)
  arcs <- arcs[!sf::st_is_empty(arcs)]
  if (!length(arcs)) return(NULL)
  fid <- if (is.null(id)) seq_len(nrow(df)) else df[[id]]
  nb  <- lapply(seq_along(arcs), function(i)
    fid[vapply(seq_along(g),
               function(j) length(sf::st_covered_by(arcs[i], bnd[j])[[1L]]) > 0,
               logical(1))])
  pick <- function(k) {
    out <- fid[rep(NA_integer_, length(nb))]
    for (i in seq_along(nb)) if (length(nb[[i]]) >= k) out[i] <- nb[[i]][[k]]
    out
  }
  rowdf <- if (is.null(by)) data.frame(matrix(nrow = length(arcs), ncol = 0))
           else df[rep(1L, length(arcs)), by, drop = FALSE]
  rowdf[[face_cols[1L]]] <- pick(1L)
  rowdf[[face_cols[2L]]] <- pick(2L)
  rowdf[[geom]] <- sf::st_as_binary(arcs, hex = TRUE)
  rownames(rowdf) <- NULL
  rowdf
}

#' Build the shared-edge topology of a polygon coverage
#'
#' Decomposes a polygon coverage into the arcs of its planar topology: each
#' border is returned once, tagged with the polygons on either side. Where the
#' raw boundaries of adjacent polygons each carry their own copy of a shared
#' edge, this nodes the unioned boundaries so a shared border is a single arc
#' carrying the identifiers of both neighbours -- the "build topology" of a GIS,
#' and the adjacency a dissolve or a coverage edit needs. An internal arc names
#' two faces; an outer arc names one and leaves the other side `NA`. Like
#' [spatial_dissolve()] it rides the **partition tier**: `x` is spilled once and
#' routed into one disjoint shard per `by` group in a single bounded pass, and
#' each group is treated as an independent coverage. Peak memory is the routing
#' budget during the pass, then one group's geometry while its arcs are built.
#'
#' Geometry travels through the engine as hex-encoded WKB in a string column and
#' the CRS is carried on the returned node; the noding is \pkg{sf}/GEOS and
#' expects projected or unprojected planar data. The \pkg{sf} package is an
#' optional dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param id Optional name of a column of `x` whose value identifies each polygon
#'   in the face columns. `NULL` (default) uses the 1-based feature order within
#'   the group.
#' @param by Character vector of attribute columns whose groups are each built as
#'   an independent coverage. `NULL` (default) treats the whole layer as one
#'   coverage.
#' @param face_cols Length-2 character vector naming the two output columns that
#'   hold the identifiers of the polygons on either side of each arc. Default
#'   `c("face1", "face2")`.
#'
#' @return A `vectra_node` of one row per arc -- the arc geometry plus the two
#'   face-identifier columns (and any `by` columns) -- carrying the input CRS and
#'   backed by temporary `.vtr` spills removed when the node is garbage-collected.
#'
#' @seealso [spatial_polygonize()] to rebuild faces from arcs,
#'   [spatial_dissolve()] to merge geometries by group, [collect_sf()] to
#'   materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' p1 <- sf::st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))))
#' p2 <- sf::st_polygon(list(rbind(c(1, 0), c(2, 0), c(2, 1), c(1, 1), c(1, 0))))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   id = c("a", "b"),
#'   geometry = sf::st_as_binary(sf::st_sfc(p1, p2), hex = TRUE)
#' ), f)
#'
#' # The shared edge appears once, tagged with both neighbours.
#' tbl(f) |> spatial_topology(id = "id") |> collect()
#' unlink(f)
#'
#' @export
spatial_topology <- function(x, id = NULL, by = NULL, geom = "geometry",
                             crs = NA, face_cols = c("face1", "face2"),
                             flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed polygon coverage)")
  if (!is.null(id) && (!is.character(id) || length(id) != 1L))
    stop("`id` must be a single column name, or NULL")
  if (!is.character(face_cols) || length(face_cols) != 2L ||
      anyNA(face_cols) || face_cols[1L] == face_cols[2L])
    stop("`face_cols` must be two distinct column names")
  crs <- .resolve_crs(x, crs)
  .partition_each(x, by, geom, crs, .topology_fn(id, by, geom, crs, face_cols),
                  flush_rows, need = id)
}

# -- eliminate (merge sliver polygons into a neighbour) -----------------------

# One group's worth of polygons -> the same coverage with every feature smaller
# than `max_area` absorbed into a neighbour. Adjacent features are found by
# intersection; each small feature is linked to the neighbour it shares the
# longest border with (or the largest-area neighbour), and an area-rooted
# union-find collapses chains of slivers so each connected run flows to its
# single largest member, whose attribute row survives. A small feature with no
# neighbour is kept unchanged, so nothing vanishes.
.eliminate_fn <- function(by, geom, crs, max_area, into) function(df) {
  sb   <- .sf_decode_chunk(df, geom, NULL, crs)
  g    <- sf::st_geometry(sb)
  n    <- length(g)
  if (!n) return(NULL)
  if (n == 1L) return(df)
  cols  <- setdiff(names(df), geom)
  area  <- as.numeric(sf::st_area(g))
  small <- area < max_area
  parent <- seq_len(n)
  find <- function(i) {
    while (parent[i] != i) { parent[i] <<- parent[parent[i]]; i <- parent[i] }
    i
  }
  if (any(small)) {
    touch <- sf::st_intersects(g)
    bnd   <- if (into == "longest_border") sf::st_boundary(g) else NULL
    for (i in which(small)) {
      cand <- setdiff(touch[[i]], i)
      if (!length(cand)) next
      if (into == "longest_border") {
        sl <- vapply(cand, function(j) {
          ix <- suppressWarnings(sf::st_intersection(bnd[i], bnd[j]))
          if (length(ix)) sum(as.numeric(sf::st_length(ix))) else 0
        }, numeric(1))
        best <- cand[which.max(sl)]
      } else {
        best <- cand[which.max(area[cand])]
      }
      ri <- find(i); rb <- find(best)
      if (ri != rb) {
        if (area[rb] >= area[ri]) parent[ri] <- rb else parent[rb] <- ri
      }
    }
  }
  roots <- vapply(seq_len(n), find, integer(1))
  ur    <- unique(roots)
  geoms <- do.call(c, lapply(ur, function(r) sf::st_union(g[roots == r])))
  rowdf <- df[ur, cols, drop = FALSE]
  rowdf[[geom]] <- sf::st_as_binary(geoms, hex = TRUE)
  rownames(rowdf) <- NULL
  rowdf
}

#' Merge sliver polygons into a neighbour
#'
#' Cleans a polygon coverage by absorbing every feature whose area is below
#' `max_area` into an adjacent feature (the QGIS "Eliminate Selected Polygons"):
#' the sliver removal a per-feature transform cannot do, because the target a
#' sliver merges into is one of its neighbours, not the sliver itself. Each small
#' feature is joined to the neighbour it shares the longest border with (or the
#' largest-area neighbour, with `into = "largest_area"`); chains of slivers
#' collapse so a connected run of small features flows to its single largest
#' member, whose attribute row survives. A small feature with no neighbour is
#' kept unchanged, so nothing vanishes. Like [spatial_dissolve()] it rides the
#' **partition tier**: `x` is spilled once and routed into one disjoint shard per
#' `by` group in a single bounded pass, and each group is cleaned as an
#' independent coverage. Peak memory is the routing budget during the pass, then
#' one group's geometry while its slivers are merged -- partition on a key whose
#' groups fit in memory. With no `by`, the whole layer is one coverage.
#'
#' Adjacency and shared-border length are \pkg{sf}/GEOS ([sf::st_intersects],
#' [sf::st_boundary], [sf::st_intersection]) and expect projected or unprojected
#' planar data; `max_area` is in CRS units squared. Geometry travels through the
#' engine as hex-encoded WKB in a string column and the CRS is carried on the
#' returned node. The \pkg{sf} package is an optional dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param max_area Area threshold in CRS units squared: a feature smaller than
#'   this is a sliver and is merged into a neighbour. Larger values absorb more.
#' @param by Character vector of attribute columns whose groups are each cleaned
#'   as an independent coverage. `NULL` (default) treats the whole layer as one
#'   coverage.
#' @param into How to pick the neighbour a sliver merges into: `"longest_border"`
#'   (default) the neighbour sharing the longest boundary, or `"largest_area"`
#'   the neighbour with the greatest area.
#'
#' @return A `vectra_node` of the cleaned coverage -- one row per surviving
#'   feature, each carrying its (largest member's) attributes and the input CRS,
#'   backed by temporary `.vtr` spills removed when the node is garbage-collected.
#'
#' @seealso [spatial_dissolve()] to merge geometries by attribute,
#'   [spatial_simplify()] for coverage-preserving simplification,
#'   [spatial_topology()] for the shared-edge adjacency, [collect_sf()] to
#'   materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' big    <- sf::st_polygon(list(rbind(
#'   c(0, 0), c(10, 0), c(10, 10), c(0, 10), c(0, 0))))
#' sliver <- sf::st_polygon(list(rbind(
#'   c(10, 0), c(10.3, 0), c(10.3, 10), c(10, 10), c(10, 0))))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   id = c("keep", "sliver"),
#'   geometry = sf::st_as_binary(sf::st_sfc(big, sliver), hex = TRUE)
#' ), f)
#'
#' # The thin sliver is absorbed into the square it borders.
#' tbl(f) |> spatial_eliminate(max_area = 5) |> collect_sf()
#' unlink(f)
#'
#' @export
spatial_eliminate <- function(x, max_area, by = NULL,
                              into = c("longest_border", "largest_area"),
                              geom = "geometry", crs = NA, flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed polygon coverage to clean)")
  if (!is.numeric(max_area) || length(max_area) != 1L || !is.finite(max_area) ||
      max_area <= 0)
    stop("`max_area` must be a single positive number (the sliver area threshold)")
  into <- match.arg(into)
  crs  <- .resolve_crs(x, crs)
  .partition_each(x, by, geom, crs, .eliminate_fn(by, geom, crs, max_area, into),
                  flush_rows)
}

# The partition tier shared with dissolve and construct. Routes `x` into one
# shard per `by` group on disk in a single bounded pass, then applies
# `group_fn(df)` to each shard's collected rows (geometry as hex-WKB in `geom`),
# accumulating the returned data frames into one node. Peak memory is the
# routing budget, then one group's geometry while `group_fn` runs.
.partition_each <- function(x, by, geom, crs, group_fn, flush_rows,
                            need = NULL) {
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed layer)")
  if (!is.null(by) && !is.character(by))
    stop("`by` must be a character vector of column names, or NULL")

  spill <- tempfile(fileext = ".vtr")
  on.exit(unlink(spill), add = TRUE)
  write_vtr(x, spill)

  schema <- .Call(C_node_schema, tbl(spill)$.node)
  miss <- setdiff(c(by, geom, need), schema$name)
  if (length(miss))
    stop(sprintf("column(s) not found in the stream: %s",
                 paste(miss, collapse = ", ")))

  budget <- getOption("vectra.partition_budget", .PARTITION_BUDGET)
  res <- .partition_router(spill, .dissolve_assign(by), budget)
  on.exit(unlink(unlist(res$runs, use.names = FALSE)), add = TRUE)

  fr  <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  acc <- .run_accumulator(fr)
  for (lab in sort(names(res$runs))) {
    df  <- collect(.concat_runs(res$runs[[lab]]))
    out <- group_fn(df)
    if (!is.null(out) && nrow(out)) acc$push(.coerce_for_vtr(out))
  }
  acc$finish(crs = crs, empty_geom = geom)
}

# Topology verbs: polygonize (lines -> faces), line_merge (segments -> maximal
# lines), simplify (coverage-preserving), locate (linear referencing), centerline
# (medial axis), topology (shared-edge arcs), eliminate (merge slivers). Geometry
# round-trips through a .vtr; CRS is carried on the node.

skip_if_not_installed("sf")

vtr_from <- function(sfobj) {
  f <- tempfile(fileext = ".vtr")
  df <- as.data.frame(sf::st_drop_geometry(sfobj))
  df$geometry <- sf::st_as_binary(sf::st_geometry(sfobj), hex = TRUE)
  write_vtr(df, f)
  f
}

# A 2x2 grid of unit cells drawn as six lines (three horizontal, three vertical).
grid_lines <- function() {
  sf::st_sfc(
    sf::st_linestring(rbind(c(0, 0), c(2, 0))),
    sf::st_linestring(rbind(c(0, 1), c(2, 1))),
    sf::st_linestring(rbind(c(0, 2), c(2, 2))),
    sf::st_linestring(rbind(c(0, 0), c(0, 2))),
    sf::st_linestring(rbind(c(1, 0), c(1, 2))),
    sf::st_linestring(rbind(c(2, 0), c(2, 2))))
}

# -- polygonize ---------------------------------------------------------------

test_that("polygonize forms the faces enclosed by a line network", {
  x <- sf::st_sf(geometry = grid_lines())
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_polygonize() |> collect_sf()
  expect_equal(nrow(d), 4L)
  expect_true(all(sf::st_geometry_type(d) == "POLYGON"))
  expect_equal(sort(as.numeric(sf::st_area(d))), rep(1, 4), tolerance = 1e-9)
})

test_that("polygonize builds one set of faces per by group", {
  g1 <- grid_lines()
  g2 <- grid_lines() + c(10, 10)
  x <- sf::st_sf(
    grp = c(rep("a", 6), rep("b", 6)),
    geometry = sf::st_sfc(c(g1, g2)))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_polygonize(by = "grp") |> collect_sf()
  expect_equal(nrow(d), 8L)
  expect_equal(sum(d$grp == "a"), 4L)
  expect_equal(sum(d$grp == "b"), 4L)
})

test_that("polygonize rejects a missing by column and a non-node input", {
  x <- sf::st_sf(geometry = grid_lines())
  f <- vtr_from(x); on.exit(unlink(f))
  expect_error(collect(spatial_polygonize(tbl(f), by = "nope")), "not found")
  expect_error(spatial_polygonize(data.frame(a = 1)), "vectra_node")
})

# -- line_merge ---------------------------------------------------------------

test_that("line_merge sews end-to-end segments into one maximal line", {
  seg <- sf::st_sfc(
    sf::st_linestring(rbind(c(0, 0), c(1, 0))),
    sf::st_linestring(rbind(c(1, 0), c(2, 0))),
    sf::st_linestring(rbind(c(2, 0), c(3, 0))))
  x <- sf::st_sf(geometry = seg)
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_line_merge() |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_equal(as.numeric(sf::st_length(d)), 3, tolerance = 1e-9)
})

test_that("line_merge keeps disconnected chains separate", {
  seg <- sf::st_sfc(
    sf::st_linestring(rbind(c(0, 0), c(1, 0))),
    sf::st_linestring(rbind(c(1, 0), c(2, 0))),
    sf::st_linestring(rbind(c(5, 5), c(6, 5))))
  x <- sf::st_sf(geometry = seg)
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_line_merge() |> collect_sf()
  expect_equal(nrow(d), 2L)
  expect_equal(sort(as.numeric(sf::st_length(d))), c(1, 2), tolerance = 1e-9)
})

# -- simplify (coverage-preserving) -------------------------------------------

test_that("coverage simplify keeps neighbours edge-matched and attributes", {
  p1 <- sf::st_polygon(list(rbind(
    c(0, 0), c(1, 0), c(1.1, 0.5), c(1, 1), c(0, 1), c(0, 0))))
  p2 <- sf::st_polygon(list(rbind(
    c(1, 0), c(2, 0), c(2, 1), c(1, 1), c(1.1, 0.5), c(1, 0))))
  x <- sf::st_sf(id = c("a", "b"), geometry = sf::st_sfc(p1, p2))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_simplify(tolerance = 0.5) |> collect_sf()
  expect_equal(nrow(d), 2L)
  expect_setequal(d$id, c("a", "b"))
  in_area  <- as.numeric(sf::st_area(sf::st_union(sf::st_sfc(p1, p2))))
  out_area <- as.numeric(sf::st_area(sf::st_union(sf::st_geometry(d))))
  # a torn coverage would leave a sliver or overlap and change the union area
  expect_equal(out_area, in_area, tolerance = 1e-9)
  inter <- suppressWarnings(
    sf::st_intersection(sf::st_geometry(d)[1], sf::st_geometry(d)[2]))
  if (length(inter)) expect_lt(as.numeric(sf::st_area(inter)), 1e-9)
})

test_that("coverage simplify drops near-collinear vertices", {
  ring <- rbind(c(0, 0), c(2, 0.05), c(4, 0), c(4, 4), c(0, 4), c(0, 0))
  x <- sf::st_sf(id = "a", geometry = sf::st_sfc(sf::st_polygon(list(ring))))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_simplify(tolerance = 0.2) |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_lt(nrow(sf::st_coordinates(d)), nrow(ring))
})

# -- locate (linear referencing) ----------------------------------------------

test_that("locate returns the measure, nearest line, and offset distance", {
  line <- sf::st_sf(road = c("main", "side"), geometry = sf::st_sfc(
    sf::st_linestring(rbind(c(0, 0), c(10, 0))),
    sf::st_linestring(rbind(c(0, 5), c(0, 15)))))
  pts <- data.frame(id = 1:2, x = c(3, 1), y = c(1, 9))
  f <- tempfile(fileext = ".vtr"); write_vtr(pts, f); on.exit(unlink(f))
  d <- tbl(f) |>
    spatial_locate(line, coords = c("x", "y"), y_id = "road") |>
    collect()
  expect_equal(d$line, c("main", "side"))
  expect_equal(d$measure, c(3, 4), tolerance = 1e-6)
  expect_equal(d$distance, c(1, 1), tolerance = 1e-6)
})

test_that("locate snap moves each point onto its nearest line", {
  line <- sf::st_sfc(sf::st_linestring(rbind(c(0, 0), c(10, 0))))
  pts <- data.frame(id = 1L, x = 3, y = 2)
  f <- tempfile(fileext = ".vtr"); write_vtr(pts, f); on.exit(unlink(f))
  d <- tbl(f) |>
    spatial_locate(line, coords = c("x", "y"), snap = TRUE) |>
    collect_sf()
  expect_true(all(sf::st_geometry_type(d) == "POINT"))
  expect_equal(unname(sf::st_coordinates(d)[1, c("X", "Y")]), c(3, 0),
               tolerance = 1e-6)
})

test_that("locate rejects a non-node input and a non-sf line", {
  line <- sf::st_sfc(sf::st_linestring(rbind(c(0, 0), c(1, 0))))
  expect_error(spatial_locate(data.frame(a = 1), line), "vectra_node")
  pts <- data.frame(x = 0, y = 0)
  f <- tempfile(fileext = ".vtr"); write_vtr(pts, f); on.exit(unlink(f))
  expect_error(spatial_locate(tbl(f), data.frame(a = 1), coords = c("x", "y")),
               "sf or sfc")
})

# -- centerline ---------------------------------------------------------------

test_that("centerline runs down the middle of a strip", {
  road <- sf::st_polygon(list(rbind(
    c(0, 0), c(10, 0), c(10, 2), c(0, 2), c(0, 0))))
  x <- sf::st_sf(geometry = sf::st_sfc(road))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_centerline(density = 0.25, prune = 0.5) |> collect_sf()
  expect_gt(nrow(d), 0L)
  expect_true(all(sf::st_geometry_type(d) == "LINESTRING"))
  # the centerline lies inside the strip and tracks mid-height (y ~ 1)
  expect_true(all(lengths(sf::st_within(
    sf::st_geometry(d), sf::st_sfc(road))) > 0))
  ys <- sf::st_coordinates(d)[, "Y"]
  expect_lt(abs(mean(ys) - 1), 0.3)
  # it spans most of the 10-unit length
  expect_gt(sum(as.numeric(sf::st_length(d))), 8)
})

test_that("centerline passes non-polygon geometry through unchanged", {
  ln <- sf::st_linestring(rbind(c(0, 0), c(5, 0)))
  x <- sf::st_sf(geometry = sf::st_sfc(ln))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_centerline() |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_equal(as.numeric(sf::st_length(d)), 5, tolerance = 1e-9)
})

test_that("centerline rejects bad density and prune", {
  x <- sf::st_sf(geometry = sf::st_sfc(
    sf::st_polygon(list(rbind(c(0, 0), c(4, 0), c(4, 2), c(0, 2), c(0, 0))))))
  f <- vtr_from(x); on.exit(unlink(f))
  expect_error(spatial_centerline(tbl(f), density = -1), "density")
  expect_error(spatial_centerline(tbl(f), prune = -1), "prune")
})

# -- topology -----------------------------------------------------------------

test_that("topology returns shared edges once with both neighbours", {
  p1 <- sf::st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))))
  p2 <- sf::st_polygon(list(rbind(c(1, 0), c(2, 0), c(2, 1), c(1, 1), c(1, 0))))
  p3 <- sf::st_polygon(list(rbind(c(0, 1), c(1, 1), c(1, 2), c(0, 2), c(0, 1))))
  x <- sf::st_sf(id = c("a", "b", "c"), geometry = sf::st_sfc(p1, p2, p3))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_topology(id = "id") |> collect_sf()
  expect_true(all(c("face1", "face2") %in% names(d)))
  expect_true(all(sf::st_geometry_type(d) == "LINESTRING"))
  nfaces <- (!is.na(d$face1)) + (!is.na(d$face2))
  # exactly two internal shared edges (a|b and a|c), the rest outer (one face)
  expect_equal(sum(nfaces == 2L), 2L)
  expect_gt(sum(nfaces == 1L), 0L)
  shared <- d[nfaces == 2L, ]
  pair <- function(r) paste(sort(c(r$face1, r$face2)), collapse = "|")
  expect_setequal(vapply(seq_len(nrow(shared)),
                         function(i) pair(shared[i, ]), character(1)),
                  c("a|b", "a|c"))
})

test_that("topology arcs rebuild the original coverage via polygonize", {
  p1 <- sf::st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))))
  p2 <- sf::st_polygon(list(rbind(c(1, 0), c(2, 0), c(2, 1), c(1, 1), c(1, 0))))
  x <- sf::st_sf(id = c("a", "b"), geometry = sf::st_sfc(p1, p2))
  f <- vtr_from(x); on.exit(unlink(f))
  arcs <- tbl(f) |> spatial_topology(id = "id") |> collect_sf()
  faces <- sf::st_collection_extract(
    sf::st_polygonize(sf::st_union(sf::st_geometry(arcs))), "POLYGON")
  expect_equal(length(faces), 2L)
  expect_equal(as.numeric(sf::st_area(sf::st_union(faces))), 2, tolerance = 1e-9)
})

test_that("topology rejects a missing id column and bad face_cols", {
  x <- sf::st_sf(geometry = sf::st_sfc(
    sf::st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))))))
  f <- vtr_from(x); on.exit(unlink(f))
  expect_error(collect(spatial_topology(tbl(f), id = "nope")), "not found")
  expect_error(spatial_topology(tbl(f), face_cols = "only_one"), "two distinct")
})

# -- eliminate (merge slivers) ------------------------------------------------

test_that("eliminate absorbs a sliver into the square it borders", {
  big <- sf::st_polygon(list(rbind(
    c(0, 0), c(10, 0), c(10, 10), c(0, 10), c(0, 0))))
  sliver <- sf::st_polygon(list(rbind(
    c(10, 0), c(10.3, 0), c(10.3, 10), c(10, 10), c(10, 0))))
  x <- sf::st_sf(id = c("keep", "sliver"), geometry = sf::st_sfc(big, sliver))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_eliminate(max_area = 5) |> collect_sf()
  expect_equal(nrow(d), 1L)
  # the survivor keeps the large feature's attributes
  expect_equal(d$id, "keep")
  # the union area is preserved (nothing dropped, nothing double-counted)
  expect_equal(as.numeric(sf::st_area(sf::st_geometry(d))), 103,
               tolerance = 1e-9)
})

test_that("eliminate collapses a chain of slivers to the largest member", {
  b  <- sf::st_polygon(list(rbind(
    c(0, 0), c(10, 0), c(10, 10), c(0, 10), c(0, 0))))
  s1 <- sf::st_polygon(list(rbind(
    c(10, 0), c(10.3, 0), c(10.3, 10), c(10, 10), c(10, 0))))
  s2 <- sf::st_polygon(list(rbind(
    c(10.3, 0), c(10.6, 0), c(10.6, 10), c(10.3, 10), c(10.3, 0))))
  x <- sf::st_sf(id = c("big", "s1", "s2"), geometry = sf::st_sfc(b, s1, s2))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_eliminate(max_area = 5) |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_equal(d$id, "big")
})

test_that("eliminate keeps a sliver with no neighbour and leaves big ones", {
  b1  <- sf::st_polygon(list(rbind(
    c(0, 0), c(10, 0), c(10, 10), c(0, 10), c(0, 0))))
  b2  <- sf::st_polygon(list(rbind(
    c(10, 0), c(20, 0), c(20, 10), c(10, 10), c(10, 0))))
  iso <- sf::st_polygon(list(rbind(
    c(50, 50), c(50.2, 50), c(50.2, 51), c(50, 51), c(50, 50))))
  x <- sf::st_sf(id = c("b1", "b2", "iso"), geometry = sf::st_sfc(b1, b2, iso))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_eliminate(max_area = 5) |> collect_sf()
  # two untouched big squares plus the unmergeable isolated sliver
  expect_equal(nrow(d), 3L)
  expect_setequal(d$id, c("b1", "b2", "iso"))
})

test_that("eliminate by-group cleans each coverage independently", {
  mk <- function(dx) {
    big <- sf::st_polygon(list(rbind(
      c(0, 0), c(10, 0), c(10, 10), c(0, 10), c(0, 0)))) + c(dx, 0)
    sl  <- sf::st_polygon(list(rbind(
      c(10, 0), c(10.3, 0), c(10.3, 10), c(10, 10), c(10, 0)))) + c(dx, 0)
    sf::st_sfc(big, sl)
  }
  x <- sf::st_sf(
    grp = c("a", "a", "b", "b"),
    geometry = sf::st_sfc(c(mk(0), mk(100))))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_eliminate(max_area = 5, by = "grp") |> collect_sf()
  expect_equal(nrow(d), 2L)
  expect_setequal(d$grp, c("a", "b"))
})

test_that("eliminate rejects a bad threshold and a non-node input", {
  x <- sf::st_sf(geometry = sf::st_sfc(
    sf::st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))))))
  f <- vtr_from(x); on.exit(unlink(f))
  expect_error(spatial_eliminate(tbl(f), max_area = -1), "positive")
  expect_error(spatial_eliminate(data.frame(a = 1), max_area = 1), "vectra_node")
})

#' Logically delete rows from a .vtr file
#'
#' Marks the specified 0-based physical row indices as deleted by writing (or
#' updating) a tombstone side file (`<path>.del`). The original `.vtr` file is
#' never modified. The next call to [tbl()] on the same path will automatically
#' exclude the deleted rows.
#'
#' Tombstone files are cumulative: calling `delete_vtr()` multiple times on the
#' same file merges all deletions (union, deduplicated). To undo deletions,
#' remove the `.del` file manually with `unlink(paste0(path, ".del"))`.
#'
#' @param path File path of the `.vtr` file to delete rows from.
#' @param row_ids A numeric vector of **0-based** physical row indices to
#'   delete. Out-of-range indices are silently ignored on read (they will
#'   never match a real row).
#'
#' @return Invisible `NULL`.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#'
#' # Delete the first and third rows (0-based indices 0 and 2)
#' delete_vtr(f, c(0, 2))
#'
#' result <- tbl(f) |> collect()
#' stopifnot(nrow(result) == nrow(mtcars) - 2L)
#'
#' unlink(c(f, paste0(f, ".del")))
#'
#' @export
delete_vtr <- function(path, row_ids) {
  check_scalar_string(path)
  if (!is.numeric(row_ids) && !is.integer(row_ids))
    stop("row_ids must be a numeric or integer vector of 0-based row indices")
  path <- normalizePath(path, mustWork = TRUE)
  .Call(C_delete_vtr, path, as.double(row_ids))
  invisible(NULL)
}

#' Print the execution plan for a vectra query
#'
#' Shows the node types, column schemas, and structure of the lazy query plan.
#'
#' @param x A `vectra_node` object.
#' @param ... Ignored.
#'
#' @return Invisible `x`.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> filter(cyl > 4) |> select(mpg, cyl) |> explain()
#' unlink(f)
#'
#' @export
explain <- function(x, ...) {
  UseMethod("explain")
}

#' @export
explain.vectra_node <- function(x, ...) {
  plan_lines <- .Call(C_node_plan, x$.node)
  schema <- .Call(C_node_schema, x$.node)

  cat("vectra execution plan\n")
  if (!is.null(x$.groups)) {
    cat(sprintf("Groups: %s\n", paste(x$.groups, collapse = ", ")))
  }
  cat("\n")
  for (line in plan_lines) cat(line, "\n")
  cat("\n")
  cat(sprintf("Output columns (%d):\n", length(schema$name)))
  for (i in seq_along(schema$name)) {
    cat(sprintf("  %s <%s>\n", schema$name[i], schema$type[i]))
  }
  cat("\n")
  print(grade_of(x))
  invisible(x)
}

# Window function support for vectra
#
# Window functions are detected inside mutate() and routed to C_window_node.
# Supported: lag(), lead(), row_number(), cumsum(), cummean(), cummin(), cummax()

# Known window function names
.win_fns <- c("lag", "lead", "row_number", "rank", "dense_rank",
              "cumsum", "cummean", "cummin", "cummax",
              "ntile", "percent_rank", "cume_dist")

# Check if an expression is a window function call
is_window_call <- function(expr) {
  if (!is.call(expr)) return(FALSE)
  fn <- as.character(expr[[1]])
  fn %in% .win_fns
}

# Resolve an order argument that may be wrapped in desc(): returns col + desc.
.parse_order_arg <- function(a) {
  if (is.call(a) && identical(as.character(a[[1]]), "desc"))
    list(col = as.character(a[[2]]), desc = TRUE)
  else
    list(col = as.character(a), desc = FALSE)
}

# Parse a window function call into a spec list for C
parse_window_spec <- function(expr, output_name) {
  fn <- as.character(expr[[1]])

  if (fn == "row_number") {
    # row_number() -> input order; row_number(col) / row_number(desc(col)) ->
    # ordered, deterministic 1..n within each partition.
    if (length(expr) >= 2) {
      pa <- .parse_order_arg(expr[[2]])
      return(list(name = output_name, kind = "row_number", col = pa$col,
                  offset = 1L, default = NULL, desc = pa$desc))
    }
    return(list(name = output_name, kind = "row_number", col = NULL,
                offset = 1L, default = NULL, desc = FALSE))
  }

  if (fn == "rank") {
    pa <- .parse_order_arg(expr[[2]])
    return(list(name = output_name, kind = "rank", col = pa$col,
                offset = 1L, default = NULL, desc = pa$desc))
  }

  if (fn == "dense_rank") {
    col <- as.character(expr[[2]])
    return(list(name = output_name, kind = "dense_rank", col = col,
                offset = 1L, default = NULL))
  }

  if (fn %in% c("lag", "lead")) {
    # lag(col, n = 1, default = NA)
    col <- as.character(expr[[2]])
    offset <- 1L
    default_val <- NULL

    args <- as.list(expr)[-1]  # drop function name
    arg_names <- names(args)

    if (length(args) >= 2) {
      # Second arg is n (positional or named)
      if (!is.null(arg_names) && !is.na(match("n", arg_names))) {
        offset <- as.integer(eval(args[[match("n", arg_names)]]))
      } else if (length(args) >= 2 && (is.null(arg_names) || arg_names[2] == "")) {
        offset <- as.integer(eval(args[[2]]))
      }
    }

    if (!is.null(arg_names) && !is.na(match("default", arg_names))) {
      default_val <- as.double(eval(args[[match("default", arg_names)]]))
    } else if (length(args) >= 3 && (is.null(arg_names) || arg_names[3] == "")) {
      default_val <- as.double(eval(args[[3]]))
    }

    return(list(name = output_name, kind = fn, col = col,
                offset = offset, default = default_val))
  }

  if (fn == "ntile") {
    # ntile(n) - divide into n buckets
    n_tiles <- as.integer(eval(expr[[2]]))
    return(list(name = output_name, kind = "ntile", col = NULL,
                offset = n_tiles, default = NULL))
  }

  if (fn == "percent_rank") {
    col <- as.character(expr[[2]])
    return(list(name = output_name, kind = "percent_rank", col = col,
                offset = 1L, default = NULL))
  }

  if (fn == "cume_dist") {
    col <- as.character(expr[[2]])
    return(list(name = output_name, kind = "cume_dist", col = col,
                offset = 1L, default = NULL))
  }

  # cumsum, cummean, cummin, cummax: single column argument
  if (fn %in% c("cumsum", "cummean", "cummin", "cummax")) {
    col <- as.character(expr[[2]])
    return(list(name = output_name, kind = fn, col = col,
                offset = 1L, default = NULL))
  }

  stop(sprintf("unsupported window function: %s", fn))
}

# Split mutate dots into window specs and regular expressions.
# Returns list(win_specs, win_names, regular_dots, regular_names)
split_window_exprs <- function(dots) {
  dot_names <- names(dots)
  win_specs <- list()
  win_names <- character(0)
  reg_dots <- list()
  reg_names <- character(0)

  for (i in seq_along(dots)) {
    if (is_window_call(dots[[i]])) {
      spec <- parse_window_spec(dots[[i]], dot_names[i])
      win_specs <- c(win_specs, list(spec))
      win_names <- c(win_names, dot_names[i])
    } else {
      reg_dots <- c(reg_dots, list(dots[[i]]))
      reg_names <- c(reg_names, dot_names[i])
    }
  }

  names(reg_dots) <- reg_names
  list(win_specs = win_specs, win_names = win_names,
       regular_dots = reg_dots, regular_names = reg_names)
}

# Create a window node from a vectra_node and window specs
create_window_node <- function(.data, win_specs) {
  key_names <- if (!is.null(.data$.groups)) .data$.groups else character(0)
  new_xptr <- .Call(C_window_node, .data$.node, key_names, win_specs)
  structure(list(.node = new_xptr, .path = .data$.path,
                 .groups = .data$.groups), class = "vectra_node")
}

# raster.R — VECR raster API (Phase 1).
#
# Four primary entry points:
#   vec_write_raster   write a matrix or 3D array to a .vec raster
#   vec_open_raster    open a .vec raster, return a metadata + handle list
#   vec_read_window    decode a window of a chosen band into a matrix
#   vec_extract_points sample band values at (x, y) points
#
# The C side accepts the storage dtype as a string ("f32", "i16", ...) and
# casts R doubles into that dtype on write. On read, every dtype is widened
# to R double; nodata pixels become NA_real_.

#' Write a raster matrix or 3D array to a .vec raster file
#'
#' Writes a row-major raster (one band) or a band-major 3D array (multi-band)
#' to the VECR raster format. Each tile is encoded as a self-describing tdc
#' block (PRED_2D + BYTE_SHUFFLE + LZ).
#'
#' @param x A numeric matrix `c(rows, cols)` for a single band, or a numeric
#'   3D array `c(rows, cols, bands)` for multi-band.
#' @param path Output file path.
#' @param dtype Storage dtype, one of `"f64"`, `"f32"`, `"i8"`, `"u8"`,
#'   `"i16"`, `"u16"`, `"i32"`, `"u32"`, `"i64"`, `"u64"`. Defaults to `"f32"`
#'   for floating-point input — `"f64"` doubles file size with no information
#'   gain for typical climate rasters.
#' @param tile_size Square tile edge in pixels. Default 512.
#' @param extent Numeric vector `c(xmin, ymin, xmax, ymax)`. Used together with
#'   the raster dimensions to derive the geotransform. Either `extent` or
#'   `gt` must be supplied for georeferenced output.
#' @param gt Numeric(6) GDAL-style geotransform. Overrides `extent` if both
#'   are given.
#' @param epsg EPSG code (integer) or 0L for none.
#' @param nodata Nodata value, or `NA_real_` to skip recording one.
#' @param band_names Optional character vector of length equal to the number
#'   of bands.
#' @param compression Compression effort, one of `"fast"` (single spec, fast
#'   encode), `"balanced"` (probe two entropy coders, ~2x encode time), or
#'   `"max"` (probe six candidate specs per tile, slowest encode but
#'   smallest file). Decode cost is unchanged across levels because each
#'   tile records its own codec spec. Default `"fast"`.
#' @return Invisible `NULL`.
#' @export
vec_write_raster <- function(x, path,
                             dtype       = "f32",
                             tile_size   = 512L,
                             extent      = NULL,
                             gt          = NULL,
                             epsg        = 0L,
                             nodata      = NA_real_,
                             band_names  = NULL,
                             compression = c("fast", "balanced", "max")) {

  if (!is.character(path) || length(path) != 1L)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = FALSE)

  if (!is.numeric(x))
    stop("x must be numeric")

  d <- dim(x)
  if (is.null(d) || length(d) < 2L || length(d) > 3L)
    stop("x must be a numeric matrix or 3D array (rows, cols[, bands])")

  rows <- d[1L]
  cols <- d[2L]
  n_bands <- if (length(d) == 3L) d[3L] else 1L

  width  <- as.integer(cols)
  height <- as.integer(rows)

  ## Geotransform derivation. R matrices are arranged with row 1 at the top
  ## of the y-axis (north), matching GeoTIFF convention.
  if (is.null(gt)) {
    if (is.null(extent)) {
      gt <- c(0, 1, 0, 0, 0, 1)  # identity
    } else {
      if (!is.numeric(extent) || length(extent) != 4L)
        stop("extent must be c(xmin, ymin, xmax, ymax)")
      xmin <- extent[1L]; ymin <- extent[2L]
      xmax <- extent[3L]; ymax <- extent[4L]
      xres <- (xmax - xmin) / cols
      yres <- (ymax - ymin) / rows
      gt <- c(xmin, xres, 0, ymax, 0, -yres)
    }
  } else {
    if (!is.numeric(gt) || length(gt) != 6L)
      stop("gt must be a numeric vector of length 6")
  }

  if (!is.null(band_names)) {
    if (!is.character(band_names) || length(band_names) != n_bands)
      stop(sprintf("band_names must be character of length %d", n_bands))
  }

  ## R matrices are column-major; the C side expects band-major then
  ## row-major within each band. Convert via aperm for 3D arrays; for
  ## matrices, t() suffices.
  if (length(d) == 2L) {
    data_vec <- as.vector(t(x))                 # row-major, single band
  } else {
    ## permute so the result is c(cols, rows, bands), then as.vector gives
    ## band-major + within each band cols-vary-fastest = row-major.
    data_vec <- as.vector(aperm(x, c(2L, 1L, 3L)))
  }
  storage.mode(data_vec) <- "double"

  compression <- match.arg(compression)
  comp_code <- switch(compression,
                      fast     = 0L,
                      balanced = 1L,
                      max      = 2L)

  .Call(C_vec_write_raster,
        path,
        data_vec,
        c(width, height, as.integer(n_bands)),
        as.character(dtype),
        as.integer(tile_size),
        as.numeric(gt),
        as.integer(epsg),
        as.numeric(nodata),
        band_names,
        comp_code)

  invisible(NULL)
}

#' Open a .vec raster
#'
#' Lazy open: parses the header and tile index but does not decode any
#' tiles. Returns a list with metadata and an external pointer handle.
#' The pointer is auto-finalized when garbage collected; call
#' `vec_close_raster()` to release earlier.
#'
#' @param path Path to a `.vec` raster file.
#' @return A `vectra_raster` list with elements:
#'   `ptr`, `width`, `height`, `n_bands`, `tile_size`, `dtype`, `gt`,
#'   `epsg`, `nodata`, `band_names`.
#' @export
vec_open_raster <- function(path) {
  if (!is.character(path) || length(path) != 1L)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = TRUE)
  out <- .Call(C_vec_open_raster, path)
  structure(out, class = "vectra_raster")
}

#' Close a .vec raster handle
#'
#' Idempotent. The handle is also auto-released by R's garbage collector.
#' @param r A `vectra_raster` returned by `vec_open_raster()`.
#' @return Invisible `NULL`.
#' @export
vec_close_raster <- function(r) {
  if (!inherits(r, "vectra_raster"))
    stop("r must be a vectra_raster")
  .Call(C_vec_close_raster, r$ptr)
  invisible(NULL)
}

#' Read a window of pixels from a .vec raster
#'
#' Decodes only the tiles overlapping the requested window. Pixels outside
#' the raster extent come back as `NA`.
#'
#' @param r A `vectra_raster` from `vec_open_raster()`.
#' @param band Band index (1-based). Default 1.
#' @param level Overview level — 0 = full resolution, 1 = half, 2 =
#'   quarter, etc. Must be < `r$n_levels` (which is 1 unless
#'   `vec_build_overviews()` has been run on the file).
#' @param cols 1-based column range `c(col_min, col_max)`. Inclusive.
#'   Coordinates are in the chosen level's pixel grid (so at level 1 the
#'   raster is half as wide). Default `c(1, level_width)`.
#' @param rows 1-based row range `c(row_min, row_max)`. Inclusive.
#'   Default `c(1, level_height)`.
#' @return A numeric matrix with `nrow = row_max - row_min + 1` and
#'   `ncol = col_max - col_min + 1`. Nodata pixels become `NA`.
#' @export
vec_read_window <- function(r,
                            band = 1L,
                            level = 0L,
                            cols = NULL,
                            rows = NULL) {
  if (!inherits(r, "vectra_raster"))
    stop("r must be a vectra_raster")
  level <- as.integer(level)
  level_width  <- max(1L, as.integer(ceiling(r$width  / 2^level)))
  level_height <- max(1L, as.integer(ceiling(r$height / 2^level)))
  if (is.null(cols)) cols <- c(1L, level_width)
  if (is.null(rows)) rows <- c(1L, level_height)
  if (length(cols) != 2L || length(rows) != 2L)
    stop("cols and rows must each be length 2")

  .Call(C_vec_read_window,
        r$ptr,
        as.integer(band),
        as.integer(level),
        as.integer(cols[1L]), as.integer(rows[1L]),
        as.integer(cols[2L]), as.integer(rows[2L]))
}

#' Extract band values at (x, y) points from a .vec raster
#'
#' @param r A `vectra_raster` from `vec_open_raster()`.
#' @param x Numeric vector of x coordinates in CRS units.
#' @param y Numeric vector of y coordinates, same length as `x`.
#' @return A `data.frame` with columns `x`, `y`, then one column per band
#'   (named after `r$band_names` if recorded, otherwise `band1`, `band2`,
#'   ...). NA marks pixels outside the raster or matching nodata.
#' @export
vec_extract_points <- function(r, x, y) {
  if (!inherits(r, "vectra_raster"))
    stop("r must be a vectra_raster")
  if (length(x) != length(y))
    stop("x and y must have the same length")
  .Call(C_vec_extract_points,
        r$ptr,
        as.numeric(x), as.numeric(y))
}

#' Build overview pyramids for a .vec raster
#'
#' Appends `n_levels - 1` reduced-resolution copies of the raster to the
#' file. Each level is computed by 2x downsampling the previous level
#' with the chosen kernel. Reading via `vec_read_window(level = L)`
#' picks tiles at level L; the file's `n_levels` is updated in place.
#'
#' @param path Path to a `.vec` raster file. The file is modified in place.
#' @param levels Total levels including level 0 (so `levels = 5` adds
#'   four overviews: levels 1..4). Must be in `[2, 16]`.
#' @param resampling One of `"nearest"`, `"average"`, `"bilinear"`,
#'   `"mode"`, `"gauss"`. `"average"` is the right choice for continuous
#'   rasters; `"mode"` for categorical/land-cover.
#' @param compression Compression effort for the new tiles. Defaults to
#'   `"fast"` because overview tiles are usually one-shot writes.
#' @return Invisible `NULL`.
#' @export
vec_build_overviews <- function(path,
                                levels,
                                resampling  = c("average", "nearest",
                                                "bilinear", "mode", "gauss"),
                                compression = c("fast", "balanced", "max")) {
  if (!is.character(path) || length(path) != 1L)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = TRUE)
  if (!is.numeric(levels) || length(levels) != 1L || levels < 2L)
    stop("levels must be a single integer >= 2")

  resampling <- match.arg(resampling)
  compression <- match.arg(compression)
  comp_code <- switch(compression, fast = 0L, balanced = 1L, max = 2L)

  .Call(C_vec_build_overviews,
        path,
        as.integer(levels),
        resampling,
        comp_code)
  invisible(NULL)
}

#' Export a .vec raster to GeoTIFF
#'
#' Writes the level-0 pixels of a `.vec` raster to a GeoTIFF file. The
#' TIFF inherits dtype, geotransform, EPSG, and nodata from the source.
#' Strip layout; the writer supports `"none"`, `"deflate"`, and `"lzw"`
#' compression. LZW also applies horizontal differencing (Predictor 2)
#' for integer pixel types, which dramatically improves compression on
#' smooth raster data and matches the layout most production GIS tools
#' produce by default. Tiled and BigTIFF output land in a follow-up.
#'
#' @param r Either a path to a `.vec` raster or a `vectra_raster` returned
#'   by `vec_open_raster()`. If a handle is passed it is left open.
#' @param path Output `.tif` path.
#' @param compression One of `"deflate"` (default), `"lzw"`, or `"none"`.
#' @return Invisible `NULL`.
#' @export
vec_to_tiff <- function(r, path,
                        compression = c("deflate", "lzw", "none")) {
  if (!is.character(path) || length(path) != 1L)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = FALSE)

  vec_path <- if (inherits(r, "vectra_raster")) {
    ## Reach into the externalptr's resolved file path. The reader doesn't
    ## currently expose its source path, so require a string for now.
    stop("vec_to_tiff currently requires the source path; pass it directly")
  } else if (is.character(r) && length(r) == 1L) {
    normalizePath(r, mustWork = TRUE)
  } else {
    stop("r must be a .vec path or a vectra_raster")
  }

  compression <- match.arg(compression)
  comp_code <- switch(compression,
                      none    = 0L,
                      deflate = 1L,
                      lzw     = 2L)

  .Call(C_vec_to_tiff, vec_path, path, comp_code)
  invisible(NULL)
}

#' Write a 4D time-cube raster to .vec
#'
#' Each (band, time) combination becomes a stack of tiles tagged with the
#' chosen time stamp. Stamps are stored as int64 in the per-tile index
#' entry; a value of `0` is reserved for "untimed" so this writer remaps
#' any caller-supplied 0 to 1 internally.
#'
#' @param x Numeric 4D array `c(rows, cols, bands, time)`.
#' @param times Numeric/integer vector with `length(times) == dim(x)[4]`,
#'   in the unit of your choice (epoch ms, year, step index).
#' @param path Output `.vec` path.
#' @param dtype Storage dtype (see `vec_write_raster`).
#' @param tile_size Tile edge in pixels.
#' @param layout Tile layout — one of `"image"` (default; one tile per
#'   `(band, time, ty, tx)`, optimal for "give me one full image at time
#'   T" reads) or `"pixel"` (Phase 6b; one tile per `(band, ty, tx)`
#'   holding the full time stack as `[tw*th, n_time]`, optimal for "give
#'   me the time series at pixel `(x, y)`" reads).
#' @param extent,gt,epsg,nodata,band_names,compression Same semantics as
#'   `vec_write_raster()`.
#' @return Invisible `NULL`.
#' @export
vec_write_time_cube <- function(x, times, path,
                                dtype       = "f32",
                                tile_size   = 512L,
                                layout      = c("image", "pixel"),
                                extent      = NULL,
                                gt          = NULL,
                                epsg        = 0L,
                                nodata      = NA_real_,
                                band_names  = NULL,
                                compression = c("fast", "balanced", "max")) {
  if (!is.numeric(x))
    stop("x must be numeric")
  d <- dim(x)
  if (is.null(d) || length(d) != 4L)
    stop("x must be a 4D array c(rows, cols, bands, time)")
  if (length(times) != d[4L])
    stop("length(times) must equal dim(x)[4]")
  if (!is.character(path) || length(path) != 1L)
    stop("path must be a single character string")
  path <- normalizePath(path, mustWork = FALSE)

  rows <- d[1L]; cols <- d[2L]
  width <- as.integer(cols); height <- as.integer(rows)
  n_bands <- as.integer(d[3L]); n_time <- as.integer(d[4L])

  if (is.null(gt)) {
    if (is.null(extent)) {
      gt <- c(0, 1, 0, 0, 0, 1)
    } else {
      xmin <- extent[1L]; ymin <- extent[2L]
      xmax <- extent[3L]; ymax <- extent[4L]
      gt <- c(xmin, (xmax - xmin) / cols, 0, ymax, 0, -(ymax - ymin) / rows)
    }
  }

  ## Permute (rows, cols, bands, time) -> (cols, rows, bands, time) so the
  ## flattened vector is row-major within each band-slice.
  data_vec <- as.vector(aperm(x, c(2L, 1L, 3L, 4L)))
  storage.mode(data_vec) <- "double"

  compression <- match.arg(compression)
  comp_code <- switch(compression, fast = 0L, balanced = 1L, max = 2L)
  layout <- match.arg(layout)

  if (layout == "image") {
    .Call(C_vec_write_time_cube,
          path,
          data_vec,
          c(width, height, n_bands, n_time),
          as.numeric(times),
          as.character(dtype),
          as.integer(tile_size),
          as.numeric(gt),
          as.integer(epsg),
          as.numeric(nodata),
          band_names,
          comp_code)
  } else {
    .Call(C_vec_write_pixel_cube,
          path,
          data_vec,
          c(width, height, n_bands, n_time),
          as.numeric(times),
          as.character(dtype),
          as.integer(tile_size),
          as.numeric(gt),
          as.integer(epsg),
          as.numeric(nodata),
          band_names,
          comp_code)
  }
  invisible(NULL)
}

#' Read the full time series at a single pixel from a .vec time cube
#'
#' Returns a numeric vector of length `n_time` — one value per time step
#' recorded in the file, in ascending time-stamp order.
#'
#' For pixel-major files (written with
#' `vec_write_time_cube(layout = "pixel")`) this is the optimal access
#' pattern: a single tile decode yields all time values for the pixel.
#' For image-major files the reader scans the index for distinct time
#' stamps, decodes one spatial tile per stamp, and extracts the pixel
#' from each — correct but `n_time` slower than the optimal layout.
#'
#' @param r A `vectra_raster` from `vec_open_raster()`.
#' @param x,y Pixel coordinates. Either both `x` and `y` (CRS units; the
#'   geotransform is used to map to col/row) or both `col` and `row`
#'   (1-based pixel indices).
#' @param col,row 1-based pixel coordinates (alternative to x/y).
#' @param band Band index (1-based).
#' @param level Overview level. Default 0.
#' @return A numeric vector of length `n_time`. NA marks pixels outside
#'   the raster or matching nodata. The corresponding time stamps can
#'   be obtained from `vec_raster_times(r, band, level)`.
#' @export
vec_read_pixel_series <- function(r, x = NULL, y = NULL,
                                  col = NULL, row = NULL,
                                  band = 1L, level = 0L) {
  if (!inherits(r, "vectra_raster"))
    stop("r must be a vectra_raster")

  if (!is.null(x) && !is.null(y)) {
    if (length(x) != 1L || length(y) != 1L)
      stop("x and y must be scalars; use vapply() for multi-pixel reads")
    gt <- r$gt
    if (gt[2L] == 0 || gt[6L] == 0)
      stop("geotransform missing scale terms; pass col/row directly")
    col <- as.integer(floor((x - gt[1L]) / gt[2L])) + 1L
    row <- as.integer(floor((y - gt[4L]) / gt[6L])) + 1L
  }
  if (is.null(col) || is.null(row))
    stop("either (x, y) or (col, row) must be supplied")

  .Call(C_vec_read_pixel_series,
        r$ptr,
        as.integer(col), as.integer(row),
        as.integer(band), as.integer(level))
}

#' Distinct time stamps stored in a .vec time cube
#'
#' Returns the ascending vector of time stamps recorded for the given
#' (band, level). Pixel-major files store one consolidated table; image-
#' major files derive the list from the per-tile time field.
#'
#' @param r A `vectra_raster`.
#' @param band 1-based band index.
#' @param level Overview level.
#' @return Numeric vector of stamps (length 0 when the file has no time
#'   information).
#' @export
vec_raster_times <- function(r, band = 1L, level = 0L) {
  if (!inherits(r, "vectra_raster"))
    stop("r must be a vectra_raster")
  out <- .Call(C_vec_raster_times, r$ptr,
               as.integer(band), as.integer(level))
  if (is.null(out)) numeric(0) else out
}

#' Tile layout of an open .vec raster
#'
#' Returns `"image"` (default Phase 6 layout — one tile per
#' `(band, time, ty, tx)`) or `"pixel"` (Phase 6b transpose layout — one
#' tile per `(band, ty, tx)` holding the full time stack).
#'
#' @param r A `vectra_raster`.
#' @return Character(1) `"image"` or `"pixel"`.
#' @export
vec_raster_layout <- function(r) {
  if (!inherits(r, "vectra_raster"))
    stop("r must be a vectra_raster")
  .Call(C_vec_raster_layout, r$ptr)
}

#' Read a single time slice from a .vec time cube
#'
#' Performs a linear scan of the index for tiles with `time == time` and
#' decodes the matching window. The lookup is O(n_tiles) per call — Phase
#' 6's optimized hash-map lookup is a follow-up.
#'
#' @param r A `vectra_raster` from `vec_open_raster()`.
#' @param time Time value to match (numeric/integer).
#' @param band Band index (1-based).
#' @param level Overview level. Default 0.
#' @param cols,rows 1-based ranges, same as `vec_read_window`.
#' @return A numeric matrix.
#' @export
vec_read_time_slice <- function(r, time, band = 1L, level = 0L,
                                cols = NULL, rows = NULL) {
  if (!inherits(r, "vectra_raster"))
    stop("r must be a vectra_raster")
  level <- as.integer(level)
  level_w <- max(1L, as.integer(ceiling(r$width  / 2^level)))
  level_h <- max(1L, as.integer(ceiling(r$height / 2^level)))
  if (is.null(cols)) cols <- c(1L, level_w)
  if (is.null(rows)) rows <- c(1L, level_h)
  ## Remap user 0 -> 1 to match the writer's reservation of 0 for "untimed".
  t <- as.numeric(time)
  if (t == 0) t <- 1
  .Call(C_vec_read_time_slice,
        r$ptr,
        as.integer(band), as.integer(level),
        t,
        as.integer(cols[1L]), as.integer(rows[1L]),
        as.integer(cols[2L]), as.integer(rows[2L]))
}

#' @export
print.vectra_raster <- function(x, ...) {
  cat("<vectra_raster>\n")
  cat(sprintf("  dimensions  : %d cols x %d rows x %d bands\n",
              x$width, x$height, x$n_bands))
  cat(sprintf("  dtype       : %s\n", x$dtype))
  cat(sprintf("  tile_size   : %d\n", x$tile_size))
  cat(sprintf("  geotransform: [%.6g, %.6g, %.6g, %.6g, %.6g, %.6g]\n",
              x$gt[1], x$gt[2], x$gt[3], x$gt[4], x$gt[5], x$gt[6]))
  if (!is.null(x$n_levels) && x$n_levels > 1L)
    cat(sprintf("  n_levels    : %d\n", x$n_levels))
  if (x$epsg > 0L) cat(sprintf("  epsg        : %d\n", x$epsg))
  if (!is.na(x$nodata)) cat(sprintf("  nodata      : %g\n", x$nodata))
  if (!is.null(x$band_names))
    cat(sprintf("  band_names  : %s\n", paste(x$band_names, collapse = ", ")))
  invisible(x)
}

# Phase 1 round-trip tests for vec_write_raster / vec_open_raster /
# vec_read_window / vec_extract_points. These exercise:
#   - lossless float and integer round-trips
#   - tile boundaries (raster larger than tile_size, edge tiles)
#   - nodata handling
#   - multi-band writes
#   - geotransform-based point extraction
#   - terra round-trip when terra is installed (smoke test only — terra is
#     a Suggests dependency)

make_raster <- function(rows, cols, bands = 1L, fn = function(b, r, c) b * 1000 + r * cols + c) {
  if (bands == 1L) {
    m <- matrix(fn(1L, rep(seq_len(rows), times = cols), rep(seq_len(cols), each = rows)),
                nrow = rows, ncol = cols)
    return(m)
  }
  arr <- array(0, dim = c(rows, cols, bands))
  for (b in seq_len(bands)) {
    for (r in seq_len(rows)) {
      for (cc in seq_len(cols)) {
        arr[r, cc, b] <- fn(b, r, cc)
      }
    }
  }
  arr
}

test_that("f64 single-band round-trip preserves values exactly", {
  m <- matrix(rnorm(50 * 50), 50, 50)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_raster(m, tmp, dtype = "f64")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)
  expect_equal(r$width,  50L)
  expect_equal(r$height, 50L)
  expect_equal(r$dtype,  "f64")
  out <- vec_read_window(r)
  expect_equal(dim(out), c(50L, 50L))
  expect_equal(out, m, tolerance = 0)
})

test_that("f32 round-trip is lossless within float32 representable values", {
  m <- matrix(seq(-1, 1, length.out = 64 * 64), 64, 64)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f32")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  out <- vec_read_window(r)
  ## f32 round-trip is lossless against the f32-quantized original.
  m32 <- matrix(as.numeric(as.single(m)), 64, 64)
  expect_equal(out, m32, tolerance = 1e-7)
})

test_that("i16 round-trip is exact", {
  m <- matrix(sample(-1000:1000, 80 * 80, replace = TRUE), 80, 80)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "i16")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  out <- vec_read_window(r)
  expect_equal(out, m, tolerance = 0)
})

test_that("u8 round-trip is exact", {
  m <- matrix(sample(0:255, 200 * 200, replace = TRUE), 200, 200)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "u8")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  out <- vec_read_window(r)
  expect_equal(out, m, tolerance = 0)
})

test_that("raster larger than tile_size with edge tiles round-trips", {
  ## 600x800 > 512 tile -> 2x2 tile grid with edge tiles 88 wide / 288 tall.
  m <- matrix(runif(600 * 800), 600, 800)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f32", tile_size = 512L)
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  expect_equal(r$tile_size, 512L)
  out <- vec_read_window(r)
  expect_equal(out, matrix(as.numeric(as.single(m)), 600, 800),
               tolerance = 1e-7)
})

test_that("partial windows match the corresponding slab of the full raster", {
  m <- matrix(rnorm(300 * 400), 300, 400)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f64", tile_size = 128L)
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })

  ## Window inside a single tile.
  win <- vec_read_window(r, rows = c(50, 100), cols = c(70, 120))
  expect_equal(dim(win), c(51L, 51L))
  expect_equal(win, m[50:100, 70:120], tolerance = 0)

  ## Window spanning multiple tiles.
  win2 <- vec_read_window(r, rows = c(100, 200), cols = c(200, 350))
  expect_equal(dim(win2), c(101L, 151L))
  expect_equal(win2, m[100:200, 200:350], tolerance = 0)
})

test_that("nodata pixels become NA on read", {
  m <- matrix(1:100, 10, 10)
  m[5, 5] <- -9999L
  m[8, 3] <- -9999L
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "i32", nodata = -9999)
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  expect_equal(r$nodata, -9999)
  out <- vec_read_window(r)
  expect_true(is.na(out[5, 5]))
  expect_true(is.na(out[8, 3]))
  expect_equal(out[1, 1], m[1, 1])
})

test_that("multi-band raster round-trips with band names", {
  arr <- array(seq_len(40 * 40 * 3) * 0.5, dim = c(40, 40, 3))
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(arr, tmp, dtype = "f32",
                   band_names = c("red", "green", "blue"))
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  expect_equal(r$n_bands, 3L)
  expect_equal(r$band_names, c("red", "green", "blue"))

  for (b in 1:3) {
    out <- vec_read_window(r, band = b)
    expect_equal(out, matrix(as.numeric(as.single(arr[, , b])), 40, 40),
                 tolerance = 1e-7,
                 info = sprintf("band %d", b))
  }
})

test_that("vec_extract_points returns pixel-center values via the geotransform", {
  ## 5 cols x 3 rows raster, extent (0,0)-(5,3); each pixel is 1x1.
  m <- matrix(1:15, nrow = 3, ncol = 5)   # row-major: row 1 = c(1, 4, 7, 10, 13)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f64",
                   extent = c(0, 0, 5, 3))
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })

  ## Pixel centers: x = 0.5,1.5,..., y = 0.5,1.5,2.5 (with row 1 at the top).
  pts <- vec_extract_points(r,
    x = c(0.5, 4.5, 2.5),
    y = c(2.5, 0.5, 1.5))   # row 1 (top), row 3 (bottom), row 2 (middle)
  expect_equal(pts$band1[1], m[1, 1])  # top-left
  expect_equal(pts$band1[2], m[3, 5])  # bottom-right
  expect_equal(pts$band1[3], m[2, 3])  # middle
})

test_that("points outside the raster come back as NA", {
  m <- matrix(1, 4, 4)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f64", extent = c(0, 0, 4, 4))
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  pts <- vec_extract_points(r, x = c(2, -1, 99), y = c(2, 2, 2))
  expect_equal(pts$band1[1], 1)
  expect_true(is.na(pts$band1[2]))
  expect_true(is.na(pts$band1[3]))
})

test_that("CRS / EPSG round-trips", {
  m <- matrix(0, 8, 8)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f64", epsg = 31287L,
                   extent = c(0, 0, 8, 8))
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  expect_equal(r$epsg, 31287L)
})

test_that("compression='balanced' and 'max' produce identical decoded output", {
  set.seed(7)
  m <- matrix(rnorm(128 * 128), 128, 128)
  ref <- matrix(as.numeric(as.single(m)), 128, 128)

  outs <- list()
  sizes <- integer(3)
  for (i in seq_along(c("fast", "balanced", "max"))) {
    level <- c("fast", "balanced", "max")[i]
    tmp <- tempfile(fileext = ".vec")
    vec_write_raster(m, tmp, dtype = "f32", compression = level)
    sizes[i] <- file.size(tmp)
    r <- vec_open_raster(tmp)
    outs[[level]] <- vec_read_window(r)
    vec_close_raster(r)
    unlink(tmp)
  }

  ## All three levels must round-trip to the same f32-quantized matrix.
  expect_equal(outs$fast,     ref, tolerance = 1e-7)
  expect_equal(outs$balanced, ref, tolerance = 1e-7)
  expect_equal(outs$max,      ref, tolerance = 1e-7)

  ## max <= balanced <= fast for at least *most* inputs (not strictly true on
  ## every random tile, but on ≥ 100kB random gaussian data it should hold).
  ## Check the inequality with a small slack to avoid flake on edge cases.
  expect_lte(sizes[3], sizes[1] + 256)   # max <= fast
})

test_that("compression='max' is non-disastrous on a constant tile", {
  ## A constant raster should hit the predictor's zero-residual fast path
  ## under every codec spec — the resulting file should be tiny under any
  ## level (well under 4 KB for a 256x256 constant tile).
  m <- matrix(7.5, 256, 256)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f32", compression = "max")
  expect_lt(file.size(tmp), 4096)
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  out <- vec_read_window(r)
  expect_equal(out, matrix(as.numeric(as.single(m)), 256, 256))
})

test_that("vec_build_overviews adds N-1 reduced levels", {
  ## A 64x64 constant raster: building 4 levels gives sizes 64, 32, 16, 8.
  m <- matrix(7.0, 64, 64)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_raster(m, tmp, dtype = "f64", tile_size = 32L)

  vec_build_overviews(tmp, levels = 4L, resampling = "average")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)
  expect_equal(r$n_levels, 4L)

  ## Each level should round-trip the (constant) value.
  for (L in 0:3) {
    out <- vec_read_window(r, level = L)
    target_w <- max(1, ceiling(64 / 2^L))
    target_h <- max(1, ceiling(64 / 2^L))
    expect_equal(dim(out), c(target_h, target_w),
                 info = sprintf("level %d", L))
    expect_true(all(out == 7.0), info = sprintf("level %d", L))
  }
})

test_that("average resampling produces correct level-1 means", {
  ## Build a 4x4 raster where each 2x2 block has a distinct value;
  ## level 1 should be the per-block average (= the block's value).
  m <- matrix(0, 4, 4)
  m[1:2, 1:2] <- 1
  m[1:2, 3:4] <- 2
  m[3:4, 1:2] <- 3
  m[3:4, 3:4] <- 4
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f64", tile_size = 64L)
  vec_build_overviews(tmp, levels = 2L, resampling = "average")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })

  l1 <- vec_read_window(r, level = 1L)
  expect_equal(dim(l1), c(2L, 2L))
  expect_equal(l1, matrix(c(1, 3, 2, 4), 2, 2))
})

test_that("nearest resampling takes the top-left pixel of each 2x2", {
  m <- matrix(c(1, 2, 3, 4,
                5, 6, 7, 8,
                9, 10, 11, 12,
                13, 14, 15, 16),
              nrow = 4, byrow = TRUE)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f64", tile_size = 64L)
  vec_build_overviews(tmp, levels = 2L, resampling = "nearest")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })

  l1 <- vec_read_window(r, level = 1L)
  expect_equal(l1, matrix(c(1, 9, 3, 11), 2, 2))
})

test_that("mode resampling picks the most-frequent value", {
  m <- matrix(c(5, 5, 8, 8,
                5, 7, 8, 9,
                3, 3, 1, 1,
                3, 4, 2, 1),
              nrow = 4, byrow = TRUE)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "i32", tile_size = 64L)
  vec_build_overviews(tmp, levels = 2L, resampling = "mode")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  l1 <- vec_read_window(r, level = 1L)
  ## Top-left 2x2 = {5,5,5,7} -> mode 5
  ## Top-right     = {8,8,8,9} -> 8
  ## Bottom-left   = {3,3,3,4} -> 3
  ## Bottom-right  = {1,1,2,1} -> 1
  expect_equal(l1, matrix(c(5L, 3L, 8L, 1L), 2, 2))
})

test_that("vec_read_window rejects an out-of-range level", {
  m <- matrix(0, 8, 8)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f64")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })
  expect_error(vec_read_window(r, level = 5L), "level")
})

test_that("vec_build_overviews refuses to add fewer levels than already exist", {
  m <- matrix(0, 32, 32)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f64")
  vec_build_overviews(tmp, levels = 3L)
  expect_error(vec_build_overviews(tmp, levels = 2L), "already")
  unlink(tmp)
})

test_that("parallel tile decode matches serial decode (sufficient tiles)", {
  ## 800x800 with tile_size 128 -> 7x7 = 49 tiles. With OMP_NUM_THREADS>1
  ## the read path uses the parallel branch; result must match.
  set.seed(2)
  m <- matrix(rnorm(800 * 800), 800, 800)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_raster(m, tmp, dtype = "f64", tile_size = 128L)
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)
  out <- vec_read_window(r)
  expect_equal(out, m, tolerance = 0)
})

test_that("Phase 5b mmap reader: many parallel tile decodes, no race", {
  ## Phase 5b: with the mmap reader the inner OpenMP loop reads tile
  ## bytes via memcpy from the mapped region, no fread critical section.
  ## Build a raster with enough tiles that >= 4 threads each get real
  ## work (256 tiles via 16x16 grid at tile_size 64), then read multiple
  ## windows back-to-back. Compares to the serial reference each time
  ## so any concurrent corruption / off-by-one in the offset arithmetic
  ## would be caught by an exact-match assertion.
  prev <- Sys.getenv("OMP_NUM_THREADS", unset = NA)
  Sys.setenv(OMP_NUM_THREADS = "4")
  on.exit({
    if (is.na(prev)) Sys.unsetenv("OMP_NUM_THREADS")
    else             Sys.setenv(OMP_NUM_THREADS = prev)
  }, add = TRUE)

  set.seed(42)
  rows <- 1024L; cols <- 1024L
  m <- matrix(rnorm(rows * cols), rows, cols)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp), add = TRUE)
  vec_write_raster(m, tmp, dtype = "f64", tile_size = 64L)

  r <- vec_open_raster(tmp)
  on.exit(vec_close_raster(r), add = TRUE)

  ## Full-window read drives the OpenMP parallel decode with 256 tiles.
  for (rep in 1:5) {
    out <- vec_read_window(r)
    expect_equal(out, m, tolerance = 0)
  }

  ## Stagger overlapping sub-windows across multiple invocations to
  ## stress the per-tile mmap pointer arithmetic.
  windows <- list(
    list(cols = c(  1L,  512L), rows = c(  1L,  512L)),
    list(cols = c( 33L,  700L), rows = c(200L, 1024L)),
    list(cols = c( 64L,  960L), rows = c( 64L,  960L)),
    list(cols = c(100L,  800L), rows = c( 50L,  900L))
  )
  for (w in windows) {
    sub <- vec_read_window(r, cols = w$cols, rows = w$rows)
    ref <- m[w$rows[1]:w$rows[2], w$cols[1]:w$cols[2]]
    expect_equal(dim(sub), dim(ref))
    expect_equal(sub, ref, tolerance = 0)
  }
})

test_that("vec_to_tiff round-trips a single-band raster via terra", {
  skip_if_not_installed("terra")
  m <- matrix(seq(-1, 1, length.out = 30 * 40), 30, 40)
  vec_path <- tempfile(fileext = ".vec")
  tiff_path <- tempfile(fileext = ".tif")
  on.exit(unlink(c(vec_path, tiff_path)))

  vec_write_raster(m, vec_path, dtype = "f32",
                   extent = c(0, 0, 40, 30), epsg = 4326L)
  vec_to_tiff(vec_path, tiff_path, compression = "deflate")

  tr <- terra::rast(tiff_path)
  expect_equal(terra::nlyr(tr), 1L)
  expect_equal(terra::ncol(tr), 40L)
  expect_equal(terra::nrow(tr), 30L)
  ## Compare a sample of pixels.
  px <- terra::values(tr)
  expect_equal(as.numeric(px), as.numeric(as.single(t(m))),
               tolerance = 1e-6)
})

test_that("vec_to_tiff exports a 4-band raster terra reads correctly", {
  skip_if_not_installed("terra")
  arr <- array(0, dim = c(20, 25, 4))
  for (b in 1:4) arr[, , b] <- matrix(seq_len(20 * 25) * b, 20, 25)
  vec_path <- tempfile(fileext = ".vec")
  tiff_path <- tempfile(fileext = ".tif")
  on.exit(unlink(c(vec_path, tiff_path)))

  vec_write_raster(arr, vec_path, dtype = "f32",
                   extent = c(0, 0, 25, 20))
  vec_to_tiff(vec_path, tiff_path, compression = "deflate")

  tr <- terra::rast(tiff_path)
  expect_equal(terra::nlyr(tr), 4L)
  for (b in 1:4) {
    layer <- terra::values(tr[[b]])
    expect_equal(as.numeric(layer),
                 as.numeric(as.single(t(arr[, , b]))),
                 tolerance = 1e-6,
                 info = sprintf("band %d", b))
  }
})

test_that("vec_to_tiff propagates nodata", {
  skip_if_not_installed("terra")
  m <- matrix(1:25, 5, 5)
  m[3, 3] <- -9999L
  vec_path <- tempfile(fileext = ".vec")
  tiff_path <- tempfile(fileext = ".tif")
  on.exit(unlink(c(vec_path, tiff_path)))

  vec_write_raster(m, vec_path, dtype = "i32", nodata = -9999)
  vec_to_tiff(vec_path, tiff_path, compression = "none")

  tr <- terra::rast(tiff_path)
  vals <- terra::values(tr)[, 1]
  ## terra reads NoData as NA when GDAL_NODATA is recognised.
  expect_true(any(is.na(vals)))
})

test_that("vec_to_tiff LZW + Predictor 2 round-trips an integer raster", {
  skip_if_not_installed("terra")
  ## A smooth integer ramp — Predictor 2 should slash the byte stream
  ## entropy (most differences become 0/1) before LZW sees it.
  m <- matrix(seq_len(40 * 30), 30, 40)
  vec_path  <- tempfile(fileext = ".vec")
  tiff_path <- tempfile(fileext = ".tif")
  on.exit(unlink(c(vec_path, tiff_path)))

  vec_write_raster(m, vec_path, dtype = "i32",
                   extent = c(0, 0, 40, 30), epsg = 4326L)
  vec_to_tiff(vec_path, tiff_path, compression = "lzw")

  tr <- terra::rast(tiff_path)
  expect_equal(terra::nlyr(tr), 1L)
  expect_equal(terra::ncol(tr), 40L)
  expect_equal(terra::nrow(tr), 30L)

  px <- as.numeric(terra::values(tr))
  expect_equal(px, as.numeric(t(m)))
})

test_that("vec_to_tiff LZW handles a multi-band 16-bit raster", {
  skip_if_not_installed("terra")
  arr <- array(0L, dim = c(20, 25, 3))
  for (b in 1:3) arr[, , b] <- matrix(seq_len(20 * 25) * b, 20, 25)
  vec_path  <- tempfile(fileext = ".vec")
  tiff_path <- tempfile(fileext = ".tif")
  on.exit(unlink(c(vec_path, tiff_path)))

  vec_write_raster(arr, vec_path, dtype = "i16",
                   extent = c(0, 0, 25, 20))
  vec_to_tiff(vec_path, tiff_path, compression = "lzw")

  tr <- terra::rast(tiff_path)
  expect_equal(terra::nlyr(tr), 3L)
  for (b in 1:3) {
    layer <- terra::values(tr[[b]])
    expect_equal(as.numeric(layer), as.numeric(t(arr[, , b])),
                 info = sprintf("band %d", b))
  }
})

test_that("vec_to_tiff LZW (no predictor) round-trips a float raster", {
  skip_if_not_installed("terra")
  ## Float pixel types: writer emits Predictor = 1 (none), since byte-wise
  ## subtraction of float bit patterns is not meaningful. terra/GDAL
  ## must still read the LZW stream back correctly.
  m <- matrix(seq(-1, 1, length.out = 30 * 40), 30, 40)
  vec_path  <- tempfile(fileext = ".vec")
  tiff_path <- tempfile(fileext = ".tif")
  on.exit(unlink(c(vec_path, tiff_path)))

  vec_write_raster(m, vec_path, dtype = "f32",
                   extent = c(0, 0, 40, 30), epsg = 4326L)
  vec_to_tiff(vec_path, tiff_path, compression = "lzw")

  tr <- terra::rast(tiff_path)
  expect_equal(terra::nlyr(tr), 1L)
  px <- as.numeric(terra::values(tr))
  expect_equal(px, as.numeric(as.single(t(m))), tolerance = 1e-6)
})

test_that("vec_to_tiff LZW emits Compression=5 and Predictor=2 for integers", {
  ## Pure on-disk tag check — does not depend on terra. We poke at the IFD
  ## for tags 259 (Compression) and 317 (Predictor) and confirm the writer
  ## set them as advertised.
  m <- matrix(seq_len(64 * 64), 64, 64)
  vec_path  <- tempfile(fileext = ".vec")
  tiff_path <- tempfile(fileext = ".tif")
  on.exit(unlink(c(vec_path, tiff_path)))

  vec_write_raster(m, vec_path, dtype = "i32")
  vec_to_tiff(vec_path, tiff_path, compression = "lzw")

  raw <- readBin(tiff_path, what = "raw",
                 n = file.info(tiff_path)$size)
  ## Header: "II", magic 42, IFD offset (LE u32 at byte 4).
  expect_equal(rawToChar(raw[1:2]), "II")
  ifd_off <- sum(as.integer(raw[5:8]) * c(1L, 256L, 65536L, 16777216L)) + 1L

  n_entries <- sum(as.integer(raw[ifd_off:(ifd_off + 1L)]) *
                     c(1L, 256L))
  base <- ifd_off + 2L
  found_compression <- found_predictor <- NA_integer_
  for (i in seq_len(n_entries)) {
    e <- ifd_off + 2L + (i - 1L) * 12L
    tag <- sum(as.integer(raw[e:(e + 1L)]) * c(1L, 256L))
    val <- sum(as.integer(raw[(e + 8L):(e + 9L)]) * c(1L, 256L))
    if (tag == 259L) found_compression <- val
    if (tag == 317L) found_predictor   <- val
  }
  expect_equal(found_compression, 5L) ## LZW
  expect_equal(found_predictor,   2L) ## horizontal differencing
})

test_that("time cube round-trips per slice", {
  ## 3 time steps x 2 bands x 8 rows x 10 cols.
  arr <- array(0, dim = c(8, 10, 2, 3))
  for (t in 1:3) {
    for (b in 1:2) {
      arr[, , b, t] <- matrix(seq_len(80) * (10 * t + b), 8, 10)
    }
  }
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_time_cube(arr, times = c(2020, 2021, 2022), path = tmp,
                      dtype = "f64")

  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)
  expect_equal(r$n_bands, 2L)

  for (t in c(2020, 2021, 2022)) {
    for (b in 1:2) {
      out <- vec_read_time_slice(r, time = t, band = b)
      expect_equal(out, arr[, , b, t - 2019],
                   info = sprintf("time=%d band=%d", t, b))
    }
  }
})

test_that("vec_read_time_slice errors on a missing time stamp", {
  arr <- array(0, dim = c(4, 4, 1, 2))
  arr[, , 1, 1] <- matrix(1:16, 4, 4)
  arr[, , 1, 2] <- matrix(101:116, 4, 4)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_time_cube(arr, times = c(1, 2), path = tmp, dtype = "i32")

  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)
  expect_error(vec_read_time_slice(r, time = 99), "no tiles match")
})

test_that("terra can ingest pixel values via point extraction (smoke test)", {
  skip_if_not_installed("terra")
  ## Build a known raster, write to .vec, sample at known points and
  ## compare to terra's interpretation of the same matrix at those points.
  set.seed(1)
  m <- matrix(rnorm(20 * 30), nrow = 20, ncol = 30)
  tmp <- tempfile(fileext = ".vec")
  vec_write_raster(m, tmp, dtype = "f32", extent = c(0, 0, 30, 20))
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) })

  tr <- terra::rast(m, extent = terra::ext(0, 30, 0, 20))
  pts <- data.frame(x = c(1.5, 15.5, 28.5),
                    y = c(19.5, 10.5, 0.5))
  ours <- vec_extract_points(r, pts$x, pts$y)$band1
  theirs <- terra::extract(tr, as.matrix(pts))[, 1]
  ## Both pipelines should report the f32-quantized matrix at pixel centers.
  expect_equal(ours, theirs, tolerance = 1e-6)
})

# ---------------------------------------------------------------------------
# Phase 6b — pixel-time-series transpose layout
# ---------------------------------------------------------------------------

# Helper: build a 4D cube where each (row, col, band, time) cell has a
# unique distinguishable value. We use the same formula in every test so
# round-trip equality checks are easy to interpret.
make_cube <- function(rows, cols, bands, n_time) {
  arr <- array(0, dim = c(rows, cols, bands, n_time))
  for (t in seq_len(n_time)) {
    for (b in seq_len(bands)) {
      arr[, , b, t] <- matrix(seq_len(rows * cols) * (10 * t + b),
                              rows, cols)
    }
  }
  arr
}

test_that("pixel-major time cube round-trips per slice", {
  arr <- make_cube(rows = 8, cols = 10, bands = 2, n_time = 3)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_time_cube(arr, times = c(2020, 2021, 2022), path = tmp,
                      dtype = "f64", layout = "pixel")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)

  expect_equal(vec_raster_layout(r), "pixel")
  expect_equal(vec_raster_times(r), c(2020, 2021, 2022))

  for (t_val in c(2020, 2021, 2022)) {
    for (b in 1:2) {
      out <- vec_read_time_slice(r, time = t_val, band = b)
      expect_equal(out, arr[, , b, t_val - 2019],
                   info = sprintf("time=%d band=%d", t_val, b))
    }
  }
})

test_that("pixel-major default layout is 'image' (no opt-in -> image)", {
  arr <- make_cube(rows = 6, cols = 6, bands = 1, n_time = 2)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_time_cube(arr, times = c(1, 2), path = tmp, dtype = "f64")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)
  expect_equal(vec_raster_layout(r), "image")
})

test_that("vec_read_pixel_series matches vec_read_time_slice (pixel layout)", {
  arr <- make_cube(rows = 7, cols = 9, bands = 2, n_time = 4)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_time_cube(arr, times = 1:4, path = tmp, dtype = "f64",
                      layout = "pixel")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)

  ## For every pixel, read_pixel_series should match the column of values
  ## you'd assemble by calling vec_read_time_slice four times.
  for (b in 1:2) {
    for (rr in c(1L, 4L, 7L)) {
      for (cc in c(1L, 5L, 9L)) {
        series <- vec_read_pixel_series(r, col = cc, row = rr, band = b)
        ## Direct check against the source array.
        expect_equal(series, arr[rr, cc, b, ],
                     info = sprintf("band=%d row=%d col=%d", b, rr, cc))
        ## Cross-check: assemble the same vector from time-slice reads.
        slice <- numeric(4)
        for (t in 1:4) {
          slice[t] <- vec_read_time_slice(r, time = t, band = b)[rr, cc]
        }
        expect_equal(series, slice, info = "slice consistency")
      }
    }
  }
})

test_that("vec_read_pixel_series falls back to image-major files", {
  ## For image-major, the function decodes one tile per distinct time
  ## stamp and extracts a single sample. Slow-but-correct.
  arr <- make_cube(rows = 5, cols = 5, bands = 1, n_time = 3)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_time_cube(arr, times = c(100, 200, 300), path = tmp,
                      dtype = "f64", layout = "image")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)

  expect_equal(vec_raster_layout(r), "image")
  series <- vec_read_pixel_series(r, col = 3L, row = 2L, band = 1L)
  expect_equal(series, arr[2, 3, 1, ])
  expect_equal(vec_raster_times(r), c(100, 200, 300))
})

test_that("pixel-major edge tiles round-trip with non-square remainders", {
  ## tile_size 4 against a 7x10 raster -> 2x3 tiles with edge tile widths
  ## of 4, 4, 2 and edge tile heights of 4, 3.
  arr <- make_cube(rows = 7, cols = 10, bands = 1, n_time = 2)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_time_cube(arr, times = c(1, 2), path = tmp,
                      dtype = "f64", tile_size = 4L, layout = "pixel")
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)

  ## Pixel in the bottom-right edge tile.
  series <- vec_read_pixel_series(r, col = 10L, row = 7L, band = 1L)
  expect_equal(series, arr[7, 10, 1, ])

  ## Full-raster window at t=2.
  out <- vec_read_time_slice(r, time = 2)
  expect_equal(out, arr[, , 1, 2])
})

test_that("pixel-major nodata round-trips through the time stack", {
  arr <- make_cube(rows = 5, cols = 5, bands = 1, n_time = 3)
  arr[3, 3, 1, 2] <- -9999    # nodata at single (row, col, time)
  arr[1, 1, 1, ] <- -9999     # nodata at all times for one pixel
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_time_cube(arr, times = c(1, 2, 3), path = tmp,
                      dtype = "i32", layout = "pixel", nodata = -9999)
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)

  ## The single-cell nodata becomes NA on read; the other two stamps
  ## round-trip the original cube values.
  s_centre <- vec_read_pixel_series(r, col = 3L, row = 3L, band = 1L)
  expect_equal(s_centre[1], arr[3, 3, 1, 1])
  expect_true(is.na(s_centre[2]))
  expect_equal(s_centre[3], arr[3, 3, 1, 3])

  ## All-nodata pixel returns NA at every time step.
  s_corner <- vec_read_pixel_series(r, col = 1L, row = 1L, band = 1L)
  expect_true(all(is.na(s_corner)))
})

test_that("pixel-major and image-major produce equivalent slices", {
  arr <- make_cube(rows = 6, cols = 8, bands = 2, n_time = 3)
  tmp_img <- tempfile(fileext = ".vec")
  tmp_pix <- tempfile(fileext = ".vec")
  on.exit(unlink(c(tmp_img, tmp_pix)))
  vec_write_time_cube(arr, times = c(10, 20, 30), path = tmp_img,
                      dtype = "f64", layout = "image")
  vec_write_time_cube(arr, times = c(10, 20, 30), path = tmp_pix,
                      dtype = "f64", layout = "pixel")
  ri <- vec_open_raster(tmp_img)
  rp <- vec_open_raster(tmp_pix)
  on.exit({ vec_close_raster(ri); vec_close_raster(rp); unlink(c(tmp_img, tmp_pix)) },
          add = TRUE)

  for (t_val in c(10, 20, 30)) {
    for (b in 1:2) {
      img <- vec_read_time_slice(ri, time = t_val, band = b)
      pix <- vec_read_time_slice(rp, time = t_val, band = b)
      expect_equal(img, pix, info = sprintf("t=%d b=%d", t_val, b))
    }
  }
})

test_that("vec_read_pixel_series via x/y coordinates uses the geotransform", {
  arr <- make_cube(rows = 4, cols = 5, bands = 1, n_time = 2)
  tmp <- tempfile(fileext = ".vec")
  on.exit(unlink(tmp))
  vec_write_time_cube(arr, times = c(1, 2), path = tmp,
                      dtype = "f64", layout = "pixel",
                      extent = c(0, 0, 5, 4))
  r <- vec_open_raster(tmp)
  on.exit({ vec_close_raster(r); unlink(tmp) }, add = TRUE)

  ## Pixel center (2.5, 1.5) -> col = 3, row = 3 (R indexing) since y=1.5
  ## maps to row = floor((1.5 - 4)/-1) + 1 = floor(2.5) + 1 = 3.
  ## col = floor((2.5 - 0)/1) + 1 = 3.
  series <- vec_read_pixel_series(r, x = 2.5, y = 1.5, band = 1L)
  expect_equal(series, arr[3, 3, 1, ])
})

# Recovery and streaming-invariance tests for mask().
#
# Ground truth is sf centre-in-polygon (the package's own pixel-centre
# convention, shared with zonal()). mask must reproduce it exactly, stay equal
# across tile sizes, and sit inside terra's any-overlap mask where terra exists.

wr <- function(m, ext = c(0, 0, ncol(m), nrow(m)), tile = 512L) {
  f <- tempfile(fileext = ".vec")
  vec_write_raster(m, f, dtype = "f64", extent = ext, tile_size = tile)
  f
}

# Centre-in-polygon truth for an n x n grid of unit cells over [0, n]^2.
centre_cover <- function(n, poly) {
  cx <- rep(seq(0.5, n - 0.5, 1), each = n)
  cy <- rep(seq(n - 0.5, 0.5, -1), times = n)
  pts <- sf::st_as_sf(data.frame(x = cx, y = cy), coords = c("x", "y"))
  matrix(lengths(sf::st_intersects(pts, poly)) > 0, n, n, byrow = TRUE)
}

test_that("mask matches sf centre-in-polygon exactly", {
  skip_if_not_installed("sf")
  vals <- matrix(1:100, 10, 10, byrow = TRUE)
  fv <- wr(vals, ext = c(0, 0, 10, 10), tile = 4L)
  on.exit(unlink(fv))
  disc <- sf::st_buffer(sf::st_sfc(sf::st_point(c(5, 5))), 3)

  m <- mask(fv, disc)
  cov <- centre_cover(10, disc)
  expect_equal(!is.na(m), cov, ignore_attr = TRUE)
  # kept pixels keep their original value
  expect_equal(m[cov], vals[cov])
})

test_that("mask inverse keeps exactly the complement", {
  skip_if_not_installed("sf")
  vals <- matrix(1:100, 10, 10, byrow = TRUE)
  fv <- wr(vals, ext = c(0, 0, 10, 10), tile = 4L)
  on.exit(unlink(fv))
  disc <- sf::st_buffer(sf::st_sfc(sf::st_point(c(5, 5))), 3)

  m <- mask(fv, disc)
  inv <- mask(fv, disc, inverse = TRUE)
  expect_equal(is.na(inv), !is.na(m), ignore_attr = TRUE)
})

test_that("mask is invariant to tile size (streaming invariance)", {
  skip_if_not_installed("sf")
  vals <- matrix(seq_len(20 * 20), 20, 20, byrow = TRUE)
  poly <- sf::st_buffer(sf::st_sfc(sf::st_point(c(10, 12))), 5)
  f1 <- wr(vals, ext = c(0, 0, 20, 20), tile = 4L)
  f2 <- wr(vals, ext = c(0, 0, 20, 20), tile = 20L)
  on.exit(unlink(c(f1, f2)))

  expect_equal(mask(f1, poly), mask(f2, poly))
})

test_that("mask kept cells are a subset of terra's any-overlap mask", {
  skip_if_not_installed("sf")
  skip_if_not_installed("terra")
  vals <- matrix(1:100, 10, 10, byrow = TRUE)
  fv <- wr(vals, ext = c(0, 0, 10, 10), tile = 4L)
  on.exit(unlink(fv))
  disc <- sf::st_buffer(sf::st_sfc(sf::st_point(c(5, 5))), 3)

  m <- mask(fv, disc)
  tr <- terra::rast(vals, extent = terra::ext(0, 10, 0, 10))
  tmv <- matrix(terra::values(terra::mask(tr, terra::vect(disc), touches = TRUE)),
                10, 10, byrow = TRUE)
  # every cell I keep, terra (any-overlap) also keeps
  expect_true(all(!is.na(tmv)[!is.na(m)]))
})

test_that("mask preserves all bands and streams to a .vec", {
  skip_if_not_installed("sf")
  arr <- array(0, dim = c(8, 8, 2))
  arr[, , 1] <- matrix(1:64, 8, 8, byrow = TRUE)
  arr[, , 2] <- matrix(64:1, 8, 8, byrow = TRUE)
  f <- tempfile(fileext = ".vec"); out <- tempfile(fileext = ".vec")
  vec_write_raster(arr, f, dtype = "f64", extent = c(0, 0, 8, 8), tile_size = 4L)
  on.exit(unlink(c(f, out)))
  poly <- sf::st_buffer(sf::st_sfc(sf::st_point(c(4, 4))), 2.5)

  h <- mask(f, poly, path = out)
  expect_s3_class(h, "vectra_raster")
  expect_equal(h$n_bands, 2L)
  b1 <- vec_read_window(h, band = 1L); b2 <- vec_read_window(h, band = 2L)
  vec_close_raster(h)
  expect_equal(is.na(b1), is.na(b2))
})

test_that("mask rejects a non-polygon mask", {
  fv <- wr(matrix(1, 4, 4))
  on.exit(unlink(fv))
  expect_error(mask(fv, data.frame(a = 1)), "sf or sfc")
})

# Recovery and streaming-invariance tests for rasterize().
#
# The reference grid is an independent, vectorised binning of the points in R
# (tapply over cell keys); the streamed C accumulator must reproduce it exactly.
# A separate terra cell-for-cell check guards the geotransform convention, and a
# multi-row-group scan proves the streamed result equals the single-batch one.

# Independent reference: bin points into an nr x nc grid (row 1 = north).
ref_grid <- function(pts, ext, nr, nc, field = NULL, fun = "count",
                     bg = NA_real_) {
  xmin <- ext[1]; xmax <- ext[3]; ymax <- ext[4]
  xres <- (xmax - xmin) / nc
  yres <- (ext[4] - ext[2]) / nr
  col <- floor((pts$x - xmin) / xres)      # 0-based, west->east
  row <- floor((ymax - pts$y) / yres)      # 0-based, north->south
  inb <- col >= 0 & col < nc & row >= 0 & row < nr
  m <- matrix(bg, nr, nc)
  idx <- which(inb)
  if (!length(idx)) return(m)
  rr <- row[idx] + 1L; cc <- col[idx] + 1L
  key <- (cc - 1L) * nr + rr               # column-major matrix index
  if (is.null(field)) {
    agg <- tapply(rep(1L, length(key)), key, sum)
  } else {
    v <- pts[[field]][idx]
    keep <- !is.na(v)
    key <- key[keep]; v <- v[keep]
    fn <- switch(fun, sum = sum, mean = mean, min = min, max = max)
    agg <- tapply(v, key, fn)
  }
  m[as.integer(names(agg))] <- as.numeric(agg)
  m
}

vtr_points <- function(pts, batch_size = NULL) {
  f <- tempfile(fileext = ".vtr")
  write_vtr(pts, f, batch_size = batch_size)
  f
}

test_that("count recovers a hand-binned grid", {
  set.seed(11)
  n <- 3000
  pts <- data.frame(x = runif(n, 0, 10), y = runif(n, 0, 10))
  f <- vtr_points(pts); on.exit(unlink(f))

  got <- tbl(f) |> rasterize(extent = c(0, 0, 10, 10), dims = c(10, 10))
  ref <- ref_grid(pts, c(0, 0, 10, 10), 10, 10)

  expect_equal(dim(got), c(10L, 10L))
  expect_equal(got, ref, ignore_attr = TRUE)
  expect_equal(sum(got), n)               # every point lands in a cell
})

test_that("sum, mean, min, max recover the binned field", {
  set.seed(12)
  n <- 4000
  pts <- data.frame(x = runif(n, 0, 8), y = runif(n, 0, 8), z = rnorm(n))
  f <- vtr_points(pts); on.exit(unlink(f))
  ext <- c(0, 0, 8, 8)

  for (fn in c("sum", "mean", "min", "max")) {
    got <- tbl(f) |>
      rasterize(extent = ext, dims = c(8, 8), field = "z", fun = fn)
    ref <- ref_grid(pts, ext, 8, 8, field = "z", fun = fn)
    expect_equal(got, ref, ignore_attr = TRUE, info = fn)
  }
})

test_that("NA field values are skipped (na.rm)", {
  pts <- data.frame(x = c(0.5, 0.5, 0.5), y = c(0.5, 0.5, 0.5),
                    z = c(1, NA, 3))
  f <- vtr_points(pts); on.exit(unlink(f))
  # single 1x1 cell over [0,1]^2
  s <- tbl(f) |> rasterize(extent = c(0, 0, 1, 1), dims = c(1, 1),
                           field = "z", fun = "sum")
  m <- tbl(f) |> rasterize(extent = c(0, 0, 1, 1), dims = c(1, 1),
                           field = "z", fun = "mean")
  cnt <- tbl(f) |> rasterize(extent = c(0, 0, 1, 1), dims = c(1, 1))
  expect_equal(as.numeric(s), 4)          # 1 + 3
  expect_equal(as.numeric(m), 2)          # mean of 1, 3
  expect_equal(as.numeric(cnt), 3)        # count tallies all points
})

test_that("empty cells take the background value", {
  pts <- data.frame(x = 0.5, y = 0.5)     # x west, y south -> bottom-left cell
  f <- vtr_points(pts); on.exit(unlink(f))
  got <- tbl(f) |> rasterize(extent = c(0, 0, 2, 2), dims = c(2, 2),
                             background = -1)
  expect_equal(got[2, 1], 1)              # row 2 = southern row got the point
  expect_equal(sum(got == -1), 3L)        # the other three are background
})

test_that("streamed multi-batch equals single-batch", {
  set.seed(13)
  n <- 5000
  pts <- data.frame(x = runif(n, 0, 20), y = runif(n, 0, 20), z = rnorm(n))
  f1 <- vtr_points(pts)                          # default one row group
  f2 <- vtr_points(pts, batch_size = 211)        # many small row groups
  on.exit(unlink(c(f1, f2)))

  for (fn in c("count", "sum", "mean", "min", "max")) {
    fld <- if (fn == "count") NULL else "z"
    a <- tbl(f1) |> rasterize(extent = c(0, 0, 20, 20), dims = c(16, 16),
                              field = fld, fun = fn)
    b <- tbl(f2) |> rasterize(extent = c(0, 0, 20, 20), dims = c(16, 16),
                              field = fld, fun = fn)
    expect_equal(a, b, ignore_attr = TRUE, info = fn)
  }
})

test_that("res and dims describe the same grid", {
  set.seed(14)
  pts <- data.frame(x = runif(500, 0, 10), y = runif(500, 0, 10))
  f <- vtr_points(pts); on.exit(unlink(f))
  by_dims <- tbl(f) |> rasterize(extent = c(0, 0, 10, 10), dims = c(20, 20))
  by_res  <- tbl(f) |> rasterize(extent = c(0, 0, 10, 10), res = 0.5)
  expect_equal(dim(by_res), c(20L, 20L))
  expect_equal(by_dims, by_res, ignore_attr = TRUE)
})

test_that("geom (sf point) input matches coords input", {
  skip_if_not_installed("sf")
  set.seed(15)
  n <- 1500
  pts <- data.frame(x = runif(n, 0, 10), y = runif(n, 0, 10), z = rnorm(n))

  f_xy <- vtr_points(pts)
  g <- sf::st_as_binary(
    sf::st_geometry(sf::st_as_sf(pts, coords = c("x", "y"))), hex = TRUE)
  f_geom <- vtr_points(data.frame(z = pts$z, geometry = g, stringsAsFactors = FALSE))
  on.exit(unlink(c(f_xy, f_geom)))

  a <- tbl(f_xy)  |> rasterize(extent = c(0, 0, 10, 10), dims = c(10, 10),
                               field = "z", fun = "mean")
  b <- tbl(f_geom) |> rasterize(extent = c(0, 0, 10, 10), dims = c(10, 10),
                                geom = "geometry", field = "z", fun = "mean")
  expect_equal(a, b, ignore_attr = TRUE)
})

test_that("rasterize matches terra::rasterize cell-for-cell", {
  skip_if_not_installed("terra")
  set.seed(16)
  n <- 4000
  pts <- data.frame(x = runif(n, 0, 12), y = runif(n, 0, 12), z = rnorm(n))
  f <- vtr_points(pts); on.exit(unlink(f))

  tmpl <- terra::rast(nrows = 12, ncols = 12,
                      xmin = 0, xmax = 12, ymin = 0, ymax = 12)
  vt <- terra::vect(pts, geom = c("x", "y"))

  # count
  got_c <- tbl(f) |> rasterize(extent = c(0, 0, 12, 12), dims = c(12, 12))
  ter_c <- terra::as.matrix(terra::rasterize(vt, tmpl, fun = "length"),
                            wide = TRUE)
  # terra leaves empty cells NA; rasterize() background is NA too.
  expect_true(all((is.na(ter_c) & is.na(got_c)) |
                  (ter_c == got_c)))

  # mean of z
  got_m <- tbl(f) |> rasterize(extent = c(0, 0, 12, 12), dims = c(12, 12),
                               field = "z", fun = "mean")
  ter_m <- terra::as.matrix(terra::rasterize(vt, tmpl, field = "z",
                                             fun = "mean"), wide = TRUE)
  both <- !is.na(ter_m) & !is.na(got_m)
  expect_true(all(is.na(ter_m) == is.na(got_m)))
  expect_equal(got_m[both], ter_m[both], tolerance = 1e-9)
})

test_that("path output writes a readable .vec raster", {
  set.seed(17)
  pts <- data.frame(x = runif(2000, 0, 10), y = runif(2000, 0, 10))
  f <- vtr_points(pts); on.exit(unlink(f))
  out <- tempfile(fileext = ".vec"); on.exit(unlink(out), add = TRUE)

  r <- tbl(f) |> rasterize(extent = c(0, 0, 10, 10), dims = c(10, 10),
                           path = out, dtype = "f64")
  expect_s3_class(r, "vectra_raster")
  expect_equal(r$width, 10L)
  expect_equal(r$height, 10L)

  mem <- ref_grid(pts, c(0, 0, 10, 10), 10, 10)
  disk <- vec_read_window(r)
  # empty cells: NA in memory, NaN->NA on read; both NA
  both <- !is.na(mem) & !is.na(disk)
  expect_equal(disk[both], mem[both], tolerance = 1e-9)
  vec_close_raster(r)
})

test_that("input validation", {
  pts <- data.frame(x = runif(10), y = runif(10))
  f <- vtr_points(pts); on.exit(unlink(f))
  expect_error(rasterize(pts), "vectra_node")
  expect_error(tbl(f) |> rasterize(extent = c(0, 0, 1, 1), dims = c(2, 2),
                                   fun = "sum"),
               "needs a `field`")
  expect_error(tbl(f) |> rasterize(extent = c(0, 0, 1, 1), dims = c(2, 2),
                                   coords = c("lon", "lat")),
               "coords column")
  expect_error(tbl(f) |> rasterize(),
               "template|extent")
})

# Recovery and streaming-invariance tests for rast_calc().
#
# The reference is the same expression evaluated on the whole resident matrices.
# The streamed strip pass must reproduce it exactly and stay equal across tile
# sizes (a streamed path must equal the resident path).

wr <- function(m, ext = c(0, 0, ncol(m), nrow(m)), tile = 512L) {
  f <- tempfile(fileext = ".vec")
  vec_write_raster(m, f, dtype = "f64", extent = ext, tile_size = tile)
  f
}

test_that("rast_calc recovers NDVI computed on resident matrices", {
  set.seed(1)
  nir <- matrix(runif(20 * 20, 10, 90), 20, 20)
  red <- matrix(runif(20 * 20, 5, 50), 20, 20)
  fn <- wr(nir, tile = 8L); fr <- wr(red, tile = 8L)
  on.exit(unlink(c(fn, fr)))

  got <- rast_calc(list(nir = fn, red = fr), (nir - red) / (nir + red))
  ref <- (nir - red) / (nir + red)
  expect_equal(unclass(got), unclass(ref), tolerance = 1e-9, ignore_attr = TRUE)
})

test_that("rast_calc is invariant to tile size (streaming invariance)", {
  set.seed(2)
  a <- matrix(rnorm(48 * 48), 48, 48)
  fa_small <- wr(a, tile = 8L); fa_big <- wr(a, tile = 48L)
  on.exit(unlink(c(fa_small, fa_big)))

  small <- rast_calc(list(a = fa_small), a * 3 - 2)
  big   <- rast_calc(list(a = fa_big),   a * 3 - 2)
  expect_equal(small, big, tolerance = 1e-12)
})

test_that("rast_calc reclassify matches a resident cut()", {
  set.seed(3)
  dem <- matrix(runif(16 * 16, 0, 100), 16, 16)
  fd <- wr(dem, tile = 4L)
  on.exit(unlink(fd))
  brks <- c(-Inf, 25, 50, 75, Inf)
  got <- rast_calc(list(dem = fd), as.integer(cut(dem, brks)))
  ref <- matrix(as.integer(cut(as.vector(dem), brks)), 16, 16)
  expect_equal(unclass(got), unclass(ref), ignore_attr = TRUE)
})

test_that("rast_calc round-trips through a .vec sink", {
  m <- matrix(1:36, 6, 6, byrow = TRUE)
  fm <- wr(m, tile = 3L)
  out <- tempfile(fileext = ".vec")
  on.exit(unlink(c(fm, out)))
  h <- rast_calc(list(m = fm), m + 100, path = out)
  expect_s3_class(h, "vectra_raster")
  back <- vec_read_window(h, band = 1L)
  vec_close_raster(h)
  expect_equal(back, m + 100, ignore_attr = TRUE)
})

test_that("rast_calc rejects misaligned rasters and bad expressions", {
  a <- wr(matrix(1, 4, 4)); b <- wr(matrix(1, 5, 5))
  on.exit(unlink(c(a, b)))
  expect_error(rast_calc(list(a = a, b = b), a + b), "share dimensions")
  expect_error(rast_calc(list(matrix(1)), 1), "named list")
  expect_error(rast_calc(list(a = a), c(1, 2, 3)), "one value per cell")
})

# Recovery and streaming-invariance tests for zonal().
#
# The reference is an independent base-R grouped summary (tapply over zone ids)
# of the same pixels. The streamed strip fold must reproduce it exactly, equal
# across tile sizes (streaming invariance), and match terra::zonal cell-for-cell
# where terra is available.

# Independent per-zone reference over two aligned matrices (row-major pixels).
ref_zonal <- function(vals, zone, fun, na.rm = TRUE) {
  v <- as.vector(vals); z <- as.vector(zone)
  keep <- !is.na(z)
  v <- v[keep]; z <- z[keep]
  if (na.rm) {
    ok <- !is.na(v); v <- v[ok]; z <- z[ok]
  }
  labs <- sort(unique(z))
  stat <- function(f) vapply(labs, function(k) {
    vv <- v[z == k]
    if (!na.rm && any(is.na(vv)) && f != "count") return(NA_real_)
    vv <- vv[!is.na(vv)]
    switch(f,
      count = length(vv), sum = sum(vv), mean = mean(vv),
      min = min(vv), max = max(vv),
      sd = if (length(vv) < 2) NA_real_ else stats::sd(vv))
  }, numeric(1))
  out <- data.frame(zone = as.numeric(labs))
  for (f in fun) out[[f]] <- stat(f)
  out
}

write_raster <- function(m, ext = c(0, 0, ncol(m), nrow(m)), tile = 512L) {
  f <- tempfile(fileext = ".vec")
  vec_write_raster(m, f, dtype = "f64", extent = ext, tile_size = tile)
  f
}

test_that("zonal recovers a hand-computed per-zone summary", {
  vals <- matrix(1:16, 4, 4, byrow = TRUE)
  zone <- matrix(c(1, 1, 2, 2, 1, 1, 2, 2,
                   3, 3, 4, 4, 3, 3, 4, 4), 4, 4, byrow = TRUE)
  fv <- write_raster(vals); fz <- write_raster(zone)
  on.exit(unlink(c(fv, fz)))

  got <- zonal(fv, fz, fun = c("mean", "sum", "count", "min", "max"))
  ref <- ref_zonal(vals, zone, c("mean", "sum", "count", "min", "max"))
  expect_equal(got, ref, tolerance = 1e-9)
  # zone 1 holds the four top-left values 1,2,5,6
  expect_equal(got$sum[got$zone == 1], 1 + 2 + 5 + 6)
})

test_that("sd matches the resident standard deviation", {
  set.seed(21)
  vals <- matrix(rnorm(40 * 40), 40, 40)
  zone <- matrix(sample(1:5, 40 * 40, replace = TRUE), 40, 40)
  fv <- write_raster(vals); fz <- write_raster(zone)
  on.exit(unlink(c(fv, fz)))

  got <- zonal(fv, fz, fun = "sd")
  ref <- ref_zonal(vals, zone, "sd")
  expect_equal(got, ref, tolerance = 1e-8)
})

test_that("streamed strips equal the single-strip result", {
  set.seed(22)
  vals <- matrix(rnorm(60 * 50), 60, 50)
  zone <- matrix(sample(1:7, 60 * 50, replace = TRUE), 60, 50)
  # small tiles -> many strips, one giant tile -> a single strip
  fv_s <- write_raster(vals, tile = 16L); fz_s <- write_raster(zone, tile = 16L)
  fv_1 <- write_raster(vals, tile = 512L); fz_1 <- write_raster(zone, tile = 512L)
  on.exit(unlink(c(fv_s, fz_s, fv_1, fz_1)))

  funs <- c("mean", "sum", "count", "min", "max", "sd")
  a <- zonal(fv_s, fz_s, fun = funs)
  b <- zonal(fv_1, fz_1, fun = funs)
  expect_equal(a, b, tolerance = 1e-9)
})

test_that("nodata pixels are skipped under na.rm = TRUE and taint under FALSE", {
  vals <- matrix(c(1, 2, NA, 4,
                   5, 6, 7, 8), 2, 4, byrow = TRUE)
  zone <- matrix(c(1, 1, 2, 2,
                   1, 1, 2, 2), 2, 4, byrow = TRUE)
  fv <- write_raster(vals); fz <- write_raster(zone)
  on.exit(unlink(c(fv, fz)))

  # zone 2 spans columns 3-4: values NA, 4, 7, 8
  keep <- zonal(fv, fz, fun = c("mean", "count"), na.rm = TRUE)
  expect_equal(keep$mean[keep$zone == 2], mean(c(4, 7, 8)))  # NA dropped
  expect_equal(keep$count[keep$zone == 2], 3)               # only non-NA cells

  prop <- zonal(fv, fz, fun = c("mean", "count"), na.rm = FALSE)
  expect_true(is.na(prop$mean[prop$zone == 2]))             # NA propagates
  expect_equal(prop$count[prop$zone == 2], 3)               # count still non-NA
})

test_that("zonal matches terra::zonal cell-for-cell", {
  skip_if_not_installed("terra")
  set.seed(23)
  vals <- matrix(rnorm(50 * 50), 50, 50)
  zone <- matrix(sample(1:6, 50 * 50, replace = TRUE), 50, 50)
  fv <- write_raster(vals); fz <- write_raster(zone)
  on.exit(unlink(c(fv, fz)))

  rv <- terra::rast(vals, extent = terra::ext(0, 50, 0, 50))
  rz <- terra::rast(zone, extent = terra::ext(0, 50, 0, 50))
  for (fn in c("mean", "sum", "min", "max")) {
    got <- zonal(fv, fz, fun = fn)
    ter <- terra::zonal(rv, rz, fun = fn, na.rm = TRUE)
    names(ter) <- c("zone", "val")
    ter <- ter[order(ter$zone), ]
    expect_equal(got[[fn]], ter$val, tolerance = 1e-9, info = fn)
  }
})

test_that("sf polygon zones assign each pixel by its centre", {
  skip_if_not_installed("sf")
  # 4x4 grid over [0,4]^2; split into left (x<2) and right (x>=2) halves.
  vals <- matrix(1:16, 4, 4, byrow = TRUE)
  fv <- write_raster(vals, ext = c(0, 0, 4, 4))
  on.exit(unlink(fv))

  sq <- function(x0, x1) sf::st_polygon(list(rbind(
    c(x0, 0), c(x1, 0), c(x1, 4), c(x0, 4), c(x0, 0))))
  zones <- sf::st_sf(zid = c(10, 20),
                     geometry = sf::st_sfc(sq(0, 2), sq(2, 4)))

  got <- zonal(fv, zones, fun = c("mean", "count"), zone_field = "zid")
  # left two columns -> zone 10, right two -> zone 20; 8 cells each
  expect_equal(got$zone, c(10, 20))
  expect_equal(got$count, c(8, 8))
  left  <- mean(as.vector(vals[, 1:2]))
  right <- mean(as.vector(vals[, 3:4]))
  expect_equal(got$mean, c(left, right), tolerance = 1e-9)
})

test_that("a vectra_raster handle works as well as a path", {
  vals <- matrix(1:9, 3, 3, byrow = TRUE)
  zone <- matrix(c(1, 1, 2, 1, 1, 2, 3, 3, 3), 3, 3, byrow = TRUE)
  fv <- write_raster(vals); fz <- write_raster(zone)
  on.exit(unlink(c(fv, fz)))
  rv <- vec_open_raster(fv); rz <- vec_open_raster(fz)
  on.exit({ vec_close_raster(rv); vec_close_raster(rz) }, add = TRUE)

  got <- zonal(rv, rz, fun = "sum")
  ref <- ref_zonal(vals, zone, "sum")
  expect_equal(got, ref, tolerance = 1e-9)
})

test_that("input validation", {
  vals <- matrix(1:4, 2, 2)
  fv <- write_raster(vals); on.exit(unlink(fv))
  expect_error(zonal(fv, 42), "vectra_raster.*path.*sf")
  expect_error(zonal(fv, fv, fun = "bogus"), "should be one of|arg")
  # mismatched dimensions
  big <- write_raster(matrix(1:9, 3, 3)); on.exit(unlink(big), add = TRUE)
  expect_error(zonal(fv, big), "dimensions")
})

# focal() -- streamed moving-window statistics over a VECR raster, checked
# against hand-computed windows, terra::focal ground truth, and the resident
# (single-strip) path. The streamed strip pass must equal the resident result
# exactly: divergence there is a bug, not "more accurate".

write_raster <- function(m, ext = c(0, 0, ncol(m), nrow(m)), tile = 512L,
                         dtype = "f64") {
  f <- tempfile(fileext = ".vec")
  vec_write_raster(m, f, dtype = dtype, extent = ext, tile_size = tile)
  f
}

test_that("focal recovers a hand-computed 3x3 window", {
  m <- matrix(1:25, 5, 5, byrow = TRUE)
  f <- write_raster(m); on.exit(unlink(f))

  s <- focal(f, w = matrix(1, 3, 3), fun = "sum", na.rm = TRUE)
  # interior cell (3,3) value 13: 3x3 block sums to 9 * 13 = 117 (linear grid)
  expect_equal(s[3, 3], sum(m[2:4, 2:4]))
  expect_equal(s[3, 3], 117)
  mn <- focal(f, w = matrix(1, 3, 3), fun = "mean", na.rm = TRUE)
  expect_equal(mn[3, 3], mean(m[2:4, 2:4]))
  # corner (1,1) under na.rm = TRUE sees only the in-raster 2x2 block
  expect_equal(s[1, 1], sum(m[1:2, 1:2]))
})

test_that("na.rm = FALSE makes edge windows NA, na.rm = TRUE fills them", {
  m <- matrix(1:25, 5, 5, byrow = TRUE)
  f <- write_raster(m); on.exit(unlink(f))

  keep <- focal(f, w = matrix(1, 3, 3), fun = "mean", na.rm = FALSE)
  expect_true(is.na(keep[1, 1]))                  # window runs off the edge
  expect_false(is.na(keep[3, 3]))                 # fully interior
  fill <- focal(f, w = matrix(1, 3, 3), fun = "mean", na.rm = TRUE)
  expect_false(is.na(fill[1, 1]))
  expect_equal(fill[3, 3], keep[3, 3])            # interior identical
})

test_that("weights scale sum and NA weights drop cells", {
  m <- matrix(1:25, 5, 5, byrow = TRUE)
  f <- write_raster(m); on.exit(unlink(f))

  w <- matrix(c(0, 1, 0, 1, 1, 1, 0, 1, 0), 3, 3, byrow = TRUE)  # plus-shape
  s <- focal(f, w = w, fun = "sum", na.rm = TRUE)
  # cell (3,3)=13: plus neighbours are 8,12,13,14,18
  expect_equal(s[3, 3], 8 + 12 + 13 + 14 + 18)

  wna <- w; wna[wna == 0] <- NA                   # NA weight == out of window
  sna <- focal(f, w = wna, fun = "sum", na.rm = TRUE)
  expect_equal(sna[3, 3], s[3, 3])

  w2 <- matrix(2, 3, 3)
  expect_equal(focal(f, w = w2, fun = "sum")[3, 3],
               2 * sum(m[2:4, 2:4]))
})

test_that("a single odd integer is shorthand for a square window", {
  m <- matrix(1:25, 5, 5, byrow = TRUE)
  f <- write_raster(m); on.exit(unlink(f))
  a <- focal(f, w = 3, fun = "sum")
  b <- focal(f, w = matrix(1, 3, 3), fun = "sum")
  expect_equal(a, b)
})

test_that("streamed strips equal the single-strip result", {
  set.seed(41)
  m <- matrix(rnorm(37 * 29), 37, 29)
  f_s <- write_raster(m, tile = 8L)               # many tile-rows
  f_1 <- write_raster(m, tile = 512L)             # a single strip
  on.exit(unlink(c(f_s, f_1)))

  for (fn in c("sum", "mean", "min", "max", "sd", "median")) {
    a <- focal(f_s, w = matrix(1, 3, 3), fun = fn, na.rm = TRUE)
    b <- focal(f_1, w = matrix(1, 3, 3), fun = fn, na.rm = TRUE)
    expect_equal(a, b, tolerance = 1e-12, info = fn)
  }
  # a 5x3 window crosses more than one tile-row halo
  a <- focal(f_s, w = matrix(1, 5, 3), fun = "mean")
  b <- focal(f_1, w = matrix(1, 5, 3), fun = "mean")
  expect_equal(a, b, tolerance = 1e-12)
})

test_that("streamed-to-.vec output equals the in-memory result", {
  set.seed(42)
  m <- matrix(rnorm(20 * 16), 20, 16)
  f <- write_raster(m, tile = 8L); on.exit(unlink(f))
  mem <- focal(f, w = matrix(1, 3, 3), fun = "mean", na.rm = TRUE)

  of <- tempfile(fileext = ".vec"); on.exit(unlink(of), add = TRUE)
  h <- focal(f, w = matrix(1, 3, 3), fun = "mean", na.rm = TRUE,
             path = of, dtype = "f64")
  on.exit(vec_close_raster(h), add = TRUE)
  disk <- vec_read_window(h, band = 1)
  expect_equal(max(abs(mem - disk), na.rm = TRUE), 0)
  expect_true(all(is.na(mem) == is.na(disk)))
})

test_that("focal matches terra::focal", {
  skip_if_not_installed("terra")
  set.seed(43)
  m <- matrix(rnorm(40 * 32), 40, 32)
  f <- write_raster(m, tile = 16L); on.exit(unlink(f))
  r <- terra::rast(m, extent = terra::ext(0, 32, 0, 40))

  for (fn in c("sum", "mean", "min", "max", "sd")) {
    got <- focal(f, w = matrix(1, 3, 3), fun = fn, na.rm = TRUE)
    ter <- matrix(terra::values(terra::focal(r, w = matrix(1, 3, 3),
                                             fun = fn, na.rm = TRUE)),
                  nrow = 40, byrow = TRUE)
    expect_equal(got, ter, tolerance = 1e-9, ignore_attr = TRUE, info = fn)
  }
  # na.rm = FALSE: both leave the border NA
  got <- focal(f, w = matrix(1, 3, 3), fun = "mean", na.rm = FALSE)
  ter <- matrix(terra::values(terra::focal(r, w = matrix(1, 3, 3),
                                           fun = "mean", na.rm = FALSE)),
                nrow = 40, byrow = TRUE)
  expect_equal(got, ter, tolerance = 1e-9, ignore_attr = TRUE)
  expect_true(all(is.na(got) == is.na(ter)))
})

test_that("a vectra_raster handle works as well as a path", {
  m <- matrix(1:16, 4, 4, byrow = TRUE)
  f <- write_raster(m); on.exit(unlink(f))
  rr <- vec_open_raster(f); on.exit(vec_close_raster(rr), add = TRUE)
  expect_equal(focal(rr, w = matrix(1, 3, 3), fun = "sum"),
               focal(f,  w = matrix(1, 3, 3), fun = "sum"))
})

test_that("focal validates inputs", {
  m <- matrix(1:16, 4, 4)
  f <- write_raster(m); on.exit(unlink(f))
  expect_error(focal(f, w = matrix(1, 2, 2)), "odd")
  expect_error(focal(f, w = 4), "odd")
  expect_error(focal(f, fun = "bogus"), "should be one of|arg")
})

# Recovery and streaming-invariance tests for proximity().
#
# Ground truth is the brute-force Euclidean distance from every cell centre to
# the nearest feature cell centre, and terra::distance where terra exists.
# proximity must reproduce both and stay equal across tile sizes.

wr <- function(m, ext = c(0, 0, ncol(m), nrow(m)), tile = 512L) {
  f <- tempfile(fileext = ".vec")
  vec_write_raster(m, f, dtype = "f64", extent = ext, tile_size = tile)
  f
}

# Brute-force distance from each cell centre to the nearest TRUE cell centre.
# `feat` is an H x W logical matrix (row 1 north); xres/yres are the cell sizes.
brute_prox <- function(feat, xres, yres) {
  H <- nrow(feat); W <- ncol(feat)
  fi <- which(feat, arr.ind = TRUE)
  out <- matrix(NA_real_, H, W)
  if (!nrow(fi)) return(out)
  for (i in seq_len(H)) for (j in seq_len(W)) {
    dx <- (j - fi[, 2L]) * xres
    dy <- (i - fi[, 1L]) * yres
    out[i, j] <- sqrt(min(dx * dx + dy * dy))
  }
  out
}

test_that("proximity matches brute-force Euclidean distance", {
  H <- 12L; W <- 12L
  m <- matrix(NA_real_, H, W)
  feats <- rbind(c(2L, 3L), c(7L, 9L), c(11L, 2L))
  for (k in seq_len(nrow(feats))) m[feats[k, 1L], feats[k, 2L]] <- 1
  f <- wr(m, ext = c(0, 0, W, H), tile = 4L)
  on.exit(unlink(f))

  d <- proximity(f)
  truth <- brute_prox(!is.na(m), 1, 1)
  expect_equal(d, truth, ignore_attr = TRUE, tolerance = 1e-6)
})

test_that("proximity scales x and y by resolution (anisotropic)", {
  H <- 10L; W <- 10L
  m <- matrix(NA_real_, H, W); m[5L, 5L] <- 1
  # extent makes 2:1 cells: xres = 2, yres = 1.
  f <- wr(m, ext = c(0, 0, 2 * W, H), tile = 4L)
  on.exit(unlink(f))

  d <- proximity(f)
  truth <- brute_prox(!is.na(m), 2, 1)
  expect_equal(d, truth, ignore_attr = TRUE, tolerance = 1e-6)
})

test_that("target selects which values are features", {
  m <- matrix(0, 6L, 6L); m[2L, 2L] <- 7; m[5L, 5L] <- 7; m[1L, 6L] <- 3
  f <- wr(m, ext = c(0, 0, 6, 6), tile = 3L)
  on.exit(unlink(f))

  d <- proximity(f, target = 7)
  truth <- brute_prox(m == 7, 1, 1)
  expect_equal(d, truth, ignore_attr = TRUE, tolerance = 1e-6)
})

test_that("proximity matches terra::distance cell-for-cell", {
  skip_if_not_installed("terra")
  set.seed(3)
  H <- 15L; W <- 15L
  m <- matrix(NA_real_, H, W)
  m[sample(H * W, 5L)] <- 1
  f <- wr(m, ext = c(0, 0, W, H), tile = 5L)
  on.exit(unlink(f))

  d <- proximity(f)
  # A projected CRS keeps terra::distance planar-Euclidean (lon/lat is geodesic).
  tr <- terra::rast(m, extent = terra::ext(0, W, 0, H), crs = "EPSG:32633")
  tdv <- matrix(terra::values(terra::distance(tr)), H, W, byrow = TRUE)
  expect_equal(d, tdv, ignore_attr = TRUE, tolerance = 1e-6)
})

test_that("proximity is invariant to tile size (streaming invariance)", {
  set.seed(5)
  H <- 16L; W <- 16L
  m <- matrix(NA_real_, H, W)
  m[cbind(sample(H, 4L), sample(W, 4L))] <- 1
  f1 <- wr(m, ext = c(0, 0, W, H), tile = 4L)
  f2 <- wr(m, ext = c(0, 0, W, H), tile = 16L)
  on.exit(unlink(c(f1, f2)))

  expect_equal(proximity(f1), proximity(f2))
})

test_that("proximity streams to a .vec and round-trips", {
  H <- 12L; W <- 12L
  m <- matrix(NA_real_, H, W); m[3L, 4L] <- 1; m[9L, 10L] <- 1
  f <- wr(m, ext = c(0, 0, W, H), tile = 4L)
  out <- tempfile(fileext = ".vec")
  on.exit(unlink(c(f, out)))

  h <- proximity(f, path = out, dtype = "f64")
  expect_s3_class(h, "vectra_raster")
  got <- vec_read_window(h)
  vec_close_raster(h)
  truth <- brute_prox(!is.na(m), 1, 1)
  expect_equal(got, truth, ignore_attr = TRUE, tolerance = 1e-6)
})

test_that("no feature anywhere yields all NA", {
  m <- matrix(NA_real_, 8L, 8L)
  f <- wr(m, ext = c(0, 0, 8, 8), tile = 4L)
  on.exit(unlink(f))
  expect_true(all(is.na(proximity(f))))

  m2 <- matrix(5, 8L, 8L)
  f2 <- wr(m2, ext = c(0, 0, 8, 8), tile = 4L)
  on.exit(unlink(f2), add = TRUE)
  expect_true(all(is.na(proximity(f2, target = 999))))
})

# terrain() -- Horn 3x3 DEM derivatives over a streamed VECR raster, checked
# against terra::terrain / terra::shade and against the resident (single-strip)
# path. The streamed strip pass must equal the resident result exactly.

write_dem <- function(m, tile = 512L, dtype = "f64") {
  f <- tempfile(fileext = ".vec")
  vec_write_raster(m, f, dtype = dtype,
                   extent = c(0, 0, ncol(m), nrow(m)), tile_size = tile)
  f
}

# A tilted surface with curvature so slope/aspect vary and no interior cell is
# flat (aspect undefined only on flats).
tilted <- function(nr, nc) {
  outer(seq_len(nr), seq_len(nc),
        function(r, c) 10 + 2 * c + 0.7 * r + 0.3 * r * c)
}

test_that("TPI and roughness recover hand-computed 3x3 summaries", {
  m <- matrix(c(1, 2, 3, 4,
                5, 6, 7, 8,
                9, 10, 12, 12,
                13, 14, 15, 16), 4, 4, byrow = TRUE)
  f <- write_dem(m); on.exit(unlink(f))

  tpi <- terrain(f, v = "TPI")
  # cell (2,2)=6: neighbours 1,2,3,5,7,9,10,12 -> mean 6.125 -> TPI -0.125
  nb <- c(m[1, 1:3], m[2, c(1, 3)], m[3, 1:3])
  expect_equal(tpi[2, 2], m[2, 2] - mean(nb))

  rough <- terrain(f, v = "roughness")
  blk <- m[1:3, 1:3]
  expect_equal(rough[2, 2], max(blk) - min(blk))
})

test_that("slope, aspect, TPI, roughness, TRI match terra::terrain", {
  skip_if_not_installed("terra")
  m <- tilted(12, 10)
  f <- write_dem(m, tile = 16L); on.exit(unlink(f))
  r <- terra::rast(m, extent = terra::ext(0, 10, 0, 12))

  for (v in c("slope", "aspect", "TPI", "roughness", "TRI")) {
    got <- terrain(f, v = v, unit = "degrees")
    ter <- matrix(terra::values(terra::terrain(r, v = v, unit = "degrees",
                                               neighbors = 8)),
                  nrow = 12, byrow = TRUE)
    expect_equal(got, ter, tolerance = 1e-7, ignore_attr = TRUE, info = v)
    expect_true(all(is.na(got) == is.na(ter)), info = v)
  }
})

test_that("slope honours the radians unit", {
  skip_if_not_installed("terra")
  m <- tilted(10, 9)
  f <- write_dem(m); on.exit(unlink(f))
  r <- terra::rast(m, extent = terra::ext(0, 9, 0, 10))
  got <- terrain(f, v = "slope", unit = "radians")
  ter <- matrix(terra::values(terra::terrain(r, v = "slope", unit = "radians")),
                nrow = 10, byrow = TRUE)
  expect_equal(got, ter, tolerance = 1e-7, ignore_attr = TRUE)
})

test_that("hillshade matches terra::shade", {
  skip_if_not_installed("terra")
  m <- tilted(11, 11)
  f <- write_dem(m); on.exit(unlink(f))
  r <- terra::rast(m, extent = terra::ext(0, 11, 0, 11))
  sl <- terra::terrain(r, v = "slope", unit = "radians")
  as <- terra::terrain(r, v = "aspect", unit = "radians")
  ter <- matrix(terra::values(terra::shade(sl, as, angle = 45, direction = 315)),
                nrow = 11, byrow = TRUE)
  got <- terrain(f, v = "hillshade", azimuth = 315, altitude = 45)
  expect_equal(got, ter, tolerance = 1e-7, ignore_attr = TRUE)
})

test_that("return follows input: matrix for one v, named list for several", {
  m <- tilted(6, 6)
  f <- write_dem(m); on.exit(unlink(f))
  one <- terrain(f, v = "slope")
  expect_true(is.matrix(one))
  many <- terrain(f, v = c("slope", "aspect", "TPI"))
  expect_type(many, "list")
  expect_equal(names(many), c("slope", "aspect", "TPI"))
  expect_equal(many$slope, one)
})

test_that("streamed strips equal the single-strip result", {
  m <- tilted(40, 33)
  f_s <- write_dem(m, tile = 8L)
  f_1 <- write_dem(m, tile = 512L)
  on.exit(unlink(c(f_s, f_1)))
  for (v in c("slope", "aspect", "hillshade", "TPI", "roughness", "TRI")) {
    a <- terrain(f_s, v = v)
    b <- terrain(f_1, v = v)
    expect_equal(a, b, tolerance = 1e-12, info = v)
  }
})

test_that("streamed-to-.vec output equals the in-memory result", {
  m <- tilted(22, 18)
  f <- write_dem(m, tile = 8L); on.exit(unlink(f))
  mem <- terrain(f, v = c("slope", "aspect"))

  of <- tempfile(fileext = ".vec"); on.exit(unlink(of), add = TRUE)
  h <- terrain(f, v = c("slope", "aspect"), path = of, dtype = "f64")
  on.exit(vec_close_raster(h), add = TRUE)
  expect_equal(h$n_bands, 2L)
  expect_equal(h$band_names, c("slope", "aspect"))
  s1 <- vec_read_window(h, band = 1)
  s2 <- vec_read_window(h, band = 2)
  expect_equal(max(abs(mem$slope - s1), na.rm = TRUE), 0)
  expect_equal(max(abs(mem$aspect - s2), na.rm = TRUE), 0)
})

test_that("terrain validates inputs", {
  m <- tilted(4, 4)
  f <- write_dem(m); on.exit(unlink(f))
  expect_error(terrain(f, v = "bogus"), "should be one of|arg")
  expect_error(terrain(f, unit = "grads"), "should be one of|arg")
})

# warp() -- resample/reproject a streamed VECR raster, checked against an
# analytic linear field (exact under bilinear/cubic), terra::project ground
# truth, and the resident (single-strip) path. The streamed strip pass must
# equal the resident result exactly: divergence there is a bug.

write_raster <- function(m, ext = c(0, 0, ncol(m), nrow(m)), tile = 512L,
                         dtype = "f64", epsg = 0L) {
  f <- tempfile(fileext = ".vec")
  vec_write_raster(m, f, dtype = dtype, extent = ext, tile_size = tile,
                   epsg = epsg)
  f
}

# Pixel-centre coordinates of a warp result matrix from its geotransform.
centres <- function(got) {
  gt <- attr(got, "gt")
  nr <- nrow(got); nc <- ncol(got)
  x <- gt[1L] + (col(got) - 0.5) * gt[2L]
  y <- gt[4L] + (row(got) - 0.5) * gt[6L]
  list(x = x, y = y)
}

# A spatial linear field is reproduced exactly by bilinear and cubic sampling.
lin_field <- function(x, y) 12 + 0.7 * x - 0.4 * y

# Build a source raster whose cell (i, j) holds the linear field at its centre.
lin_raster <- function(nr, nc, tile = 512L) {
  m <- matrix(0, nr, nc)
  for (i in seq_len(nr)) for (j in seq_len(nc)) {
    xc <- j - 0.5; yc <- nr - (i - 0.5)
    m[i, j] <- lin_field(xc, yc)
  }
  write_raster(m, tile = tile)
}

test_that("warping a raster onto its own grid reproduces it", {
  set.seed(11)
  m <- matrix(rnorm(20 * 16), 20, 16)
  f  <- write_raster(m, tile = 8L)
  ft <- write_raster(m, tile = 8L)
  on.exit(unlink(c(f, ft)))
  for (meth in c("near", "bilinear", "cubic")) {
    w <- warp(f, ft, method = meth)
    expect_equal(max(abs(w - m), na.rm = TRUE), 0, info = meth)
  }
  # near has no halo, so the identity warp is fully defined.
  expect_false(anyNA(warp(f, ft, method = "near")))
})

test_that("bilinear and cubic recover an analytic linear field", {
  f <- lin_raster(24, 20, tile = 8L); on.exit(unlink(f))
  # A finer, shifted target grid inside the source so kernels stay in-bounds.
  spec <- list(extent = c(2, 2, 18, 22), res = 0.5)
  for (meth in c("bilinear", "cubic")) {
    got <- warp(f, spec, method = meth)
    cc <- centres(got)
    exp <- lin_field(cc$x, cc$y)
    int <- !is.na(got)
    expect_equal(max(abs(got[int] - exp[int])), 0, tolerance = 1e-9, info = meth)
  }
})

test_that("nearest neighbour picks the containing source cell", {
  m <- matrix(1:36, 6, 6, byrow = TRUE)
  f <- write_raster(m); on.exit(unlink(f))
  # 2x2 target over the same extent (3x coarsen): target centres land on the
  # source centres x = 1.5/4.5, y = 4.5/1.5 -> source rows 2/5, cols 2/5.
  got <- warp(f, list(extent = c(0, 0, 6, 6), dims = c(2, 2)), method = "near")
  expect_equal(dim(got), c(2L, 2L))
  expect_equal(got[1, 1], m[2, 2])   # NW
  expect_equal(got[1, 2], m[2, 5])   # NE
  expect_equal(got[2, 1], m[5, 2])   # SW
  expect_equal(got[2, 2], m[5, 5])   # SE
})

test_that("targets beyond the source extent come back NA", {
  m <- matrix(1:16, 4, 4, byrow = TRUE)
  f <- write_raster(m); on.exit(unlink(f))
  got <- warp(f, list(extent = c(-4, -4, 8, 8), res = 1), method = "near")
  expect_true(is.na(got[1, 1]))                 # NW corner is off the source
  expect_false(is.na(got[6, 6]))                # inside the source footprint
})

test_that("near, bilinear and cubic match terra::project", {
  skip_if_not_installed("terra")
  z <- outer(1:30, 1:24, function(r, c) 5 + 0.3 * c + 0.2 * r + 0.01 * r * c)
  f <- write_raster(z, tile = 8L); on.exit(unlink(f))
  rs <- terra::rast(z, extent = terra::ext(0, 24, 0, 30), crs = "EPSG:3857")
  # Read the template's actual grid back so warp uses an identical extent+dims
  # (terra may adjust resolution to fit the extent).
  tmpl <- terra::rast(extent = terra::ext(0.3, 23.1, 0.5, 29.7),
                      resolution = 0.8, crs = "EPSG:3857")
  e <- as.vector(terra::ext(tmpl))
  spec <- list(extent = c(e[1], e[3], e[2], e[4]),
               dims = c(terra::ncol(tmpl), terra::nrow(tmpl)))
  for (meth in c("near", "bilinear", "cubic")) {
    got <- warp(f, spec, method = meth)
    ter <- matrix(terra::values(terra::project(rs, tmpl, method = meth)),
                  nrow = nrow(got), byrow = TRUE)
    int <- !is.na(got) & !is.na(ter)
    # terra stores results as float32, so compare at single-precision tolerance.
    expect_equal(max(abs(got[int] - ter[int])), 0, tolerance = 1e-5, info = meth)
  }
})

test_that("reprojection across CRS matches terra::project", {
  skip_if_not_installed("terra")
  skip_if_not_installed("sf")
  z <- outer(1:40, 1:32, function(r, c) 100 + 0.5 * c + 0.3 * r)
  # Source in Web Mercator over a small patch near the equator.
  f <- write_raster(z, ext = c(0, 0, 32000, 40000), tile = 8L, epsg = 3857)
  rs <- terra::rast(z, extent = terra::ext(0, 32000, 0, 40000), crs = "EPSG:3857")
  on.exit(unlink(f))

  tmpl <- terra::project(rs, "EPSG:4326")          # terra picks the target grid
  e <- as.vector(terra::ext(tmpl))
  spec <- list(crs = 4326, extent = c(e[1], e[3], e[2], e[4]),
               dims = c(terra::ncol(tmpl), terra::nrow(tmpl)))
  got <- warp(f, spec, method = "bilinear")
  ter <- matrix(terra::values(terra::project(rs, tmpl, method = "bilinear")),
                nrow = nrow(got), byrow = TRUE)
  int <- !is.na(got) & !is.na(ter)
  expect_gt(sum(int), 0.5 * length(got))           # most cells defined on both
  expect_equal(max(abs(got[int] - ter[int])), 0, tolerance = 1e-3)
})

test_that("streamed strips equal the single-strip result", {
  set.seed(12)
  m <- matrix(rnorm(37 * 29), 37, 29)
  f_s <- write_raster(m, tile = 8L)
  f_1 <- write_raster(m, tile = 512L)
  on.exit(unlink(c(f_s, f_1)))
  spec <- list(extent = c(1, 2, 26, 33), res = 0.7)
  for (meth in c("near", "bilinear", "cubic")) {
    a <- warp(f_s, spec, method = meth)
    b <- warp(f_1, spec, method = meth)
    expect_equal(a, b, tolerance = 1e-12, info = meth)
  }
})

test_that("streamed-to-.vec output equals the in-memory result", {
  set.seed(13)
  m <- matrix(rnorm(22 * 18), 22, 18)
  f <- write_raster(m, tile = 8L); on.exit(unlink(f))
  spec <- list(extent = c(1, 1, 17, 21), res = 0.6)
  mem <- warp(f, spec, method = "bilinear")

  of <- tempfile(fileext = ".vec"); on.exit(unlink(of), add = TRUE)
  h <- warp(f, spec, method = "bilinear", path = of, dtype = "f64")
  on.exit(vec_close_raster(h), add = TRUE)
  expect_equal(h$n_bands, 1L)
  disk <- vec_read_window(h, band = 1)
  expect_equal(max(abs(mem - disk), na.rm = TRUE), 0)
  expect_true(all(is.na(mem) == is.na(disk)))
})

test_that("a vectra_raster handle works as well as a path", {
  m <- matrix(1:16, 4, 4, byrow = TRUE)
  f <- write_raster(m); on.exit(unlink(f))
  rr <- vec_open_raster(f); on.exit(vec_close_raster(rr), add = TRUE)
  spec <- list(extent = c(0, 0, 4, 4), res = 0.5)
  expect_equal(warp(rr, spec, method = "bilinear"),
               warp(f,  spec, method = "bilinear"))
})

test_that("warp validates inputs", {
  m <- matrix(1:16, 4, 4)
  f <- write_raster(m); on.exit(unlink(f))
  expect_error(warp(f, list(extent = c(0, 0, 4, 4)), method = "near"),
               "res|dims")
  expect_error(warp(f, list(crs = 3857), method = "near"), "extent|res")
  expect_error(warp(f, 42L), "template")
  expect_error(warp(f, list(extent = c(0, 0, 4, 4), res = 1), method = "spline"),
               "should be one of|arg")
})

# Recovery and streaming-invariance tests for contours().
#
# The defining property of a contour at level L is that the raster, sampled at
# every vertex of the line, equals L. The recovery test checks exactly that
# (the value-based ground truth, robust to segmentation and ordering). Plus
# total contour length is equal across tile sizes (streaming invariance).

wr <- function(m, ext = c(0, 0, ncol(m), nrow(m)), tile = 512L) {
  f <- tempfile(fileext = ".vec")
  vec_write_raster(m, f, dtype = "f64", extent = ext, tile_size = tile)
  f
}

test_that("contour vertices sample back to their level", {
  skip_if_not_installed("sf")
  z <- outer(1:30, 1:30, function(r, c) 0.5 * r + 0.8 * c)
  f <- wr(z, ext = c(0, 0, 30, 30), tile = 8L)
  on.exit(unlink(f))

  levs <- c(15, 25, 35)
  iso <- collect_sf(contours(f, levels = levs))
  expect_gt(nrow(iso), 0L)

  r <- vec_open_raster(f)
  on.exit(vec_close_raster(r), add = TRUE)
  bad <- 0
  for (i in seq_len(nrow(iso))) {
    co <- sf::st_coordinates(iso[i, ])
    s <- vec_extract_points(r, co[, "X"], co[, "Y"])[[3]]
    bad <- bad + sum(abs(s - iso$level[i]) > 1.0, na.rm = TRUE)
  }
  expect_equal(bad, 0)
})

test_that("contour total length is invariant to tile size", {
  skip_if_not_installed("sf")
  z <- outer(1:24, 1:24, function(r, c) r + c)
  f1 <- wr(z, ext = c(0, 0, 24, 24), tile = 6L)
  f2 <- wr(z, ext = c(0, 0, 24, 24), tile = 24L)
  on.exit(unlink(c(f1, f2)))

  levs <- c(15, 25, 35)
  l1 <- sum(as.numeric(sf::st_length(collect_sf(contours(f1, levs)))))
  l2 <- sum(as.numeric(sf::st_length(collect_sf(contours(f2, levs)))))
  expect_equal(l1, l2, tolerance = 1e-6)
})

test_that("a planar field gives one line per level of the expected length", {
  skip_if_not_installed("sf")
  # z = x + y over [0,20]^2: level L is the segment x + y = L, length L*sqrt(2)
  # truncated to the square. For L = 20 (the diagonal) length = 20*sqrt(2).
  z <- outer(1:20, 1:20, function(r, c) r + c)
  f <- wr(z, ext = c(0, 0, 20, 20), tile = 5L)
  on.exit(unlink(f))

  iso <- collect_sf(contours(f, levels = 20))
  expect_equal(nrow(iso), 1L)
  len <- sum(as.numeric(sf::st_length(iso)))
  # pixel-centre grid spans centres 0.5..19.5, so the traced diagonal is a bit
  # shorter than the full square diagonal; check it is close to 19*sqrt(2).
  expect_equal(len, 19 * sqrt(2), tolerance = 0.5)
})

test_that("merge = FALSE returns the raw per-cell segments", {
  skip_if_not_installed("sf")
  z <- outer(1:10, 1:10, function(r, c) r + c)
  f <- wr(z, ext = c(0, 0, 10, 10), tile = 10L)
  on.exit(unlink(f))

  seg <- collect_sf(contours(f, levels = 10, merge = FALSE))
  mer <- collect_sf(contours(f, levels = 10, merge = TRUE))
  expect_gt(nrow(seg), nrow(mer))
  # merging does not change total length
  expect_equal(sum(as.numeric(sf::st_length(seg))),
               sum(as.numeric(sf::st_length(mer))), tolerance = 1e-6)
})

test_that("contours rejects empty levels", {
  f <- wr(matrix(1, 4, 4))
  on.exit(unlink(f))
  expect_error(contours(f, levels = numeric(0)), "numeric vector")
})

#' Geometry functions inside mutate(), filter(), and summarise()
#'
#' vectra recognizes a family of `st_*` geometry functions directly inside
#' expression verbs. They run on the GEOS C library straight off the hex-WKB
#' geometry column, one row at a time, with no per-batch round-trip through
#' \pkg{sf}: `tbl(f) |> filter(st_area(geometry) > 1e6)` prunes the stream in C,
#' and `mutate(centroid = st_centroid(geometry))` adds a new hex-WKB geometry
#' column. The computation is planar, in the geometry's own coordinate units,
#' exactly as the streaming spatial verbs are.
#'
#' These names are interpreted by the expression engine; they are not exported R
#' functions and are not called as such. They are available only inside
#' [mutate()], [transmute()], [filter()], and a grouped [summarise()] over a
#' `vectra_node`. The geometry argument is the hex-WKB column (named `geometry`
#' by convention). A function that returns a geometry produces another hex-WKB
#' column; materialize it with [collect_sf()] (point it at the column with
#' `geom =`), or write it with [write_tiff()]/`sf::st_write()`.
#'
#' @section Measures (return a number):
#' \describe{
#'   \item{`st_area(g)`}{Area of polygonal geometry (0 for lines and points).}
#'   \item{`st_length(g)`, `st_perimeter(g)`}{Length of a line, or the perimeter
#'     (boundary length) of a polygon. The two names are aliases.}
#'   \item{`st_x(g)`, `st_y(g)`}{Coordinate of a point geometry; `NA` for a
#'     non-point.}
#'   \item{`st_npoints(g)`}{Number of coordinates in the geometry.}
#'   \item{`st_ngeometries(g)`}{Number of sub-geometries in a collection or
#'     multi-geometry (1 for a single geometry).}
#'   \item{`st_distance(a, b)`}{Shortest planar distance between two geometries
#'     (see the binary second argument below).}
#' }
#'
#' @section Predicates (return TRUE / FALSE):
#' Unary: `st_is_valid(g)`, `st_is_empty(g)`, `st_is_simple(g)`.
#'
#' Binary topological relations, each taking a second geometry: `st_intersects`,
#' `st_within`, `st_contains`, `st_overlaps`, `st_touches`, `st_crosses`,
#' `st_equals`, `st_disjoint`, `st_covers`, `st_covered_by`. Used in [filter()]
#' they keep the rows where the relation holds.
#'
#' @section Type (returns a string):
#' `st_geometry_type(g)` gives the GEOS geometry type name (`"Point"`,
#' `"Polygon"`, `"MultiPolygon"`, ...).
#'
#' @section Transforms (return a geometry):
#' \describe{
#'   \item{`st_centroid(g)`}{Area (or length, or vertex) centroid.}
#'   \item{`st_point_on_surface(g)`}{A point guaranteed to lie on the geometry.}
#'   \item{`st_boundary(g)`}{The topological boundary.}
#'   \item{`st_envelope(g)`}{The axis-aligned bounding rectangle.}
#'   \item{`st_convex_hull(g)`}{The convex hull.}
#'   \item{`st_make_valid(g)`}{Repair an invalid geometry.}
#'   \item{`st_buffer(g, dist)`}{Buffer by `dist` (round joins, 8 segments per
#'     quadrant).}
#'   \item{`st_simplify(g, tol)`}{Topology-preserving Douglas-Peucker
#'     simplification with tolerance `tol`.}
#' }
#'
#' @section The second geometry of a binary op:
#' For `st_distance` and the binary predicates, the second argument can be
#' another geometry column (compared row by row), a constant `sf`/`sfc` object
#' (a multi-feature object is unioned to one geometry), or a hex-WKB string. A
#' constant is parsed once and reused across every row, so testing a whole
#' stream against one area of interest stays cheap.
#'
#' @section Missing geometry:
#' A missing (`NA`) or unparseable geometry, or an operation that has no answer
#' (a coordinate of a non-point, distance to a missing geometry), yields `NA`
#' for that row rather than an error.
#'
#' @name geom_expressions
#' @seealso [mutate()], [filter()], [collect_sf()] to materialize a geometry
#'   result as `sf`; [spatial_map()] for an arbitrary per-feature \pkg{sf}
#'   transform; [spatial_filter()] and [spatial_join()] for relating a stream to
#'   a resident reference layer.
#' @examples
#' if (requireNamespace("sf", quietly = TRUE)) {
#'   nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#'   f <- tempfile(fileext = ".vtr")
#'   write_vtr(data.frame(
#'     NAME     = nc$NAME,
#'     geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
#'   ), f)
#'
#'   # measures: add area and perimeter columns
#'   tbl(f) |>
#'     mutate(area = st_area(geometry), perim = st_perimeter(geometry)) |>
#'     select(NAME, area, perim) |>
#'     collect() |>
#'     head()
#'
#'   # predicate: keep counties intersecting an area of interest
#'   aoi <- sf::st_as_sfc(sf::st_bbox(
#'     c(xmin = -81.5, ymin = 36.2, xmax = -80.5, ymax = 36.6)))
#'   tbl(f) |> filter(st_intersects(geometry, aoi)) |> collect() |> nrow()
#'
#'   # transform: replace each county with its centroid, materialize as sf
#'   tbl(f) |>
#'     mutate(geometry = st_centroid(geometry)) |>
#'     select(NAME, geometry) |>
#'     collect_sf()
#'
#'   unlink(f)
#' }
NULL

# Convert a scalar R value to a literal expression node
.env_val_to_literal <- function(varname, val) {
  if (is.logical(val) && length(val) == 1) return(list(kind = "lit_logical", value = val))
  if (is.integer(val) && length(val) == 1) return(list(kind = "lit_integer", value = val))
  if (is.double(val) && length(val) == 1) return(list(kind = "lit_double", value = val))
  if (is.character(val) && length(val) == 1) return(list(kind = "lit_string", value = val))
  stop(sprintf(".env$%s must be a scalar logical/integer/double/string, got %s of length %d",
               varname, class(val)[1], length(val)))
}

# NSE expression capture -> serialized list for C bridge

# ---------------------------------------------------------------------------
# Dispatch table: function name -> handler
# Each handler receives (fn, expr, env, cols) and returns a serialized list,
# or NULL if the function name is not handled by that group.
# ---------------------------------------------------------------------------

# Arithmetic operators: +, -, *, /, %%
.serialize_arith <- function(fn, expr, env, cols) {
  if (length(expr) == 2 && fn == "-") {
    return(list(kind = "negate",
                operand = serialize_expr(expr[[2]], env, cols)))
  }
  op <- if (fn == "%%") "%" else fn
  list(kind = "arith", op = op,
       left = serialize_expr(expr[[2]], env, cols),
       right = serialize_expr(expr[[3]], env, cols))
}

# Comparison operators: ==, !=, <, <=, >, >=
.serialize_cmp <- function(fn, expr, env, cols) {
  list(kind = "cmp", op = fn,
       left = serialize_expr(expr[[2]], env, cols),
       right = serialize_expr(expr[[3]], env, cols))
}

# Boolean operators: &, &&, |, ||, !
.serialize_bool <- function(fn, expr, env, cols) {
  if (fn == "!" ) {
    return(list(kind = "bool", op = "!",
                operand = serialize_expr(expr[[2]], env, cols)))
  }
  op <- if (fn %in% c("&", "&&")) "&" else "|"
  list(kind = "bool", op = op,
       left = serialize_expr(expr[[2]], env, cols),
       right = serialize_expr(expr[[3]], env, cols))
}

# Math functions: abs, sqrt, log, exp, floor, ceiling, round, log2, log10,
#                 sign, trunc, pmin, pmax
.serialize_math <- function(fn, expr, env, cols) {
  if (fn == "pmin" || fn == "pmax") {
    return(list(kind = fn,
                left = serialize_expr(expr[[2]], env, cols),
                right = serialize_expr(expr[[3]], env, cols)))
  }
  fn_char <- switch(fn,
    abs = "a", sqrt = "s", log = "l", exp = "e",
    floor = "f", ceiling = "c", round = "r",
    log2 = "2", log10 = "t", sign = "g", trunc = "u")
  list(kind = "math_unary", fn = fn_char,
       operand = serialize_expr(expr[[2]], env, cols))
}

# String functions: nchar, substr, substring, grepl, gsub, sub, str_extract,
#   tolower, toupper, trimws, paste, paste0, startsWith, endsWith
.serialize_string <- function(fn, expr, env, cols) {
  if (fn == "nchar") {
    return(list(kind = "nchar",
                operand = serialize_expr(expr[[2]], env, cols)))
  }

  if (fn == "substr" || fn == "substring") {
    return(list(kind = "substr",
                operand = serialize_expr(expr[[2]], env, cols),
                start = serialize_expr(expr[[3]], env, cols),
                stop = serialize_expr(expr[[4]], env, cols)))
  }

  if (fn == "grepl") {
    pattern <- expr[[2]]
    x <- expr[[3]]
    if (!is.character(pattern))
      stop("grepl: pattern must be a string literal")
    fixed <- TRUE
    nms <- names(expr)
    if (!is.null(nms)) {
      fi <- match("fixed", nms)
      if (!is.na(fi)) fixed <- isTRUE(eval(expr[[fi]], env))
    }
    return(list(kind = "grepl",
                pattern = as.character(pattern),
                operand = serialize_expr(x, env, cols),
                fixed = fixed))
  }

  if (fn %in% c("tolower", "toupper", "trimws")) {
    return(list(kind = fn,
                operand = serialize_expr(expr[[2]], env, cols)))
  }

  if (fn == "paste0") {
    args <- lapply(as.list(expr)[-1], serialize_expr, env = env, cols = cols)
    return(list(kind = "paste", args = args, sep = NULL))
  }

  if (fn == "paste") {
    call_args <- as.list(expr)[-1]
    nms <- names(call_args)
    sep <- " "
    data_args <- call_args
    if (!is.null(nms)) {
      si <- match("sep", nms)
      if (!is.na(si)) {
        sep <- as.character(eval(call_args[[si]], env))
        data_args <- call_args[-si]
      }
      ci <- match("collapse", names(data_args))
      if (!is.na(ci)) data_args <- data_args[-ci]
    }
    args <- lapply(data_args, serialize_expr, env = env, cols = cols)
    return(list(kind = "paste", args = args, sep = sep))
  }

  if (fn == "startsWith") {
    prefix <- expr[[3]]
    if (!is.character(prefix)) stop("startsWith: prefix must be a string literal")
    return(list(kind = "startsWith", prefix = as.character(prefix),
                operand = serialize_expr(expr[[2]], env, cols)))
  }

  if (fn == "endsWith") {
    suffix <- expr[[3]]
    if (!is.character(suffix)) stop("endsWith: suffix must be a string literal")
    return(list(kind = "endsWith", suffix = as.character(suffix),
                operand = serialize_expr(expr[[2]], env, cols)))
  }

  if (fn == "gsub" || fn == "sub") {
    pattern <- expr[[2]]
    replacement <- expr[[3]]
    x <- expr[[4]]
    if (!is.character(pattern)) stop(paste0(fn, ": pattern must be a string literal"))
    if (!is.character(replacement)) stop(paste0(fn, ": replacement must be a string literal"))
    fixed <- TRUE
    nms <- names(expr)
    if (!is.null(nms)) {
      fi <- match("fixed", nms)
      if (!is.na(fi)) fixed <- isTRUE(eval(expr[[fi]], env))
    }
    return(list(kind = fn,
                pattern = as.character(pattern),
                replacement = as.character(replacement),
                operand = serialize_expr(x, env, cols),
                fixed = fixed))
  }

  if (fn == "str_extract") {
    x <- expr[[2]]
    pattern <- expr[[3]]
    if (!is.character(pattern))
      stop("str_extract: pattern must be a string literal")
    return(list(kind = "str_extract",
                pattern = as.character(pattern),
                operand = serialize_expr(x, env, cols)))
  }

  NULL
}

# Control flow: case_when, coalesce, ifelse, if_else
.serialize_control_flow <- function(fn, expr, env, cols) {
  if (fn == "case_when") {
    call_args <- as.list(expr)[-1]
    nms <- names(call_args)
    cases <- list()
    default_expr <- NULL
    for (i in seq_along(call_args)) {
      nm <- if (!is.null(nms)) nms[i] else ""
      a <- call_args[[i]]
      if (nm == ".default") {
        default_expr <- serialize_expr(a, env, cols)
      } else if (is.call(a) && identical(a[[1]], as.name("~"))) {
        lhs <- a[[2]]
        rhs <- a[[3]]
        if (is.logical(lhs) && isTRUE(lhs)) {
          default_expr <- serialize_expr(rhs, env, cols)
        } else {
          cases[[length(cases) + 1]] <- list(
            cond = serialize_expr(lhs, env, cols),
            val  = serialize_expr(rhs, env, cols))
        }
      } else {
        stop("case_when: each argument must be a formula (condition ~ value) or .default")
      }
    }
    return(list(kind = "case_when", cases = cases, default = default_expr))
  }

  if (fn == "coalesce") {
    args <- lapply(as.list(expr)[-1], serialize_expr, env = env, cols = cols)
    return(list(kind = "coalesce", args = args))
  }

  # if_else / ifelse
  list(kind = "if_else",
       cond = serialize_expr(expr[[2]], env, cols),
       then_expr = serialize_expr(expr[[3]], env, cols),
       else_expr = serialize_expr(expr[[4]], env, cols))
}

# Type casting: as.numeric, as.double, as.integer, as.character, as.logical
.serialize_cast <- function(fn, expr, env, cols) {
  to <- switch(fn,
    as.numeric = , as.double = "double",
    as.integer = "int64",
    as.character = "string",
    as.logical = "bool")
  list(kind = "cast", to = to,
       operand = serialize_expr(expr[[2]], env, cols))
}

# Date functions: year, month, day, hour, minute, second, as.Date, as.POSIXct
.serialize_date <- function(fn, expr, env, cols) {
  if (fn %in% c("year", "month", "day", "hour", "minute", "second")) {
    part_char <- switch(fn, year = "Y", month = "M", day = "D",
                        hour = "h", minute = "m", second = "s")
    return(list(kind = "date_part", part = part_char,
                operand = serialize_expr(expr[[2]], env, cols)))
  }

  if (fn == "as.Date") {
    arg <- expr[[2]]
    if (is.character(arg)) {
      d <- as.Date(arg)
      return(list(kind = "lit_double", value = as.double(d)))
    }
    return(list(kind = "as_date",
                operand = serialize_expr(arg, env, cols)))
  }

  # as.POSIXct
  arg <- expr[[2]]
  if (is.character(arg)) {
    tz <- "UTC"
    if (length(expr) >= 3) {
      arg_names <- names(expr)
      if (!is.null(arg_names)) {
        tz_idx <- match("tz", arg_names)
        if (!is.na(tz_idx)) tz <- as.character(eval(expr[[tz_idx]], env))
      }
    }
    d <- as.POSIXct(arg, tz = tz)
    return(list(kind = "lit_double", value = as.double(d)))
  }
  list(kind = "cast", to = "double",
       operand = serialize_expr(arg, env, cols))
}

# Fuzzy matching: levenshtein, levenshtein_norm, dl_dist, dl_dist_norm,
#                 jaro_winkler
.serialize_fuzzy <- function(fn, expr, env, cols) {
  args <- as.list(expr[-1])

  if (fn %in% c("levenshtein", "dl_dist")) {
    res <- list(kind = fn,
                operand = serialize_expr(args[[1]], env, cols),
                pattern = serialize_expr(args[[2]], env, cols))
    nm <- names(args)
    if (length(args) >= 3) {
      md <- if (!is.null(nm) && "max_dist" %in% nm) args[["max_dist"]] else args[[3]]
      res$max_dist <- eval(md, env)
    }
    return(res)
  }

  # levenshtein_norm, dl_dist_norm, jaro_winkler
  list(kind = fn,
       operand = serialize_expr(args[[1]], env, cols),
       pattern = serialize_expr(args[[2]], env, cols))
}

# Set operations: %in%, between
.serialize_set_ops <- function(fn, expr, env, cols) {
  if (fn == "%in%") {
    set_val <- eval(expr[[3]], env)
    return(list(kind = "in",
                operand = serialize_expr(expr[[2]], env, cols),
                set = set_val))
  }
  # between
  x <- serialize_expr(expr[[2]], env, cols)
  left <- serialize_expr(expr[[3]], env, cols)
  right <- serialize_expr(expr[[4]], env, cols)
  list(kind = "bool", op = "&",
       left = list(kind = "cmp", op = ">=", left = x, right = left),
       right = list(kind = "cmp", op = "<=", left = x, right = right))
}

# Graph/lookup: resolve, propagate
.serialize_graph <- function(fn, expr, env, cols) {
  if (fn == "resolve") {
    if (length(expr) != 4)
      stop("resolve() requires exactly 3 arguments: fk_col, pk_col, value_col")
    return(list(kind = "resolve",
                fk = serialize_expr(expr[[2]], env, cols),
                pk = serialize_expr(expr[[3]], env, cols),
                val = serialize_expr(expr[[4]], env, cols)))
  }
  # propagate
  if (length(expr) != 4)
    stop("propagate() requires exactly 3 arguments: parent_fk, pk_col, seed_expr")
  list(kind = "propagate",
       parent_fk = serialize_expr(expr[[2]], env, cols),
       pk = serialize_expr(expr[[3]], env, cols),
       seed = serialize_expr(expr[[4]], env, cols))
}

# Geometry functions: libgeos ops over a hex-WKB geometry column, run inside the
# C engine off the column with no per-batch sf round-trip. Each name maps to a
# single-character op code the C side dispatches on (see src/expr_geom.c).
#   unary      geom -> double / bool / string / geometry
#   parameter  geom, numeric -> geometry         (buffer, simplify)
#   binary     geom, geom    -> double / bool     (distance, predicates)
.GEOM_UNARY <- c(
  st_area = "A", st_length = "L", st_perimeter = "L",
  st_x = "X", st_y = "Y", st_npoints = "n", st_ngeometries = "g",
  st_is_valid = "v", st_is_empty = "m", st_is_simple = "s",
  st_geometry_type = "t",
  st_centroid = "c", st_point_on_surface = "o", st_boundary = "b",
  st_envelope = "e", st_convex_hull = "h", st_make_valid = "M")
.GEOM_PARAM <- c(st_buffer = "B", st_simplify = "S")
.GEOM_BINARY <- c(
  st_distance = "D", st_intersects = "i", st_within = "w", st_contains = "C",
  st_overlaps = "O", st_touches = "T", st_crosses = "R", st_equals = "Q",
  st_disjoint = "J", st_covers = "K", st_covered_by = "V")

# Resolve a binary op's second geometry: a geometry column reference, a constant
# sf/sfc object (converted once to hex-WKB), or a hex-WKB string literal.
.serialize_geom_arg <- function(a, env, cols) {
  if (is.name(a) && !is.null(cols) && as.character(a) %in% cols)
    return(serialize_expr(a, env, cols))
  val <- tryCatch(eval(a, env), error = function(e) NULL)
  if (inherits(val, "sf") || inherits(val, "sfc")) {
    if (!requireNamespace("sf", quietly = TRUE))
      stop("a constant sf/sfc geometry argument needs the sf package installed")
    g <- sf::st_geometry(val)
    if (length(g) > 1) g <- sf::st_union(g)
    hex <- sf::st_as_binary(g, hex = TRUE)
    return(list(kind = "lit_string", value = as.character(hex)[1]))
  }
  if (is.character(val) && length(val) == 1)
    return(list(kind = "lit_string", value = val))
  stop("second geometry argument must be a geometry column, an sf/sfc object, ",
       "or a hex-WKB string")
}

.serialize_geom <- function(fn, expr, env, cols) {
  if (fn %in% names(.GEOM_UNARY)) {
    return(list(kind = "geom", fn = unname(.GEOM_UNARY[fn]),
                operand = serialize_expr(expr[[2]], env, cols)))
  }
  if (fn %in% names(.GEOM_PARAM)) {
    return(list(kind = "geom", fn = unname(.GEOM_PARAM[fn]),
                operand = serialize_expr(expr[[2]], env, cols),
                param = serialize_expr(expr[[3]], env, cols)))
  }
  # binary: distance / topological predicate
  list(kind = "geom", fn = unname(.GEOM_BINARY[fn]),
       operand = serialize_expr(expr[[2]], env, cols),
       other = .serialize_geom_arg(expr[[3]], env, cols))
}

# ---------------------------------------------------------------------------
# Dispatch registry: maps function names to handler functions.
# Adding a new expression type = one entry here + one handler above.
# ---------------------------------------------------------------------------
.expr_dispatch <- new.env(parent = emptyenv())

# Populate dispatch table
local({
  register <- function(names, handler) {
    for (nm in names) .expr_dispatch[[nm]] <- handler
  }

  register(c("+", "-", "*", "/", "%%"),               .serialize_arith)
  register(c("==", "!=", "<", "<=", ">", ">="),       .serialize_cmp)
  register(c("&", "&&", "|", "||", "!"),              .serialize_bool)
  register(c("abs", "sqrt", "log", "exp", "floor",
             "ceiling", "round", "log2", "log10",
             "sign", "trunc", "pmin", "pmax"),         .serialize_math)
  register(c("nchar", "substr", "substring", "grepl",
             "tolower", "toupper", "trimws", "paste0",
             "paste", "startsWith", "endsWith", "gsub",
             "sub", "str_extract"),                     .serialize_string)
  register(c("case_when", "coalesce",
             "if_else", "ifelse"),                      .serialize_control_flow)
  register(c("as.numeric", "as.double", "as.integer",
             "as.character", "as.logical"),              .serialize_cast)
  register(c("year", "month", "day", "hour", "minute",
             "second", "as.Date", "as.POSIXct"),        .serialize_date)
  register(c("levenshtein", "levenshtein_norm",
             "dl_dist", "dl_dist_norm",
             "jaro_winkler"),                            .serialize_fuzzy)
  register(c("%in%", "between"),                        .serialize_set_ops)
  register(c("resolve", "propagate"),                   .serialize_graph)
  register(c(names(.GEOM_UNARY), names(.GEOM_PARAM),
             names(.GEOM_BINARY)),                       .serialize_geom)
})

# ---------------------------------------------------------------------------
# Main entry point
# ---------------------------------------------------------------------------
serialize_expr <- function(expr, env = parent.frame(), cols = NULL) {
  # Bare names: column refs, constants, or env lookups

  if (is.name(expr)) {
    name <- as.character(expr)
    if (name == "TRUE") return(list(kind = "lit_logical", value = TRUE))
    if (name == "FALSE") return(list(kind = "lit_logical", value = FALSE))
    if (name == "NA") return(list(kind = "lit_na"))
    if (name == "NA_real_") return(list(kind = "lit_na"))
    if (name == "NA_integer_") return(list(kind = "lit_na"))
    if (name == "NA_character_") return(list(kind = "lit_na"))
    if (!is.null(cols) && !(name %in% cols)) {
      val <- tryCatch(get(name, envir = env), error = function(e) NULL)
      if (!is.null(val)) return(.env_val_to_literal(name, val))
    }
    return(list(kind = "col_ref", name = name))
  }

  # Scalar literals
  if (is.numeric(expr) || is.logical(expr) || is.character(expr)) {
    val <- expr
    if (is.logical(val) && is.na(val)) return(list(kind = "lit_na"))
    if (is.logical(val)) return(list(kind = "lit_logical", value = val))
    if (is.integer(val) && is.na(val)) return(list(kind = "lit_na"))
    if (is.integer(val)) return(list(kind = "lit_integer", value = val))
    if (is.double(val) && is.na(val)) return(list(kind = "lit_na"))
    if (is.double(val)) return(list(kind = "lit_double", value = val))
    if (is.character(val) && is.na(val)) return(list(kind = "lit_na"))
    if (is.character(val)) return(list(kind = "lit_string", value = val))
  }

  if (!is.call(expr))
    stop(sprintf("unsupported expression type: %s", typeof(expr)))

  # .env$varname or .env[["varname"]] — evaluate in caller's environment
  if (length(expr) == 3) {
    op <- expr[[1]]
    lhs <- expr[[2]]
    if (is.name(lhs) && identical(as.character(lhs), ".env") &&
        (identical(op, quote(`$`)) || identical(op, quote(`[[`)))) {
      varname <- if (identical(op, quote(`$`))) as.character(expr[[3]]) else eval(expr[[3]], env)
      val <- get(varname, envir = env)
      return(.env_val_to_literal(varname, val))
    }
  }

  fn <- as.character(expr[[1]])

  # Parentheses — pass through
  if (fn == "(") return(serialize_expr(expr[[2]], env, cols))

  # is.na — standalone because it doesn't group with any category

  if (fn == "is.na") {
    return(list(kind = "is_na",
                operand = serialize_expr(expr[[2]], env, cols)))
  }

  # Dispatch to handler

  handler <- .expr_dispatch[[fn]]
  if (!is.null(handler)) return(handler(fn, expr, env, cols))

  stop(sprintf("unsupported function in expression: %s", fn))
}

# Combine multiple filter expressions with &
combine_predicates <- function(exprs, env, cols = NULL) {
  if (length(exprs) == 0) stop("no filter expressions provided")
  result <- serialize_expr(exprs[[1]], env, cols)
  for (i in seq_along(exprs)[-1]) {
    result <- list(kind = "bool", op = "&",
                   left = result,
                   right = serialize_expr(exprs[[i]], env, cols))
  }
  result
}

# network.R - streamed network analysis over a resident line graph.
#
# Tier 5 of the spatial roadmap: routing and reachability over a line network.
# Unlike the geometry verbs, this needs a graph (nodes at line endpoints, edges
# weighted by length or a cost column) and a shortest-path solver, not a geometry
# stream. It still fits vectra's shape by reusing the resident-`y` streamed-`x`
# family (spatial_knn, spatial_locate): spatial_network() builds the node-edge
# graph once and holds it resident in an external pointer; spatial_route() and
# spatial_service_area() stream origin (and destination) batches past it.
#
# The graph build and the Dijkstra solver are native C (src/network.c); the
# geometry side - endpoint extraction, snapping origins to nodes, assembling a
# route line - is sf, matching how the rest of this family handles geometry. No
# igraph dependency: the solver is a binary-heap Dijkstra over a CSR adjacency.

# Normalise a one-way direction code column to "B" (both ways), "FT" (from->to
# only, the digitised direction), "TF" (to->from only), or "N" (closed). Accepts
# the textual codes and the +/-/0 sign convention; NA becomes "B".
.network_dir_code <- function(v) {
  cc <- toupper(trimws(as.character(v)))
  cc[is.na(cc) | cc == "" | cc %in% c("BOTH", "B", "TRUE")] <- "B"
  cc[cc %in% c("FT", "F", "+", "1")] <- "FT"
  cc[cc %in% c("TF", "T", "-")]      <- "TF"
  cc[cc %in% c("N", "NONE", "0", "FALSE", "CLOSED")] <- "N"
  bad <- !cc %in% c("B", "FT", "TF", "N")
  if (any(bad))
    stop("unknown direction code(s): ",
         paste(unique(v[bad]), collapse = ", "),
         "; use B / FT / TF / N (or +/-/0)", call. = FALSE)
  cc
}

# Snap a set of query geometries to their nearest graph node, returning 1-based
# node ids (indices into the network's node table).
.network_snap <- function(network, g) {
  g <- .align_resident_crs(g, network$crs)
  as.integer(sf::st_nearest_feature(g, network$node_pts))
}

# Coerce a destination argument (sf / sfc) to an sfc of points in the network CRS.
.network_dest <- function(to, crs) {
  if (inherits(to, "sf")) g <- sf::st_geometry(to)
  else if (inherits(to, "sfc")) g <- to
  else stop("`to` must be an sf or sfc object", call. = FALSE)
  .align_resident_crs(g, crs)
}

# Rebuild route line geometry from C's edge-id paths. Each path is the sequence
# of directed-edge ids (0-based) the solver walked; an edge maps to a source line
# (network$coords_list) traversed forward or reversed. Coincident join vertices
# are dropped. An empty path (unreachable, or origin == destination) yields an
# empty linestring.
.route_geometry <- function(network, paths, crs) {
  cl    <- network$coords_list
  eline <- network$edges$line
  erev  <- network$edges$rev
  geoms <- lapply(paths, function(p) {
    if (!length(p)) return(sf::st_linestring())
    full <- NULL
    for (e in p) {
      i <- e + 1L
      m <- cl[[eline[i]]]
      if (erev[i]) m <- m[nrow(m):1L, , drop = FALSE]
      if (is.null(full)) {
        full <- m
      } else {
        if (full[nrow(full), 1L] == m[1L, 1L] &&
            full[nrow(full), 2L] == m[1L, 2L])
          m <- m[-1L, , drop = FALSE]
        full <- rbind(full, m)
      }
    }
    if (is.null(full) || nrow(full) < 2L) sf::st_linestring()
    else sf::st_linestring(full)
  })
  sf::st_sfc(geoms, crs = .as_crs(crs))
}

# Build one origin's service-area geometry for a cost band `b`, given the node
# cost vector `ncost` (Inf where unreached). "nodes" = the reachable nodes as a
# multipoint, "lines" = the edges with both endpoints reachable as a
# multilinestring, "polygon" = the convex hull of the reachable nodes (the
# generalised service area); a band reaching fewer than three nodes yields an
# empty polygon.
.service_geometry <- function(network, ncost, b, output) {
  in_b <- ncost <= b
  if (output == "lines") {
    sel <- which(in_b[network$line_from] & in_b[network$line_to])
    if (!length(sel)) return(sf::st_multilinestring())
    return(sf::st_multilinestring(network$coords_list[sel]))
  }
  idx <- which(in_b)
  if (!length(idx))
    return(if (output == "polygon") sf::st_polygon() else sf::st_multipoint())
  pts <- network$node_xy[idx, , drop = FALSE]
  if (output == "nodes") return(sf::st_multipoint(pts))
  if (nrow(pts) < 3L) return(sf::st_polygon())
  sf::st_convex_hull(sf::st_multipoint(pts))
}

#' Build a routable network graph from a line layer
#'
#' Builds the node-edge graph a shortest-path query needs: nodes at line
#' endpoints, edges weighted by geometry length or a cost column. The graph is
#' held resident in an external pointer and passed to [spatial_route()] and
#' [spatial_service_area()], which stream origin (and destination) batches past
#' it. This is the network counterpart of a resident `sf` `y` in [spatial_knn()]:
#' the graph is the resident budget (bounded by the network size), while the
#' query side scales by streaming.
#'
#' Endpoints within `tolerance` of each other are snapped to a single node, so a
#' layer whose touching lines do not share exactly-equal endpoint coordinates
#' still connects. The input must be (multi)linestrings; a `MULTILINESTRING` is
#' split into its parts, each becoming one edge with the source attributes. Lines
#' are treated as already split at their junctions (the usual road-network
#' convention); two lines that merely cross in their interiors are not connected
#' unless they share an endpoint. The graph and the Dijkstra solver are native C;
#' \pkg{sf} (an optional dependency, Suggests) supplies only the geometry.
#'
#' @param lines An `sf` or `sfc` object of (multi)linestrings, or a `vectra_node`
#'   carrying a hex-WKB line geometry column (it is materialised).
#' @param weight Name of a column in `lines` holding each edge's traversal cost.
#'   `NULL` (default) uses the geometry length ([sf::st_length()]).
#' @param directed If `TRUE`, edges are one-way (by `direction`, or the digitised
#'   from->to direction). `FALSE` (default) makes every edge two-way.
#' @param direction Name of a column of one-way codes, used when `directed`:
#'   `"B"` two-way, `"FT"` from->to only, `"TF"` to->from only, `"N"` closed (the
#'   `+`/`-`/`0` sign convention is also accepted). `NULL` uses `"FT"` for every
#'   line, or `"B"` when `weight_to` is given.
#' @param weight_to Name of a column holding the reverse-direction cost on a
#'   two-way edge of a directed graph. `NULL` reuses `weight`.
#' @param tolerance Endpoints within this distance (CRS units) are snapped to one
#'   node. `0` (default) joins only exactly-coincident endpoints.
#' @param geom,crs Geometry column name and CRS, as in [spatial_map()]. The CRS
#'   defaults to the one `lines` carries.
#'
#' @return A `vectra_network` object: the resident graph (an external pointer
#'   plus the node coordinates, edge table, and source-line geometry needed to
#'   snap queries and rebuild routes). Print it to see the node, edge, and
#'   connected-component counts.
#'
#' @seealso [spatial_route()] for shortest paths and origin-destination costs,
#'   [spatial_service_area()] for reachability and isochrones.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' # A small grid of streets.
#' mk <- function(x1, y1, x2, y2)
#'   sf::st_linestring(rbind(c(x1, y1), c(x2, y2)))
#' streets <- sf::st_sfc(
#'   mk(0, 0, 1, 0), mk(1, 0, 2, 0), mk(0, 0, 0, 1),
#'   mk(0, 1, 1, 1), mk(1, 0, 1, 1), mk(1, 1, 2, 1), mk(2, 0, 2, 1))
#' net <- spatial_network(streets)
#' net
#'
#' @export
spatial_network <- function(lines, weight = NULL, directed = FALSE,
                            direction = NULL, weight_to = NULL,
                            tolerance = 0, geom = "geometry", crs = NA) {
  .check_sf()
  if (inherits(lines, "vectra_node")) {
    node_crs <- lines$.crs
    lines <- collect_sf(lines, geom = geom, crs = crs)
    if (identical(crs, NA) || is.null(crs)) crs <- node_crs
  }
  if (inherits(lines, "sfc")) lines <- sf::st_sf(geometry = lines)
  if (!inherits(lines, "sf"))
    stop("`lines` must be an sf/sfc object or a vectra_node", call. = FALSE)
  if (identical(crs, NA) || is.null(crs)) crs <- sf::st_crs(lines)
  crs <- .as_crs(crs)

  gt <- as.character(sf::st_geometry_type(sf::st_geometry(lines)))
  if (!all(gt %in% c("LINESTRING", "MULTILINESTRING")))
    stop("`lines` must be LINESTRING or MULTILINESTRING geometry", call. = FALSE)
  # explode multilinestrings into single linestring edges, replicating attributes
  lines <- suppressWarnings(
    sf::st_cast(sf::st_cast(lines, "MULTILINESTRING"), "LINESTRING"))
  g <- sf::st_geometry(lines)

  empt <- sf::st_is_empty(g)
  if (any(empt)) {
    warning(sprintf("dropping %d empty line geometr%s", sum(empt),
                    if (sum(empt) == 1L) "y" else "ies"), call. = FALSE)
    lines <- lines[!empt, , drop = FALSE]
    g <- sf::st_geometry(lines)
  }
  n_lines <- length(g)
  if (n_lines == 0L) stop("`lines` has no usable line geometry", call. = FALSE)

  # edge weights
  if (is.null(weight)) {
    w <- as.numeric(sf::st_length(g))
  } else {
    if (!weight %in% names(lines))
      stop(sprintf("weight column '%s' not found in `lines`", weight),
           call. = FALSE)
    w <- as.numeric(lines[[weight]])
  }
  if (any(!is.finite(w) | w < 0))
    stop("edge weights must be finite and non-negative", call. = FALSE)

  # endpoints + per-line coordinate matrices, from one coordinate pass
  co <- sf::st_coordinates(g)
  L1 <- co[, "L1"]
  coords_list <- lapply(split(seq_len(nrow(co)), L1),
                        function(ix) co[ix, c("X", "Y"), drop = FALSE])
  coords_list <- coords_list[order(as.integer(names(coords_list)))]
  names(coords_list) <- NULL
  first <- !duplicated(L1)
  last  <- !duplicated(L1, fromLast = TRUE)
  a_xy <- co[first, c("X", "Y"), drop = FALSE]   # start of each line
  b_xy <- co[last,  c("X", "Y"), drop = FALSE]   # end of each line

  # snap endpoints to node ids: round to the tolerance grid, hash the coordinate
  allxy <- rbind(a_xy, b_xy)
  rxy <- if (tolerance > 0) round(allxy / tolerance) * tolerance else allxy
  key <- paste(rxy[, 1L], rxy[, 2L], sep = "\r")
  uk  <- unique(key)
  node_of <- match(key, uk)                       # 1-based node id per endpoint
  n_nodes <- length(uk)
  rep_idx <- match(uk, key)                        # first occurrence per node
  node_xy <- allxy[rep_idx, , drop = FALSE]        # representative node coords

  line_from <- node_of[seq_len(n_lines)]
  line_to   <- node_of[n_lines + seq_len(n_lines)]

  # expand into directed edges, carrying each edge's source line and a reversed
  # flag so a route path can rebuild geometry
  if (!directed) {
    dfrom <- c(line_from, line_to)
    dto   <- c(line_to,   line_from)
    dw    <- c(w, w)
    dline <- c(seq_len(n_lines), seq_len(n_lines))
    drev  <- c(rep(FALSE, n_lines), rep(TRUE, n_lines))
  } else {
    wt <- if (is.null(weight_to)) w else {
      if (!weight_to %in% names(lines))
        stop(sprintf("weight_to column '%s' not found in `lines`", weight_to),
             call. = FALSE)
      as.numeric(lines[[weight_to]])
    }
    if (any(!is.finite(wt) | wt < 0))
      stop("reverse weights must be finite and non-negative", call. = FALSE)
    code <- if (is.null(direction)) {
      rep(if (is.null(weight_to)) "FT" else "B", n_lines)
    } else {
      if (!direction %in% names(lines))
        stop(sprintf("direction column '%s' not found in `lines`", direction),
             call. = FALSE)
      .network_dir_code(lines[[direction]])
    }
    fwd <- code %in% c("FT", "B")
    bwd <- code %in% c("TF", "B")
    dfrom <- c(line_from[fwd], line_to[bwd])
    dto   <- c(line_to[fwd],   line_from[bwd])
    dw    <- c(w[fwd],         wt[bwd])
    dline <- c(which(fwd),     which(bwd))
    drev  <- c(rep(FALSE, sum(fwd)), rep(TRUE, sum(bwd)))
  }

  ptr <- .Call(C_network_build, as.integer(dfrom - 1L), as.integer(dto - 1L),
               as.double(dw), as.integer(n_nodes))

  node_pts <- sf::st_as_sf(
    data.frame(.x = node_xy[, 1L], .y = node_xy[, 2L]),
    coords = c(".x", ".y"), crs = crs)

  structure(
    list(ptr = ptr,
         n_nodes = n_nodes,
         node_xy = node_xy,
         node_pts = node_pts,
         coords_list = coords_list,
         line_from = line_from,
         line_to = line_to,
         edges = list(line = as.integer(dline), rev = drev),
         crs = crs,
         directed = directed,
         stats = .Call(C_network_stats, ptr)),
    class = "vectra_network")
}

#' @export
print.vectra_network <- function(x, ...) {
  st <- x$stats
  cat("<vectra_network>\n")
  cat(sprintf("  nodes:      %d\n", st[1L]))
  cat(sprintf("  edges:      %d (%s)\n", st[2L],
              if (x$directed) "directed" else "undirected"))
  cat(sprintf("  components: %d\n", st[3L]))
  crs <- tryCatch(x$crs$input, error = function(e) NA)
  if (!is.null(crs) && !is.na(crs)) cat(sprintf("  crs:        %s\n", crs))
  invisible(x)
}

#' Shortest paths and origin-destination costs over a network
#'
#' Streams a layer of origins `x` past a resident [spatial_network()] and, for
#' each origin, finds the shortest path to one or more destinations `to`. Each
#' origin and destination is snapped to its nearest graph node; the solver
#' (native-C Dijkstra) returns one row per (origin, destination) pair with the
#' total cost and, by default, the route geometry. With `geometry = FALSE` only
#' the cost is returned, so a destination set per origin yields the
#' origin-destination cost matrix in long form. The billion-origin stream never
#' materialises; the graph stays resident.
#'
#' An unreachable destination returns an infinite cost and an empty geometry
#' rather than dropping the row, so a cost matrix stays rectangular. Snapping is
#' to the nearest node, so place origins and destinations on or near the network;
#' costs are in the units of the network's `weight` (or CRS length units when the
#' graph was built from geometry length). The \pkg{sf} package is an optional
#' dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param x A `vectra_node` of origin features (point geometry, or `coords`).
#' @param network A `vectra_network` from [spatial_network()].
#' @param to An `sf`/`sfc` of destination points. One destination routes every
#'   origin to it; several produce one row per (origin, destination).
#' @param to_id Optional column in `to` identifying each destination in the
#'   output. `NULL` (default) uses the 1-based destination index.
#' @param geometry If `TRUE` (default) each row carries the route line; if
#'   `FALSE` only the cost column (the cost-matrix form, no geometry).
#' @param cost_col,dest_col Names of the output cost and destination-identifier
#'   columns. Defaults `"cost"`, `"destination"`.
#'
#' @return A `vectra_node`: one row per (origin, destination) carrying `x`'s
#'   attributes, the destination id, the cost, and (when `geometry = TRUE`) the
#'   route geometry. Backed by temporary `.vtr` spills removed when the node is
#'   garbage-collected, and carrying the network CRS.
#'
#' @seealso [spatial_network()] to build the graph, [spatial_service_area()] for
#'   reachability, [collect_sf()] to materialize routes as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' mk <- function(x1, y1, x2, y2)
#'   sf::st_linestring(rbind(c(x1, y1), c(x2, y2)))
#' streets <- sf::st_sfc(
#'   mk(0, 0, 1, 0), mk(1, 0, 2, 0), mk(0, 0, 0, 1),
#'   mk(0, 1, 1, 1), mk(1, 0, 1, 1), mk(1, 1, 2, 1), mk(2, 0, 2, 1))
#' net <- spatial_network(streets)
#'
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(id = 1:2, x = c(0, 0), y = c(0, 1)), f)
#' dest <- sf::st_sfc(sf::st_point(c(2, 1)))
#'
#' tbl(f) |>
#'   spatial_route(net, to = dest, coords = c("x", "y")) |>
#'   collect_sf()
#' unlink(f)
#'
#' @export
spatial_route <- function(x, network, to, to_id = NULL, geometry = TRUE,
                          cost_col = "cost", dest_col = "destination",
                          geom = "geometry", coords = NULL, crs = NA,
                          out_geom = NULL, flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed origins)", call. = FALSE)
  if (!inherits(network, "vectra_network"))
    stop("`network` must be a vectra_network from spatial_network()",
         call. = FALSE)
  crs <- .resolve_crs(x, crs)
  if (is.na(.as_crs(crs))) crs <- network$crs

  dst_g    <- .network_dest(to, crs)
  dst_node <- .network_snap(network, dst_g)
  nd <- length(dst_node)
  if (nd == 0L) stop("`to` has no destinations", call. = FALSE)
  did <- if (is.null(to_id)) seq_len(nd) else {
    if (!inherits(to, "sf") || !to_id %in% names(to))
      stop(sprintf("`to_id` column '%s' not found in `to`", to_id),
           call. = FALSE)
    to[[to_id]]
  }

  out_geom <- out_geom %||% (if (is.null(coords)) geom else "geometry")
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  acc <- .run_accumulator(fr)
  nxt <- .batch_cursor(x)
  repeat {
    chunk <- nxt(); if (is.null(chunk)) break
    sb <- .sf_decode_chunk(chunk, geom, coords, crs)
    no <- nrow(sb)
    if (no == 0L) next
    o_node <- .network_snap(network, sf::st_geometry(sb))

    src_pairs <- rep(o_node,   each  = nd)
    dst_pairs <- rep(dst_node, times = no)
    res <- .Call(C_network_route, network$ptr,
                 as.integer(src_pairs - 1L), as.integer(dst_pairs - 1L),
                 isTRUE(geometry))

    oi <- rep(seq_len(no), each = nd)
    df <- as.data.frame(sf::st_drop_geometry(sb))[oi, , drop = FALSE]
    rownames(df) <- NULL
    df[[dest_col]] <- did[rep(seq_len(nd), times = no)]
    df[[cost_col]] <- res$cost

    if (isTRUE(geometry)) {
      rg  <- .route_geometry(network, res$paths, crs)
      out <- sf::st_sf(df, geometry = rg)
      acc$push(.sf_encode_result(out, out_geom))
    } else {
      acc$push(.coerce_for_vtr(df))
    }
  }
  acc$finish(crs = .as_crs(crs), empty_geom = out_geom)
}

#' Service areas and isochrones over a network
#'
#' Streams a layer of origins `x` past a resident [spatial_network()] and, for
#' each origin, finds every part of the network reachable within a cost budget -
#' the service area, or, with several budgets, nested travel-cost isochrone bands.
#' Each origin is snapped to its nearest graph node and a budget-bounded Dijkstra
#' (native C) collects the reachable nodes; the reachable set is returned as the
#' convex hull (`output = "polygon"`, the generalised service area), the
#' reachable edges (`"lines"`), or the reachable nodes (`"nodes"`).
#'
#' A vector `cost` returns one row per (origin, band), each band the area
#' reachable within that budget, so the rows nest from the smallest budget out.
#' Costs are in the network's `weight` units. The convex-hull polygon is a
#' generalisation; use `output = "lines"` for the exact reachable network. The
#' \pkg{sf} package is an optional dependency (Suggests).
#'
#' @inheritParams spatial_route
#' @param cost A cost budget (scalar), or several budgets for nested isochrone
#'   bands (e.g. `c(5, 10, 15)`).
#' @param output `"polygon"` (default) for the convex hull of the reachable
#'   nodes, `"lines"` for the reachable edges, or `"nodes"` for the reachable
#'   nodes as a multipoint.
#' @param band_col Name of the output column holding each row's budget. Default
#'   `"band"`.
#'
#' @return A `vectra_node`: one row per (origin, band) carrying `x`'s attributes,
#'   the band value, and the service-area geometry. Backed by temporary `.vtr`
#'   spills removed when the node is garbage-collected, and carrying the network
#'   CRS.
#'
#' @seealso [spatial_network()] to build the graph, [spatial_route()] for
#'   shortest paths, [collect_sf()] to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' mk <- function(x1, y1, x2, y2)
#'   sf::st_linestring(rbind(c(x1, y1), c(x2, y2)))
#' streets <- sf::st_sfc(
#'   mk(0, 0, 1, 0), mk(1, 0, 2, 0), mk(0, 0, 0, 1),
#'   mk(0, 1, 1, 1), mk(1, 0, 1, 1), mk(1, 1, 2, 1), mk(2, 0, 2, 1))
#' net <- spatial_network(streets)
#'
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(id = 1L, x = 0, y = 0), f)
#'
#' tbl(f) |>
#'   spatial_service_area(net, cost = c(1, 2), output = "lines",
#'                        coords = c("x", "y")) |>
#'   collect_sf()
#' unlink(f)
#'
#' @export
spatial_service_area <- function(x, network, cost,
                                 output = c("polygon", "lines", "nodes"),
                                 band_col = "band",
                                 geom = "geometry", coords = NULL, crs = NA,
                                 out_geom = NULL, flush_rows = NULL) {
  .check_sf()
  output <- match.arg(output)
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed origins)", call. = FALSE)
  if (!inherits(network, "vectra_network"))
    stop("`network` must be a vectra_network from spatial_network()",
         call. = FALSE)
  bands <- sort(unique(as.numeric(cost)))
  if (!length(bands) || any(!is.finite(bands) | bands < 0))
    stop("`cost` must be one or more finite, non-negative budgets", call. = FALSE)
  nb <- length(bands)
  max_budget <- max(bands)

  crs <- .resolve_crs(x, crs)
  if (is.na(.as_crs(crs))) crs <- network$crs
  out_geom <- out_geom %||% "geometry"
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  acc <- .run_accumulator(fr)
  nxt <- .batch_cursor(x)
  repeat {
    chunk <- nxt(); if (is.null(chunk)) break
    sb <- .sf_decode_chunk(chunk, geom, coords, crs)
    no <- nrow(sb)
    if (no == 0L) next
    o_node <- .network_snap(network, sf::st_geometry(sb))
    res <- .Call(C_network_service, network$ptr,
                 as.integer(o_node - 1L), as.double(max_budget))

    geoms <- vector("list", no * nb)
    k <- 0L
    for (i in seq_len(no)) {
      ri <- res[[i]]
      ncost <- rep(Inf, network$n_nodes)
      if (length(ri[[1L]])) ncost[ri[[1L]] + 1L] <- ri[[2L]]
      for (b in bands) {
        k <- k + 1L
        geoms[[k]] <- .service_geometry(network, ncost, b, output)
      }
    }
    oi <- rep(seq_len(no), each = nb)
    df <- as.data.frame(sf::st_drop_geometry(sb))[oi, , drop = FALSE]
    rownames(df) <- NULL
    df[[band_col]] <- rep(bands, times = no)
    out <- sf::st_sf(df, geometry = sf::st_sfc(geoms, crs = .as_crs(crs)))
    acc$push(.sf_encode_result(out, out_geom))
  }
  acc$finish(crs = .as_crs(crs), empty_geom = out_geom)
}

# Network analysis: spatial_network / spatial_route / spatial_service_area.
#
# Recovery tests against hand-computed shortest paths on a known graph (and
# against igraph when available), route geometry reconstruction, service-area
# reachable sets, directed one-way handling, unreachable Inf, and the streaming
# invariance that a multi-batch origin stream equals a single-batch result.

skip_if_no_sf <- function() skip_if_not_installed("sf")

# A 3x2 lattice of unit street segments:
#   (0,1)--(1,1)--(2,1)
#     |      |      |
#   (0,0)--(1,0)--(2,0)
# Every edge has length 1, so shortest-path cost == Manhattan grid distance.
lattice_streets <- function() {
  mk <- function(x1, y1, x2, y2)
    sf::st_linestring(rbind(c(x1, y1), c(x2, y2)))
  sf::st_sfc(
    mk(0, 0, 1, 0), mk(1, 0, 2, 0),          # bottom row
    mk(0, 1, 1, 1), mk(1, 1, 2, 1),          # top row
    mk(0, 0, 0, 1), mk(1, 0, 1, 1), mk(2, 0, 2, 1)  # verticals
  )
}

write_points <- function(df) {
  f <- tempfile(fileext = ".vtr")
  write_vtr(df, f)
  f
}

test_that("spatial_network reports node, edge, component counts", {
  skip_if_no_sf()
  net <- spatial_network(lattice_streets())
  expect_s3_class(net, "vectra_network")
  expect_identical(net$stats[1], 6L)        # 6 nodes
  expect_identical(net$stats[2], 14L)       # 7 lines * 2 directed edges
  expect_identical(net$stats[3], 1L)        # one connected component
})

test_that("route cost matches hand-computed grid distance", {
  skip_if_no_sf()
  net <- spatial_network(lattice_streets())
  f <- write_points(data.frame(id = 1L, x = 0, y = 0))
  dest <- sf::st_sfc(sf::st_point(c(2, 1)))   # opposite corner

  res <- tbl(f) |>
    spatial_route(net, to = dest, coords = c("x", "y"), geometry = FALSE) |>
    collect()
  # (0,0) -> (2,1): three unit steps
  expect_equal(res$cost, 3)
  unlink(f)
})

test_that("origin-destination cost matrix is one row per pair", {
  skip_if_no_sf()
  net <- spatial_network(lattice_streets())
  f <- write_points(data.frame(id = 1:2, x = c(0, 2), y = c(0, 0)))
  dests <- sf::st_sf(
    name = c("tl", "tr"),
    geometry = sf::st_sfc(sf::st_point(c(0, 1)), sf::st_point(c(2, 1))))

  res <- tbl(f) |>
    spatial_route(net, to = dests, to_id = "name", geometry = FALSE,
                  coords = c("x", "y")) |>
    collect()
  expect_equal(nrow(res), 4L)               # 2 origins x 2 destinations
  # origin (0,0): to tl(0,1)=1, to tr(2,1)=3; origin (2,0): tl=3, tr=1
  o1 <- res[res$id == 1L, ]
  expect_equal(o1$cost[o1$destination == "tl"], 1)
  expect_equal(o1$cost[o1$destination == "tr"], 3)
  o2 <- res[res$id == 2L, ]
  expect_equal(o2$cost[o2$destination == "tl"], 3)
  expect_equal(o2$cost[o2$destination == "tr"], 1)
  unlink(f)
})

test_that("route geometry reconstructs a path of the right length", {
  skip_if_no_sf()
  net <- spatial_network(lattice_streets())
  f <- write_points(data.frame(id = 1L, x = 0, y = 0))
  dest <- sf::st_sfc(sf::st_point(c(2, 1)))

  sfres <- tbl(f) |>
    spatial_route(net, to = dest, coords = c("x", "y")) |>
    collect_sf()
  expect_equal(nrow(sfres), 1L)
  expect_true(as.character(sf::st_geometry_type(sfres)) == "LINESTRING")
  # the reconstructed line length equals the routing cost
  expect_equal(as.numeric(sf::st_length(sfres)), 3, tolerance = 1e-9)
  unlink(f)
})

test_that("unreachable destination returns Inf cost, not a dropped row", {
  skip_if_no_sf()
  # two disconnected components: a left segment and a far-away right segment
  mk <- function(x1, y1, x2, y2)
    sf::st_linestring(rbind(c(x1, y1), c(x2, y2)))
  streets <- sf::st_sfc(mk(0, 0, 1, 0), mk(10, 0, 11, 0))
  net <- spatial_network(streets)
  expect_identical(net$stats[3], 2L)        # two components

  f <- write_points(data.frame(id = 1L, x = 0, y = 0))
  dest <- sf::st_sfc(sf::st_point(c(11, 0)))
  res <- tbl(f) |>
    spatial_route(net, to = dest, coords = c("x", "y"), geometry = FALSE) |>
    collect()
  expect_equal(nrow(res), 1L)
  expect_true(is.infinite(res$cost))
  unlink(f)
})

test_that("directed one-way edges block the reverse direction", {
  skip_if_no_sf()
  mk <- function(x1, y1, x2, y2)
    sf::st_linestring(rbind(c(x1, y1), c(x2, y2)))
  # a -> b -> c one-way chain (digitised direction is left to right)
  streets <- sf::st_sf(
    geometry = sf::st_sfc(mk(0, 0, 1, 0), mk(1, 0, 2, 0)))
  net <- spatial_network(streets, directed = TRUE)

  f <- write_points(data.frame(id = 1:2, x = c(0, 2), y = c(0, 0)))
  # forward origin (0,0) -> (2,0) reachable; reverse (2,0) -> (0,0) is not
  res <- tbl(f) |>
    spatial_route(net, to = sf::st_sfc(sf::st_point(c(2, 0)), sf::st_point(c(0, 0))),
                  geometry = FALSE, coords = c("x", "y")) |>
    collect()
  fwd <- res[res$id == 1L & res$destination == 1L, ]   # (0,0)->(2,0)
  rev <- res[res$id == 2L & res$destination == 2L, ]   # (2,0)->(0,0)
  expect_equal(fwd$cost, 2)
  expect_true(is.infinite(rev$cost))
  unlink(f)
})

test_that("service area reachable nodes match the budget", {
  skip_if_no_sf()
  net <- spatial_network(lattice_streets())
  f <- write_points(data.frame(id = 1L, x = 0, y = 0))

  sfres <- tbl(f) |>
    spatial_service_area(net, cost = 1, output = "nodes",
                         coords = c("x", "y")) |>
    collect_sf()
  expect_equal(nrow(sfres), 1L)
  pts <- sf::st_coordinates(sfres)[, c("X", "Y")]
  # within cost 1 of (0,0): itself, (1,0), (0,1)
  reached <- unique(pts)
  expect_equal(nrow(reached), 3L)
  unlink(f)
})

test_that("service area bands nest from small to large budget", {
  skip_if_no_sf()
  net <- spatial_network(lattice_streets())
  f <- write_points(data.frame(id = 1L, x = 0, y = 0))

  sfres <- tbl(f) |>
    spatial_service_area(net, cost = c(1, 2), output = "nodes",
                         coords = c("x", "y")) |>
    collect_sf()
  expect_equal(nrow(sfres), 2L)             # two bands
  n1 <- nrow(unique(sf::st_coordinates(sfres[sfres$band == 1, ])[, c("X", "Y")]))
  n2 <- nrow(unique(sf::st_coordinates(sfres[sfres$band == 2, ])[, c("X", "Y")]))
  expect_equal(n1, 3L)                       # itself + 2 neighbours
  expect_gte(n2, n1)                         # larger budget reaches at least as many
  unlink(f)
})

test_that("streaming is invariant to batch size", {
  skip_if_no_sf()
  net <- spatial_network(lattice_streets())
  pts <- data.frame(id = 1:6,
                    x = c(0, 1, 2, 0, 1, 2),
                    y = c(0, 0, 0, 1, 1, 1))
  dest <- sf::st_sfc(sf::st_point(c(2, 1)))

  f <- write_points(pts)
  one <- tbl(f) |>
    spatial_route(net, to = dest, coords = c("x", "y"), geometry = FALSE) |>
    collect()
  unlink(f)

  # force several row groups by writing a small batch_size
  f2 <- tempfile(fileext = ".vtr")
  write_vtr(pts, f2, batch_size = 2L)
  many <- tbl(f2) |>
    spatial_route(net, to = dest, coords = c("x", "y"), geometry = FALSE) |>
    collect()
  unlink(f2)

  one <- one[order(one$id), ]
  many <- many[order(many$id), ]
  expect_equal(many$cost, one$cost)
})

test_that("route matches igraph shortest paths when available", {
  skip_if_no_sf()
  skip_if_not_installed("igraph")
  net <- spatial_network(lattice_streets())

  # build the same undirected graph in igraph, weight 1 per edge
  edges <- rbind(
    c(1, 2), c(2, 3), c(4, 5), c(5, 6),      # node ids depend on snap order;
    c(1, 4), c(2, 5), c(3, 6))               # mirror lattice connectivity by coord
  # map lattice coords to vectra node ids via nearest-node snapping
  corner <- function(x, y) .network_snap(net, sf::st_sfc(sf::st_point(c(x, y))))
  ids <- list(bl = corner(0, 0), bm = corner(1, 0), br = corner(2, 0),
              tl = corner(0, 1), tm = corner(1, 1), tr = corner(2, 1))
  g <- igraph::make_empty_graph(n = net$n_nodes, directed = FALSE)
  add <- function(a, b) igraph::edges(c(a, b))
  el <- c(ids$bl, ids$bm, ids$bm, ids$br, ids$tl, ids$tm, ids$tm, ids$tr,
          ids$bl, ids$tl, ids$bm, ids$tm, ids$br, ids$tr)
  g <- igraph::add_edges(g, el)
  igraph::E(g)$weight <- 1

  f <- write_points(data.frame(id = 1L, x = 0, y = 0))
  res <- tbl(f) |>
    spatial_route(net, to = sf::st_sfc(sf::st_point(c(2, 1))),
                  geometry = FALSE, coords = c("x", "y")) |>
    collect()
  ig <- igraph::distances(g, v = ids$bl, to = ids$tr)[1, 1]
  expect_equal(res$cost, ig)
  unlink(f)
})

# spatial.R - streamed spatial operations via sf.
#
# vectra has no geometry type and no GEOS: topology stays sf's job. These verbs
# instead stream a larger-than-RAM query through the engine one batch at a time,
# hand each batch to sf for the actual geometry work, and spill the transformed
# batches back to disk as a fresh lazy node. Geometry rides through the engine as
# hex-encoded WKB in an ordinary string column, so no new VecType is needed; the
# CRS is carried on the returned node (the .vtr file stores no CRS).
#
# A shared run-file accumulator (.run_accumulator) plus the .spatial_stream
# engine drive the front doors:
#   spatial_map(x, fn)          per-feature transforms (buffer, transform, ...)
#   spatial_join(x, y, join)    big-x streamed against a resident small-y sf
#   spatial_filter(x, y, pred)  select-by-location: keep big-x rows matching y
#   spatial_clip(x, mask)       clip / erase big-x geometry against a mask
#   spatial_overlay(x)          self-overlay a resident polygon layer into pieces
# This mirrors offload(by = ...): batch cursor -> per-batch work -> run-files ->
# a ConcatNode with a finalizer that clears the temp spills.

.SPATIAL_FLUSH <- 5e5      # transformed rows buffered before a run-file flush

.check_sf <- function() {
  if (!requireNamespace("sf", quietly = TRUE))
    stop("spatial operations require the 'sf' package; install it with ",
         "install.packages(\"sf\")", call. = FALSE)
}

# Default sentinel-aware CRS resolution: an explicit `crs` wins; otherwise
# inherit the CRS the upstream node carries (so spatial_map() |> spatial_join()
# keeps the projection without re-stating it).
.resolve_crs <- function(x, crs) {
  if ((identical(crs, NA) || is.null(crs)) && !is.null(x$.crs)) return(x$.crs)
  crs
}

# -- geometry transport (hex WKB in a string column) --------------------------

# Decode one engine batch (a data.frame) into an sf object. Either a geometry
# column of hex WKB / WKT strings (`geom`), or point coordinates assembled from
# two numeric columns (`coords = c("x", "y")`), in which case the coordinate
# columns are kept so downstream work can still see them.
# Normalize any CRS input (including the logical-NA default) to an sf crs
# object, since sf rejects a bare logical NA where it wants NA_crs_.
.as_crs <- function(crs) {
  if (identical(crs, NA) || is.null(crs)) sf::st_crs(NA) else sf::st_crs(crs)
}

# Give a resident sf/sfc the stream's CRS when it carries none, so a per-batch
# predicate or overlay does not reject on an NA-vs-known CRS mismatch. When both
# sides already carry a CRS they are left untouched and sf enforces the match.
.align_resident_crs <- function(y, crs) {
  cc <- .as_crs(crs)
  if (is.na(sf::st_crs(y)) && !is.na(cc)) y <- sf::st_set_crs(y, cc)
  y
}

# Reduce a CRS (an EPSG integer, an sf crs object, a WKT/proj string) to an
# EPSG code for the raster header, or 0 when none is known. Only touches sf when
# the input is not already a bare integer, so the coordinate-only rasterize path
# stays sf-free.
.crs_to_epsg <- function(crs) {
  if (is.null(crs) || (length(crs) == 1L && is.na(crs))) return(0L)
  if (is.numeric(crs) && length(crs) == 1L) return(as.integer(crs))
  if (requireNamespace("sf", quietly = TRUE)) {
    e <- tryCatch(sf::st_crs(crs)$epsg, error = function(e) NA_integer_)
    if (!is.null(e) && !is.na(e)) return(as.integer(e))
  }
  0L
}

# Resolve the target grid for rasterize() into width/height + a north-up GDAL
# geotransform. A `vectra_raster` template borrows its geometry and CRS; an
# explicit extent is paired with either `dims = c(nrow, ncol)` or `res` (a
# scalar or c(xres, yres)). With `res`, the cell counts are rounded to fit the
# extent exactly, so the extent stays authoritative and cells stay uniform.
.rasterize_grid <- function(template, extent, res, dims, crs) {
  if (inherits(template, "vectra_raster")) {
    gt   <- template$gt
    w    <- as.integer(template$width)
    h    <- as.integer(template$height)
    tcrs <- if (!is.null(template$epsg) && template$epsg > 0L)
      as.integer(template$epsg) else NA
    if (is.null(crs) || (length(crs) == 1L && is.na(crs))) crs <- tcrs
    xres <- gt[2L]; yres <- -gt[6L]
    ext  <- c(gt[1L], gt[4L] - yres * h, gt[1L] + xres * w, gt[4L])
    return(list(width = w, height = h, gt = gt, crs = crs,
                epsg = .crs_to_epsg(crs), extent = ext, res = c(xres, yres)))
  }
  if (is.numeric(template) && length(template) == 4L && is.null(extent))
    extent <- template
  if (is.null(extent) || !is.numeric(extent) || length(extent) != 4L)
    stop("supply `template` (a vectra_raster) or ",
         "`extent = c(xmin, ymin, xmax, ymax)`")
  xmin <- extent[1L]; ymin <- extent[2L]; xmax <- extent[3L]; ymax <- extent[4L]
  if (!(xmax > xmin) || !(ymax > ymin))
    stop("extent must have xmax > xmin and ymax > ymin")
  if (!is.null(dims)) {
    if (length(dims) != 2L) stop("dims must be c(nrow, ncol)")
    h <- as.integer(dims[1L]); w <- as.integer(dims[2L])
  } else if (!is.null(res)) {
    if (length(res) == 1L) res <- c(res, res)
    w <- as.integer(round((xmax - xmin) / res[1L]))
    h <- as.integer(round((ymax - ymin) / res[2L]))
  } else {
    stop("supply either `res` or `dims` together with `extent`")
  }
  if (is.na(w) || is.na(h) || w <= 0L || h <= 0L)
    stop("the derived grid has non-positive dimensions")
  xres <- (xmax - xmin) / w
  yres <- (ymax - ymin) / h
  gt <- c(xmin, xres, 0, ymax, 0, -yres)
  list(width = w, height = h, gt = gt, crs = crs, epsg = .crs_to_epsg(crs),
       extent = c(xmin, ymin, xmax, ymax), res = c(xres, yres))
}

.sf_decode_chunk <- function(chunk, geom, coords, crs) {
  crs <- .as_crs(crs)
  if (!is.null(coords)) {
    miss <- setdiff(coords, names(chunk))
    if (length(miss))
      stop(sprintf("coords column(s) not found: %s",
                   paste(miss, collapse = ", ")))
    return(sf::st_as_sf(chunk, coords = coords, crs = crs, remove = FALSE))
  }
  if (!geom %in% names(chunk))
    stop(sprintf("geometry column '%s' not found; pass geom= or coords=", geom))
  # st_as_sfc.character parses WKT only; hex WKB must be tagged so the WKB
  # reader runs, which round-trips coordinates losslessly (WKT would round to
  # the print precision).
  g <- sf::st_as_sfc(structure(chunk[[geom]], class = "WKB"), EWKB = FALSE)
  g <- sf::st_set_crs(g, crs)
  rest <- chunk[setdiff(names(chunk), geom)]
  sf::st_sf(rest, geometry = g)
}

# Coerce a data.frame's columns to types the .vtr writer accepts (int64 /
# double / bool / string). sf attribute columns can arrive as factors, dates, or
# units; those collapse to character. List columns other than the geometry we
# already encoded are an error.
.coerce_for_vtr <- function(df) {
  for (nm in names(df)) {
    col <- df[[nm]]
    if (is.factor(col) || inherits(col, "Date") || inherits(col, "POSIXt")) {
      df[[nm]] <- as.character(col)
    } else if (is.integer(col) || is.double(col) ||
               is.logical(col) || is.character(col)) {
      # already writable
    } else if (is.list(col)) {
      stop(sprintf("column '%s' is a list column and cannot be written; ",
                   "drop it before the spatial step", nm))
    } else {
      df[[nm]] <- as.character(col)
    }
  }
  df
}

# Encode the per-batch sf result back into a writable data.frame: the active
# geometry becomes hex WKB in `out_geom`, the rest are coerced attribute columns.
# A plain data.frame (geometry deliberately dropped, e.g. a summary) passes
# through unchanged apart from type coercion.
.sf_encode_result <- function(res, out_geom) {
  if (inherits(res, "sfc")) res <- sf::st_sf(geometry = res)
  if (!inherits(res, "sf")) {
    if (is.data.frame(res)) return(.coerce_for_vtr(as.data.frame(res)))
    stop("spatial batch function must return an sf object, sfc, or data.frame")
  }
  g  <- sf::st_geometry(res)
  df <- as.data.frame(sf::st_drop_geometry(res))
  df[[out_geom]] <- sf::st_as_binary(g, hex = TRUE)
  .coerce_for_vtr(df)
}

# -- native libgeos compute paths ---------------------------------------------

# Worker count for the per-batch GEOS loops, mirroring the overlay convention.
.spatial_threads <- function()
  max(as.integer(getOption("vectra.spatial_threads",
                           min(parallel::detectCores(), 8L))), 1L)

# Parse a resident sf/sfc side once into a GEOS locator (an external pointer
# holding the parsed geometries plus an STRtree), reused across every streamed
# batch. Geometry crosses to C as raw WKB; sf is touched only here, on the small
# resident side, not in the per-batch loop.
.geos_locator <- function(y) {
  g <- if (inherits(y, "sf")) sf::st_geometry(y) else y
  .Call(C_geos_locator_build, sf::st_as_binary(g, EWKB = FALSE))
}

# Native GEOS computes in the planar frame. sf computes geographic coordinates
# on the sphere (s2) when `sf_use_s2()` is on, so the native path matches the sf
# path only for projected data, or when s2 is off (both then planar via GEOS),
# or when the CRS is unknown (sf is planar too). Geographic data with s2 on
# keeps the spherical sf path.
.geos_planar_ok <- function(crs) {
  ll <- tryCatch(sf::st_is_longlat(crs), error = function(e) NA)
  if (is.na(ll)) return(TRUE)
  !ll || !sf::sf_use_s2()
}

# Map an sf binary-predicate function to the native predicate code the C filter
# understands, or NA when it is one the native path does not cover (the caller
# then falls back to the sf loop). NULL means the default, st_intersects.
.geos_pred_code <- function(predicate) {
  if (is.null(predicate)) return(0L)
  known <- c(st_intersects = 0L, st_within = 1L, st_contains = 2L,
             st_overlaps = 3L, st_covers = 4L, st_covered_by = 5L,
             st_touches = 6L, st_crosses = 7L, st_equals = 8L,
             st_disjoint = 9L, st_is_within_distance = 10L)
  for (nm in names(known))
    if (identical(predicate, getExportedValue("sf", nm))) return(unname(known[[nm]]))
  NA_integer_
}

# Pull the two coordinate columns from a batch as doubles (the raw-point input to
# the native locate kernel), erroring if either is missing.
.batch_xy <- function(chunk, coords) {
  miss <- setdiff(coords, names(chunk))
  if (length(miss))
    stop(sprintf("coords column(s) not found: %s", paste(miss, collapse = ", ")))
  list(x = as.numeric(chunk[[coords[1L]]]), y = as.numeric(chunk[[coords[2L]]]))
}

# Build one streamed spatial-join batch from the per-row match lists the C join
# returns. `matches` is a list with one integer vector of 1-based resident-row
# indices per left row; each left row is replicated once per match and the
# resident attributes `y_attrs` attached. With `left`, an unmatched left row is
# kept once padded with NA resident columns; otherwise it is dropped. Attribute
# names shared by both sides get `suffix`, mirroring [sf::st_join()]. The left
# geometry column rides through untouched (the left side is never decoded).
.geos_join_assemble <- function(chunk, matches, y_attrs, left, suffix) {
  nmatch <- lengths(matches)
  reps   <- if (left) pmax(nmatch, 1L) else nmatch
  left_idx  <- rep.int(seq_len(nrow(chunk)), reps)
  right_idx <- unlist(lapply(seq_along(matches), function(i) {
    mi <- matches[[i]]
    if (length(mi)) mi else if (left) NA_integer_ else integer(0)
  }), use.names = FALSE)
  if (is.null(right_idx)) right_idx <- integer(0)

  lout <- chunk[left_idx, , drop = FALSE]
  rout <- y_attrs[right_idx, , drop = FALSE]
  shared <- intersect(names(lout), names(rout))
  if (length(shared)) {
    names(lout)[match(shared, names(lout))] <- paste0(shared, suffix[1L])
    names(rout)[match(shared, names(rout))] <- paste0(shared, suffix[2L])
  }
  out <- cbind(lout, rout, stringsAsFactors = FALSE)
  rownames(out) <- NULL
  out
}

# -- self-overlay (QGIS-style Union) ------------------------------------------

# Fixed-precision grid (in CRS units) for the self-overlay. Snapping input
# coordinates to a grid this fine merges the near-duplicate boundaries that
# overlapping polygons share, so the noded arrangement has no sliver faces. The
# grid is derived from coordinate magnitude (not extent), keeping it stable
# against far-flung outliers and invariant to units. Returns 0 when no sensible
# grid exists (degenerate or all-zero coordinates), which disables snapping.
.overlay_grid_mag <- function(mag) {
  if (!is.finite(mag) || mag == 0) return(0)
  mag * 3e-8
}
.overlay_grid <- function(x) .overlay_grid_mag(max(abs(sf::st_bbox(x))))

# Per-input coverage error of an overlay result. For a correct disjoint
# decomposition the piece areas covering each input must sum to that input's
# area; this is the invariant the snapping grid is verified against. Uses the
# piece and input areas the C engine returns, so no geometry is re-measured.
# Returns a data.frame keyed by the input's chunk-local index, so the caller can
# map back to global rows and name the worst offenders, not just the scalar max.
.overlay_coverage_detail <- function(origin, parea, iarea) {
  if (!length(origin))
    return(data.frame(src = integer(0), cov = numeric(0),
                      iarea = numeric(0), err = numeric(0)))
  agg <- rowsum(parea, group = origin)
  cov <- numeric(length(iarea))
  cov[as.integer(rownames(agg))] <- agg[, 1L]
  data.frame(src = seq_along(iarea), cov = cov, iarea = iarea,
             err = abs(cov - iarea) / pmax(iarea, 1))
}

# Memory model. The overlay cost of a tile is driven by the noding of its
# boundary linework, which can balloon far past the input where polygons overlap
# densely. The layer is therefore tiled so no tile is large enough to blow up: a
# tile that lands on a dense blob keeps subdividing. Peak memory is bounded by
# `mem_limit`; lowering it tightens memory at the cost of more tiles.
#
# Tile size is a throughput knob with an interior optimum, not "bigger is faster".
# Smaller tiles replicate a feature spanning several tiles into each (more clip and
# parse work), while larger tiles node more linework per tile, and noding cost is
# superlinear in the linework gathered. Measured on a dense ~470k-feature layer the
# total wall time is flat for per-tile inputs around 1-5 MB and rises on either
# side; a budget far above that (tens of GB of working set) runs slower, not faster.
# The default therefore targets that per-tile size and lets the budget scale with
# the thread count -- more threads run more tiles at once, so the working set grows
# with parallelism -- rather than scaling with installed RAM.
.OVERLAY_TILE_TGT  <- 2.6e6 # target per-tile input bytes (throughput sweet spot)
.OVERLAY_FACTOR    <- 24    # per-tile working set ~ tile input bytes * this
.OVERLAY_FEAT_CAP  <- 3000  # also cap features per tile (node count, not just bytes)
.OVERLAY_MAX_DEPTH <- 16    # quadtree depth cap (coincident features can't split)

# Total physical RAM in bytes, or NA if it cannot be determined. Cached per
# session. Used only to RELAX limits when memory is plentiful and to cap them
# when it is scarce; every caller has a safe fixed fallback for the NA case, so
# nothing about correctness or the peak guarantee depends on detection working.
.sys_ram_cache <- new.env(parent = emptyenv())
.sys_ram_bytes <- function() {
  if (!is.null(.sys_ram_cache$v)) return(.sys_ram_cache$v)
  v <- tryCatch({
    sys <- Sys.info()[["sysname"]]
    if (identical(sys, "Linux")) {
      mt <- grep("^MemTotal:", readLines("/proc/meminfo", n = 64L), value = TRUE)
      if (length(mt)) as.numeric(sub("\\D+(\\d+).*", "\\1", mt[1L])) * 1024 else NA_real_
    } else if (identical(sys, "Darwin")) {
      as.numeric(system2("sysctl", c("-n", "hw.memsize"), stdout = TRUE, stderr = FALSE))
    } else if (identical(sys, "Windows")) {
      pick <- function(out) {
        num <- suppressWarnings(as.numeric(grep("[0-9]", out, value = TRUE)))
        num <- num[is.finite(num) & num > 0]
        if (length(num)) num[1L] else NA_real_
      }
      out <- tryCatch(system2("wmic", c("ComputerSystem", "get", "TotalPhysicalMemory"),
                              stdout = TRUE, stderr = FALSE), error = function(e) character(0))
      r <- pick(out)
      if (is.na(r)) {  # wmic is absent on newer Windows; fall back to CIM
        out <- tryCatch(system2("powershell",
          c("-NoProfile", "-Command",
            "(Get-CimInstance Win32_ComputerSystem).TotalPhysicalMemory"),
          stdout = TRUE, stderr = FALSE), error = function(e) character(0))
        r <- pick(out)
      }
      r
    } else NA_real_
  }, error = function(e) NA_real_)
  if (length(v) != 1L || !is.finite(v) || v <= 0) v <- NA_real_
  .sys_ram_cache$v <- v
  v
}

# Default overlay working-set budget. tile_bytes = mem / (threads * FACTOR), so to
# hold tile_bytes at the target the budget scales with the thread count. Peak
# working set is then target * FACTOR * threads, modest and proportional to the
# parallelism actually used. When total RAM is known the budget is also capped at
# half of it, so a many-core / low-RAM machine cannot scale the working set past
# what it can hold. An explicit mem_limit always wins.
.overlay_mem_default <- function(nthreads) {
  thread_scaled <- .OVERLAY_TILE_TGT * .OVERLAY_FACTOR * max(as.integer(nthreads), 1L)
  ram <- .sys_ram_bytes()
  if (is.na(ram)) thread_scaled else min(thread_scaled, 0.5 * ram)
}

# Features per parse chunk. The parse materializes the raw WKB for its chunk only,
# so the chunk count bounds the transient input copy that coexists with the
# accumulating cleaned WKB. Scales with total RAM (bigger chunks, fewer round
# trips, on roomy machines) around a 16 GB reference, with a safe fixed fallback.
.OVERLAY_PARSE_CHUNK <- 50000L
# RAM-scaled read/parse batch size. Keeping peak low is the goal, so this only
# scales DOWN on a memory-constrained machine (a bigger batch on a roomy machine
# just saves a few read round trips while raising the peak, which defeats the
# point); the batch never grows past the reference size.
.overlay_chunk_ram <- function() {
  ram <- .sys_ram_bytes()
  if (is.na(ram)) .OVERLAY_PARSE_CHUNK
  else as.integer(.OVERLAY_PARSE_CHUNK * max(0.25, min(1, ram / 16e9)))
}
# Batch size for reads/parse, RAM-scaled, overridable, and (sf path) capped at n.
.overlay_chunk_size <- function()
  as.integer(getOption("vectra.overlay_parse_chunk", .overlay_chunk_ram()))
.overlay_parse_chunk <- function(n)
  max(1L, min(as.integer(n), .overlay_chunk_size()))

# Coordinate magnitude of a GeoPackage layer from its declared extent
# (gpkg_contents), or NA if it cannot be read. Lets the file path derive the
# snapping grid without scanning geometry.
.overlay_source_mag <- function(path, layer) {
  ext <- tryCatch(
    sf::st_read(path, quiet = TRUE,
      query = sprintf(
        "SELECT min_x, min_y, max_x, max_y FROM gpkg_contents WHERE table_name = '%s'",
        layer)),
    error = function(e) NULL)
  if (is.null(ext) || nrow(ext) == 0L) return(NA_real_)
  v <- suppressWarnings(as.numeric(unlist(ext[1L, c("min_x", "min_y", "max_x", "max_y")])))
  v <- v[is.finite(v)]
  if (!length(v)) NA_real_ else max(abs(v))
}

# Common input path for spatial_overlay(): produce cleaned geometry (cwkb),
# bounding boxes, and attributes from either an in-memory sf object or a file
# source read in feature batches. Each batch is parsed once and the raw WKB is
# released before the next, so the transient input copy stays small. The file
# path never holds the whole layer as an sf object, so peak memory tracks the
# cleaned geometry rather than the (much larger) cost of materializing the source.
.overlay_ingest <- function(x, vars, piece, grid, nthreads,
                            layer = NULL, query = NULL, read_chunk = NULL) {
  sel_attrs <- function(df) {
    df <- as.data.frame(df)
    if (!is.null(vars)) {
      miss <- setdiff(vars, names(df))
      if (length(miss))
        stop(sprintf("vars not found in `x`: %s", paste(miss, collapse = ", ")))
      df <- df[, vars, drop = FALSE]
    }
    if (piece %in% names(df))
      stop(sprintf("piece column '%s' already exists in `x`; pass piece=", piece))
    df
  }

  if (inherits(x, "sf")) {
    crs <- sf::st_crs(x)
    n   <- nrow(x)
    if (n == 0L) stop("`x` has no geometries to overlay")
    attrs <- sel_attrs(sf::st_drop_geometry(x))
    if (is.null(grid)) grid <- .overlay_grid_mag(max(abs(sf::st_bbox(x))))
    xgeom <- sf::st_geometry(x)
    bbox  <- matrix(NA_real_, n, 4L)
    cwkb  <- vector("list", n)
    chunk <- read_chunk %||% .overlay_parse_chunk(n)
    for (s in seq.int(1L, n, by = chunk)) {
      e  <- min(s + chunk - 1L, n)
      w  <- sf::st_as_binary(xgeom[s:e], EWKB = FALSE)
      pr <- .Call(C_overlay_parse, w, as.double(grid), as.integer(nthreads))
      bbox[s:e, ] <- pr[[1L]]
      cwkb[s:e]   <- pr[[2L]]
      w <- NULL; pr <- NULL
    }
    return(list(attrs = attrs, crs = crs, n = n,
                cwkb = cwkb, bbox = bbox, grid = as.double(grid)))
  }

  if (!is.character(x) || length(x) != 1L)
    stop("`x` must be an sf object or a single file path")
  if (!file.exists(x)) stop(sprintf("file not found: %s", x))
  if (is.null(layer) && is.null(query))
    stop("reading from a file needs `layer=` (a layer name) or `query=` (SQL)")
  make_q <- function(off) {
    if (is.null(query))
      sprintf('SELECT * FROM "%s" LIMIT %d OFFSET %d', layer, chunk, off)
    else
      sprintf("SELECT * FROM (%s) LIMIT %d OFFSET %d", query, chunk, off)
  }

  if (is.null(grid)) {
    mag <- if (!is.null(layer)) .overlay_source_mag(x, layer) else NA_real_
    if (is.na(mag))
      stop("cannot derive the snapping grid from this source; pass grid= explicitly")
    grid <- .overlay_grid_mag(mag)
  }

  chunk <- read_chunk %||% .overlay_chunk_size()
  cw <- list(); bb <- list(); at <- list(); crs <- NULL; off <- 0L; ci <- 0L
  repeat {
    b  <- sf::st_read(x, query = make_q(off), quiet = TRUE)
    nb <- nrow(b)
    if (nb == 0L) break
    if (is.null(crs)) crs <- sf::st_crs(b)
    ci <- ci + 1L
    at[[ci]] <- sel_attrs(sf::st_drop_geometry(b))
    w  <- sf::st_as_binary(sf::st_geometry(b), EWKB = FALSE)
    pr <- .Call(C_overlay_parse, w, as.double(grid), as.integer(nthreads))
    bb[[ci]] <- pr[[1L]]; cw[[ci]] <- pr[[2L]]
    off <- off + nb
    b <- NULL; w <- NULL; pr <- NULL
    if (nb < chunk) break
  }
  if (ci == 0L) stop("the source returned no features to overlay")
  cwkb <- unlist(cw, recursive = FALSE, use.names = FALSE)
  list(attrs = do.call(rbind, at), crs = crs, n = length(cwkb),
       cwkb = cwkb, bbox = do.call(rbind, bb), grid = as.double(grid))
}

# Adaptive quadtree over feature bounding boxes: subdivide the extent until each
# leaf tile is within the byte and feature budgets, then return the leaves as
# overlay jobs (`idx` = global rows whose bbox meets the tile, `rect` = clip
# bounds). A feature spanning a tile edge is replicated into both tiles and
# clipped there, so pieces are split along the grid but coverage and labels stay
# exact.
.overlay_tiles <- function(bbox, idx, rect, budget_bytes, budget_feat, wbytes, depth) {
  if (length(idx) <= 1L ||
      depth >= .OVERLAY_MAX_DEPTH ||
      (length(idx) <= budget_feat && sum(wbytes[idx]) <= budget_bytes))
    return(list(list(idx = idx, rect = rect)))
  xm <- (rect[1L] + rect[3L]) / 2; ym <- (rect[2L] + rect[4L]) / 2
  quads <- list(c(rect[1L], rect[2L], xm, ym), c(xm, rect[2L], rect[3L], ym),
                c(rect[1L], ym, xm, rect[4L]), c(xm, ym, rect[3L], rect[4L]))
  # Index this cell's features only; pulling whole bbox columns here would copy the
  # entire layer at every quadtree node.
  xmin <- bbox[idx, 1L]; ymin <- bbox[idx, 2L]; xmax <- bbox[idx, 3L]; ymax <- bbox[idx, 4L]
  sels <- lapply(quads, function(q)
    idx[xmin <= q[3L] & xmax >= q[1L] & ymin <= q[4L] & ymax >= q[2L]])
  # Stop if splitting does not separate the features (large mutually-overlapping
  # features whose bounding boxes all span the cell): subdividing would replicate
  # them into every child without shrinking any, exploding the tile count. The
  # cell stays one tile; clipping still shrinks each feature to it.
  if (max(vapply(sels, length, integer(1))) >= length(idx))
    return(list(list(idx = idx, rect = rect)))
  out <- list()
  for (i in seq_along(quads))
    if (length(sels[[i]]))
      out <- c(out, .overlay_tiles(bbox, sels[[i]], quads[[i]],
                                   budget_bytes, budget_feat, wbytes, depth + 1L))
  out
}

# -- run-file accumulator (shared spill machinery) ----------------------------

# Buffer writable data.frames, flush them to `.vtr` spill files once `flush_rows`
# rows are pending, and finalize into a lazy ConcatNode whose temporary spills
# are removed when the node is garbage-collected. This is the single place the
# streamed spatial verbs (map / join / filter / clip / overlay) turn a sequence
# of per-batch data.frames into a lazy node. `push()` records the schema from the
# first frame it sees (even an empty one), so `finish()` can emit a correctly
# typed empty node when nothing matched. Returns a list of `push` and `finish`.
.run_accumulator <- function(flush_rows) {
  st <- new.env(parent = emptyenv())
  st$buf <- list(); st$buffered <- 0L; st$runs <- character(0); st$template <- NULL

  flush <- function() {
    if (!length(st$buf)) return(invisible())
    df <- if (length(st$buf) == 1L) st$buf[[1]] else do.call(rbind, st$buf)
    rf <- tempfile(fileext = ".vtr"); write_vtr(df, rf)
    st$runs <- c(st$runs, rf); st$buf <- list(); st$buffered <- 0L
  }

  push <- function(df) {
    if (is.null(st$template)) st$template <- df[0, , drop = FALSE]
    if (nrow(df)) {
      st$buf <- c(st$buf, list(df))
      st$buffered <- st$buffered + nrow(df)
      if (st$buffered >= flush_rows) flush()
    }
    invisible()
  }

  finish <- function(crs, empty_geom = "geometry") {
    flush()
    if (!length(st$runs)) {
      tmpl <- st$template
      if (is.null(tmpl))
        tmpl <- stats::setNames(
          data.frame(character(0), stringsAsFactors = FALSE), empty_geom)
      rf <- tempfile(fileext = ".vtr"); write_vtr(tmpl, rf); st$runs <- rf
    }
    node <- .concat_runs(st$runs)
    reg <- new.env(parent = emptyenv()); reg$paths <- st$runs
    reg.finalizer(reg, function(e) try(unlink(e$paths), silent = TRUE),
                  onexit = TRUE)
    node$.reg <- reg
    node$.crs <- crs
    node
  }

  list(push = push, finish = finish)
}

# -- the streaming engine -----------------------------------------------------

# Pull `x` one batch at a time, run `batch_fn` (an sf-in / sf-out function) on
# each, encode the result, and accumulate to run-files. Returns a lazy ConcatNode
# carrying the output CRS (the first sf/sfc result's CRS, else the input `crs`).
.spatial_stream <- function(x, batch_fn, geom, coords, crs, out_geom,
                            flush_rows) {
  nxt <- .batch_cursor(x)
  acc <- .run_accumulator(flush_rows)
  out_crs <- NULL
  repeat {
    chunk <- nxt(); if (is.null(chunk)) break
    sb  <- .sf_decode_chunk(chunk, geom, coords, crs)
    res <- batch_fn(sb)
    if (is.null(out_crs) && (inherits(res, "sf") || inherits(res, "sfc")))
      out_crs <- sf::st_crs(res)
    acc$push(.sf_encode_result(res, out_geom))
  }
  acc$finish(crs = if (!is.null(out_crs)) out_crs else crs, empty_geom = out_geom)
}

# Decompose one sfg into its single-part components: a MULTI* into its parts, a
# GEOMETRYCOLLECTION recursively into the parts of its members, and an already
# single-part geometry into itself. An empty multi-geometry has no parts, so it
# passes through as a single row rather than vanishing.
.parts_of <- function(g) {
  out <- switch(class(g)[2L],
    MULTIPOLYGON    = lapply(unclass(g), sf::st_polygon),
    MULTILINESTRING = lapply(unclass(g), sf::st_linestring),
    MULTIPOINT      = {
      m <- unclass(g)
      if (!length(m)) list()
      else lapply(seq_len(nrow(m)), function(i) sf::st_point(m[i, ]))
    },
    GEOMETRYCOLLECTION = unlist(lapply(g, .parts_of), recursive = FALSE),
    list(g))
  if (length(out)) out else list(g)
}

# Explode one decoded sf batch: replace each multipart feature with one row per
# part, replicating the attributes. `part` (when set) numbers the parts within
# each input feature, 1-based.
.explode_batch <- function(sb, part) {
  if (!is.null(part) && part %in% names(sb))
    stop(sprintf("part column '%s' already exists in `x`; pass part=", part))
  if (nrow(sb) == 0L) {
    if (!is.null(part)) sb[[part]] <- integer(0)
    return(sb)
  }
  g     <- sf::st_geometry(sb)
  parts <- lapply(g, .parts_of)
  np    <- lengths(parts)
  out   <- sb[rep.int(seq_len(nrow(sb)), np), , drop = FALSE]
  sf::st_geometry(out) <- sf::st_sfc(unlist(parts, recursive = FALSE),
                                     crs = sf::st_crs(g))
  if (!is.null(part))
    out[[part]] <- unlist(lapply(np, seq_len), use.names = FALSE)
  out
}

# -- front doors --------------------------------------------------------------

#' Stream a query through an sf transform
#'
#' Applies a per-feature \pkg{sf} operation (buffer, centroid, area, CRS
#' transform, simplify, ...) to a lazy vectra query one batch at a time and
#' returns a new lazy node. The engine pulls one batch, hands it to `fn` as an
#' `sf` object, encodes the result back into the stream, and spills to disk, so
#' peak memory is one batch regardless of result size. This is the streaming,
#' larger-than-RAM counterpart to running the same `sf` call on a whole
#' in-memory table.
#'
#' Geometry travels through the engine as hex-encoded WKB in an ordinary string
#' column (vectra has no native geometry type), and the coordinate reference
#' system is carried on the returned node rather than in the `.vtr` file. Use
#' [collect_sf()] to materialize the result as an `sf` object, or [collect()]
#' to get the underlying data.frame with the WKB string column.
#'
#' Topology is delegated entirely to \pkg{sf}/GEOS; vectra only supplies the
#' streaming. The `sf` package is an optional dependency (Suggests).
#'
#' @param x A `vectra_node` (from [tbl()], [tbl_tiff()], any verb chain, ...).
#'   It is consumed by the stream.
#' @param fn A function (or purrr-style formula such as `~ sf::st_buffer(.x,
#'   1000)`) taking one `sf` batch and returning an `sf` object, `sfc`, or plain
#'   data.frame. The active geometry of the return becomes the output geometry.
#' @param geom Name of the input geometry column holding hex-WKB or WKT strings.
#'   Default `"geometry"`. Ignored when `coords` is given.
#' @param coords Optional length-2 character vector naming the x and y
#'   coordinate columns to assemble point geometry from (e.g. `c("x", "y")`),
#'   for inputs such as [tiff_extract_points()] output. The coordinate columns
#'   are retained.
#' @param crs Coordinate reference system of the input geometry, in any form
#'   [sf::st_crs()] accepts (EPSG integer, WKT, proj string). Defaults to the
#'   CRS the upstream node carries, or unknown.
#' @param out_geom Name of the output geometry column. Defaults to `geom`
#'   (or `"geometry"` when `coords` is used).
#' @param flush_rows Transformed rows buffered before a spill flush. Larger
#'   values mean fewer, bigger temporary files. Defaults to
#'   `getOption("vectra.spatial_flush", 5e5)`.
#'
#' @return A `vectra_node` backed by temporary `.vtr` spills (removed when the
#'   node is garbage-collected), carrying the output CRS for [collect_sf()].
#'
#' @seealso [spatial_join()] to join a streamed side against a resident `sf`
#'   object, [collect_sf()] to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   NAME = nc$NAME,
#'   geometry = sf::st_as_binary(sf::st_centroid(sf::st_geometry(nc)),
#'                               hex = TRUE)
#' ), f)
#'
#' # Buffer every county centroid by 0.1 degree, streaming.
#' buffered <- tbl(f) |>
#'   spatial_map(~ sf::st_buffer(.x, 0.1), crs = sf::st_crs(nc))
#' collect_sf(buffered)
#' unlink(f)
#'
#' @export
spatial_map <- function(x, fn, geom = "geometry", coords = NULL, crs = NA,
                        out_geom = NULL, flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (build one with tbl(), tbl_csv(), ...)")
  fn <- rlang::as_function(fn)
  crs <- .resolve_crs(x, crs)
  if (is.null(out_geom)) out_geom <- if (is.null(coords)) geom else "geometry"
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  .spatial_stream(x, fn, geom, coords, crs, out_geom, fr)
}

#' Explode multipart geometries into single-part features
#'
#' Streams a lazy vectra query and splits every multipart geometry into its
#' component single-part geometries: a `MULTIPOLYGON` becomes one row per
#' polygon, a `MULTILINESTRING` one row per linestring, a `MULTIPOINT` one row
#' per point, and a `GEOMETRYCOLLECTION` one row per member (recursively). The
#' attributes of the source feature are copied onto each part. Already
#' single-part geometries pass through unchanged, as does an empty geometry
#' (kept as one row). This is the streaming counterpart of the QGIS
#' "multipart to singleparts" tool and of [sf::st_cast()] to a single-part type.
#'
#' One batch is decoded, exploded, and spilled at a time, so peak memory tracks
#' one batch and its parts, not the whole layer.
#'
#' @inheritParams spatial_map
#' @param part Optional name of an integer column numbering the parts within each
#'   source feature, 1-based. Default `NULL` adds no such column.
#'
#' @return A `vectra_node` of single-part features, backed by temporary `.vtr`
#'   spills (removed when the node is garbage-collected) and carrying the input
#'   CRS.
#'
#' @seealso [spatial_map()] for per-feature transforms, [spatial_dissolve()] to
#'   merge features the other way, [collect_sf()] to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' mp <- sf::st_multipolygon(list(
#'   list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))),
#'   list(rbind(c(2, 2), c(3, 2), c(3, 3), c(2, 3), c(2, 2)))))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   id = 1L,
#'   geometry = sf::st_as_binary(sf::st_sfc(mp), hex = TRUE)
#' ), f)
#'
#' # One row per polygon, attributes copied, parts numbered.
#' tbl(f) |> spatial_explode(part = "part_id") |> collect_sf()
#' unlink(f)
#'
#' @export
spatial_explode <- function(x, geom = "geometry", crs = NA, out_geom = NULL,
                            part = NULL, flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (build one with tbl(), tbl_csv(), ...)")
  if (!is.null(part) && (!is.character(part) || length(part) != 1L))
    stop("`part` must be a single column name, or NULL")
  crs <- .resolve_crs(x, crs)
  if (is.null(out_geom)) out_geom <- geom
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  .spatial_stream(x, function(sb) .explode_batch(sb, part),
                  geom, coords = NULL, crs = crs, out_geom = out_geom,
                  flush_rows = fr)
}

#' Spatial join a streamed query against a resident sf object
#'
#' Streams a large left side `x` through the engine and joins each batch against
#' a small right side `y` held resident in memory, using an \pkg{sf} binary
#' predicate (`st_intersects` by default). This is the spatial analogue of a
#' hash join with the small side on the build side: the billion-row left stream
#' never materializes, while `y` (admin polygons, habitat patches, ...) stays in
#' RAM. The dominant real workload it serves is tagging huge point sets with the
#' polygon they fall in.
#'
#' For the recognised predicates -- the topological ones (intersects, within,
#' contains, overlaps, covers, covered by, touches, crosses), equals,
#' within-distance ([sf::st_is_within_distance], radius passed as `dist =`), and
#' nearest feature ([sf::st_nearest_feature]) -- on projected or unprojected
#' planar data, the match runs natively on the GEOS C API straight off the
#' hex-WKB column: `y` is parsed once into a spatial index, each batch's matches
#' come back from C, and `y`'s attributes are attached in R without decoding the
#' left side to \pkg{sf}. Coordinate-assembled (`coords`) point input runs
#' natively too, building each point in C (the emitted point geometry is built in
#' C as well). Geographic coordinates with spherical geometry on
#' (`sf::sf_use_s2()`), a disjoint join (whose matches are the bounding-box
#' complement an index cannot prune), and other extra [sf::st_join()] arguments
#' use \pkg{sf} instead, preserving its semantics.
#'
#' When both sides are larger than RAM, pass `partition = grid(cellsize)` and a
#' streamed `vectra_node` as `y`: both inputs are binned to a uniform spatial
#' grid, then joined one shard at a time. Each left feature is assigned to the
#' single grid cell of its reference point while each right feature is
#' replicated to every cell its bounding box overlaps, so a left row is emitted
#' exactly once and the result equals the resident join. This is exact for point
#' left geometries (the dominant case -- tagging a huge point set with the
#' polygon it falls in) and finds, for an extended left feature, the matches
#' whose right bounding box overlaps the left reference cell; choose a `cellsize`
#' larger than the left features for an extended-on-extended join. The partition
#' path serves topological predicates (intersects, within, contains, overlaps,
#' covers, covered by). It also serves [sf::st_nearest_feature]: because nearest
#' is not local to one cell, each left feature then searches its own cell and the
#' eight around it, so the true nearest is found when it lies within one cell of
#' the left reference cell (pick a `cellsize` at least the largest expected
#' nearest distance). Topology and CRS handling are \pkg{sf}'s; vectra supplies
#' the stream and the grid partition.
#'
#' @inheritParams spatial_map
#' @param y The right side of the join: an `sf` object held resident (the
#'   default), or -- when `partition` is given -- a streamed `vectra_node`.
#' @param join An \pkg{sf} binary predicate function, e.g. [sf::st_intersects]
#'   (default), [sf::st_within], [sf::st_contains], [sf::st_nearest_feature].
#' @param left If `TRUE` (default) keep every left row (left join); if `FALSE`
#'   keep only matches (inner join).
#' @param suffix Length-2 character vector disambiguating columns present on
#'   both sides. Default `c(".x", ".y")`.
#' @param partition Optional [grid()] specification enabling the two-sided
#'   streamed path, in which `y` is itself a `vectra_node`. Default `NULL` keeps
#'   the resident-`y` path.
#' @param y_geom,y_coords Geometry transport for a streamed `y` under
#'   `partition`: the name of `y`'s hex-WKB geometry column (`y_geom`, default
#'   the left `geom`), or a length-2 character vector of `y`'s coordinate columns
#'   (`y_coords`). Ignored without `partition`.
#' @param ... Further arguments passed to [sf::st_join()].
#'
#' @return A `vectra_node` of the joined stream, backed by temporary `.vtr`
#'   spills and carrying the left CRS.
#'
#' @seealso [spatial_map()] for per-feature transforms, [collect_sf()] to
#'   materialize as `sf`, [offload()] to partition both-sides-huge joins.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#'
#' # A stream of points, stored with x/y coordinate columns.
#' set.seed(1)
#' pts <- sf::st_coordinates(sf::st_sample(nc, 200))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), f)
#'
#' # Tag each point with the county it falls in, streaming.
#' tagged <- tbl(f) |>
#'   spatial_join(nc["NAME"], join = sf::st_intersects,
#'                coords = c("x", "y"), crs = sf::st_crs(nc))
#' head(collect(tagged))
#'
#' # Both sides streamed: bin to a grid and join per shard. Here y is a
#' # vectra_node rather than a resident sf object.
#' g <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   NAME = nc$NAME,
#'   geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
#' ), g)
#' tagged2 <- tbl(f) |>
#'   spatial_join(tbl(g), coords = c("x", "y"), crs = sf::st_crs(nc),
#'                partition = grid(0.5))
#' head(collect(tagged2))
#' unlink(c(f, g))
#'
#' @export
spatial_join <- function(x, y, join = NULL, geom = "geometry", coords = NULL,
                         crs = NA, left = TRUE, suffix = c(".x", ".y"),
                         partition = NULL, y_geom = NULL, y_coords = NULL,
                         out_geom = NULL, flush_rows = NULL, ...) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed left side)")
  if (is.null(join)) join <- sf::st_intersects
  crs <- .resolve_crs(x, crs)
  if (is.null(out_geom)) out_geom <- if (is.null(coords)) geom else "geometry"
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  dots <- list(...)

  if (!is.null(partition)) {
    if (!inherits(partition, "vectra_grid"))
      stop("`partition` must be a grid() specification, or NULL")
    if (!inherits(y, "vectra_node"))
      stop("with `partition`, `y` must be a vectra_node (the streamed right side)")
    if (is.null(y_geom)) y_geom <- geom
    return(.spatial_join_partition(x, y, join, partition, geom, coords,
                                   y_geom, y_coords, crs, left, suffix,
                                   out_geom, fr, dots))
  }

  if (!inherits(y, "sf"))
    stop("`y` must be an sf object (the resident right side of the join)")

  # Native path: a recognised predicate and planar-equivalent data. `y` is parsed
  # once into a GEOS locator; each batch's per-row matches come back from C and
  # the resident attributes are joined on in R, so the streamed left side is
  # never decoded to sf. `prep_batch(chunk, loc, nt)` returns a list of the
  # (possibly geometry-augmented) left `chunk` and its per-row match lists.
  run_native_join <- function(prep_batch) {
    loc     <- .geos_locator(y)
    y_attrs <- .coerce_for_vtr(as.data.frame(sf::st_drop_geometry(y)))
    nt      <- .spatial_threads()
    acc     <- .run_accumulator(fr)
    nxt     <- .batch_cursor(x)
    repeat {
      chunk <- nxt(); if (is.null(chunk)) break
      pb   <- prep_batch(chunk, loc, nt)
      out  <- .geos_join_assemble(pb$chunk, pb$matches, y_attrs, left, suffix)
      if (!identical(out_geom, geom) && geom %in% names(out))
        names(out)[match(geom, names(out))] <- out_geom
      gcol <- if (out_geom %in% names(out)) out_geom
              else if (geom %in% names(out)) geom else NULL
      if (!is.null(gcol))
        out <- out[c(setdiff(names(out), gcol), gcol)]  # geometry last, as st_join
      acc$push(.coerce_for_vtr(out))
    }
    acc$finish(crs = crs, empty_geom = out_geom)
  }

  # The streamed geometry reaches C as the batch's hex-WKB column (geom=) or as
  # raw point coordinates built into points in C (coords=); both share the match
  # assembly above. within-distance takes its radius through `...` (`dist =`);
  # any other extra st_join argument, or disjoint (whose matches are the bbox
  # complement the index cannot prune), falls back to sf.
  if (.geos_planar_ok(crs)) {
    nearest <- !length(dots) && identical(join, sf::st_nearest_feature)
    code <- if (nearest) NA_integer_ else .geos_pred_code(join)
    dist_val <- 0
    if (!is.na(code) && code == 10L) {
      dist_val <- dots$dist
      if (is.null(dist_val) || length(setdiff(names(dots), "dist")))
        code <- NA_integer_
    } else if (length(dots) || (!is.na(code) && code == 9L)) {
      code <- NA_integer_
    }
    if (nearest || !is.na(code)) {
      d <- as.numeric(dist_val)
      hex_matches <- function(hex, loc, nt)
        if (nearest) {
          near <- .Call(C_geos_nearest, loc, hex, nt)
          lapply(near, function(k) if (is.na(k)) integer(0) else k)
        } else .Call(C_geos_join, loc, hex, code, d, nt)
      xy_matches <- function(xy, loc, nt)
        if (nearest) {
          near <- .Call(C_geos_locate_xy, loc, xy$x, xy$y, 11L, 0, FALSE, nt)
          lapply(near, function(k) if (is.na(k)) integer(0) else k)
        } else .Call(C_geos_locate_xy, loc, xy$x, xy$y, code, d, TRUE, nt)
      if (is.null(coords))
        return(run_native_join(function(chunk, loc, nt) {
          if (!geom %in% names(chunk))
            stop(sprintf("geometry column '%s' not found; pass geom= or coords=", geom))
          list(chunk = chunk,
               matches = hex_matches(as.character(chunk[[geom]]), loc, nt))
        }))
      return(run_native_join(function(chunk, loc, nt) {
        xy <- .batch_xy(chunk, coords)
        chunk[[out_geom]] <- .Call(C_geos_points_to_hex, xy$x, xy$y)
        list(chunk = chunk, matches = xy_matches(xy, loc, nt))
      }))
    }
  }

  batch_fn <- function(sb)
    do.call(sf::st_join,
            c(list(sb, y, join = join, left = left, suffix = suffix), dots))
  .spatial_stream(x, batch_fn, geom, coords, crs, out_geom, fr)
}

# -- two-sided partitioned spatial join (grid partition) ----------------------

#' Define a uniform grid for a partitioned spatial join
#'
#' Describes the regular grid that [spatial_join()] uses to partition two
#' streamed layers for the both-sides-larger-than-RAM case. Cell `(cx, cy)`
#' covers `[origin_x + cx * cellsize_x, origin_x + (cx + 1) * cellsize_x)` and
#' likewise in y, so a coordinate maps to the cell `floor((coord - origin) /
#' cellsize)`. Pick a `cellsize` comparable to the scale of the join (large
#' enough that most cells hold a workable shard, small enough that one cell's
#' features fit in memory); for an extended-on-extended join choose it larger
#' than the left features.
#'
#' @param cellsize Cell size: a single number for square cells, or
#'   `c(cellsize_x, cellsize_y)`.
#' @param origin Grid origin `c(x0, y0)` (a cell corner). Default `c(0, 0)`.
#'
#' @return A `vectra_grid` specification to pass as `spatial_join(partition =)`.
#'
#' @seealso [spatial_join()] for the join it partitions.
#'
#' @examples
#' grid(1000)
#' grid(c(0.5, 0.25), origin = c(-180, -90))
#'
#' @export
grid <- function(cellsize, origin = c(0, 0)) {
  if (!is.numeric(cellsize) || !length(cellsize) %in% 1:2 ||
      any(!is.finite(cellsize)) || any(cellsize <= 0))
    stop("`cellsize` must be one or two positive numbers")
  if (length(cellsize) == 1L) cellsize <- c(cellsize, cellsize)
  if (!is.numeric(origin) || length(origin) != 2L || any(!is.finite(origin)))
    stop("`origin` must be a length-2 numeric c(x0, y0)")
  structure(list(cellsize = as.numeric(cellsize[1:2]),
                 origin = as.numeric(origin)),
            class = "vectra_grid")
}

#' @export
print.vectra_grid <- function(x, ...) {
  cat(sprintf("<vectra grid: cellsize %g x %g, origin (%g, %g)>\n",
              x$cellsize[1L], x$cellsize[2L], x$origin[1L], x$origin[2L]))
  invisible(x)
}

# Integer cell index of a coordinate along one axis.
.grid_ix <- function(v, o, s) as.integer(floor((v - o) / s))

# Cell label "cx:cy" from integer cell indices.
.grid_label <- function(cx, cy) paste(cx, cy, sep = ":")

# Per-feature bounding boxes of an sf/sfc batch as a 4-column matrix, derived
# from the flattened vertices (sf returns the feature index in the last
# st_coordinates column for lines/polygons/multipoints), so the whole batch is
# one vectorized pass. POINT geometry carries no index column (just X, Y), so
# each row is its own degenerate bbox.
.feature_bbox <- function(g) {
  co <- sf::st_coordinates(sf::st_geometry(g))
  if (ncol(co) <= 2L)
    return(cbind(xmin = co[, "X"], xmax = co[, "X"],
                 ymin = co[, "Y"], ymax = co[, "Y"]))
  fid <- co[, ncol(co)]
  cbind(xmin = tapply(co[, "X"], fid, min), xmax = tapply(co[, "X"], fid, max),
        ymin = tapply(co[, "Y"], fid, min), ymax = tapply(co[, "Y"], fid, max))
}

# Reference-cell label (one per row) for the left side: the cell each point or
# bbox centre falls in. Coordinate columns stay sf-free; WKB geometry is decoded.
.left_cell_labels <- function(chunk, geom, coords, crs, g) {
  if (!is.null(coords)) {
    x <- as.numeric(chunk[[coords[1L]]]); y <- as.numeric(chunk[[coords[2L]]])
  } else {
    bb <- .feature_bbox(.sf_decode_chunk(chunk, geom, NULL, crs))
    x <- (bb[, "xmin"] + bb[, "xmax"]) / 2
    y <- (bb[, "ymin"] + bb[, "ymax"]) / 2
  }
  .grid_label(.grid_ix(x, g$origin[1L], g$cellsize[1L]),
              .grid_ix(y, g$origin[2L], g$cellsize[2L]))
}

# Replicate right-side rows to every grid cell their geometry's bounding box
# overlaps, returning the augmented data.frame (original columns plus `.cell`).
# A point geometry hits one cell; a polygon spanning several is repeated once
# per cell so the per-shard join is complete.
.right_replicate <- function(chunk, geom, coords, crs, g) {
  if (!is.null(coords)) {
    x <- as.numeric(chunk[[coords[1L]]]); y <- as.numeric(chunk[[coords[2L]]])
    xmin <- x; xmax <- x; ymin <- y; ymax <- y
  } else {
    bb <- .feature_bbox(.sf_decode_chunk(chunk, geom, NULL, crs))
    xmin <- bb[, "xmin"]; xmax <- bb[, "xmax"]
    ymin <- bb[, "ymin"]; ymax <- bb[, "ymax"]
  }
  cx0 <- .grid_ix(xmin, g$origin[1L], g$cellsize[1L])
  cx1 <- .grid_ix(xmax, g$origin[1L], g$cellsize[1L])
  cy0 <- .grid_ix(ymin, g$origin[2L], g$cellsize[2L])
  cy1 <- .grid_ix(ymax, g$origin[2L], g$cellsize[2L])
  nx  <- cx1 - cx0 + 1L; ny <- cy1 - cy0 + 1L
  reps <- as.numeric(nx) * as.numeric(ny)
  row <- rep.int(seq_len(nrow(chunk)), reps)
  lab <- character(sum(reps))
  at <- 0L
  for (i in seq_len(nrow(chunk))) {
    cxs <- cx0[i]:cx1[i]; cys <- cy0[i]:cy1[i]
    li  <- .grid_label(rep(cxs, times = length(cys)),
                       rep(cys, each = length(cxs)))
    lab[(at + 1L):(at + length(li))] <- li
    at <- at + length(li)
  }
  out <- chunk[row, , drop = FALSE]
  out[[".cell"]] <- lab
  rownames(out) <- NULL
  out
}

# Stream a cursor, augment each batch into rows carrying a `.cell` label, and
# route them to per-cell run-files flushed when the budget is crossed. Returns
# the named list of run-file paths per cell label. Handles the row replication
# the right side needs (augment may return more rows than it received).
.cell_router <- function(cursor, augment, budget) {
  st <- new.env(parent = emptyenv())
  st$buffers <- list(); st$runs <- list(); st$buffered <- 0
  flush_one <- function(lab) {
    bufs <- st$buffers[[lab]]
    if (is.null(bufs) || !length(bufs)) return(invisible())
    df <- if (length(bufs) == 1L) bufs[[1L]] else do.call(rbind, bufs)
    rf <- tempfile(fileext = ".vtr"); write_vtr(df, rf)
    st$runs[[lab]] <- c(st$runs[[lab]], rf); st$buffers[[lab]] <- NULL
  }
  flush_all <- function() {
    for (lab in names(st$buffers)) flush_one(lab); st$buffered <- 0
  }
  repeat {
    chunk <- cursor(); if (is.null(chunk)) break
    aug <- augment(chunk)
    if (!nrow(aug)) next
    idx <- split(seq_len(nrow(aug)), aug[[".cell"]])
    for (lab in names(idx))
      st$buffers[[lab]] <- c(st$buffers[[lab]],
                             list(aug[idx[[lab]], , drop = FALSE]))
    st$buffered <- st$buffered + nrow(aug)
    if (st$buffered >= budget) flush_all()
  }
  flush_all()
  st$runs
}

# Drop the internal `.cell` routing column before a shard is decoded to sf.
.drop_cell <- function(df) df[setdiff(names(df), ".cell")]

# Right-side run files for a left cell, gathered over the 3x3 neighbourhood. The
# nearest-feature join is not local to a single cell, so a left point searches
# its own cell and the eight around it; the true nearest is found when it lies
# within one cell of the left reference cell (pick a cellsize at least the
# largest expected nearest distance).
.halo_runs <- function(rruns, lab) {
  cc <- as.integer(strsplit(lab, ":", fixed = TRUE)[[1L]])
  labs <- character(0)
  for (dx in -1:1) for (dy in -1:1)
    labs <- c(labs, .grid_label(cc[1L] + dx, cc[2L] + dy))
  unlist(rruns[intersect(labs, names(rruns))], use.names = FALSE)
}

.spatial_join_partition <- function(x, y, join, g, geom, coords,
                                    y_geom, y_coords, crs, left, suffix,
                                    out_geom, fr, dots) {
  budget <- getOption("vectra.partition_budget", .PARTITION_BUDGET)
  halo <- identical(join, sf::st_nearest_feature)

  lruns <- .cell_router(
    .batch_cursor(x),
    function(chunk) {
      chunk[[".cell"]] <- .left_cell_labels(chunk, geom, coords, crs, g)
      .coerce_for_vtr(chunk)
    }, budget)
  on.exit(unlink(unlist(lruns, use.names = FALSE)), add = TRUE)

  # An empty right sf with y's attribute schema, for left cells with no right
  # shard (so a left-join still pads them with NA right columns).
  ytmpl <- new.env(parent = emptyenv()); ytmpl$sf <- NULL
  rruns <- .cell_router(
    .batch_cursor(y),
    function(chunk) {
      if (is.null(ytmpl$sf))
        ytmpl$sf <- .sf_decode_chunk(chunk, y_geom, y_coords, crs)[0, ]
      .coerce_for_vtr(.right_replicate(chunk, y_geom, y_coords, crs, g))
    }, budget)
  on.exit(unlink(unlist(rruns, use.names = FALSE)), add = TRUE)

  acc <- .run_accumulator(fr)
  for (lab in names(lruns)) {
    lsf <- .sf_decode_chunk(.drop_cell(collect(.concat_runs(lruns[[lab]]))),
                            geom, coords, crs)
    rpaths <- if (halo) .halo_runs(rruns, lab) else rruns[[lab]]
    rsf <- if (!is.null(rpaths) && length(rpaths))
      .sf_decode_chunk(.drop_cell(collect(.concat_runs(rpaths))),
                       y_geom, y_coords, crs)
    else ytmpl$sf
    if (is.null(rsf)) {
      if (!left) next
      rsf <- sf::st_sf(geometry = sf::st_sfc(crs = .as_crs(crs)))
    }
    if (!left && !nrow(rsf)) next
    res <- do.call(sf::st_join,
                   c(list(lsf, rsf, join = join, left = left, suffix = suffix),
                     dots))
    acc$push(.sf_encode_result(res, out_geom))
  }
  acc$finish(crs = crs, empty_geom = out_geom)
}

#' Keep streamed rows by their spatial relation to a resident layer
#'
#' Streams a large left side `x` through the engine and keeps each row whose
#' geometry satisfies an \pkg{sf} binary predicate against a small resident
#' layer `y` (select by location). This is the spatial counterpart to a
#' [semi_join()]: rows are filtered, never duplicated, and no columns are added,
#' so the output carries `x`'s schema unchanged. With `negate = TRUE` it keeps
#' the rows that do *not* match (select by location, inverted). The billion-row
#' left stream never materializes; `y` (a study region, habitat patches, a
#' coastline buffer, ...) stays resident.
#'
#' For the recognised predicates -- the topological ones (intersects, within,
#' contains, overlaps, covers, covered by, touches, crosses), equals, disjoint,
#' and within-distance ([sf::st_is_within_distance], whose radius is passed as
#' `dist =`) -- on projected or unprojected planar data, the test runs natively
#' on the GEOS C API straight off the hex-WKB column: `y` is parsed once into a
#' spatial index and each batch is tested in C, with no per-batch round-trip
#' through \pkg{sf}. Coordinate-assembled (`coords`) point input runs natively
#' too, building each point in C rather than through \pkg{sf}; disjoint is the
#' one exception there (its matches are the bounding boxes the index prunes
#' away) and keeps the \pkg{sf} loop, as it does for the join. Geographic
#' coordinates with spherical geometry on (`sf::sf_use_s2()`) and any other
#' predicate use \pkg{sf} instead, preserving its semantics. When `y` carries no
#' CRS it inherits the stream's so the predicate does not reject on a mismatch.
#'
#' @inheritParams spatial_map
#' @param y An `sf` or `sfc` object: the resident locator layer to test against.
#' @param predicate An \pkg{sf} binary predicate function, e.g.
#'   [sf::st_intersects] (default), [sf::st_within], [sf::st_covered_by],
#'   [sf::st_is_within_distance]. A left row is kept when the predicate reports
#'   at least one match against `y`.
#' @param negate If `TRUE`, keep the rows with no match instead (the inverted
#'   select-by-location). Default `FALSE`.
#' @param ... Further arguments passed to `predicate`, e.g. `dist =` for
#'   [sf::st_is_within_distance].
#'
#' @return A `vectra_node` of the kept rows with `x`'s schema, backed by
#'   temporary `.vtr` spills and carrying the input CRS.
#'
#' @seealso [spatial_join()] to tag rows with `y`'s attributes, [spatial_clip()]
#'   to cut geometry against a mask, [filter()] for attribute predicates.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#' region <- nc[nc$NAME %in% c("Ashe", "Alleghany", "Surry"), "NAME"]
#'
#' set.seed(1)
#' pts <- sf::st_coordinates(sf::st_sample(nc, 300))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), f)
#'
#' # Keep only the points that fall inside the three-county region, streaming.
#' inside <- tbl(f) |>
#'   spatial_filter(region, coords = c("x", "y"), crs = sf::st_crs(nc))
#' nrow(collect(inside))
#' unlink(f)
#'
#' @export
spatial_filter <- function(x, y, predicate = NULL, negate = FALSE,
                           geom = "geometry", coords = NULL, crs = NA,
                           flush_rows = NULL, ...) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed left side)")
  if (!inherits(y, "sf") && !inherits(y, "sfc"))
    stop("`y` must be an sf or sfc object (the resident locator layer)")
  crs <- .resolve_crs(x, crs)
  y   <- .align_resident_crs(y, crs)
  fr  <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  dots <- list(...)
  acc <- .run_accumulator(fr)
  empty_geom <- if (is.null(coords)) geom else "geometry"

  # Native path: a recognised predicate and hex-WKB geometry. The resident side
  # is parsed once into a GEOS locator and each batch is tested in C straight off
  # the column. `st_is_within_distance` takes its radius through `...` (`dist =`);
  # any other extra predicate argument, coordinate-assembled (`coords`) input, or
  # an unrecognised predicate falls back to the per-batch sf loop.
  code <- .geos_pred_code(predicate)
  dist_val <- 0
  if (!is.na(code) && code == 10L) {
    dist_val <- dots$dist
    if (is.null(dist_val) || length(setdiff(names(dots), "dist")))
      code <- NA_integer_
  } else if (length(dots)) {
    code <- NA_integer_
  }
  if (is.null(coords) && !is.na(code) && .geos_planar_ok(crs)) {
    loc <- .geos_locator(y)
    nt  <- .spatial_threads()
    nxt <- .batch_cursor(x)
    repeat {
      chunk <- nxt(); if (is.null(chunk)) break
      if (!geom %in% names(chunk))
        stop(sprintf("geometry column '%s' not found; pass geom= or coords=", geom))
      hit <- .Call(C_geos_filter, loc, as.character(chunk[[geom]]),
                   code, isTRUE(negate), as.numeric(dist_val), nt)
      acc$push(.coerce_for_vtr(chunk[hit, , drop = FALSE]))
    }
    return(acc$finish(crs = crs, empty_geom = empty_geom))
  }

  # Native coords path: raw x/y point columns matched in C with no per-batch sf.
  # Disjoint (its matches are the bbox complement the index cannot prune) keeps
  # the sf loop, as it does for the join.
  if (!is.null(coords) && !is.na(code) && code != 9L && .geos_planar_ok(crs)) {
    loc <- .geos_locator(y)
    nt  <- .spatial_threads()
    nxt <- .batch_cursor(x)
    repeat {
      chunk <- nxt(); if (is.null(chunk)) break
      xy  <- .batch_xy(chunk, coords)
      idx <- .Call(C_geos_locate_xy, loc, xy$x, xy$y, code,
                   as.numeric(dist_val), FALSE, nt)
      hit <- !is.na(idx)
      if (negate) hit <- !hit
      acc$push(.coerce_for_vtr(chunk[hit, , drop = FALSE]))
    }
    return(acc$finish(crs = crs, empty_geom = empty_geom))
  }

  if (is.null(predicate)) predicate <- sf::st_intersects
  nxt <- .batch_cursor(x)
  repeat {
    chunk <- nxt(); if (is.null(chunk)) break
    sb  <- .sf_decode_chunk(chunk, geom, coords, crs)
    hit <- lengths(do.call(predicate, c(list(sb, y), dots))) > 0L
    if (negate) hit <- !hit
    acc$push(.coerce_for_vtr(chunk[hit, , drop = FALSE]))
  }
  acc$finish(crs = crs, empty_geom = empty_geom)
}

#' Clip or erase a streamed layer against a resident mask
#'
#' Streams a large layer `x` through the engine and cuts each batch's geometry
#' against a small resident `mask` (a study boundary, a buffer, a set of
#' patches). By default this clips -- the intersection with the mask, the GIS
#' "Clip" tool -- keeping only the parts of `x` that fall inside `mask`. With
#' `erase = TRUE` it instead erases -- the difference, the "Erase"/"Difference"
#' tool -- keeping the parts of `x` outside `mask`. The mask is dissolved to a
#' single geometry once and held resident while the billion-row left stream
#' flows past one batch at a time.
#'
#' Geometry travels through the engine as hex-encoded WKB in a string column and
#' the CRS is carried on the returned node; use [collect_sf()] to materialize.
#' On projected or unprojected planar data the cut runs natively on the GEOS C
#' API straight off the hex-WKB column (the mask parsed once); geographic
#' coordinates with spherical geometry on (`sf::sf_use_s2()`) and
#' coordinate-assembled (`coords`) input cut through \pkg{sf} instead. When
#' `mask` carries no CRS it inherits the stream's.
#'
#' @inheritParams spatial_map
#' @param mask An `sf` or `sfc` object whose dissolved geometry clips (or, with
#'   `erase = TRUE`, erases) the stream.
#' @param erase If `TRUE`, keep the parts of `x` *outside* `mask` (difference)
#'   rather than inside (intersection). Default `FALSE`.
#'
#' @return A `vectra_node` of the cut geometry with `x`'s attributes, backed by
#'   temporary `.vtr` spills and carrying the input CRS.
#'
#' @seealso [spatial_filter()] to keep whole features by location without
#'   cutting them, [spatial_map()] for per-feature transforms, [collect_sf()].
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#' mask <- sf::st_union(nc[nc$NAME %in% c("Ashe", "Alleghany"), ])
#'
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   NAME = nc$NAME,
#'   geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
#' ), f)
#'
#' # Clip every county polygon to the two-county mask, streaming.
#' clipped <- tbl(f) |> spatial_clip(mask, crs = sf::st_crs(nc))
#' collect_sf(clipped)
#' unlink(f)
#'
#' @export
spatial_clip <- function(x, mask, erase = FALSE, geom = "geometry",
                         coords = NULL, crs = NA, out_geom = NULL,
                         flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed layer to clip)")
  if (!inherits(mask, "sf") && !inherits(mask, "sfc"))
    stop("`mask` must be an sf or sfc object (the resident clip/erase mask)")
  crs    <- .resolve_crs(x, crs)
  mask   <- .align_resident_crs(mask, crs)
  mask_u <- sf::st_union(sf::st_geometry(mask))   # one resident mask geometry
  if (is.null(out_geom)) out_geom <- if (is.null(coords)) geom else "geometry"
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)

  # Native path: cut each batch's hex-WKB geometry against the resident mask in
  # C (intersection to clip, difference to erase), parsing the mask once. The
  # coordinate-assembled (`coords`) and spherical (geographic + s2) cases fall
  # back to the per-batch sf loop.
  if (is.null(coords) && .geos_planar_ok(crs)) {
    loc <- .geos_locator(mask_u)
    nt  <- .spatial_threads()
    acc <- .run_accumulator(fr)
    nxt <- .batch_cursor(x)
    repeat {
      chunk <- nxt(); if (is.null(chunk)) break
      if (!geom %in% names(chunk))
        stop(sprintf("geometry column '%s' not found; pass geom= or coords=", geom))
      cut  <- .Call(C_geos_clip, loc, as.character(chunk[[geom]]),
                    isTRUE(erase), nt)
      keep <- !is.na(cut)
      df   <- chunk[keep, setdiff(names(chunk), geom), drop = FALSE]
      df[[out_geom]] <- cut[keep]
      acc$push(.coerce_for_vtr(df))
    }
    return(acc$finish(crs = crs, empty_geom = out_geom))
  }

  op <- if (erase) sf::st_difference else sf::st_intersection
  # st_intersection/st_difference warn once per batch that attributes are
  # assumed spatially constant; that is the intended behaviour here, so mute
  # just that warning rather than letting it repeat for every streamed batch.
  batch_fn <- function(sb)
    withCallingHandlers(
      op(sb, mask_u),
      warning = function(w) {
        if (grepl("assumed.*spatially constant", conditionMessage(w)))
          invokeRestart("muffleWarning")
      })
  .spatial_stream(x, batch_fn, geom, coords, crs, out_geom, fr)
}

# -- snapping and topology cleanup --------------------------------------------

# Snap one batch's geometry to a fixed-precision grid using the same C
# snap-rounding the overlay noder uses internally: each geometry crosses to C as
# WKB, is rounded to the `size` lattice and made valid, and comes back as cleaned
# WKB. The count is preserved (one cleaned geometry per input), so attributes
# ride through untouched.
.snap_grid_batch <- function(sb, size, nt) {
  g  <- sf::st_geometry(sb)
  w  <- sf::st_as_binary(g, EWKB = FALSE)
  pr <- .Call(C_overlay_parse, w, as.double(size), as.integer(nt))
  g2 <- sf::st_as_sfc(structure(pr[[2L]], class = "WKB"), EWKB = FALSE)
  sf::st_set_geometry(sb, sf::st_set_crs(g2, sf::st_crs(g)))
}

#' Snap a streamed layer's coordinates to a fixed grid
#'
#' Rounds every coordinate of a streamed layer to a regular grid of spacing
#' `size` (in CRS units) and repairs the result, one batch at a time. This is the
#' fixed-precision snap-rounding the overlay noder
#' ([spatial_overlay()]) applies internally, exposed as a standalone verb: it
#' merges near-coincident vertices and removes the slivers that floating-point
#' coordinates leave between shared boundaries, so a layer can be cleaned (or
#' pre-noded to a common precision) without running a full overlay. Snapping is
#' done in C straight off the hex-WKB column; one cleaned geometry comes back per
#' input feature, so attributes ride through untouched.
#'
#' Geometry travels through the engine as hex-encoded WKB in a string column and
#' the CRS is carried on the returned node; use [collect_sf()] to materialize.
#' The \pkg{sf} package is an optional dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param size Grid spacing in CRS units (a positive number). Coordinates are
#'   rounded to the nearest multiple; a larger `size` snaps more aggressively.
#'
#' @return A `vectra_node` of the snapped geometry with `x`'s attributes, backed
#'   by temporary `.vtr` spills (removed when the node is garbage-collected) and
#'   carrying the input CRS.
#'
#' @seealso [spatial_snap()] to snap toward another layer instead of a grid,
#'   [spatial_overlay()] whose noding uses the same snap-rounding, [collect_sf()].
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' p <- sf::st_polygon(list(rbind(c(0.04, 0.03), c(1.02, 0.01),
#'                                c(0.98, 1.03), c(0.01, 0.97), c(0.04, 0.03))))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   id = 1L, geometry = sf::st_as_binary(sf::st_sfc(p), hex = TRUE)
#' ), f)
#'
#' # Snap the jittered corners back onto a 0.1 grid.
#' tbl(f) |> spatial_snap_grid(0.1) |> collect_sf()
#' unlink(f)
#'
#' @export
spatial_snap_grid <- function(x, size, geom = "geometry", crs = NA,
                              out_geom = NULL, flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed layer to snap)")
  if (!is.numeric(size) || length(size) != 1L || !is.finite(size) || size <= 0)
    stop("`size` must be a single positive number (the grid spacing)")
  crs <- .resolve_crs(x, crs)
  if (is.null(out_geom)) out_geom <- geom
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  nt <- .spatial_threads()
  .spatial_stream(x, function(sb) .snap_grid_batch(sb, size, nt),
                  geom, coords = NULL, crs = crs, out_geom = out_geom,
                  flush_rows = fr)
}

#' Snap a streamed layer toward a resident reference layer
#'
#' Streams a large layer `x` through the engine and snaps each batch's vertices
#' toward a small resident reference layer `y` when they lie within `tolerance`
#' (in CRS units), one batch at a time (the QGIS "snap geometries to layer").
#' Vertices and edges of `x` closer than `tolerance` to `y` are pulled onto `y`,
#' which closes the small gaps and overshoots between two layers that should
#' share a boundary. The reference layer stays resident while the billion-row
#' left stream flows past; the snap itself is \pkg{sf}'s [sf::st_snap()].
#'
#' Geometry travels through the engine as hex-encoded WKB in a string column and
#' the CRS is carried on the returned node; use [collect_sf()] to materialize.
#' When `y` carries no CRS it inherits the stream's. The \pkg{sf} package is an
#' optional dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param y An `sf` or `sfc` object: the resident reference layer to snap toward.
#' @param tolerance Snapping distance in CRS units (a positive number). Vertices
#'   and edges of `x` within this distance of `y` are moved onto `y`.
#'
#' @return A `vectra_node` of the snapped geometry with `x`'s attributes, backed
#'   by temporary `.vtr` spills (removed when the node is garbage-collected) and
#'   carrying the input CRS.
#'
#' @seealso [spatial_snap_grid()] to snap to a grid instead of a layer,
#'   [spatial_clip()] for the resident-mask streaming pattern, [collect_sf()].
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' ref <- sf::st_sfc(sf::st_linestring(rbind(c(0, 0), c(10, 0))))
#' line <- sf::st_linestring(rbind(c(0, 0.2), c(5, 0.1), c(10, 0.2)))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   id = 1L, geometry = sf::st_as_binary(sf::st_sfc(line), hex = TRUE)
#' ), f)
#'
#' # Pull the near-zero vertices down onto the reference line.
#' tbl(f) |> spatial_snap(ref, tolerance = 0.5) |> collect_sf()
#' unlink(f)
#'
#' @export
spatial_snap <- function(x, y, tolerance, geom = "geometry", coords = NULL,
                         crs = NA, out_geom = NULL, flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed layer to snap)")
  if (!inherits(y, "sf") && !inherits(y, "sfc"))
    stop("`y` must be an sf or sfc object (the resident reference layer)")
  if (!is.numeric(tolerance) || length(tolerance) != 1L ||
      !is.finite(tolerance) || tolerance <= 0)
    stop("`tolerance` must be a single positive number (the snap distance)")
  crs  <- .resolve_crs(x, crs)
  y    <- .align_resident_crs(y, crs)
  yg   <- sf::st_geometry(y)
  if (is.null(out_geom)) out_geom <- if (is.null(coords)) geom else "geometry"
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  batch_fn <- function(sb) sf::st_snap(sb, yg, tolerance = tolerance)
  .spatial_stream(x, batch_fn, geom, coords, crs, out_geom, fr)
}

# -- line / polygon smoothing (Chaikin corner-cutting) ------------------------

# One Chaikin pass over an open polyline: each segment P_i -> P_{i+1} yields two
# points at 1/4 and 3/4 of the way, cutting the corner. `keep_ends` pins the
# first and last vertices so an open line is not shortened at its tips. Repeated
# `iterations` times; a line too short to cut is returned unchanged.
.chaikin_open <- function(m, iterations, keep_ends) {
  for (it in seq_len(iterations)) {
    n <- nrow(m)
    if (n < 3L) break
    i <- seq_len(n - 1L)
    q <- 0.75 * m[i, , drop = FALSE] + 0.25 * m[i + 1L, , drop = FALSE]
    r <- 0.25 * m[i, , drop = FALSE] + 0.75 * m[i + 1L, , drop = FALSE]
    body <- matrix(0, 2L * length(i), ncol(m))
    body[seq.int(1L, by = 2L, length.out = length(i)), ] <- q
    body[seq.int(2L, by = 2L, length.out = length(i)), ] <- r
    m <- if (keep_ends) rbind(m[1L, , drop = FALSE], body, m[n, , drop = FALSE])
         else body
  }
  m
}

# One Chaikin pass over a closed ring (first vertex repeated as the last). The
# cut runs cyclically over every edge, including the closing one, and the result
# is re-closed; corner-cutting shrinks the ring slightly, as Chaikin does.
.chaikin_ring <- function(m, iterations) {
  for (it in seq_len(iterations)) {
    p <- m[-nrow(m), , drop = FALSE]
    k <- nrow(p)
    if (k < 3L) break
    nxt  <- p[c(2:k, 1L), , drop = FALSE]
    q <- 0.75 * p + 0.25 * nxt
    r <- 0.25 * p + 0.75 * nxt
    body <- matrix(0, 2L * k, ncol(p))
    body[seq.int(1L, by = 2L, length.out = k), ] <- q
    body[seq.int(2L, by = 2L, length.out = k), ] <- r
    m <- rbind(body, body[1L, , drop = FALSE])
  }
  m
}

# Smooth one sfg: open polylines cut with `keep_ends`, polygon rings cut
# cyclically, multi/collection types recursed. Point geometry is returned as-is.
.smooth_sfg <- function(g, iterations, keep_ends) {
  switch(class(g)[2L],
    LINESTRING = sf::st_linestring(.chaikin_open(unclass(g), iterations, keep_ends)),
    MULTILINESTRING = sf::st_multilinestring(
      lapply(unclass(g), .chaikin_open, iterations, keep_ends)),
    POLYGON = sf::st_polygon(lapply(unclass(g), .chaikin_ring, iterations)),
    MULTIPOLYGON = sf::st_multipolygon(lapply(unclass(g), function(rings)
      lapply(rings, .chaikin_ring, iterations))),
    GEOMETRYCOLLECTION = sf::st_geometrycollection(
      lapply(g, .smooth_sfg, iterations, keep_ends)),
    g)
}

.smooth_batch <- function(sb, iterations, keep_ends) {
  if (nrow(sb) == 0L) return(sb)
  g  <- sf::st_geometry(sb)
  g2 <- sf::st_sfc(lapply(g, .smooth_sfg, iterations, keep_ends),
                   crs = sf::st_crs(g))
  sf::st_set_geometry(sb, g2)
}

#' Smooth streamed line and polygon geometry
#'
#' Rounds the corners of every line and polygon in a streamed layer by Chaikin
#' corner-cutting, one batch at a time. Each iteration replaces every vertex with
#' two points a quarter and three-quarters of the way along its adjacent edges,
#' so sharp angles become a sequence of short chamfers that read as a smooth
#' curve; more `iterations` give a smoother result with more vertices. Open lines
#' keep their endpoints (`keep_ends`); polygon rings are cut cyclically and shrink
#' slightly, as Chaikin smoothing does. Point geometry passes through unchanged.
#'
#' The smoothing is computed directly on the coordinates (no GEOS call), so it is
#' dependency-light; \pkg{sf} is used only to decode and rebuild each batch.
#' Geometry travels through the engine as hex-encoded WKB in a string column and
#' the CRS is carried on the returned node; use [collect_sf()] to materialize.
#'
#' @inheritParams spatial_map
#' @param iterations Number of corner-cutting passes (a positive integer). Each
#'   pass roughly doubles the vertex count. Default `2`.
#' @param keep_ends If `TRUE` (default), pin the first and last vertex of an open
#'   line so it is not shortened at its tips. Ignored for closed rings.
#'
#' @return A `vectra_node` of the smoothed geometry with `x`'s attributes, backed
#'   by temporary `.vtr` spills (removed when the node is garbage-collected) and
#'   carrying the input CRS.
#'
#' @seealso [spatial_map()] for per-feature transforms such as densifying with
#'   `~ sf::st_segmentize(.x, dfMaxLength)` or sampling points along a line with
#'   `~ sf::st_line_sample(.x, n)`, [collect_sf()] to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' zig <- sf::st_linestring(rbind(c(0, 0), c(1, 1), c(2, 0), c(3, 1), c(4, 0)))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   id = 1L, geometry = sf::st_as_binary(sf::st_sfc(zig), hex = TRUE)
#' ), f)
#'
#' # Smooth the zig-zag with three corner-cutting passes.
#' tbl(f) |> spatial_smooth(iterations = 3) |> collect_sf()
#' unlink(f)
#'
#' @export
spatial_smooth <- function(x, iterations = 2L, keep_ends = TRUE,
                           geom = "geometry", crs = NA, out_geom = NULL,
                           flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed layer to smooth)")
  if (!is.numeric(iterations) || length(iterations) != 1L ||
      !is.finite(iterations) || iterations < 1)
    stop("`iterations` must be a single positive integer")
  iterations <- as.integer(iterations)
  crs <- .resolve_crs(x, crs)
  if (is.null(out_geom)) out_geom <- geom
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  .spatial_stream(x, function(sb) .smooth_batch(sb, iterations, keep_ends),
                  geom, coords = NULL, crs = crs, out_geom = out_geom,
                  flush_rows = fr)
}

# -- k-nearest neighbours (with distances) ------------------------------------

# Find the k nearest resident-y features for each row of one decoded left batch
# and return one row per (left, neighbour) pair: the left columns replicated,
# plus the neighbour's rank (1 = nearest), its identifier `yid`, and the
# distance. Distances come from the full left-by-y distance matrix (y is the
# small resident side), reduced to the k smallest per left row.
.knn_batch <- function(sb, yg, yid, k, rank_col, id_col, dist_col) {
  ny <- length(yg)
  kk <- min(as.integer(k), ny)
  nl <- nrow(sb)
  if (nl == 0L || kk == 0L) {
    out <- sb[integer(0), , drop = FALSE]
    out[[rank_col]] <- integer(0)
    out[[id_col]]   <- yid[integer(0)]
    out[[dist_col]] <- numeric(0)
    return(out)
  }
  d <- matrix(as.numeric(sf::st_distance(sb, yg)), nrow = nl)
  ord <- t(apply(d, 1L, order))[, seq_len(kk), drop = FALSE]
  left_idx <- rep(seq_len(nl), each = kk)
  nb_pos   <- as.integer(t(ord))
  out <- sb[left_idx, , drop = FALSE]
  out[[rank_col]] <- rep.int(seq_len(kk), nl)
  out[[id_col]]   <- yid[nb_pos]
  out[[dist_col]] <- d[cbind(left_idx, nb_pos)]
  rownames(out) <- NULL
  out
}

#' k nearest neighbours of a streamed layer, with distances
#'
#' Streams a large left side `x` through the engine and, for each feature, finds
#' the `k` nearest features of a small resident layer `y`, returning one row per
#' (left, neighbour) pair with the neighbour's rank, identifier, and distance.
#' Where [spatial_join()] with [sf::st_nearest_feature] attaches only the single
#' nearest match, this returns the top `k` and the distances themselves -- the
#' nearest-`k` query and the building block of a distance matrix. The billion-row
#' left stream never materializes; `y` (the candidate neighbours) stays resident.
#'
#' Distances are \pkg{sf}'s [sf::st_distance()]: planar (CRS units) on projected
#' or unprojected planar data, great-circle (metres) on geographic coordinates
#' with spherical geometry on (`sf::sf_use_s2()`). Each batch forms its
#' left-by-`y` distance matrix, so `y` should be the small side; when `y` carries
#' no CRS it inherits the stream's. The left geometry rides through unchanged
#' (replicated once per neighbour). The \pkg{sf} package is an optional
#' dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param y An `sf` or `sfc` object: the resident candidate-neighbour layer.
#' @param k Number of nearest neighbours to return per left feature (capped at
#'   the number of `y` features). Default `1`.
#' @param y_id Optional name of a column in `y` whose value identifies each
#'   neighbour in the output. Default `NULL` uses `y`'s 1-based row index.
#' @param id_col,dist_col,rank_col Names of the output columns holding the
#'   neighbour identifier, the distance, and the 1-based rank (1 = nearest).
#'   Defaults `"neighbor"`, `"distance"`, `"rank"`.
#'
#' @return A `vectra_node` of one row per (left, neighbour) pair -- `x`'s columns
#'   (geometry included) plus the rank, neighbour identifier, and distance --
#'   backed by temporary `.vtr` spills (removed when the node is garbage-
#'   collected) and carrying the input CRS.
#'
#' @seealso [spatial_join()] for a nearest-feature attribute join, [collect_sf()]
#'   to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#' towns <- sf::st_centroid(sf::st_geometry(nc))[1:5]
#' towns <- sf::st_sf(town = nc$NAME[1:5], geometry = towns)
#'
#' set.seed(1)
#' pts <- sf::st_coordinates(sf::st_sample(nc, 100))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), f)
#'
#' # The two nearest towns to each point, with distances.
#' tbl(f) |>
#'   spatial_knn(towns, k = 2, coords = c("x", "y"), crs = sf::st_crs(nc),
#'               y_id = "town") |>
#'   collect() |> head()
#' unlink(f)
#'
#' @export
spatial_knn <- function(x, y, k = 1L, geom = "geometry", coords = NULL,
                        crs = NA, y_id = NULL, id_col = "neighbor",
                        dist_col = "distance", rank_col = "rank",
                        out_geom = NULL, flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed left side)")
  if (!inherits(y, "sf") && !inherits(y, "sfc"))
    stop("`y` must be an sf or sfc object (the resident neighbour layer)")
  if (!is.numeric(k) || length(k) != 1L || !is.finite(k) || k < 1)
    stop("`k` must be a single positive integer")
  crs <- .resolve_crs(x, crs)
  y   <- .align_resident_crs(y, crs)
  yg  <- sf::st_geometry(y)
  yid <- if (is.null(y_id)) seq_len(length(yg)) else {
    if (!inherits(y, "sf") || !y_id %in% names(y))
      stop(sprintf("`y_id` column '%s' not found in `y`", y_id))
    y[[y_id]]
  }
  if (is.null(out_geom)) out_geom <- if (is.null(coords)) geom else "geometry"
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  batch_fn <- function(sb)
    .knn_batch(sb, yg, yid, k, rank_col, id_col, dist_col)
  .spatial_stream(x, batch_fn, geom, coords, crs, out_geom, fr)
}

# -- split / line intersection (cut against a resident blade) -----------------

# Split one geometry by the resident blade. A polygon's boundary is merged with
# the blade, noded, and re-polygonized; the faces whose interior point falls
# inside the original polygon are the pieces. A line is merged with the blade,
# noded, and cast to single segments; the arcs that lie on the original line are
# the pieces. A geometry the blade does not cut returns as a single piece, so a
# feature never vanishes. Returns an sfc of one or more pieces.
.split_one <- function(g, blade_u, crs) {
  gs <- sf::st_sfc(g, crs = crs)
  switch(class(g)[2L],
    POLYGON = ,
    MULTIPOLYGON = {
      b     <- sf::st_sfc(sf::st_boundary(g), crs = crs)
      noded <- sf::st_node(sf::st_union(c(b, blade_u)))
      faces <- sf::st_collection_extract(sf::st_polygonize(noded), "POLYGON")
      if (!length(faces)) return(gs)
      ip   <- sf::st_point_on_surface(faces)
      keep <- faces[lengths(sf::st_within(ip, gs)) > 0L]
      if (length(keep)) keep else gs
    },
    LINESTRING = ,
    MULTILINESTRING = {
      noded <- sf::st_node(sf::st_union(c(gs, blade_u)))
      segs  <- sf::st_cast(noded, "LINESTRING", warn = FALSE)
      if (!length(segs)) return(gs)
      keep <- segs[lengths(sf::st_covered_by(segs, gs)) > 0L]
      if (length(keep)) keep else gs
    },
    gs)
}

# Intersection points of one geometry with the resident blade, as a single
# (multi)point sfg; an empty geometry when they do not cross.
.cross_points <- function(g, blade_u, crs) {
  ip <- suppressWarnings(sf::st_intersection(sf::st_sfc(g, crs = crs), blade_u))
  if (!length(ip)) return(sf::st_multipoint())
  pts <- suppressWarnings(sf::st_collection_extract(ip, "POINT"))
  if (!length(pts)) return(sf::st_multipoint())
  sf::st_combine(pts)[[1L]]
}

.split_batch <- function(sb, blade_u, crs, extract) {
  if (nrow(sb) == 0L) return(sb)
  crs <- .as_crs(crs)
  g <- sf::st_geometry(sb)
  if (extract == "points") {
    pts <- sf::st_sfc(lapply(seq_along(g),
                             function(i) .cross_points(g[[i]], blade_u, crs)),
                      crs = crs)
    keep <- !sf::st_is_empty(pts)
    out  <- sb[keep, , drop = FALSE]
    sf::st_geometry(out) <- pts[keep]
    out
  } else {
    pieces <- lapply(seq_along(g), function(i) .split_one(g[[i]], blade_u, crs))
    np  <- lengths(pieces)
    out <- sb[rep.int(seq_len(nrow(sb)), np), , drop = FALSE]
    sf::st_geometry(out) <- sf::st_sfc(unlist(pieces, recursive = FALSE),
                                       crs = crs)
    rownames(out) <- NULL
    out
  }
}

#' Split a streamed layer by a resident blade, or return its crossing points
#'
#' Streams a large layer `x` through the engine and cuts each batch's geometry
#' against a small resident `blade` layer (the QGIS "split with lines"), one
#' batch at a time. With `extract = "pieces"` (the default) every feature is
#' divided where the blade crosses it -- a polygon into the faces the blade carves
#' out, a line into the arcs between crossings -- and each piece is emitted as its
#' own row with the source attributes copied; a feature the blade does not cross
#' passes through as a single piece. With `extract = "points"` the verb instead
#' returns, per feature, the points where it meets the blade (the "line
#' intersections" tool), dropping features that do not cross.
#'
#' The split is built from \pkg{sf}/GEOS noding and polygonization, so it expects
#' projected or unprojected planar data; geographic coordinates are best
#' projected first. The blade is dissolved to one geometry once and held resident
#' while the left stream flows past. Geometry travels through the engine as
#' hex-encoded WKB in a string column and the CRS is carried on the returned
#' node; when `blade` carries no CRS it inherits the stream's. The \pkg{sf}
#' package is an optional dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param blade An `sf` or `sfc` object whose geometry cuts the stream (typically
#'   lines, but any geometry whose boundary can node `x`).
#' @param extract `"pieces"` (default) to emit the split pieces, or `"points"`
#'   to emit the intersection points of each feature with the blade.
#'
#' @return A `vectra_node`: with `extract = "pieces"`, one row per piece carrying
#'   `x`'s attributes; with `extract = "points"`, one row per crossing feature
#'   carrying its intersection points. Backed by temporary `.vtr` spills (removed
#'   when the node is garbage-collected) and carrying the input CRS.
#'
#' @seealso [spatial_clip()] to cut against a mask without dividing into pieces,
#'   [spatial_overlay()] to node two polygon layers into a partition,
#'   [collect_sf()] to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' sq <- sf::st_polygon(list(rbind(c(0, 0), c(4, 0), c(4, 4), c(0, 4), c(0, 0))))
#' blade <- sf::st_sfc(sf::st_linestring(rbind(c(2, -1), c(2, 5))))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   id = 1L, geometry = sf::st_as_binary(sf::st_sfc(sq), hex = TRUE)
#' ), f)
#'
#' # Split the square into two halves along the blade.
#' tbl(f) |> spatial_split(blade) |> collect_sf()
#' unlink(f)
#'
#' @export
spatial_split <- function(x, blade, extract = c("pieces", "points"),
                          geom = "geometry", crs = NA, out_geom = NULL,
                          flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed layer to split)")
  if (!inherits(blade, "sf") && !inherits(blade, "sfc"))
    stop("`blade` must be an sf or sfc object (the resident cutting layer)")
  extract <- match.arg(extract)
  crs   <- .resolve_crs(x, crs)
  blade <- .align_resident_crs(blade, crs)
  blade_u <- sf::st_set_crs(sf::st_union(sf::st_geometry(blade)), .as_crs(crs))
  if (is.null(out_geom)) out_geom <- geom
  fr <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  .spatial_stream(x, function(sb) .split_batch(sb, blade_u, crs, extract),
                  geom, coords = NULL, crs = crs, out_geom = out_geom,
                  flush_rows = fr)
}

# -- dissolve (aggregate geometries by group) ---------------------------------

# Composite group label for a batch: one string per row joining the `by`
# column values (unit separator, unlikely to collide), or a constant when no
# `by` is given (dissolve the whole layer into one feature).
.dissolve_assign <- function(by) {
  if (is.null(by)) return(function(chunk) rep("all", nrow(chunk)))
  function(chunk) {
    parts <- lapply(by, function(b) as.character(chunk[[b]]))
    do.call(paste, c(parts, sep = ""))
  }
}

#' Dissolve geometries by group
#'
#' Unions the geometries within each `by` group into a single feature (the GIS
#' "Dissolve" tool), optionally summarising attributes. Unlike the streamed
#' per-batch verbs, dissolve needs every geometry of a group together to union
#' them, so it rides the **partition tier**: `x` is spilled once and routed into
#' one disjoint shard per group in a single bounded pass, then each shard is read
#' in and unioned with \pkg{sf}. Peak memory is the routing budget during the
#' pass, then one group's geometries while it is unioned -- partition the input
#' on a key whose groups fit in memory. With no `by`, the whole layer dissolves
#' into one feature.
#'
#' Geometry travels through the engine as hex-encoded WKB in a string column and
#' the CRS is carried on the returned node; use [collect_sf()] to materialize.
#' On projected or unprojected planar data each group is unioned natively on the
#' GEOS C API straight off the hex-WKB column; geographic coordinates with
#' spherical geometry on (`sf::sf_use_s2()`), or any extra [sf::st_union()]
#' arguments (e.g. `is_coverage = TRUE`), union through \pkg{sf} instead. The
#' \pkg{sf} package is an optional dependency (Suggests).
#'
#' @inheritParams spatial_map
#' @param by Character vector of attribute columns to dissolve within: one
#'   output feature per distinct combination of their values. `NULL` (default)
#'   dissolves the entire layer into a single feature.
#' @param ... Further arguments passed to [sf::st_union()] (e.g.
#'   `is_coverage = TRUE`).
#' @param .fun Optional named list of attribute summaries. Each element is a
#'   function taking the group's data.frame and returning a length-1 value; the
#'   list name becomes the output column (e.g.
#'   `.fun = list(total = function(d) sum(d$pop))`). Default `NULL` keeps only
#'   the `by` columns and the dissolved geometry.
#'
#' @return A `vectra_node` of one row per group -- the `by` columns, any `.fun`
#'   summaries, and the dissolved geometry -- backed by temporary `.vtr` spills
#'   removed when the node is garbage-collected, carrying the input CRS for
#'   [collect_sf()].
#'
#' @seealso [spatial_overlay()] to split overlaps apart rather than merge them,
#'   [offload()] for the partition tier this rides on, [collect_sf()].
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#' nc$band <- nc$SID74 > 5            # an attribute to dissolve within
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   band = nc$band, BIR74 = nc$BIR74,
#'   geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
#' ), f)
#'
#' # Merge the counties into two features by `band`, summing births.
#' merged <- tbl(f) |>
#'   spatial_dissolve(by = "band", crs = sf::st_crs(nc),
#'                    .fun = list(births = function(d) sum(d$BIR74)))
#' collect_sf(merged)
#' unlink(f)
#'
#' @export
spatial_dissolve <- function(x, by = NULL, ..., geom = "geometry", crs = NA,
                             .fun = NULL, flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed layer to dissolve)")
  if (!is.null(by) && !is.character(by))
    stop("`by` must be a character vector of column names, or NULL")
  if (!is.null(.fun) && (!is.list(.fun) || is.null(names(.fun)) ||
                         any(names(.fun) == "")))
    stop("`.fun` must be a named list of functions, or NULL")
  crs <- .resolve_crs(x, crs)
  dots <- list(...)

  group_fn <- function(df) {
    # Native union off the hex-WKB column. Extra st_union arguments (e.g.
    # is_coverage = TRUE) are not expressible through GEOSUnaryUnion, and
    # geographic data with s2 on unions on the sphere, so both union through sf
    # instead; the planar projected case unions natively.
    if (length(dots) || !.geos_planar_ok(crs)) {
      sb    <- .sf_decode_chunk(df, geom, NULL, crs)
      u     <- do.call(sf::st_union, c(list(sf::st_geometry(sb)), dots))
      u_hex <- sf::st_as_binary(u, hex = TRUE)
    } else {
      u_hex <- .Call(C_geos_union_hex, as.character(df[[geom]]))
    }
    row <- if (is.null(by)) df[1, character(0), drop = FALSE]
           else df[1, by, drop = FALSE]
    if (!is.null(.fun))
      for (nm in names(.fun)) row[[nm]] <- .fun[[nm]](df)
    row[[geom]] <- u_hex
    rownames(row) <- NULL
    row
  }
  .partition_each(x, by, geom, crs, group_fn, flush_rows)
}

# -- set-wise geometry constructions ------------------------------------------

# The constructions that need every feature of a group at once. Each maps the
# group's combined geometry `gu` (a length-1 sfc) to the construction: the
# enclosing kinds return one geometry, the tessellation kinds (voronoi,
# delaunay) return one polygon per cell. `pole` is the centre of the maximum
# inscribed circle (the point inside the shape farthest from its edges, the QGIS
# "pole of inaccessibility"). The inscribed-circle and pole kinds need a positive
# tolerance; the caller supplies one derived from the extent when none is given.
.CONSTRUCT_KINDS <- c("convex_hull", "concave_hull", "envelope", "oriented_box",
                      "enclosing_circle", "inscribed_circle", "pole",
                      "voronoi", "delaunay")

# Tolerance for the kinds that require one. inscribed_circle/pole take a small
# fraction of the bounding-box diagonal so the circle is found to a sensible
# precision; the tessellations accept 0 (GEOS picks its own snap).
.construct_tol <- function(gu, kind) {
  if (!kind %in% c("inscribed_circle", "pole")) return(0)
  bb <- sf::st_bbox(gu)
  d  <- sqrt((bb[["xmax"]] - bb[["xmin"]])^2 + (bb[["ymax"]] - bb[["ymin"]])^2)
  if (is.finite(d) && d > 0) d * 1e-3 else 1e-6
}

# Build the construction for one group. `gu` is the group's combined geometry
# (an sfc of length 1); returns an sfc of the result (length 1 for the enclosing
# kinds, one polygon per cell for the tessellations). The inscribed-circle path
# drops the empty companion geometry sf returns alongside the circle.
.construct_group <- function(gu, kind, ratio, allow_holes, tol) {
  switch(kind,
    convex_hull      = sf::st_convex_hull(gu),
    concave_hull     = sf::st_concave_hull(gu, ratio = ratio,
                                           allow_holes = allow_holes),
    envelope         = sf::st_as_sfc(sf::st_bbox(gu)),
    oriented_box     = sf::st_minimum_rotated_rectangle(gu),
    enclosing_circle = sf::st_minimum_bounding_circle(gu),
    inscribed_circle = {
      ic <- sf::st_inscribed_circle(gu, dTolerance = tol)
      ic[!sf::st_is_empty(ic)]
    },
    pole = {
      ic <- sf::st_inscribed_circle(gu, dTolerance = tol)
      ic <- ic[!sf::st_is_empty(ic)]
      suppressWarnings(sf::st_centroid(ic))
    },
    voronoi  = sf::st_collection_extract(
                 sf::st_voronoi(gu, dTolerance = tol), "POLYGON"),
    delaunay = sf::st_collection_extract(
                 sf::st_triangulate(gu, dTolerance = tol), "POLYGON"))
}

#' Build a set-wise geometry construction, optionally per group
#'
#' Constructs one geometry (or a tessellation) from a whole set of features --
#' the constructions a per-feature [spatial_map()] cannot express because they
#' need every feature in scope at once. Like [spatial_dissolve()] it rides the
#' **partition tier**: `x` is spilled once and routed into one disjoint shard per
#' `by` group in a single bounded pass, then each shard's geometry is combined
#' and the construction built with \pkg{sf}. With no `by`, the whole layer yields
#' one construction. Peak memory is the routing budget during the pass, then one
#' group's geometry while it is built -- partition on a key whose groups fit in
#' memory.
#'
#' `kind` selects the construction:
#' \describe{
#'   \item{`"convex_hull"`}{the convex hull of the set.}
#'   \item{`"concave_hull"`}{the concave hull (`ratio`, `allow_holes`).}
#'   \item{`"envelope"`}{the axis-aligned bounding rectangle.}
#'   \item{`"oriented_box"`}{the minimum-area rotated bounding rectangle.}
#'   \item{`"enclosing_circle"`}{the minimum bounding circle.}
#'   \item{`"inscribed_circle"`}{the maximum inscribed circle (largest circle
#'     that fits inside the set's union).}
#'   \item{`"pole"`}{the pole of inaccessibility -- the centre of the maximum
#'     inscribed circle, the point inside the shape farthest from its edges.}
#'   \item{`"voronoi"`}{the Voronoi tessellation, one polygon per cell.}
#'   \item{`"delaunay"`}{the Delaunay triangulation, one polygon per triangle.}
#' }
#' The enclosing kinds and `pole` emit one feature per group; `voronoi` and
#' `delaunay` emit one feature per cell, each carrying the group's `by` values.
#'
#' Geometry travels through the engine as hex-encoded WKB in a string column and
#' the CRS is carried on the returned node; use [collect_sf()] to materialize.
#' Topology is \pkg{sf}/GEOS throughout (an optional dependency, Suggests); some
#' constructions need projected coordinates.
#'
#' @inheritParams spatial_map
#' @param kind The construction to build; one of the values above.
#' @param by Character vector of attribute columns to construct within: one
#'   construction (or tessellation) per distinct combination of their values.
#'   `NULL` (default) builds a single construction from the whole layer.
#' @param ratio For `kind = "concave_hull"`, the concaveness in `[0, 1]` (1 is
#'   the convex hull). Default `0.3`.
#' @param allow_holes For `kind = "concave_hull"`, whether the hull may contain
#'   holes. Default `FALSE`.
#' @param tolerance Distance tolerance for the kinds that take one
#'   (`"inscribed_circle"`, `"pole"`, `"voronoi"`, `"delaunay"`). `0` (default)
#'   lets the inscribed-circle kinds derive a tolerance from the extent and the
#'   tessellations use the GEOS default.
#'
#' @return A `vectra_node` of the construction -- one row per group for the
#'   enclosing kinds, one row per cell for the tessellations -- carrying the
#'   `by` columns and the input CRS, backed by temporary `.vtr` spills removed
#'   when the node is garbage-collected.
#'
#' @seealso [spatial_dissolve()] to merge a group into one feature,
#'   [spatial_map()] for per-feature transforms, [collect_sf()] to materialize.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#' nc$band <- nc$SID74 > 5
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   band = nc$band,
#'   geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
#' ), f)
#'
#' # One convex hull per band.
#' tbl(f) |>
#'   spatial_construct("convex_hull", by = "band", crs = sf::st_crs(nc)) |>
#'   collect_sf()
#' unlink(f)
#'
#' @export
spatial_construct <- function(x, kind = .CONSTRUCT_KINDS, by = NULL,
                              geom = "geometry", crs = NA, ratio = 0.3,
                              allow_holes = FALSE, tolerance = 0,
                              flush_rows = NULL) {
  .check_sf()
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (the streamed layer to construct from)")
  kind <- match.arg(kind)
  if (!is.null(by) && !is.character(by))
    stop("`by` must be a character vector of column names, or NULL")
  crs <- .resolve_crs(x, crs)

  group_fn <- function(df) {
    sb <- .sf_decode_chunk(df, geom, NULL, crs)
    gu <- sf::st_union(sf::st_geometry(sb))
    tol <- if (tolerance > 0) tolerance else .construct_tol(gu, kind)
    out <- .construct_group(gu, kind, ratio, allow_holes, tol)
    out <- out[!sf::st_is_empty(out)]
    if (!length(out)) return(NULL)
    rowdf <- if (is.null(by)) data.frame(matrix(nrow = length(out), ncol = 0))
             else df[rep(1L, length(out)), by, drop = FALSE]
    rowdf[[geom]] <- sf::st_as_binary(out, hex = TRUE)
    rownames(rowdf) <- NULL
    rowdf
  }
  .partition_each(x, by, geom, crs, group_fn, flush_rows)
}

#' Rasterize a streamed point layer onto a fixed grid
#'
#' Folds a larger-than-RAM stream of points into a fixed raster grid one batch
#' at a time. The grid (`template`) is held resident in memory while the points
#' flow past the engine, so peak memory is the grid plus one batch regardless of
#' how many points there are -- the streaming counterpart to running
#' `terra::rasterize()` on a point set that has to fit in RAM. Each point's
#' coordinate is mapped to its grid cell through the raster geotransform and the
#' per-cell value is accumulated in C.
#'
#' The reduction `fun` is a monoid over the points falling in each cell:
#' `"count"` tallies points (no `field` needed); `"sum"`, `"mean"`, `"min"`,
#' `"max"` aggregate a numeric `field`. Cells that receive no point take the
#' `background` value (`NA` by default). This is the *monoid fold* tier of the
#' spatial toolbox: bounded memory, a single streaming pass, no spill.
#'
#' Points arrive either as two numeric coordinate columns (`coords`, the default
#' and fully \pkg{sf}-free path -- the headline larger-than-RAM case) or decoded
#' from a hex-WKB point-geometry column (`geom`, which needs \pkg{sf}). Geometry
#' input is expected to be points (one coordinate per row); line and polygon
#' coverage rasterization is out of scope here.
#'
#' @param x A `vectra_node` streaming the points (from [tbl()], [tbl_csv()], any
#'   verb chain). It is consumed by the stream.
#' @param template Optional grid to borrow geometry and CRS from: a
#'   `vectra_raster` (from [vec_open_raster()]), or a numeric
#'   `c(xmin, ymin, xmax, ymax)` extent. When omitted, supply `extent` with
#'   `res` or `dims`.
#' @param field Name of a numeric column to aggregate. Required for every `fun`
#'   except `"count"` (which ignores it).
#' @param fun Reduction over the points in each cell: one of `"count"`, `"sum"`,
#'   `"mean"`, `"min"`, `"max"`. `NA` values in `field` are skipped.
#' @param extent Numeric `c(xmin, ymin, xmax, ymax)` defining the grid extent
#'   when no `template` is given.
#' @param res Cell size: a single number for square cells, or `c(xres, yres)`.
#'   The cell counts are rounded to fit `extent` exactly. Supply `res` or `dims`.
#' @param dims Grid shape `c(nrow, ncol)`, an alternative to `res`.
#' @param coords Length-2 character vector naming the x and y coordinate
#'   columns. Default `c("x", "y")`. Ignored when `geom` is supplied.
#' @param geom Name of a hex-WKB point-geometry column to rasterize instead of
#'   coordinate columns. Requires \pkg{sf}.
#' @param crs Coordinate reference system recorded on the output, in any form
#'   [sf::st_crs()] accepts or a bare EPSG integer. Defaults to the template's,
#'   then the node's, else unknown.
#' @param background Value for cells that receive no point. Default `NA_real_`.
#' @param path Optional output path. When given, the grid is written to a `.vec`
#'   raster via [vec_write_raster()] and the opened [vec_open_raster()] handle is
#'   returned invisibly. When `NULL`, the grid is returned in memory.
#' @param dtype Storage dtype for the `.vec` output (see [vec_write_raster()]).
#'   Default `"f32"`.
#'
#' @return When `path` is `NULL`, a numeric matrix with `nrow` grid rows
#'   (row 1 northmost) and `ncol` grid columns, carrying `gt`, `extent`, `res`,
#'   `crs`, and `fun` attributes. When `path` is given, the written
#'   `vectra_raster` handle (invisibly).
#'
#' @seealso [vec_write_raster()] and [vec_to_tiff()] for raster output,
#'   [spatial_join()] to instead tag points with polygon attributes.
#'
#' @examples
#' set.seed(1)
#' n <- 1e4
#' pts <- data.frame(x = runif(n, 0, 10), y = runif(n, 0, 10), z = rnorm(n))
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(pts, f)
#'
#' # Point density on a 10x10 grid, streamed: the grid is resident, the
#' # points are not.
#' counts <- tbl(f) |> rasterize(extent = c(0, 0, 10, 10), dims = c(10, 10))
#' counts
#'
#' # Mean of z per cell.
#' zmean <- tbl(f) |>
#'   rasterize(extent = c(0, 0, 10, 10), dims = c(10, 10),
#'             field = "z", fun = "mean")
#' unlink(f)
#'
#' @export
rasterize <- function(x, template = NULL, field = NULL,
                      fun = c("count", "sum", "mean", "min", "max"),
                      extent = NULL, res = NULL, dims = NULL,
                      coords = c("x", "y"), geom = NULL, crs = NA,
                      background = NA_real_, path = NULL, dtype = "f32") {
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (build one with tbl(), tbl_csv(), ...)")
  fun <- match.arg(fun)
  fun_code <- switch(fun, count = 0L, sum = 1L, mean = 2L, min = 3L, max = 4L)
  if (fun != "count" && is.null(field))
    stop(sprintf("fun = \"%s\" needs a `field` column to aggregate", fun))

  crs  <- .resolve_crs(x, crs)
  grid <- .rasterize_grid(template, extent, res, dims, crs)
  crs  <- grid$crs

  use_geom <- !is.null(geom)
  if (use_geom) .check_sf()

  acc <- .Call(C_rasterize_new,
               c(as.integer(grid$width), as.integer(grid$height)),
               as.numeric(grid$gt), fun_code)

  nxt <- .batch_cursor(x)
  repeat {
    chunk <- nxt(); if (is.null(chunk)) break
    if (use_geom) {
      sb <- .sf_decode_chunk(chunk, geom, NULL, crs)
      xy <- sf::st_coordinates(sf::st_geometry(sb))
      if (nrow(xy) != nrow(chunk))
        stop("geom rasterize expects point geometry (one coordinate per row)")
      xs <- as.numeric(xy[, "X"]); ys <- as.numeric(xy[, "Y"])
    } else {
      miss <- setdiff(coords, names(chunk))
      if (length(miss))
        stop(sprintf("coords column(s) not found: %s",
                     paste(miss, collapse = ", ")))
      xs <- as.numeric(chunk[[coords[1L]]])
      ys <- as.numeric(chunk[[coords[2L]]])
    }
    vals <- NULL
    if (!is.null(field)) {
      if (!field %in% names(chunk))
        stop(sprintf("field column '%s' not found", field))
      vals <- as.numeric(chunk[[field]])
    }
    .Call(C_rasterize_push, acc, xs, ys, vals)
  }

  m <- .Call(C_rasterize_finish, acc, as.numeric(background))
  attr(m, "gt")     <- grid$gt
  attr(m, "extent") <- grid$extent
  attr(m, "res")    <- grid$res
  attr(m, "crs")    <- crs
  attr(m, "fun")    <- fun

  if (!is.null(path)) {
    vec_write_raster(m, path, dtype = dtype, gt = grid$gt,
                     epsg = grid$epsg, nodata = as.numeric(background))
    return(invisible(vec_open_raster(path)))
  }
  m
}

# -- zonal statistics (per-zone raster summaries) -----------------------------

# Per-zone running moments, keyed by zone label. n / s (sum) / ss (sum of
# squares) combine additively across strips; mn / mx combine by pmin / pmax;
# `na` records zones tainted by an NA value when na.rm = FALSE. Each is a named
# numeric vector indexed by the zone label, grown by union as new zones appear.
.zonal_acc <- function() {
  e <- new.env(parent = emptyenv())
  e$n <- e$s <- e$ss <- e$mn <- e$mx <- stats::setNames(numeric(0), character(0))
  e$na <- character(0)
  e
}

.zonal_merge_add <- function(a, b) {
  if (!length(a)) return(b)
  nm  <- union(names(a), names(b))
  out <- stats::setNames(numeric(length(nm)), nm)
  out[names(a)] <- out[names(a)] + a
  out[names(b)] <- out[names(b)] + b
  out
}

.zonal_merge_fun <- function(a, b, f) {
  if (!length(a)) return(b)
  nm <- union(names(a), names(b))
  oa <- stats::setNames(rep(NA_real_, length(nm)), nm); oa[names(a)] <- a
  ob <- stats::setNames(rep(NA_real_, length(nm)), nm); ob[names(b)] <- b
  stats::setNames(f(oa, ob, na.rm = TRUE), nm)
}

# Fold one strip's (zone, value) pixels into the accumulator. `z` and `v` are
# aligned vectors; NA handling has already split out tainted zones for the
# na.rm = FALSE path via `tainted`.
.zonal_update <- function(e, z, v, tainted = character(0)) {
  if (length(tainted)) e$na <- union(e$na, tainted)
  keep <- !is.na(z) & !is.na(v)
  z <- z[keep]; v <- v[keep]
  if (!length(z)) return(invisible())
  zc  <- as.character(z)
  agg <- rowsum(cbind(rep.int(1, length(v)), v, v * v), zc)
  labs <- rownames(agg)
  e$n  <- .zonal_merge_add(e$n,  stats::setNames(agg[, 1L], labs))
  e$s  <- .zonal_merge_add(e$s,  stats::setNames(agg[, 2L], labs))
  e$ss <- .zonal_merge_add(e$ss, stats::setNames(agg[, 3L], labs))
  e$mn <- .zonal_merge_fun(e$mn, tapply(v, zc, min), pmin)
  e$mx <- .zonal_merge_fun(e$mx, tapply(v, zc, max), pmax)
  invisible()
}

# Resolve a value/zone raster argument that may be a path or an open handle.
# Returns list(r, close) so the caller closes only handles it opened.
.zonal_open <- function(x, what) {
  if (inherits(x, "vectra_raster")) return(list(r = x, close = FALSE))
  if (is.character(x) && length(x) == 1L)
    return(list(r = vec_open_raster(x), close = TRUE))
  stop(sprintf("`%s` must be a vectra_raster or a path to a .vec raster", what))
}

# Pixel-centre coordinates for a strip of rows r0:r1 over the full width, in the
# column-major order that as.vector() of the read window produces (row within
# strip varies fastest). gt is the value raster's north-up geotransform.
.zonal_strip_xy <- function(gt, r0, r1, width) {
  h <- r1 - r0 + 1L
  rows_idx <- rep.int(r0:r1, width)
  cols_idx <- rep(seq_len(width), each = h)
  list(x = gt[1L] + (cols_idx - 0.5) * gt[2L],
       y = gt[4L] + (rows_idx - 0.5) * gt[6L])
}

#' Summarise raster values within zones
#'
#' Reduces a raster to one summary row per zone, streaming the raster one
#' tile-row strip at a time so the whole grid never has to be resident. Zones
#' come either from a second raster aligned to the value grid (each pixel's zone
#' is that raster's value, the \code{terra::zonal} pattern) or from an \pkg{sf}
#' polygon layer (each pixel is assigned the polygon its centre falls in). The
#' per-zone running moments (count, sum, sum of squares, min, max) are folded in
#' memory as strips arrive, so peak memory is one strip plus the small per-zone
#' table regardless of raster size. This is the *monoid fold* tier of the
#' spatial toolbox: bounded memory, a single streaming pass, no spill.
#'
#' `sd` is derived from the streamed moments
#' (`sqrt((sum2 - sum^2 / n) / (n - 1))`), so it needs no second pass. With
#' `na.rm = TRUE` (the default) nodata pixels are skipped; with `na.rm = FALSE`
#' any nodata pixel in a zone makes that zone's `sum`/`mean`/`min`/`max`/`sd`
#' `NA`, matching the resident behaviour. `count` always reports the number of
#' non-nodata cells in the zone.
#'
#' Polygon zones assign each pixel centre to a polygon natively on the GEOS C
#' API (the polygons parsed once into a spatial index, every strip's centres
#' located in C), so \pkg{sf} is touched only to read the polygons in; geographic
#' polygons with spherical geometry on (`sf::sf_use_s2()`) keep the \pkg{sf}
#' point-in-polygon path. Raster zones are fully \pkg{sf}-free. The zone raster
#' must share the value raster's dimensions and geotransform.
#'
#' @param raster A `vectra_raster` (from [vec_open_raster()]) or a path to a
#'   `.vec` raster holding the values to summarise.
#' @param zones The zones to summarise within: a `vectra_raster` / `.vec` path
#'   aligned to `raster` (zone id per pixel), or an `sf`/`sfc` polygon layer.
#' @param fun One or more of `"mean"`, `"sum"`, `"count"`, `"min"`, `"max"`,
#'   `"sd"`. Each becomes a column in the result. Default `"mean"`.
#' @param band Band of the value `raster` to summarise (1-based). Default 1.
#' @param zone_band Band of a raster `zones` holding the zone ids. Default 1.
#' @param zone_field For an `sf` `zones` layer, the column giving each polygon's
#'   zone id. Default `NULL` uses the polygon row index `1:n`.
#' @param na.rm If `TRUE` (default) skip nodata pixels; if `FALSE` let a nodata
#'   pixel propagate `NA` to its zone's statistics.
#'
#' @return A data.frame with a `zone` column (sorted) followed by one column per
#'   `fun`, one row per zone.
#'
#' @seealso [rasterize()] to build a value raster from streamed points,
#'   [vec_open_raster()] to open the inputs.
#'
#' @examples
#' # A value raster and an aligned 2x2-block zone raster on a 4x4 grid.
#' vals <- matrix(1:16, 4, 4, byrow = TRUE)
#' zone <- matrix(c(1, 1, 2, 2, 1, 1, 2, 2,
#'                  3, 3, 4, 4, 3, 3, 4, 4), 4, 4, byrow = TRUE)
#' fv <- tempfile(fileext = ".vec"); fz <- tempfile(fileext = ".vec")
#' vec_write_raster(vals, fv, dtype = "f64", extent = c(0, 0, 4, 4))
#' vec_write_raster(zone, fz, dtype = "f64", extent = c(0, 0, 4, 4))
#'
#' zonal(fv, fz, fun = c("mean", "sum", "count"))
#' unlink(c(fv, fz))
#'
#' @export
zonal <- function(raster, zones, fun = "mean", band = 1L, zone_band = 1L,
                  zone_field = NULL, na.rm = TRUE) {
  fun <- match.arg(fun, c("mean", "sum", "count", "min", "max", "sd"),
                   several.ok = TRUE)

  vh <- .zonal_open(raster, "raster")
  r  <- vh$r
  if (vh$close) on.exit(vec_close_raster(r), add = TRUE)
  gt <- r$gt; W <- as.integer(r$width); H <- as.integer(r$height)
  ts <- max(1L, as.integer(r$tile_size))

  use_raster_zones <- inherits(zones, "vectra_raster") ||
    (is.character(zones) && length(zones) == 1L)
  if (use_raster_zones) {
    zh <- .zonal_open(zones, "zones")
    zr <- zh$r
    if (zh$close) on.exit(vec_close_raster(zr), add = TRUE)
    if (zr$width != W || zr$height != H)
      stop("zone raster must match the value raster's dimensions")
    if (!isTRUE(all.equal(as.numeric(zr$gt), as.numeric(gt))))
      stop("zone raster geotransform must match the value raster's")
    numeric_zone <- TRUE
  } else {
    if (!inherits(zones, "sf") && !inherits(zones, "sfc"))
      stop("`zones` must be a vectra_raster, a .vec path, or an sf/sfc layer")
    .check_sf()
    zid <- if (inherits(zones, "sfc") || is.null(zone_field)) {
      seq_len(length(sf::st_geometry(zones)))
    } else {
      if (!zone_field %in% names(zones))
        stop(sprintf("zone_field '%s' not found in `zones`", zone_field))
      zones[[zone_field]]
    }
    numeric_zone <- is.numeric(zid)
    zcrs <- sf::st_crs(zones)
    if (is.na(zcrs) && !is.null(r$epsg) && r$epsg > 0L)
      zcrs <- sf::st_crs(r$epsg)
    zones_g <- sf::st_geometry(zones)
    if (is.na(sf::st_crs(zones_g)) && !is.na(zcrs))
      zones_g <- sf::st_set_crs(zones_g, zcrs)
    # Native point-in-polygon for each pixel centre, planar-gated like the other
    # GEOS paths; spherical geographic zones (s2 on) keep the sf intersects loop.
    native_zones <- .geos_planar_ok(zcrs)
    if (native_zones) {
      zloc <- .geos_locator(zones_g)
      znt  <- .spatial_threads()
    }
  }

  acc <- .zonal_acc()
  r0  <- 1L
  while (r0 <= H) {
    r1 <- min(r0 + ts - 1L, H)
    vm <- vec_read_window(r, band = band, cols = c(1L, W), rows = c(r0, r1))
    v  <- as.vector(vm)

    if (use_raster_zones) {
      zm <- vec_read_window(zr, band = zone_band, cols = c(1L, W),
                            rows = c(r0, r1))
      z <- as.vector(zm)
    } else if (native_zones) {
      xy <- .zonal_strip_xy(gt, r0, r1, W)
      first <- .Call(C_geos_locate_xy, zloc, as.numeric(xy$x), as.numeric(xy$y),
                     0L, 0, FALSE, znt)
      z <- zid[first]
    } else {
      xy  <- .zonal_strip_xy(gt, r0, r1, W)
      pts <- sf::st_as_sf(data.frame(x = xy$x, y = xy$y),
                          coords = c("x", "y"), crs = zcrs)
      hit <- sf::st_intersects(pts, zones_g)
      first <- vapply(hit, function(h) if (length(h)) h[1L] else NA_integer_,
                      integer(1L))
      z <- zid[first]
    }

    tainted <- character(0)
    if (!na.rm) {
      bad <- !is.na(z) & is.na(v)
      if (any(bad)) tainted <- unique(as.character(z[bad]))
    }
    .zonal_update(acc, z, v, tainted)
    r0 <- r1 + 1L
  }

  zlab <- names(acc$n)
  if (!length(zlab))
    return(.zonal_frame(character(0), acc, fun, numeric_zone, na.rm))
  ord <- if (numeric_zone) order(as.numeric(zlab)) else order(zlab)
  .zonal_frame(zlab[ord], acc, fun, numeric_zone, na.rm)
}

# Assemble the per-zone result frame from the accumulated moments, deriving each
# requested statistic and blanking na.rm = FALSE tainted zones.
.zonal_frame <- function(zlab, e, fun, numeric_zone, na.rm) {
  n  <- e$n[zlab]; s <- e$s[zlab]; ss <- e$ss[zlab]
  mn <- e$mn[zlab]; mx <- e$mx[zlab]
  zone <- if (numeric_zone) as.numeric(zlab) else zlab
  out <- data.frame(zone = zone, stringsAsFactors = FALSE)
  for (fn in fun) {
    out[[fn]] <- switch(fn,
      count = as.numeric(n),
      sum   = as.numeric(s),
      mean  = s / n,
      min   = as.numeric(mn),
      max   = as.numeric(mx),
      sd    = sqrt(pmax(0, (ss - s * s / n) / (n - 1))))
  }
  if (!na.rm && length(e$na) && nrow(out)) {
    bad <- as.character(out$zone) %in% e$na
    for (fn in setdiff(fun, "count")) out[[fn]][bad] <- NA_real_
  }
  rownames(out) <- NULL
  out
}

# -- focal / terrain (moving-window raster derivatives) -----------------------

# Drive a haloed tile-row strip pass over a VECR raster. For each output
# tile-row the input is read expanded by `radh` rows (the halo), `compute` maps
# the strip to an out_h x (W*nout) result, and each of the nout output bands is
# either streamed to a .vec (never the whole band resident) or assembled into an
# in-memory matrix. Returns the opened handle (path) or a list of nout matrices.
# Output side shared by the streamed raster kernels (focal/terrain/warp). The
# returned object writes one output tile-row at a time -- either streamed to a
# new .vec (path) or folded into in-memory matrices -- so an op never holds the
# whole output band. `write(ty, r0, r1, os)` takes an out_h x (W*nout) strip
# (derivative k in columns [(k-1)*W+1, k*W]); `finish()` returns the opened
# handle (streamed) or the list of attributed matrices (in memory).
.raster_sink <- function(W, H, nout, gt, epsg, TS,
                         path, dtype, band_names, comp_code) {
  writer <- NULL; acc <- NULL
  if (!is.null(path)) {
    path <- normalizePath(path, mustWork = FALSE)
    writer <- .Call(C_vecr_writer_open, path,
                    c(W, H, as.integer(nout)), as.character(dtype),
                    TS, as.numeric(gt), epsg, NA_real_, band_names, comp_code)
  } else {
    acc <- replicate(nout, matrix(NA_real_, H, W), simplify = FALSE)
  }
  list(
    write = function(ty, r0, r1, os) {
      for (k in seq_len(nout)) {
        strip_k <- os[, ((k - 1L) * W + 1L):(k * W), drop = FALSE]
        if (!is.null(writer)) {
          .Call(C_vecr_writer_write_strip, writer, as.integer(k),
                as.integer(ty), strip_k)
        } else {
          acc[[k]][r0:r1, ] <<- strip_k
        }
      }
    },
    finish = function() {
      if (!is.null(writer)) {
        .Call(C_vecr_writer_finish, writer)
        return(invisible(vec_open_raster(path)))
      }
      ext <- c(gt[1L], gt[4L] + H * gt[6L], gt[1L] + W * gt[2L], gt[4L])
      lapply(acc, function(m) {
        attr(m, "gt")     <- gt
        attr(m, "extent") <- ext
        attr(m, "crs")    <- if (epsg > 0L) epsg else NA
        m
      })
    }
  )
}

.raster_focal_run <- function(r, band, radh, nout, compute,
                              path, dtype, band_names, comp_code) {
  gt <- r$gt; W <- as.integer(r$width); H <- as.integer(r$height)
  TS <- max(1L, as.integer(r$tile_size))
  epsg <- if (!is.null(r$epsg)) as.integer(r$epsg) else 0L
  sink <- .raster_sink(W, H, nout, gt, epsg, TS, path, dtype, band_names, comp_code)

  tiles_y <- (H + TS - 1L) %/% TS
  for (ty in seq_len(tiles_y) - 1L) {
    r0 <- ty * TS + 1L
    r1 <- min(r0 + TS - 1L, H)
    in_r0 <- max(1L, r0 - radh)
    in_r1 <- min(H, r1 + radh)
    vm <- vec_read_window(r, band = band, cols = c(1L, W), rows = c(in_r0, in_r1))
    in_h  <- in_r1 - in_r0 + 1L
    top   <- r0 - in_r0
    out_h <- r1 - r0 + 1L
    os <- compute(vm, as.integer(in_h), as.integer(top), as.integer(out_h))
    sink$write(ty, r0, r1, os)
  }
  sink$finish()
}

# Normalize a focal window argument to an odd-by-odd weight matrix.
.focal_window <- function(w) {
  if (is.numeric(w) && is.null(dim(w)) && length(w) == 1L) {
    if (w < 1 || w %% 2 == 0)
      stop("`w` given as a single number must be a positive odd integer")
    w <- matrix(1, w, w)
  }
  if (!is.matrix(w) || !is.numeric(w))
    stop("`w` must be a numeric weight matrix or a single odd integer")
  if (nrow(w) %% 2 == 0 || ncol(w) %% 2 == 0)
    stop("`w` must have an odd number of rows and columns")
  w
}

#' Moving-window (focal) statistics over a streamed raster
#'
#' Applies a moving window to a `.vec` raster, reading the input one tile-row
#' strip at a time -- each strip expanded by the kernel radius (a halo read) so
#' window neighbours are available without ever holding the whole grid resident.
#' The per-window statistic is computed in C. When `path` is given the output is
#' streamed straight back to a new `.vec` one tile-row at a time, so neither the
#' input nor the output band is ever fully in memory; this is the raster op that
#' runs out of core where an in-memory engine needs the whole raster at once.
#'
#' This is the *sort / partition* tier of the spatial toolbox: bounded to one
#' haloed strip at a time, exploiting tile locality.
#'
#' The window `w` is a numeric weight matrix with odd dimensions (or a single
#' odd integer `k`, shorthand for a `k x k` matrix of ones). `NA` weights mark
#' cells outside the window. For `fun = "sum"`/`"mean"` the weights scale the
#' values (sum is `sum(w * x)`, mean is `sum(w * x) / sum(w)`); for the other
#' statistics a finite weight only marks membership. With `na.rm = TRUE` (the
#' default) nodata cells inside the window are skipped; with `na.rm = FALSE` any
#' nodata cell -- including a window that runs off the raster edge -- makes the
#' result `NA`, matching the resident behaviour.
#'
#' @param x A `vectra_raster` (from [vec_open_raster()]) or a path to a `.vec`
#'   raster.
#' @param w A numeric weight matrix with odd dimensions, or a single positive
#'   odd integer `k` for a `k x k` window of ones. Default `matrix(1, 3, 3)`.
#' @param fun Window statistic: one of `"sum"`, `"mean"`, `"min"`, `"max"`,
#'   `"sd"`, `"median"`. Default `"mean"`.
#' @param na.rm Skip nodata cells inside the window (`TRUE`, default) or let
#'   them propagate `NA` (`FALSE`).
#' @param band Band to read (1-based). Default 1.
#' @param path Optional output `.vec` path. When given the result is streamed to
#'   disk and the opened [vec_open_raster()] handle is returned invisibly; when
#'   `NULL` the result is returned as an in-memory matrix.
#' @param dtype Storage dtype for `.vec` output (see [vec_write_raster()]).
#'   Default `"f32"`.
#' @param compression Compression effort for `.vec` output. Default `"fast"`.
#'
#' @return When `path` is `NULL`, a numeric matrix (row 1 northmost) carrying
#'   `gt`, `extent`, `crs`, and `fun` attributes. When `path` is given, the
#'   written `vectra_raster` handle (invisibly).
#'
#' @seealso [terrain()] for DEM derivatives built on the same strip pass,
#'   [zonal()] for per-zone summaries.
#'
#' @examples
#' m <- matrix(1:36, 6, 6, byrow = TRUE)
#' f <- tempfile(fileext = ".vec")
#' vec_write_raster(m, f, dtype = "f64", extent = c(0, 0, 6, 6))
#'
#' # 3x3 mean smoother; edge cells see off-raster neighbours.
#' focal(f, w = matrix(1, 3, 3), fun = "mean")
#' unlink(f)
#'
#' @export
focal <- function(x, w = matrix(1, 3, 3),
                  fun = c("mean", "sum", "min", "max", "sd", "median"),
                  na.rm = TRUE, band = 1L, path = NULL, dtype = "f32",
                  compression = c("fast", "balanced", "max")) {
  fun <- match.arg(fun)
  fun_code <- switch(fun, sum = 0L, mean = 1L, min = 2L, max = 3L,
                     sd = 4L, median = 5L)
  w <- .focal_window(w)
  kh <- nrow(w); kw <- ncol(w)
  weights <- as.numeric(t(w))
  radh <- (kh - 1L) %/% 2L
  comp_code <- switch(match.arg(compression), fast = 0L, balanced = 1L, max = 2L)

  vh <- .zonal_open(x, "x"); r <- vh$r
  if (vh$close) on.exit(vec_close_raster(r), add = TRUE)
  W <- as.integer(r$width)
  kdims <- c(as.integer(kh), as.integer(kw))

  compute <- function(vm, in_h, top, out_h)
    .Call(C_focal_strip, vm, c(in_h, W), weights, kdims,
          fun_code, isTRUE(na.rm), top, out_h)

  res <- .raster_focal_run(r, as.integer(band), radh, 1L, compute,
                           path, dtype, NULL, comp_code)
  if (!is.null(path)) return(res)
  m <- res[[1L]]; attr(m, "fun") <- fun; m
}

#' Terrain derivatives from a streamed elevation raster
#'
#' Computes DEM derivatives from a `.vec` elevation raster with Horn's 3x3
#' method, on the same haloed tile-row strip pass as [focal()] -- the input is
#' read one strip at a time and, when `path` is given, the outputs are streamed
#' straight back to a multi-band `.vec`. Matches \pkg{terra}'s
#' `terrain()` / `shade()` conventions.
#'
#' @param x A `vectra_raster` (from [vec_open_raster()]) or a path to a `.vec`
#'   elevation raster.
#' @param v Derivatives to compute, any of `"slope"`, `"aspect"`,
#'   `"hillshade"`, `"TPI"` (topographic position index), `"roughness"`,
#'   `"TRI"` (terrain ruggedness index). The return follows the input: one
#'   matrix for a single `v`, a named list for several.
#' @param unit Angular unit for `slope` and `aspect`: `"degrees"` (default) or
#'   `"radians"`.
#' @param azimuth,altitude Sun position for `"hillshade"`, in degrees. Defaults
#'   315 (NW) and 45.
#' @param band Band to read (1-based). Default 1.
#' @param path Optional output `.vec` path (one band per `v`, named after `v`).
#'   When given the result is streamed to disk and the opened
#'   [vec_open_raster()] handle is returned invisibly; when `NULL` the result is
#'   returned in memory.
#' @param dtype Storage dtype for `.vec` output (see [vec_write_raster()]).
#'   Default `"f32"`.
#' @param compression Compression effort for `.vec` output. Default `"fast"`.
#'
#' @return When `path` is `NULL`: a numeric matrix for a single `v`, or a named
#'   list of matrices for several, each carrying `gt`, `extent`, and `crs`
#'   attributes (row 1 northmost). When `path` is given, the written multi-band
#'   `vectra_raster` handle (invisibly).
#'
#' @details Slope and aspect use the Horn (1981) finite-difference gradient over
#'   the 3x3 neighbourhood; `aspect` is degrees clockwise from north (flat cells
#'   return 90). `hillshade` is the cosine of the incidence angle for the given
#'   sun position, clamped at 0. `TPI` is the cell minus the mean of its eight
#'   neighbours; `roughness` is the range over the 3x3; `TRI` is the mean
#'   absolute difference to the eight neighbours. Cells whose 3x3 neighbourhood
#'   touches a nodata value or the raster edge return `NA`.
#'
#' @seealso [focal()] for arbitrary moving windows.
#'
#' @examples
#' # A tilted surface so slope and aspect are well defined.
#' z <- outer(1:8, 1:8, function(r, c) 10 + 2 * c + r)
#' f <- tempfile(fileext = ".vec")
#' vec_write_raster(z, f, dtype = "f64", extent = c(0, 0, 8, 8))
#'
#' slp <- terrain(f, v = "slope")
#' deriv <- terrain(f, v = c("slope", "aspect", "hillshade"))
#' names(deriv)
#' unlink(f)
#'
#' @export
terrain <- function(x, v = c("slope", "aspect", "hillshade",
                             "TPI", "roughness", "TRI"),
                    unit = c("degrees", "radians"),
                    azimuth = 315, altitude = 45,
                    band = 1L, path = NULL, dtype = "f32",
                    compression = c("fast", "balanced", "max")) {
  v <- match.arg(v, several.ok = TRUE)
  codes <- c(slope = 0L, aspect = 1L, hillshade = 2L,
             TPI = 3L, roughness = 4L, TRI = 5L)
  which_codes <- as.integer(unname(codes[v]))
  unit_code <- if (match.arg(unit) == "radians") 1L else 0L
  sun <- c(as.numeric(azimuth), as.numeric(altitude))
  comp_code <- switch(match.arg(compression), fast = 0L, balanced = 1L, max = 2L)

  vh <- .zonal_open(x, "x"); r <- vh$r
  if (vh$close) on.exit(vec_close_raster(r), add = TRUE)
  W <- as.integer(r$width)
  gt <- r$gt
  res <- c(abs(gt[2L]), abs(gt[6L]))
  nout <- length(v)

  compute <- function(vm, in_h, top, out_h)
    .Call(C_terrain_strip, vm, c(in_h, W), which_codes, top, out_h,
          res, unit_code, sun)

  out <- .raster_focal_run(r, as.integer(band), 1L, nout, compute,
                           path, dtype, v, comp_code)
  if (!is.null(path)) return(out)
  names(out) <- v
  if (nout == 1L) out[[1L]] else out
}

# Resolve the warp target grid to list(W, H, gt, epsg). `template` is either a
# vectra_raster / .vec path whose grid is borrowed wholesale, or a list spec
# list(crs=, extent=, res=, dims=). With no `extent` the target extent is the
# source's corners projected into the target CRS (needs sf for a real
# reprojection); `res` or `dims` then sets the spacing.
.warp_grid <- function(template, src) {
  if (inherits(template, "vectra_raster") ||
      (is.character(template) && length(template) == 1L)) {
    th <- .zonal_open(template, "template")
    t  <- th$r
    g  <- list(W = as.integer(t$width), H = as.integer(t$height),
               gt = as.numeric(t$gt),
               epsg = if (!is.null(t$epsg)) as.integer(t$epsg) else 0L)
    if (th$close) vec_close_raster(t)
    return(g)
  }
  if (!is.list(template))
    stop("`template` must be a vectra_raster, a .vec path, or a list(crs=, extent=, res=)")

  epsg <- template$crs
  if (is.null(epsg)) epsg <- if (!is.null(src$epsg)) src$epsg else 0L
  epsg <- as.integer(epsg)
  res <- template$res; extent <- template$extent; dims <- template$dims

  if (is.null(extent)) {
    if (is.null(res))
      stop("a list `template` without `extent` needs `res` to set the grid")
    src_epsg <- if (!is.null(src$epsg)) as.integer(src$epsg) else 0L
    sg <- as.numeric(src$gt); sW <- src$width; sH <- src$height
    cx <- c(0, sW, 0, sW); cy <- c(0, 0, sH, sH)
    X <- sg[1L] + cx * sg[2L] + cy * sg[3L]
    Y <- sg[4L] + cx * sg[5L] + cy * sg[6L]
    if (epsg > 0L && src_epsg > 0L && epsg != src_epsg) {
      .check_sf()
      p <- sf::sf_project(sf::st_crs(src_epsg), sf::st_crs(epsg), cbind(X, Y))
      X <- p[, 1L]; Y <- p[, 2L]
    }
    extent <- c(min(X), min(Y), max(X), max(Y))
  }
  xmin <- extent[1L]; ymin <- extent[2L]; xmax <- extent[3L]; ymax <- extent[4L]

  if (!is.null(res)) {
    xres <- res[1L]; yres <- if (length(res) >= 2L) res[2L] else res[1L]
    W <- max(1L, as.integer(round((xmax - xmin) / xres)))
    H <- max(1L, as.integer(round((ymax - ymin) / yres)))
  } else if (!is.null(dims)) {
    W <- as.integer(dims[1L]); H <- as.integer(dims[2L])
    xres <- (xmax - xmin) / W; yres <- (ymax - ymin) / H
  } else {
    stop("a list `template` needs `res` or `dims`")
  }
  list(W = W, H = H, gt = c(xmin, xres, 0, ymax, 0, -yres), epsg = epsg)
}

#' Resample or reproject a streamed raster onto a target grid
#'
#' Warps a `.vec` raster onto a target grid, walking the *output* one tile-row
#' strip at a time. For each strip the target pixel-centre coordinates are built,
#' projected into the source coordinate reference system when the two CRSs differ
#' (delegated to PROJ via \pkg{sf}), mapped through the source geotransform to
#' fractional source pixels, and sampled from the bounded source window those
#' coordinates fall in. The output is assembled in memory or streamed straight
#' back to a new `.vec`, so the whole output grid is never resident; the source
#' is read in bounded windows rather than held whole.
#'
#' This is the *sort / partition* tier of the spatial toolbox: each output strip
#' reads the source window it projects onto. For a mild reprojection or a plain
#' resample that window is a thin band; a strong reprojection can make it large,
#' but the output stays streamed throughout.
#'
#' Sampling follows the GDAL / \pkg{terra} convention (pixel centres at
#' half-integer coordinates). `"near"` takes the nearest source cell;
#' `"bilinear"` the 2x2 weighted mean; `"cubic"` the 4x4 cubic convolution
#' (Catmull-Rom, a = -0.5). A target cell whose sampling kernel reaches outside
#' the source extent, or touches a nodata cell, comes back `NA`.
#'
#' Reprojection happens only when both rasters carry a known EPSG code and the
#' codes differ; otherwise `warp()` resamples within a shared CRS and needs no
#' \pkg{sf}.
#'
#' @param x A `vectra_raster` (from [vec_open_raster()]) or a path to a `.vec`
#'   raster to warp.
#' @param template The target grid: a `vectra_raster` / `.vec` path whose grid
#'   and CRS are borrowed, or a list `list(crs =, extent =, res =, dims =)`.
#'   With `crs` and `res` but no `extent`, the target extent is the source's
#'   corners projected into `crs`.
#' @param method Resampling method: `"near"`, `"bilinear"`, or `"cubic"`.
#'   Default `"near"`.
#' @param band Band to warp (1-based). Default 1.
#' @param path Optional output `.vec` path. When given the result is streamed to
#'   disk and the opened [vec_open_raster()] handle is returned invisibly; when
#'   `NULL` the result is returned as an in-memory matrix.
#' @param dtype Storage dtype for `.vec` output (see [vec_write_raster()]).
#'   Default `"f32"`.
#' @param compression Compression effort for `.vec` output. Default `"fast"`.
#'
#' @return When `path` is `NULL`, a numeric matrix on the target grid (row 1
#'   northmost) carrying `gt`, `extent`, and `crs` attributes. When `path` is
#'   given, the written `vectra_raster` handle (invisibly).
#'
#' @seealso [rasterize()] to build a raster from streamed vector features,
#'   [focal()] for moving-window statistics.
#'
#' @examples
#' z <- outer(1:8, 1:8, function(r, c) r + 2 * c)
#' f <- tempfile(fileext = ".vec")
#' vec_write_raster(z, f, dtype = "f64", extent = c(0, 0, 8, 8))
#'
#' # Resample onto a finer grid over the same extent.
#' fine <- warp(f, list(extent = c(0, 0, 8, 8), res = 0.5), method = "bilinear")
#' dim(fine)
#' unlink(f)
#'
#' @export
warp <- function(x, template, method = c("near", "bilinear", "cubic"),
                 band = 1L, path = NULL, dtype = "f32",
                 compression = c("fast", "balanced", "max")) {
  method_code <- switch(match.arg(method), near = 0L, bilinear = 1L, cubic = 2L)
  comp_code <- switch(match.arg(compression), fast = 0L, balanced = 1L, max = 2L)

  vh <- .zonal_open(x, "x"); src <- vh$r
  if (vh$close) on.exit(vec_close_raster(src), add = TRUE)
  band <- as.integer(band)

  tg <- .warp_grid(template, src)
  W <- tg$W; H <- tg$H; gt_t <- tg$gt; epsg_t <- tg$epsg
  gt_s <- as.numeric(src$gt)
  sW <- as.integer(src$width); sH <- as.integer(src$height)
  epsg_s <- if (!is.null(src$epsg)) as.integer(src$epsg) else 0L

  reproject <- epsg_t > 0L && epsg_s > 0L && epsg_t != epsg_s
  if (reproject) .check_sf()

  det <- gt_s[2L] * gt_s[6L] - gt_s[3L] * gt_s[5L]
  if (abs(det) < .Machine$double.eps)
    stop("source geotransform is not invertible")
  margin <- method_code   # near = 0, bilinear = 1, cubic = 2 kernel half-width

  TS <- max(1L, as.integer(src$tile_size))
  sink <- .raster_sink(W, H, 1L, gt_t, epsg_t, TS, path, dtype, NULL, comp_code)

  tiles_y <- (H + TS - 1L) %/% TS
  for (ty in seq_len(tiles_y) - 1L) {
    r0 <- ty * TS + 1L
    r1 <- min(r0 + TS - 1L, H)
    out_h <- r1 - r0 + 1L

    # Target pixel centres for this strip, column-major (output row fastest).
    grow <- rep.int(seq.int(0L, out_h - 1L), W)
    gcol <- rep(seq.int(0L, W - 1L), each = out_h)
    gr <- (r0 - 1L) + grow
    Xt <- gt_t[1L] + (gcol + 0.5) * gt_t[2L] + (gr + 0.5) * gt_t[3L]
    Yt <- gt_t[4L] + (gcol + 0.5) * gt_t[5L] + (gr + 0.5) * gt_t[6L]

    if (reproject) {
      p <- sf::sf_project(sf::st_crs(epsg_t), sf::st_crs(epsg_s),
                          cbind(Xt, Yt), keep = TRUE)
      Xs <- p[, 1L]; Ys <- p[, 2L]
    } else {
      Xs <- Xt; Ys <- Yt
    }

    dx <- Xs - gt_s[1L]; dy <- Ys - gt_s[4L]
    sx <- (dx * gt_s[6L] - dy * gt_s[3L]) / det   # fractional source col (edge)
    sy <- (dy * gt_s[2L] - dx * gt_s[5L]) / det   # fractional source row (edge)
    sx[!is.finite(sx)] <- NA_real_
    sy[!is.finite(sy)] <- NA_real_

    fin <- !is.na(sx) & !is.na(sy)
    if (any(fin)) {
      cmin <- max(0L, as.integer(floor(min(sx[fin]) - 0.5)) - margin - 1L)
      cmax <- min(sW - 1L, as.integer(ceiling(max(sx[fin]) - 0.5)) + margin + 1L)
      rmin <- max(0L, as.integer(floor(min(sy[fin]) - 0.5)) - margin - 1L)
      rmax <- min(sH - 1L, as.integer(ceiling(max(sy[fin]) - 0.5)) + margin + 1L)
    } else {
      cmin <- 1L; cmax <- 0L; rmin <- 1L; rmax <- 0L
    }

    if (cmax >= cmin && rmax >= rmin) {
      win <- vec_read_window(src, band = band,
                             cols = c(cmin + 1L, cmax + 1L),
                             rows = c(rmin + 1L, rmax + 1L))
      win_h <- rmax - rmin + 1L; win_w <- cmax - cmin + 1L
      os <- .Call(C_warp_strip, win,
                  c(as.integer(win_h), as.integer(win_w)),
                  c(cmin, rmin), sx, sy, method_code, c(out_h, W))
    } else {
      os <- matrix(NA_real_, out_h, W)
    }
    sink$write(ty, r0, r1, os)
  }

  out <- sink$finish()
  if (!is.null(path)) return(out)
  out[[1L]]
}

#' Materialize a spatial query as an sf object
#'
#' Collects a `vectra_node` (typically the result of [spatial_map()] or
#' [spatial_join()]) into memory and rebuilds an `sf` object from its hex-WKB
#' geometry column. The CRS defaults to the one carried on the node.
#'
#' This is the spatial counterpart to [collect()]: use it when the final result
#' fits in memory as `sf`. For a result still larger than RAM, keep it as a node
#' and write it out with [write_vtr()] (the geometry stays as a WKB string
#' column) or reduce it with [collect_chunked()].
#'
#' @param x A `vectra_node` with a hex-WKB / WKT geometry column, or a
#'   data.frame already collected from one.
#' @param geom Name of the geometry column. Default `"geometry"`.
#' @param crs Override the coordinate reference system. Defaults to the CRS the
#'   node carries, or unknown.
#'
#' @return An `sf` object.
#'
#' @seealso [spatial_map()], [spatial_join()], [collect()].
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(data.frame(
#'   NAME = nc$NAME,
#'   geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
#' ), f)
#' result <- tbl(f) |> spatial_map(~ sf::st_centroid(.x), crs = sf::st_crs(nc))
#' collect_sf(result)
#' unlink(f)
#'
#' @export
collect_sf <- function(x, geom = "geometry", crs = NULL) {
  .check_sf()
  node_crs <- if (inherits(x, "vectra_node")) x$.crs else NULL
  df <- if (inherits(x, "vectra_node")) collect(x) else as.data.frame(x)
  if (is.null(crs)) crs <- node_crs
  crs <- .as_crs(crs)
  if (!geom %in% names(df))
    stop(sprintf("geometry column '%s' not found; pass geom=", geom))
  g <- sf::st_as_sfc(structure(df[[geom]], class = "WKB"), EWKB = FALSE)
  g <- sf::st_set_crs(g, crs)
  rest <- df[setdiff(names(df), geom)]
  sf::st_sf(rest, geometry = g)
}

#' Stream a vectra node's geometry to a vector file
#'
#' An [sf::st_write()] method (also reached through [sf::write_sf()]) for a
#' `vectra_node`: writes the result a batch at a time, appending each, so the
#' whole layer is never held in memory. This is the streaming counterpart to
#' `collect_sf(x) |> sf::st_write(...)` -- that route materializes every feature
#' as an `sf` object first, which for a multi-million-feature result dominates
#' memory; this route's peak is one batch.
#'
#' @param obj A `vectra_node` whose rows carry a hex-WKB geometry column (from
#'   [spatial_overlay()], a grouped [slice_min()] / [slice_max()] resolution, a
#'   `.vtr` scan, ...). It is consumed by the stream.
#' @param dsn Destination data source name (file path).
#' @param layer Layer name. `NULL` lets \pkg{sf} derive it from `dsn`.
#' @param ... Unused; for S3 generic compatibility.
#' @param geom Name of the hex-WKB geometry column. Default `"geometry"`.
#' @param crs CRS to tag the output with. `NULL` takes the CRS carried on the
#'   node.
#' @param delete_dsn If `TRUE`, remove an existing `dsn` before writing.
#' @param quiet Passed to [sf::st_write()].
#'
#' @return The `dsn`, invisibly.
#' @seealso [collect_sf()] to materialize the whole result as one `sf` object.
#' @exportS3Method sf::st_write
st_write.vectra_node <- function(obj, dsn, layer = NULL, ..., geom = "geometry",
                                 crs = NULL, delete_dsn = FALSE, quiet = TRUE) {
  .check_sf()
  if (is.null(crs)) crs <- obj$.crs
  crs <- .as_crs(crs)
  if (isTRUE(delete_dsn) && file.exists(dsn)) unlink(dsn)
  nxt   <- .batch_cursor(obj)
  first <- TRUE
  repeat {
    df <- nxt()
    if (is.null(df)) break
    if (!geom %in% names(df))
      stop(sprintf("geometry column '%s' not found; pass geom=", geom))
    g   <- sf::st_as_sfc(structure(df[[geom]], class = "WKB"), EWKB = FALSE)
    g   <- sf::st_set_crs(g, crs)
    sfb <- sf::st_sf(df[setdiff(names(df), geom)], geometry = g)
    sf::st_write(sfb, dsn, layer = layer, append = !first, quiet = quiet)
    first <- FALSE
  }
  if (first) stop("the query produced no rows to write")
  invisible(dsn)
}

#' Self-overlay a polygon layer into disjoint pieces (QGIS-style Union)
#'
#' Splits a polygon layer along all its own overlaps into disjoint pieces and
#' returns a lazy node with one row per piece per covering polygon: where `k`
#' polygons overlap, that piece appears `k` times, each row carrying one source
#' polygon's attributes. This is the union overlay GIS tools expose as
#' "Union (single layer)", with the overlap retained once per contributing
#' feature rather than dissolved. Resolve the duplicates with a grouped
#' [slice_min()] / [slice_max()] -- for example earliest designation year wins:
#' `group_by(piece_id) |> slice_min(year)`.
#'
#' The topology is done once with \pkg{sf}/GEOS and tiled over connected overlap
#' clusters (disjoint clusters never share a piece, so the tiling is exact and
#' bounded in memory), then the exploded pieces are streamed to a `.vtr` and
#' handed back as a lazy node. Geometry rides through the engine as hex-encoded
#' WKB in a string column; the CRS is carried on the node for [collect_sf()].
#'
#' The overlay runs on a fixed-precision model: coordinates are snapped to a
#' grid derived from their own magnitude so the pieces come out disjoint and
#' their areas reconstruct the union of the inputs, instead of drifting by the
#' fraction of a percent that floating-point sliver artefacts on invalid input
#' otherwise introduce. Inputs are also passed through [sf::st_make_valid()].
#'
#' With a second layer `y`, the same machinery overlays two layers instead of
#' self-unioning one: both layers are noded together into one planar partition,
#' and each piece carries the attributes of the `x` record and the `y` record
#' that cover it. `how` selects which pieces to keep -- the intersection (pieces
#' covered by both), the union (every piece of either), `x` split by `y`
#' (`"identity"`), or the parts in exactly one layer (`"symdiff"`). With
#' `y = NULL` (the default) the function self-unions `x` and `how` is ignored.
#'
#' @param x An `sf` object with polygon or multipolygon geometry, or a single
#'   path to a vector file (e.g. a GeoPackage). A path is read in feature batches
#'   via `layer` / `query`, so the whole layer is never held in memory at once --
#'   peak memory then tracks the cleaned geometry, not the source size, which lets
#'   a larger-than-RAM layer overlay on a modest machine.
#' @param y Optional second layer to overlay `x` against, in the same forms `x`
#'   accepts (an `sf` object or a file path read via `layer_y` / `query_y`). It
#'   must share the CRS of `x`. `NULL` (the default) self-unions `x`.
#' @param vars Character vector of attribute columns of `x` to carry onto each
#'   piece. Default `NULL` keeps them all; name a subset to keep the streamed
#'   output narrow.
#' @param vars_y Character vector of attribute columns of `y` to carry onto each
#'   piece (two-layer overlay only). Default `NULL` keeps them all. A name shared
#'   with an `x` column is disambiguated with a `.x` / `.y` suffix in the output.
#' @param how For a two-layer overlay, which pieces to keep: `"intersection"`
#'   (covered by both layers; the default), `"union"` (every piece of either,
#'   the absent side's attributes filled with `NA`), `"identity"` (all of `x`,
#'   split by `y`, with `y`'s attributes where it covers and `NA` elsewhere), or
#'   `"symdiff"` (pieces in exactly one layer). Ignored when `y = NULL`.
#' @param piece Name of the integer piece-id column added to the output (the key
#'   you group by to resolve overlaps). Default `"piece_id"`.
#' @param geom Name of the output hex-WKB geometry column. Default `"geometry"`.
#' @param grid Fixed-precision snapping grid size in CRS units. Coordinates are
#'   snapped to this grid before noding so near-duplicate shared boundaries merge
#'   into one. `NULL` (the default) derives it from coordinate magnitude
#'   (`max(abs(st_bbox(x))) * 3e-8`), which suits projected layers. Pass a number
#'   to override when that default is too coarse for fine geometry (or too coarse
#'   because an outlier coordinate inflated the magnitude), or `0` to disable
#'   snapping entirely.
#' @param precision Fixed-precision grid size, in CRS units, for noding the
#'   boundary linework. Noding on a fixed grid is deterministic and avoids the
#'   floating noder's repair-and-retry on dense overlapping linework, which is
#'   what makes a large dense layer feasible to overlay. It is far finer than
#'   `grid` so intersection points are not collapsed. `NULL` (the default)
#'   derives it from coordinate magnitude (`max(abs(st_bbox(x))) * 1e-13`); pass
#'   a number to override, or `0` to node in floating precision.
#' @param dedup Overlay one representative per group of byte-identical cleaned
#'   geometries and fan the per-record attributes back onto its pieces afterwards.
#'   Duplicates add no faces, so the result is identical; this only removes the
#'   redundant noding when many records are stacked over one site (common in
#'   WDPA-style data). `TRUE` by default; set `FALSE` to overlay every record.
#' @param flush_rows Exploded rows buffered before a spill flush. Defaults to
#'   `getOption("vectra.spatial_flush", 5e5)`.
#' @param mem_limit Approximate peak working-set budget in bytes, bounding the
#'   per-tile size (`tile_bytes = mem_limit / (threads * 24)`). It is a throughput
#'   knob with an interior optimum, not "bigger is faster": too small replicates
#'   features across many tiles, too large nodes too much linework per tile (a
#'   superlinear cost), and a budget of tens of GB runs slower than the default on
#'   a dense layer. Lower it for tighter memory. Defaults via
#'   `getOption("vectra.overlay_mem_limit", ...)` to a value that scales with
#'   `threads` to hold the per-tile size near its measured optimum.
#' @param threads Number of OpenMP threads for the per-component overlay within a
#'   chunk. `0` (the default, via `getOption("vectra.overlay_threads", 0)`) uses
#'   all available cores.
#' @param quiet If `FALSE`, show a text progress bar over the overlay chunks.
#' @param layer When `x` is a file path, the name of the layer to read. Ignored
#'   for an `sf` `x`. Supply this or `query`.
#' @param query When `x` is a file path, a SQL statement selecting the features to
#'   overlay (read in batches via `LIMIT`/`OFFSET`); use it instead of `layer`
#'   for a subset or join. With `query` and no `layer`, pass `grid` explicitly,
#'   since the layer extent cannot be read from the file metadata.
#' @param layer_y,query_y The `layer` / `query` equivalents for a file-path `y`.
#' @param read_chunk Features per read/parse batch. `NULL` (default) sizes it
#'   from available RAM. Smaller batches lower peak memory; larger ones do fewer
#'   round trips.
#'
#' @return A `vectra_node` over the exploded overlay, backed by temporary `.vtr`
#'   spills removed when the node is garbage-collected, carrying the CRS of `x`
#'   for [collect_sf()]. For a self-union it is one row per piece per covering
#'   polygon; for a two-layer overlay one row per piece per covering
#'   `x`-record / `y`-record pair, with the columns of both layers.
#'
#' @seealso [slice_min()] / [slice_max()] to resolve each piece to one winner,
#'   [collect_sf()] to materialize as `sf`.
#'
#' @examplesIf requireNamespace("sf", quietly = TRUE)
#' # Two overlapping squares designated in different years.
#' sq <- function(a, b) sf::st_polygon(list(rbind(
#'   c(a, 0), c(b, 0), c(b, 1), c(a, 1), c(a, 0))))
#' polys <- sf::st_sf(year = c(1990L, 2010L),
#'                    geometry = sf::st_sfc(sq(0, 2), sq(1, 3)))
#'
#' # Split into disjoint pieces; earliest year wins where they overlap.
#' first <- spatial_overlay(polys) |>
#'   group_by(piece_id) |>
#'   slice_min(year, n = 1, with_ties = FALSE) |>
#'   collect_sf()
#' first
#'
#' # Two-layer overlay: intersect the squares with a zone layer, keeping both
#' # sets of attributes on each overlapping piece.
#' zones <- sf::st_sf(zone = c("A", "B"),
#'                    geometry = sf::st_sfc(sq(0, 1.5), sq(1.5, 3)))
#' inter <- spatial_overlay(polys, zones, how = "intersection") |> collect_sf()
#' inter
#'
#' @export
spatial_overlay <- function(x, y = NULL, vars = NULL, vars_y = NULL,
                            how = c("intersection", "union", "identity", "symdiff"),
                            piece = "piece_id",
                            geom = "geometry", grid = NULL, precision = NULL,
                            dedup = TRUE, flush_rows = NULL,
                            mem_limit = NULL, threads = NULL, quiet = TRUE,
                            layer = NULL, query = NULL,
                            layer_y = NULL, query_y = NULL, read_chunk = NULL) {
  .check_sf()
  how <- match.arg(how)
  two <- !is.null(y)

  # Validate an explicit grid / precision up front; NULL means derive them.
  if (!is.null(grid)) {
    if (!is.numeric(grid) || length(grid) != 1L || !is.finite(grid) || grid < 0)
      stop("`grid` must be a single non-negative number (CRS units), or NULL to derive it")
    grid <- as.double(grid)
  }
  if (!is.null(precision)) {
    if (!is.numeric(precision) || length(precision) != 1L || !is.finite(precision) || precision < 0)
      stop("`precision` must be a single non-negative number (CRS units), or NULL to derive it")
    precision <- as.double(precision)
  }

  # Overlay is CPU-bound and the tiles are load-balanced across the pool, so use
  # every core by default; peak memory is bounded by the per-tile budget times the
  # running threads, not by the thread count.
  nthreads <- threads   %||% getOption("vectra.overlay_threads",
                                       max(parallel::detectCores(), 1L))
  nthreads <- max(as.integer(nthreads), 1L)
  mem      <- mem_limit %||% getOption("vectra.overlay_mem_limit",
                                       .overlay_mem_default(nthreads))

  # Read and parse the input: an in-memory sf object, or a file source (a path
  # with layer= or query=) read in feature batches so the whole layer is never
  # materialized at once. Either way the geometry is repaired, made areal, snapped
  # to the grid, and returned as cleaned WKB plus bounding boxes; geometry stays
  # compact (raw WKB) in R and never touches sf again on the compute boundary. For
  # a two-layer overlay both layers are ingested onto the same grid and stacked, so
  # the downstream noding, components, dedup, and tiling treat them as one set; a
  # per-input `side` tag (1 = x, 2 = y) drives the attribute fan-out at the end.
  ingx  <- .overlay_ingest(x, vars, piece, grid, nthreads,
                           layer = layer, query = query, read_chunk = read_chunk)
  crs   <- ingx$crs; grid <- ingx$grid
  if (two) {
    ingy <- .overlay_ingest(y, vars_y, piece, grid, nthreads,
                            layer = layer_y, query = query_y, read_chunk = read_chunk)
    if (!isTRUE(crs == ingy$crs))
      stop("`x` and `y` must share a CRS; transform one first (e.g. sf::st_transform()).")
    nx <- ingx$n; ny <- ingy$n; n <- nx + ny
    cwkb <- c(ingx$cwkb, ingy$cwkb)
    bbox <- rbind(ingx$bbox, ingy$bbox)
    side_of  <- c(rep.int(1L, nx), rep.int(2L, ny))   # which layer each input is
    local_of <- c(seq_len(nx), seq_len(ny))           # its row within that layer
    # Disambiguate shared column names (dplyr-style .x / .y) and pre-rename the
    # per-layer attribute frames so the fan-out can combine them without collision.
    nm_x <- names(ingx$attrs); nm_y <- names(ingy$attrs)
    ax <- ingx$attrs; if (length(ax)) names(ax) <- ifelse(nm_x %in% nm_y, paste0(nm_x, ".x"), nm_x)
    ay <- ingy$attrs; if (length(ay)) names(ay) <- ifelse(nm_y %in% nm_x, paste0(nm_y, ".y"), nm_y)
    out_template <- ax[0, , drop = FALSE]
    for (nm in names(ay)) out_template[[nm]] <- ay[[nm]][0]
    out_template[[piece]] <- integer(0)
    out_template[[geom]]  <- character(0)
    ingy <- NULL
  } else {
    attrs <- ingx$attrs; n <- ingx$n
    cwkb  <- ingx$cwkb;  bbox <- ingx$bbox
  }
  ingx  <- NULL
  if (!any(!is.na(bbox[, 1L]))) stop("the inputs have no parseable geometries to overlay")

  # Noding precision. Boundaries are noded at a fixed grid (snap rounding): this is
  # deterministic and avoids the floating noder's repair-and-retry on dense overlapping
  # linework, which is what makes a large dense layer feasible to node. The grid is
  # far finer than the cleaning grid so intersection points are not collapsed; it is
  # set well above the floating-point resolution at the layer's extent.
  if (is.null(precision)) {
    mag       <- max(abs(bbox), na.rm = TRUE)
    precision <- if (is.finite(mag) && mag > 0) mag * 1e-13 else 0
  }

  comp   <- .Call(C_overlay_components, bbox)
  wbytes <- as.numeric(lengths(cwkb))

  # Deduplicate identical cleaned geometry before overlaying. After snapping, the
  # many records stacked over one site are byte-identical, and duplicates add no
  # faces -- so the overlay runs on one representative per group and the per-record
  # attributes are fanned back onto its pieces afterwards. The arrangement is
  # unchanged; only the redundant noding/clipping work is removed. `mem_by_dk[[k]]`
  # holds the original rows the k-th distinct input stands for.
  n_orig <- n
  if (isTRUE(dedup)) {
    grp       <- .Call(C_overlay_group, cwkb)
    rep_idx   <- which(!duplicated(grp))
    mem_by_dk <- unname(split(seq_len(n_orig), grp))
  } else {
    rep_idx   <- seq_len(n_orig)
    mem_by_dk <- as.list(seq_len(n_orig))
  }
  comp <- comp[rep_idx]; bbox <- bbox[rep_idx, , drop = FALSE]
  cwkb <- cwkb[rep_idx]; wbytes <- wbytes[rep_idx]; n <- length(rep_idx)

  # Connected components (from bounding boxes). Each component is one overlay job;
  # only the few components too large for the memory budget are tiled over their
  # own extent. Most components are small, so there is no clipping or replication
  # and the fast exact path is taken.
  tile_bytes <- max(mem / (nthreads * .OVERLAY_FACTOR), 1e6)
  no_clip <- rep(NA_real_, 4L)
  # Build each component's jobs into its own list, then flatten once. Growing a
  # single `jobs` list per component is quadratic over the many components a large
  # layer splits into; the per-component lists keep it linear.
  per_comp <- lapply(split(seq_len(n), comp), function(rows) {
    if (length(rows) <= .OVERLAY_FEAT_CAP && sum(wbytes[rows]) <= tile_bytes)
      return(list(list(idx = rows, rect = no_clip)))
    ext <- c(min(bbox[rows, 1L]), min(bbox[rows, 2L]),
             max(bbox[rows, 3L]), max(bbox[rows, 4L]))
    .overlay_tiles(bbox, rows, ext, tile_bytes, .OVERLAY_FEAT_CAP, wbytes, 0L)
  })
  jobs <- unlist(per_comp, recursive = FALSE, use.names = FALSE)
  # Process the heaviest tiles first. A tile's cost tracks its feature bytes, and
  # the dense archipelago tiles cost far more than the rest; running them first
  # keeps every thread busy instead of stranding one on a giant tile at the tail.
  if (length(jobs) > 1L) {
    jcost <- vapply(jobs, function(t) sum(wbytes[t$idx]), numeric(1))
    jobs <- jobs[order(jcost, decreasing = TRUE)]
  }
  if (!quiet)
    message(sprintf(paste0("spatial_overlay: %d inputs (%d distinct), %d components, %d jobs, ",
                           "%d threads, grid=%.4g, noding=%.3g"),
                    n_orig, n, max(comp), length(jobs), nthreads, grid, precision))

  fr        <- flush_rows %||% getOption("vectra.spatial_flush", .SPATIAL_FLUSH)
  acc       <- .run_accumulator(fr)
  # Seed the schema from the combined template so an empty two-layer result (e.g.
  # an intersection with no overlaps) still finishes as a correctly typed node.
  if (two) acc$push(.coerce_for_vtr(out_template))
  piece_off <- 0L
  cov_err   <- 0
  worst     <- NULL                                  # top offending inputs by coverage error

  # Turn one batch's arrangement (pieces x covering inputs) into output rows.
  # Self-union: each piece-row is fanned to one output row per original record the
  # covering distinct input stands for, carrying that record's attributes.
  emit_single <- function(geoms, origin, fid, gi, poff) {
    members <- mem_by_dk[gi[origin]]                 # original rows per piece-row
    mult    <- lengths(members)
    rep_row <- rep.int(seq_along(origin), mult)      # piece-row index per fanned row
    pid     <- poff + fid
    df  <- attrs[unlist(members, use.names = FALSE), , drop = FALSE]
    df[[piece]] <- pid[rep_row]                      # rows of one face share a piece id
    df[[geom]]  <- geoms[rep_row]
    rownames(df) <- NULL
    df
  }

  # Two-layer: group the covering inputs of each face by layer, then combine the
  # x-records and y-records covering a face per `how`. A face's covering rows are
  # the (x record) x (y record) pairs; faces touching only one layer fill the
  # other side with NA. All output rows of a face share its piece id, and the
  # face geometry is taken once (every covering row of a face holds the same one).
  emit_two <- function(geoms, origin, fid, gi, poff) {
    members  <- mem_by_dk[gi[origin]]
    rows     <- unlist(members, use.names = FALSE)   # original rows covering each piece-row
    face_rep <- rep.int(fid, lengths(members))       # face id (batch-local, 1-based) per row
    sd       <- side_of[rows]; loc <- local_of[rows]
    nf   <- max(fid)
    fdup <- !duplicated(fid)
    geom_by_face <- character(nf)
    geom_by_face[fid[fdup]] <- geoms[fdup]           # one representative geometry per face

    ix <- data.frame(face = face_rep[sd == 1L], loc = loc[sd == 1L])  # x coverage
    iy <- data.frame(face = face_rep[sd == 2L], loc = loc[sd == 2L])  # y coverage

    block <- function(face_vec, lx, ly) {
      if (!length(face_vec)) return(NULL)
      m  <- length(face_vec)
      d  <- ax[if (is.null(lx)) rep(NA_integer_, m) else lx, , drop = FALSE]
      dy <- ay[if (is.null(ly)) rep(NA_integer_, m) else ly, , drop = FALSE]
      for (nm in names(dy)) d[[nm]] <- dy[[nm]]
      d[[piece]] <- poff + face_vec
      d[[geom]]  <- geom_by_face[face_vec]
      rownames(d) <- NULL
      d
    }

    out <- list()
    if (how != "symdiff" && nrow(ix) && nrow(iy)) {  # pieces covered by both layers
      mm <- merge(ix, iy, by = "face", suffixes = c(".x", ".y"))
      if (nrow(mm)) out <- c(out, list(block(mm$face, mm$loc.x, mm$loc.y)))
    }
    fx <- unique(ix$face); fy <- unique(iy$face)
    if (how %in% c("union", "identity", "symdiff")) {  # pieces with only x coverage
      xof <- setdiff(fx, fy)
      if (length(xof)) { s <- ix[ix$face %in% xof, ]; out <- c(out, list(block(s$face, s$loc, NULL))) }
    }
    if (how %in% c("union", "symdiff")) {              # pieces with only y coverage
      yof <- setdiff(fy, fx)
      if (length(yof)) { s <- iy[iy$face %in% yof, ]; out <- c(out, list(block(s$face, NULL, s$loc))) }
    }
    out <- out[!vapply(out, is.null, logical(1))]
    if (!length(out)) out_template else do.call(rbind, out)
  }

  emit_fn <- if (two) emit_two else emit_single

  # Overlay one batch of tiles, map pieces back to global rows, stream to spill.
  run_batch <- function(batch) {
    job <- rep.int(seq_along(batch), vapply(batch, function(t) length(t$idx), integer(1)))
    gi  <- unlist(lapply(batch, `[[`, "idx"), use.names = FALSE)
    rct <- unlist(lapply(batch, `[[`, "rect"), use.names = FALSE)
    res <- .Call(C_overlay_run, cwkb[gi], as.integer(job), as.double(rct),
                 as.integer(nthreads), as.double(precision))
    geoms <- res[[1L]]; origin <- res[[2L]]; parea <- res[[3L]]
    iarea <- res[[4L]]; fid <- res[[5L]]
    if (length(geoms)) {
      df <- emit_fn(geoms, origin, fid, gi, piece_off)
      acc$push(.coerce_for_vtr(df))
      piece_off <<- piece_off + max(fid)
      det <- .overlay_coverage_detail(origin, parea, iarea)
      if (nrow(det)) cov_err <<- max(cov_err, max(det$err))
      bad <- det[det$err > 1e-4, , drop = FALSE]
      if (nrow(bad)) {
        bad$row <- rep_idx[gi[bad$src]]               # distinct chunk -> original row
        worst <<- rbind(worst, bad[, c("row", "err", "iarea", "cov")])
        worst <<- worst[utils::head(order(worst$err, decreasing = TRUE), 50L), , drop = FALSE]
      }
    }
  }
  # A batch is overlaid with one thread per tile and load-balanced across the pool
  # only within the batch, so each batch must hold many more tiles than threads, or
  # a batch that gathers several large tiles leaves most threads idle on its tail.
  # Batch by a fixed tile count, not input bytes: a feature spanning many tiles is
  # shared in memory but its bytes recur in every tile, so a byte budget collapses
  # to a handful of tiles wherever such features dominate. Peak working set stays
  # bounded by tile_bytes times the running threads, independent of the batch size.
  batch_tiles <- max(256L, 32L * nthreads)
  pb <- if (!quiet) utils::txtProgressBar(0, length(jobs), style = 3) else NULL

  njob <- length(jobs)
  for (start in seq.int(1L, njob, by = batch_tiles)) {
    last <- min(start + batch_tiles - 1L, njob)
    run_batch(jobs[start:last])
    if (!is.null(pb)) utils::setTxtProgressBar(pb, last)
  }
  if (!is.null(pb)) close(pb)
  if (piece_off == 0L) stop("overlay produced no polygonal pieces")
  if (cov_err > 1e-4) {
    rows <- if (!is.null(worst))
      paste(utils::head(worst$row[order(worst$err, decreasing = TRUE)], 5L), collapse = ", ")
    else "unknown"
    warning(sprintf(paste0("spatial_overlay: coverage rel-err %.2e exceeds 1e-4; %d input(s) did not ",
                           "reconstruct from their pieces (worst rows in `x`: %s). Their geometry is ",
                           "finer than, or invalid after, the snapping grid (%.4g); pass grid= to adjust."),
                    cov_err, if (is.null(worst)) 0L else nrow(worst), rows, grid),
            call. = FALSE)
  }
  attr_node <- acc$finish(crs = crs, empty_geom = geom)
  if (!is.null(worst))
    attr(attr_node, "coverage_offenders") <-
      worst[order(worst$err, decreasing = TRUE), , drop = FALSE]
  attr_node
}

# Native libgeos compute paths for spatial_filter / spatial_clip /
# spatial_dissolve, checked head-to-head against the sf path. With crs = NA the
# geometry math is planar on both sides, so the native C path (taken when the
# predicate is recognised and the CRS is planar) must equal the sf path. The sf
# path is forced by wrapping the predicate in a closure that is not identical()
# to the exported sf function, which the native dispatcher therefore skips.

skip_if_not_installed("sf")

sq <- function(xmin, xmax, ymin, ymax)
  sf::st_polygon(list(rbind(c(xmin, ymin), c(xmax, ymin), c(xmax, ymax),
                            c(xmin, ymax), c(xmin, ymin))))

# A stream of polygons with varied relations to a resident locator layer.
stream_polys <- sf::st_sfc(
  sq(0, 1, 0, 1),        # inside A
  sq(0, 2, 0, 2),        # equals A
  sq(0.5, 2.5, 0, 1),    # straddles A's edge
  sq(1.8, 3.2, 0, 1),    # touches/near B
  sq(3, 5, 0, 2),        # equals B
  sq(3.2, 3.8, 0.2, 0.8),# inside B
  sq(6, 7, 6, 7))        # disjoint from both
resident <- sf::st_sf(rid = c("A", "B"),
                      geometry = sf::st_sfc(sq(0, 2, 0, 2), sq(3, 5, 0, 2)))

write_stream <- function() {
  df <- data.frame(pid = seq_along(stream_polys),
                   geometry = sf::st_as_binary(stream_polys, hex = TRUE))
  f <- tempfile(fileext = ".vtr")
  write_vtr(df, f)
  f
}

test_that("native filter equals sf filter for every recognised predicate", {
  f <- write_stream(); on.exit(unlink(f))
  preds <- list(
    intersects  = sf::st_intersects,
    within      = sf::st_within,
    contains    = sf::st_contains,
    overlaps    = sf::st_overlaps,
    covers      = sf::st_covers,
    covered_by  = sf::st_covered_by,
    touches     = sf::st_touches,
    crosses     = sf::st_crosses)
  for (nm in names(preds)) {
    p <- preds[[nm]]
    wrap <- function(a, b) p(a, b)        # not identical() to p -> sf path
    native <- collect(spatial_filter(tbl(f), resident, predicate = p, crs = NA))
    viasf  <- collect(spatial_filter(tbl(f), resident, predicate = wrap, crs = NA))
    expect_equal(sort(native$pid), sort(viasf$pid), info = nm)
    nativeN <- collect(spatial_filter(tbl(f), resident, predicate = p,
                                      negate = TRUE, crs = NA))
    viasfN  <- collect(spatial_filter(tbl(f), resident, predicate = wrap,
                                      negate = TRUE, crs = NA))
    expect_equal(sort(nativeN$pid), sort(viasfN$pid), info = paste0("!", nm))
  }
})

test_that("native join equals sf join, with multi-match attribute duplication", {
  f <- write_stream(); on.exit(unlink(f))
  # Overlapping resident features so a straddling left polygon matches both,
  # exercising the row replication the join performs.
  resj <- sf::st_sf(rid = c("C", "D"),
                    geometry = sf::st_sfc(sq(0, 3, 0, 3), sq(1, 4, 0, 2)))
  key <- function(d)
    sort(paste(d$pid, ifelse(is.na(d$rid), "NA", d$rid), sep = "|"))
  preds <- list(intersects = sf::st_intersects, within = sf::st_within,
                contains = sf::st_contains, covered_by = sf::st_covered_by)
  for (nm in names(preds)) {
    p <- preds[[nm]]
    wrap <- function(a, b) p(a, b)          # not identical() to p -> sf path
    for (lf in c(TRUE, FALSE)) {
      nat <- collect(spatial_join(tbl(f), resj, join = p, crs = NA, left = lf))
      ref <- collect(spatial_join(tbl(f), resj, join = wrap, crs = NA, left = lf))
      expect_equal(key(nat), key(ref), info = paste0(nm, " left=", lf))
      expect_equal(names(nat)[order(names(nat))],
                   names(ref)[order(names(ref))], info = paste0(nm, " cols"))
    }
  }
})

test_that("native clip and erase equal sf, geometry-for-geometry", {
  f <- write_stream(); on.exit(unlink(f))
  mask <- sf::st_sfc(sq(1, 4, 0.5, 1.5))

  clip_nat <- collect_sf(spatial_clip(tbl(f), mask, crs = NA))
  ref <- sf::st_sf(pid = seq_along(stream_polys), geometry = stream_polys)
  clip_ref <- suppressWarnings(sf::st_intersection(ref, sf::st_union(mask)))
  clip_nat <- clip_nat[order(clip_nat$pid), ]
  clip_ref <- clip_ref[order(clip_ref$pid), ]
  expect_equal(clip_nat$pid, clip_ref$pid)
  expect_equal(as.numeric(sf::st_area(clip_nat)),
               as.numeric(sf::st_area(clip_ref)), tolerance = 1e-9)

  erase_nat <- collect_sf(spatial_clip(tbl(f), mask, erase = TRUE, crs = NA))
  erase_ref <- suppressWarnings(sf::st_difference(ref, sf::st_union(mask)))
  erase_nat <- erase_nat[order(erase_nat$pid), ]
  erase_ref <- erase_ref[order(erase_ref$pid), ]
  expect_equal(erase_nat$pid, erase_ref$pid)
  expect_equal(sum(as.numeric(sf::st_area(erase_nat))),
               sum(as.numeric(sf::st_area(erase_ref))), tolerance = 1e-9)
})

test_that("native dissolve equals sf union per group", {
  band <- c(1, 1, 1, 2, 2, 2, 2)
  df <- data.frame(band = band, w = seq_along(band),
                   geometry = sf::st_as_binary(stream_polys, hex = TRUE))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f)

  got <- collect_sf(spatial_dissolve(tbl(f), by = "band", crs = NA,
                                     .fun = list(wsum = function(d) sum(d$w))))
  got <- got[order(got$band), ]
  ref_area <- vapply(sort(unique(band)), function(b)
    as.numeric(sf::st_area(sf::st_union(stream_polys[band == b]))), numeric(1))
  expect_equal(nrow(got), length(ref_area))
  expect_equal(as.numeric(sf::st_area(got)), ref_area, tolerance = 1e-9)
  expect_equal(got$wsum[order(got$band)],
               c(sum((seq_along(band))[band == 1]),
                 sum((seq_along(band))[band == 2])))
})

test_that("native path is gated to planar; spherical geographic falls back", {
  # planar_ok: NA crs and projected -> native; geographic + s2 on -> sf.
  expect_true(vectra:::.geos_planar_ok(sf::st_crs(NA)))
  expect_true(vectra:::.geos_planar_ok(sf::st_crs(32119)))     # projected
  old <- sf::sf_use_s2(TRUE); on.exit(sf::sf_use_s2(old))
  expect_false(vectra:::.geos_planar_ok(sf::st_crs(4326)))     # lon/lat + s2
  sf::sf_use_s2(FALSE)
  expect_true(vectra:::.geos_planar_ok(sf::st_crs(4326)))      # lon/lat, s2 off
})

test_that("predicate codes map the native set; others are NA (sf fallback)", {
  expect_equal(vectra:::.geos_pred_code(NULL), 0L)
  expect_equal(vectra:::.geos_pred_code(sf::st_within), 1L)
  expect_equal(vectra:::.geos_pred_code(sf::st_equals), 8L)
  expect_equal(vectra:::.geos_pred_code(sf::st_disjoint), 9L)
  expect_equal(vectra:::.geos_pred_code(sf::st_is_within_distance), 10L)
  expect_true(is.na(vectra:::.geos_pred_code(sf::st_nearest_feature)))
  expect_true(is.na(vectra:::.geos_pred_code(function(a, b) NULL)))
})

test_that("native filter equals / disjoint / within_distance match sf", {
  f <- write_stream(); on.exit(unlink(f))
  for (nm in c("equals", "disjoint")) {
    p <- getExportedValue("sf", paste0("st_", nm))
    wrap <- function(a, b) p(a, b)            # not identical() to p -> sf path
    nat <- collect(spatial_filter(tbl(f), resident, predicate = p, crs = NA))
    ref <- collect(spatial_filter(tbl(f), resident, predicate = wrap, crs = NA))
    expect_equal(sort(nat$pid), sort(ref$pid), info = nm)
    natN <- collect(spatial_filter(tbl(f), resident, predicate = p,
                                   negate = TRUE, crs = NA))
    refN <- collect(spatial_filter(tbl(f), resident, predicate = wrap,
                                   negate = TRUE, crs = NA))
    expect_equal(sort(natN$pid), sort(refN$pid), info = paste0("!", nm))
  }
  pw <- sf::st_is_within_distance
  wrapw <- function(a, b, ...) pw(a, b, ...)
  for (d in c(0.1, 0.5, 1.5)) {
    nat <- collect(spatial_filter(tbl(f), resident, predicate = pw,
                                  dist = d, crs = NA))
    ref <- collect(spatial_filter(tbl(f), resident, predicate = wrapw,
                                  dist = d, crs = NA))
    expect_equal(sort(nat$pid), sort(ref$pid), info = paste0("within ", d))
  }
})

test_that("native join equals / within_distance / nearest match sf", {
  f <- write_stream(); on.exit(unlink(f))
  resj <- sf::st_sf(rid = c("C", "D"),
                    geometry = sf::st_sfc(sq(0, 3, 0, 3), sq(1, 4, 0, 2)))
  key <- function(d)
    sort(paste(d$pid, ifelse(is.na(d$rid), "NA", d$rid), sep = "|"))
  for (lf in c(TRUE, FALSE)) {
    pe <- sf::st_equals; wrape <- function(a, b) pe(a, b)
    nat <- collect(spatial_join(tbl(f), resj, join = pe, crs = NA, left = lf))
    ref <- collect(spatial_join(tbl(f), resj, join = wrape, crs = NA, left = lf))
    expect_equal(key(nat), key(ref), info = paste0("equals left=", lf))

    pw <- sf::st_is_within_distance; wrapw <- function(a, b, ...) pw(a, b, ...)
    nat <- collect(spatial_join(tbl(f), resj, join = pw, dist = 0.5,
                                crs = NA, left = lf))
    ref <- collect(spatial_join(tbl(f), resj, join = wrapw, dist = 0.5,
                                crs = NA, left = lf))
    expect_equal(key(nat), key(ref), info = paste0("within left=", lf))
  }
  pn <- sf::st_nearest_feature; wrapn <- function(a, b) pn(a, b)
  nat <- collect(spatial_join(tbl(f), resj, join = pn, crs = NA))
  ref <- collect(spatial_join(tbl(f), resj, join = wrapn, crs = NA))
  expect_equal(key(nat), key(ref))
})

# Raw-coordinate point input: spatial_filter / spatial_join build points in C
# (C_geos_locate_xy) instead of decoding the batch to sf, and the join emits the
# point geometry from C_geos_points_to_hex. Same crs = NA planar gate, same sf
# path forced by a non-identical predicate wrapper.
pt_stream <- data.frame(
  pid = 1:7,
  x = c(0.5, 1.0, 2.5, 4.0, 6.5, 1.5, 3.5),
  y = c(0.5, 1.0, 0.5, 1.0, 6.5, 0.5, 1.0))

write_points <- function() {
  f <- tempfile(fileext = ".vtr"); write_vtr(pt_stream, f); f
}

test_that("native coords filter matches sf for points", {
  f <- write_points(); on.exit(unlink(f))
  for (nm in c("intersects", "within", "covered_by")) {
    p <- getExportedValue("sf", paste0("st_", nm))
    wrap <- function(a, b) p(a, b)
    nat <- collect(spatial_filter(tbl(f), resident, predicate = p,
                                  coords = c("x", "y"), crs = NA))
    ref <- collect(spatial_filter(tbl(f), resident, predicate = wrap,
                                  coords = c("x", "y"), crs = NA))
    expect_equal(sort(nat$pid), sort(ref$pid), info = nm)
  }
  pw <- sf::st_is_within_distance; wrapw <- function(a, b, ...) pw(a, b, ...)
  for (d in c(0.3, 0.6, 2.0)) {
    nat <- collect(spatial_filter(tbl(f), resident, predicate = pw,
                                  dist = d, coords = c("x", "y"), crs = NA))
    ref <- collect(spatial_filter(tbl(f), resident, predicate = wrapw,
                                  dist = d, coords = c("x", "y"), crs = NA))
    expect_equal(sort(nat$pid), sort(ref$pid), info = paste0("within ", d))
  }
})

test_that("native coords join matches sf, point geometry included", {
  f <- write_points(); on.exit(unlink(f))
  resj <- sf::st_sf(rid = c("C", "D"),
                    geometry = sf::st_sfc(sq(0, 3, 0, 3), sq(1, 4, 0, 2)))
  key <- function(d)
    sort(paste(d$pid, ifelse(is.na(d$rid), "NA", d$rid), sep = "|"))
  for (lf in c(TRUE, FALSE)) {
    pj <- sf::st_intersects; wrapj <- function(a, b) pj(a, b)
    nat <- collect(spatial_join(tbl(f), resj, join = pj, coords = c("x", "y"),
                                crs = NA, left = lf))
    ref <- collect(spatial_join(tbl(f), resj, join = wrapj, coords = c("x", "y"),
                                crs = NA, left = lf))
    expect_equal(key(nat), key(ref), info = paste0("intersects left=", lf))
    expect_equal(names(nat)[order(names(nat))],
                 names(ref)[order(names(ref))], info = paste0("cols left=", lf))
  }
  # within-distance and nearest
  pw <- sf::st_is_within_distance; wrapw <- function(a, b, ...) pw(a, b, ...)
  nat <- collect(spatial_join(tbl(f), resj, join = pw, dist = 0.6,
                              coords = c("x", "y"), crs = NA))
  ref <- collect(spatial_join(tbl(f), resj, join = wrapw, dist = 0.6,
                              coords = c("x", "y"), crs = NA))
  expect_equal(key(nat), key(ref), info = "within")
  pn <- sf::st_nearest_feature; wrapn <- function(a, b) pn(a, b)
  nat <- collect(spatial_join(tbl(f), resj, join = pn, coords = c("x", "y"),
                              crs = NA))
  ref <- collect(spatial_join(tbl(f), resj, join = wrapn, coords = c("x", "y"),
                              crs = NA))
  expect_equal(key(nat), key(ref), info = "nearest")

  # the emitted point geometry equals the sf-encoded points, row for row
  o   <- order(paste(nat$pid, nat$rid))
  oR  <- order(paste(ref$pid, ref$rid))
  gn  <- sf::st_as_sfc(structure(nat$geometry[o],  class = "WKB"), EWKB = FALSE)
  gr  <- sf::st_as_sfc(structure(ref$geometry[oR], class = "WKB"), EWKB = FALSE)
  expect_equal(unname(sf::st_coordinates(gn)), unname(sf::st_coordinates(gr)))
})

# spatial_explode(): multipart -> single-part, one row per component, attributes
# replicated. Geometry round-trips through a .vtr; CRS is carried on the node.

skip_if_not_installed("sf")

vtr_from <- function(sfobj) {
  f <- tempfile(fileext = ".vtr")
  df <- as.data.frame(sf::st_drop_geometry(sfobj))
  df$geometry <- sf::st_as_binary(sf::st_geometry(sfobj), hex = TRUE)
  write_vtr(df, f)
  f
}

mp2 <- function() sf::st_multipolygon(list(
  list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))),
  list(rbind(c(2, 2), c(3, 2), c(3, 3), c(2, 3), c(2, 2)))))

test_that("a multipolygon explodes to one row per polygon, attributes copied", {
  x <- sf::st_sf(id = 7L, k = "a", geometry = sf::st_sfc(mp2()))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_explode() |> collect_sf()
  expect_equal(nrow(d), 2L)
  expect_equal(d$id, c(7L, 7L))
  expect_equal(d$k, c("a", "a"))
  expect_true(all(sf::st_geometry_type(d) == "POLYGON"))
  expect_equal(sort(as.numeric(sf::st_area(d))), c(1, 1), tolerance = 1e-9)
})

test_that("part= numbers the parts within each source feature, 1-based", {
  x <- sf::st_sf(id = c(1L, 2L), geometry = sf::st_sfc(mp2(), mp2()))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_explode(part = "pid") |> collect()
  expect_equal(nrow(d), 4L)
  expect_equal(d$pid, c(1L, 2L, 1L, 2L))
})

test_that("single-part geometries pass through unchanged", {
  poly <- sf::st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(poly))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_explode() |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_equal(as.numeric(sf::st_area(d)), 1, tolerance = 1e-9)
})

test_that("multilinestring and multipoint explode by part", {
  mls <- sf::st_multilinestring(list(rbind(c(0, 0), c(1, 1)),
                                     rbind(c(2, 2), c(3, 3))))
  mpt <- sf::st_multipoint(rbind(c(0, 0), c(1, 1), c(2, 2)))
  xl <- sf::st_sf(id = 1L, geometry = sf::st_sfc(mls))
  xp <- sf::st_sf(id = 1L, geometry = sf::st_sfc(mpt))
  fl <- vtr_from(xl); fp <- vtr_from(xp); on.exit(unlink(c(fl, fp)))
  dl <- tbl(fl) |> spatial_explode() |> collect_sf()
  dp <- tbl(fp) |> spatial_explode() |> collect_sf()
  expect_equal(nrow(dl), 2L)
  expect_true(all(sf::st_geometry_type(dl) == "LINESTRING"))
  expect_equal(nrow(dp), 3L)
  expect_true(all(sf::st_geometry_type(dp) == "POINT"))
})

test_that("a geometry collection explodes recursively into its members", {
  gc <- sf::st_geometrycollection(list(
    sf::st_point(c(0, 0)),
    sf::st_multipoint(rbind(c(1, 1), c(2, 2)))))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(gc))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_explode() |> collect_sf()
  expect_equal(nrow(d), 3L)
  expect_true(all(sf::st_geometry_type(d) == "POINT"))
})

test_that("a CRS passed to the verb is carried onto the exploded node", {
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(mp2()))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_explode(crs = 3857) |> collect_sf()
  expect_equal(sf::st_crs(d), sf::st_crs(3857))
})

test_that("explode then total area equals the source area", {
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(mp2()))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_explode() |> collect_sf()
  expect_equal(sum(as.numeric(sf::st_area(d))),
               as.numeric(sf::st_area(sf::st_sfc(mp2()))), tolerance = 1e-9)
})

test_that("a clashing part column name is rejected", {
  x <- sf::st_sf(pid = 1L, geometry = sf::st_sfc(mp2()))
  f <- vtr_from(x); on.exit(unlink(f))
  expect_error(collect(spatial_explode(tbl(f), part = "pid")), "already exists")
})

test_that("a non-vectra_node input is rejected", {
  expect_error(spatial_explode(data.frame(a = 1)), "vectra_node")
})

# spatial_filter (select by location) and spatial_clip (clip / erase) — streamed
# vector ops checked against sf in-memory ground truth, with planar CRS = NA so
# intersections and areas are exact Cartesian and the streamed path must equal
# the resident path.

skip_if_not_installed("sf")

make_square <- function(xmin, xmax, ymin, ymax) {
  sf::st_polygon(list(rbind(
    c(xmin, ymin), c(xmax, ymin), c(xmax, ymax),
    c(xmin, ymax), c(xmin, ymin))))
}

# ---- spatial_filter ---------------------------------------------------------

test_that("spatial_filter keeps exactly the points inside a resident region", {
  region <- sf::st_sfc(make_square(0, 1, 0, 1))
  pts <- data.frame(
    id = 1:5,
    x  = c(0.5, 0.9, 1.5, 2.0, 0.1),
    y  = c(0.5, 0.1, 0.5, 0.5, 0.9))   # ids 1, 2, 5 inside; 3, 4 outside
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(pts, f)

  res <- tbl(f) |>
    spatial_filter(region, coords = c("x", "y"), crs = NA) |>
    collect()
  expect_equal(sort(res$id), c(1L, 2L, 5L))
  # Schema preserved: no geometry column introduced for coords input.
  expect_equal(sort(names(res)), c("id", "x", "y"))
})

test_that("spatial_filter negate keeps the complement", {
  region <- sf::st_sfc(make_square(0, 1, 0, 1))
  pts <- data.frame(id = 1:5,
                    x = c(0.5, 0.9, 1.5, 2.0, 0.1),
                    y = c(0.5, 0.1, 0.5, 0.5, 0.9))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(pts, f)

  res <- tbl(f) |>
    spatial_filter(region, coords = c("x", "y"), crs = NA, negate = TRUE) |>
    collect()
  expect_equal(sort(res$id), c(3L, 4L))
})

test_that("spatial_filter matches sf across multi-batch streaming", {
  set.seed(11)
  region <- sf::st_sfc(make_square(0, 1, 0, 1))
  n <- 4000L
  pts <- data.frame(id = seq_len(n),
                    x = runif(n, -0.5, 2),
                    y = runif(n, -0.5, 2))
  sb_all <- sf::st_as_sf(pts, coords = c("x", "y"), crs = NA, remove = FALSE)
  want <- pts$id[lengths(sf::st_intersects(sb_all, region)) > 0L]

  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(pts, f, batch_size = 500L)            # several read batches
  old <- options(vectra.spatial_flush = 700)      # several spill flushes
  on.exit(options(old), add = TRUE)

  got <- tbl(f) |>
    spatial_filter(region, coords = c("x", "y"), crs = NA) |>
    collect()
  expect_equal(sort(got$id), sort(want))
})

test_that("spatial_filter works on a hex-WKB column and preserves schema", {
  polys <- sf::st_sfc(make_square(0, 1, 0, 1),
                      make_square(2, 3, 0, 1),
                      make_square(0.5, 1.5, 0, 1))
  df <- data.frame(pid = 1:3,
                   geometry = sf::st_as_binary(polys, hex = TRUE))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f)

  probe <- sf::st_sfc(make_square(0.9, 1.1, 0, 1))   # overlaps pid 1 and 3
  res <- tbl(f) |> spatial_filter(probe, crs = NA) |> collect()
  expect_equal(sort(res$pid), c(1L, 3L))
  expect_equal(sort(names(res)), c("geometry", "pid"))
})

# ---- spatial_clip -----------------------------------------------------------

test_that("spatial_clip clips streamed polygons to a resident mask", {
  polys <- sf::st_sf(
    pid = c("A", "B"),
    geometry = sf::st_sfc(make_square(0, 2, 0, 1), make_square(2, 4, 0, 1)))
  df <- data.frame(
    pid = polys$pid,
    geometry = sf::st_as_binary(sf::st_geometry(polys), hex = TRUE))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f)

  mask <- sf::st_sfc(make_square(1, 3, 0, 1))
  out <- tbl(f) |> spatial_clip(mask, crs = NA) |> collect_sf()
  out <- out[order(out$pid), ]
  # A clip = (1,2)x(0,1); B clip = (2,3)x(0,1): each area 1.
  expect_equal(sort(out$pid), c("A", "B"))
  expect_equal(as.numeric(sf::st_area(out)), c(1, 1), tolerance = 1e-9)
})

test_that("spatial_clip erase keeps the part outside the mask", {
  polys <- sf::st_sf(
    pid = c("A", "B"),
    geometry = sf::st_sfc(make_square(0, 2, 0, 1), make_square(2, 4, 0, 1)))
  df <- data.frame(
    pid = polys$pid,
    geometry = sf::st_as_binary(sf::st_geometry(polys), hex = TRUE))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f)

  mask <- sf::st_sfc(make_square(1, 3, 0, 1))
  out <- tbl(f) |> spatial_clip(mask, erase = TRUE, crs = NA) |> collect_sf()
  out <- out[order(out$pid), ]
  # A erase = (0,1)x(0,1); B erase = (3,4)x(0,1): each area 1.
  expect_equal(as.numeric(sf::st_area(out)), c(1, 1), tolerance = 1e-9)
})

test_that("spatial_clip matches sf across multi-batch streaming", {
  k <- 60L
  xs <- seq_len(k)
  geoms <- do.call(sf::st_sfc,
                   lapply(xs, function(i) make_square(i, i + 1, 0, 1)))
  df <- data.frame(pid = xs, geometry = sf::st_as_binary(geoms, hex = TRUE))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f, batch_size = 7L)
  old <- options(vectra.spatial_flush = 11)
  on.exit(options(old), add = TRUE)

  mask <- sf::st_sfc(make_square(10.5, 40.5, 0, 1))
  got <- tbl(f) |> spatial_clip(mask, crs = NA) |> collect_sf()

  resident <- sf::st_sf(pid = xs, geometry = geoms)
  ref <- suppressWarnings(sf::st_intersection(resident, sf::st_union(mask)))
  expect_equal(sort(got$pid), sort(ref$pid))
  expect_equal(sum(as.numeric(sf::st_area(got))),
               sum(as.numeric(sf::st_area(ref))), tolerance = 1e-9)
})

# ---- validation -------------------------------------------------------------

test_that("spatial_filter and spatial_clip validate inputs", {
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(data.frame(x = 1, y = 1), f)
  region <- sf::st_sfc(make_square(0, 1, 0, 1))
  expect_error(spatial_filter(42, region), "must be a vectra_node")
  expect_error(spatial_filter(tbl(f), data.frame(a = 1)), "must be an sf or sfc")
  expect_error(spatial_clip(42, region), "must be a vectra_node")
  expect_error(spatial_clip(tbl(f), data.frame(a = 1)), "must be an sf or sfc")
})

# spatial_split(): cut a streamed layer by a resident blade into pieces, or
# return the crossing points.

skip_if_not_installed("sf")

vtr_from <- function(sfobj) {
  f <- tempfile(fileext = ".vtr")
  df <- as.data.frame(sf::st_drop_geometry(sfobj))
  df$geometry <- sf::st_as_binary(sf::st_geometry(sfobj), hex = TRUE)
  write_vtr(df, f)
  f
}

sq    <- function() sf::st_polygon(list(
  rbind(c(0, 0), c(4, 0), c(4, 4), c(0, 4), c(0, 0))))
vblade <- function() sf::st_sfc(sf::st_linestring(rbind(c(2, -1), c(2, 5))))

test_that("a polygon is split into the faces the blade carves out", {
  x <- sf::st_sf(id = 1L, k = "a", geometry = sf::st_sfc(sq()))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_split(vblade()) |> collect_sf()
  expect_equal(nrow(d), 2L)
  expect_equal(sort(as.numeric(sf::st_area(d))), c(8, 8), tolerance = 1e-9)
  expect_equal(d$id, c(1L, 1L))           # attributes copied onto each piece
  expect_equal(d$k, c("a", "a"))
  expect_equal(sum(as.numeric(sf::st_area(d))), 16, tolerance = 1e-9)
})

test_that("a line is split at the blade crossing", {
  line <- sf::st_linestring(rbind(c(0, 2), c(4, 2)))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(line))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_split(vblade()) |> collect_sf()
  expect_equal(nrow(d), 2L)
  expect_true(all(sf::st_geometry_type(d) == "LINESTRING"))
  expect_equal(sort(as.numeric(sf::st_length(d))), c(2, 2), tolerance = 1e-9)
})

test_that("a feature the blade misses passes through as one piece", {
  far <- sf::st_polygon(list(
    rbind(c(10, 10), c(12, 10), c(12, 12), c(10, 12), c(10, 10))))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(far))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_split(vblade()) |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_equal(as.numeric(sf::st_area(d)), 4, tolerance = 1e-9)
})

test_that("extract = 'points' returns the crossing points and drops misses", {
  line_hit  <- sf::st_linestring(rbind(c(0, 2), c(4, 2)))      # crosses x = 2
  line_miss <- sf::st_linestring(rbind(c(0, 9), c(1, 9)))      # never reaches blade
  x <- sf::st_sf(id = c(1L, 2L),
                 geometry = sf::st_sfc(line_hit, line_miss))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_split(vblade(), extract = "points") |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_equal(d$id, 1L)
  xy <- sf::st_coordinates(d)
  expect_equal(unname(xy[1, c("X", "Y")]), c(2, 2), tolerance = 1e-9)
})

test_that("multiple blades split a polygon into a grid of pieces", {
  blades <- sf::st_sfc(
    sf::st_linestring(rbind(c(2, -1), c(2, 5))),
    sf::st_linestring(rbind(c(-1, 2), c(5, 2))))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(sq()))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_split(blades) |> collect_sf()
  expect_equal(nrow(d), 4L)
  expect_equal(sort(as.numeric(sf::st_area(d))), rep(4, 4), tolerance = 1e-9)
})

test_that("a CRS passed to the verb is carried onto the pieces", {
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(sq()))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_split(vblade(), crs = 3857) |> collect_sf()
  expect_equal(sf::st_crs(d), sf::st_crs(3857))
})

test_that("invalid inputs are rejected", {
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(sq()))
  f <- vtr_from(x); on.exit(unlink(f))
  expect_error(spatial_split(data.frame(a = 1), vblade()), "vectra_node")
  expect_error(spatial_split(tbl(f), data.frame(a = 1)), "sf")
  expect_error(collect(spatial_split(tbl(f), vblade(), extract = "nope")),
               "should be one of")
})

# Streamed spatial operations (spatial_map / spatial_join / collect_sf).
# Planar CRS = NA throughout so st_intersects / areas are exact Cartesian and
# the recovered values can be checked against hand-computed truth.

skip_if_not_installed("sf")

make_square <- function(xmin, xmax, ymin, ymax) {
  sf::st_polygon(list(rbind(
    c(xmin, ymin), c(xmax, ymin), c(xmax, ymax),
    c(xmin, ymax), c(xmin, ymin))))
}

# Two disjoint unit squares with a gap between them at x in (1, 2).
two_squares <- function() {
  sf::st_sf(
    pid = c("A", "B"),
    geometry = sf::st_sfc(make_square(0, 1, 0, 1),
                          make_square(2, 3, 0, 1)))
}

test_that("spatial_join recovers known point-in-polygon membership", {
  polys <- two_squares()

  # Truth: 1->A, 2->B, 3->none (gap), 4->A.
  pts <- data.frame(
    id = 1:4,
    x  = c(0.5, 2.5, 1.5, 0.1),
    y  = c(0.5, 0.5, 0.5, 0.9))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(pts, f)

  res <- tbl(f) |>
    spatial_join(polys, join = sf::st_intersects,
                 coords = c("x", "y"), crs = NA) |>
    collect()
  res <- res[order(res$id), ]

  expect_equal(nrow(res), 4L)
  expect_equal(res$pid, c("A", "B", NA, "A"))
  # Left join keeps the unmatched gap point and the coordinate columns.
  expect_equal(res$x, c(0.5, 2.5, 1.5, 0.1))
  expect_true("geometry" %in% names(res))
})

test_that("spatial_join inner (left = FALSE) drops unmatched rows", {
  polys <- two_squares()
  pts <- data.frame(id = 1:3, x = c(0.5, 1.5, 2.5), y = c(0.5, 0.5, 0.5))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(pts, f)

  res <- tbl(f) |>
    spatial_join(polys, coords = c("x", "y"), crs = NA, left = FALSE) |>
    collect()
  res <- res[order(res$id), ]

  expect_equal(res$id, c(1L, 3L))
  expect_equal(res$pid, c("A", "B"))
})

test_that("spatial_join survives multi-batch streaming and many spill flushes", {
  polys <- two_squares()

  set.seed(42)
  n <- 5000L
  # Half the points land in square A, half in square B, none in the gap.
  inA <- seq_len(n) %% 2L == 0L
  x <- ifelse(inA, runif(n, 0, 1), runif(n, 2, 3))
  y <- runif(n, 0, 1)
  pts <- data.frame(id = seq_len(n), x = x, y = y, want = ifelse(inA, "A", "B"))

  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(pts, f, batch_size = 1000L)   # 5 row groups -> 5 read batches

  old <- options(vectra.spatial_flush = 1500)  # force multiple run-file flushes
  on.exit(options(old), add = TRUE)

  res <- tbl(f) |>
    spatial_join(polys["pid"], coords = c("x", "y"), crs = NA) |>
    collect()

  expect_equal(nrow(res), n)
  res <- res[order(res$id), ]
  expect_equal(res$pid, res$want)
})

test_that("spatial_map buffers a point to a disk of the right area", {
  # A single point at the origin, geometry carried as hex WKB.
  pt <- sf::st_sfc(sf::st_point(c(0, 0)))
  df <- data.frame(id = 1L, geometry = sf::st_as_binary(pt, hex = TRUE))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f)

  out <- tbl(f) |> spatial_map(~ sf::st_buffer(.x, 1), crs = NA)
  sf_out <- collect_sf(out)

  expect_s3_class(sf_out, "sf")
  expect_equal(nrow(sf_out), 1L)
  # st_buffer's polygonal disk under-approximates pi; just check it is close.
  expect_equal(as.numeric(sf::st_area(sf_out)), pi, tolerance = 0.01)
})

test_that("spatial_map round-trips geometry unchanged under the identity map", {
  geoms <- sf::st_sfc(
    sf::st_point(c(1, 2)),
    sf::st_point(c(3, 4)),
    sf::st_point(c(5, 6)))
  df <- data.frame(id = 1:3, geometry = sf::st_as_binary(geoms, hex = TRUE))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f)

  out <- tbl(f) |> spatial_map(~ .x, crs = NA)
  sf_out <- collect_sf(out)
  sf_out <- sf_out[order(sf_out$id), ]

  coords <- sf::st_coordinates(sf_out)
  expect_equal(coords[, "X"], c(1, 3, 5))
  expect_equal(coords[, "Y"], c(2, 4, 6))
})

test_that("collect_sf rebuilds the carried CRS", {
  geoms <- sf::st_sfc(sf::st_point(c(0, 0)), crs = 4326)
  df <- data.frame(id = 1L, geometry = sf::st_as_binary(geoms, hex = TRUE))
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f)

  out <- tbl(f) |> spatial_map(~ .x, crs = 4326)
  sf_out <- collect_sf(out)
  expect_equal(sf::st_crs(sf_out), sf::st_crs(4326))
})

test_that("spatial_join composes with offload() partitioning (both-sides-huge)", {
  polys <- two_squares()

  set.seed(7)
  n <- 2000L
  inA <- seq_len(n) %% 2L == 0L
  pts <- data.frame(
    id  = seq_len(n),
    x   = ifelse(inA, runif(n, 0, 1), runif(n, 2, 3)),
    y   = runif(n, 0, 1),
    cell = ifelse(inA, "west", "east"))   # a spatial grid key
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(pts, f)

  # Single-pass reference.
  ref <- tbl(f) |>
    spatial_join(polys["pid"], coords = c("x", "y"), crs = NA) |>
    collect()
  ref <- ref[order(ref$id), ]

  # Partition the stream by the grid key, join each shard independently,
  # then recombine. The union must reproduce the single-pass result.
  part <- offload(tbl(f), by = "cell")
  shard_res <- lapply(unclass(part), function(nd)
    nd |> spatial_join(polys["pid"], coords = c("x", "y"), crs = NA) |> collect())
  got <- do.call(rbind, shard_res)
  got <- got[order(got$id), ]

  expect_equal(nrow(got), n)
  expect_equal(got$pid, ref$pid)
})

test_that("spatial verbs validate their inputs", {
  polys <- two_squares()
  expect_error(spatial_map(42, ~ .x), "must be a vectra_node")
  expect_error(spatial_join(42, polys), "must be a vectra_node")

  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(data.frame(x = 1, y = 1), f)
  expect_error(spatial_join(tbl(f), data.frame(a = 1)),
               "must be an sf object")
})

# spatial_smooth(): Chaikin corner-cutting over streamed lines and polygons.

skip_if_not_installed("sf")

vtr_from <- function(sfobj) {
  f <- tempfile(fileext = ".vtr")
  df <- as.data.frame(sf::st_drop_geometry(sfobj))
  df$geometry <- sf::st_as_binary(sf::st_geometry(sfobj), hex = TRUE)
  write_vtr(df, f)
  f
}

zig <- function() sf::st_linestring(
  rbind(c(0, 0), c(1, 1), c(2, 0), c(3, 1), c(4, 0)))

test_that("smoothing a line adds vertices and keeps the endpoints", {
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(zig()))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_smooth(iterations = 3) |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_true(all(sf::st_geometry_type(d) == "LINESTRING"))
  co <- sf::st_coordinates(d)[, c("X", "Y")]
  expect_gt(nrow(co), 5L)                          # more vertices than the input
  expect_equal(unname(co[1, ]), c(0, 0), tolerance = 1e-9)
  expect_equal(unname(co[nrow(co), ]), c(4, 0), tolerance = 1e-9)
})

test_that("smoothing cuts the apex below the original and adds vertices", {
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(zig()))
  f <- vtr_from(x); on.exit(unlink(f))
  stat <- function(it) {
    d <- tbl(f) |> spatial_smooth(iterations = it) |> collect_sf()
    co <- sf::st_coordinates(d)
    list(peak = max(co[, "Y"]), n = nrow(co))
  }
  s1 <- stat(1); s3 <- stat(3)
  expect_lt(s1$peak, 1)                            # below the original apex y = 1
  expect_lt(s3$peak, 1)
  expect_gt(s3$n, s1$n)                            # more passes add more vertices
})

test_that("keep_ends = FALSE drops the original endpoints", {
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(zig()))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_smooth(iterations = 1, keep_ends = FALSE) |> collect_sf()
  co <- sf::st_coordinates(d)[, c("X", "Y")]
  expect_false(isTRUE(all.equal(unname(co[1, ]), c(0, 0))))
})

test_that("a polygon stays a closed valid polygon after smoothing", {
  sq <- sf::st_polygon(list(rbind(c(0, 0), c(4, 0), c(4, 4), c(0, 4), c(0, 0))))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(sq))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_smooth(iterations = 2) |> collect_sf()
  expect_true(all(sf::st_geometry_type(d) == "POLYGON"))
  expect_true(all(sf::st_is_valid(d)))
  # corner-cutting shrinks the ring but keeps most of the area
  a <- as.numeric(sf::st_area(d))
  expect_lt(a, 16); expect_gt(a, 12)
})

test_that("multilinestring smooths each part and attributes ride through", {
  mls <- sf::st_multilinestring(list(
    rbind(c(0, 0), c(1, 1), c(2, 0)),
    rbind(c(5, 5), c(6, 4), c(7, 5))))
  x <- sf::st_sf(id = 9L, k = "p", geometry = sf::st_sfc(mls))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_smooth(iterations = 2) |> collect_sf()
  expect_equal(d$id, 9L)
  expect_equal(d$k, "p")
  expect_true(all(sf::st_geometry_type(d) == "MULTILINESTRING"))
})

test_that("point geometry passes through unchanged", {
  pt <- sf::st_point(c(3, 4))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(pt))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_smooth() |> collect_sf()
  expect_equal(unname(sf::st_coordinates(d)[1, c("X", "Y")]), c(3, 4))
})

test_that("a CRS passed to the verb is carried", {
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(zig()))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_smooth(crs = 3857) |> collect_sf()
  expect_equal(sf::st_crs(d), sf::st_crs(3857))
})

test_that("invalid inputs are rejected", {
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(zig()))
  f <- vtr_from(x); on.exit(unlink(f))
  expect_error(spatial_smooth(data.frame(a = 1)), "vectra_node")
  expect_error(collect(spatial_smooth(tbl(f), iterations = 0)), "positive")
})

# spatial_overlay(x, y, how=): two-layer overlay into a planar partition carrying
# attributes from both layers. Planar (CRS NA / 3857) so areas are exact.
#
# Fixtures are two unit-height squares that overlap on [1,2]:
#   x = [0,2] x [0,1]   y = [1,3] x [0,1]
# so the partition is three unit pieces: x-only [0,1], both [1,2], y-only [2,3].

skip_if_not_installed("sf")

mk <- function(x0, x1, y0 = 0, y1 = 1) sf::st_polygon(list(rbind(
  c(x0, y0), c(x1, y0), c(x1, y1), c(x0, y1), c(x0, y0))))

xl <- sf::st_sf(xv = "A", geometry = sf::st_sfc(mk(0, 2)))
yl <- sf::st_sf(yv = "Z", geometry = sf::st_sfc(mk(1, 3)))

areas <- function(s) sort(as.numeric(sf::st_area(s)))

test_that("intersection keeps only the overlap, with both layers' attributes", {
  d <- spatial_overlay(xl, yl, how = "intersection") |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_equal(sum(as.numeric(sf::st_area(d))), 1, tolerance = 1e-4)
  expect_equal(d$xv, "A")
  expect_equal(d$yv, "Z")
})

test_that("union keeps every piece, absent side filled with NA", {
  d <- spatial_overlay(xl, yl, how = "union") |> collect_sf()
  expect_equal(nrow(d), 3L)
  expect_equal(areas(d), c(1, 1, 1), tolerance = 1e-4)
  expect_equal(sum(is.na(d$xv)), 1L)                    # the y-only piece
  expect_equal(sum(is.na(d$yv)), 1L)                    # the x-only piece
  expect_equal(sum(!is.na(d$xv) & !is.na(d$yv)), 1L)    # the overlap
})

test_that("identity keeps all of x split by y, drops y-only pieces", {
  d <- spatial_overlay(xl, yl, how = "identity") |> collect_sf()
  expect_equal(nrow(d), 2L)
  expect_equal(sum(as.numeric(sf::st_area(d))), 2, tolerance = 1e-4)
  expect_true(all(d$xv == "A"))                         # every piece has an x record
  expect_equal(sum(is.na(d$yv)), 1L)                    # the part of x outside y
})

test_that("symdiff keeps the parts in exactly one layer", {
  d <- spatial_overlay(xl, yl, how = "symdiff") |> collect_sf()
  expect_equal(nrow(d), 2L)
  expect_equal(sum(as.numeric(sf::st_area(d))), 2, tolerance = 1e-4)
  expect_true(all(xor(is.na(d$xv), is.na(d$yv))))       # exactly one side per piece
})

test_that("intersection matches sf::st_intersection on area", {
  old <- sf::sf_use_s2(FALSE); on.exit(sf::sf_use_s2(old))
  ref  <- suppressWarnings(sf::st_intersection(xl, yl))
  ours <- spatial_overlay(xl, yl, how = "intersection") |> collect_sf()
  expect_equal(sum(as.numeric(sf::st_area(ours))),
               sum(as.numeric(sf::st_area(ref))), tolerance = 1e-4)
})

test_that("shared column names are disambiguated with .x / .y", {
  x2 <- sf::st_sf(id = 1L, geometry = sf::st_sfc(mk(0, 2)))
  y2 <- sf::st_sf(id = 9L, geometry = sf::st_sfc(mk(1, 3)))
  d  <- spatial_overlay(x2, y2, how = "intersection") |> collect()
  expect_true(all(c("id.x", "id.y") %in% names(d)))
  expect_equal(d$id.x, 1L)
  expect_equal(d$id.y, 9L)
})

test_that("vars and vars_y select the carried columns per layer", {
  x3 <- sf::st_sf(a = 1L, b = 2L, geometry = sf::st_sfc(mk(0, 2)))
  y3 <- sf::st_sf(c = 3L, d = 4L, geometry = sf::st_sfc(mk(1, 3)))
  d  <- spatial_overlay(x3, y3, vars = "a", vars_y = "c", how = "intersection") |> collect()
  expect_true(all(c("a", "c") %in% names(d)))
  expect_false(any(c("b", "d") %in% names(d)))
})

test_that("a CRS mismatch is rejected", {
  xc <- sf::st_sf(g = 1L, geometry = sf::st_sfc(mk(0, 2), crs = 3857))
  yc <- sf::st_sf(h = 2L, geometry = sf::st_sfc(mk(1, 3), crs = 4326))
  expect_error(spatial_overlay(xc, yc), "share a CRS")
})

test_that("the shared CRS is carried onto the overlay node", {
  xc <- sf::st_sf(g = 1L, geometry = sf::st_sfc(mk(0, 2), crs = 3857))
  yc <- sf::st_sf(h = 2L, geometry = sf::st_sfc(mk(1, 3), crs = 3857))
  ov <- spatial_overlay(xc, yc, how = "intersection")
  expect_equal(sf::st_crs(collect_sf(ov)), sf::st_crs(3857))
})

test_that("a non-overlapping intersection is a typed empty node", {
  xa <- sf::st_sf(xv = "A", geometry = sf::st_sfc(mk(0, 1)))
  yb <- sf::st_sf(yv = "Z", geometry = sf::st_sfc(mk(5, 6)))
  d  <- spatial_overlay(xa, yb, how = "intersection") |> collect()
  expect_equal(nrow(d), 0L)
  expect_true(all(c("xv", "yv", "piece_id", "geometry") %in% names(d)))
})

# A self-overlay reached two ways: spatial_overlay(x) keeps one piece-row per
# covering source (so the overlap row appears once per year), while
# spatial_overlay(x, x, how="union") collapses each face to one geometry and
# pairs the covering x- and y-records. The partition itself -- the set of pieces
# and their areas -- is the same; only the row multiplicity differs.
test_that("union against a copy reproduces the self-union partition", {
  geoms <- sf::st_sfc(mk(0, 2), mk(1, 3))
  x <- sf::st_sf(year = c(1L, 2L), geometry = geoms)
  piece_areas <- function(o) {
    d <- collect(o)
    d <- d[!duplicated(d$piece_id), ]
    g <- sf::st_as_sfc(structure(d$geometry, class = "WKB"), EWKB = FALSE)
    sort(round(as.numeric(sf::st_area(g)), 6))
  }
  self <- piece_areas(spatial_overlay(x, vars = "year"))
  both <- piece_areas(spatial_overlay(x, x, vars = "year", vars_y = "year", how = "union"))
  expect_equal(both, self, tolerance = 1e-6)
})

test_that("reading y from a GeoPackage matches the in-memory overlay", {
  xc <- sf::st_sf(xv = "A", geometry = sf::st_sfc(mk(0, 2), crs = 3857))
  yc <- sf::st_sf(yv = "Z", geometry = sf::st_sfc(mk(1, 3), crs = 3857))
  gp <- tempfile(fileext = ".gpkg"); on.exit(unlink(gp))
  sf::st_write(yc, gp, "lyr", quiet = TRUE)
  pull <- function(o) {
    d <- as.data.frame(collect(o))
    d[order(d$piece_id), c("piece_id", "xv", "yv", "geometry")]
  }
  ref  <- pull(spatial_overlay(xc, yc, how = "union"))
  file <- pull(spatial_overlay(xc, gp, layer_y = "lyr", how = "union"))
  expect_equal(nrow(file), nrow(ref))
  expect_identical(file$geometry, ref$geometry)
  expect_identical(file$xv, ref$xv)
  expect_identical(file$yv, ref$yv)
})

# P2a + P2b: tdc-backed encode/decode bridges — round-trip via tdc.
#
# Exercises src/vtr_codec_tdc.c. Encodes an R vector through
# vtr_encode_column_tdc, then decodes the resulting tdc block record via
# vtr_decode_column_tdc_into. Verifies bit-identical round-trip across
# REALSXP / INTSXP / LGLSXP.
#
# String round-trip is intentionally absent: tdc v0 has no public size
# query for variable-width payloads, so the bridge returns
# TDC_E_UNSUPPORTED for VEC_STRING. Strings land alongside the tdc API
# extension (tracked in VECTRA_REWIRE.md follow-ups).

VTR_COMPRESS_NONE  <- 0L
VTR_COMPRESS_FAST  <- 1L
VTR_COMPRESS_SMALL <- 2L

# Match the SEXPTYPE codes hard-coded in vtr_codec_tdc.c::r_sxp_to_vectype.
# These are stable R constants (Rinternals.h).
SXP_LGL  <- 10L
SXP_INT  <- 13L
SXP_REAL <- 14L

roundtrip <- function(x, comp_level) {
  rt <- switch(typeof(x),
    "double"  = SXP_REAL,
    "integer" = SXP_INT,
    "logical" = SXP_LGL,
    stop("unsupported R type: ", typeof(x))
  )
  raw_bytes <- .Call("C_tdc_encode_column", x, comp_level, PACKAGE = "vectra")
  decoded   <- .Call("C_tdc_decode_column", raw_bytes, length(x), rt,
                     PACKAGE = "vectra")
  list(raw = raw_bytes, decoded = decoded)
}

test_that("REALSXP round-trips at every comp_level", {
  set.seed(42)
  cases <- list(
    monotone     = as.double(seq_len(1024)),
    random       = runif(1024, -100, 100),
    constant     = rep(3.14, 512),
    mixed_signs  = rnorm(2000),
    small        = c(1.0, 2.0, 3.0)
  )
  for (level in c(VTR_COMPRESS_NONE, VTR_COMPRESS_FAST, VTR_COMPRESS_SMALL)) {
    for (nm in names(cases)) {
      x <- cases[[nm]]
      rt <- roundtrip(x, level)
      expect_identical(rt$decoded, x,
                       info = sprintf("REALSXP case=%s level=%d", nm, level))
    }
  }
})

test_that("INTSXP round-trips at every comp_level", {
  set.seed(7)
  cases <- list(
    monotone = seq_len(2048),
    random   = sample.int(.Machine$integer.max, 1024) - 1L,
    constant = rep(42L, 256),
    small    = c(-1L, 0L, 1L, 2L, 3L)
  )
  for (level in c(VTR_COMPRESS_NONE, VTR_COMPRESS_FAST, VTR_COMPRESS_SMALL)) {
    for (nm in names(cases)) {
      x <- cases[[nm]]
      rt <- roundtrip(x, level)
      expect_identical(rt$decoded, x,
                       info = sprintf("INTSXP case=%s level=%d", nm, level))
    }
  }
})

test_that("LGLSXP round-trips at every comp_level", {
  set.seed(13)
  cases <- list(
    all_true  = rep(TRUE, 1024),
    all_false = rep(FALSE, 1024),
    mixed     = sample(c(TRUE, FALSE), 4096, replace = TRUE),
    small     = c(TRUE, FALSE, TRUE)
  )
  for (level in c(VTR_COMPRESS_NONE, VTR_COMPRESS_FAST, VTR_COMPRESS_SMALL)) {
    for (nm in names(cases)) {
      x <- cases[[nm]]
      rt <- roundtrip(x, level)
      expect_identical(rt$decoded, x,
                       info = sprintf("LGLSXP case=%s level=%d", nm, level))
    }
  }
})

test_that("FAST compresses a low-entropy double vector below raw size", {
  x <- as.double(rep(seq_len(64), 256))     # 16384 doubles, very repetitive
  raw_size <- length(x) * 8
  rt <- roundtrip(x, VTR_COMPRESS_FAST)
  expect_lt(length(rt$raw), raw_size)
  expect_identical(rt$decoded, x)
})

test_that("NONE produces a passthrough block (no entropy stage)", {
  x <- runif(256)
  rt_none <- roundtrip(x, VTR_COMPRESS_NONE)
  rt_fast <- roundtrip(x, VTR_COMPRESS_FAST)
  expect_gte(length(rt_none$raw), length(x) * 8)
  expect_identical(rt_none$decoded, x)
  expect_identical(rt_fast$decoded, x)
})

test_that("empty vectors round-trip cleanly", {
  for (x in list(double(0), integer(0), logical(0))) {
    rt <- roundtrip(x, VTR_COMPRESS_FAST)
    expect_identical(rt$decoded, x)
  }
})

# spatial_snap_grid() rounds coordinates to a lattice (the overlay noder's snap,
# exposed); spatial_snap() pulls vertices toward a resident reference layer.

skip_if_not_installed("sf")

vtr_from <- function(sfobj) {
  f <- tempfile(fileext = ".vtr")
  df <- as.data.frame(sf::st_drop_geometry(sfobj))
  df$geometry <- sf::st_as_binary(sf::st_geometry(sfobj), hex = TRUE)
  write_vtr(df, f)
  f
}

test_that("snap_grid rounds jittered coordinates onto the grid", {
  p <- sf::st_polygon(list(rbind(c(0.04, 0.03), c(1.02, 0.01),
                                 c(0.98, 1.03), c(0.01, 0.97), c(0.04, 0.03))))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(p))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_snap_grid(0.1) |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_equal(d$id, 1L)
  co <- sf::st_coordinates(d)[, c("X", "Y")]
  expect_equal(co, round(co / 0.1) * 0.1, tolerance = 1e-9)
})

test_that("snap_grid keeps one cleaned feature per input and its attributes", {
  ps <- sf::st_sfc(
    sf::st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0)))),
    sf::st_polygon(list(rbind(c(2, 2), c(3, 2), c(3, 3), c(2, 3), c(2, 2)))))
  x <- sf::st_sf(id = c(10L, 20L), k = c("a", "b"), geometry = ps)
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_snap_grid(0.5) |> collect_sf()
  expect_equal(nrow(d), 2L)
  expect_equal(d$id, c(10L, 20L))
  expect_equal(d$k, c("a", "b"))
})

test_that("snap_grid carries a CRS passed to the verb", {
  p <- sf::st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(p))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_snap_grid(0.5, crs = 3857) |> collect_sf()
  expect_equal(sf::st_crs(d), sf::st_crs(3857))
})

test_that("snap_grid rejects a non-positive size", {
  p <- sf::st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))))
  x <- sf::st_sf(geometry = sf::st_sfc(p))
  f <- vtr_from(x); on.exit(unlink(f))
  expect_error(collect(spatial_snap_grid(tbl(f), 0)), "positive")
})

test_that("snap pulls near vertices onto a reference layer", {
  ref  <- sf::st_sfc(sf::st_linestring(rbind(c(0, 0), c(10, 0))))
  line <- sf::st_linestring(rbind(c(0, 0.2), c(5, 0.1), c(10, 0.2)))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(line))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_snap(ref, tolerance = 0.5) |> collect_sf()
  # the two endpoints sit within tolerance of the reference line, so they snap
  # onto y = 0
  ends <- sf::st_coordinates(d)
  expect_equal(min(abs(ends[, "Y"])), 0, tolerance = 1e-9)
  expect_lt(min(ends[, "Y"]), 0.2)
})

test_that("snap leaves vertices beyond the tolerance untouched", {
  ref  <- sf::st_sfc(sf::st_linestring(rbind(c(0, 0), c(10, 0))))
  line <- sf::st_linestring(rbind(c(0, 5), c(10, 5)))
  x <- sf::st_sf(id = 1L, geometry = sf::st_sfc(line))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_snap(ref, tolerance = 0.5) |> collect_sf()
  expect_equal(unique(sf::st_coordinates(d)[, "Y"]), 5, tolerance = 1e-9)
})

test_that("snap rejects a non-sf reference and a non-positive tolerance", {
  line <- sf::st_linestring(rbind(c(0, 0), c(1, 1)))
  x <- sf::st_sf(geometry = sf::st_sfc(line))
  f <- vtr_from(x); on.exit(unlink(f))
  ref <- sf::st_sfc(sf::st_linestring(rbind(c(0, 0), c(10, 0))))
  expect_error(spatial_snap(tbl(f), data.frame(a = 1), tolerance = 1), "sf")
  expect_error(collect(spatial_snap(tbl(f), ref, tolerance = -1)), "positive")
})

test_that("a non-vectra_node input is rejected by both verbs", {
  ref <- sf::st_sfc(sf::st_linestring(rbind(c(0, 0), c(10, 0))))
  expect_error(spatial_snap_grid(data.frame(a = 1), 1), "vectra_node")
  expect_error(spatial_snap(data.frame(a = 1), ref, tolerance = 1), "vectra_node")
})

# Recovery test for the two-sided nearest-feature join on the grid partition.
#
# st_nearest_feature is not local to a cell, so the partition path searches the
# left point's 3x3 cell neighbourhood. With a cellsize at least the largest
# nearest distance, the streamed result must equal the resident st_join, and be
# equal across batch sizes (streaming invariance).

write_points <- function(df) {
  f <- tempfile(fileext = ".vtr")
  write_vtr(df, f)
  f
}

write_target_wkb <- function(df) {
  g <- sf::st_as_binary(sf::st_geometry(sf::st_as_sf(df, coords = c("x", "y"))),
                        hex = TRUE)
  f <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(tid = df$tid, geometry = g), f)
  f
}

test_that("two-sided nearest matches the resident nearest join", {
  skip_if_not_installed("sf")
  set.seed(1)
  n <- 300
  left <- data.frame(id = seq_len(n), x = runif(n, 0, 10), y = runif(n, 0, 10))
  targ <- data.frame(tid = 1:9,
                     x = rep(c(1.5, 5, 8.5), 3),
                     y = rep(c(1.5, 5, 8.5), each = 3))
  fl <- write_points(left); ft <- write_target_wkb(targ)
  on.exit(unlink(c(fl, ft)))

  got <- collect(tbl(fl) |>
    spatial_join(tbl(ft), coords = c("x", "y"),
                 join = sf::st_nearest_feature, partition = grid(5)))
  got <- got[order(got$id), ]

  lsf <- sf::st_as_sf(left, coords = c("x", "y"))
  tsf <- sf::st_as_sf(targ, coords = c("x", "y"))
  ref <- sf::st_join(lsf, tsf, join = sf::st_nearest_feature)
  ref <- ref[order(ref$id), ]

  expect_equal(got$tid, ref$tid)
})

test_that("two-sided nearest is invariant to the partition budget", {
  skip_if_not_installed("sf")
  set.seed(2)
  n <- 200
  left <- data.frame(id = seq_len(n), x = runif(n, 0, 20), y = runif(n, 0, 20))
  targ <- data.frame(tid = 1:16,
                     x = rep(seq(2.5, 17.5, 5), 4),
                     y = rep(seq(2.5, 17.5, 5), each = 4))
  fl <- write_points(left); ft <- write_target_wkb(targ)
  on.exit(unlink(c(fl, ft)))

  run <- function(budget) {
    old <- options(vectra.partition_budget = budget); on.exit(options(old))
    r <- collect(tbl(fl) |>
      spatial_join(tbl(ft), coords = c("x", "y"),
                   join = sf::st_nearest_feature, partition = grid(7)))
    r[order(r$id), "tid"]
  }
  expect_equal(run(20), run(1e6))
})

# spatial_knn(): k nearest resident-y features per streamed left row, returned
# long with rank, identifier, and distance.

skip_if_not_installed("sf")

vtr_pts <- function(df) {
  f <- tempfile(fileext = ".vtr")
  write_vtr(df, f)
  f
}

# four resident neighbours on the x-axis at x = 0, 1, 2, 3
nbrs <- function() sf::st_sf(
  town = c("a", "b", "c", "d"),
  geometry = sf::st_sfc(sf::st_point(c(0, 0)), sf::st_point(c(1, 0)),
                        sf::st_point(c(2, 0)), sf::st_point(c(3, 0))))

test_that("k=1 returns the single nearest neighbour and its distance", {
  f <- vtr_pts(data.frame(id = c(1L, 2L), x = c(0.1, 2.9), y = c(0, 0)))
  on.exit(unlink(f))
  d <- tbl(f) |> spatial_knn(nbrs(), k = 1, coords = c("x", "y")) |> collect()
  expect_equal(nrow(d), 2L)
  expect_equal(d$rank, c(1L, 1L))
  expect_equal(d$neighbor, c(1L, 4L))         # nearest to 0.1 is x=0, to 2.9 is x=3
  expect_equal(d$distance, c(0.1, 0.1), tolerance = 1e-9)
})

test_that("k=2 returns two ranked neighbours per left feature in order", {
  f <- vtr_pts(data.frame(id = 1L, x = 0.2, y = 0))
  on.exit(unlink(f))
  d <- tbl(f) |> spatial_knn(nbrs(), k = 2, coords = c("x", "y")) |> collect()
  expect_equal(nrow(d), 2L)
  expect_equal(d$rank, c(1L, 2L))
  expect_equal(d$neighbor, c(1L, 2L))         # x=0 then x=1
  expect_equal(d$distance, c(0.2, 0.8), tolerance = 1e-9)
  expect_true(all(diff(d$distance) >= 0))     # non-decreasing with rank
})

test_that("y_id labels neighbours by a y column", {
  f <- vtr_pts(data.frame(id = 1L, x = 0.1, y = 0))
  on.exit(unlink(f))
  d <- tbl(f) |>
    spatial_knn(nbrs(), k = 1, coords = c("x", "y"), y_id = "town") |> collect()
  expect_equal(d$neighbor, "a")
})

test_that("k is capped at the number of y features", {
  f <- vtr_pts(data.frame(id = 1L, x = 0.5, y = 0))
  on.exit(unlink(f))
  d <- tbl(f) |> spatial_knn(nbrs(), k = 10, coords = c("x", "y")) |> collect()
  expect_equal(nrow(d), 4L)
  expect_equal(sort(d$neighbor), 1:4)
})

test_that("custom output column names are honoured", {
  f <- vtr_pts(data.frame(id = 1L, x = 0.1, y = 0))
  on.exit(unlink(f))
  d <- tbl(f) |>
    spatial_knn(nbrs(), k = 1, coords = c("x", "y"),
                id_col = "nb", dist_col = "dkm", rank_col = "rk") |> collect()
  expect_true(all(c("nb", "dkm", "rk") %in% names(d)))
})

test_that("left attributes ride through, replicated per neighbour", {
  f <- vtr_pts(data.frame(id = 7L, tag = "z", x = 0.2, y = 0))
  on.exit(unlink(f))
  d <- tbl(f) |> spatial_knn(nbrs(), k = 3, coords = c("x", "y")) |> collect()
  expect_equal(d$id, rep(7L, 3))
  expect_equal(d$tag, rep("z", 3))
})

test_that("hex-WKB point geometry input works and carries CRS", {
  pts <- sf::st_sfc(sf::st_point(c(0.1, 0)), sf::st_point(c(2.9, 0)), crs = 3857)
  x <- sf::st_sf(id = c(1L, 2L), geometry = pts)
  f <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(id = x$id,
                       geometry = sf::st_as_binary(pts, hex = TRUE)), f)
  on.exit(unlink(f))
  y <- sf::st_set_crs(nbrs(), 3857)
  d <- tbl(f) |> spatial_knn(y, k = 1, crs = 3857) |> collect_sf()
  expect_equal(d$neighbor, c(1L, 4L))
  expect_equal(sf::st_crs(d), sf::st_crs(3857))
})

test_that("invalid inputs are rejected", {
  f <- vtr_pts(data.frame(id = 1L, x = 0, y = 0))
  on.exit(unlink(f))
  expect_error(spatial_knn(data.frame(a = 1), nbrs()), "vectra_node")
  expect_error(spatial_knn(tbl(f), data.frame(a = 1)), "sf")
  expect_error(collect(spatial_knn(tbl(f), nbrs(), k = 0, coords = c("x", "y"))),
               "positive")
  expect_error(
    collect(spatial_knn(tbl(f), nbrs(), coords = c("x", "y"), y_id = "nope")),
    "not found")
})

# spatial_construct(): set-wise constructions on the partition tier. Geometry
# round-trips through a .vtr; CRS is carried on the node.

skip_if_not_installed("sf")

vtr_from <- function(sfobj) {
  f <- tempfile(fileext = ".vtr")
  df <- as.data.frame(sf::st_drop_geometry(sfobj))
  df$geometry <- sf::st_as_binary(sf::st_geometry(sfobj), hex = TRUE)
  write_vtr(df, f)
  f
}

# A square cloud of points (corners + an interior point) around (ox, oy).
pt_cloud <- function(ox = 0, oy = 0) {
  xy <- rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0.4, 0.6))
  xy[, 1] <- xy[, 1] + ox; xy[, 2] <- xy[, 2] + oy
  sf::st_sfc(lapply(seq_len(nrow(xy)), function(i) sf::st_point(xy[i, ])))
}

test_that("convex hull of the whole layer is one polygon of the right area", {
  x <- sf::st_sf(geometry = pt_cloud())
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_construct("convex_hull") |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_true(all(sf::st_geometry_type(d) == "POLYGON"))
  expect_equal(as.numeric(sf::st_area(d)), 1, tolerance = 1e-9)
})

test_that("by= builds one construction per group", {
  x <- sf::st_sf(
    g = c(rep("a", 5), rep("b", 5)),
    geometry = c(pt_cloud(0, 0), pt_cloud(10, 10)))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_construct("convex_hull", by = "g") |> collect_sf()
  expect_equal(nrow(d), 2L)
  expect_setequal(d$g, c("a", "b"))
  expect_equal(sort(as.numeric(sf::st_area(d))), c(1, 1), tolerance = 1e-9)
})

test_that("envelope is the axis-aligned bounding rectangle", {
  x <- sf::st_sf(geometry = pt_cloud())
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_construct("envelope") |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_equal(as.numeric(sf::st_area(d)), 1, tolerance = 1e-9)
})

test_that("oriented box is no larger than the envelope", {
  set.seed(1)
  xy <- cbind(runif(40), runif(40))
  x <- sf::st_sf(geometry = sf::st_sfc(
    lapply(seq_len(nrow(xy)), function(i) sf::st_point(xy[i, ]))))
  f <- vtr_from(x); on.exit(unlink(f))
  ob <- tbl(f) |> spatial_construct("oriented_box") |> collect_sf()
  env <- tbl(f) |> spatial_construct("envelope") |> collect_sf()
  expect_true(all(sf::st_geometry_type(ob) == "POLYGON"))
  expect_lte(as.numeric(sf::st_area(ob)), as.numeric(sf::st_area(env)) + 1e-9)
})

test_that("the enclosing circle covers every input point", {
  x <- sf::st_sf(geometry = pt_cloud())
  f <- vtr_from(x); on.exit(unlink(f))
  circ <- tbl(f) |> spatial_construct("enclosing_circle") |> collect_sf()
  # the bounding circle passes through the extreme points; a tiny buffer absorbs
  # the segmentation error so the boundary points count as covered
  circ_b <- sf::st_buffer(sf::st_geometry(circ), 1e-6)
  inside <- sf::st_covers(circ_b, pt_cloud())[[1]]
  expect_equal(length(inside), 5L)
})

test_that("the inscribed circle lies inside the shape", {
  poly <- sf::st_polygon(list(rbind(c(0, 0), c(4, 0), c(4, 2), c(0, 2), c(0, 0))))
  x <- sf::st_sf(geometry = sf::st_sfc(poly))
  f <- vtr_from(x); on.exit(unlink(f))
  ic <- tbl(f) |> spatial_construct("inscribed_circle") |> collect_sf()
  expect_equal(nrow(ic), 1L)
  expect_true(all(sf::st_geometry_type(ic) == "POLYGON"))
  covered <- lengths(sf::st_covered_by(sf::st_geometry(ic), sf::st_sfc(poly)))
  expect_equal(covered, 1L)
})

test_that("the pole of inaccessibility is a point inside the shape", {
  poly <- sf::st_polygon(list(rbind(c(0, 0), c(4, 0), c(4, 2), c(0, 2), c(0, 0))))
  x <- sf::st_sf(geometry = sf::st_sfc(poly))
  f <- vtr_from(x); on.exit(unlink(f))
  p <- tbl(f) |> spatial_construct("pole") |> collect_sf()
  expect_equal(nrow(p), 1L)
  expect_true(all(sf::st_geometry_type(p) == "POINT"))
  expect_equal(lengths(sf::st_within(sf::st_geometry(p), sf::st_sfc(poly))), 1L)
  # the deepest interior point of a 4x2 box sits on its centre line (y = 1)
  expect_equal(unname(sf::st_coordinates(p)[1, "Y"]), 1, tolerance = 1e-2)
})

test_that("voronoi and delaunay emit one polygon per cell", {
  x <- sf::st_sf(geometry = pt_cloud())
  f <- vtr_from(x); on.exit(unlink(f))
  v <- tbl(f) |> spatial_construct("voronoi") |> collect_sf()
  d <- tbl(f) |> spatial_construct("delaunay") |> collect_sf()
  expect_gt(nrow(v), 1L)
  expect_true(all(sf::st_geometry_type(v) == "POLYGON"))
  expect_gt(nrow(d), 1L)
  expect_true(all(sf::st_geometry_type(d) == "POLYGON"))
})

test_that("a tessellation repeats the group's by values onto every cell", {
  x <- sf::st_sf(
    g = c(rep("a", 5), rep("b", 5)),
    geometry = c(pt_cloud(0, 0), pt_cloud(10, 10)))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_construct("delaunay", by = "g") |> collect_sf()
  expect_setequal(d$g, c("a", "b"))
  expect_true(all(d$g %in% c("a", "b")))
  expect_gt(sum(d$g == "a"), 0L)
  expect_gt(sum(d$g == "b"), 0L)
})

test_that("concave hull returns a polygon", {
  set.seed(2)
  xy <- cbind(runif(60), runif(60))
  x <- sf::st_sf(geometry = sf::st_sfc(
    lapply(seq_len(nrow(xy)), function(i) sf::st_point(xy[i, ]))))
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_construct("concave_hull", ratio = 0.5) |> collect_sf()
  expect_equal(nrow(d), 1L)
  expect_true(all(sf::st_geometry_type(d) %in% c("POLYGON", "MULTIPOLYGON")))
})

test_that("a CRS passed to the verb is carried onto the construction", {
  x <- sf::st_sf(geometry = pt_cloud())
  f <- vtr_from(x); on.exit(unlink(f))
  d <- tbl(f) |> spatial_construct("convex_hull", crs = 3857) |> collect_sf()
  expect_equal(sf::st_crs(d), sf::st_crs(3857))
})

test_that("a missing by column is rejected", {
  x <- sf::st_sf(geometry = pt_cloud())
  f <- vtr_from(x); on.exit(unlink(f))
  expect_error(collect(spatial_construct(tbl(f), "convex_hull", by = "nope")),
               "not found")
})

test_that("a non-vectra_node input is rejected", {
  expect_error(spatial_construct(data.frame(a = 1), "convex_hull"), "vectra_node")
})

# spatial_dissolve (aggregate geometries by group) -- the streamed partition-tier
# dissolve checked against sf in-memory ground truth. Planar CRS = NA so union
# areas are exact Cartesian and the streamed path must equal the resident path.

skip_if_not_installed("sf")

make_square <- function(xmin, xmax, ymin, ymax) {
  sf::st_polygon(list(rbind(
    c(xmin, ymin), c(xmax, ymin), c(xmax, ymax),
    c(xmin, ymax), c(xmin, ymin))))
}

# Two overlapping squares in band A (union area 3) and two disjoint squares in
# band B (union area 2); a `val` attribute to summarise.
fixture <- function() {
  geoms <- sf::st_sfc(
    make_square(0, 2, 0, 1), make_square(1, 3, 0, 1),   # band A, overlap [1,2]
    make_square(0, 1, 2, 3), make_square(2, 3, 2, 3))   # band B, disjoint
  df <- data.frame(
    band = c("A", "A", "B", "B"),
    val  = c(10, 20, 5, 7),
    geometry = sf::st_as_binary(geoms, hex = TRUE),
    stringsAsFactors = FALSE)
  list(geoms = geoms, df = df)
}

test_that("dissolve unions geometry within each group, matching sf", {
  fx <- fixture()
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(fx$df, f)

  out <- tbl(f) |> spatial_dissolve(by = "band", crs = NA) |> collect_sf()
  out <- out[order(out$band), ]
  expect_equal(out$band, c("A", "B"))
  # one feature per group, geometry = union of that group's squares
  refA <- sf::st_union(fx$geoms[1:2])
  refB <- sf::st_union(fx$geoms[3:4])
  expect_equal(as.numeric(sf::st_area(out)),
               c(as.numeric(sf::st_area(refA)),
                 as.numeric(sf::st_area(refB))), tolerance = 1e-9)
  expect_equal(as.numeric(sf::st_area(out)), c(3, 2), tolerance = 1e-9)
})

test_that(".fun summarises attributes per group", {
  fx <- fixture()
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(fx$df, f)

  out <- tbl(f) |>
    spatial_dissolve(by = "band", crs = NA,
                     .fun = list(total = function(d) sum(d$val),
                                 n = function(d) nrow(d))) |>
    collect_sf()
  out <- out[order(out$band), ]
  expect_equal(sort(names(out)),
               sort(c("band", "total", "n", "geometry")))
  expect_equal(out$total, c(30, 12))   # A: 10+20, B: 5+7
  expect_equal(out$n, c(2, 2))
})

test_that("by = NULL dissolves the whole layer into one feature", {
  fx <- fixture()
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(fx$df, f)

  out <- tbl(f) |> spatial_dissolve(crs = NA) |> collect_sf()
  expect_equal(nrow(out), 1L)
  # bands A and B are disjoint from each other: total area 3 + 2 = 5
  expect_equal(as.numeric(sf::st_area(out)), 5, tolerance = 1e-9)
})

test_that("multi-column by groups on the value combination", {
  geoms <- sf::st_sfc(
    make_square(0, 1, 0, 1), make_square(1, 2, 0, 1),   # r1/c1, r1/c2
    make_square(0, 1, 2, 3), make_square(1, 2, 2, 3))   # r2/c1, r2/c2
  df <- data.frame(
    region = c("r1", "r1", "r2", "r2"),
    cls    = c("c1", "c2", "c1", "c1"),
    geometry = sf::st_as_binary(geoms, hex = TRUE),
    stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f)

  out <- tbl(f) |>
    spatial_dissolve(by = c("region", "cls"), crs = NA) |> collect_sf()
  key <- paste(out$region, out$cls)
  ord <- order(key)
  out <- out[ord, ]; key <- key[ord]
  # three groups: (r1,c1), (r1,c2), (r2,c1 = two squares unioned)
  expect_equal(key, c("r1 c1", "r1 c2", "r2 c1"))
  expect_equal(as.numeric(sf::st_area(out)), c(1, 1, 2), tolerance = 1e-9)
})

test_that("streamed dissolve equals the single-batch resident dissolve", {
  set.seed(31)
  k <- 120L
  # a strip of unit squares, each tagged into one of 4 bands
  geoms <- do.call(sf::st_sfc,
                   lapply(seq_len(k), function(i) make_square(i, i + 1, 0, 1)))
  band  <- sample(letters[1:4], k, replace = TRUE)
  df <- data.frame(band = band,
                   geometry = sf::st_as_binary(geoms, hex = TRUE),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(df, f, batch_size = 9L)                 # several read batches
  old <- options(vectra.spatial_flush = 13,
                 vectra.partition_budget = 17)      # several routing flushes
  on.exit(options(old), add = TRUE)

  got <- tbl(f) |> spatial_dissolve(by = "band", crs = NA) |> collect_sf()
  got <- got[order(got$band), ]

  resident <- sf::st_sf(band = band, geometry = geoms)
  ref <- lapply(sort(unique(band)), function(b)
    sf::st_union(sf::st_geometry(resident[resident$band == b, ])))
  expect_equal(got$band, sort(unique(band)))
  expect_equal(as.numeric(sf::st_area(got)),
               vapply(ref, function(g) as.numeric(sf::st_area(g)), numeric(1)),
               tolerance = 1e-9)
})

test_that("spatial_dissolve validates inputs", {
  f <- tempfile(fileext = ".vtr"); on.exit(unlink(f))
  write_vtr(data.frame(band = "A",
                       geometry = "01010000000000000000000000"), f)
  expect_error(spatial_dissolve(42), "must be a vectra_node")
  expect_error(spatial_dissolve(tbl(f), by = 1), "character vector")
  expect_error(spatial_dissolve(tbl(f), .fun = list(function(d) 1)),
               "named list")
  expect_error(spatial_dissolve(tbl(f), by = "nope"), "not found")
})

# spatial_join(partition = grid()) -- the two-sided streamed path. Both inputs
# are vectra_nodes binned to a grid and joined per shard. For point left
# geometries the result must equal the resident sf::st_join, and a single huge
# cell must reproduce the resident spatial_join exactly (streaming invariance).

# A square polygon of side `s` with lower-left corner (x0, y0).
sq <- function(x0, y0, s) sf::st_polygon(list(rbind(
  c(x0, y0), c(x0 + s, y0), c(x0 + s, y0 + s), c(x0, y0 + s), c(x0, y0))))

# Four corner squares (with one overlapping the first) over [0, 20], leaving
# gaps so some points stay unmatched and the overlap yields one-to-many hits.
make_polys <- function() {
  sf::st_sf(
    poly = 1:5,
    geometry = sf::st_sfc(
      sq(0, 0, 8), sq(10, 0, 8), sq(0, 10, 8), sq(10, 10, 8), sq(4, 4, 8)))
}

write_points <- function(pts) {
  f <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), f)
  f
}

write_polys_wkb <- function(polys) {
  f <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(
    poly = polys$poly,
    geometry = sf::st_as_binary(sf::st_geometry(polys), hex = TRUE)), f)
  f
}

# Collect a join node down to a comparable (id, poly) table, row-ordered.
key_table <- function(node) {
  d <- collect(node)
  d <- d[order(d$id, d$poly, na.last = TRUE), c("id", "poly"), drop = FALSE]
  rownames(d) <- NULL
  d
}

test_that("partitioned point-in-polygon join equals the resident join", {
  skip_if_not_installed("sf")
  set.seed(20)
  pts   <- cbind(runif(400, -2, 22), runif(400, -2, 22))
  polys <- make_polys()
  fp <- write_points(pts); fg <- write_polys_wkb(polys)
  on.exit(unlink(c(fp, fg)))

  resident <- tbl(fp) |>
    spatial_join(polys["poly"], coords = c("x", "y"), join = sf::st_intersects)
  parted <- tbl(fp) |>
    spatial_join(tbl(fg), coords = c("x", "y"), join = sf::st_intersects,
                 partition = grid(5))

  expect_equal(key_table(parted), key_table(resident))
})

test_that("a single huge cell reproduces the resident join exactly", {
  skip_if_not_installed("sf")
  set.seed(21)
  pts   <- cbind(runif(250, 0, 20), runif(250, 0, 20))
  polys <- make_polys()
  fp <- write_points(pts); fg <- write_polys_wkb(polys)
  on.exit(unlink(c(fp, fg)))

  resident <- tbl(fp) |>
    spatial_join(polys["poly"], coords = c("x", "y"))
  one_cell <- tbl(fp) |>
    spatial_join(tbl(fg), coords = c("x", "y"), partition = grid(1000))

  expect_equal(key_table(one_cell), key_table(resident))
})

test_that("an inner partitioned join drops the unmatched left rows", {
  skip_if_not_installed("sf")
  set.seed(22)
  pts   <- cbind(runif(300, -2, 22), runif(300, -2, 22))
  polys <- make_polys()
  fp <- write_points(pts); fg <- write_polys_wkb(polys)
  on.exit(unlink(c(fp, fg)))

  resident <- tbl(fp) |>
    spatial_join(polys["poly"], coords = c("x", "y"), left = FALSE)
  parted <- tbl(fp) |>
    spatial_join(tbl(fg), coords = c("x", "y"), left = FALSE,
                 partition = grid(4))

  kt <- key_table(parted)
  expect_false(anyNA(kt$poly))                 # inner: every row matched
  expect_equal(kt, key_table(resident))
})

test_that("the overlap region produces one row per covering polygon", {
  skip_if_not_installed("sf")
  # A point at (6, 6) sits in both square 1 (0,0,8) and square 5 (4,4,8).
  pts   <- cbind(c(6, 15), c(6, 2))
  polys <- make_polys()
  fp <- write_points(pts); fg <- write_polys_wkb(polys)
  on.exit(unlink(c(fp, fg)))

  parted <- tbl(fp) |>
    spatial_join(tbl(fg), coords = c("x", "y"), partition = grid(3))
  kt <- key_table(parted)
  expect_equal(sort(kt$poly[kt$id == 1]), c(1, 5))   # both covering squares
  expect_equal(kt$poly[kt$id == 2], 2)               # only square 2
})

test_that("partitioned join carries the result through collect_sf", {
  skip_if_not_installed("sf")
  set.seed(23)
  pts   <- cbind(runif(80, 0, 20), runif(80, 0, 20))
  polys <- make_polys()
  fp <- write_points(pts); fg <- write_polys_wkb(polys)
  on.exit(unlink(c(fp, fg)))

  out <- tbl(fp) |>
    spatial_join(tbl(fg), coords = c("x", "y"), crs = 4326,
                 partition = grid(6))
  sf_out <- collect_sf(out)
  expect_s3_class(sf_out, "sf")
  expect_equal(as.integer(sf::st_crs(sf_out)$epsg), 4326L)
})

test_that("grid() validates its arguments", {
  expect_error(grid(-1), "positive")
  expect_error(grid(0), "positive")
  expect_error(grid(c(1, 2, 3)), "one or two")
  expect_error(grid(1, origin = 0), "length-2")
  g <- grid(c(2, 4), origin = c(1, 1))
  expect_s3_class(g, "vectra_grid")
  expect_equal(g$cellsize, c(2, 4))
})

test_that("spatial_join rejects a bad partition or non-node y", {
  skip_if_not_installed("sf")
  fp <- write_points(cbind(1, 1)); on.exit(unlink(fp))
  expect_error(
    tbl(fp) |> spatial_join(make_polys(), coords = c("x", "y"),
                            partition = grid(1)),
    "vectra_node")
  expect_error(
    tbl(fp) |> spatial_join(tbl(fp), coords = c("x", "y"),
                            partition = 5),
    "grid")
})

# Grouping and aggregation verbs: group_by, summarise, ungroup, count, tally
# Includes internal helper: parse_agg_expr

#' Group a vectra query by columns
#'
#' @param .data A `vectra_node` object.
#' @param ... Grouping column names (unquoted).
#'
#' @return A `vectra_node` with grouping information stored.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> group_by(cyl) |> summarise(avg = mean(mpg)) |> collect()
#' unlink(f)
#'
#' @export
group_by <- function(.data, ...) {
  UseMethod("group_by")
}

#' @export
group_by.vectra_node <- function(.data, ...) {
  grp_exprs <- eval(substitute(alist(...)))
  grp_names <- vapply(grp_exprs, as.character, character(1))
  structure(list(.node = .data$.node, .path = .data$.path,
                 .groups = grp_names),
            class = "vectra_node")
}

#' Summarise grouped data
#'
#' @param .data A grouped `vectra_node` (from [group_by()]).
#' @param ... Named aggregation expressions using `n()`, `sum()`, `mean()`,
#'   `min()`, `max()`, `sd()`, `var()`, `first()`, `last()`, `any()`, `all()`,
#'   `median()`, `n_distinct()`.
#' @param .groups How to handle groups in the result. One of `"drop_last"`
#'   (default), `"drop"`, or `"keep"`.
#'
#' @return A `vectra_node` with one row per group.
#'
#' @details
#' Aggregation is hash-based by default. When the engine detects it is
#' advantageous, it switches to a sort-based path that can spill to disk,
#' keeping memory bounded regardless of group count.
#'
#' All aggregation functions accept `na.rm = TRUE` to skip NA values.
#' Without `na.rm`, any NA in a group poisons the result (returns NA).
#' R-matching edge cases: `sum(na.rm = TRUE)` on all-NA returns 0,
#' `mean(na.rm = TRUE)` on all-NA returns NaN, `min/max(na.rm = TRUE)` on
#' all-NA returns Inf/-Inf with a warning.
#'
#' This is a materializing operation.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> group_by(cyl) |> summarise(avg_mpg = mean(mpg)) |> collect()
#' unlink(f)
#'
#' @export
summarise <- function(.data, ..., .groups = NULL) {
  UseMethod("summarise")
}

#' @export
summarise.vectra_node <- function(.data, ..., .groups = NULL) {
  dots <- eval(substitute(alist(...)))
  # Expand across() calls
  schema <- .Call(C_node_schema, .data$.node)
  proxy <- schema_proxy(schema)
  dots <- expand_across(dots, schema$name, parent.frame(), proxy)
  dot_names <- names(dots)
  if (is.null(dot_names) || any(dot_names == ""))
    stop("all summarise expressions must be named")

  key_names <- .data$.groups
  if (is.null(key_names)) key_names <- character(0)

  # Parse agg expressions, detecting nested expressions like mean(x + y)
  agg_specs <- vector("list", length(dots))
  mutate_exprs <- list()  # nested exprs that need a hidden mutate
  for (i in seq_along(dots)) {
    parsed <- parse_agg_expr(dots[[i]], dot_names[i])
    if (!is.null(parsed$.nested_expr)) {
      # The inner expression needs a hidden mutate column
      tmp_name <- parsed$.nested_col
      mutate_exprs[[tmp_name]] <- parsed$.nested_expr
      parsed$.nested_expr <- NULL
      parsed$.nested_col <- NULL
    }
    agg_specs[[i]] <- parsed
  }

  # Insert hidden mutate node if there are nested expressions
  node <- .data
  if (length(mutate_exprs) > 0) {
    cur_schema <- .Call(C_node_schema, node$.node)
    existing_names <- cur_schema$name

    out_names <- existing_names
    out_exprs <- vector("list", length(existing_names))

    for (tmp_nm in names(mutate_exprs)) {
      out_names <- c(out_names, tmp_nm)
      out_exprs <- c(out_exprs, list(
        serialize_expr(mutate_exprs[[tmp_nm]], parent.frame(), existing_names)))
    }

    new_xptr <- .Call(C_project_node, node$.node, out_names, out_exprs)
    node <- structure(list(.node = new_xptr, .path = node$.path,
                           .groups = node$.groups), class = "vectra_node")
  }

  # Check for R-fallback aggregations (median, n_distinct)
  has_fallback <- any(vapply(agg_specs, function(s) isTRUE(s$.r_fallback), logical(1)))
  if (has_fallback) {
    df <- collect(node)
    .eval_agg <- function(spec, chunk) {
      col <- if (!is.null(spec$col)) chunk[[spec$col]] else NULL
      switch(spec$kind,
        n = nrow(chunk),
        sum = sum(col, na.rm = spec$na_rm),
        mean = mean(col, na.rm = spec$na_rm),
        min = min(col, na.rm = spec$na_rm),
        max = max(col, na.rm = spec$na_rm),
        sd = sd(col, na.rm = spec$na_rm),
        var = var(col, na.rm = spec$na_rm),
        first = col[!is.na(col)][1],
        last = rev(col[!is.na(col)])[1],
        any = any(as.logical(col), na.rm = spec$na_rm),
        all = all(as.logical(col), na.rm = spec$na_rm),
        median = median(col, na.rm = spec$na_rm),
        n_distinct = length(unique(col[!is.na(col)])))
    }
    if (is.null(key_names) || length(key_names) == 0) {
      results <- list()
      for (i in seq_along(agg_specs)) {
        results[[agg_specs[[i]]$name]] <- .eval_agg(agg_specs[[i]], df)
      }
      return(as.data.frame(results, stringsAsFactors = FALSE))
    } else {
      split_idx <- interaction(df[key_names], drop = TRUE)
      pieces <- split(df, split_idx, drop = TRUE)
      result_list <- lapply(pieces, function(chunk) {
        row <- chunk[1, key_names, drop = FALSE]
        for (i in seq_along(agg_specs)) {
          row[[agg_specs[[i]]$name]] <- .eval_agg(agg_specs[[i]], chunk)
        }
        row
      })
      result <- do.call(rbind, result_list)
      rownames(result) <- NULL
      return(result)
    }
  }

  # Remove .r_fallback flags before passing to C
  agg_specs <- lapply(agg_specs, function(s) { s$.r_fallback <- NULL; s })

  new_xptr <- .Call(C_group_agg_node, node$.node, key_names, agg_specs)

  # Determine residual grouping
  if (is.null(.groups)) .groups <- "drop_last"
  result_groups <- switch(.groups,
    drop_last = if (length(key_names) > 1) key_names[-length(key_names)] else NULL,
    drop = NULL,
    keep = key_names,
    stop(sprintf(".groups must be 'drop_last', 'drop', or 'keep', got '%s'", .groups))
  )

  structure(list(.node = new_xptr, .path = .data$.path,
                 .groups = result_groups), class = "vectra_node")
}

#' @rdname summarise
#' @export
summarize <- summarise

#' Remove grouping from a vectra query
#'
#' @param x A `vectra_node` object.
#' @param ... Ignored.
#'
#' @return An ungrouped `vectra_node`.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> group_by(cyl) |> ungroup()
#' unlink(f)
#'
#' @export
ungroup <- function(x, ...) {
  UseMethod("ungroup")
}

#' @export
ungroup.vectra_node <- function(x, ...) {
  structure(list(.node = x$.node, .path = x$.path), class = "vectra_node")
}

#' Count observations by group
#'
#' @param x A `vectra_node` object.
#' @param ... Grouping columns (unquoted).
#' @param wt Column to weight by (unquoted). If `NULL`, counts rows.
#' @param sort If `TRUE`, sort output in descending order of `n`.
#' @param name Name of the count column (default `"n"`).
#'
#' @return A `vectra_node` with group columns and a count column.
#'
#' @details
#' Equivalent to `group_by(...) |> summarise(n = n())`. When `wt` is
#' provided, uses `sum(wt)` instead of `n()`. When `sort = TRUE`, results
#' are sorted in descending order of the count column.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' tbl(f) |> count(cyl) |> collect()
#' unlink(f)
#'
#' @export
count <- function(x, ..., wt = NULL, sort = FALSE, name = NULL) {
  UseMethod("count")
}

#' @export
count.vectra_node <- function(x, ..., wt = NULL, sort = FALSE, name = NULL) {
  if (!is.logical(sort) || length(sort) != 1 || is.na(sort))
    stop(sprintf("sort must be TRUE or FALSE, got %s", deparse(sort)))
  if (!is.null(name) && (!is.character(name) || length(name) != 1))
    stop(sprintf("name must be NULL or a single string, got %s of length %d", class(name)[1], length(name)))
  grp_exprs <- eval(substitute(alist(...)))
  grp_names <- vapply(grp_exprs, as.character, character(1))
  cnt_name <- if (!is.null(name)) name else "n"
  wt_expr <- substitute(wt)

  # Build the grouped summarise
  node <- x
  if (length(grp_names) > 0) {
    node <- structure(list(.node = node$.node, .path = node$.path,
                           .groups = grp_names), class = "vectra_node")
  }

  if (is.null(wt_expr) || identical(wt_expr, quote(NULL))) {
    agg_specs <- list(list(name = cnt_name, kind = "n", col = NULL, na_rm = FALSE))
  } else {
    wt_name <- as.character(wt_expr)
    agg_specs <- list(list(name = cnt_name, kind = "sum", col = wt_name, na_rm = FALSE))
  }

  new_xptr <- .Call(C_group_agg_node, node$.node, grp_names, agg_specs)
  if (sort) {
    sort_xptr <- .Call(C_sort_node, new_xptr, cnt_name, TRUE)
    return(structure(list(.node = sort_xptr, .path = node$.path), class = "vectra_node"))
  }
  structure(list(.node = new_xptr, .path = node$.path), class = "vectra_node")
}

#' @rdname count
#' @export
tally <- function(x, wt = NULL, sort = FALSE, name = NULL) {
  UseMethod("tally")
}

#' @export
tally.vectra_node <- function(x, wt = NULL, sort = FALSE, name = NULL) {
  if (!is.logical(sort) || length(sort) != 1 || is.na(sort))
    stop(sprintf("sort must be TRUE or FALSE, got %s", deparse(sort)))
  if (!is.null(name) && (!is.character(name) || length(name) != 1))
    stop(sprintf("name must be NULL or a single string, got %s of length %d", class(name)[1], length(name)))
  cnt_name <- if (!is.null(name)) name else "n"
  wt_expr <- substitute(wt)
  key_names <- if (!is.null(x$.groups)) x$.groups else character(0)

  if (is.null(wt_expr) || identical(wt_expr, quote(NULL))) {
    agg_specs <- list(list(name = cnt_name, kind = "n", col = NULL, na_rm = FALSE))
  } else {
    wt_name <- as.character(wt_expr)
    agg_specs <- list(list(name = cnt_name, kind = "sum", col = wt_name, na_rm = FALSE))
  }

  new_xptr <- .Call(C_group_agg_node, x$.node, key_names, agg_specs)
  if (sort) {
    sort_xptr <- .Call(C_sort_node, new_xptr, cnt_name, TRUE)
    return(structure(list(.node = sort_xptr, .path = x$.path), class = "vectra_node"))
  }
  structure(list(.node = new_xptr, .path = x$.path), class = "vectra_node")
}

# Parse an aggregation expression like sum(x), mean(y, na.rm = TRUE), n()
# Supports nested expressions: mean(x + y) auto-inserts a hidden mutate column.
parse_agg_expr <- function(expr, output_name) {
  if (!is.call(expr))
    stop(sprintf("summarise expression '%s' must be a function call", output_name))

  fn_call <- expr[[1]]
  # Accept namespace-qualified calls like vectra::n() or dplyr:::sum()
  if (is.call(fn_call) && length(fn_call) == 3L && is.name(fn_call[[1L]]) &&
      as.character(fn_call[[1L]]) %in% c("::", ":::")) {
    fn <- as.character(fn_call[[3L]])
  } else if (is.name(fn_call)) {
    fn <- as.character(fn_call)
  } else {
    stop(sprintf(
      "summarise expression '%s' must call a named aggregation function, got %s",
      output_name, deparse(fn_call)))
  }
  valid_aggs <- c("n", "sum", "mean", "min", "max", "sd", "var", "first", "last",
                   "any", "all", "median", "n_distinct")
  if (!fn %in% valid_aggs)
    stop(sprintf("unknown aggregation function: %s. Use one of: %s",
                 fn, paste(valid_aggs, collapse = ", ")))

  if (fn == "n") {
    return(list(name = output_name, kind = "n", col = NULL, na_rm = FALSE))
  }

  # Extract column argument
  col_arg <- expr[[2]]

  # Check for na.rm argument
  na_rm <- FALSE
  if (length(expr) >= 3) {
    arg_names <- names(expr)
    if (!is.null(arg_names)) {
      idx <- match("na.rm", arg_names)
      if (!is.na(idx)) {
        na_rm <- isTRUE(eval(expr[[idx]]))
      }
    }
  }

  # median and n_distinct are now native C aggregations
  if (fn == "median") {
    col_name <- if (is.name(col_arg)) as.character(col_arg) else NULL
    if (is.null(col_name))
      stop("median() requires a simple column reference, not an expression")
    return(list(name = output_name, kind = "median", col = col_name,
                na_rm = na_rm))
  }

  if (fn == "n_distinct") {
    col_name <- if (is.name(col_arg)) as.character(col_arg) else NULL
    if (is.null(col_name))
      stop("n_distinct() requires a simple column reference, not an expression")
    return(list(name = output_name, kind = "n_distinct", col = col_name,
                na_rm = FALSE))
  }

  if (is.name(col_arg)) {
    # Simple column reference
    col_name <- as.character(col_arg)
    return(list(name = output_name, kind = fn, col = col_name, na_rm = na_rm))
  }

  # Nested expression: e.g. mean(x + y) or sum(x * 2)
  # Generate a temp column name and return the inner expression for mutate
  tmp_name <- paste0(".vectra_tmp_", output_name)
  list(name = output_name, kind = fn, col = tmp_name, na_rm = na_rm,
       .nested_expr = col_arg, .nested_col = tmp_name)
}

# --- NA poisoning: without na.rm, any NA in group poisons the result ---

test_that("sum without na.rm returns NA when group has NAs", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, NA, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(s = sum(x)) |> collect()
  expect_true(is.na(result$s[result$g == "a"]))
  expect_equal(result$s[result$g == "b"], 3)
})

test_that("mean without na.rm returns NA when group has NAs", {
  df <- data.frame(g = c("a", "a", "b", "b"),
                   x = c(10.0, NA, 3.0, 7.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(m = mean(x)) |> collect()
  expect_true(is.na(result$m[result$g == "a"]))
  expect_equal(result$m[result$g == "b"], 5)
})

test_that("min/max without na.rm returns NA when group has NAs", {
  df <- data.frame(g = c("a", "a"), x = c(5.0, NA), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  r <- tbl(f) |> group_by(g) |> summarise(lo = min(x), hi = max(x)) |> collect()
  expect_true(is.na(r$lo))
  expect_true(is.na(r$hi))
})

# --- na.rm = TRUE on all-NA groups (R-matching semantics) ---

test_that("sum(na.rm=TRUE) on all-NA group returns 0", {
  df <- data.frame(g = c("a", "a"), x = c(NA_real_, NA_real_),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    group_by(g) |>
    summarise(s = sum(x, na.rm = TRUE)) |> collect()
  expect_equal(result$s, 0)
})

test_that("mean(na.rm=TRUE) on all-NA group returns NaN", {
  df <- data.frame(g = c("a", "a"), x = c(NA_real_, NA_real_),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    group_by(g) |>
    summarise(m = mean(x, na.rm = TRUE)) |> collect()
  expect_true(is.nan(result$m))
})

test_that("min(na.rm=TRUE) on all-NA group returns Inf", {
  df <- data.frame(g = c("a", "a"), x = c(NA_real_, NA_real_),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  expect_warning(
    result <- tbl(f) |>
      group_by(g) |>
      summarise(lo = min(x, na.rm = TRUE)) |> collect(),
    "no non-missing"
  )
  expect_equal(result$lo, Inf)
})

test_that("max(na.rm=TRUE) on all-NA group returns -Inf", {
  df <- data.frame(g = c("a", "a"), x = c(NA_real_, NA_real_),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  expect_warning(
    result <- tbl(f) |>
      group_by(g) |>
      summarise(hi = max(x, na.rm = TRUE)) |> collect(),
    "no non-missing"
  )
  expect_equal(result$hi, -Inf)
})

# --- na.rm = TRUE with mixed valid + NA values ---

test_that("sum(na.rm=TRUE) skips NAs and sums valid values", {
  df <- data.frame(g = c("a", "a", "a"),
                   x = c(1.0, NA, 3.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    group_by(g) |>
    summarise(s = sum(x, na.rm = TRUE)) |> collect()
  expect_equal(result$s, 4)
})

test_that("mean(na.rm=TRUE) computes mean of non-NA values only", {
  df <- data.frame(g = c("a", "a", "a"),
                   x = c(10.0, NA, 20.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    group_by(g) |>
    summarise(m = mean(x, na.rm = TRUE)) |> collect()
  expect_equal(result$m, 15)
})

# --- n() counts all rows including NAs ---

test_that("n() counts all rows including NA values", {
  df <- data.frame(g = c("a", "a", "a"),
                   x = c(1.0, NA, 3.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(cnt = n()) |> collect()
  expect_equal(result$cnt, 3)
})

# --- Integer aggregations ---

test_that("sum of integer column returns correct double", {
  df <- data.frame(g = c("a", "a", "b"), x = c(100L, 200L, 50L))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(s = sum(x)) |> collect()
  expect_equal(result$s[result$g == "a"], 300)
  expect_equal(result$s[result$g == "b"], 50)
})

# --- count ---

test_that("count with sort = TRUE sorts descending", {
  df <- data.frame(g = c("a", "a", "a", "b", "b", "c"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> count(g, sort = TRUE) |> collect()
  expect_equal(result$n, c(3, 2, 1))
  expect_equal(result$g, c("a", "b", "c"))
})

test_that("count with wt sums the weight column", {
  df <- data.frame(g = c("a", "a", "b"), w = c(10.0, 20.0, 5.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> count(g, wt = w) |> collect()
  expect_equal(result$n[result$g == "a"], 30)
  expect_equal(result$n[result$g == "b"], 5)
})

test_that("count with custom name and sort = TRUE works", {
  df <- data.frame(g = c("x", "x", "y"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> count(g, sort = TRUE, name = "cnt") |> collect()
  expect_equal(result$cnt, c(2, 1))
  expect_true("cnt" %in% names(result))
})

test_that("count with no groups counts total rows", {
  df <- data.frame(x = c(1.0, 2.0, 3.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> count() |> collect()
  expect_equal(result$n, 3)
})

# --- tally ---

test_that("tally with sort = TRUE sorts descending", {
  df <- data.frame(g = c("a", "a", "a", "b", "b", "c"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> tally(sort = TRUE) |> collect()
  expect_equal(result$n, c(3, 2, 1))
})

test_that("tally with wt sums the weight column", {
  df <- data.frame(g = c("a", "a", "b"), w = c(10.0, 20.0, 5.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> tally(wt = w) |> collect()
  expect_equal(result$n[result$g == "a"], 30)
  expect_equal(result$n[result$g == "b"], 5)
})

# --- Date roundtrip ---

test_that("Date columns roundtrip through write_vtr/collect", {
  df <- data.frame(d = as.Date(c("2024-01-01", "2024-06-15", NA)),
                   x = c(1, 2, 3))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  expect_s3_class(result$d, "Date")
  expect_equal(result$d, df$d)
  expect_equal(result$x, df$x)
})

# --- POSIXct roundtrip ---

test_that("POSIXct columns roundtrip through write_vtr/collect", {
  df <- data.frame(
    ts = as.POSIXct(c("2024-01-01 12:00:00", "2024-06-15 18:30:00", NA),
                    tz = "UTC"),
    x = c(1, 2, 3)
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  expect_s3_class(result$ts, "POSIXct")
  expect_equal(result$ts, df$ts)
})

# --- Factor roundtrip ---

test_that("factor columns roundtrip through write_vtr/collect", {
  df <- data.frame(g = factor(c("a", "b", "a", NA), levels = c("a", "b", "c")),
                   x = c(1, 2, 3, 4))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> collect()
  expect_s3_class(result$g, "factor")
  expect_equal(levels(result$g), c("a", "b", "c"))
  expect_equal(as.character(result$g), c("a", "b", "a", NA))
})

# --- where() in select ---

test_that("select with where(is.numeric)", {
  df <- data.frame(a = 1.0, b = "x", c = 2L, stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(where(is.numeric)) |> collect()
  expect_true(all(c("a", "c") %in% names(result)))
  expect_false("b" %in% names(result))
})

test_that("select with where(is.character)", {
  df <- data.frame(a = 1.0, b = "x", c = 2L, stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(where(is.character)) |> collect()
  expect_equal(names(result), "b")
})

# --- bind_rows coercion ---

test_that("bind_rows coerces int to double across nodes", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(x = c(1L, 2L)), f1)
  write_vtr(data.frame(x = c(3.5, 4.5)), f2)
  result <- bind_rows(tbl(f1), tbl(f2)) |> collect()
  expect_equal(nrow(result), 4)
  expect_equal(result$x, c(1, 2, 3.5, 4.5))
})

# --- across with where() ---

test_that("across with where(is.numeric) in summarise", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, 2.0, 3.0),
                   y = c("p", "q", "r"), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    group_by(g) |>
    summarise(across(where(is.numeric), sum)) |>
    collect()
  expect_equal(result$x[result$g == "a"], 3)
})

# --- Date with multi-rowgroup ---

test_that("Date roundtrip with multiple row groups", {
  dates <- as.Date("2020-01-01") + 0:99
  df <- data.frame(d = dates, x = seq_along(dates))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f, batch_size = 30)
  result <- tbl(f) |> collect()
  expect_s3_class(result$d, "Date")
  expect_equal(result$d, df$d)
})

test_that("group_by + summarise with count", {
  df <- data.frame(g = c("a", "b", "a", "b", "a"), x = 1:5,
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(cnt = n()) |> collect()
  # First-seen order: "a" first, "b" second
  expect_equal(result$g, c("a", "b"))
  expect_equal(result$cnt, c(3, 2))
})

test_that("group_by + summarise with sum", {
  df <- data.frame(g = c("a", "a", "b", "b"), x = c(1.0, 2.0, 3.0, 4.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(total = sum(x)) |> collect()
  expect_equal(result$g, c("a", "b"))
  expect_equal(result$total, c(3, 7))
})

test_that("group_by + summarise with mean", {
  df <- data.frame(g = c("a", "a", "b", "b"), x = c(10.0, 20.0, 30.0, 40.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(avg = mean(x)) |> collect()
  expect_equal(result$avg, c(15, 35))
})

test_that("group_by + summarise with min and max", {
  df <- data.frame(g = c("a", "a", "b", "b"), x = c(5.0, 1.0, 8.0, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    group_by(g) |>
    summarise(lo = min(x), hi = max(x)) |>
    collect()
  expect_equal(result$lo, c(1, 3))
  expect_equal(result$hi, c(5, 8))
})

test_that("group_by with NA key values", {
  df <- data.frame(g = c("a", NA, "a", NA), x = c(1.0, 2.0, 3.0, 4.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(total = sum(x)) |> collect()
  expect_equal(nrow(result), 2)
  # "a" group and NA group
  expect_equal(result$total[result$g == "a" & !is.na(result$g)], 4)
  expect_equal(result$total[is.na(result$g)], 6)
})

test_that("multiple grouping columns", {
  df <- data.frame(
    a = c("x", "x", "y", "y"),
    b = c(1L, 2L, 1L, 2L),
    v = c(10.0, 20.0, 30.0, 40.0)
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(a, b) |> summarise(s = sum(v)) |> collect()
  expect_equal(nrow(result), 4)
})

test_that("summarise with na.rm", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, NA, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    group_by(g) |>
    summarise(total = sum(x, na.rm = TRUE)) |>
    collect()
  expect_equal(result$total[result$g == "a"], 1)
  expect_equal(result$total[result$g == "b"], 3)
})

test_that("summarise without na.rm gives NA for all-NA group", {
  df <- data.frame(g = c("a", "a"), x = c(NA_real_, NA_real_),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    group_by(g) |>
    summarise(total = sum(x)) |>
    collect()
  expect_true(is.na(result$total))
})

test_that("string key arena survives multiple resizes", {
  # Initial arena capacity is 64. 200 unique groups forces resizes at 65 and 129.
  # Regression test for UAF in arena_ensure when string data was aliased.
  set.seed(1)
  n <- 2000
  n_groups <- 200
  df <- data.frame(
    g = sample(paste0("grp_", seq_len(n_groups)), n, replace = TRUE),
    x = rnorm(n),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(sx = sum(x), n = n()) |> collect()
  expect_equal(nrow(result), n_groups)
  expect_equal(sum(result$n), n)
  # Verify against R base
  ref <- aggregate(x ~ g, data = df, FUN = sum)
  ref <- ref[match(result$g, ref$g), ]
  expect_equal(result$sx, ref$x, tolerance = 1e-10)

  # Chain a downstream operation on the grouped result to exercise
  # post-resize hash probing with the result keys still intact.
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(f2), add = TRUE)
  write_vtr(result, f2)
  result2 <- tbl(f2) |> filter(n > 5) |> collect()
  expect_true(all(result2$n > 5))
  expect_true(nrow(result2) > 0)
})

test_that("filter then group_by then summarise", {
  df <- data.frame(
    g = c("a", "b", "a", "b", "a"),
    x = c(1.0, 2.0, 3.0, 4.0, 5.0),
    stringsAsFactors = FALSE
  )
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    filter(x > 2) |>
    group_by(g) |>
    summarise(s = sum(x)) |>
    collect()
  expect_equal(result$s[result$g == "a"], 8)
  expect_equal(result$s[result$g == "b"], 4)
})

test_that("summarise accepts namespace-qualified aggregation calls", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, 2.0, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)

  # vectra::n() and vectra::sum() should work the same as bare n() / sum()
  r1 <- tbl(f) |> vectra::summarize(n = vectra::n()) |> collect()
  expect_equal(r1$n, 3)

  r2 <- tbl(f) |> group_by(g) |>
    summarise(cnt = vectra::n(), total = vectra::sum(x)) |> collect()
  expect_equal(r2$cnt, c(2, 1))
  expect_equal(r2$total, c(3, 3))

  # Unknown namespace-qualified function still errors with a clean message
  expect_error(
    tbl(f) |> summarise(z = vectra::nope(x)) |> collect(),
    "unknown aggregation function: nope"
  )
})

# --- tidyselect helpers in select ---

test_that("select with starts_with", {
  df <- data.frame(mpg = 1, cyl = 2, disp = 3)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(starts_with("m")) |> collect()
  expect_equal(names(result), "mpg")
})

test_that("select with ends_with", {
  df <- data.frame(mpg = 1, cyl = 2, disp = 3)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(ends_with("g")) |> collect()
  expect_equal(names(result), "mpg")
})

test_that("select with contains", {
  df <- data.frame(mpg = 1, cyl = 2, disp = 3)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(contains("is")) |> collect()
  expect_equal(names(result), "disp")
})

test_that("select with everything()", {
  df <- data.frame(a = 1, b = 2, c = 3)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(everything()) |> collect()
  expect_equal(names(result), c("a", "b", "c"))
})

test_that("select with negation (deselect)", {
  df <- data.frame(a = 1, b = 2, c = 3)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(-b) |> collect()
  expect_equal(names(result), c("a", "c"))
})

test_that("select with renaming", {
  df <- data.frame(a = 1, b = 2)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(alpha = a, b) |> collect()
  expect_equal(names(result), c("alpha", "b"))
  expect_equal(result$alpha, 1)
})

test_that("select with last_col()", {
  df <- data.frame(a = 1, b = 2, c = 3)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> select(last_col()) |> collect()
  expect_equal(names(result), "c")
})

# --- across in summarise ---

test_that("across in summarise applies function to selected columns", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, 2.0, 3.0),
                   y = c(10.0, 20.0, 30.0), stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    group_by(g) |>
    summarise(across(c(x, y), mean)) |>
    collect()
  expect_equal(result$x[result$g == "a"], 1.5)
  expect_equal(result$y[result$g == "a"], 15)
})

test_that("across with named function list", {
  df <- data.frame(g = c("a", "a", "b"), x = c(1.0, 2.0, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |>
    group_by(g) |>
    summarise(across(x, list(total = sum, avg = mean))) |>
    collect()
  expect_true("x_total" %in% names(result))
  expect_true("x_avg" %in% names(result))
  expect_equal(result$x_total[result$g == "a"], 3)
  expect_equal(result$x_avg[result$g == "a"], 1.5)
})

# --- any / all ---

test_that("any aggregation works", {
  df <- data.frame(g = c("a", "a", "b", "b"),
                   x = c(TRUE, FALSE, FALSE, FALSE))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(a = any(x)) |> collect()
  expect_equal(result$a[result$g == "a"], 1)
  expect_equal(result$a[result$g == "b"], 0)
})

test_that("all aggregation works", {
  df <- data.frame(g = c("a", "a", "b", "b"),
                   x = c(TRUE, TRUE, TRUE, FALSE))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(a = all(x)) |> collect()
  expect_equal(result$a[result$g == "a"], 1)
  expect_equal(result$a[result$g == "b"], 0)
})

# --- median ---

test_that("median aggregation works", {
  df <- data.frame(g = c("a", "a", "a", "b", "b"),
                   x = c(1.0, 3.0, 2.0, 10.0, 20.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(m = median(x)) |> collect()
  expect_equal(result$m[result$g == "a"], 2)
  expect_equal(result$m[result$g == "b"], 15)
})

test_that("median with na.rm = TRUE", {
  df <- data.frame(g = c("a", "a", "a"), x = c(1.0, NA, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(m = median(x, na.rm = TRUE)) |> collect()
  expect_equal(result$m, 2)
})

# --- n_distinct ---

test_that("n_distinct counts unique values", {
  df <- data.frame(g = c("a", "a", "a", "b", "b"),
                   x = c(1.0, 1.0, 2.0, 3.0, 3.0),
                   stringsAsFactors = FALSE)
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> group_by(g) |> summarise(nd = n_distinct(x)) |> collect()
  expect_equal(result$nd[result$g == "a"], 2)
  expect_equal(result$nd[result$g == "b"], 1)
})

# --- slice ---

test_that("slice selects rows by position", {
  df <- data.frame(x = c(10.0, 20.0, 30.0, 40.0, 50.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice(2, 4)
  expect_equal(nrow(result), 2)
  expect_equal(result$x, c(20, 40))
})

test_that("slice with negative indices removes rows", {
  df <- data.frame(x = c(10.0, 20.0, 30.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> slice(-2)
  expect_equal(nrow(result), 2)
  expect_equal(result$x, c(10, 30))
})

# --- cross_join ---

test_that("cross_join produces Cartesian product", {
  f1 <- tempfile(fileext = ".vtr")
  f2 <- tempfile(fileext = ".vtr")
  on.exit(unlink(c(f1, f2)))
  write_vtr(data.frame(a = c(1, 2)), f1)
  write_vtr(data.frame(b = c("x", "y", "z"), stringsAsFactors = FALSE), f2)
  result <- cross_join(tbl(f1), tbl(f2))
  expect_equal(nrow(result), 6)
  expect_true("a" %in% names(result))
  expect_true("b" %in% names(result))
})

# --- ntile ---

test_that("ntile divides into buckets", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 4.0, 5.0, 6.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(bucket = ntile(3)) |> collect()
  expect_equal(result$bucket, c(1, 1, 2, 2, 3, 3))
})

# --- percent_rank ---

test_that("percent_rank computes relative rank", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 4.0, 5.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(pr = percent_rank(x)) |> collect()
  expect_equal(result$pr, c(0, 0.25, 0.5, 0.75, 1.0))
})

# --- cume_dist ---

test_that("cume_dist computes cumulative distribution", {
  df <- data.frame(x = c(1.0, 2.0, 3.0, 4.0, 5.0))
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(df, f)
  result <- tbl(f) |> mutate(cd = cume_dist(x)) |> collect()
  expect_equal(result$cd, c(0.2, 0.4, 0.6, 0.8, 1.0))
})

# offload.R - disk-backed streaming: the offload functor.
#
# collect_chunked() and chunk_feeder() consume a query with a bounded in-RAM
# accumulator (one pass, O(1) state). offload() lifts that into the regime where
# the state itself spills to disk. It is an endofunctor on query streams that is
# the identity on values and changes only the memory and I/O profile.
#
# One front door, two return shapes:
#   offload(x)            -> a replay-cache node (same rows as x, now on disk,
#                            replayed by a disk scan instead of a pipeline re-run)
#   offload(x, by = ...)  -> a partition: disjoint per-key shards, so a model
#                            whose rows couple within a group becomes a set of
#                            independent per-shard fits.
# The external-sort instance is the existing arrange() (already disk-backed).

.PARTITION_BUDGET <- 1e6        # rows buffered before a routing flush

# -- cost grade ---------------------------------------------------------------

new_offload_grade <- function(tier, passes, peak, io, note = NULL) {
  structure(list(tier = tier, passes = passes, peak = peak, io = io,
                 note = note), class = "offload_grade")
}

#' @export
print.offload_grade <- function(x, ...) {
  cat(sprintf("<offload grade: %s>\n", x$tier))
  cat(sprintf("  passes over data : %s\n", x$passes))
  cat(sprintf("  peak memory      : %s\n", x$peak))
  cat(sprintf("  I/O cost         : %s\n", x$io))
  if (!is.null(x$note)) cat(sprintf("  note             : %s\n", x$note))
  invisible(x)
}

.scan_grade <- function()
  new_offload_grade("streaming scan", "1 per consumption (lazy)",
                    "O(one batch)", "O(n) per pass",
                    "plain query node; re-reading re-runs the upstream pipeline")

# Internal: the grade of any stream (offloaded or plain). Surfaced through
# print() and explain(), not as a separate exported verb.
grade_of <- function(x) {
  if (inherits(x, "offload_grade")) return(x)
  g <- attr(x, ".grade", exact = TRUE)
  if (is.null(g) && is.list(x)) g <- x$.grade
  if (is.null(g)) .scan_grade() else g
}

# -- offload(): replay cache and partition ------------------------------------

.replay_node <- function(spill, durable) {
  node <- tbl(spill)
  node$.grade <- new_offload_grade(
    tier = "replay cache",
    passes = "1 to spill, then O(1) re-reads",
    peak = "O(one batch)",
    io = "O(n) to write, O(n) per replay",
    note = if (durable) "durable spill"
           else "temp spill, removed when the node is garbage-collected")
  class(node) <- c("vectra_offload", "vectra_node")
  if (!durable) {
    reg <- new.env(parent = emptyenv())
    reg$path <- spill
    reg.finalizer(reg, function(e) try(unlink(e$path), silent = TRUE),
                  onexit = TRUE)
    node$.reg <- reg
  }
  node
}

#' Spill a query to disk and stream it back (the offload functor)
#'
#' Materializes a query once to disk and returns a stream that holds the same
#' rows, so every later pass is a disk scan instead of a re-run of the upstream
#' pipeline. The materialization streams batch by batch, so peak memory stays at
#' one batch regardless of result size. This is the bridge from the bounded
#' single-pass world of [collect_chunked()] to out-of-core fits.
#'
#' With no `by`, `offload()` returns a **replay cache**: a `vectra_node` backed
#' by one `.vtr` file. Feed it to a pull-based consumer such as
#' `biglm::bigglm()` through [chunk_feeder()], which accepts an offloaded node
#' directly, so each iteratively reweighted pass reads the prepared columns from
#' disk rather than rebuilding them. Bake the selects and mutates into the query
#' you offload, and replay does no further work.
#'
#' With `by`, `offload()` returns a **partition**: the rows split into disjoint
#' shards, one per key value (discrete key) or per value range (`method =
#' "range"`, or any numeric key), written in a single streaming pass. A
#' partition prints as a list of shards and behaves like one: `length()`,
#' `names()` (the keys), `p[["key"]]` (a shard node), and `lapply(p, ...)` all
#' work. Fold it with [collect_chunked()] (supplying `combine`). The union of
#' the shards reproduces the input; row totals are checked.
#'
#' @param x A `vectra_node` to materialize.
#' @param by Optional name (string) of a partition key column. When supplied,
#'   the result is a partition rather than a single node.
#' @param n Number of buckets for `method = "range"` or `"hash"`. Ignored for a
#'   one-shard-per-value partition.
#' @param method Partition strategy: `"auto"` (default; one shard per value for
#'   a discrete key, `n` ranges for a numeric key), `"level"` (one shard per
#'   distinct value), `"range"` (`n` equal-width value ranges), or `"hash"`
#'   (`n` buckets by a stable hash of the key, co-locating each key).
#' @param path Optional file path for a durable replay-cache spill (used only
#'   when `by` is `NULL`). When `NULL` a temporary file is used and removed when
#'   the returned node is garbage-collected.
#' @param compress Compression for spill files, passed to [write_vtr()]:
#'   `"fast"` (default), `"small"`, or `"none"`.
#'
#' @return A `vectra_node` (no `by`) or a `vectra_partition` (with `by`), each
#'   carrying a cost grade shown by [print()] and [explain()].
#'
#' @seealso [chunk_feeder()] (accepts an offloaded node), [collect_chunked()]
#'   for the partitioned monoidal reduce, and [arrange()] for the external-sort
#'   instance.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#'
#' # Replay cache: same rows, now on disk.
#' s <- offload(tbl(f) |> filter(cyl > 4) |> select(mpg, wt, hp))
#' nrow(collect(s))
#'
#' # Partition by a key: a list of per-shard nodes.
#' p <- offload(tbl(f), by = "cyl")
#' names(p)
#' length(p)
#' nrow(collect(p[[1]]))
#' unlink(f)
#'
#' @export
offload <- function(x, by = NULL, n = NULL,
                    method = c("auto", "level", "range", "hash"),
                    path = NULL, compress = c("fast", "small", "none")) {
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (build one with tbl(), tbl_csv(), ...)")
  method <- match.arg(method)
  compress <- match.arg(compress)

  if (is.null(by)) {
    durable <- !is.null(path)
    spill <- if (durable) normalizePath(path, mustWork = FALSE)
             else tempfile(fileext = ".vtr")
    write_vtr(x, spill, compress = compress)
    return(.replay_node(spill, durable))
  }

  if (!is.character(by) || length(by) != 1)
    stop("`by` must be a single column name (string)")

  spill <- tempfile(fileext = ".vtr")
  on.exit(unlink(spill), add = TRUE)        # intermediate; shards are run-files
  write_vtr(x, spill, compress = compress)

  schema <- .Call(C_node_schema, tbl(spill)$.node)
  if (!(by %in% schema$name))
    stop(sprintf("column '%s' not found in the query", by))
  is_double <- identical(schema$type[match(by, schema$name)], "double")

  if (method == "auto") method <- if (is_double) "range" else "level"
  if (method %in% c("range", "hash") && is.null(n))
    n <- if (method == "range") 8L else 16L

  budget <- getOption("vectra.partition_budget", .PARTITION_BUDGET)
  spec <- .partition_spec(spill, by, method, n)
  res <- .partition_router(spill, spec$assign, budget)

  labels <- spec$order(names(res$runs))
  shards <- lapply(labels, function(lab) .concat_runs(res$runs[[lab]]))
  names(shards) <- labels
  counts <- vapply(labels, function(lab) res$counts[[lab]], numeric(1))

  total <- res$n
  if (sum(counts) != total)
    stop(sprintf("partition lost rows: %g in shards vs %g in input",
                 sum(counts), total))

  all_runs <- unlist(res$runs, use.names = FALSE)
  reg <- new.env(parent = emptyenv())
  reg$paths <- all_runs
  reg.finalizer(reg, function(e) try(unlink(e$paths), silent = TRUE),
                onexit = TRUE)

  p <- shards
  attr(p, "by") <- by
  attr(p, ".counts") <- stats::setNames(counts, labels)
  attr(p, ".reg") <- reg
  attr(p, ".grade") <- new_offload_grade(
    tier = sprintf("partition by '%s' (%s)", by, method),
    passes = "1 spill + 1 routing pass",
    peak = sprintf("O(routing budget = %g rows), or O(one shard) when collected",
                   budget),
    io = "O(n)",
    note = "localizes coupling so each shard fits in RAM")
  class(p) <- c("vectra_partition", "list")
  p
}

# Build the per-strategy row-to-label assignment and a label ordering.
.partition_spec <- function(spill, by, method, n) {
  if (method == "level") {
    return(list(
      assign = function(chunk) {
        k <- chunk[[by]]
        lab <- as.character(k); lab[is.na(k)] <- "<NA>"; lab
      },
      order = function(labs) {
        non_na <- sort(setdiff(labs, "<NA>"))
        c(non_na, intersect("<NA>", labs))
      }))
  }
  if (method == "hash") {
    return(list(
      assign = function(chunk) {
        k <- chunk[[by]]
        lab <- sprintf("hash %d", as.integer(.hash_chr(as.character(k)) %% n))
        lab[is.na(k)] <- "<NA>"; lab
      },
      order = function(labs) {
        h <- grep("^hash ", labs, value = TRUE)
        h <- h[order(as.integer(sub("^hash ", "", h)))]
        c(h, intersect("<NA>", labs))
      }))
  }
  # range
  mm <- collect_chunked(
    tbl(spill),
    function(a, ch) {
      v <- ch[[by]]; v <- v[!is.na(v)]
      if (!length(v)) return(a)
      c(min(a[1], min(v)), max(a[2], max(v)))
    },
    .init = c(Inf, -Inf))
  lo <- mm[1]; hi <- mm[2]
  if (!is.finite(lo))
    stop(sprintf("cannot range-partition: '%s' has no non-NA values", by))
  breaks <- if (lo == hi) c(lo, hi) else seq(lo, hi, length.out = n + 1L)
  nb <- length(breaks) - 1L
  range_lab <- vapply(seq_len(nb), function(i)
    sprintf("[%g, %g%s", breaks[i], breaks[i + 1L],
            if (i < nb) ")" else "]"), character(1))
  list(
    assign = function(chunk) {
      v <- chunk[[by]]
      i <- findInterval(v, breaks, rightmost.closed = TRUE, all.inside = TRUE)
      lab <- range_lab[i]; lab[is.na(v)] <- "<NA>"; lab
    },
    order = function(labs) c(intersect(range_lab, labs),
                             intersect("<NA>", labs)))
}

# Stable per-value hash of a character vector (value-stable across batches).
.hash_chr <- function(s) {
  u <- unique(s)
  hu <- vapply(u, function(x) {
    if (is.na(x)) return(0)
    cs <- utf8ToInt(x); h <- 0
    for (c in cs) h <- (h * 131 + c) %% 2147483629
    h
  }, numeric(1))
  unname(hu[match(s, u)])
}

# One streaming pass: route each batch's rows to per-label buffers, flush to a
# run-file when the buffered row count crosses the budget. Each label ends as a
# set of run-files (a lazy concat downstream). Bounded by the budget.
.partition_router <- function(spill, assign, budget) {
  st <- new.env(parent = emptyenv())
  st$buffers <- list(); st$runs <- list(); st$counts <- list()
  st$buffered <- 0; st$n <- 0

  flush_one <- function(label) {
    bufs <- st$buffers[[label]]
    if (is.null(bufs) || !length(bufs)) return(invisible())
    df <- if (length(bufs) == 1) bufs[[1]] else do.call(rbind, bufs)
    rf <- tempfile(fileext = ".vtr")
    write_vtr(df, rf)
    st$runs[[label]] <- c(st$runs[[label]], rf)
    st$counts[[label]] <- (st$counts[[label]] %||% 0) + nrow(df)
    st$buffers[[label]] <- NULL
  }
  flush_all <- function() {
    for (lab in names(st$buffers)) flush_one(lab)
    st$buffered <- 0
  }

  nxt <- .batch_cursor(tbl(spill))
  repeat {
    chunk <- nxt(); if (is.null(chunk)) break
    st$n <- st$n + nrow(chunk)
    idx <- split(seq_len(nrow(chunk)), assign(chunk))
    for (lab in names(idx))
      st$buffers[[lab]] <- c(st$buffers[[lab]],
                             list(chunk[idx[[lab]], , drop = FALSE]))
    st$buffered <- st$buffered + nrow(chunk)
    if (st$buffered >= budget) flush_all()
  }
  flush_all()
  list(runs = st$runs, counts = st$counts, n = st$n)
}

`%||%` <- function(a, b) if (is.null(a)) b else a

.concat_runs <- function(paths) {
  nodes <- lapply(paths, tbl)
  if (length(nodes) == 1) nodes[[1]] else do.call(bind_rows, nodes)
}

#' @export
print.vectra_partition <- function(x, ...) {
  by <- attr(x, "by"); counts <- attr(x, ".counts")
  cat(sprintf("<vectra partition: %d shards by '%s'>\n", length(x), by))
  for (lab in names(x))
    cat(sprintf("  %-18s %g rows\n", lab, counts[[lab]]))
  print(grade_of(x))
  invisible(x)
}

# -- monoidal reduce over a partition -----------------------------------------

#' @rdname collect_chunked
#' @export
collect_chunked.vectra_partition <- function(x, f, .init = NULL,
                                             combine = NULL,
                                             commutative = FALSE) {
  if (!is.function(f))
    stop("`f` must be a function of (acc, chunk)")
  if (is.null(combine))
    stop("folding a partition needs `combine` to merge per-shard accumulators; ",
         "pass combine = (a monoid: `.init` is its identity, `combine` its op)")
  if (!is.function(combine))
    stop("`combine` must be a function of (acc, acc)")
  if (length(x) == 0) return(.init)
  accs <- lapply(unclass(x), function(nd)
    collect_chunked(nd, f, .init = .init))
  Reduce(combine, accs)
}

# -- per-shard application ----------------------------------------------------

#' Apply a function to each shard of a partition
#'
#' Run a function once per shard of a partition (`offload(x, by = ...)`) and
#' gather the results. Each shard is read into memory as a data.frame and passed
#' to `.f` together with its key, so a model that couples rows within a group
#' becomes a set of independent per-shard fits. This is the per-group
#' counterpart to [collect_chunked()], which instead merges every shard into a
#' single accumulator.
#'
#' `group_map()` returns a named list, one element per shard keyed by the shard
#' key, and places no constraint on what `.f` returns. Use it for per-group
#' results that do not rebind into a table, such as fitted models.
#'
#' `group_modify()` expects `.f` to return a data.frame for each shard and binds
#' those frames into one. When a shard's result does not already carry the
#' partition key column, the key is added as a leading column (named after the
#' partition's `by`), so every row records the shard it came from. Use it for
#' per-group summaries that recombine into a single table.
#'
#' Each shard is materialized in full before `.f` sees it, so partition the
#' query on a key whose groups fit in memory. For a reduction that stays bounded
#' without ever holding a whole group, fold the partition with
#' [collect_chunked()] instead.
#'
#' @param .data A `vectra_partition` from [offload()] with a `by` key.
#' @param .f A function applied to each shard. It receives the shard as a
#'   data.frame and the shard key (a string) as its first two arguments; any
#'   further arguments in `...` follow. A purrr-style formula such as
#'   `~ lm(y ~ x, .x)` also works, with `.x` the shard data and `.y` the key.
#'   For `group_modify()`, `.f` must return a data.frame.
#' @param ... Additional arguments passed on to `.f`.
#'
#' @return `group_map()` returns a named list with one element per shard.
#'   `group_modify()` returns a single data.frame: the per-shard results
#'   row-bound, with the shard key restored as a column when `.f` dropped it.
#'
#' @seealso [offload()] to build a partition, and [collect_chunked()] for the
#'   partitioned monoidal reduce.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#' p <- offload(tbl(f), by = "cyl")
#'
#' # One fit per shard, returned as a named list keyed by cyl.
#' fits <- group_map(p, function(d, cyl) coef(lm(mpg ~ wt, data = d)))
#' fits
#'
#' # Per-shard summaries recombined into one table, key restored as a column.
#' group_modify(p, function(d, cyl)
#'   data.frame(n = nrow(d), mean_mpg = mean(d$mpg)))
#' unlink(f)
#'
#' @export
group_map <- function(.data, .f, ...) {
  UseMethod("group_map")
}

#' @rdname group_map
#' @export
group_map.vectra_partition <- function(.data, .f, ...) {
  .f <- rlang::as_function(.f)
  dots <- list(...)
  shards <- unclass(.data)
  Map(function(nd, key) do.call(.f, c(list(collect(nd), key), dots)),
      shards, names(shards))
}

#' @rdname group_map
#' @export
group_modify <- function(.data, .f, ...) {
  UseMethod("group_modify")
}

#' @rdname group_map
#' @export
group_modify.vectra_partition <- function(.data, .f, ...) {
  .f <- rlang::as_function(.f)
  dots <- list(...)
  by <- attr(.data, "by")
  shards <- unclass(.data)
  parts <- Map(function(nd, key) {
    res <- do.call(.f, c(list(collect(nd), key), dots))
    if (!is.data.frame(res))
      stop("`.f` must return a data.frame for each shard (group_modify)")
    if (!(by %in% names(res))) {
      keycol <- stats::setNames(
        data.frame(rep(key, nrow(res)), stringsAsFactors = FALSE), by)
      res <- cbind(keycol, res)
    }
    res
  }, shards, names(shards))
  out <- do.call(rbind, unname(parts))
  rownames(out) <- NULL
  out
}

# The offload functor and its laws. offload() must be the identity on values
# (only the memory/cost profile changes), replay must be repeatable, and a
# partitioned monoidal reduce must equal the single-pass fold. One verb: no `by`
# gives a replay-cache node, `by` gives a partition (a list of shard nodes).

test_that("offload is the identity on values", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(mtcars, f)

  full <- tbl(f) |> filter(cyl > 4) |> select(mpg, wt, hp) |> collect()
  spilled <- offload(tbl(f) |> filter(cyl > 4) |> select(mpg, wt, hp))
  expect_equal(collect(spilled), full)
  expect_s3_class(spilled, "vectra_offload")
  expect_s3_class(spilled, "vectra_node")
})

test_that("offload preserves string, factor, and Date columns", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  df <- data.frame(
    s = rep(c("alpha", "beta", "gamma"), length.out = 120),
    g = factor(rep(c("lo", "hi"), length.out = 120), levels = c("lo", "hi")),
    d = as.Date("2020-01-01") + 0:119,
    v = as.double(1:120),
    stringsAsFactors = FALSE
  )
  write_vtr(df, f, batch_size = 25)

  full <- tbl(f) |> collect()
  got  <- collect(offload(tbl(f)))
  expect_equal(got$s, full$s)
  expect_identical(levels(got$g), levels(full$g))
  expect_equal(as.character(got$g), as.character(full$g))
  expect_s3_class(got$d, "Date")
  expect_equal(got$d, full$d)
  expect_equal(got$v, full$v)
})

test_that("an offloaded node reports its replay-cache grade", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(mtcars, f)
  expect_output(print(offload(tbl(f))), "replay cache")
  expect_output(print(tbl(f)), "vectra query node")   # plain node, no grade tier
})

test_that("chunk_feeder replays an offloaded node on every pass", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(mtcars, f)

  feed <- chunk_feeder(offload(tbl(f) |> select(mpg, wt, hp)))
  drain <- function() {
    feed(reset = TRUE)
    rows <- 0L
    while (!is.null(ch <- feed(reset = FALSE))) rows <- rows + nrow(ch)
    rows
  }
  expect_equal(drain(), nrow(mtcars))
  expect_equal(drain(), nrow(mtcars))   # repeatable from disk
})

test_that("offload feeds bigglm identically to the raw query", {
  skip_if_not_installed("biglm")
  f <- tempfile(fileext = ".csv")
  on.exit(unlink(f))
  set.seed(7)
  n <- 1500
  bio1 <- rnorm(n); bio12 <- rnorm(n)
  presence <- rbinom(n, 1, plogis(-0.3 + 1.2 * bio1 - 0.8 * bio12))
  write.csv(data.frame(presence = as.double(presence), bio1 = bio1,
                       bio12 = bio12), f, row.names = FALSE)

  src_raw <- function() tbl_csv(f, batch_size = 256) |>
    select(presence, bio1, bio12)
  s <- offload(tbl_csv(f, batch_size = 256) |> select(presence, bio1, bio12))

  fit_raw <- biglm::bigglm(presence ~ bio1 + bio12, data = chunk_feeder(src_raw),
                           family = binomial(), maxit = 25)
  fit_off <- biglm::bigglm(presence ~ bio1 + bio12, data = chunk_feeder(s),
                           family = binomial(), maxit = 25)
  expect_equal(unname(coef(fit_off)), unname(coef(fit_raw)), tolerance = 1e-8)
})

test_that("offload(by=) on a discrete key is a true partition (list-like)", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  df <- data.frame(g = rep(c("a", "b", "c"), length.out = 150),
                   x = as.double(1:150), stringsAsFactors = FALSE)
  write_vtr(df, f, batch_size = 20)

  p <- offload(tbl(f), by = "g")
  expect_s3_class(p, "vectra_partition")
  expect_setequal(names(p), c("a", "b", "c"))     # names() are the keys
  expect_equal(length(p), 3L)

  rebuilt <- do.call(rbind, lapply(p, collect))   # list idioms work
  expect_equal(sort(rebuilt$x), sort(df$x))        # union reproduces input
  for (i in seq_along(p))
    expect_equal(length(unique(collect(p[[i]])$g)), 1L)
})

test_that("offload(by=) auto-range-partitions a continuous key", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  set.seed(5)
  df <- data.frame(x = rnorm(400), y = as.double(1:400))
  write_vtr(df, f, batch_size = 32)

  p <- offload(tbl(f), by = "x")                  # double -> range (auto)
  expect_gt(length(p), 1L)
  rebuilt <- do.call(rbind, lapply(p, collect))
  expect_equal(sort(rebuilt$y), sort(df$y))
})

test_that("offload(by=, method='hash') co-locates keys and loses no rows", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  df <- data.frame(g = rep(letters[1:7], length.out = 350),
                   x = as.double(1:350), stringsAsFactors = FALSE)
  write_vtr(df, f, batch_size = 16)

  p <- offload(tbl(f), by = "g", method = "hash", n = 3)
  expect_lte(length(p), 3L)
  rebuilt <- do.call(rbind, lapply(p, collect))
  expect_equal(sort(rebuilt$x), sort(df$x))
  # each distinct key lands in exactly one shard (co-location)
  shard_of <- unlist(lapply(seq_along(p), function(i)
    setNames(rep(i, length(unique(collect(p[[i]])$g))),
             unique(collect(p[[i]])$g))))
  expect_equal(length(shard_of), length(unique(df$g)))
})

test_that("multi-flush routing (tiny budget) still reproduces the input", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  df <- data.frame(g = rep(c("a", "b"), length.out = 500),
                   x = as.double(1:500), stringsAsFactors = FALSE)
  write_vtr(df, f, batch_size = 25)

  # Force many flushes so shards become multi-run concats.
  old <- options(vectra.partition_budget = 50)
  on.exit(options(old), add = TRUE)
  p <- offload(tbl(f), by = "g")
  rebuilt <- do.call(rbind, lapply(p, collect))
  expect_equal(sort(rebuilt$x), sort(df$x))
  expect_equal(sum(attr(p, ".counts")), nrow(df))
})

test_that("partitioned monoidal reduce equals the single-pass fold", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  df <- data.frame(g = rep(c("a", "b", "c", "d"), length.out = 200),
                   x = as.double(1:200), stringsAsFactors = FALSE)
  write_vtr(df, f, batch_size = 16)

  p <- offload(tbl(f), by = "g")
  streamed <- collect_chunked(p, function(acc, chunk) acc + sum(chunk$x),
                              .init = 0, combine = `+`)
  single <- collect_chunked(tbl(f), function(acc, chunk) acc + sum(chunk$x),
                            .init = 0)
  expect_equal(streamed, sum(df$x))
  expect_equal(streamed, single)
})

test_that("partitioned X'X reduce reproduces lm() (monoid law, order-free)", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  set.seed(42)
  df <- data.frame(g = rep(c("a", "b", "c"), length.out = 600),
                   wt = rnorm(600, 3, 1), hp = rnorm(600, 150, 40),
                   stringsAsFactors = FALSE)
  df$mpg <- 30 - 5 * df$wt - 0.05 * df$hp + rnorm(600, 0, 1)
  write_vtr(df, f, batch_size = 50)

  accumulate <- function(acc, chunk) {
    X <- cbind(1, chunk$wt, chunk$hp); y <- chunk$mpg
    list(XtX = acc$XtX + crossprod(X), Xty = acc$Xty + crossprod(X, y))
  }
  combine <- function(a, b) list(XtX = a$XtX + b$XtX, Xty = a$Xty + b$Xty)

  p <- offload(tbl(f) |> select(g, mpg, wt, hp), by = "g")
  acc <- collect_chunked(p, accumulate,
                         .init = list(XtX = matrix(0, 3, 3), Xty = matrix(0, 3, 1)),
                         combine = combine, commutative = TRUE)
  beta_part <- drop(solve(acc$XtX, acc$Xty))
  beta_lm <- unname(coef(lm(mpg ~ wt + hp, data = df)))
  expect_equal(beta_part, beta_lm, tolerance = 1e-8)
})

test_that("offload validates its arguments", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(mtcars, f)

  expect_error(offload(mtcars), "vectra_node")
  expect_error(offload(tbl(f), by = c("a", "b")), "single column")
  expect_error(offload(tbl(f), by = "nope"), "not found")

  expect_error(chunk_feeder(42), "function")
  expect_error(chunk_feeder(tbl(f)), "offload")          # plain node rejected
  expect_error(chunk_feeder(offload(tbl(f))), NA)        # offloaded accepted

  p <- offload(tbl(f), by = "cyl")
  expect_error(collect_chunked(p, function(a, c) a + nrow(c), .init = 0L),
               "combine")
})

# group_map / group_modify: per-shard application over a partition. group_map
# must fit each shard exactly as lm() on that subset (recovery), key the result
# list by shard, and forward extra args; group_modify must rebind per-shard
# frames and restore the key column.

test_that("group_map fits each shard and matches lm() on that subset", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  set.seed(11)
  df <- data.frame(g = rep(c("a", "b", "c"), length.out = 900),
                   x = rnorm(900), stringsAsFactors = FALSE)
  df$y <- 2 + 3 * df$x + match(df$g, c("a", "b", "c")) + rnorm(900, 0, 0.1)
  write_vtr(df, f, batch_size = 64)

  p <- offload(tbl(f), by = "g")
  fits <- group_map(p, function(d, key) coef(lm(y ~ x, data = d)))

  expect_type(fits, "list")
  expect_setequal(names(fits), c("a", "b", "c"))   # list keyed by shard
  for (g in c("a", "b", "c")) {
    ref <- coef(lm(y ~ x, data = df[df$g == g, , drop = FALSE]))
    expect_equal(unname(fits[[g]]), unname(ref), tolerance = 1e-8)
  }
})

test_that("group_map accepts a purrr-style formula and forwards extra args", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  df <- data.frame(g = rep(c("a", "b"), length.out = 200),
                   x = as.double(1:200), stringsAsFactors = FALSE)
  write_vtr(df, f, batch_size = 16)

  p <- offload(tbl(f), by = "g")
  counts <- group_map(p, ~ nrow(.x))               # .x = shard, .y = key
  expect_equal(sum(unlist(counts)), nrow(df))

  trimmed <- group_map(p, function(d, key, top) head(d, top), top = 2)
  expect_true(all(vapply(trimmed, nrow, integer(1)) == 2L))
})

test_that("group_modify recombines shard tables and restores the key column", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  set.seed(12)
  df <- data.frame(g = rep(c("a", "b", "c"), length.out = 450),
                   y = rnorm(450), stringsAsFactors = FALSE)
  write_vtr(df, f, batch_size = 32)

  p <- offload(tbl(f), by = "g")
  out <- group_modify(p, function(d, key)
    data.frame(n = nrow(d), mean_y = mean(d$y)))

  expect_s3_class(out, "data.frame")
  expect_true("g" %in% names(out))                 # by-key restored as a column
  expect_setequal(out$g, c("a", "b", "c"))
  expect_equal(sum(out$n), nrow(df))
  for (g in c("a", "b", "c"))
    expect_equal(out$mean_y[out$g == g], mean(df$y[df$g == g]),
                 tolerance = 1e-12)
})

test_that("group_modify keeps a key the result already carries (no duplicate)", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  df <- data.frame(g = rep(c("a", "b", "c"), length.out = 300),
                   x = as.double(1:300), stringsAsFactors = FALSE)
  write_vtr(df, f, batch_size = 24)

  p <- offload(tbl(f), by = "g")
  out <- group_modify(p, function(d, key) d[d$x %% 2 == 0, , drop = FALSE])

  expect_equal(sum(names(out) == "g"), 1L)         # not duplicated
  expect_setequal(unique(out$g), c("a", "b", "c"))
  expect_equal(sort(out$x), sort(df$x[df$x %% 2 == 0]))
})

test_that("group_modify requires a data.frame from .f", {
  f <- tempfile(fileext = ".vtr")
  on.exit(unlink(f))
  write_vtr(mtcars, f)

  p <- offload(tbl(f), by = "cyl")
  expect_error(group_modify(p, function(d, key) mean(d$mpg)), "data.frame")
})

# chunk.R - streaming consumption of a query one batch at a time.
#
# collect() materializes the whole result into one data.frame. These verbs
# instead hand the result to R in bounded pieces: collect_chunked() folds a
# function over the stream (vectra drives the iteration), and chunk_feeder()
# turns a query into a resettable generator that a pull-based consumer such as
# biglm::bigglm() drives itself. Both sit on the same C cursor (C_node_optimize
# once, then C_node_next_batch per piece), so there is one streaming path.

# Open a streaming cursor over a node. Optimizes the plan once, then returns a
# nullary function that yields the next non-empty chunk as a data.frame, or
# NULL at end of stream. The node is consumed: once drained it cannot be reused
# (build a fresh node to stream again).
.batch_cursor <- function(x) {
  if (!inherits(x, "vectra_node"))
    stop("`x` must be a vectra_node (build one with tbl(), tbl_csv(), ...)")
  node <- x$.node
  .Call(C_node_optimize, node)
  function() .Call(C_node_next_batch, node)
}

#' Fold a function over a query, one batch at a time
#'
#' Streams a lazy query through R in bounded pieces and reduces them with `f`,
#' instead of materializing the whole result the way [collect()] does. The
#' engine pulls one batch (a data.frame of up to a few hundred thousand rows)
#' at a time; `f` is called as `f(acc, chunk)` and its return value becomes the
#' accumulator for the next batch. Peak memory is one batch plus whatever the
#' accumulator holds, so a result far larger than RAM can be reduced to a small
#' summary in a single pass.
#'
#' This is the streaming counterpart to a fold (`Reduce()`): use it when the
#' query returns more rows than fit in memory but the *reduction* is small. A
#' running count, per-group sufficient statistics, the cross-products `X'X` and
#' `X'y` behind a linear fit, an online mean or histogram - all accumulate in
#' bounded space across the stream. When you instead need the model-fitting
#' consumer to drive the iteration (and to re-read the data on each pass, as an
#' iteratively reweighted GLM does), use [chunk_feeder()].
#'
#' @param x A `vectra_node` (from [tbl()], [tbl_csv()], [tbl_tiff()], ... and
#'   any chain of verbs). It is consumed: after `collect_chunked()` returns, the
#'   stream is drained and `x` cannot be collected again.
#' @param f A function of two arguments `function(acc, chunk)` returning the
#'   updated accumulator. `chunk` is a data.frame holding the next batch of
#'   rows.
#' @param .init Initial accumulator value. Passed to `f` with the first batch
#'   and returned unchanged if the query yields no rows. When `combine` is
#'   supplied this is also the monoid identity (the value `combine` leaves
#'   unchanged).
#' @param combine Optional function `function(acc, acc)` that merges two
#'   accumulators. Supplying it declares the reduction a monoid with `.init` as
#'   identity, which is what lets the fold run over the independent shards of a
#'   partition (`offload(x, by = ...)`) and have the partial results merged
#'   correctly. For a plain node the stream is a single sequence, so `combine`
#'   is not needed and is ignored.
#' @param commutative Logical; declare that `combine` does not depend on the
#'   order of its arguments. Lets a partitioned fold merge shards in any order
#'   (no stable sort required). Default `FALSE`.
#'
#' @return The final accumulator. For a node: `f` applied left-to-right across
#'   every batch, seeded with `.init`. For a partition: each shard folded with
#'   `f`/`.init`, then those per-shard accumulators merged with `combine`.
#'
#' @seealso [chunk_feeder()] for pull-based consumers such as `biglm::bigglm()`,
#'   [offload()] for the replay cache and the partitioned monoidal reduce,
#'   [group_map()] and [group_modify()] for per-shard application, and
#'   [collect()] to materialize the full result.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#'
#' # Row count without materializing the result.
#' collect_chunked(tbl(f), function(acc, chunk) acc + nrow(chunk), .init = 0L)
#'
#' # Accumulate the normal-equation pieces X'X and X'y for an exact OLS fit
#' # of mpg ~ wt + hp, in one streaming pass.
#' acc <- collect_chunked(
#'   tbl(f) |> select(mpg, wt, hp),
#'   function(acc, chunk) {
#'     X <- cbind(1, chunk$wt, chunk$hp)
#'     y <- chunk$mpg
#'     list(XtX = acc$XtX + crossprod(X), Xty = acc$Xty + crossprod(X, y))
#'   },
#'   .init = list(XtX = matrix(0, 3, 3), Xty = matrix(0, 3, 1))
#' )
#' solve(acc$XtX, acc$Xty)            # same as coef(lm(mpg ~ wt + hp, mtcars))
#' unlink(f)
#'
#' @export
collect_chunked <- function(x, f, .init = NULL, combine = NULL,
                            commutative = FALSE) {
  UseMethod("collect_chunked")
}

#' @rdname collect_chunked
#' @export
collect_chunked.default <- function(x, f, .init = NULL, combine = NULL,
                                    commutative = FALSE) {
  stop("`x` must be a vectra_node or a vectra_partition (build one with tbl(), ",
       "tbl_csv(), offload(by = ...), ...)")
}

#' @rdname collect_chunked
#' @export
collect_chunked.vectra_node <- function(x, f, .init = NULL, combine = NULL,
                                        commutative = FALSE) {
  if (!is.function(f))
    stop("`f` must be a function of (acc, chunk)")
  nxt <- .batch_cursor(x)
  acc <- .init
  repeat {
    chunk <- nxt()
    if (is.null(chunk)) break
    acc <- f(acc, chunk)
  }
  acc
}

#' Turn a query into a resettable chunk generator
#'
#' Wraps a query so a pull-based consumer can read it one chunk at a time and
#' re-read it from the start as many times as needed. The returned closure
#' follows the `data(reset)` protocol that `biglm::bigglm()` expects: called
#' with `reset = TRUE` it rewinds to the beginning of the data, and called with
#' `reset = FALSE` it returns the next chunk as a data.frame, or `NULL` once the
#' data is exhausted. This lets `bigglm()` fit a generalized linear model on a
#' dataset larger than RAM, streaming each iteratively reweighted pass through
#' the engine without ever holding the full design matrix.
#'
#' Because a vectra node is consumed as it streams, re-reading requires a fresh
#' node on each pass. `chunk_feeder()` accepts either form: a *factory*, a
#' function of no arguments that returns a new node each time it is called; or an
#' offloaded node from [offload()], which is backed by a file and replays from
#' disk directly. On every `reset = TRUE` a fresh stream is started, so the same
#' query is replayed on each pass.
#'
#' Prefer feeding an [offload()] of the prepared query: the pipeline (scan,
#' joins, mutate) runs once into the spill, and every reweighted pass is then a
#' disk scan of the prepared columns rather than a re-run of the pipeline.
#'
#' @param .source Either a function of no arguments returning a fresh
#'   `vectra_node` each time it is called (e.g. `function() tbl_csv("occ.csv") |>
#'   select(presence, bio1, bio12)`), or an offloaded node from [offload()].
#'   Every chunk must contain all variables the consumer's formula references.
#'
#' @return A function `function(reset = FALSE)`. With `reset = TRUE` it rewinds
#'   and returns `invisible(NULL)`; with `reset = FALSE` it returns the next
#'   chunk as a data.frame, or `NULL` at end of stream.
#'
#' @seealso [offload()] for the replay cache, and [collect_chunked()] for
#'   single-pass reductions that vectra drives.
#'
#' @examples
#' f <- tempfile(fileext = ".vtr")
#' write_vtr(mtcars, f)
#'
#' feed <- chunk_feeder(function() tbl(f) |> select(mpg, wt, hp))
#' feed(reset = TRUE)       # rewind to the start of the stream
#' first <- feed()          # first chunk as a data.frame
#' head(first)
#'
#' \donttest{
#' # Out-of-core GLM: prepare once with offload(), then bigglm() replays it.
#' if (requireNamespace("biglm", quietly = TRUE)) {
#'   s <- offload(tbl(f) |> select(mpg, wt, hp))
#'   fit <- biglm::bigglm(mpg ~ wt + hp, data = chunk_feeder(s),
#'                        family = gaussian())
#'   coef(fit)
#' }
#' }
#' unlink(f)
#'
#' @export
chunk_feeder <- function(.source) {
  factory <-
    if (is.function(.source)) {
      .source
    } else if (inherits(.source, "vectra_offload")) {
      nd <- .source; force(nd)
      function() tbl(nd$.path)        # replay the spill from disk
    } else {
      stop("`.source` must be a function returning a fresh vectra_node, or an ",
           "offloaded node from offload() (which replays from disk)")
    }
  nxt <- NULL
  function(reset = FALSE) {
    if (reset) {
      nxt <<- .batch_cursor(factory())
      return(invisible(NULL))
    }
    if (is.null(nxt)) nxt <<- .batch_cursor(factory())
    nxt()
  }
}

# P3 + P4: tdc-backed row-group container — round-trip via vtr1_tdc.
#
# Exercises src/vtr1_tdc.c. Builds a multi-column data.frame, splits
# it into row groups via the writer, opens the resulting tdc container
# and reads it back, then verifies bit-identical equality column by
# column.
#
# String columns land in P4d via tdc_decode_block_varlen; the dedicated
# round-trip cases live at the bottom of this file (basic, NA, multi-
# rowgroup, unicode, varied length).

VTR_COMPRESS_NONE  <- 0L
VTR_COMPRESS_FAST  <- 1L
VTR_COMPRESS_SMALL <- 2L

vtr_tdc_roundtrip <- function(df, rowgroup_size, comp_level,
                              annotations = NULL) {
  path <- tempfile(fileext = ".vtdc")
  on.exit(unlink(path), add = TRUE)
  .Call("C_write_vtr_tdc", path, df,
        as.integer(rowgroup_size), as.integer(comp_level),
        annotations,
        PACKAGE = "vectra")
  out <- .Call("C_read_vtr_tdc", path, PACKAGE = "vectra")
  attr(out, "row.names") <- .set_row_names(length(out[[1]]))
  class(out) <- "data.frame"
  out
}

test_that("multi-column data.frame round-trips byte-exactly across comp levels", {
  set.seed(1)
  n <- 5000L
  df <- data.frame(
    x_dbl = rnorm(n, 100, 25),
    x_int = sample.int(.Machine$integer.max, n) - 1L,
    x_lgl = sample(c(TRUE, FALSE), n, replace = TRUE),
    x_seq = as.double(seq_len(n)),
    stringsAsFactors = FALSE
  )

  for (level in c(VTR_COMPRESS_NONE, VTR_COMPRESS_FAST, VTR_COMPRESS_SMALL)) {
    for (rg in c(64L, 1024L, n)) {
      rt <- vtr_tdc_roundtrip(df, rg, level)
      expect_identical(names(rt), names(df),
                       info = sprintf("level=%d rg=%d", level, rg))
      for (col in names(df)) {
        expect_identical(rt[[col]], df[[col]],
                         info = sprintf("col=%s level=%d rg=%d",
                                        col, level, rg))
      }
    }
  }
})

test_that("single-rowgroup write matches the input exactly", {
  df <- data.frame(
    a = as.double(1:1024),
    b = 1024:1L,
    c = rep(c(TRUE, FALSE), 512),
    stringsAsFactors = FALSE
  )
  rt <- vtr_tdc_roundtrip(df, 4096L, VTR_COMPRESS_FAST)
  expect_identical(rt$a, df$a)
  expect_identical(rt$b, df$b)
  expect_identical(rt$c, df$c)
})

test_that("rowgroup size that does not divide n_rows works", {
  df <- data.frame(
    v = as.double(seq_len(2050)),
    stringsAsFactors = FALSE
  )
  rt <- vtr_tdc_roundtrip(df, 256L, VTR_COMPRESS_FAST)
  expect_identical(rt$v, df$v)
})

test_that("constant column compresses below raw and round-trips", {
  df <- data.frame(
    k = rep(3.14, 8192),
    stringsAsFactors = FALSE
  )
  path <- tempfile(fileext = ".vtdc")
  on.exit(unlink(path), add = TRUE)
  .Call("C_write_vtr_tdc", path, df, 8192L, VTR_COMPRESS_FAST, NULL,
        PACKAGE = "vectra")
  raw_size <- length(df$k) * 8
  expect_lt(file.info(path)$size, raw_size)
  rt <- .Call("C_read_vtr_tdc", path, PACKAGE = "vectra")
  expect_identical(rt$k, df$k)
})

# ----- P4a: per-column user annotations propagate end-to-end -----------------

test_that("user annotations round-trip through the schema slot", {
  df <- data.frame(
    a = as.double(1:32),
    b = 1:32L,
    c = rep(c(TRUE, FALSE), 16),
    stringsAsFactors = FALSE
  )
  ann <- c("factor|low|mid|high", NA, "units=kg")

  path <- tempfile(fileext = ".vtdc")
  on.exit(unlink(path), add = TRUE)
  .Call("C_write_vtr_tdc", path, df, 32L, VTR_COMPRESS_FAST, ann,
        PACKAGE = "vectra")

  back <- .Call("C_read_vtr_tdc_annotations", path, PACKAGE = "vectra")
  expect_identical(back, ann)

  # Empty-string annotation collapses to NA on the read side.
  ann2 <- c("", "x", "")
  .Call("C_write_vtr_tdc", path, df, 32L, VTR_COMPRESS_FAST, ann2,
        PACKAGE = "vectra")
  back2 <- .Call("C_read_vtr_tdc_annotations", path, PACKAGE = "vectra")
  expect_identical(back2, c(NA_character_, "x", NA_character_))
})

test_that("annotation roundtrip survives multi-rowgroup writes", {
  df <- data.frame(
    x = as.double(1:1000),
    y = rep(c(TRUE, FALSE), 500),
    stringsAsFactors = FALSE
  )
  ann <- c("scale=0.5", "factor|TRUE|FALSE")
  path <- tempfile(fileext = ".vtdc")
  on.exit(unlink(path), add = TRUE)
  .Call("C_write_vtr_tdc", path, df, 128L, VTR_COMPRESS_FAST, ann,
        PACKAGE = "vectra")
  back <- .Call("C_read_vtr_tdc_annotations", path, PACKAGE = "vectra")
  expect_identical(back, ann)
  # Body data still round-trips byte-exactly with annotations attached.
  rt <- .Call("C_read_vtr_tdc", path, PACKAGE = "vectra")
  expect_identical(rt$x, df$x)
  expect_identical(rt$y, df$y)
})

# ----- P4a: per-column statistics round-trip ---------------------------------

test_that("stats round-trip min/max/null_count for double/int/bool", {
  df <- data.frame(
    d = c(1.5, 2.5, 3.5, 4.5, 5.5),
    i = c(10L, 20L, 30L, 40L, 50L),
    b = c(TRUE, TRUE, FALSE, TRUE, FALSE),
    stringsAsFactors = FALSE
  )
  path <- tempfile(fileext = ".vtdc")
  on.exit(unlink(path), add = TRUE)
  .Call("C_write_vtr_tdc", path, df, 5L, VTR_COMPRESS_FAST, NULL,
        PACKAGE = "vectra")

  stats <- .Call("C_read_vtr_tdc_stats", path, PACKAGE = "vectra")
  expect_length(stats, 1L)        # one rowgroup
  expect_length(stats[[1]], 3L)   # three columns

  # double
  v <- stats[[1]][[1]]
  expect_equal(v[1], 1)             # has_stats
  expect_equal(v[2], 1.5)           # min
  expect_equal(v[3], 5.5)           # max
  expect_equal(v[4], 0)             # null_count

  # int (stored at int64 granularity in stats)
  v <- stats[[1]][[2]]
  expect_equal(v[1], 1)
  expect_equal(v[2], 10)
  expect_equal(v[3], 50)
  expect_equal(v[4], 0)

  # bool: layout is {has_false, has_true} in the {min, max} slots
  v <- stats[[1]][[3]]
  expect_equal(v[1], 1)
  expect_equal(v[2], 1)             # has_false
  expect_equal(v[3], 1)             # has_true
  expect_equal(v[4], 0)
})

test_that("stats split correctly across multiple rowgroups", {
  df <- data.frame(
    x = c(seq(1, 100), seq(201, 300)) * 1.0,  # rg1 in [1,100], rg2 in [201,300]
    stringsAsFactors = FALSE
  )
  path <- tempfile(fileext = ".vtdc")
  on.exit(unlink(path), add = TRUE)
  .Call("C_write_vtr_tdc", path, df, 100L, VTR_COMPRESS_FAST, NULL,
        PACKAGE = "vectra")

  stats <- .Call("C_read_vtr_tdc_stats", path, PACKAGE = "vectra")
  expect_length(stats, 2L)

  v1 <- stats[[1]][[1]]
  expect_equal(v1[2], 1)
  expect_equal(v1[3], 100)

  v2 <- stats[[2]][[1]]
  expect_equal(v2[2], 201)
  expect_equal(v2[3], 300)
})

test_that("constant column produces equal min and max", {
  df <- data.frame(
    k = rep(42.0, 256),
    stringsAsFactors = FALSE
  )
  path <- tempfile(fileext = ".vtdc")
  on.exit(unlink(path), add = TRUE)
  .Call("C_write_vtr_tdc", path, df, 256L, VTR_COMPRESS_FAST, NULL,
        PACKAGE = "vectra")
  stats <- .Call("C_read_vtr_tdc_stats", path, PACKAGE = "vectra")
  v <- stats[[1]][[1]]
  expect_equal(v[1], 1)
  expect_equal(v[2], 42)
  expect_equal(v[3], 42)
})

test_that("bool-only-true column reports has_false=0", {
  df <- data.frame(
    b = rep(TRUE, 128),
    stringsAsFactors = FALSE
  )
  path <- tempfile(fileext = ".vtdc")
  on.exit(unlink(path), add = TRUE)
  .Call("C_write_vtr_tdc", path, df, 128L, VTR_COMPRESS_FAST, NULL,
        PACKAGE = "vectra")
  stats <- .Call("C_read_vtr_tdc_stats", path, PACKAGE = "vectra")
  v <- stats[[1]][[1]]
  expect_equal(v[1], 1)
  expect_equal(v[2], 0)             # has_false
  expect_equal(v[3], 1)             # has_true
})

# ----- P4b: NA round-trip via validity bitmap --------------------------------

test_that("NA values round-trip for double / int / logical, single rowgroup", {
  df <- data.frame(
    d = c(1.5, NA_real_, 3.5, NA_real_, 5.5),
    i = c(NA_integer_, 20L, 30L, NA_integer_, 50L),
    b = c(TRUE, NA, FALSE, NA, TRUE),
    stringsAsFactors = FALSE
  )
  rt <- vtr_tdc_roundtrip(df, 5L, VTR_COMPRESS_FAST)
  expect_identical(rt$d, df$d)
  expect_identical(rt$i, df$i)
  expect_identical(rt$b, df$b)
})

test_that("NA values round-trip across rowgroup boundaries", {
  set.seed(42)
  n <- 1000L
  d <- rnorm(n)
  d[sample.int(n, 200)] <- NA_real_
  i <- sample.int(.Machine$integer.max, n) - 1L
  i[sample.int(n, 150)] <- NA_integer_
  b <- sample(c(TRUE, FALSE), n, replace = TRUE)
  b[sample.int(n, 100)] <- NA
  df <- data.frame(d = d, i = i, b = b, stringsAsFactors = FALSE)

  for (rg in c(64L, 128L, 333L, n)) {
    for (level in c(VTR_COMPRESS_NONE, VTR_COMPRESS_FAST,
                    VTR_COMPRESS_SMALL)) {
      rt <- vtr_tdc_roundtrip(df, rg, level)
      expect_identical(rt$d, df$d, info = sprintf("rg=%d lvl=%d", rg, level))
      expect_identical(rt$i, df$i, info = sprintf("rg=%d lvl=%d", rg, level))
      expect_identical(rt$b, df$b, info = sprintf("rg=%d lvl=%d", rg, level))
    }
  }
})

test_that("all-NA column round-trips correctly", {
  df <- data.frame(
    d = rep(NA_real_, 256),
    i = rep(NA_integer_, 256),
    b = rep(NA, 256),
    stringsAsFactors = FALSE
  )
  rt <- vtr_tdc_roundtrip(df, 64L, VTR_COMPRESS_FAST)
  expect_identical(rt$d, df$d)
  expect_identical(rt$i, df$i)
  expect_identical(rt$b, df$b)
})

test_that("NA at rowgroup boundary positions survives", {
  # NAs deliberately placed at first / last index of each rowgroup.
  rg <- 32L
  n  <- 128L
  d <- as.double(seq_len(n))
  edges <- c(1L, rg, rg + 1L, 2L * rg, 2L * rg + 1L, n)
  d[edges] <- NA_real_
  df <- data.frame(d = d, stringsAsFactors = FALSE)
  rt <- vtr_tdc_roundtrip(df, rg, VTR_COMPRESS_FAST)
  expect_identical(rt$d, df$d)
})

# ----- P4c: parallel reader correctness --------------------------------------

test_that("many-rowgroup file decodes correctly under parallel reader", {
  # 200 rowgroups exercises OpenMP scheduling and per-thread FILE* handling.
  set.seed(7)
  n <- 20000L
  df <- data.frame(
    d = rnorm(n),
    i = sample.int(.Machine$integer.max, n) - 1L,
    b = sample(c(TRUE, FALSE), n, replace = TRUE),
    stringsAsFactors = FALSE
  )
  # Salt with NAs every 73rd row across all three types.
  na_idx <- seq.int(1L, n, by = 73L)
  df$d[na_idx] <- NA_real_
  df$i[na_idx] <- NA_integer_
  df$b[na_idx] <- NA

  rt <- vtr_tdc_roundtrip(df, 100L, VTR_COMPRESS_FAST)  # 200 rowgroups
  expect_identical(rt$d, df$d)
  expect_identical(rt$i, df$i)
  expect_identical(rt$b, df$b)
})

test_that("null_count stat reflects NA presence", {
  df <- data.frame(
    d = c(1.0, NA_real_, 3.0, NA_real_, 5.0),
    i = c(NA_integer_, NA_integer_, 30L, 40L, 50L),
    b = c(TRUE, NA, NA, NA, FALSE),
    stringsAsFactors = FALSE
  )
  path <- tempfile(fileext = ".vtdc")
  on.exit(unlink(path), add = TRUE)
  .Call("C_write_vtr_tdc", path, df, 5L, VTR_COMPRESS_FAST, NULL,
        PACKAGE = "vectra")
  stats <- .Call("C_read_vtr_tdc_stats", path, PACKAGE = "vectra")
  expect_equal(stats[[1]][[1]][4], 2)  # d: 2 NAs
  expect_equal(stats[[1]][[2]][4], 2)  # i: 2 NAs
  expect_equal(stats[[1]][[3]][4], 3)  # b: 3 NAs

  # min/max ignore NAs.
  expect_equal(stats[[1]][[1]][2], 1)
  expect_equal(stats[[1]][[1]][3], 5)
  expect_equal(stats[[1]][[2]][2], 30)
  expect_equal(stats[[1]][[2]][3], 50)
})

# ---------------------------------------------------------------------------
# P4d — VEC_STRING round-trip
# ---------------------------------------------------------------------------

test_that("string column round-trips byte-exactly across rowgroup sizes", {
  s <- c("alpha", "beta", "", "gamma", "delta", "epsilon", "zeta", "eta")
  df <- data.frame(s = s, stringsAsFactors = FALSE)
  for (rg in c(1L, 2L, 3L, 8L, 64L)) {
    rt <- vtr_tdc_roundtrip(df, rg, VTR_COMPRESS_FAST)
    expect_identical(rt$s, df$s, info = sprintf("rg=%d", rg))
  }
})

test_that("string column with NA_character_ round-trips and counts nulls", {
  df <- data.frame(
    s = c("a", NA_character_, "", "longer string", NA_character_, "z"),
    stringsAsFactors = FALSE
  )
  path <- tempfile(fileext = ".vtdc")
  on.exit(unlink(path), add = TRUE)
  .Call("C_write_vtr_tdc", path, df, 6L, VTR_COMPRESS_FAST, NULL,
        PACKAGE = "vectra")
  out <- .Call("C_read_vtr_tdc", path, PACKAGE = "vectra")
  expect_identical(out$s, df$s)

  stats <- .Call("C_read_vtr_tdc_stats", path, PACKAGE = "vectra")
  expect_equal(stats[[1]][[1]][4], 2)  # 2 NA strings
})

test_that("string column survives many rowgroups (boundary stress)", {
  s <- sprintf("row-%05d", 0:999)
  df <- data.frame(s = s, stringsAsFactors = FALSE)
  rt <- vtr_tdc_roundtrip(df, 137L, VTR_COMPRESS_FAST)  # 8 rowgroups
  expect_identical(rt$s, df$s)
})

test_that("unicode and varied-length strings round-trip", {
  df <- data.frame(
    s = c("café", "naïve", "日本語", "emoji \U1F600",
          strrep("x", 1000), "", "single"),
    stringsAsFactors = FALSE
  )
  rt <- vtr_tdc_roundtrip(df, 3L, VTR_COMPRESS_FAST)
  expect_identical(rt$s, df$s)
})

test_that("string columns mix with numeric/logical columns in one container", {
  set.seed(7)
  n <- 500L
  df <- data.frame(
    name  = sprintf("item-%03d", seq_len(n)),
    value = rnorm(n),
    flag  = sample(c(TRUE, FALSE), n, replace = TRUE),
    stringsAsFactors = FALSE
  )
  rt <- vtr_tdc_roundtrip(df, 73L, VTR_COMPRESS_FAST)
  expect_identical(rt$name,  df$name)
  expect_identical(rt$value, df$value)
  expect_identical(rt$flag,  df$flag)
})

---
title: "Offloading: streaming, monoids, and out-of-core fits"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Offloading: streaming, monoids, and out-of-core fits}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
set.seed(1)
has_biglm <- requireNamespace("biglm", quietly = TRUE)
library(vectra)
```

This article works through one question: when can a computation run over data
that does not fit in memory, and at what cost. The answer turns out to be a
clean piece of algebra, and `vectra` exposes it through two pieces:
`collect_chunked()` for bounded folds, and `offload()` for spilling a query to
disk (or, with a key, splitting one into shards). Along the way the same
principle explains both what streams cheaply and what needs a sort.

## Two cost models

There are two distinct settings, and conflating them causes most of the
confusion about what is possible.

The **streaming model** assumes one pass over the data, working memory much
smaller than the data, and no ability to store or re-read the input. In this
setting some exact computations are provably impossible: computing the exact
number of distinct values, or the exact median, requires memory proportional to
the data. These are information-theoretic lower bounds, and no cleverness
removes them. The escape is to approximate (sketches such as HyperLogLog and
quantile sketches).

The **external-memory model** assumes bounded RAM and a disk to spill to, with
cost measured in input/output volume. Any
computable reduction over data that fits on disk is feasible; the lower bounds
above do not bind, because they assumed the data could not be stored. The exact
median is a sort followed by reading the middle element. The exact distinct
count is a sort followed by a scan. What varies between reductions is the I/O
cost, not whether they can be done. The only hard limit is data that exceeds the
disk.

`vectra` lives in the second model. The design question is which cost tier a
reduction lands in, and the rest of this article is about reading that tier off
the structure of the reduction.

## When a reduction streams in one pass

The cheapest tier is a single pass with a fixed-size accumulator. A reduction
belongs here exactly when it is a *fold whose accumulator is a bounded monoid*:

```
reduce(data) = combine over chunks of ( accumulate over rows )
```

with three properties carrying their weight:

* **Associativity** of `combine`. `combine(combine(a, b), c)` equals
  `combine(a, combine(b, c))`. This is what lets the data be split into chunks,
  reduced independently, and merged. Algebraically the reduction descends from
  the free monoid (sequences under concatenation): it does not depend on where
  the chunk boundaries fall.
* **Commutativity** of `combine`, when present. The result does not depend on
  the order of the chunks, so the reduction descends to the multiset and the
  pieces may be processed in any order.
* **Bounded carrier**. The accumulator size does not grow with the number of
  rows, so peak memory is one batch plus the accumulator.

`collect_chunked()` is the fold. The engine pulls one batch, applies
`f(acc, chunk)`, frees the batch, and continues, so a result far larger than RAM
reduces to a small summary in bounded memory.

```{r fold-setup}
f <- tempfile(fileext = ".vtr")
df <- data.frame(g = rep(c("a", "b", "c"), length.out = 6000),
                 wt = rnorm(6000, 3, 1), hp = rnorm(6000, 150, 40))
df$mpg <- 30 - 5 * df$wt - 0.05 * df$hp + rnorm(6000, 0, 1)
write_vtr(df, f)
```

A row count is the simplest monoid (the accumulator is an integer, `combine` is
addition):

```{r fold-count}
collect_chunked(tbl(f), function(acc, chunk) acc + nrow(chunk), .init = 0L)
```

The normal-equation pieces `X'X` and `X'y` form a monoid too: each is a matrix,
and addition is the `combine`. Accumulating them in one pass gives an exact
least-squares fit without ever holding the design matrix.

```{r fold-ols}
acc <- collect_chunked(
  tbl(f) |> select(mpg, wt, hp),
  function(acc, chunk) {
    X <- cbind(1, chunk$wt, chunk$hp); y <- chunk$mpg
    list(XtX = acc$XtX + crossprod(X), Xty = acc$Xty + crossprod(X, y))
  },
  .init = list(XtX = matrix(0, 3, 3), Xty = matrix(0, 3, 1))
)
drop(solve(acc$XtX, acc$Xty))
coef(lm(mpg ~ wt + hp, data = df))
```

The two agree because summing per-batch cross-products is associative: the fit
is a monoid homomorphism from the data to the `3 x 3` accumulator.

## The offload functor

A generalized linear model needs several passes over the data. Iteratively reweighted
least squares reweights every row by the current coefficient estimate, so the
data must be read once per iteration. `biglm::bigglm()` does exactly this, and
`chunk_feeder()` hands it a resettable stream. The catch is that re-reading a
query re-runs the whole pipeline (scan, joins, mutate) on every pass.

`offload()` removes that cost. It materializes a query once to a `.vtr` file and
returns a node that streams from the file. The returned node holds the same rows
as the input; the only change is where the bytes live and that replaying is now
a disk scan. In categorical terms it is an endofunctor on query streams that is
the identity on values and changes only the cost profile:

```{r offload-identity}
s <- offload(tbl(f) |> select(mpg, wt, hp))

identical(collect(s), collect(tbl(f) |> select(mpg, wt, hp)))
s
```

Printing an offloaded node shows its cost grade, the label for what the
stream costs. A plain node reports a streaming scan; an offloaded node reports
the replay cache. `explain()` shows the same grade alongside the plan.

Feeding the replay cache to a fit is the point. `chunk_feeder()` accepts an
offloaded node directly, so each reweighted pass reads the prepared columns from
disk rather than rebuilding them:

```{r offload-bigglm, eval = has_biglm}
s <- offload(tbl(f) |> select(mpg, wt, hp))
fit <- biglm::bigglm(mpg ~ wt + hp, data = chunk_feeder(s),
                     family = gaussian())
coef(fit)
```

```{r offload-bigglm-note, echo = FALSE, results = "asis", eval = !has_biglm}
cat("> The out-of-core GLM example needs the **biglm** package, which is not",
    "installed, so it was not run here.\n")
```

Because the spill holds exactly the modelling columns, replay does no further
work. Bake the selects and mutates into the query you offload, and the per-pass
cost is one scan of the prepared table.

## Localizing coupling by sorting and partitioning

The one-pass tier covers reductions that decompose over independent rows. Many
models do not: a mixed model couples rows within a group, a Cox model couples
each event with everyone still at risk, a model with a `scale()` or spline term
couples through a quantity estimated from all the data. The move that rescues
them is to reorder so the coupling becomes local, then handle each local piece
on its own.

Two orderings cover most cases. Sorting by a key makes a cumulative coupling
into a running aggregate; partitioning by a key makes a within-group coupling
into independent blocks.

`arrange()` is already an external merge sort: it spills sorted runs to `.vtr`
and merges them with a bounded-memory heap, so sorting larger-than-RAM data is
built in.

```{r arrange}
tbl(f) |> arrange(desc(mpg)) |> head(3) |> collect()
```

`offload(x, by = ...)` splits a query into disjoint shards on disk in a single
streaming pass: one per key value for a discrete key, or per value range
(`method = "range"`) or hash bucket (`method = "hash"`) for a continuous or
high-cardinality key. The union of the shards reproduces the input, and row
totals are checked.

```{r partition}
p <- offload(tbl(f), by = "g")
p
```

Each shard is an ordinary node that fits in memory on its own, so a per-group
model is a function run over the shards. `group_map()` reads each shard into a
data.frame, hands it to the function with its key, and returns the results
keyed by shard:

```{r partition-fits}
fits <- group_map(p, function(d, g) coef(lm(mpg ~ wt + hp, data = d)))
fits
```

When the per-shard result is itself a data.frame, `group_modify()` binds the
pieces into one table and restores the key as a column:

```{r partition-modify}
group_modify(p, function(d, g)
  data.frame(n = nrow(d), mean_mpg = mean(d$mpg)))
```

### The monoidal reduce, and the law it rests on

Partitioning also lets a single global reduction run as independent per-shard
folds whose results merge. This is where the algebra returns: merging partial
accumulators needs `combine`, and `collect_chunked()` takes it as an argument.
Supplying `combine` declares the reduction a monoid, with `.init` as its
identity.

```{r partition-reduce}
accumulate <- function(acc, chunk) {
  X <- cbind(1, chunk$wt, chunk$hp); y <- chunk$mpg
  list(XtX = acc$XtX + crossprod(X), Xty = acc$Xty + crossprod(X, y))
}
combine <- function(a, b) list(XtX = a$XtX + b$XtX, Xty = a$Xty + b$Xty)

acc <- collect_chunked(
  offload(tbl(f) |> select(g, mpg, wt, hp), by = "g"),
  accumulate,
  .init = list(XtX = matrix(0, 3, 3), Xty = matrix(0, 3, 1)),
  combine = combine, commutative = TRUE
)
drop(solve(acc$XtX, acc$Xty))
coef(lm(mpg ~ wt + hp, data = df))
```

The partitioned reduce reproduces the global fit, and that is the correctness
law made concrete. Offloading preserves the answer exactly when the reduction
factors through the quotient the spill plan quotients by. Associativity of
`combine` lets the chunk boundaries move (the reduction descends from the free
monoid). Commutativity lets the shards merge in any order (it descends to the
multiset). Here `combine` is matrix addition, which is both, so the partition
order is irrelevant and `commutative = TRUE` records that.

The declared algebra also fixes the cheapest legal plan:

| declared | what it permits | cheapest plan |
|---|---|---|
| semigroup (associative `combine`) | moving chunk boundaries | sequential spill and merge |
| monoid (identity `.init` too) | empty and parallel merges | partition, k-way merge |
| commutative monoid | reordering shards | shuffle, no stable sort |

## The cost tiers

Reading the structure of a reduction gives its tier directly:

| tier | I/O | peak memory | examples |
|---|---|---|---|
| monoid fold | `O(n)`, one pass | `O(1)` | sum, count, mean, `X'X` |
| sort or partition localizable | `O(n log n)` | `O(1)` | median, distinct, per-group fits, Cox |
| all-to-all, no structure | `O(n^2)` or more | varies | dense kernel or Gaussian-process Gram |

Every tier is feasible with a disk; the tier sets the cost. The first is
`collect_chunked()` on a plain node. The second is `arrange()` or
`offload(x, by = ...)` ahead of a fold or a per-shard fit. The third still runs,
at quadratic I/O, and is usually attacked with approximation (inducing points,
low rank).

## A note on the abstraction

It is tempting to reach for heavier type theory here, since the correctness
condition is a statement about quotients. The lighter framing is the one that
fits the engine. The equality that matters, `collect(offload(x))` equals
`collect(x)`, is a plain proposition with no higher structure, and nothing about
it is checkable inside C or R at compile time. So the contract is a declared
monoid plus property tests: the test suite asserts the identity on values,
replay equivalence, and that a partitioned reduce equals the single-pass fold.
Those tests are the working proof of the quotient law.

## Choosing a path

* A bounded summary of a large result (count, mean, sufficient statistics): fold
  it with `collect_chunked()`.
* An iterative fit (a GLM through `biglm::bigglm()`): `offload()` the prepared
  query once and feed it to `chunk_feeder()`.
* A model that couples within groups, or a global reduce you want to merge from
  pieces: `offload(x, by = ...)`, then `group_map()` for a fit per shard,
  `group_modify()` to recombine per-shard tables, or a monoidal reduce with
  `collect_chunked(..., combine = )`.
* Data that must be in key order: `arrange()`, which sorts externally.

```{r cleanup, include = FALSE}
unlink(f)
```

---
title: "String Operations and Fuzzy Matching"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{String Operations and Fuzzy Matching}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Introduction

String-heavy datasets are common in ecology. Species names arrive with typos,
observer notes contain free-text descriptions, and site labels mix upper and
lower case in ways that break joins. Cleaning and matching these strings is
often the slowest part of an analysis pipeline, both in wall-clock time and in
the number of lines of R code required.

vectra handles all string operations inside its C engine, operating directly
on raw byte buffers rather than going through R's internal `CHARSXP` string
pool. When you write `mutate(name_clean = tolower(trimws(species)))`, that
expression is serialized into an execution plan and evaluated in C as data
flows through the pipeline. No temporary R character vectors are allocated, no
intermediate copies exist, and the global string cache that R maintains for
every unique character value is bypassed entirely.

This architecture has two practical consequences. First, string operations
scale linearly with data size and do not suffer from R's O(n log n) string
interning overhead. Second, the CPU-intensive operations (regex matching,
edit-distance computation) can be parallelized with OpenMP because the C
engine owns the memory layout. For datasets with more than a thousand rows,
vectra automatically distributes regex `grepl`, `levenshtein`, `dl_dist`, and
`jaro_winkler` across available cores.

This vignette walks through every string operation vectra supports, from
basic transformations through regex pattern matching to fuzzy joins. We will
use ecological data throughout: species names with realistic errors, site
labels from field surveys, and observer notes that need cleaning before
analysis. The examples are deliberately small so that outputs fit on screen,
but every operation shown here works identically on datasets with millions of
rows, streaming through memory one row group at a time without materializing
the full result until `collect()` is called.


## Basic string functions

We will start with a small dataset of tree observations where names need
cleaning. The data has trailing spaces, inconsistent case, and a few columns
we will want to slice apart.

```{r basic-data}
library(vectra)

trees <- data.frame(
  species  = c("Quercus robur  ", "  fagus sylvatica", "ACER platanoides",
               "Betula pendula", "  Pinus SYLVESTRIS  "),
  site     = c("plot_A01", "plot_B12", "plot_A01", "plot_C03", "plot_B12"),
  observer = c("J. Smith", "A. Mueller", "J. Smith",
               "B. Novak", "A. Mueller"),
  dbh_cm   = c(45.2, 38.1, 22.7, 31.0, 55.3),
  stringsAsFactors = FALSE
)

f <- tempfile(fileext = ".vtr")
write_vtr(trees, f)
```

### Trimming whitespace

`trimws()` strips leading and trailing whitespace from every value in a
column. The C engine operates on the raw bytes, so the overhead per string is
a single pass to find the first and last non-space characters. We apply it
inside `mutate()` just as we would in base R.

```{r trimws}
tbl(f) |>
  mutate(species_clean = trimws(species)) |>
  select(species, species_clean) |>
  collect()
```

The original column is preserved alongside the cleaned version. If we no
longer need it, a subsequent `select()` drops it from the pipeline before any
downstream node sees it.

### Case conversion

`toupper()` and `tolower()` convert ASCII characters in place. These are pure
byte-level operations (A-Z mapped to a-z and vice versa), so they are
effectively free compared to the I/O cost of reading the column from disk.

```{r case}
tbl(f) |>
  mutate(species_lower = tolower(trimws(species)),
         species_upper = toupper(trimws(species))) |>
  select(species_lower, species_upper) |>
  collect()
```

Chaining `tolower(trimws(...))` creates a single expression node in the plan.
The engine evaluates the innermost function first and passes the result
directly to the outer one, with no intermediate allocation visible from R.

### String length

`nchar()` returns the byte length of each string as a double (vectra's
integer type). This is useful for filtering out suspiciously short or long
entries that might indicate data-entry errors.

```{r nchar}
tbl(f) |>
  mutate(name_len = nchar(trimws(species))) |>
  select(species, name_len) |>
  collect()
```

Because vectra operates on raw bytes, `nchar()` returns the byte length, not
the number of Unicode code points. For pure ASCII text (English letters,
digits, common punctuation), bytes and characters are the same thing. But for
multi-byte UTF-8 characters such as accented letters ("ü" = 2 bytes, "é" = 2
bytes) or CJK ideographs (3 bytes each), the reported length will be higher
than what `base::nchar()` returns in R. The same applies to `substr()`: its
start and stop positions refer to byte offsets. In ecological data, this
rarely matters because taxonomic names and site codes are overwhelmingly
ASCII, but it is worth keeping in mind if your data contains place names with
diacritics or observer names in non-Latin scripts.

### Extracting substrings

`substr(x, start, stop)` and `substring(x, first, last)` both extract a
portion of each string by position. Positions are 1-based, following R
convention. These are useful for pulling fixed-width codes out of identifier
columns.

```{r substr}
tbl(f) |>
  mutate(plot_letter = substr(site, 6, 6),
         plot_number = substr(site, 7, 8)) |>
  select(site, plot_letter, plot_number) |>
  collect()
```

The site codes follow a `plot_X##` pattern, so character 6 is the letter and
characters 7-8 are the numeric part. Both are returned as strings.

### Prefix and suffix tests

`startsWith()` and `endsWith()` return a logical column. They take a literal
string argument (not a column reference) and match it against the beginning or
end of each value. Because the comparison is a simple memcmp on the byte
prefix or suffix, it runs faster than a regex equivalent.

```{r starts-ends}
tbl(f) |>
  filter(startsWith(site, "plot_A")) |>
  select(species, site) |>
  collect()
```

```{r endswith}
tbl(f) |>
  mutate(is_mueller = endsWith(observer, "Mueller")) |>
  select(observer, is_mueller) |>
  collect()
```

`startsWith` and `endsWith` are particularly useful in `filter()` for
subsetting data by site code prefixes or file path suffixes, since the C
engine can evaluate them without compiling a regex.


## String concatenation

`paste()` and `paste0()` concatenate an arbitrary number of arguments into a
single string column. These are N-ary functions: you can pass two columns,
five columns, or a mix of columns and literal strings. The C engine's
two-pass string builder computes the total output length across all arguments
first, allocates a single buffer, and fills it in one sweep.

### paste0: no separator

`paste0()` joins its arguments with no separator, which is the common case
for building identifiers or composite keys.

```{r paste0}
tbl(f) |>
  mutate(record_id = paste0(site, "_", trimws(species))) |>
  select(record_id) |>
  collect()
```

Here we pass three arguments: the `site` column, a literal underscore, and
the trimmed species name. The engine handles mixed column-and-literal
arguments natively.

### paste: custom separator

`paste()` uses a space as the default separator, matching base R behaviour.
The `sep` parameter overrides this.

```{r paste-sep}
tbl(f) |>
  mutate(label = paste(observer, site, sep = " @ ")) |>
  select(label) |>
  collect()
```

### Multi-column paste

When building composite labels from more than two columns, passing all of
them in a single `paste()` call is cleaner than nesting multiple `paste0()`
calls.

```{r paste-multi}
tbl(f) |>
  mutate(full_label = paste(observer, site,
                            trimws(species), sep = " | ")) |>
  select(full_label) |>
  collect()
```

The engine constructs each output string in a single allocation regardless of
how many arguments are passed, so there is no performance penalty for
concatenating five columns versus two.


## Pattern matching: fixed strings

vectra's `grepl()`, `gsub()`, and `sub()` accept a `fixed` parameter that
controls whether the pattern is treated as a literal byte sequence or as a
POSIX regular expression. When `fixed = TRUE` (the default in vectra), the
engine uses a direct substring search, which avoids the cost of compiling a
regex automaton. For exact substring tests and simple replacements, this is
the right choice.

Note that vectra defaults to `fixed = TRUE`, which differs from base R where
`fixed` defaults to `FALSE`. We chose this default because in our experience
the majority of string filtering and replacement tasks in data cleaning
involve literal substrings, not regex patterns. When you need actual regex,
pass `fixed = FALSE` explicitly. This is covered in the next section.

### Filtering with grepl

`grepl(pattern, x, fixed = TRUE)` returns a boolean column that is `TRUE`
wherever the pattern appears as a literal substring. We use it here to find
all records where the observer's surname contains "Smith".

```{r grepl-fixed}
tbl(f) |>
  filter(grepl("Smith", observer, fixed = TRUE)) |>
  collect()
```

Only one row matches. The fixed-string search scans each byte buffer with a
Boyer-Moore-style skip, so it scales well even on columns with long string
values.

### Replacing with gsub and sub

`gsub(pattern, replacement, x, fixed = TRUE)` replaces every occurrence of
the pattern; `sub()` replaces only the first. Both accept `fixed = TRUE` for
literal matching and return a new string column.

A common cleaning task is standardizing separators. Suppose our site codes
sometimes use underscores and sometimes hyphens.

```{r gsub-fixed}
messy_sites <- data.frame(
  site = c("plot_A01", "plot-B12", "plot_A01", "plot-C03", "plot_B12"),
  stringsAsFactors = FALSE
)
f2 <- tempfile(fileext = ".vtr")
write_vtr(messy_sites, f2)

tbl(f2) |>
  mutate(site_clean = gsub("-", "_", site, fixed = TRUE)) |>
  collect()
```

Every hyphen becomes an underscore. Since `fixed = TRUE` is the default,
writing `gsub("-", "_", site)` without the explicit argument produces the
same result.

### sub: first match only

`sub()` replaces only the first occurrence. This is useful when a pattern
might appear multiple times and we want to change just the leading instance.

```{r sub-fixed}
notes <- data.frame(
  note = c("tree dead, bark loose, dead branches",
           "alive, healthy canopy",
           "dead standing, no bark"),
  stringsAsFactors = FALSE
)
f3 <- tempfile(fileext = ".vtr")
write_vtr(notes, f3)

tbl(f3) |>
  mutate(note_edited = sub("dead", "DEAD", note, fixed = TRUE)) |>
  select(note, note_edited) |>
  collect()
```

In the first row, only the first "dead" is uppercased; the second "dead" in
"dead branches" remains unchanged. `gsub` would have replaced both.


## Pattern matching: regex

When literal substring matching is not enough, vectra supports full POSIX
regular expressions by passing `fixed = FALSE`. The C engine compiles each
regex pattern once per batch using `regcomp()` and applies it across all rows.
For batches with more than 1000 rows, the matching loop is parallelized with
OpenMP, with each thread compiling its own copy of the regex for thread
safety.

The regex dialect is POSIX Extended Regular Expressions (ERE), the same
dialect used by `grep -E` on Unix systems. This means character classes like
`[A-Z]`, quantifiers `+`, `*`, `?`, alternation with `|`, and grouping with
`()` all work as expected. Backreferences in replacement strings (`\\1`,
`\\2`) refer to captured groups. Perl-specific extensions like lookaheads,
lookbehinds, and non-greedy quantifiers (`*?`) are not supported; this is a
deliberate constraint that keeps the regex engine simple and predictable.

### Regex filtering with grepl

We will build a dataset with species names that have various formatting
issues. The regex `^[A-Z][a-z]+ [a-z]+$` matches properly formatted binomial
names (capital genus, lowercase epithet, single space).

```{r grepl-regex}
species_list <- data.frame(
  name = c("Quercus robur", "fagus sylvatica", "ACER PLATANOIDES",
           "Betula pendula", "Pinus  sylvestris", "Tilia cordata"),
  stringsAsFactors = FALSE
)
f4 <- tempfile(fileext = ".vtr")
write_vtr(species_list, f4)

tbl(f4) |>
  mutate(valid_format = grepl("^[A-Z][a-z]+ [a-z]+$",
                              name, fixed = FALSE)) |>
  collect()
```

Three names fail: "fagus sylvatica" (lowercase genus), "ACER PLATANOIDES"
(all caps), and "Pinus  sylvestris" (double space). This kind of validation
pass is a useful first step before attempting joins against a reference
taxonomy.

### Regex replacement with gsub

`gsub()` with `fixed = FALSE` supports POSIX regex in the pattern and
backreferences in the replacement string. Backreferences `\\1`, `\\2`, etc.
refer to captured groups in the pattern.

Suppose we want to reformat species names from "Genus epithet" to
"G. epithet" for compact display labels.

```{r gsub-regex}
tbl(f4) |>
  mutate(short_name = gsub("^([A-Za-z])[a-z]+ ([a-z]+)$",
                           "\\1. \\2",
                           name, fixed = FALSE)) |>
  select(name, short_name) |>
  collect()
```

The pattern captures the first letter of the genus as group `\\1` and the
entire epithet as group `\\2`. Rows that do not match the pattern are returned
unchanged, which is consistent with base R's `gsub` behaviour.

### sub with regex: first match only

`sub()` with regex works the same way but replaces only the first match. This
is handy for stripping a leading prefix that varies in length.

```{r sub-regex}
records <- data.frame(
  code = c("ID:001-Quercus", "ID:042-Fagus",
           "ID:007-Betula", "ID:113-Pinus"),
  stringsAsFactors = FALSE
)
f5 <- tempfile(fileext = ".vtr")
write_vtr(records, f5)

tbl(f5) |>
  mutate(genus = sub("^ID:[0-9]+-", "", code, fixed = FALSE)) |>
  select(code, genus) |>
  collect()
```

The regex `^ID:[0-9]+-` matches the literal "ID:" prefix, one or more digits,
and the trailing hyphen. Replacing it with an empty string leaves just the
genus name.

### str_extract: capturing structured parts

`str_extract(x, pattern)` returns the first match of the regex pattern, or
`NA` if no match is found. When the pattern contains a capture group (round
brackets), `str_extract` returns the contents of that group rather than the
full match.

This is particularly useful for pulling structured components out of
semi-formatted strings. Here we extract numeric plot identifiers from site
codes.

```{r str-extract}
sites <- data.frame(
  site_code = c("Forest_Plot_042", "Meadow_Transect_007",
                "Forest_Plot_113", "Wetland_Quad_019"),
  stringsAsFactors = FALSE
)
f6 <- tempfile(fileext = ".vtr")
write_vtr(sites, f6)

tbl(f6) |>
  mutate(plot_num = str_extract(site_code, "([0-9]+)")) |>
  select(site_code, plot_num) |>
  collect()
```

The capture group `([0-9]+)` matches one or more digits. Since there is a
group, `str_extract` returns the group content. Without the parentheses, it
would return the full match, which in this case is the same.

A more interesting example extracts the habitat type from the site code. The
pattern uses a group to capture everything before the first underscore.

```{r str-extract-habitat}
tbl(f6) |>
  mutate(habitat = str_extract(site_code, "^([A-Za-z]+)_")) |>
  select(site_code, habitat) |>
  collect()
```

The group captures "Forest", "Meadow", "Forest", "Wetland" while the
trailing underscore is part of the match but not part of the group. This
approach is cleaner than combining `substr` with a computed position.


## Fuzzy string matching

Exact string matching breaks down when data comes from multiple sources with
inconsistent spelling. A field botanist writes "Qurecus robur" (transposed
letters), a database export has "Fagus silvatica" (single-letter substitution),
and a citizen-science platform records "quercus ROBUR" (case mismatch). Fuzzy
string matching quantifies how different two strings are and lets us match
records that are close enough.

vectra provides three families of distance functions, all usable inside
`mutate()` and `filter()`. Each takes two string arguments (columns or
literals) and returns a numeric distance or similarity score. The raw variants
(`levenshtein`, `dl_dist`) return integer edit counts; the normalized variants
(`levenshtein_norm`, `dl_dist_norm`) divide by the length of the longer
string to produce a value between 0 and 1; and `jaro_winkler` returns a
similarity score where 1 means identical.

### Levenshtein distance

`levenshtein(x, y)` counts the minimum number of single-character edits
(insertions, deletions, substitutions) needed to transform `x` into `y`. The
result is a non-negative integer. `levenshtein_norm(x, y)` divides by the
length of the longer string, producing a value between 0 (identical) and 1
(completely different).

```{r levenshtein-data}
typos <- data.frame(
  field_name = c("Qurecus robur", "Fagus silvatica",
                 "Acer platanodes", "Betula pendula",
                 "Pinus sylvestrs"),
  ref_name   = c("Quercus robur", "Fagus sylvatica",
                 "Acer platanoides", "Betula pendula",
                 "Pinus sylvestris"),
  stringsAsFactors = FALSE
)
f7 <- tempfile(fileext = ".vtr")
write_vtr(typos, f7)

tbl(f7) |>
  mutate(lev = levenshtein(field_name, ref_name),
         lev_norm = levenshtein_norm(field_name, ref_name)) |>
  collect()
```

"Qurecus robur" vs "Quercus robur" has a Levenshtein distance of 2 (the
transposed "re" to "er" counts as two operations: delete and insert, or two
substitutions depending on the alignment). "Betula pendula" matches exactly,
giving a distance of 0.

### Damerau-Levenshtein distance

`dl_dist(x, y)` extends Levenshtein by counting transpositions of two
adjacent characters as a single edit operation. This better reflects the
kinds of errors humans make when typing. `dl_dist_norm(x, y)` is the
normalized variant.

```{r dl-dist}
tbl(f7) |>
  mutate(dl = dl_dist(field_name, ref_name),
         dl_norm = dl_dist_norm(field_name, ref_name)) |>
  collect()
```

Compare the first row: "Qurecus" vs "Quercus" is a transposition of "r" and
"e". Levenshtein counts this as 2 edits, but Damerau-Levenshtein counts it as
1. For human-typed data, DL distance is usually the better default because
transposition errors are common.

### Jaro-Winkler similarity

`jaro_winkler(x, y)` returns a similarity score between 0 and 1, where 1
means the strings are identical. Unlike the edit-distance functions, higher
values indicate better matches. Jaro-Winkler gives extra weight to matching
prefixes, making it particularly effective for personal names and short
strings where the first few characters are usually correct.

```{r jaro-winkler}
tbl(f7) |>
  mutate(jw = jaro_winkler(field_name, ref_name)) |>
  collect()
```

The scores are all above 0.9 for these examples because the errors are small
relative to string length. Jaro-Winkler is less informative for long strings
with small differences (everything scores above 0.95) but excels at
distinguishing between short strings like surnames.

### Choosing a distance metric

The choice depends on the error profile of your data:

- **Damerau-Levenshtein** (`dl_dist` / `dl_dist_norm`): the best general
  default. Handles insertions, deletions, substitutions, and transpositions.
  Use this for species names, place names, and any free-text field where typos
  are the primary error source.

- **Levenshtein** (`levenshtein` / `levenshtein_norm`): appropriate when
  transpositions are rare or when you need compatibility with systems that
  use classical Levenshtein. The scores will be slightly more conservative
  (higher distance) than DL for transposed strings.

- **Jaro-Winkler** (`jaro_winkler`): best for short strings (under 15
  characters) and personal names. The prefix weighting helps when the first
  few characters are reliable and errors tend to occur later in the string.
  Less useful for long species names where a single-character change barely
  moves the score.

### Filtering by distance threshold

In practice, we often want to flag records that exceed some distance threshold
from a reference value. Here we filter for species entries that are within a
normalized DL distance of 0.15 from "Quercus robur".

```{r filter-fuzzy}
field_data <- data.frame(
  species = c("Qurecus robur", "Quercus robor", "Fagus sylvatica",
              "Quercus robur", "Quercis rubur", "Betula pendula"),
  plot    = c("A1", "A2", "B1", "A3", "B2", "C1"),
  stringsAsFactors = FALSE
)
f8 <- tempfile(fileext = ".vtr")
write_vtr(field_data, f8)

tbl(f8) |>
  mutate(dist = dl_dist_norm(species, "Quercus robur")) |>
  filter(dist < 0.15) |>
  collect()
```

Only entries close to "Quercus robur" survive the filter. "Fagus sylvatica"
and "Betula pendula" are too distant, and "Quercis rubur" sits right at the
boundary (whether it passes depends on the exact normalized distance).

### Ranking by similarity

When matching against multiple candidates, sorting by distance gives us the
best matches first.

```{r rank-fuzzy}
tbl(f8) |>
  mutate(sim = jaro_winkler(species, "Quercus robur")) |>
  arrange(desc(sim)) |>
  collect()
```

The exact match scores 1.0, and the candidates with small typos rank next.
This pattern is useful for building a shortlist of probable matches to review
manually.


## Fuzzy joins

Fuzzy matching within `mutate()` works when we already have paired columns to
compare. More often, we have two separate tables and need to find which rows
in a messy dataset correspond to which rows in a clean reference. This is
where `fuzzy_join()` comes in.

### Basic fuzzy join

`fuzzy_join(x, y, by, method, max_dist)` joins two vectra tables using
approximate string matching. The `by` argument is a named character vector of
length 1 specifying which columns to compare. The result includes all columns
from both sides plus a `fuzzy_dist` column with the normalized distance of
each match.

```{r fuzzy-join-basic}
ref <- data.frame(
  canonical = c("Quercus robur", "Fagus sylvatica",
                "Acer platanoides", "Betula pendula",
                "Pinus sylvestris", "Tilia cordata"),
  family    = c("Fagaceae", "Fagaceae", "Sapindaceae",
                "Betulaceae", "Pinaceae", "Malvaceae"),
  stringsAsFactors = FALSE
)
f_ref <- tempfile(fileext = ".vtr")
write_vtr(ref, f_ref)

messy <- data.frame(
  field_species = c("Qurecus robur", "Fagus silvatica",
                    "Acer platanodes", "Betla pendula",
                    "Pinis sylvestris"),
  count         = c(12L, 7L, 3L, 15L, 9L),
  stringsAsFactors = FALSE
)
f_messy <- tempfile(fileext = ".vtr")
write_vtr(messy, f_messy)

fuzzy_join(
  tbl(f_messy), tbl(f_ref),
  by = c("field_species" = "canonical"),
  method = "dl",
  max_dist = 0.25
) |> collect()
```

Each messy field name finds its closest match in the reference table. The
`fuzzy_dist` column shows how far apart they are, which is useful for
flagging matches that might need manual verification (say, anything above
0.15).

### Methods: dl, levenshtein, jw

The `method` parameter selects the distance algorithm. The three options are
`"dl"` (Damerau-Levenshtein, the default), `"levenshtein"`, and `"jw"`
(Jaro-Winkler). When using `"jw"`, `max_dist` is interpreted as the maximum
*dissimilarity* (i.e., `1 - jaro_winkler_score`), so `max_dist = 0.2` keeps
matches with a Jaro-Winkler score of 0.8 or higher.

```{r fuzzy-join-jw}
fuzzy_join(
  tbl(f_messy), tbl(f_ref),
  by = c("field_species" = "canonical"),
  method = "jw",
  max_dist = 0.15
) |> collect()
```

Jaro-Winkler produces a tighter match set here because it penalizes
differences in the prefix more heavily.

### Blocking for performance

Without blocking, `fuzzy_join` compares every probe row against every
build row. For large tables this is quadratic. The `block_by` parameter
specifies a column for exact-match blocking: only pairs that share the same
blocking key are compared.

In a taxonomic context, the natural blocking key is the genus. We split the
genus from both sides and block on it, so "Qurecus robur" is only compared
against reference rows where the genus starts with "Q" (or more precisely,
where the blocking column matches exactly).

```{r fuzzy-join-blocked}
ref_blocked <- data.frame(
  genus     = c("Quercus", "Fagus", "Acer",
                "Betula", "Pinus", "Tilia"),
  canonical = c("Quercus robur", "Fagus sylvatica",
                "Acer platanoides", "Betula pendula",
                "Pinus sylvestris", "Tilia cordata"),
  family    = c("Fagaceae", "Fagaceae", "Sapindaceae",
                "Betulaceae", "Pinaceae", "Malvaceae"),
  stringsAsFactors = FALSE
)
f_ref2 <- tempfile(fileext = ".vtr")
write_vtr(ref_blocked, f_ref2)

messy_blocked <- data.frame(
  genus_field   = c("Quercus", "Fagus", "Acer",
                    "Betula", "Pinus"),
  field_species = c("Qurecus robur", "Fagus silvatica",
                    "Acer platanodes", "Betla pendula",
                    "Pinis sylvestris"),
  count         = c(12L, 7L, 3L, 15L, 9L),
  stringsAsFactors = FALSE
)
f_messy2 <- tempfile(fileext = ".vtr")
write_vtr(messy_blocked, f_messy2)

fuzzy_join(
  tbl(f_messy2), tbl(f_ref2),
  by = c("field_species" = "canonical"),
  method = "dl",
  max_dist = 0.25,
  block_by = c("genus_field" = "genus")
) |> collect()
```

With blocking, each probe row is compared against at most one reference row
(the one sharing its genus) instead of all six. On real datasets with
thousands of reference species, this reduces the number of distance
computations by orders of magnitude.

### Thread control

The `n_threads` parameter controls how many OpenMP threads the distance
computations use. The default is 4, which is a reasonable starting point. On
a workstation with many cores, increasing it to 8 or 16 can help for large
joins. On a laptop where other processes need CPU time, reducing it to 2 is
reasonable.

```{r fuzzy-join-threads, eval = FALSE}
fuzzy_join(
  tbl(f_messy), tbl(f_ref),
  by = c("field_species" = "canonical"),
  method = "dl",
  max_dist = 0.25,
  n_threads = 8L
) |> collect()
```


## Block lookups

`fuzzy_join()` is a streaming operation: the probe side flows through the
pipeline and the build side is materialized into a hash table internally. But
sometimes we need to look up the same reference repeatedly with different
query sets, and rebuilding the hash table each time is wasteful.

`materialize()` converts a vectra node into a persistent in-memory block. Once
materialized, we can probe it with `block_lookup()` (exact matching) or
`block_fuzzy_lookup()` (fuzzy matching) as many times as needed without
re-reading the source data.

### Exact lookups

`block_lookup(block, column, keys)` performs a hash-based exact match on a
string column. The hash index is built lazily on the first call and cached for
subsequent calls.

```{r block-lookup}
blk <- tbl(f_ref) |>
  select(canonical, family) |>
  materialize()

block_lookup(blk, "canonical",
             c("Quercus robur", "Betula pendula"))
```

The result includes a `query_idx` column that maps each result row back to
the input key vector (1-based). Queries that match multiple rows in the block
return all matches.

### Case-insensitive lookups

Field data often has inconsistent capitalization. The `ci = TRUE` parameter
performs case-insensitive matching without requiring a separate lowercased
column in the reference.

```{r block-lookup-ci}
block_lookup(blk, "canonical",
             c("quercus robur", "BETULA PENDULA"),
             ci = TRUE)
```

Both queries match despite the case mismatch. Internally, the engine builds a
separate case-folded hash index on first use and caches it alongside the
case-sensitive one.

### Fuzzy lookups

`block_fuzzy_lookup()` computes string distances between query keys and every
row in the block (or, with blocking, only rows sharing a blocking key). The
interface mirrors `fuzzy_join()` but operates on a materialized block.

```{r block-fuzzy}
blk2 <- tbl(f_ref2) |>
  select(genus, canonical, family) |>
  materialize()

block_fuzzy_lookup(
  blk2, "canonical",
  keys = c("Qurecus robur", "Fagus silvatica"),
  method = "dl",
  max_dist = 0.2,
  block_col = "genus",
  block_keys = c("Quercus", "Fagus"),
  n_threads = 2L
)
```

The `block_col` and `block_keys` arguments work together: the first query
("Qurecus robur") is only compared against rows where `genus == "Quercus"`,
and the second query against rows where `genus == "Fagus"`. This is the
same blocking logic as `fuzzy_join()` but applied to a materialized block
that can be reused across multiple lookup calls without re-reading disk.


## Performance

String operations in vectra fall into three performance tiers. Understanding
which tier an operation belongs to helps predict where the bottleneck will be
in a pipeline.

### Tier 1: byte-level operations (essentially free)

`toupper`, `tolower`, `trimws`, `nchar`, `substr`, `startsWith`, `endsWith`,
and `paste`/`paste0` all operate on raw bytes with O(n) single-pass
algorithms. Their cost is negligible compared to disk I/O. On a modern SSD,
reading a 100 MB string column takes roughly 50 ms; applying `tolower` to it
adds less than 1 ms, so in practice these operations contribute negligible
overhead compared to the I/O cost of reading the data from disk.

### Tier 2: pattern matching (depends on pattern complexity)

`grepl`, `gsub`, `sub`, and `str_extract` compile a POSIX regex (or set up a
fixed-string scanner) and apply it to every row. Fixed-string matching
(`fixed = TRUE`) is in tier 1 territory. Regex matching cost depends on the
pattern: a simple alternation like `^(A|B|C)$` is nearly as fast as fixed,
while a pattern with nested quantifiers like `(.+?)*` can be exponentially
slow on pathological inputs. For batches larger than 1000 rows, the engine
distributes the matching loop across OpenMP threads with per-thread regex
compilation, so doubling the available cores roughly halves the wall time.

The practical advice: use `fixed = TRUE` (or omit the parameter, since it
defaults to `TRUE`) whenever the pattern is a literal string. Reserve
`fixed = FALSE` for cases where you genuinely need regex features like
character classes, alternations, or capture groups.

### Tier 3: fuzzy distance computation (CPU-bound)

`levenshtein`, `dl_dist`, and `jaro_winkler` compute an O(m*n) dynamic
programming alignment for each pair of strings, where m and n are the string
lengths. For column-vs-column comparisons in `mutate()`, this is O(rows *
m * n). For fuzzy joins without blocking, it is O(probe_rows * build_rows *
m * n). Blocking reduces this to O(probe_rows * avg_partition_size * m * n),
which is why it matters so much. A loose `max_dist` does not change the
number of distance computations, but it increases the number of rows that
survive the threshold, and in a many-to-many fuzzy join this can cause a
combinatorial explosion of output rows. If both tables contain 10,000 rows
and a generous threshold matches each probe row to 50 build rows, the result
has 500,000 rows. Tightening `max_dist` or adding a `block_by` column are
the two most effective ways to keep the output manageable.

All three distance functions use OpenMP parallelization when the input exceeds
1000 rows. The work is distributed with `schedule(dynamic, 64)`, meaning
threads grab 64-row chunks as they finish, which balances load well when
string lengths vary.

### Two-pass string building

String-producing expressions (everything that outputs a new string column)
use a two-pass strategy. The first pass computes the output length for each
row without writing any bytes. The second pass allocates a single contiguous
buffer and fills it. This avoids the realloc-and-copy pattern that R uses
internally, which can fragment memory and cause O(n^2) behaviour on long
strings. For `paste()` with many arguments or `gsub` with complex
replacements, the two-pass approach keeps memory allocation linear.


## Practical guidance

### Which distance metric for which use case

For taxonomic name matching, `dl_dist_norm` with a threshold of 0.15-0.20 is
the workhorse. It catches transpositions (the most common typo type in Latin
binomials), single-letter substitutions, and dropped characters. If your
reference list is large and you can block on genus, the false-positive rate at
0.20 is low enough for automated matching with a manual review pass for
distances above 0.10.

For matching personal names (observers, collectors, authors),
`jaro_winkler` with a threshold of 0.10-0.15 (remember, this is
dissimilarity, so 0.10 means a similarity score of 0.90 or higher) is the
better choice. Names are short, and the prefix weighting in Jaro-Winkler
exploits the fact that people rarely mistype the first letter of their own
name.

For free-text fields where errors include missing words, reordered words, or
abbreviations, none of these character-level metrics work well. Consider
preprocessing with `tolower`, `trimws`, and `gsub` to normalize the strings
before computing distances.

### Choosing max_dist thresholds

Start conservative (low `max_dist`) and increase gradually. A `max_dist` of
0.10 with DL distance catches most single-character typos in strings of 10+
characters. At 0.20, you start matching strings that differ by two edits,
which increases recall but also increases false positives. Above 0.30, the
match set usually contains too much noise to be useful without manual
filtering.

For Jaro-Winkler, the scale is inverted (higher score = better match). A
`max_dist` of 0.10 (score >= 0.90) is a reasonable starting point. Below 0.05
(score >= 0.95), you catch only very minor differences.

A useful workflow is to run the fuzzy join once with a generous threshold, sort
the results by `fuzzy_dist`, and examine the matches near the boundary. This
gives an empirical sense of where true matches end and false positives begin
for your specific data.

### Handling encoding issues

vectra operates on raw bytes and assumes UTF-8 encoding. If your input data
is Latin-1 or another encoding, convert it to UTF-8 before writing to `.vtr`
format. In R, `iconv(x, from = "latin1", to = "UTF-8")` handles this. Once
the data is in UTF-8, all string operations work correctly on ASCII characters
and pass through multi-byte sequences without corruption, though the distance
metrics count bytes rather than Unicode code points for multi-byte characters.

This byte-level behaviour has a practical implication for distance
computation: if your strings contain accented characters encoded as multi-byte
UTF-8 sequences (e.g., "u" + combining accent), the edit distance between
"u" and "ü" will be larger than 1 because the engine sees multiple bytes
changing, not a single character. For Latin-alphabet ecological data (species
names, place names), this is rarely a problem because accented characters are
uncommon in standardized taxonomic databases. If your data does include
accented characters, consider normalizing them to ASCII before computing
distances (e.g., replacing "ü" with "u" via `gsub`).

### Common cleaning patterns

A typical cleaning pipeline chains several string operations before matching.
Here is a pattern we use repeatedly for species name standardization.

```{r cleaning-pattern}
raw_names <- data.frame(
  species = c("  Quercus ROBUR ", "fagus sylvatica.",
              "Acer platanoides (L.)", "BETULA   pendula"),
  stringsAsFactors = FALSE
)
f9 <- tempfile(fileext = ".vtr")
write_vtr(raw_names, f9)

tbl(f9) |>
  mutate(clean = tolower(trimws(species))) |>
  mutate(clean = gsub(".", "", clean, fixed = TRUE)) |>
  mutate(clean = gsub(" +", " ", clean, fixed = FALSE)) |>
  select(species, clean) |>
  collect()
```

The pipeline strips whitespace, lowercases everything, removes periods,
and collapses multiple spaces into one. Each `mutate()` call creates a
streaming node, so chaining them costs no extra memory. After this cleaning, the
strings are ready for exact or fuzzy matching against a standardized reference.

### Combining approaches

The most effective matching strategy layers exact and fuzzy methods. Start
with an exact join on cleaned names (catching the easy cases at near-zero
cost), then run a fuzzy join on the unmatched remainder. This avoids spending
CPU cycles on distance computations for records that match perfectly after
basic cleaning.

```{r layered-matching}
ref_clean <- data.frame(
  canonical = c("quercus robur", "fagus sylvatica",
                "acer platanoides", "betula pendula"),
  status    = c("accepted", "accepted", "accepted", "accepted"),
  stringsAsFactors = FALSE
)
f_refc <- tempfile(fileext = ".vtr")
write_vtr(ref_clean, f_refc)

cleaned <- tbl(f9) |>
  mutate(clean = tolower(trimws(species))) |>
  mutate(clean = gsub(".", "", clean, fixed = TRUE)) |>
  mutate(clean = gsub(" +", " ", clean, fixed = FALSE))

# Step 1: exact join on cleaned names
exact <- left_join(
  cleaned, tbl(f_refc),
  by = c("clean" = "canonical")
) |> collect()

exact
```

Records where `status` is `NA` did not match exactly and would go to a second
pass with `fuzzy_join()`. In this example all four match after cleaning, which
is the ideal outcome: the fuzzy join handles only the genuinely ambiguous
cases.

This layered strategy reflects how taxonomic name resolution works in
practice. The Global Biodiversity Information Facility (GBIF) backbone
taxonomy, for instance, contains over 7 million names. Running a fuzzy join
against the full list for every input record would be prohibitively expensive.
Cleaning first, exact-matching second, and fuzzy-matching only the remainder
keeps the computational cost proportional to the number of genuinely
problematic records rather than the total dataset size. In a typical
citizen-science dataset, 80-90% of records match exactly after basic
normalization, leaving only 10-20% for the distance computation stage.

---
title: "Streaming spatial operations"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Streaming spatial operations}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
has_sf <- requireNamespace("sf", quietly = TRUE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = has_sf,
  fig.width = 7,
  fig.height = 4.2,
  out.width = "100%",
  dpi = 96
)
```

This vignette is the hands-on companion to
[Out-of-core GIS with vectra](spatial.html). That article lays out the cost
model; this one runs the streaming vector verbs end to end on a real layer, so
every block below is code you can execute and check. It needs the optional
[sf](https://r-spatial.github.io/sf/) package, which supplies the geometry
engine vectra streams around.

```{r libs, message = FALSE}
library(vectra)
library(sf)
```

## The problem

A desktop GIS holds one layer in memory, runs a tool, and writes the result.
That model breaks the moment the working layer is bigger than RAM: a national
occurrence set, a continental road network, every parcel in a country. The
geometry alone runs to tens of gigabytes before any operation touches it.

vectra keeps the same toolbox but changes what stays resident. A query is pulled
through the C engine in fixed-size batches; each spatial step works on the batch
in front of it and spills the transformed batch back to disk as a fresh lazy
node. Peak memory is one batch plus whatever small layer the step compares
against, so a billion-row layer flows past a fixed budget. The verbs in this
vignette are that streaming envelope: `spatial_map()`, `spatial_filter()`,
`spatial_clip()`, `spatial_join()`, `spatial_dissolve()`, `spatial_overlay()`,
and `rasterize()`.

## How geometry travels

vectra has no geometry type. A geometry rides through the engine as hex-encoded
WKB in an ordinary string column, and the coordinate reference system is carried
on the returned node rather than written into the `.vtr` file. Topology stays
with sf and the GEOS library it links: vectra contributes the streaming, the
spill machinery, and a native fast path, not the geometry algorithms.

One streamed step is a loop. Pull a batch, decode its WKB column into an `sf`
object, run the operation, encode the result back into a WKB string column, and
append it to a run-file. When the loop finishes, the run-files become a single
lazy node you can carry on querying. The memory a run holds is

```
peak  =  one batch  +  the resident comparison layer
```

independent of how many batches stream past. The batch size for the spill is
`flush_rows` (default 500,000 transformed rows); the resident layer is the small
`y` a join or filter tests against.

To make the examples concrete, load the North Carolina counties shipped with sf
and project them to the state plane in metres. Projecting matters here: in a
planar CRS distances, areas, and buffers are Euclidean, and the recognised
operations run natively on the GEOS C API straight off the WKB column instead of
decoding each batch back to `sf`.

```{r load-nc}
nc <- st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
nc <- st_transform(nc, 32119)        # NAD83 / North Carolina, metres
crs_nc <- st_crs(nc)
nrow(nc)
```

Write the polygons to a `.vtr` file with their geometry as a hex-WKB column.
This is the shape every streamed layer takes: attribute columns plus one string
column of WKB. In a real out-of-core workflow this file would be written once
from a large source and reused; here it is 100 counties standing in for the
billion-row case.

```{r write-poly}
f_poly <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME  = nc$NAME,
  BIR74 = nc$BIR74,
  SID74 = nc$SID74,
  geometry = st_as_binary(st_geometry(nc), hex = TRUE)
), f_poly)
```

`tbl(f_poly)` now opens a lazy node over that file. Nothing is read until a
verb pulls it. Two functions bring a streamed result back to R: `collect()`
returns the underlying data.frame, geometry still a WKB string; `collect_sf()`
decodes that string and reattaches the node's CRS, giving an ordinary `sf`
object ready to plot or hand to any sf function.

## Per-feature transforms with spatial_map

The widest door is `spatial_map()`: apply any per-feature sf operation to a
streamed layer one batch at a time. Buffering, simplifying, computing centroids,
reprojecting, extracting coordinates, measuring area. The function you pass takes
one `sf` batch and returns an `sf` object, an `sfc`, or a plain data.frame. The
active geometry of the return becomes the streamed geometry.

Buffer every county centroid by 15 km. The first step builds a centroid layer
on disk, the second streams a buffer over it.

```{r map-buffer}
cent <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME = nc$NAME,
  geometry = st_as_binary(st_centroid(st_geometry(nc)), hex = TRUE)
), cent)

buffered <- tbl(cent) |>
  spatial_map(~ st_buffer(.x, 15000), crs = crs_nc)
buffered
```

`buffered` is a lazy node, not a materialized layer: the buffer has not run yet.
`collect_sf()` pulls it through and decodes the geometry.

```{r map-buffer-plot}
b_sf <- collect_sf(buffered)
plot(st_geometry(nc), border = "grey80", col = NA,
     main = "County centroids buffered by 15 km")
plot(st_geometry(b_sf), border = "#3366cc", col = "#3366cc22", add = TRUE)
```

The formula syntax (`~ st_buffer(.x, 15000)`) is rlang's: `.x` is the batch. A
named function works identically and is clearer for anything longer than one
call. The return can also be a plain data.frame rather than a geometry, which
turns the spatial step into a streamed feature-summary with the geometry
dropped. That is the way to compute per-feature area over a layer too large to
hold whole.

```{r map-area}
areas <- tbl(f_poly) |>
  spatial_map(~ data.frame(NAME = .x$NAME,
                           area_km2 = as.numeric(st_area(.x)) / 1e6),
              crs = crs_nc)
head(collect(areas))
```

Because the result is a node, it chains. A buffer feeding a filter feeding a
join is three streamed steps, each spilling to its own run-files, with the CRS
carried forward so you state it once. `flush_rows` controls the spill batch:
raise it for fewer, larger temporary files when rows are small, lower it when a
single feature's geometry is heavy. The default of 500,000 suits point and
small-polygon work.

## Smooth geometry with spatial_smooth

`spatial_smooth()` rounds the corners of every line and polygon by Chaikin
corner-cutting, one batch at a time. Each pass replaces a vertex with two points
along its adjacent edges, so sharp angles become a smooth curve; more
`iterations` give a smoother result with more vertices. It is computed directly
on the coordinates with no GEOS call, and open lines keep their endpoints.

```{r smooth}
zig <- st_linestring(rbind(c(0, 0), c(1, 1), c(2, 0), c(3, 1), c(4, 0)))
f_zig <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1L, geometry = st_as_binary(st_sfc(zig), hex = TRUE)), f_zig)

tbl(f_zig) |> spatial_smooth(iterations = 3) |> collect_sf()
unlink(f_zig)
```

## Select by location with spatial_filter

The most-used vector tool keeps the features standing in a spatial relation to
another layer: points inside a study region, parcels touching a road, patches
within a buffer of the coast. `spatial_filter()` streams the large layer and
tests each batch against a small resident layer with an sf predicate, keeping or
dropping whole features. It is the spatial counterpart to `semi_join()`: rows
are filtered, never duplicated, and no columns are added, so the output carries
the input schema unchanged.

Sample 500 points across the state and store them as plain x/y columns. Point
input as coordinate columns is the headline case and the fully sf-free path:
each point is built in C and matched against the resident layer's spatial index
with no per-batch round-trip through sf.

```{r sample-points}
set.seed(1)
pts <- st_coordinates(st_sample(st_union(nc), 500))
fp <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), fp)
```

Keep the points falling inside a five-county region in the northwest.

```{r filter-region}
region <- nc[nc$NAME %in% c("Ashe", "Alleghany", "Surry", "Wilkes", "Watauga"),
             "NAME"]

keep_xy <- tbl(fp) |>
  spatial_filter(region, coords = c("x", "y"), crs = crs_nc) |>
  collect()
nrow(keep_xy)
```

Twenty-five of the 500 points land in the region. Plot the kept points against
the dropped ones.

```{r filter-plot}
plot(st_geometry(nc), border = "grey85", col = NA, main = "Select by location")
plot(st_geometry(region), border = "#cc3344", col = "#cc334411", add = TRUE)
points(pts, pch = 16, cex = 0.5, col = "grey70")
points(keep_xy$x, keep_xy$y, pch = 16, cex = 0.6, col = "#cc3344")
```

`negate = TRUE` inverts the test, keeping the 475 points outside the region. A
different `predicate` changes the relation: `st_within`, `st_covered_by`,
`st_touches`, `st_crosses`, and the rest of the sf binary predicates are all
accepted. `st_is_within_distance` takes its radius through `dist`, in CRS units,
and keeps every feature within that distance of the layer.

```{r filter-distance}
near <- tbl(fp) |>
  spatial_filter(region, predicate = st_is_within_distance,
                 coords = c("x", "y"), crs = crs_nc, dist = 30000)
nrow(collect(near))
```

Sixty points sit within 30 km of the region, more than the 25 strictly inside.
For the recognised predicates on planar data the test runs in C off the WKB
column or the raw coordinates; an unrecognised predicate, geographic data with
spherical geometry switched on, or an extra predicate argument falls back to the
per-batch sf loop, which preserves sf's exact semantics at the cost of decoding
each batch.

## Cut geometry with spatial_clip

`spatial_filter()` keeps or drops whole features; `spatial_clip()` cuts their
geometry along a mask. Clipping intersects each batch with the mask and keeps the
part inside it, the streamed equivalent of `st_intersection()` against a fixed
boundary. `erase = TRUE` keeps the part outside instead.

```{r clip}
mask_region <- st_union(st_geometry(region))

clipped <- tbl(f_poly) |> spatial_clip(mask_region, crs = crs_nc)
c_sf <- collect_sf(clipped)
nrow(c_sf)
```

Twelve counties have area inside the region's bounding shape, and each comes back
trimmed to the part that overlaps it. Plot the clipped slivers over the full
state.

```{r clip-plot}
plot(st_geometry(nc), border = "grey85", col = NA,
     main = "Counties clipped to the region")
plot(st_geometry(c_sf), border = "#2a9d5c", col = "#2a9d5c33", add = TRUE)
```

The mask is unioned once into a single resident geometry, then each batch is cut
against it in C. As with the join and filter, the cut runs natively on the WKB
column for planar data and through `st_intersection()` / `st_difference()` for
the geographic-with-s2 and coordinate-assembled cases. Erase reverses the
keep: `tbl(f_poly) |> spatial_clip(mask_region, erase = TRUE)` returns the 95
counties with any area outside the region, each trimmed to that outside part.

## Split geometry with spatial_split

`spatial_split()` cuts each feature against a small resident `blade` layer, the
QGIS "split with lines". A polygon is divided into the faces the blade carves
out, a line into the arcs between crossings, and each piece is emitted as its own
row with the source attributes copied. A feature the blade does not cross passes
through whole. With `extract = "points"` the verb returns the crossing points
instead.

```{r split}
square <- st_polygon(list(rbind(c(0, 0), c(4, 0), c(4, 4), c(0, 4), c(0, 0))))
blade  <- st_sfc(st_linestring(rbind(c(2, -1), c(2, 5))))

f_sq <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1L, geometry = st_as_binary(st_sfc(square), hex = TRUE)), f_sq)

tbl(f_sq) |> spatial_split(blade) |> collect_sf()
unlink(f_sq)
```

## Tag features with spatial_join

To attach a layer's attributes rather than filter or cut, `spatial_join()`
streams the large left side and joins each batch against a small resident right
side. The dominant workload is tagging a huge point set with the polygon it
falls in: which county each occurrence sits in, which census tract each address
belongs to. The billion-row left stream never materializes while the polygon
layer stays in RAM.

```{r join-tag}
tagged <- tbl(fp) |>
  spatial_join(nc["NAME"], coords = c("x", "y"), crs = crs_nc)
tdf <- collect(tagged)
head(tdf[, c("id", "NAME")])
```

Every point now carries the `NAME` of its county. The default predicate is
`st_intersects` and the default is a left join, so an unmatched left row is kept
once with NA in the right columns; `left = FALSE` drops the unmatched rows for an
inner join. Colour the points by the county they were tagged with.

```{r join-plot}
tagged_sf <- st_as_sf(tdf, coords = c("x", "y"), crs = crs_nc)
plot(st_geometry(nc), border = "grey85", col = NA,
     main = "Points tagged with their county")
plot(tagged_sf["NAME"], pch = 16, cex = 0.5, add = TRUE)
```

Columns present on both sides are disambiguated with `suffix` (default
`c(".x", ".y")`), exactly as `st_join()` does. A different `join` predicate
changes the relation; `st_nearest_feature` finds the closest right feature to
each left one, which is how you snap points to the nearest road or station.

```{r join-nearest}
ncent <- st_sf(NAME = nc$NAME, geometry = st_centroid(st_geometry(nc)))
nearest <- tbl(fp) |>
  spatial_join(ncent, join = st_nearest_feature,
               coords = c("x", "y"), crs = crs_nc)
nrow(collect(nearest))
```

### When both sides are larger than RAM

The resident-`y` path assumes the right side fits in memory. When it does not,
pass `partition = grid(cellsize)` and a streamed `vectra_node` as `y`. Both
inputs are binned to a uniform grid and joined one shard at a time, so neither
side is ever fully resident. A coordinate maps to the cell

```
cell  =  floor( (coord - origin) / cellsize )
```

per axis. Each left feature is assigned to the single cell of its reference
point; each right feature is replicated to every cell its bounding box overlaps.
A left row is therefore emitted exactly once, and the result equals the resident
join for point left geometry.

```{r join-partition}
g_poly <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME = nc$NAME,
  geometry = st_as_binary(st_geometry(nc), hex = TRUE)
), g_poly)

tagged2 <- tbl(fp) |>
  spatial_join(tbl(g_poly), coords = c("x", "y"), crs = crs_nc,
               partition = grid(80000))
t2 <- collect(tagged2)
sum(!is.na(t2$NAME))
```

All 500 points are tagged, matching the resident join. The `cellsize` is the one
tuning knob: large enough that one cell's features fit in memory, small enough
that the shards stay balanced. For point-in-polygon tagging any cell larger than
the polygons works; for an extended-on-extended join choose it larger than the
left features. The nearest-feature predicate is not local to one cell, so under
`partition` a left point searches its own cell and the eight around it; set
`cellsize` at least as large as the largest expected nearest distance.

### Same answer as resident sf

Streaming changes the memory profile, not the result. The streamed join returns
exactly what `st_join()` returns on the whole layer held in memory, feature for
feature. Run both on the 500-point set and compare.

```{r join-equality}
streamed <- collect(
  tbl(fp) |> spatial_join(nc["NAME"], coords = c("x", "y"), crs = crs_nc))

resident <- st_join(
  st_as_sf(collect(tbl(fp)), coords = c("x", "y"), crs = crs_nc, remove = FALSE),
  nc["NAME"], join = st_intersects)

all.equal(streamed$NAME[order(streamed$id)],
          resident$NAME[order(resident$id)])
```

The county tags match for every point. The equality is the contract: a streamed
verb is the resident sf call run batch by batch, so the choice between them is a
memory decision, not an accuracy trade-off.

### At scale

The point of streaming shows when the layer stops fitting in memory. Scatter
200,000 points across the state's bounding box and filter them to the five-county
region. Only one batch is resident at a time, so the same code that ran on 500
points runs on 200,000 with a flat memory profile.

```{r filter-scale}
set.seed(42)
bb <- st_bbox(nc)
n_big <- 2e5
big <- data.frame(id = seq_len(n_big),
                  x = runif(n_big, bb["xmin"], bb["xmax"]),
                  y = runif(n_big, bb["ymin"], bb["ymax"]))
fbig <- tempfile(fileext = ".vtr")
write_vtr(big, fbig)

kept <- tbl(fbig) |>
  spatial_filter(region, coords = c("x", "y"), crs = crs_nc) |>
  collect()
nrow(kept)
```

A few thousand of the 200,000 land in the five counties. At a billion points the
file would be larger but the resident set would not grow: the engine still pulls
one batch, tests it in C against the region's spatial index, and moves on. That
is the whole proposition. The operation that a desktop GIS can only run on a
layer it can open is the same operation, run past a fixed memory budget.

```{r cleanup-scale, include = FALSE}
unlink(fbig)
```

## Nearest features with spatial_knn

Where `spatial_join()` attaches the single nearest feature, `spatial_knn()`
returns the `k` nearest features per streamed point, one row per pair, each with
the rank (1 is nearest) and the distance. The candidate layer `y` is held
resident while the points stream.

```{r knn}
towns <- suppressWarnings(st_centroid(st_geometry(nc)))[1:5]
towns <- st_sf(town = nc$NAME[1:5], geometry = towns)

set.seed(1)
pts <- suppressWarnings(st_coordinates(st_sample(nc, 100)))
f_pts <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), f_pts)

tbl(f_pts) |>
  spatial_knn(towns, k = 2, coords = c("x", "y"), crs = crs_nc, y_id = "town") |>
  collect() |> head()
unlink(f_pts)
```

## Merge geometries with spatial_dissolve

Dissolve unions the geometries within each group into one feature, the GIS
"Dissolve" tool: counties into states, parcels into ownership blocks. Unlike the
per-batch verbs, it needs every geometry of a group together to union them, so it
rides a different tier. The input is spilled once and routed into one disjoint
shard per group in a single bounded pass; each shard is then read back and
unioned. Peak memory is the routing budget during the pass, then one group's
geometries while that group unions. Partition on a key whose groups fit in
memory.

Split the counties into two bands by their 1974 SIDS count and merge each band
into a single feature, summing births along the way.

```{r dissolve}
nc$band <- ifelse(nc$SID74 > 5, "high", "low")
fb <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  band = nc$band, BIR74 = nc$BIR74,
  geometry = st_as_binary(st_geometry(nc), hex = TRUE)
), fb)

merged <- tbl(fb) |>
  spatial_dissolve(by = "band", crs = crs_nc,
                   .fun = list(births = function(d) sum(d$BIR74)))
m_sf <- collect_sf(merged)
m_sf
```

Two features come back, one per band, each carrying the summed `births` of its
counties. The `.fun` argument is a named list of summaries; each function takes
the group's data.frame and returns one value, and the list name becomes the
output column. Plot the two bands.

```{r dissolve-plot}
plot(m_sf["band"], main = "Counties dissolved into two SIDS bands")
```

With no `by`, the whole layer dissolves into one feature, which is the fast way
to compute a layer's outline out of core. On planar data each group is unioned
natively off the WKB column; geographic data with s2 on, or an extra `st_union()`
argument such as `is_coverage = TRUE`, unions through sf instead.

## Split overlaps apart with spatial_overlay

Dissolve merges overlapping geometry; overlay splits it. `spatial_overlay()`
takes a polygon layer that overlaps itself and cuts it along every overlap into
disjoint pieces, returning one row per piece per covering polygon. Where `k`
polygons overlap, that piece appears `k` times, each row carrying one source
polygon's attributes. This is the operation a GIS exposes as "Union (single
layer)": the overlap is retained once per contributing feature rather than
dissolved away. It answers questions like "which protected areas, designated in
which years, cover this exact patch of sea".

Three overlapping squares designated in different years make the smallest
meaningful example.

```{r overlay}
sq <- function(a, b) st_polygon(list(rbind(
  c(a, 0), c(b, 0), c(b, 1), c(a, 1), c(a, 0))))
polys <- st_sf(year = c(1990L, 2010L, 2000L),
               geometry = st_sfc(sq(0, 2), sq(1, 3), sq(1.5, 3.5)))

pieces <- collect_sf(spatial_overlay(polys))
nrow(pieces)
length(unique(pieces$piece_id))
```

The three squares decompose into five disjoint pieces, returned as nine rows
because the overlapping pieces are repeated once per covering square. The
`piece_id` column keys the pieces: rows sharing an id are the same patch of
ground seen from different source polygons. Resolve the duplication with a
grouped slice. Earliest designation year wins:

```{r overlay-resolve}
first <- spatial_overlay(polys) |>
  group_by(piece_id) |>
  slice_min(year, n = 1, with_ties = FALSE) |>
  collect_sf()
nrow(first)

plot(first["year"], main = "Overlay pieces, earliest year wins")
```

Five pieces remain, one per disjoint patch, each labelled with the earliest year
that covers it. Swapping `slice_min` for `slice_max` gives the latest; any
grouped reduction works, because the pieces are an ordinary streamed node by this
point.

Three properties make the overlay trustworthy on real data. First, correctness
is checked, not assumed: for a valid decomposition the piece areas covering an
input must sum to that input's area, and the engine verifies this invariant per
batch, warning if coverage drifts past a relative `1e-4` rather than returning
silently wrong. Second, memory is bounded by tiling: a connected cluster of many
large overlapping polygons can node into far more linework than the input, so
clusters too large for the budget are tiled over their own extent and clipped,
with each feature cleaned exactly once. Third, the snapping grid is explicit:
coordinates are snapped to a grid derived from their magnitude so the
near-duplicate boundaries that overlapping polygons share coincide, which is what
keeps the pieces disjoint instead of leaving sliver faces.

Unlike the other verbs, `spatial_overlay()` takes a resident `sf` object, not a
lazy node: building the overlap graph needs the geometries in memory. The
exploded result, typically several times larger than the input, is what streams
to disk. `mem_limit` caps the peak working set and `threads` sets the parallel
overlay width; raise both for speed on a big machine, lower `mem_limit` for
tighter memory.

### A second layer

Passing a second layer `y` nodes two layers into one planar partition and
carries the attributes of the covering `x`-record and `y`-record onto each
piece. A `how` argument selects which pieces to keep: `"intersection"` (covered
by both, the default), `"union"` (every piece of either, the absent side filled
with `NA`), `"identity"` (all of `x` split by `y`), or `"symdiff"` (pieces in
exactly one layer).

```{r overlay-two}
zones <- st_sf(zone = c("A", "B"),
               geometry = st_sfc(sq(0, 1.5), sq(1.5, 3)))

inter <- spatial_overlay(polys, zones, how = "intersection") |> collect_sf()
inter
```

Each piece now carries both its `year` (from the squares) and its `zone` (from
the zone layer). `vars_y` narrows the carried `y` columns, and a name shared
with an `x` column is disambiguated with a `.x` / `.y` suffix. A file-path `y`
is read in batches with `layer_y` / `query_y`, the same way `x` is.

## Materializing and round-tripping

Three exits bring a streamed result out. `collect()` returns the data.frame with
geometry still a WKB string, useful when the next step is a
non-spatial verb or another `.vtr` write. `collect_sf()` decodes the WKB and
reattaches the node's CRS, giving an `sf` object. `write_vtr()` on a node streams
the result straight to a new `.vtr` file without ever holding it whole, so a
multi-step spatial pipeline can land its output on disk under the same fixed
memory budget it ran in.

```{r roundtrip}
out <- tempfile(fileext = ".vtr")
tbl(fp) |>
  spatial_filter(region, coords = c("x", "y"), crs = crs_nc) |>
  write_vtr(out)
nrow(collect(tbl(out)))
```

The CRS lives on the node, not in the file. A pipeline that opens projected data,
maps, filters, and joins keeps the projection because each verb carries it
forward; you state `crs =` once at the first step that needs it, or let it
inherit from the upstream node. The `.vtr` file itself stores only the WKB bytes,
so reopening a written file and calling `collect_sf()` needs the CRS supplied
again if you want it labelled.

## The cost model

Each verb states what it holds resident, so the toolbox reads as a cost model.
Three tiers cover everything here.

| Tier | Verbs | Resident set |
|---|---|---|
| Monoid fold | `spatial_map`, `spatial_filter`, `spatial_clip`, `rasterize` | one batch + small `y` |
| Partition | `spatial_dissolve`, partitioned `spatial_join` | routing budget, then one shard |
| All-to-all | `spatial_overlay` | one overlap cluster (tiled) |

The fold tier is the cheapest: one batch at a time, no spill, memory flat across
the whole stream. The partition tier spills the input once and processes it one
group or shard at a time, so its peak is set by the largest group rather than the
whole layer. The all-to-all tier is inherently resident because the operation is
global, and the overlay bounds it by tiling overlap clusters. Operations that are
global by nature and not tiled, such as Voronoi diagrams, Delaunay triangulation,
convex hulls, and neighbour graphs, are not streamed at all: `collect_sf()` the
layer and run sf or terra on it.

A second axis is whether a step runs natively or through sf. The recognised
predicates and operations run in C straight off the WKB column when the data is
planar, which means projected, or geographic with spherical geometry off, or of
unknown CRS. Geographic data with `sf::sf_use_s2()` on keeps the spherical sf
path so the answer matches sf exactly. The native path parses the resident layer
once into a spatial index and tests each batch in C with no decode; the fallback
decodes each batch to `sf`. For the large planar workloads vectra targets, the
native path is the common case, which is why the examples project up front.

Four options tune the machinery:

- `vectra.spatial_flush` (default 500,000): rows buffered before a spill flush.

- `vectra.partition_budget`: rows held while routing a dissolve or partitioned
  join before flushing shards.
- `vectra.overlay_mem_limit` (default 2 GB): the overlay's peak working-set
  budget.
- `vectra.overlay_threads` and `vectra.spatial_threads`: worker counts for the
  overlay and the per-batch GEOS loops.

## Practical guidance

A few rules of thumb, with the numbers that drive them.

**Project before you stream.** A projected CRS puts the recognised verbs on the
native GEOS path and makes distances and areas Euclidean. Geographic data with
s2 on works but decodes every batch to sf and computes on the sphere, which is
correct but slower; reach for it only when spherical accuracy over large extents
matters more than throughput.

**Size the join grid to the data, not the machine.** For point-in-polygon
tagging set `grid(cellsize)` to anything larger than your polygons; the join is
exact for point left geometry at any such size. For an extended-on-extended join
make `cellsize` larger than the left features. For a partitioned nearest-feature
join make it at least the largest expected nearest distance, because the search
only reaches the eight neighbouring cells.

**Match `flush_rows` to feature weight.** The 500,000 default suits points and
small polygons. For heavy geometry, dense coastlines, detailed admin boundaries,
lower it so a spill batch is a workable size; for millions of tiny point rows,
raise it to cut the number of temporary files.

**Partition dissolve on a key whose groups fit.** Dissolve holds one group's
geometry while it unions. Dissolving a country into one outline is fine; grouping
by a key with a single enormous group is not, because that group must be resident
to union. Split the key finer if a group blows the budget.

**Trust the overlay's coverage warning.** If `spatial_overlay()` warns that
coverage drifted past `1e-4`, the named input features are finer than the
snapping grid or invalid after it. Pass a smaller `grid =` for fine geometry, or
inspect the `coverage_offenders` attribute on the result for the worst rows. A
clean run means the pieces reconstruct the inputs exactly.

When the verbs map to tiers and resident sets, the choice of tool is a memory
decision:

| Need | Verb | Holds resident |
|---|---|---|
| Per-feature transform | `spatial_map` | one batch |
| Keep / drop by location | `spatial_filter` | one batch + locator |
| Cut geometry to a mask | `spatial_clip` | one batch + mask |
| Tag with attributes | `spatial_join` | one batch + polygons |
| Tag, both sides huge | `spatial_join(partition=)` | one grid shard |
| Merge by group | `spatial_dissolve` | one group |
| Split self-overlaps | `spatial_overlay` | one overlap cluster |

**When not to stream.** For a layer that already fits in memory, sf is simpler
and faster: there is no gain in spilling 10,000 features through run-files when
`st_join()` runs in a blink. Streaming earns its overhead when the layer is
larger than RAM, or when a step in a longer pipeline would otherwise force a full
materialization. For genuinely global operations, Voronoi, Delaunay, hulls,
neighbour graphs, there is nothing to stream: collect the layer and run the
resident tool. The streaming envelope covers the operations whose memory can be
bounded, and the cost model names the ones it cannot.

```{r cleanup, include = FALSE}
unlink(c(f_poly, cent, fp, g_poly, fb, out))
```

## Where to go next

- [Out-of-core GIS with vectra](spatial.html) for the cost model and the raster
  half of the toolbox: `zonal()`, `focal()`, `terrain()`, `warp()`,
  `polygonize()`, `contours()`, `proximity()`.
- [Larger-than-RAM strategy](large-data.html) for the spill and partition tiers
  the spatial verbs reuse.
- [Species distribution modelling](sdm.html) for an end-to-end ecological
  workflow built on these verbs.
- The function reference for every spatial verb and its arguments.
```

---
title: "Species Distribution Models"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Species Distribution Models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
set.seed(1)
has_biglm <- requireNamespace("biglm", quietly = TRUE)
library(vectra)
```

## What vectra does for a distribution model

A species distribution model relates occurrence records to environmental
predictors. The predictors usually arrive as raster layers (climate, terrain,
land cover), and the records as point coordinates. The first step of almost
every workflow is the same: sample each raster at the occurrence coordinates to
build a table of predictor values, then fit a model to that table.

vectra covers two parts of this pipeline directly. It reads `GeoTIFF` rasters
(including tiled and `BigTIFF` layouts) and samples them at point coordinates
with `tiff_extract_points()`, reading only the strips that contain query
points, so a sparse point set over a continental raster touches a small
fraction of the file. And it consumes a query one batch at a time, so a model
matrix that exceeds memory can still be fitted: `chunk_feeder()` streams the
data through `biglm::bigglm()` one chunk per iteration.

The rest of this article assumes two things are handled before extraction:

* **Vector geometry.** vectra streams the vector verbs (`spatial_join()`,
  `spatial_filter()`, `spatial_clip()`, `spatial_overlay()`, `rasterize()`, ...)
  around `sf`, one batch at a time -- see
  [Streaming spatial operations](streaming-spatial.html). It carries no geometry
  type of its own, so the topology comes from `sf` and GEOS. This article stays
  on the raster-sampling path and assumes any range-map intersection or
  study-region buffer is prepared before extraction.
* **Reprojection.** vectra reads the coordinate reference system of a `GeoTIFF`
  with `tiff_crs()` but does not transform coordinates. Occurrence points must
  already be in the raster's CRS when they reach `tiff_extract_points()`.

## Align coordinates before extracting

`tiff_extract_points()` maps each coordinate to a pixel through the raster's
affine geotransform. It does this in the raster's own CRS, so points recorded
in a different CRS must be reprojected first. Read the raster CRS with
`tiff_crs()`, then transform the points to match:

```{r reproject, eval = FALSE}
# Occurrences recorded in lon/lat (EPSG:4326), raster in a projected CRS.
target <- tiff_crs("predictors.tif")$epsg          # e.g. 3035 (LAEA Europe)

pts <- sf::st_as_sf(occ, coords = c("lon", "lat"), crs = 4326)
pts <- sf::st_transform(pts, target)
xy  <- sf::st_coordinates(pts)

env <- tiff_extract_points("predictors.tif", x = xy[, 1], y = xy[, 2])
```

When the points and the raster already share a CRS this step is unnecessary.
The examples below work in lon/lat throughout.

## Build a predictor table

We synthesize a small two-layer climate raster over Central Europe and write it
to a `GeoTIFF` with an EPSG:4326 GeoKey. Layer `band1` stands in for mean annual
temperature and `band2` for annual precipitation.

```{r raster}
nx <- 80; ny <- 50
xmin <- 5; xmax <- 17; ymin <- 45; ymax <- 55
xres <- (xmax - xmin) / nx
yres <- (ymax - ymin) / ny

g <- expand.grid(col = 0:(nx - 1), row = 0:(ny - 1))
g$x <- xmin + (g$col + 0.5) * xres          # pixel centres
g$y <- ymax - (g$row + 0.5) * yres          # row 1 is the northern edge
g$band1 <- 14 - 0.6 * (g$y - 50) + 0.2 * (g$x - 11) + rnorm(nrow(g), 0, 0.3)
g$band2 <- 750 + 25 * (g$y - 50) - 8 * (g$x - 11) + rnorm(nrow(g), 0, 10)

tif <- tempfile(fileext = ".tif")
write_tiff(g[, c("x", "y", "band1", "band2")], tif, crs = 4326L)

tiff_crs(tif)
```

Now we sample the raster at a set of survey sites. `tiff_extract_points()`
returns one row per point with the input coordinates and one column per band.
Points outside the raster come back as `NA`.

```{r extract}
N <- 1500
sites <- data.frame(
  x = runif(N, xmin + 0.3, xmax - 0.3),
  y = runif(N, ymin + 0.3, ymax - 0.3)
)

env <- tiff_extract_points(tif, sites)
head(env)
```

We standardize the two predictors once and store the prepared columns, then
draw a presence/absence outcome from them. Standardizing before the data reaches
the model lets every streaming batch share the same transform, which the
out-of-core fit below relies on. With the generating coefficients known, each
fit can be checked against them.

```{r outcome}
occ <- data.frame(
  bio1  = as.numeric(scale(env$band1)),   # standardized once, then stored
  bio12 = as.numeric(scale(env$band2))
)
eta <- -0.2 + 1.3 * occ$bio1 - 0.9 * occ$bio12
occ$presence <- rbinom(N, 1, plogis(eta))
table(occ$presence)
```

## Fit in memory when the table fits

For most studies the extracted table has thousands to a few million rows and a
handful of predictors, which fits in memory comfortably even when the source
rasters are enormous. The ordinary path is to extract, then fit with `glm()`:

```{r glm}
fit <- glm(presence ~ bio1 + bio12, data = occ, family = binomial())
round(coef(fit), 3)
```

The recovered coefficients sit close to the values used to generate the data
(intercept -0.2, `bio1` 1.3, `bio12` -0.9).

## Stream when the table is larger than memory

Dense background sampling, many predictors, or a model fitted across a whole
continent can push the predictor table past available RAM. Here vectra streams
the table through the model in bounded pieces. We write the modelling frame to a
`.vtr` file to stand in for a table too large to hold in memory.

```{r write-vtr}
vtr <- tempfile(fileext = ".vtr")
write_vtr(occ, vtr)
```

### A single pass with collect_chunked()

`collect_chunked()` folds a function over the query one batch at a time. It is
the right tool for any summary that accumulates in bounded space across the
data. Here we compute the class balance and per-class predictor means in a
single streaming pass, never holding more than one batch:

```{r collect-chunked}
acc <- collect_chunked(
  tbl(vtr),
  function(acc, chunk) {
    p <- chunk$presence == 1
    acc$n      <- acc$n      + nrow(chunk)
    acc$n_pres <- acc$n_pres + sum(p)
    acc$sum1   <- acc$sum1   + tapply(chunk$bio1, p, sum)[c("FALSE", "TRUE")]
    acc
  },
  .init = list(n = 0L, n_pres = 0L, sum1 = c("FALSE" = 0, "TRUE" = 0))
)
acc$n_pres / acc$n                      # prevalence
```

The same pattern accumulates the cross-products `X'X` and `X'y` behind a linear
model, giving an exact least-squares fit in one pass. A generalized linear
model needs more: each iteratively reweighted step reweights the rows by the
current coefficient estimate, so the data must be read once per iteration.

### An out-of-core GLM with chunk_feeder()

`chunk_feeder()` exposes the query as a resettable generator that
`biglm::bigglm()` drives itself, re-reading the stream on every iteration.
Because a vectra node is consumed as it streams, the feeder takes a factory: a
function that returns a fresh node each time the stream is rewound.

```{r bigglm, eval = has_biglm}
src <- function() tbl(vtr) |> select(presence, bio1, bio12)

fit_stream <- biglm::bigglm(
  presence ~ bio1 + bio12,
  data = chunk_feeder(src),
  family = binomial()
)
round(coef(fit_stream), 3)
```

The streamed fit matches the in-memory `glm()` to within its convergence
tolerance, while the engine never holds more than one batch of the predictor
table at a time. The same call scales to a `.vtr` or CSV far larger than RAM:
only the factory and the formula change.

`bigglm()` rebuilds the model matrix from each batch as it arrives. A formula
term whose definition is estimated from the data, such as `scale()`, `poly()`,
`ns()`, or `bs()`, is recomputed on every batch and takes a different value on
each one. Prepare these terms before writing the table: standardize predictors
as we did above, set explicit `Boundary.knots` on splines, and make sure every
factor level appears in the stream. The formula then names prepared columns and
each batch yields the same transform.

```{r bigglm-note, echo = FALSE, results = "asis", eval = !has_biglm}
cat("> The out-of-core GLM example needs the **biglm** package, which is not",
    "installed, so it was not run here. Install it with",
    "`install.packages(\"biglm\")` to reproduce the streamed fit.\n")
```

```{r cleanup, include = FALSE}
unlink(c(tif, vtr))
```

## Transferability: is a projection environment novel?

A model calibrated in one region and projected to another (a different area, or
the same area under a future climate) is only trustworthy where the projection
environment resembles the calibration environment. Mobility-oriented parity
(MOP; Owens et al. 2013) measures that resemblance directly. For each projection
site it reports the mean distance, in predictor space, to the nearest fraction
of the calibration cloud: small where the site sits inside the training
conditions, larger where it is environmentally novel and the model would be
extrapolating.

`feature_knn()` computes that distance while streaming the projection side, so a
continental projection grid never has to be held in memory. The calibration
cloud is the resident reference; each projection batch is scored against it. The
predictors here are already standardized, so a plain Euclidean distance is
meaningful; `metric = "mahalanobis"` whitens by the calibration covariance when
the predictors are correlated.

```{r feature-knn}
calib <- occ[, c("bio1", "bio12")]          # the standardized training cloud

# Two sites inside the training conditions and one well outside them.
proj <- data.frame(
  bio1  = c(0.0, 0.5, 4.0),
  bio12 = c(0.0, -0.5, -4.0)
)
pv <- tempfile(fileext = ".vtr")
write_vtr(proj, pv)

tbl(pv) |>
  feature_knn(calib, percentage = 5) |>
  collect()
unlink(pv)
```

The third site, far from any training condition, is scored with a much larger
distance than the two that sit within the cloud.

When the projection is itself a stack of environmental rasters,
`rast_feature_distance()` folds the same distance over the projection grid,
reading the reference raster once and walking the projection one tile-row strip
at a time, so the surface is written aligned to the grid and computed
out-of-core. The strict-extrapolation half of a novelty diagnostic such as MOP,
the count of predictors whose value falls outside the range each takes over the
calibration cloud, is already native: it is a per-band `min`/`max` reduce over
the calibration raster followed by a `rast_calc()` expression on the projection
raster, so the distance surface and the extrapolation count compose without a
dedicated verb.

## Choosing a path

Extract with `tiff_extract_points()` in every case. After that:

* If the predictor table fits in memory, `collect()` it and fit with `glm()`,
  `mgcv::gam()`, or any modelling function. This is the common case.
* If a single-pass summary is all you need (class balance, sufficient
  statistics, an online mean), use `collect_chunked()`.
* If the table itself exceeds memory and you need a full GLM, feed it to
  `biglm::bigglm()` through `chunk_feeder()`.
* To judge whether a projection is within the calibrated conditions, score it
  with `feature_knn()`, or `rast_feature_distance()` for a streaming raster
  novelty surface.
```

---
title: "Coverage and topology"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Coverage and topology}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
has_sf <- requireNamespace("sf", quietly = TRUE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = has_sf
)
```

The per-feature verbs in
[Streaming spatial operations](streaming-spatial.html) treat each geometry on
its own. A second family of verbs reads a whole feature set at once: the
constructions and topology operations that need every feature in a group
together, not one row at a time. These ride a shared partition tier. A `by`
argument routes the layer into one shard per group, each shard is processed as
an independent coverage, and the shards stream so the resident budget is one
group rather than the whole layer.

This vignette covers building polygons from lines, cleaning a polygon coverage,
decomposing geometry into its parts, set-wise constructions, and linear
referencing. Every block runs on the optional
[sf](https://r-spatial.github.io/sf/) package.

```{r libs, message = FALSE}
library(vectra)
library(sf)
```

## From lines to polygons

`spatial_polygonize()` builds the polygonal faces enclosed by a line network,
the inverse of taking polygon boundaries. A group's lines are unioned and noded,
then the faces of that arrangement are returned, one per row.

```{r polygonize}
grid <- st_sfc(
  st_linestring(rbind(c(0, 0), c(2, 0))),
  st_linestring(rbind(c(0, 1), c(2, 1))),
  st_linestring(rbind(c(0, 2), c(2, 2))),
  st_linestring(rbind(c(0, 0), c(0, 2))),
  st_linestring(rbind(c(1, 0), c(1, 2))),
  st_linestring(rbind(c(2, 0), c(2, 2))))

f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(geometry = st_as_binary(grid, hex = TRUE)), f)

tbl(f) |> spatial_polygonize() |> collect_sf()
```

`spatial_line_merge()` is the line counterpart of a dissolve: it sews segments
that meet end to end into maximal linestrings, one row per chain. Segments
meeting at a junction of degree greater than two stay separate.

```{r line-merge}
seg <- st_sfc(
  st_linestring(rbind(c(0, 0), c(1, 0))),
  st_linestring(rbind(c(1, 0), c(2, 0))),
  st_linestring(rbind(c(2, 0), c(3, 0))))

g <- tempfile(fileext = ".vtr")
write_vtr(data.frame(geometry = st_as_binary(seg, hex = TRUE)), g)

tbl(g) |> spatial_line_merge() |> collect_sf()
```

## Cleaning a coverage

Real coverages arrive with jittered shared borders, sub-pixel slivers, and more
detail than a map needs. Three verbs clean them while keeping adjacent polygons
edge-matched.

`spatial_snap_grid()` rounds coordinates to a regular grid and repairs the
result, one batch at a time. It is the fixed-precision snap-rounding that
`spatial_overlay()` applies internally, exposed as a standalone verb, so a layer
can be pre-noded to a common precision without running a full overlay.

```{r snap-grid}
p <- st_polygon(list(rbind(c(0.04, 0.03), c(1.02, 0.01),
                           c(0.98, 1.03), c(0.01, 0.97), c(0.04, 0.03))))
h <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = 1L, geometry = st_as_binary(st_sfc(p), hex = TRUE)), h)

tbl(h) |> spatial_snap_grid(0.1) |> collect_sf()
```

`spatial_snap()` instead pulls a streamed layer onto a resident reference layer:
vertices and edges within `tolerance` of `y` are moved onto it, so two layers
that should share a boundary line up exactly.

```{r snap}
ref  <- st_sfc(st_linestring(rbind(c(0, 0), c(10, 0))))
line <- st_linestring(rbind(c(0, 0.2), c(5, 0.1), c(10, 0.2)))

sn <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = 1L, geometry = st_as_binary(st_sfc(line), hex = TRUE)), sn)

tbl(sn) |> spatial_snap(ref, tolerance = 0.5) |> collect_sf()
```

`spatial_eliminate()` cleans a polygon coverage by absorbing every feature
smaller than `max_area` into a neighbour, the QGIS "Eliminate". Each sliver joins
the neighbour it shares the longest border with, or the largest-area neighbour
with `into = "largest_area"`. An area-rooted union-find collapses chains of
slivers so a connected run flows to its single largest member, whose attributes
survive.

```{r eliminate}
big    <- st_polygon(list(rbind(
  c(0, 0), c(10, 0), c(10, 10), c(0, 10), c(0, 0))))
sliver <- st_polygon(list(rbind(
  c(10, 0), c(10.3, 0), c(10.3, 10), c(10, 10), c(10, 0))))

e <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = c("keep", "sliver"),
  geometry = st_as_binary(st_sfc(big, sliver), hex = TRUE)), e)

tbl(e) |> spatial_eliminate(max_area = 5) |> collect_sf()
```

The thin sliver is gone, absorbed into the square it bordered, and the survivor
keeps the `keep` attributes.

`spatial_simplify()` simplifies a polygon coverage without tearing shared edges.
Boundaries are unioned so a shared border is one line, noded into arcs, each arc
simplified once with its junction endpoints pinned, then re-polygonized. Adjacent
polygons stay edge-matched with no slivers. This is the topology-preserving
simplification a per-feature `spatial_map(~ sf::st_simplify())` cannot give,
because that simplifies each polygon's copy of a shared border independently.

```{r simplify}
p1 <- st_polygon(list(rbind(
  c(0, 0), c(1, 0), c(1, 0.5), c(1, 1), c(0, 1), c(0, 0))))
p2 <- st_polygon(list(rbind(
  c(1, 0), c(2, 0), c(2, 1), c(1, 1), c(1, 0.5), c(1, 0))))

s <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = c("a", "b"),
  geometry = st_as_binary(st_sfc(p1, p2), hex = TRUE)), s)

tbl(s) |> spatial_simplify(tolerance = 0.6) |> collect_sf()
```

## Decomposing geometry

`spatial_explode()` splits every multipart geometry into its single-part
components, a `MULTIPOLYGON` into one row per polygon and so on, copying the
source attributes onto each part. An optional `part` argument names a 1-based
part-index column. It is the streaming "multipart to singleparts" tool and the
inverse of `spatial_dissolve()`.

```{r explode}
mp <- st_multipolygon(list(
  list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))),
  list(rbind(c(2, 2), c(3, 2), c(3, 3), c(2, 3), c(2, 2)))))

m <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1L, geometry = st_as_binary(st_sfc(mp), hex = TRUE)), m)

tbl(m) |> spatial_explode(part = "part_id") |> collect_sf()
```

`spatial_topology()` decomposes a polygon coverage into the arcs of its planar
topology. The unioned boundaries are noded so a shared border becomes one arc,
tagged with the identifiers of the (up to two) polygons on either side: two for
an internal shared edge, one for an outer edge. It is the inverse of
`spatial_polygonize()`.

```{r topology}
q1 <- st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))))
q2 <- st_polygon(list(rbind(c(1, 0), c(2, 0), c(2, 1), c(1, 1), c(1, 0))))

tp <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = c("a", "b"),
  geometry = st_as_binary(st_sfc(q1, q2), hex = TRUE)), tp)

tbl(tp) |> spatial_topology(id = "id") |> collect()
```

The shared edge between `a` and `b` appears once, with both face columns filled;
each outer edge appears once with the second face `NA`.

`spatial_centerline()` traces the centerline (medial axis) of each polygon from
the Voronoi diagram of its densified boundary: the Voronoi edges that fall inside
the polygon are its skeleton, merged into maximal lines. `density` sets the
boundary sampling and `prune` drops the short spurs the skeleton grows toward
convex corners. It is the usual approximation for a river or road centerline from
a filled shape; non-polygon geometry passes through unchanged.

```{r centerline}
strip <- st_polygon(list(rbind(
  c(0, 0), c(10, 0), c(10, 2), c(0, 2), c(0, 0))))

cl <- tempfile(fileext = ".vtr")
write_vtr(data.frame(geometry = st_as_binary(st_sfc(strip), hex = TRUE)), cl)

tbl(cl) |> spatial_centerline(density = 0.25, prune = 0.5) |> collect_sf()
```

## Set-wise constructions

`spatial_construct()` builds a geometry construction from a whole feature set,
the constructions a per-feature `spatial_map()` cannot express. A `kind`
argument selects it: `"convex_hull"`, `"concave_hull"`, `"envelope"`,
`"oriented_box"`, `"enclosing_circle"`, `"inscribed_circle"`, `"pole"` (the pole
of inaccessibility, the centre of the maximum inscribed circle), `"voronoi"`,
and `"delaunay"`. A `by` argument builds one construction per group; the
enclosing kinds emit one feature per group, the tessellations one polygon per
cell.

```{r construct-hull}
nc <- st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
nc$band <- nc$SID74 > 5

n <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  band = nc$band,
  geometry = st_as_binary(st_geometry(nc), hex = TRUE)), n)

tbl(n) |>
  spatial_construct("convex_hull", by = "band", crs = st_crs(nc)) |>
  collect_sf()
```

A tessellation kind returns one polygon per cell, each carrying the group's `by`
values:

```{r construct-voronoi}
xy <- rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0.4, 0.6))
cloud <- st_sfc(lapply(seq_len(nrow(xy)), function(i) st_point(xy[i, ])))

v <- tempfile(fileext = ".vtr")
write_vtr(data.frame(geometry = st_as_binary(cloud, hex = TRUE)), v)

tbl(v) |> spatial_construct("voronoi") |> collect_sf()
```

## Linear referencing

`spatial_locate()` places streamed points along a resident line layer: each
point gets its nearest line's identifier, the measure (distance along that line
to the point's projection), and the perpendicular offset, with an optional `snap`
onto the line. The inverse, a measure back to a point, is
`sf::st_line_interpolate()` through `spatial_map()`.

```{r locate}
roads <- st_sf(road = c("main", "side"), geometry = st_sfc(
  st_linestring(rbind(c(0, 0), c(10, 0))),
  st_linestring(rbind(c(0, 5), c(0, 15)))))

pts <- data.frame(id = 1:2, x = c(3, 1), y = c(1, 9))
l <- tempfile(fileext = ".vtr")
write_vtr(pts, l)

tbl(l) |>
  spatial_locate(roads, coords = c("x", "y"), y_id = "road") |>
  collect()
```

Point 1 is one unit off `main` at measure 3; point 2 is one unit off `side` at
measure 4. With `snap = TRUE` each point's geometry is replaced by its
projection onto the matched line.

```{r cleanup, include = FALSE}
unlink(c(f, g, h, sn, e, s, m, tp, cl, n, v, l))
```

# vectra — Notes for AI Coding Agents

vectra is a self-contained R columnar query engine (pure C11 backend) for
larger-than-RAM data. It exposes dplyr-style verbs that build a lazy plan; the
plan executes only when `collect()` is called. Sources include vectra's native
`.vtr` format, CSV(.gz), SQLite, Excel (.xlsx), and GeoTIFF.

## Core workflow

```r
library(vectra)

# 1. Open a source -> lazy `vectra_node`
node <- tbl_csv("measurements.csv")

# 2. Compose verbs (still lazy)
q <- node |>
  filter(temperature > 30, year >= 2020) |>
  group_by(station) |>
  summarise(avg_temp = mean(temperature), n = n())

# 3. Execute
df <- collect(q)        # returns a data.frame
explain(q)              # inspect optimized plan without executing
```

## Sources (all return a `vectra_node`)

- `tbl(path)` — `.vtr` file. A hash index (`.vtri` sidecar) is opened for the column a pushed-down equality predicate filters on.
- `tbl_csv(path, batch_size = 65536)` — CSV / `.csv.gz`; types inferred from first 1000 rows.
- `tbl_sqlite(path, table, batch_size = 65536)` — SQLite, no `DBI` dependency.
- `tbl_xlsx(path, sheet = 1L, batch_size = 65536)` — requires `openxlsx2`; sheet is read fully into memory.
- `tbl_tiff(path, batch_size = 256)` — pixels become rows with columns `x`, `y`, `band1`, `band2`, ...

## Sinks

- `write_vtr(x, path, compress = c("fast","small","none"), batch_size = NULL, col_types = NULL, quantize = NULL, spatial = NULL)` — atomic write; streams from a node.
- `append_vtr(x, path)` — appends a row group; **schema (names, types, order) must match exactly**.
- `delete_vtr(path, row_ids)` and `diff_vtr(old_path, new_path, key_col)` — logical (key-based) diff, returns `list(added = vectra_node, deleted = key_vector)`.
- `write_csv(x, path)`, `write_sqlite(x, path, table)`.
- `write_tiff(x, path, compress = FALSE, pixel_type = "float64", metadata = NULL, crs = NULL, tiled = FALSE, tile_size = 256L, bigtiff = "auto")` — `compress` is logical; `pixel_type` ∈ `"int8"`, `"int16"`, `"int32"`, `"uint8"`, `"uint16"`, `"float32"`, `"float64"`; `tile_size` must be a multiple of 16; `crs` accepts an integer EPSG, `"EPSG:4326"`, or `list(epsg=, citation=)`.

## Star schemas

```r
s <- vtr_schema(
  fact    = tbl("obs.vtr"),
  species = link("sp_id", tbl("species.vtr")),     # unnamed key = same name in both
  site    = link(c(site_id = "id"), tbl("sites.vtr"))  # named key = remap
)
lookup(s, value, species$name, site$habitat, .join = "left", .report = TRUE) |> collect()
```

`link()` and `vtr_schema()` accept **only file-backed nodes** (created by
`tbl()` / `tbl_csv()` / `tbl_sqlite()`). Nodes that come out of verbs have no
file path and will be rejected.

## Indexes and materialized blocks

- `create_index(path, column, ci = FALSE)` writes a `.vtri` sidecar; pass a character vector (in any order) for composite indexes. `has_index(path, column)` reports whether one exists and is usable against the store as it now stands. `append_vtr(along = "rows")` rebuilds the sidecars it invalidates; an index that stops matching its store is ignored rather than used to prune, so rebuild it with `create_index()` when `has_index()` turns `FALSE`. `explain()` names the index a scan will probe.
- `materialize(node)` returns a `vectra_block` (in memory, reusable). Probe with `block_lookup(block, column, keys, ci = FALSE)` or `block_fuzzy_lookup(block, column, keys, method = c("dl","levenshtein","jw"), max_dist = 0.2, block_col = NULL, block_keys = NULL, n_threads = 4L)`.

## Out-of-core model fitting and per-group work

The engine streams, so a result larger than memory reduces to a summary in one
pass, and per-group work splits into shards that each fit in memory.

- `collect_chunked(x, f, .init = NULL, combine = NULL, commutative = FALSE)` — folds `f(acc, chunk)` left-to-right over every batch of a node, holding one batch plus the accumulator. On a `vectra_partition` it folds each shard then merges the per-shard accumulators with `combine`. Use for running counts, per-group sufficient statistics, or the `X'X` / `X'y` cross-products behind an exact OLS fit.
- `chunk_feeder(.source)` — wraps a query as a resettable generator following the `data(reset = TRUE/FALSE)` protocol `biglm::bigglm()` expects. `.source` is either a factory (`function()` returning a fresh node) or an `offload()`ed node (replays from disk). Drives a GLM fit on larger-than-RAM data.
- `offload(x, by = NULL, n = NULL, method = c("auto","level","range","hash"), path = NULL, compress = c("fast","small","none"))` — with no `by`, spills a prepared query to disk once and returns a `vectra_node` that replays from the file, so an iterative consumer re-reads the spill instead of rebuilding the pipeline. With a `by` key it returns a `vectra_partition`: a named list of per-shard nodes, one per key value.
- `group_map(.data, .f, ...)` — runs `.f` on each shard of a partition, passing the shard as a data.frame and its key as a string (purrr-style `~` formulas work: `.x` is the shard, `.y` the key). `.f` is arbitrary R: any per-group computation that needs the whole group in memory at once. Returns a named list keyed by shard.
- `group_modify(.data, .f, ...)` — same, but `.f` returns a data.frame; the per-shard results are row-bound into one table with the shard key restored as a column.

```r
# Streaming OLS: accumulate X'X and X'y in one pass, never holding the matrix
acc <- tbl("survey.vtr") |> select(mpg, wt, hp) |>
  collect_chunked(function(acc, chunk) {
    X <- cbind(1, chunk$wt, chunk$hp)
    list(XtX = acc$XtX + crossprod(X), Xty = acc$Xty + crossprod(X, chunk$mpg))
  }, .init = list(XtX = matrix(0, 3, 3), Xty = matrix(0, 3, 1)))
solve(acc$XtX, acc$Xty)

# Out-of-core GLM: prepare once, then bigglm re-reads the spill on every pass
s <- offload(tbl("occ.vtr") |> select(presence, bio1, bio12))
biglm::bigglm(presence ~ bio1 + bio12, data = chunk_feeder(s), family = binomial())

# Per-key shards: group_map runs any function on each in-memory shard (here a GLM)
p <- offload(tbl("occ.vtr"), by = "region")
fits <- group_map(p, function(d, region)
  glm(presence ~ bio1 + bio12, data = d, family = binomial()))
```

## Constraints and common mistakes

- Verbs do nothing until a terminal runs them: `collect()`, `collect_chunked()`, `pull()`, `reframe()`, `glimpse()`, `explain()`, `group_map()` / `group_modify()`, or a `write_*()`.
- `mean(x, na.rm = TRUE)` etc. — engine ignores `na.rm` at the moment; NA propagation follows R semantics by default.
- TIFF rows always carry lowercase `x`, `y`, `band1`, `band2`. Reconstruct rasters with `terra::rast(df, type = "xyz")`.
- `tiff_extract_points(path, x, y)` accepts either two numeric vectors **or** a data.frame/matrix with columns named `x` and `y`.
- `lookup(s, ...)` uses `dim$col` syntax (it parses the `$` call), not `dim.col`.
- `summarize` is an alias of `summarise` — both are exported.
- `tbl_xlsx` is **not streaming**; the whole sheet is loaded. Use `tbl_csv` or convert to `.vtr` for large data.
- `append_vtr` is not fully atomic. For safety-critical writes, use `write_vtr` (which is).

## Use cases / keywords

ETL across CSV/SQLite/TIFF/.vtr; out-of-core dplyr; star-schema lookups;
streaming aggregations; out-of-core OLS/GLM fitting (`collect_chunked` and
`chunk_feeder` + `biglm`); per-key sharding for any per-group function
(`offload(by=)` + `group_map`); hash + zone-map predicate pushdown; fuzzy string
matching (Levenshtein / Damerau-Levenshtein / Jaro-Winkler) inside the engine;
GeoTIFF point sampling without `terra`; integer raster output with quantization
and spatial predictor encoding.

---
title: "Network analysis"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Network analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
has_sf <- requireNamespace("sf", quietly = TRUE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = has_sf
)
```

A line layer becomes a routable graph when its segments are joined at their
endpoints and weighted by a traversal cost. vectra builds that graph once and
holds it resident in a `vectra_network` object, the network counterpart of the
small resident `sf` layer the streaming verbs compare against. Origins then
stream past the resident graph: each batch of start points is routed in native
C, so a national set of origins flows past a fixed memory budget while the graph
stays in memory.

The graph and the shortest-path solver are pure C, a binary-heap Dijkstra over a
compressed adjacency, with no `igraph` dependency. The solver parallelises over a
batch of origins with OpenMP.

```{r libs, message = FALSE}
library(vectra)
library(sf)
```

## Building a network

`spatial_network()` takes a line layer and returns a `vectra_network`. Endpoints
are snapped to shared nodes (exactly coincident by default, or within
`tolerance` CRS units), and each edge is weighted by its geometry length unless a
`weight` column names another cost.

The examples use a small grid of unit streets.

```{r build}
mk <- function(x1, y1, x2, y2)
  st_linestring(rbind(c(x1, y1), c(x2, y2)))

streets <- st_sfc(
  mk(0, 0, 1, 0), mk(1, 0, 2, 0), mk(2, 0, 3, 0),   # bottom row
  mk(0, 1, 1, 1), mk(1, 1, 2, 1), mk(2, 1, 3, 1),   # middle row
  mk(0, 2, 1, 2), mk(1, 2, 2, 2), mk(2, 2, 3, 2),   # top row
  mk(0, 0, 0, 1), mk(0, 1, 0, 2),                   # left verticals
  mk(1, 0, 1, 1), mk(1, 1, 1, 2),
  mk(2, 0, 2, 1), mk(2, 1, 2, 2),
  mk(3, 0, 3, 1), mk(3, 1, 3, 2))                   # right verticals

net <- spatial_network(streets)
net
```

## Shortest routes

`spatial_route()` streams a layer of origin points past the network and returns
the shortest path from each origin to one or more destinations. With
`geometry = FALSE` it returns only the cost, so a set of destinations per origin
is an origin-destination cost matrix in long form.

```{r route-cost}
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = 1L, x = 0, y = 0), f)

dest <- st_sfc(st_point(c(3, 2)))

tbl(f) |>
  spatial_route(net, to = dest, coords = c("x", "y"), geometry = FALSE) |>
  collect()
```

The cost is the five unit steps from the bottom-left corner to the top-right.
With `geometry = TRUE` (the default) each row also carries the route line, ready
to materialise with `collect_sf()`:

```{r route-geom}
tbl(f) |>
  spatial_route(net, to = dest, coords = c("x", "y")) |>
  collect_sf()
```

Several destinations, named through `to_id`, give one row per origin and
destination, the cost matrix in long form:

```{r od-matrix}
g <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = 1:2, x = c(0, 3), y = c(0, 0)), g)

dests <- st_sf(
  name = c("top_left", "top_right"),
  geometry = st_sfc(st_point(c(0, 2)), st_point(c(3, 2))))

tbl(g) |>
  spatial_route(net, to = dests, to_id = "name", geometry = FALSE,
                coords = c("x", "y")) |>
  collect()
```

Each origin carries its own attributes (`id` here) onto every route row.
Unreachable pairs return an infinite cost rather than dropping the row.

## Service areas and isochrones

`spatial_service_area()` returns, per origin, the part of the network reachable
within a cost budget. A vector of budgets returns nested travel-cost bands, one
row per origin and band. The `output` argument selects the shape: the reachable
nodes as a multipoint, the reachable edges as lines, or their convex hull as a
service-area polygon.

```{r service-nodes}
tbl(f) |>
  spatial_service_area(net, cost = c(1, 2), output = "nodes",
                       coords = c("x", "y")) |>
  collect_sf()
```

The first band reaches the start node and its two immediate neighbours; the
second band reaches everything within two steps. Asking for `output = "polygon"`
wraps each band in its hull, the usual drive-time catchment:

```{r service-polygon}
tbl(f) |>
  spatial_service_area(net, cost = 2, output = "polygon",
                       coords = c("x", "y")) |>
  collect_sf() |>
  st_area()
```

## Weighted and directed networks

By default every edge is two-way and weighted by length. A `weight` column sets
a different traversal cost (a travel time, a toll), and `directed = TRUE` makes
edges one-way. The `direction` column then carries one-way codes (`"B"`
two-way, `"FT"` along the digitised direction, `"TF"` against it, `"N"` closed),
and `weight_to` gives the reverse cost on a two-way edge.

```{r weighted}
streets_df <- st_sf(
  cost = runif(length(streets), 1, 3),
  geometry = streets)

wnet <- spatial_network(streets_df, weight = "cost")

tbl(f) |>
  spatial_route(wnet, to = dest, coords = c("x", "y"), geometry = FALSE) |>
  collect()
```

The route now minimises the summed `cost` weight instead of the step count.

## The streaming model

The graph is the resident budget: it is built once and held in the
`vectra_network` object for the life of the queries against it. The origin layer
is what streams. Each batch of origins is solved against the resident graph with
one Dijkstra per origin, run in parallel across the batch, so the query side
scales by streaming while memory tracks the graph plus one batch. This is the
network analogue of the resident-`y` pattern the
[streaming spatial verbs](streaming-spatial.html) use for a small locator layer.

```{r cleanup, include = FALSE}
unlink(c(f, g))
```

---
title: "Geometry functions in expressions"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Geometry functions in expressions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
has_sf <- requireNamespace("sf", quietly = TRUE)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = has_sf
)
```

vectra carries geometry through the engine as hex-encoded WKB in an ordinary
string column. The verbs in
[Streaming spatial operations](streaming-spatial.html) wrap whole sf operations
around that column. This vignette covers the other half of the spatial surface:
a family of `st_*` functions that work inside the expression verbs themselves,
so a measure, a predicate, or a geometry transform is just another term in
`mutate()`, `filter()`, or `summarise()`.

These functions run on the GEOS C library straight off the WKB column, one row
at a time, with no per-batch round-trip through sf. `filter(st_area(geometry) >
1e6)` prunes the stream in C, and `mutate(here = st_centroid(geometry))` adds a
new WKB geometry column that any later verb can read. The per-row decode is
parallelised with OpenMP, so a measure over a large layer uses every core.

```{r libs, message = FALSE}
library(vectra)
library(sf)
```

## A layer to work on

The examples use the North Carolina counties shipped with sf. Writing the layer
to a `.vtr` is the usual first step: the geometry becomes a hex-WKB string
column (named `geometry` by convention), and the attributes ride alongside it.

```{r data}
nc <- st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)

f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME     = nc$NAME,
  BIR74    = nc$BIR74,
  geometry = st_as_binary(st_geometry(nc), hex = TRUE)
), f)

tbl(f)
```

The counties are stored in longitude and latitude, so every measure below is
planar in those units, the same convention the streaming verbs follow. Project
the layer first if you need metric areas or distances.

## Measures

A measure reads a geometry and returns a number, so it drops into `mutate()` as
an ordinary column.

```{r measures}
tbl(f) |>
  mutate(area   = st_area(geometry),
         perim  = st_perimeter(geometry),
         nverts = st_npoints(geometry)) |>
  select(NAME, area, perim, nverts) |>
  collect() |>
  head()
```

`st_length()` returns the boundary length of a polygon (an alias of
`st_perimeter()`) and the line length of a linestring. `st_ngeometries()`
counts the parts of a multi-geometry. `st_x()` and `st_y()` read the coordinate
of a point and return `NA` for anything that is not a point, which makes them
most useful on a centroid:

```{r coords}
tbl(f) |>
  mutate(centroid = st_centroid(geometry),
         cx = st_x(centroid),
         cy = st_y(centroid)) |>
  select(NAME, cx, cy) |>
  collect() |>
  head()
```

A geometry-valued function such as `st_centroid()` produces a new WKB column
(`centroid` above), and the next term reads it like any other column. That is
the whole mechanism: geometry in, geometry or a scalar out, all as columns.

## Predicates

A unary predicate tests one geometry: `st_is_valid()`, `st_is_empty()`,
`st_is_simple()`. A binary predicate tests a topological relation against a
second geometry: `st_intersects()`, `st_within()`, `st_contains()`,
`st_overlaps()`, `st_touches()`, `st_crosses()`, `st_equals()`,
`st_disjoint()`, `st_covers()`, `st_covered_by()`.

In `filter()` a predicate keeps the rows where the relation holds, the
geometry-expression form of select-by-location:

```{r predicate-filter}
aoi <- st_as_sfc(st_bbox(c(xmin = -80, ymin = 35, xmax = -78, ymax = 36.5)),
                 crs = st_crs(nc))

tbl(f) |>
  filter(st_intersects(geometry, aoi)) |>
  collect() |>
  nrow()
```

The second geometry here is a constant `sf` object. It is parsed once and shared
read-only across every row, so testing a whole stream against one area of
interest stays cheap. A multi-feature object is unioned to a single geometry
first.

In `mutate()` the same call returns a logical column, ready for a later verb:

```{r predicate-mutate}
tbl(f) |>
  mutate(near_raleigh = st_intersects(geometry, aoi)) |>
  filter(near_raleigh) |>
  select(NAME) |>
  collect() |>
  head()
```

## Distance

`st_distance()` returns the shortest planar distance to a second geometry,
again a constant or another column:

```{r distance}
raleigh <- st_sfc(st_point(c(-78.64, 35.78)), crs = st_crs(nc))

tbl(f) |>
  mutate(centroid   = st_centroid(geometry),
         d_raleigh  = st_distance(centroid, raleigh)) |>
  select(NAME, d_raleigh) |>
  arrange(d_raleigh) |>
  collect() |>
  head()
```

When the second argument is a geometry column instead of a constant, the
distance is computed row by row between the two columns.

## Aggregating a measure

Because a measure is an ordinary numeric column, it aggregates like one. A
grouped `summarise()` over a measure is a zonal total computed entirely in the
stream:

```{r summarise}
tbl(f) |>
  mutate(area = st_area(geometry)) |>
  summarise(total_area = sum(area), counties = n()) |>
  collect()
```

## Transforms

A transform returns a geometry, so it builds a new WKB column. Materialise it
with `collect_sf()`, which reads the WKB column back into an `sf` object (point
it at the column with `geom =`, and pass the `crs` the layer was stored in).

```{r transforms}
hulls <- tbl(f) |>
  mutate(geometry = st_convex_hull(geometry)) |>
  select(NAME, geometry) |>
  collect_sf(crs = st_crs(nc))

hulls
```

The transform set is `st_centroid()`, `st_point_on_surface()` (a point
guaranteed to lie on the geometry), `st_boundary()`, `st_envelope()` (the
bounding rectangle), `st_convex_hull()`, `st_make_valid()` (repair an invalid
geometry), and two parameterised forms: `st_buffer(g, dist)` and
`st_simplify(g, tol)`. Buffering each county and reading the areas back:

```{r buffer}
tbl(f) |>
  mutate(geometry = st_buffer(geometry, 0.1)) |>
  select(NAME, geometry) |>
  collect_sf(crs = st_crs(nc)) |>
  st_area() |>
  head()
```

`st_geometry_type()` returns the GEOS type name (`"Point"`, `"Polygon"`,
`"MultiPolygon"`, and so on) as a string column.

## The second geometry of a binary op

For `st_distance()` and the binary predicates, the second argument can be:

- another geometry column, compared row by row;

- a constant `sf` or `sfc` object, parsed once and reused across the stream (a
  multi-feature object is unioned to one geometry);
- a hex-WKB string.

## Missing geometry

A missing (`NA`) or unparseable geometry, or an operation with no answer (the
coordinate of a non-point, the distance to a missing geometry), yields `NA` for
that row rather than stopping the query:

```{r na}
g <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1:4,
  geometry = c(st_as_binary(st_geometry(nc)[1:3], hex = TRUE), NA)
), g)

tbl(g) |>
  mutate(area = st_area(geometry)) |>
  collect()
```

## Where this fits

The `st_*` expressions are the scalar, per-row layer of vectra's spatial
surface. They cover measures, predicates, and the common single-geometry
transforms at the price of a column term, with no sf object built per batch. For
an arbitrary per-feature transform that has no `st_*` form, reach for
`spatial_map()`, which runs any sf-in, sf-out function over each feature. For
constructions that read a whole feature set at once (dissolves, overlays, hulls
of a group, planar topology), the set-wise `spatial_*` verbs in
[Streaming spatial operations](streaming-spatial.html) and
[Coverage and topology](coverage-topology.html) are the tools.

```{r cleanup, include = FALSE}
unlink(c(f, g))
```

---
title: "Star Schemas and Lookup"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Star Schemas and Lookup}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## The flat-table problem

Ecological analyses almost always involve multiple tables. Observations live in
one file, species traits in another, site metadata in a third, climate variables
in a fourth. The natural instinct is to join everything into a single wide
data.frame early on, then work from that flat table for the rest of the
analysis.

This works for small datasets. At scale, it creates problems.

A biodiversity monitoring program with 50 million observations, 12,000 species
each carrying 40 trait columns, 3,000 sites with 25 metadata fields, and a
climate grid with 19 bioclimatic variables produces a flat table with over 80
columns per row. Most analyses use 5 to 10 of those columns. The remaining 70+
columns burn memory, slow down scans, and make column names collide across
tables (requiring `.x` and `.y` suffixes that propagate through downstream
code). When a new trait column is added to the species reference, every script
that built the flat table needs updating.

Relational databases solved this decades ago with foreign keys and normalized
schemas. The data stays in separate tables; queries join them on demand, pulling
only the columns the query needs. vectra brings the same pattern to file-based
analytical workflows with three functions: `link()`, `vtr_schema()`, and
`lookup()`.


## The star schema concept

A star schema organizes data around a central **fact table** (the primary
dataset with measurements or events) linked to multiple **dimension tables**
(reference data that enriches the facts). The fact table holds foreign keys that
point into each dimension. In database terminology, the fact table sits at the
center and dimensions radiate outward like points of a star.

For ecological data, the mapping is direct:

| Role | Table | Key | Columns |
|:-----|:------|:----|:--------|
| Fact | observations | sp_id, site_id, date | count, biomass, cover |
| Dimension | species | sp_id | name, family, order, red_list_status, ... |
| Dimension | sites | site_id | habitat, elevation, lat, lon, country, ... |
| Dimension | climate | site_id | bio1, bio2, ..., bio19 |
| Dimension | traits | sp_id | body_mass, diet, dispersal, ... |

Each dimension table has a unique key. The fact table references those keys but
stores only the measurements. To answer "what is the average count per habitat
type?", we need exactly two columns from the site dimension (site_id and
habitat) and one from the fact table (count). A flat table would have loaded all
25 site columns and all 40 trait columns into memory for no reason.


## Setting up a schema

We will build a schema from three tables: field observations of tree species
across monitoring sites, a species reference with taxonomic and conservation
data, and a site metadata table.

```{r data-setup}
library(vectra)

# Fact table: field observations
obs_path <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  obs_id  = 1:12,
  sp_id   = c(1, 2, 3, 1, 2, 4, 3, 1, 5, 2, 3, 1),
  site_id = c(1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4),
  count   = c(5, 12, 3, 8, 15, 2, 7, 20, 1, 9, 4, 11),
  dbh_cm  = c(35, 22, 48, 31, 19, 55, 42, 28, 12, 25, 39, 33)
), obs_path)

# Dimension: species
sp_path <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  sp_id       = 1:4,
  name        = c("Quercus robur", "Fagus sylvatica",
                   "Pinus sylvestris", "Abies alba"),
  family      = c("Fagaceae", "Fagaceae", "Pinaceae", "Pinaceae"),
  red_list    = c("LC", "LC", "LC", "NT"),
  shade_tol   = c(0.4, 0.8, 0.2, 0.7),
  max_height  = c(40, 45, 35, 55),
  stringsAsFactors = FALSE
), sp_path)

# Dimension: sites
site_path <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  site_id   = 1:4,
  site_name = c("Wienerwald A", "Wienerwald B",
                "Donau-Auen", "Neusiedlersee"),
  habitat   = c("Deciduous", "Deciduous", "Riparian", "Steppe"),
  elev_m    = c(450, 520, 155, 120),
  annual_precip_mm = c(750, 780, 550, 600),
  stringsAsFactors = FALSE
), site_path)
```

`link()` describes how a dimension table connects to the fact table. The first
argument is the key column (the column name shared between fact and dimension),
the second is a `vectra_node` pointing to the dimension file.

```{r links}
sp_link   <- link("sp_id", tbl(sp_path))
site_link <- link("site_id", tbl(site_path))
```

`vtr_schema()` ties the fact table and its dimension links together. Each link
gets a name that becomes the alias used in `lookup()` calls.

```{r schema}
s <- vtr_schema(
  fact = tbl(obs_path),
  sp   = sp_link,
  site = site_link
)
s
```

The print output shows the fact table's column count and each dimension with its
key. The schema object is lightweight. It holds pointers to the underlying
files, not copies of the data.


## Looking up dimension columns

`lookup()` is the verb that resolves columns through the schema. Bare names
refer to fact columns. The `dim$col` syntax refers to a specific column in a
named dimension.

```{r lookup-basic}
lookup(s, count, sp$name, site$habitat, .report = FALSE) |> collect()
```

The result has 12 rows (one per observation) and exactly 3 columns. No trait
columns, no climate data, no site coordinates were loaded. vectra built the
join tree internally: `left_join` the species dimension on `sp_id`, then
`left_join` the site dimension on `site_id`, then project down to the three
requested columns.

Requesting columns from only one dimension skips the other entirely. The
species file is never opened in this call:

```{r lookup-one-dim}
lookup(s, count, dbh_cm, site$habitat, site$elev_m, .report = FALSE) |>
  collect()
```


## Match reporting

By default, `lookup()` checks each referenced dimension for unmatched keys
before building the join tree. The check runs an `anti_join` on fresh node
copies, so it does not consume the lazy nodes used for the actual result.

Our fact table contains `sp_id = 5` (row 9), which has no entry in the species
dimension. The report catches this:

```{r report}
result <- lookup(s, count, sp$name) |> collect()
```

One observation out of 12 had an `sp_id` that the species table did not
recognize. The message names the dimension, shows how many rows were
unmatched, and previews the offending key values. For a left join, those rows
survive with NA in the dimension columns:

```{r report-na}
result
```

Row 9 has `name = NA` because sp_id 5 does not exist in the species reference.
The match report makes this visible at query time rather than three pipeline
stages later when an aggregation silently drops NA groups.

When all keys match, the report confirms it:

```{r report-ok}
lookup(s, count, site$habitat) |> collect()
```

All 12 observations have valid site_ids.

To suppress the report (useful inside functions or loops where the message would
be noise), set `.report = FALSE`:

```{r report-off}
lookup(s, count, sp$name, .report = FALSE) |> collect()
```


## Named keys

Sometimes the fact table and dimension table use different column names for the
same logical key. The species dimension might call it `species_id` while the
fact table calls it `sp_id`.

`link()` accepts named character vectors, the same `c("fact_col" = "dim_col")`
syntax used by `left_join()`:

```{r named-keys}
# Dimension with a different key name
sp2_path <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  species_code = 1:4,
  latin_name   = c("Quercus robur", "Fagus sylvatica",
                    "Pinus sylvestris", "Abies alba"),
  stringsAsFactors = FALSE
), sp2_path)

s2 <- vtr_schema(
  fact = tbl(obs_path),
  sp   = link(c("sp_id" = "species_code"), tbl(sp2_path))
)

lookup(s2, count, sp$latin_name, .report = FALSE) |> collect()
```

The named key tells vectra that the fact table's `sp_id` maps to the
dimension's `species_code`. The result column is named `latin_name`, matching
the dimension.

Composite keys (joining on multiple columns) work the same way. A temporal
dimension keyed by both site and year would use
`link(c("site_id", "year"), tbl(temporal_path))`.


## Join modes

The `.join` parameter controls whether unmatched fact rows are kept or dropped.
The default is `"left"`, which preserves every fact row and fills unmatched
dimension columns with NA. This is the safe default for exploratory work: no
data disappears silently.

`"inner"` drops fact rows with no dimension match. This is useful when the
analysis requires complete records across all referenced dimensions.

```{r join-inner}
# Only observations with known species
lookup(s, count, sp$name, .join = "inner", .report = FALSE) |> collect()
```

The 12-row fact table shrinks to 11 rows. The observation with `sp_id = 5`
(which had no species match) is gone. With an inner join, the match report
becomes less critical because the join itself enforces completeness. But it
still flags the issue before data goes missing:

```{r join-inner-report}
lookup(s, count, sp$name, .join = "inner") |> collect()
```


## Reusing the schema

The schema object does not hold live data. It stores file paths and reopens
fresh scan nodes each time `lookup()` is called. This means the same schema
works across multiple analyses without invalidating previous results.

```{r reuse}
# Analysis 1: species composition by habitat
a1 <- lookup(s, sp$name, site$habitat, .report = FALSE) |> collect()

# Analysis 2: stem diameter by elevation
a2 <- lookup(s, dbh_cm, site$elev_m, .report = FALSE) |> collect()

# Analysis 3: conservation status across sites
a3 <- lookup(s, count, sp$red_list, site$site_name, .report = FALSE) |>
  collect()
```

```{r reuse-show}
a1
a2
a3
```

Three different column selections from the same schema, each building its own
join tree internally. No flat table required.


## Practical patterns

### Pattern 1: filtering before lookup

`lookup()` works on the fact table as registered in the schema. To filter the
fact table before looking up dimensions, apply the filter to the source file and
register a new schema:

```{r pattern-filter}
s_large <- vtr_schema(
  fact = tbl(obs_path) |> filter(count >= 5),
  sp   = link("sp_id", tbl(sp_path)),
  site = link("site_id", tbl(site_path))
)
lookup(s_large, count, sp$name, site$habitat, .report = FALSE) |> collect()
```

The filter runs lazily inside the join tree. Only observations with count >= 5
reach the join nodes.


### Pattern 2: aggregation after lookup

Because `lookup()` returns a `vectra_node`, it composes with all downstream
verbs. Group by a dimension column and aggregate:

```{r pattern-agg}
lookup(s, count, sp$family, .report = FALSE) |>
  group_by(family) |>
  summarise(total = sum(count), n_obs = n()) |>
  collect()
```

This pipeline scans the fact table, joins only the species dimension (to get
`family`), groups on it, and computes the aggregation. The site dimension is
never touched.

```{r pattern-agg2}
lookup(s, count, site$habitat, .report = FALSE) |>
  group_by(habitat) |>
  summarise(mean_count = mean(count), max_count = max(count)) |>
  collect()
```


### Pattern 3: multiple dimensions in one aggregation

Crossing two dimension columns in a grouping creates a two-way summary:

```{r pattern-cross}
lookup(s, count, sp$family, site$habitat, .report = FALSE) |>
  group_by(family, habitat) |>
  summarise(total = sum(count)) |>
  collect()
```


### Pattern 4: writing results back

Lookup results can be written directly to any output format, since the return
value is a standard `vectra_node`:

```{r pattern-write}
out_path <- tempfile(fileext = ".vtr")
lookup(s, count, sp$name, site$habitat, .report = FALSE) |>
  write_vtr(out_path)

tbl(out_path) |> collect()
```

The write streams through the join tree batch by batch. The full joined result
never needs to exist in memory at once.


## When not to use a schema

Schemas are most valuable when multiple analyses query the same set of linked
tables with different column selections. For a one-off join where the column
set is known upfront, a direct `left_join()` call is simpler and equally
efficient.

Schemas also require file-backed nodes. Tables created from in-memory
data.frames (without writing to a `.vtr` or `.csv` first) cannot be registered
as schema links, because the engine needs to reopen fresh scan nodes from
file paths. If the dimension data lives only in memory, write it to a tempfile
first or use `left_join()` directly.

The sweet spot is any project where the same fact table is analyzed repeatedly
against a stable set of dimension tables, each time needing a different
slice of columns. Environmental monitoring, biodiversity databases, long-running
survey programs, species distribution modelling pipelines: all fit the pattern.

For those workflows, registering the schema once (in a project setup script or
at the top of an analysis) replaces dozens of `left_join()` calls scattered
across the codebase, makes column provenance explicit (`sp$name` is
unambiguous in a way that a bare `name` column in a 100-column flat table
is not), and catches broken keys before they propagate.


```{r cleanup, include = FALSE}
unlink(c(obs_path, sp_path, site_path, sp2_path, out_path))
```

---
title: "vectra Engine Reference"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{vectra Engine Reference}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This document defines vectra's public contract: which operations are supported,
how types and coercion work, what streams and what materializes, and where
guarantees stop. It is the reference for what vectra promises.

See `?tbl`, `?filter`, `?left_join`, and `?explain` for function-level
documentation.

## What vectra is

vectra is an R-native columnar query engine for datasets that don't fit in
memory. It provides dplyr-style verbs backed by a pure C11 pull-based execution
engine and a custom on-disk format (`.vtr`). All operations are lazy until
`collect()` materializes results as an R data.frame.

vectra is not a dplyr backend plugin. It defines its own S3 generics. The verbs
share names and semantics with dplyr, but do not depend on it.

## Execution model

### Pull-based pipeline

Every verb (`filter`, `select`, `mutate`, ...) builds a plan node. Nodes form a
tree. No data moves until `collect()` calls the root node's `next_batch()`
function, which pulls data through the tree one **row group** at a time.

A row group (internally: `VecBatch`) is a set of columnar arrays, typically
thousands to millions of rows. Each column is a typed array with a validity
bitmap for NA support.

```r
# Nothing executes here --- just builds a plan tree
plan <- tbl("data.vtr") |>
  filter(x > 0) |>
  select(id, x, y) |>
  mutate(z = x + y)

# Data flows when you call collect()
result <- collect(plan)

# Or inspect the plan without executing it
explain(plan)
```

### Selection vectors (zero-copy filtering)

`filter()` does not copy rows. It attaches a **selection vector** to the batch:
an integer array indexing which physical rows pass the predicate. Downstream
nodes read only the selected rows. This avoids memory allocation and copying for
selective filters.

### Columnar storage

Data is stored and processed column-by-column, not row-by-row. This means
operations that touch few columns (e.g. `select(id, x)` on a 100-column table)
only read the columns they need from disk.

## Data sources

| Function | Format | Streaming |
|:---------|:-------|:----------|
| `tbl(path)` | `.vtr` (vectra native) | yes, row-group-at-a-time |
| `tbl_csv(path)` | CSV | yes, batch-at-a-time |
| `tbl_sqlite(path, table)` | SQLite | yes, batch-at-a-time |
| `tbl_tiff(path)` | GeoTIFF | yes, row-strip-at-a-time |

All sources produce the same `vectra_node` object. The query engine does not
know or care which source is upstream.

### Output sinks

| Function | Format | Streaming |
|:---------|:-------|:----------|
| `collect()` | R data.frame | materializes full result in R memory |
| `write_vtr(df, path)` | `.vtr` | writes from data.frame |
| `write_csv(x, path)` | CSV | streams batch-by-batch |
| `write_sqlite(x, path, table)` | SQLite | streams batch-by-batch |
| `write_tiff(x, path, pixel_type)` | GeoTIFF | streams batch-by-batch; `pixel_type`: float64/float32/int16/int32/uint8/uint16 |

## Supported verbs

### Transformation verbs

| Verb | Streams | Notes |
|:-----|:--------|:------|
| `filter(...)` | yes | Zero-copy via selection vector |
| `select(...)` | yes | Full tidyselect: `starts_with()`, `where()`, `-col`, etc. |
| `mutate(...)` | yes | Arithmetic, comparison, boolean, `is.na()`, `nchar()`, `substr()`, `grepl()`, math (`abs`, `sqrt`, `log`, `exp`, `floor`, `ceiling`, `round`, `log2`, `log10`, `sign`, `trunc`), `if_else()`, `between()`, `%in%`, type casting (`as.numeric`), `tolower()`, `toupper()`, `trimws()`, `paste0()`, `gsub()`, `sub()`, `startsWith()`, `endsWith()`, `pmin()`, `pmax()`, `year()`, `month()`, `day()`, `hour()`, `minute()`, `second()`, `as.Date()` |
| `transmute(...)` | yes | Like `mutate()` but drops unmentioned columns |
| `rename(...)` | yes | Full tidyselect rename support |
| `relocate(...)` | yes | Reorder columns with `.before` / `.after` |

### Aggregation verbs

| Verb | Streams | Notes |
|:-----|:--------|:------|
| `group_by(...)` | metadata only | Attaches grouping info; no data moves |
| `summarise(...)` | **materializes** | Hash-based or sort-based aggregation |
| `ungroup()` | metadata only | Removes grouping |
| `count(...)` | **materializes** | Sugar for `group_by() |> summarise(n = n())` |
| `tally(...)` | **materializes** | Like `count()` on existing groups |

Supported aggregation functions: `n()`, `sum()`, `mean()`, `min()`, `max()`, `sd()`, `var()`, `first()`, `last()`, `any()`, `all()`, `median()`, `n_distinct()`.
All accept `na.rm = TRUE`.

### Ordering verbs

| Verb | Streams | Notes |
|:-----|:--------|:------|
| `arrange(...)` | **materializes** | External merge sort with 1 GB spill budget |
| `slice_head(n)` | yes | Limit node, stops after n rows |
| `slice_tail(n)` | **materializes** | Must see all rows to take last n |
| `slice_min(order_by, n)` | partial | Heap-based top-N; `with_ties = TRUE` (default) includes ties |
| `slice_max(order_by, n)` | partial | Heap-based top-N; `with_ties = TRUE` (default) includes ties |
| `head(n)` | yes | Alias for `slice_head() |> collect()` |
| `slice(...)` | **materializes** | Select or exclude rows by position (positive or negative indices) |
| `distinct(...)` | **materializes** | Uses hash-based grouping |

### Join verbs

| Verb | Streams | Notes |
|:-----|:--------|:------|
| `inner_join(x, y)` | **build materializes right** | Hash join; left streams |
| `left_join(x, y)` | **build materializes right** | Hash join; left streams |
| `right_join(x, y)` | **build materializes left** | Implemented as swapped left join |
| `full_join(x, y)` | **build materializes right** | Hash join + finalize pass |
| `semi_join(x, y)` | **build materializes right** | Hash join; returns left rows only |
| `anti_join(x, y)` | **build materializes right** | Hash join; returns non-matching left rows |
| `cross_join(x, y)` | **materializes** | Cartesian product; no key columns required |

All joins support: `by = "col"`, `by = c("a" = "b")`, `by = NULL` (natural
join), and `suffix = c(".x", ".y")`.

### Window functions

Available inside `mutate()`:

| Function | Description |
|:---------|:------------|
| `row_number()` | Sequential row number (respects groups) |
| `rank(col)` | Min rank with gaps for ties (like `dplyr::min_rank()`) |
| `dense_rank(col)` | Consecutive rank without gaps |
| `lag(col, n, default)` | Previous value |
| `lead(col, n, default)` | Next value |
| `cumsum(col)` | Cumulative sum |
| `cummean(col)` | Cumulative mean |
| `cummin(col)` | Cumulative minimum |
| `cummax(col)` | Cumulative maximum |
| `ntile(n)` | Divide rows into n roughly equal buckets |
| `percent_rank(col)` | Relative rank scaled to [0, 1] |
| `cume_dist(col)` | Cumulative distribution (proportion of values <= current) |

Window functions respect `group_by()` partitions. They **materialize** all data
within each partition.

### Other verbs

| Verb | Streams | Notes |
|:-----|:--------|:------|
| `pull(var)` | **materializes** | Collects one column as a vector |
| `bind_rows(...)` | yes | Streaming concat if schemas are compatible |
| `bind_cols(...)` | **materializes** | Collects all inputs, then `cbind()` |
| `across(...)` | n/a | Column expansion helper for `mutate()`/`summarise()` |
| `explain()` | n/a | Prints the plan tree |
| `glimpse()` | **materializes** (preview only) | Shows column types and first few values |

### tidyselect support

`select()`, `rename()`, `relocate()`, `distinct()`, and `across(.cols)` support
the full tidyselect vocabulary:

- `starts_with()`, `ends_with()`, `contains()`, `matches()`

- `everything()`, `last_col()`

- `all_of()`, `any_of()`

- `where()` predicates (e.g. `where(is.numeric)`)

- `-column` negation

`where()` works because vectra builds a 0-row typed proxy data.frame from the
schema, giving tidyselect enough type information to evaluate predicates.

## Supported types

### Base types

| C type | R input | R output (default) | R output (bit64 mode) |
|:-------|:--------|:-------------------|:----------------------|
| `int64` | integer | double | integer64 |
| `double` | double | double | double |
| `bool` | logical | logical | logical |
| `string` | character | character | character |

R's 32-bit `integer` is widened to 64-bit `int64` on write. On read, `int64` is
returned as `double` by default (R has no native 64-bit integer). Set
`options(vectra.int64 = "bit64")` to get `bit64::integer64` output instead.

### Annotated types

The `.vtr` format version 2 stores per-column annotations that preserve R type
metadata through the write/read cycle:

| R class | Annotation | Storage | Roundtrip |
|:--------|:-----------|:--------|:----------|
| Date | `"Date"` | double (days since epoch) | exact |
| POSIXct | `"POSIXct\|tz"` | double (seconds since epoch) | exact (tz preserved) |
| factor | `"factor\|lev1\|lev2\|..."` | string | exact (levels + order preserved) |

Annotations are metadata. The underlying C engine operates on the base types
only. Type restoration happens at `collect()` time.

Date and POSIXct columns support component extraction via `year()`, `month()`,
`day()`, `hour()`, `minute()`, `second()` in `mutate()` and `filter()`
expressions. Use `as.Date("2020-01-01")` as a literal in filter comparisons.
Date arithmetic (adding/subtracting days) works via standard `+` and `-`.

## Coercion rules

### Arithmetic and comparison expressions

The coercion hierarchy for numeric operations is:

```
bool < int64 < double
```

When an expression combines two different numeric types, the narrower type is
promoted to the wider type before evaluation. String columns cannot participate
in arithmetic or comparison with numeric columns; this is an error.

### Join key coercion

Join keys follow the same hierarchy. If the left key is `int64` and the right
key is `double`, the `int64` side is coerced to `double` for hashing and
comparison. The coercion happens internally; the output column retains the
original left-side type.

Joining a `string` key against a numeric key is an error.

### bind_rows coercion

When `bind_rows()` combines tables with different column types, it computes the
common type using the same `bool < int64 < double` hierarchy. Per-batch
coercion happens at the C level during streaming --- no R fallback needed.

If column names differ across inputs, the R fallback path is used: all inputs
are collected, aligned by column name, and combined with `rbind()`.

## NA semantics

### Storage

NAs are tracked by a per-column validity bitmap. Every column of every type
supports NA values. The bitmap is bit-packed (1 bit per row, 1 = valid).

### Propagation

- **Arithmetic**: `NA + x = NA`, `NA * x = NA`

- **Comparison**: `NA > x = NA`, `x == NA = NA`

- **Boolean**: `NA & FALSE = FALSE`, `NA & TRUE = NA`, `NA | TRUE = TRUE`, `NA | FALSE = NA`

- **Aggregation**: NAs are included by default; use `na.rm = TRUE` to exclude

- **Joins**: NA keys never match (same as SQL NULL semantics)

- **Window functions**: `cumsum()` and friends propagate NA forward

### is.na()

`is.na(col)` is supported in `filter()` and `mutate()` expressions. It returns
a boolean column based on the validity bitmap.

## Ordering guarantees

- `filter()`, `select()`, `mutate()`, `rename()`, `relocate()`: preserve input
  order
- `arrange()`: produces a total order (stable sort)

- `group_by() |> summarise()`: output order is **not guaranteed** (hash-based
  path) or sorted by key (sort-based path); do not depend on either
- `distinct()`: output order is not guaranteed

- Joins: probe-side order is preserved within each batch; build-side order is
  not guaranteed
- `bind_rows()`: child order is preserved (first child's rows, then second's,
  etc.)

## Streaming vs materializing

### Streaming nodes (constant memory per batch)

- Scan (`.vtr`, CSV, SQLite, TIFF)

- Filter

- Project (select / mutate / rename / relocate / transmute)

- Limit (slice_head, head)

- Concat (bind_rows)

### Materializing nodes

These nodes buffer data in memory:

| Node | What it buffers | Bounded by |
|:-----|:----------------|:-----------|
| Sort (arrange) | All input rows | 1 GB memory budget, then spills to disk |
| GroupAgg (summarise) | Hash table of groups + accumulators | Number of distinct groups |
| TopN (slice_min/max) | Heap of n rows | Requested n |
| Window | All rows per partition | Partition size |
| Join (build phase) | Right-side table in hash table | Right-side row count |

### External sort (spill-to-disk)

`arrange()` accumulates incoming batches into column builders in memory. After
each batch, the sort node estimates total memory usage across all builders. When
the estimate exceeds the memory budget (default 1 GB, defined as
`DEFAULT_MEM_BUDGET`), the node flushes the accumulated data as a sorted run:

1. The builders are finished into arrays.

2. The arrays are sorted in-place using a parallel merge sort (OpenMP task
   spawning above 32,768 rows, sequential below).
3. The sorted data is written to a temporary `.vtr` file in the system temp
   directory, split into row groups of 65,536 rows each.
4. The builders are reset and accumulation continues.

This spill cycle repeats as many times as needed. A 10 GB dataset with a 1 GB
budget produces approximately 10 spill files.

Once all input is consumed, the sort node enters the merge phase. It opens all
spill files as `Vtr1File` readers and runs a k-way merge using a min-heap. Each
heap entry holds a reference to one merge run (one spill file) and tracks the
current row group and cursor position within that run. The merge emits batches
of up to 65,536 rows, reading row groups from spill files on demand. Peak
memory during the merge phase is proportional to k (number of runs) times the
row group size, not the total dataset size.

If all data fits within the 1 GB budget, no spill occurs. The node sorts in
memory and emits the result directly. This is the common case for datasets
under ~100 million rows (depending on column count and types).

The sort-based `group_by() |> summarise()` path (used internally when the
engine detects it is advantageous) also benefits from this spill mechanism.
Temporary spill files are deleted when the sort node is freed.

### Join memory model

Joins use a **build-right, probe-left** hash join:

1. **Build phase**: The entire right-side table is materialized into a hash
   table in memory. Key columns are hashed using FNV-1a. For single-key joins,
   the hash is computed directly on the key bytes. For multi-key joins, per-key
   hashes are combined via XOR and FNV prime multiplication. The hash table uses
   open addressing. Key hashing is OpenMP-parallelized for batches above 32,768
   rows.
2. **Probe phase**: Left-side batches stream through one at a time, probing the
   hash table. For each left row, the engine computes the key hash, looks up
   the hash table, and verifies key equality (hash collisions are resolved by
   comparison). Probe-side hashing is also parallelized.
3. **Finalize phase** (full_join only): Unmatched right-side rows are emitted in
   chunks of 65,536 rows. A matched-flag array tracks which right-side rows
   participated in the join.

The memory cost of a join is proportional to the right-side table size (data
arrays plus hash table overhead). The left side streams and does not accumulate.
For this reason, place the smaller table on the right side of the join. The
`right_join` verb handles this automatically by swapping sides internally and
remapping columns in the output.

## The .vtr file format

The `.vtr` format is vectra's native binary columnar format. It is designed for
fast sequential reads with row-group-level granularity.

### Layout

```
Header:
  magic bytes ("VTR1")
  version (uint16: 1–4)
  n_cols, n_rowgroups
  per-column: name + type byte [+ annotation string in v2+]
  row group index (byte offsets)

Row groups (repeated):
  per-column:
    validity bitmap (bit-packed)
    [v4] encoding tag (1 byte) + compression tag (1 byte)
    [v4] uncompressed_size (uint32)
    typed data array (int64/double/bool/string)
  [v3+] per-column statistics (min/max)
```

### Version history

- **Version 1**: Base format with typed columns and validity bitmaps.

- **Version 2**: Adds per-column annotation strings for Date, POSIXct, and
  factor roundtripping.
- **Version 3**: Adds per-column per-rowgroup statistics (min/max) enabling
  zone-map predicate pushdown.
- **Version 4** (current): Adds a two-layer encoding and compression stack.
  Writing always produces v4. All versions (v1–v4) are readable.

### Encoding and compression (v4)

v4 applies two transformations to each column chunk (one column in one row
group), in order: an **encoding** step and a **compression** step.

Encoding transforms data logically to expose redundancy. The encoder picks the
best encoding per column per row group automatically:

| Encoding | Applies to | Condition | Mechanism |
|:---------|:-----------|:----------|:----------|
| PLAIN | all types | default | Raw bytes, no transformation |
| DICTIONARY | string | unique ratio < 50% | Builds a string dictionary; stores indices as RLE (run-length encoded) |
| DELTA | int64 | monotonically increasing | Stores first value + deltas (all >= 0) |

DICTIONARY encoding is the most impactful for typical categorical string
columns. The encoder counts distinct values in a single pass using an
open-addressing hash table (70% load factor, dynamic resizing). If fewer than
half the values are unique, it emits a dictionary (offset array + packed
strings) followed by RLE-encoded indices. The RLE step collapses runs of
repeated indices into (value, count) pairs, which is effective when rows with
the same category are clustered (e.g. after a sort). If more than half the
values are unique, the encoder aborts dictionary encoding and falls back to
PLAIN with zero overhead.

DELTA encoding stores an initial value followed by the difference between
consecutive values. It targets auto-increment IDs, timestamps, and other
monotonic integer sequences where the deltas are small and compress well.

Compression squeezes bytes physically after encoding. vectra uses a built-in
LZ77 compressor (LZ-VTR), approximately 120 lines of C with no external
dependencies. The compressor uses a 3-byte minimum match, 256-byte maximum
offset, and a hash table for match finding. It skips column chunks smaller than
64 bytes (not worth the overhead). If compression does not reduce size, the
chunk is stored uncompressed. There is no configuration knob; the format always
attempts compression on eligible chunks.

The two-layer design is intentional. Encoding and compression solve different
problems. DICTIONARY and DELTA reduce the entropy of the data (fewer distinct
byte patterns, smaller integer ranges). LZ-VTR then exploits the reduced
entropy at the byte level. Applying both layers yields better ratios than
either layer alone, particularly for RLE-encoded dictionary indices where long
runs of identical small integers compress to near zero.

## Query optimizer

`explain()` runs the optimizer before printing so you see the actual execution
plan. Two optimization passes run automatically:

### Predicate pushdown

When a `FilterNode` sits above a `ScanNode` reading a `.vtr` file (v3+), the
filter predicate is attached to the scan. The scan then applies up to three
pruning strategies on its first `next_batch()` call, in priority order:

1. **Hash index pushdown** (highest priority). If a `.vtri` sidecar index
   exists for the predicate column(s), the scan probes the index to build a
   row-group bitmap. Row groups not in the bitmap are skipped entirely. This
   handles `==` and `%in%` predicates. For composite indexes, AND-combined
   equality predicates on the indexed columns are matched and probed as a
   single composite key. See the Hash indexes section below for details.

2. **Binary search on sorted columns**. If the `.vtr` file records that a
   column is sorted and the predicate is a simple comparison (`==`, `<`, `<=`,
   `>`, `>=`) against a literal, the scan binary-searches the row group stats
   to find the first and last row groups that could contain matching rows. The
   scan range is narrowed to `[first_rg, last_rg)`. For AND-combined
   predicates on the same sorted column (e.g. `x >= 10 & x < 100`), both
   bounds are applied. For OR-combined predicates, the union of both ranges is
   used.

3. **Zone-map pruning** (applied per row group during iteration). Each row
   group in a v3+ file stores per-column min/max statistics. Before reading a
   row group's data from disk, the scan evaluates the pushed-down predicate
   against the row group's stats. If the predicate is provably false for the
   entire row group (e.g. `filter(x > 100)` on a row group where max(x) = 50),
   that row group is skipped entirely without touching the underlying bytes.
   Zone-map pruning handles comparison operators on numeric and string columns,
   AND/OR combinations, and `%in%` predicates (checking whether any set value
   falls within the row group's min/max range). String zone maps use a packed
   8-byte prefix representation for efficient comparison.

These three strategies compose. The scan first applies hash index and binary
search to narrow the set of candidate row groups, then checks zone-map stats on
each candidate before reading data. In `explain()` output, predicate pushdown
appears as `predicate pushdown` and `v3 stats` annotations on the ScanNode.

### Column pruning

The optimizer walks the plan tree top-down and determines which columns each
node actually needs from its child. The required column set at each node is the
union of: columns referenced in the node's own expressions (filter predicates,
mutate expressions, aggregation functions), columns passed through to the
parent, and join key columns. At scan nodes, unneeded columns are excluded from
disk reads by setting a column mask. For a 100-column `.vtr` file where only 3
columns are needed, this means 97 columns are never deserialized, never
decompressed, and never decoded. This is visible in `explain()` as
`3/100 cols (pruned)`. Column pruning applies to all `.vtr` scans regardless
of format version.

### Hidden mutate insertion

When `summarise()` contains nested expressions like `mean(x + y)`, the
optimizer auto-inserts a `ProjectNode` (visible as `hidden mutate` in
`explain()`) to compute the intermediate expression `x + y` before aggregation.
This is necessary because the C-level aggregation nodes (GroupAgg) operate on
single columns, not on expression trees. The hidden mutate computes the
intermediate column and gives it a generated name, which the aggregation node
then references. The same mechanism applies to `sum(log(x))`, `n_distinct(a +
b)`, and similar patterns. The inserted node is visible in `explain()` output
but invisible to the user's pipeline.

## explain() contract

`explain()` prints the optimized plan tree without executing it. The output
shows:

- Node types in execution order (leaf to root)

- Per-node annotations: streaming/materializing, column pruning, predicate
  pushdown, v3 stats, hidden mutate
- Grouping columns if present

- Output schema (column names and types)

```r
tbl("data.vtr") |>
  filter(x > 0) |>
  select(id, x) |>
  explain()
#> vectra execution plan
#>
#> ProjectNode [streaming]
#>   FilterNode [streaming]
#>     ScanNode [streaming, 2/5 cols (pruned), predicate pushdown, v3 stats]
#>
#> Output columns (2):
#>   id <int64>
#>   x <double>
```

The plan tree is a description of what will happen, not a guarantee of how it
will happen internally. Node ordering and naming may change between versions.

## Hash indexes (.vtri)

vectra supports persistent on-disk hash indexes stored as `.vtri` sidecar files
alongside `.vtr` data files. These indexes enable O(1) row group pruning for
equality predicates, turning full-table scans into targeted reads.

### Creating indexes

```r
# Single-column index
create_index(tbl("data.vtr"), "species")

# Case-insensitive index
create_index(tbl("data.vtr"), "species", ci = TRUE)

# Composite (multi-column) index
create_index(tbl("data.vtr"), c("country", "year"))

# Check if an index exists
has_index(tbl("data.vtr"), "species")
```

`create_index()` reads the `.vtr` file, hashes every value in the indexed
column(s) per row group, and writes a `.vtri` file. The file name encodes the
indexed columns: `data.species.vtri` for a single-column index,
`data.country_year.vtri` for a composite.

### Index format

The `.vtri` format is a chained hash table mapping key hashes to row group
indices:

- **v1** (single-column): stores one column index, a case-insensitive flag,
  hash entries, row group entries, a slot heads array, and a next-pointer chain.
  Hashing uses FNV-1a on the raw column bytes. For case-insensitive indexes,
  the hash folds ASCII uppercase to lowercase before hashing.
- **v2** (composite): stores multiple column indices. The composite hash
  combines per-column FNV-1a hashes via XOR and multiplication with the FNV
  prime, producing a single 64-bit hash per row group entry.

### How the scan uses indexes

When a `ScanNode` opens a `.vtr` file, it checks for `.vtri` sidecar files
matching the columns referenced in the pushed-down predicate. For single-column
indexes, it matches `==` predicates and `%in%` predicates. For composite
indexes, it matches AND-combined equality predicates where every indexed column
has an `== literal` clause.

On match, the scan probes the index to produce a row-group bitmap. For `%in%`
predicates, the scan probes once per set element and ORs the bitmaps together.
Row groups with a 0 bit in the bitmap are never read from disk. This is the
first pruning step, applied before binary search and zone-map checks.

### Performance characteristics

Index probing is O(k) where k is the number of query keys (1 for `==`, n for
`%in%`). Each probe is a hash computation plus a chain walk in the slot array.
The index file is memory-mapped at open time, so repeated probes pay no I/O
cost. For tables with many row groups and selective equality predicates, index
pushdown can reduce I/O by orders of magnitude compared to zone-map pruning
alone.

## Materialized blocks

`materialize()` consumes a vectra node and stores the result as a persistent
in-memory columnar block. Unlike nodes, which are consumed on `collect()` and
cannot be reused, blocks persist and support repeated lookups.

```r
blk <- tbl("backbone.vtr") |>
  select(taxonID, canonicalName, genus) |>
  materialize()
```

### Exact lookups

`block_lookup()` performs hash-based lookups on a string column. Hash indices
are built lazily on first use and cached for subsequent calls. The return value
is a data.frame with a `query_idx` column (1-based position in the input keys
vector) plus all columns from the block.

```r
hits <- block_lookup(blk, "canonicalName",
                     c("Quercus robur", "Pinus sylvestris"))

# Case-insensitive
hits <- block_lookup(blk, "canonicalName",
                     c("quercus robur"), ci = TRUE)
```

### Fuzzy lookups

`block_fuzzy_lookup()` computes string distances between query keys and a
block column. Three distance methods are available: Damerau-Levenshtein
(`"dl"`, default), Levenshtein (`"levenshtein"`), and Jaro-Winkler (`"jw"`).
Results are filtered by a maximum normalized distance threshold (default 0.2).

```r
fuzzy <- block_fuzzy_lookup(blk, "canonicalName",
                            c("Quercus robar", "Pinus silvestris"),
                            method = "dl", max_dist = 0.2)
```

An optional blocking column reduces the search space by requiring exact matches
on a second column before computing distances. This is useful for taxonomic
lookups where genus is known:

```r
fuzzy <- block_fuzzy_lookup(blk, "canonicalName",
                            c("Quercus robar"),
                            block_col = "genus",
                            block_keys = c("Quercus"),
                            n_threads = 4L)
```

Fuzzy lookups are OpenMP-parallelized. The `n_threads` parameter controls the
thread count (default 4).

### Use case

Materialized blocks are designed for repeated lookups against a reference table.
The typical pattern is: load a backbone or codelist once with `materialize()`,
then probe it many times with different query vectors. The block stays in memory
across calls, and its internal hash index (for exact lookups) is built once and
reused.

## OpenMP parallelization

The C engine uses OpenMP for CPU-bound operations where the per-element cost is
high enough to justify thread management overhead. Parallelism is conditional:
operations fall back to single-threaded execution when the batch size is below a
threshold or when the platform does not support OpenMP.

### Which operations parallelize

| Operation | Threshold | Schedule | Notes |
|:----------|:----------|:---------|:------|
| `filter()` (selection vector build) | 32,768 rows | parallel prefix sum | Two-phase: count matches per thread, then write at offsets |
| `grepl()` with regex | 1,000 rows | `dynamic, 64` | Per-thread regex compilation for thread safety |
| `levenshtein()`, `dl_dist()`, `jaro_winkler()` | 1,000 rows | `dynamic, 64` | Fuzzy string distance in mutate expressions |
| Sort (merge sort) | 32,768 rows | `task` | Recursive task spawning for parallel merge sort |
| Sort (key extraction) | 32,768 rows | `static` | Parallel extraction of sort keys into index arrays |
| Join (build-side hashing) | 32,768 rows | `static` | Parallel hash computation for build-side keys |
| Join (probe-side hashing) | 32,768 rows | `static` | Parallel hash computation for probe-side keys |
| Window (data copy) | 32,768 rows | `static` | Parallel copy of partition data |
| Window (group dispatch) | 64 groups | `dynamic` | Parallel window computation across groups |
| Collect (column append) | 8 columns | `static` | Parallel column-by-column append to R vectors |
| Zone-map stat computation | 32,768 rows | `static` with `reduction` | Parallel min/max scan during write |
| Block fuzzy lookup | always (if > 0 keys) | `dynamic, 1` per key / `dynamic, 16` per row | Parallelized by query key and by block row |
| Literal fill (broadcast) | 32,768 rows | `static` | Parallel fill for constant columns |

The general-purpose threshold is 32,768 rows (`VEC_OMP_THRESHOLD`). String
operations use a lower threshold of 1,000 rows because their per-element cost
(regex compilation, edit distance matrices) is much higher than arithmetic.

### Thread safety

Regex operations (`grepl` with regex patterns) compile the POSIX regex once per
thread. Each thread owns its own `regex_t` instance, allocated in thread-local
scope inside the `#pragma omp parallel` block. This avoids both contention and
the overhead of recompiling the regex per row.

The filter node uses a parallel prefix sum to build the selection vector without
locking. Each thread counts matches in its chunk, a sequential scan computes
prefix offsets, then each thread writes selected indices at its computed offset.

## Current limitations

- **slice_tail materializes**: There is no reverse-scan optimization.

- **distinct with .keep_all**: Falls back to R when `.keep_all = TRUE` with a
  column subset.

- **Predicate pushdown is .vtr only**: CSV, SQLite, and TIFF scans do not
  benefit from predicate pushdown, column pruning, or hash index acceleration.

- **No SIMD**: Arithmetic and comparison operations use scalar loops. The
  compiler may auto-vectorize some patterns, but there are no explicit SIMD
  intrinsics.

- **OpenMP availability varies**: On macOS, R ships without OpenMP by default.
  Users must install `libomp` (e.g. via Homebrew) and configure the compiler
  flags. On Windows (rtools) and Linux, OpenMP is typically available out of the
  box. When OpenMP is not available, all operations run single-threaded with no
  functional difference.

## Fallback behavior

vectra has these fallback paths to base R:

1. **bind_rows with mismatched column names**: If column names differ across
   inputs, all tables are collected and combined via `rbind()` in R.
2. **distinct with .keep_all and column subset**: Falls back to `duplicated()`
   in R (emits a message).
3. **slice_tail**: Must see all rows to take the last n; returns a data.frame.

4. **slice_min/slice_max with `with_ties = TRUE`** (the default): Collects all
   data to identify ties at the boundary; returns a data.frame.
5. **reframe**: Always collects and evaluates in R; returns a data.frame.

All other operations execute entirely in C. There is no silent fallback to dplyr
or any other package.

## Grouping preservation

All verbs preserve `group_by()` metadata: `filter()`, `select()`, `mutate()`,
`rename()`, `relocate()`, `arrange()`, and `transmute()` pass grouping through.
`rename()` additionally updates group column names to match the rename.
`summarise()` drops grouping according to its `.groups` argument.

## Package conflicts

vectra defines its own S3 generics for dplyr-like verbs (`filter`, `select`,
`mutate`, etc.) and utility functions (`glimpse`, `collect`). If dplyr is also
loaded, whichever package was attached last will mask the other's generics.
vectra's methods will still dispatch correctly on `vectra_node` objects
regardless of masking order.

---
title: "Getting Started with vectra"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with vectra}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Introduction

vectra is a columnar query engine for R. It stores data in a custom binary format (`.vtr`), reads it lazily, and processes it batch-by-batch through a C11 execution engine. The user-facing API looks like dplyr: you chain `filter()`, `mutate()`, `group_by()`, `summarise()`, and the rest. But nothing actually runs until you call `collect()`. That call pulls row groups through a tree of plan nodes, each consuming and producing fixed-size batches. Memory use stays roughly constant regardless of how large the file is.

The execution model is pull-based. When `collect()` fires, the root node calls `next_batch()` on its child, which calls its own child, and so on down to the scan node that reads from disk. Each node transforms one batch and passes it upstream. Filtering uses selection vectors rather than copying rows, so a filter over a billion-row file touches only the matching indices. Sorting spills to temporary files and merges via a k-way min-heap when data exceeds 1 GB. Aggregation uses hash tables that grow proportionally to the number of groups, not the number of input rows. The engine knows four column types: 64-bit integers, doubles, booleans, and variable-length strings, each with a separate validity bitmap for NA tracking.

The question of when to reach for vectra over base R, dplyr, data.table, or arrow comes down to working set size and workflow. If your data fits comfortably in RAM and you prefer in-memory semantics, dplyr and data.table are mature, fast, and well-supported. Arrow's `open_dataset()` handles partitioned Parquet and multi-file scans well and interoperates with the broader Arrow ecosystem (Spark, DuckDB, Flight). vectra fills a different niche: dplyr syntax on single-file storage with memory-bounded execution and no heavy compiled dependencies. The entire C codebase compiles from C11 source with no external library dependencies — compression and deflate support are vendored in the package itself. There is no C++ compiler requirement, no Boost, no protobuf. A 20 GB `.vtr` file processes on a laptop with 8 GB of RAM because the scan node reads one row group at a time and the downstream pipeline never holds more than two batches in memory simultaneously.

The on-disk format uses dictionary encoding for low-cardinality string columns, delta encoding for monotonic integer sequences, and Zstandard compression at the byte level. File sizes are competitive with Parquet for typical tabular workloads. Zone-map statistics (per-row-group min/max) are stored in the header, enabling the scan node to skip irrelevant row groups before decompressing any column data. Optional hash indexes stored as `.vtri` sidecar files accelerate equality predicates to O(1) row group lookup.

This vignette is a taste test. Each section introduces a feature area with runnable examples, then points to the relevant deep-dive vignette for full coverage. All code uses tempfiles and built-in datasets, so you can run every chunk on any machine with vectra installed.

## Writing and reading data

vectra uses a binary columnar format with the `.vtr` extension. A `.vtr` file contains a header (format version, column schema, and row group count), a row group index for fast seeking, and then the actual column data per row group. Each column within a row group is stored as a typed array with a validity bitmap for NA values. The current format version (v4) applies a two-stage encoding stack per column per row group. The first stage is a logical encoding: DICTIONARY for string columns where fewer than 50% of values are unique (the dictionary itself is stored once, and index runs are RLE-compressed), DELTA for int64 columns with monotonically increasing values (storing first value plus deltas), or PLAIN for everything else. The second stage is byte-level compression: Zstandard when the compressed output is smaller than the input and the raw data exceeds 64 bytes, or NONE otherwise. The file reader handles v1 through v4 transparently; the writer always produces v4.

Write any data.frame to `.vtr` with `write_vtr()`. Open it lazily with `tbl()`, which returns a `vectra_node` object. No data is read at this point. The `tbl()` call reads only the file header to learn the schema and row group layout.

```{r write-read}
library(vectra)

f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

node <- tbl(f)
node
```

The print method shows the schema and file path but does not scan any data. Call `collect()` to pull all batches through the plan and materialize the result as a data.frame.

```{r collect}
tbl(f) |> collect() |> head()
```

For large datasets, `write_vtr()` accepts a `batch_size` parameter that controls how many rows go into each row group. Smaller row groups give finer-grained zone-map pruning (the scan node skips row groups whose min/max statistics exclude your filter predicate) but add per-group overhead. The default, one row group per data.frame, works well for most in-memory writes. For streaming writes from a pipeline, the upstream batch size is preserved.

```{r write-batch-size}
f_batched <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f_batched, batch_size = 10)
tbl(f_batched) |> collect() |> nrow()
```

The row count is the same. The difference is internal layout: 4 row groups of approximately 10 rows each instead of 1 row group of 32 rows. With smaller row groups, zone-map pruning becomes more selective. If `mpg` ranges from 10.4 to 33.9 across the full dataset but only 10.4 to 15.2 within one particular row group, a `filter(mpg > 20)` can skip that row group entirely without decompressing any column data. The tradeoff is overhead: each row group carries its own encoding metadata and compression frame, so very small row groups (under a few hundred rows) waste space on headers. For most analytical workloads, row groups between 10,000 and 100,000 rows strike a good balance.

## Filtering and selecting

`filter()` keeps rows matching one or more conditions. Comma-separated conditions inside a single `filter()` call are combined with AND. Use `|` for OR within a single expression. You can also chain multiple `filter()` calls, which stacks FilterNodes in the plan. The engine evaluates them sequentially, one batch at a time, so the second filter only sees rows that passed the first.

```{r filter-and}
tbl(f) |>
  filter(cyl == 6, mpg > 19) |>
  select(mpg, cyl, hp, wt) |>
  collect()
```

The C engine evaluates the predicate batch-by-batch and sets a selection vector on each `VecBatch`. The selection vector is an integer array of indices that passed the predicate. Downstream nodes (project, sort, aggregation) read only those indices, so no column data is copied or compacted at the filter stage. This zero-copy approach means filtering a 50 GB file with a 1% hit rate allocates memory proportional to 1% of one batch, not 1% of the entire file.

NA values in comparisons follow SQL semantics: `NA == 6` evaluates to NA, not FALSE. Rows where the predicate evaluates to NA are excluded from the selection vector, the same as rows that evaluate to FALSE. To explicitly test for NA, use `is.na()` in the predicate.

OR conditions and parenthesized expressions work as expected.

```{r filter-or}
tbl(f) |>
  filter(cyl == 4 | cyl == 8) |>
  select(mpg, cyl) |>
  collect() |>
  head()
```

For set membership, `%in%` is accelerated with a hash set when the right-hand side has more than a handful of values. On a column with 10 million rows and a set of 500 target values, the hash-based `%in%` is roughly 10x faster than repeated OR comparisons.

```{r filter-in}
tbl(f) |>
  filter(cyl %in% c(4, 6)) |>
  select(mpg, cyl) |>
  collect() |>
  head()
```

`select()` picks columns by name or by tidyselect helper. The full tidyselect vocabulary works: `starts_with()`, `ends_with()`, `contains()`, `matches()` (regex), `where()` (type predicate), `everything()`, `last_col()`, and negation via `-` or `!`. Column renaming inside `select()` also works (`select(miles = mpg)` renames on the fly). Under the hood, `select()` inserts a ProjectNode that prunes columns from the batch before they reach the next verb, so columns you drop are never decompressed from disk in subsequent row groups.

```{r select-helpers}
tbl(f) |>
  select(starts_with("d"), mpg) |>
  collect() |>
  head()
```

Negation drops columns.

```{r select-negate}
tbl(f) |>
  select(-am, -vs, -gear, -carb) |>
  collect() |>
  head()
```

`explain()` prints the query plan tree. We can see how `filter()` and `select()` map to plan nodes, and where predicate pushdown applies.

```{r explain-filter}
tbl(f) |>
  filter(cyl > 4) |>
  select(mpg, cyl, hp) |>
  explain()
```

The ProjectNode prunes columns before the data reaches R. The FilterNode sits below it, and the ScanNode reads from disk. When a hash index exists on the filter column, the scan node probes the `.vtri` sidecar index and builds a bitmap of row groups that could contain matching rows. Row groups not in the bitmap are skipped entirely. This composes with zone-map pruning: even without an explicit index, the scan node compares the filter predicate against per-row-group min/max statistics and skips row groups that cannot match. Between these two mechanisms, a point query on an indexed column over a 10 GB file with 1000 row groups might read only 1 row group.

## Transforming columns

`mutate()` adds or replaces columns. All expression evaluation happens in the C engine: the R layer captures the expression as an abstract syntax tree (via `substitute()`), serializes it into a nested list structure, and passes it across the `.Call()` bridge. The C side walks the expression tree and evaluates it element-wise over the batch. Supported operators include arithmetic (`+`, `-`, `*`, `/`, `%%`), comparison (`==`, `!=`, `<`, `<=`, `>`, `>=`), boolean (`&`, `|`, `!`), and unary negation. There is no R-level evaluation per row, which is why `mutate()` on 100 million rows runs at C speed regardless of expression complexity.

```{r mutate-arith}
tbl(f) |>
  mutate(kpl = mpg * 0.425144, hp_per_wt = hp / wt) |>
  select(mpg, kpl, hp, wt, hp_per_wt) |>
  collect() |>
  head()
```

Math functions are evaluated natively in C: `abs()`, `sqrt()`, `log()`, `log2()`, `log10()`, `exp()`, `floor()`, `ceiling()`, `round()`, `sign()`, `trunc()`. For element-wise binary min/max, `pmin()` and `pmax()` take two arguments (columns or literals) and return the smaller or larger of each pair. These compose with arithmetic to build clamping expressions.

```{r mutate-math}
tbl(f) |>
  mutate(
    log_hp = log(hp),
    hp_floor = floor(hp / 10) * 10,
    bounded = pmin(pmax(mpg, 15), 25)
  ) |>
  select(hp, log_hp, hp_floor, mpg, bounded) |>
  collect() |>
  head()
```

`transmute()` works like `mutate()` but drops all columns except the new ones.

```{r transmute}
tbl(f) |>
  transmute(
    efficiency = mpg / wt,
    power_ratio = hp / disp
  ) |>
  collect() |>
  head()
```

Type casting converts between vectra's four internal column types. `as.numeric()` and `as.double()` cast to double, `as.integer()` casts to int64, `as.character()` casts to string, and `as.logical()` casts to bool. Casting follows R semantics: `as.integer()` on a double truncates toward zero, `as.numeric()` on a string parses the number or returns NA, and `as.character()` on a numeric formats it as a decimal string. These casts happen per-element in C and respect the validity bitmap, so NA values stay NA after casting.

```{r mutate-cast}
tbl(f) |>
  mutate(cyl_str = as.character(cyl)) |>
  select(cyl, cyl_str) |>
  collect() |>
  head(3)
```

`is.na()` tests for missing values by checking the validity bitmap directly, without touching the data buffer. `if_else()` and `case_when()` handle conditional logic. Unlike base R's `ifelse()`, `if_else()` is type-strict: both branches must have the same type, and the result preserves NA from the condition (matching dplyr semantics). `case_when()` evaluates formulas top to bottom and assigns the value from the first matching condition. `coalesce()` returns the first non-NA value from an arbitrary number of columns or literals. `between()` is shorthand for `x >= left & x <= right`, compiled to a single AND expression internally.

```{r mutate-control}
tbl(f) |>
  mutate(
    size = case_when(
      cyl == 4 ~ "small",
      cyl == 6 ~ "medium",
      cyl == 8 ~ "large"
    ),
    mpg_class = if_else(mpg > 20, "high", "low"),
    in_range = between(hp, 100, 200)
  ) |>
  select(cyl, size, mpg, mpg_class, hp, in_range) |>
  collect() |>
  head()
```

`case_when()` accepts a `.default` argument for the fallback value. Formulas are evaluated top to bottom; the first matching condition wins.

`coalesce()` takes any number of arguments and returns the first non-NA value for each row. It works with both column references and literal values.

```{r mutate-coalesce}
df_na <- data.frame(
  a = c(NA, 2, NA, 4),
  b = c(10, NA, NA, 40),
  stringsAsFactors = FALSE
)
f_na <- tempfile(fileext = ".vtr")
write_vtr(df_na, f_na)

tbl(f_na) |>
  mutate(filled = coalesce(a, b, 0)) |>
  collect()
```

## String operations

vectra evaluates string functions in C using a two-pass strategy. The first pass walks the input strings and computes the total output byte count. The second pass allocates a single contiguous buffer of that size and fills it. This avoids per-row `malloc`/`realloc` overhead that would otherwise dominate runtime on columns with millions of short strings. For regex operations (`grepl` with `fixed = FALSE`, `gsub`, `sub`, `str_extract`), the engine uses POSIX regex compiled once per batch (or once per thread under OpenMP), supporting full backreference syntax (`\\1`, `\\2`, etc.) and capture groups.

We will use a small dataset with character columns for the examples. The patterns here are representative of what you would do on larger string columns; the C engine processes them identically regardless of batch size.

```{r string-data}
people <- data.frame(
  name = c("  Alice  ", "Bob", "Charlie Brown", "Diana"),
  city = c("Amsterdam", "Berlin", "Chicago", "Dublin"),
  email = c("alice@example.com", "bob@test.org",
            "charlie.b@work.net", "diana@example.com"),
  stringsAsFactors = FALSE
)
fs <- tempfile(fileext = ".vtr")
write_vtr(people, fs)
```

`nchar()` returns the number of characters in each string (not bytes; the engine is UTF-8 aware for multi-byte characters). `substr()` extracts a substring by 1-indexed start and stop positions, matching base R semantics. `trimws()` strips leading and trailing whitespace (spaces, tabs, newlines).

```{r string-basic}
tbl(fs) |>
  mutate(
    name_trimmed = trimws(name),
    name_len = nchar(trimws(name)),
    city_prefix = substr(city, 1, 3)
  ) |>
  select(name_trimmed, name_len, city_prefix) |>
  collect()
```

`toupper()` and `tolower()` convert case. `startsWith()` and `endsWith()` test prefixes and suffixes, returning a boolean column.

```{r string-case}
tbl(fs) |>
  mutate(
    city_upper = toupper(city),
    is_example = endsWith(email, "example.com"),
    starts_a = startsWith(city, "A")
  ) |>
  select(city_upper, email, is_example, starts_a) |>
  collect()
```

`grepl()` tests for pattern matches. The default is fixed matching (literal substring search), which is the fastest path and uses a simple `strstr()` internally. Pass `fixed = FALSE` for POSIX regex, which compiles the pattern once per batch and applies it element-wise. On columns with more than 1000 rows, regex `grepl` is OpenMP-parallelized with per-thread regex compilation for thread safety.

```{r string-grepl}
tbl(fs) |>
  mutate(has_at = grepl("@example", email)) |>
  select(email, has_at) |>
  collect()
```

`gsub()` and `sub()` perform string replacement. `gsub()` replaces all occurrences of the pattern; `sub()` replaces only the first. Like `grepl()`, the default is fixed matching. Pass `fixed = FALSE` for regex with full backreference support. In regex mode, the replacement string can reference capture groups with `\\1`, `\\2`, and so on, following the POSIX extended regex specification.

```{r string-gsub}
tbl(fs) |>
  mutate(domain = gsub(".*@", "", email, fixed = FALSE)) |>
  select(email, domain) |>
  collect()
```

`str_extract()` pulls out the first match of a regex pattern and returns it as a string. When there is no match, the result is NA. If the pattern contains a capture group (parentheses), `str_extract()` returns the captured portion rather than the full match, which makes it possible to extract structured substrings like domains from URLs or species epithets from binomial names.

```{r string-extract}
tbl(fs) |>
  mutate(user = str_extract(email, "^[^@]+")) |>
  select(email, user) |>
  collect()
```

`paste()` and `paste0()` concatenate columns and literals. Both are N-ary: they accept any number of arguments, mixing column references and string literals freely. `paste()` takes a `sep` argument (default `" "`) that is inserted between each pair of arguments; `paste0()` concatenates with no separator. Internally, these use the two-pass allocation strategy: the first pass sums up the output lengths across all arguments, allocates once, then the second pass writes the concatenated strings into the pre-sized buffer.

```{r string-paste}
tbl(fs) |>
  mutate(
    greeting = paste0("Hello, ", trimws(name), "!"),
    label = paste(trimws(name), city, sep = " - ")
  ) |>
  select(greeting, label) |>
  collect()
```

For a complete reference of string operations, regex patterns, and performance characteristics on large columns, see `vignette("string-ops")`.

## Aggregation

`group_by()` marks columns as grouping keys. It is metadata-only: no plan node is created, no data is scanned, and the grouping information simply attaches to the `vectra_node` as an R-level attribute. The C engine only learns about the groups when `summarise()` (or equivalently `summarize()`) is called, at which point a GroupAggNode is created in the plan. This node builds a hash table keyed on the group columns and accumulates aggregation results in a single pass through the input batches. Each batch's rows are hashed, looked up in the hash table, and their values are folded into the running aggregation state. Once all batches are consumed, the hash table entries are flushed as the output.

vectra supports 14 aggregation functions, all evaluated in C: `n()`, `sum()`, `mean()`, `min()`, `max()`, `sd()`, `var()`, `first()`, `last()`, `any()`, `all()`, `median()`, `n_distinct()`. All except `n()` take a column name as their first argument. All except `n()` and `n_distinct()` accept `na.rm = TRUE` to skip NA values. Without `na.rm`, any NA in a group poisons the result: `sum()` returns NA, `mean()` returns NA, and so on. This matches base R semantics. With `na.rm = TRUE`, the behavior also matches R: `sum()` on all-NA returns 0, `mean()` on all-NA returns NaN, and `min()`/`max()` on all-NA return Inf/-Inf with a warning.

```{r summarise-basic}
tbl(f) |>
  group_by(cyl) |>
  summarise(
    count = n(),
    avg_mpg = mean(mpg),
    total_hp = sum(hp),
    best_mpg = max(mpg)
  ) |>
  collect()
```

The result has one row per group. Groups are not guaranteed to be in any particular order unless you add an `arrange()` afterward.

`sd()` and `var()` compute sample standard deviation and variance using the Welford one-pass algorithm (denominator `n - 1`, matching R's `sd()` and `var()`). `first()` and `last()` return the first and last non-NA values encountered in scan order within each group. `any()` and `all()` reduce boolean (or boolean-coercible) columns, returning TRUE/FALSE per group. These are useful for flagging groups that contain at least one positive observation or checking whether all rows in a group satisfy a condition.

```{r summarise-advanced}
tbl(f) |>
  group_by(cyl) |>
  summarise(
    mpg_sd = sd(mpg),
    mpg_var = var(mpg),
    first_hp = first(hp),
    last_hp = last(hp)
  ) |>
  collect()
```

`median()` uses a per-group dynamic array and quicksort internally. `n_distinct()` uses a per-group open-addressing hash set with FNV-1a hashing at 70% load factor.

```{r summarise-median}
tbl(f) |>
  group_by(cyl) |>
  summarise(
    med_mpg = median(mpg),
    unique_gears = n_distinct(gear)
  ) |>
  collect()
```

For quick counts, `count()` and `tally()` are convenient shorthands. `count()` takes grouping columns inline; `tally()` uses existing groups.

```{r count}
tbl(f) |>
  count(cyl, sort = TRUE) |>
  collect()
```

```{r tally}
tbl(f) |>
  group_by(gear) |>
  tally() |>
  collect()
```

`across()` applies the same function or set of functions to multiple columns at once. It works inside both `summarise()` and `mutate()`, accepting tidyselect column specifications (including helpers like `where(is.numeric)`, `starts_with()`, etc.) and functions as either a single function, a formula (`~ .x + 1`), or a named list of functions. The `.names` argument controls output column naming with `{.col}` and `{.fn}` glue-style placeholders, so you get predictable names like `mpg_avg`, `hp_total` without manual repetition. Under the hood, `across()` is expanded at the R level into individual aggregation expressions before the plan reaches C, so there is no runtime overhead compared to writing each expression by hand.

```{r across-summarise}
tbl(f) |>
  group_by(cyl) |>
  summarise(across(c(mpg, hp, wt), mean)) |>
  collect()
```

With multiple functions and a naming pattern:

```{r across-multi}
tbl(f) |>
  group_by(cyl) |>
  summarise(across(
    c(mpg, hp),
    list(avg = mean, total = sum),
    .names = "{.col}_{.fn}"
  )) |>
  collect()
```

`ungroup()` removes grouping metadata from a node. This is occasionally needed when you want to apply a second `summarise()` with different groups, or when you want window functions to operate on the full dataset rather than per-group.

```{r ungroup}
tbl(f) |>
  group_by(cyl, gear) |>
  summarise(n = n(), .groups = "keep") |>
  ungroup() |>
  arrange(desc(n)) |>
  collect()
```

The `.groups` argument to `summarise()` controls what happens to grouping after aggregation. `"drop_last"` (the default) peels off the last grouping column, `"drop"` removes all groups, and `"keep"` preserves them.

## Sorting and slicing

`arrange()` sorts rows. Wrap column names in `desc()` for descending order. Sort is stable: rows with equal sort keys preserve their original order.

```{r arrange}
tbl(f) |>
  select(mpg, cyl, hp) |>
  arrange(cyl, desc(mpg)) |>
  collect() |>
  head(8)
```

Under the hood, `arrange()` uses an external merge sort with a 1 GB memory budget. When data exceeds that limit, sorted runs spill to temporary `.vtr` files and merge via a k-way min-heap. NAs sort last in ascending order and first in descending order.

`slice_head()` and `slice_tail()` return the first or last `n` rows. `slice_head()` is streaming (it creates a LimitNode that stops pulling after `n` rows). `slice_tail()` must materialize all rows to know the total count.

```{r slice-head}
tbl(f) |>
  slice_head(n = 5) |>
  collect()
```

`slice_min()` and `slice_max()` use an optimized top-N algorithm (a bounded heap) that avoids a full sort. This matters when you want the 10 smallest values from a 100-million-row dataset: the heap holds at most 10 entries at any point.

```{r slice-min}
tbl(f) |>
  select(mpg, cyl, hp) |>
  slice_min(order_by = mpg, n = 3) |>
  collect()
```

By default, `slice_min()` and `slice_max()` include ties. The result may contain more than `n` rows if multiple rows share the boundary value. Use `with_ties = FALSE` for exactly `n` rows.

```{r slice-no-ties}
tbl(f) |>
  select(mpg, cyl) |>
  slice_min(order_by = cyl, n = 3, with_ties = FALSE) |>
  collect()
```

```{r slice-max}
tbl(f) |>
  select(mpg, cyl, hp) |>
  slice_max(order_by = hp, n = 4, with_ties = FALSE) |>
  collect()
```

## Joins

vectra implements hash joins using a build-right, probe-left strategy. The entire right-side table is consumed and stored in a hash table keyed on the join columns. Then left-side batches stream through one at a time, probing the hash table for matches. Memory cost is proportional to the number of rows in the right-side table, so the standard practice is to put the smaller table on the right. For a left table with 500 million rows and a right table with 50,000 rows, the hash table holds 50,000 entries and the left side streams through in constant memory.

Seven join types are available. `left_join()` keeps all left rows and fills unmatched right columns with NA. `inner_join()` keeps only rows with matches on both sides. `right_join()` keeps all right rows. `full_join()` keeps everything from both sides, filling NA where there is no match. `semi_join()` keeps left rows that have at least one match on the right, without adding any right-side columns. `anti_join()` keeps left rows that have no match. `cross_join()` produces the Cartesian product (every left row paired with every right row). NA keys never match in any join type, following SQL NULL semantics.

```{r join-setup}
cyl_info <- data.frame(
  cyl = c(4, 6, 8),
  engine_type = c("inline", "v-type", "v-type"),
  stringsAsFactors = FALSE
)
f_cyl <- tempfile(fileext = ".vtr")
write_vtr(cyl_info, f_cyl)
```

`left_join()` keeps all rows from the left table. Unmatched right-side columns are filled with NA.

```{r left-join}
tbl(f) |>
  select(mpg, cyl, hp) |>
  left_join(tbl(f_cyl), by = "cyl") |>
  collect() |>
  head()
```

`inner_join()` keeps only rows with matching keys on both sides. `semi_join()` keeps left-side rows that have a match on the right but does not add any right-side columns. `anti_join()` keeps left-side rows that have no match.

```{r semi-anti}
tbl(f) |>
  select(mpg, cyl) |>
  anti_join(
    tbl(f_cyl) |> filter(engine_type == "v-type"),
    by = "cyl"
  ) |>
  collect() |>
  head()
```

That returns only the 4-cylinder cars, since both 6- and 8-cylinder entries are in the "v-type" subset.

For joins on differently named columns, use a named vector in `by`. Multi-column keys work by passing multiple names.

```{r join-named}
ratings <- data.frame(
  cylinders = c(4, 6, 8),
  rating = c("A", "B", "C"),
  stringsAsFactors = FALSE
)
f_rat <- tempfile(fileext = ".vtr")
write_vtr(ratings, f_rat)

tbl(f) |>
  select(mpg, cyl) |>
  inner_join(tbl(f_rat), by = c("cyl" = "cylinders")) |>
  collect() |>
  head()
```

The `suffix` argument (default `c(".x", ".y")`) controls how non-key columns with the same name on both sides are disambiguated. For example, if both tables have a column called `value`, the output will contain `value.x` from the left and `value.y` from the right. Multi-column keys work by passing a character vector to `by`, either unnamed (same column names on both sides) or named (mapping left names to right names).

Key types are auto-coerced following the `bool < int64 < double` promotion hierarchy. If the left key is int64 and the right key is double, the int64 values are promoted to double before hashing. Joining a string column against a numeric column is an error and fails immediately with a descriptive message.

vectra also supports `fuzzy_join()` for approximate string matching between tables. It computes string distances between key columns and keeps pairs within a normalized distance threshold.

```{r fuzzy-join}
ref_species <- data.frame(
  canonical = c("Quercus robur", "Quercus petraea",
                 "Fagus sylvatica"),
  code = c("QR", "QP", "FS"),
  stringsAsFactors = FALSE
)
query_species <- data.frame(
  name = c("Quercus robur", "Qurecus petraea",
           "Fagus sylvatca"),
  stringsAsFactors = FALSE
)
f_ref <- tempfile(fileext = ".vtr")
f_query <- tempfile(fileext = ".vtr")
write_vtr(ref_species, f_ref)
write_vtr(query_species, f_query)

tbl(f_query) |>
  fuzzy_join(
    tbl(f_ref),
    by = c("name" = "canonical"),
    method = "dl",
    max_dist = 0.15
  ) |>
  collect()
```

The `method` argument accepts `"dl"` (Damerau-Levenshtein), `"levenshtein"`, or `"jw"` (Jaro-Winkler). The `block_by` parameter restricts comparisons to rows sharing an exact-match value on a second column, which can reduce runtime by orders of magnitude on large reference tables. For full documentation, see `vignette("joins")`.

## Window functions

Window functions compute values row-by-row in the context of the full dataset or a per-group partition. Unlike aggregation, which collapses groups to single rows, window functions preserve the original row count. They are detected inside `mutate()` by function name and automatically routed to a dedicated WindowNode in the execution plan. The WindowNode is a materializing operation: it consumes all input rows (per partition) before producing output, because functions like `rank()` need to see all values before assigning ranks.

vectra supports 12 window functions organized in three families. Ranking functions: `row_number()` (no argument, sequential 1..n), `rank()` and `dense_rank()` (take a column argument, with `rank()` leaving gaps on ties and `dense_rank()` not), `ntile(n)` (divides rows into `n` roughly equal buckets), `percent_rank()` and `cume_dist()` (take a column argument, return normalized rank as a double between 0 and 1). Offset functions: `lag()` and `lead()`. Cumulative functions: `cumsum()`, `cummean()`, `cummin()`, `cummax()`.

```{r window-rank}
tbl(f) |>
  select(mpg, cyl, hp) |>
  slice_head(n = 8) |>
  mutate(
    rn = row_number(),
    mpg_rank = rank(mpg),
    mpg_dense = dense_rank(mpg)
  ) |>
  collect()
```

`lag()` and `lead()` shift a column forward or backward by `n` positions (default 1). `lag(mpg)` gives each row the `mpg` value from the previous row; `lead(mpg)` gives the value from the next row. The `default` argument specifies the fill value when the offset falls outside the partition boundary (the first row for `lag`, the last row for `lead`). Without `default`, these positions are NA.

```{r window-lag-lead}
tbl(f) |>
  select(mpg, hp) |>
  slice_head(n = 6) |>
  mutate(
    prev_mpg = lag(mpg),
    next_mpg = lead(mpg),
    prev2_hp = lag(hp, n = 2, default = 0)
  ) |>
  collect()
```

Cumulative functions: `cumsum()`, `cummean()`, `cummin()`, `cummax()`.

```{r window-cum}
tbl(f) |>
  select(mpg, hp) |>
  slice_head(n = 6) |>
  mutate(
    running_hp = cumsum(hp),
    running_avg = cummean(mpg),
    running_min = cummin(mpg)
  ) |>
  collect()
```

Window functions respect grouping. When a node is grouped, partitions are computed independently.

```{r window-grouped}
tbl(f) |>
  select(mpg, cyl) |>
  group_by(cyl) |>
  mutate(rn = row_number(), pct = percent_rank(mpg)) |>
  slice_head(n = 10) |>
  collect()
```

Each `cyl` group gets its own `row_number()` sequence starting from 1 and its own `percent_rank()` distribution. The grouped window node partitions the data by the group keys, materializes each partition independently, computes the window functions within the partition, and then concatenates the results. If a `mutate()` call mixes window functions with regular expressions (e.g., `mutate(rn = row_number(), double_mpg = mpg * 2)`), the R layer splits them: window functions go to a WindowNode, regular expressions go to a ProjectNode stacked on top. The split is invisible to the caller.

## Dates and times

Date and POSIXct columns roundtrip through vectra. Internally, dates are stored as doubles: days since the Unix epoch (1970-01-01) for Date objects, seconds since epoch for POSIXct. The R class attribute is preserved in the `.vtr` schema metadata, so when you `collect()`, the output columns have the correct R class restored. Component extraction functions work in `mutate()` and `filter()`: `year()`, `month()`, `day()` for date parts, and `hour()`, `minute()`, `second()` for time parts on POSIXct columns. These are evaluated in C by computing the Gregorian calendar decomposition from the epoch offset, so there is no R-level overhead per row.

```{r date-data}
events <- data.frame(
  event_date = as.Date(c("2020-03-15", "2020-07-01",
                          "2021-01-15", "2021-06-30")),
  event_time = as.POSIXct(c("2020-03-15 09:30:00",
                             "2020-07-01 14:00:00",
                             "2021-01-15 08:15:00",
                             "2021-06-30 17:45:00"),
                           tz = "UTC"),
  value = c(10, 20, 30, 40)
)
fd <- tempfile(fileext = ".vtr")
write_vtr(events, fd)
```

Extract date components and aggregate by year.

```{r date-extract}
tbl(fd) |>
  mutate(
    yr = year(event_date),
    mo = month(event_date),
    dy = day(event_date)
  ) |>
  group_by(yr) |>
  summarise(total = sum(value)) |>
  collect()
```

Extract time components from POSIXct.

```{r time-extract}
tbl(fd) |>
  mutate(
    hr = hour(event_time),
    mn = minute(event_time)
  ) |>
  select(event_time, hr, mn) |>
  collect()
```

Filter by date using `as.Date()` to convert a string literal to a date value at parse time. The C engine receives the numeric date value and compares directly.

```{r date-filter}
tbl(fd) |>
  filter(event_date >= as.Date("2021-01-01")) |>
  collect()
```

Date arithmetic works through the numeric representation. Adding 30 to a Date column advances it by 30 days. Subtracting two date columns gives a difference in days. You can combine date extraction with aggregation: `group_by(yr = year(event_date)) |> summarise(total = sum(value))` works as expected because `mutate()` expressions are allowed inline inside `group_by()` through the standard dplyr convention of named expressions.

```{r date-arith}
tbl(fd) |>
  mutate(plus_30 = event_date + 30) |>
  select(event_date, plus_30) |>
  collect()
```

The `plus_30` column is numeric (days since epoch). To interpret it as a Date in R after collecting, wrap it with `as.Date(x, origin = "1970-01-01")`.

## String similarity

vectra computes edit distances directly in the C engine, with OpenMP parallelization for columns exceeding 1000 rows. Three distance algorithms are available.

`levenshtein()` counts insertions, deletions, and substitutions. `dl_dist()` (Damerau-Levenshtein) additionally counts transpositions of adjacent characters as a single edit, which better captures common typos like "hte" for "the". `jaro_winkler()` returns a similarity score from 0 to 1, where 1 means identical. It weights early-character matches more heavily, making it well-suited for short strings like personal names.

```{r similarity-data}
species <- data.frame(
  name = c("Quercus robur", "Quercus rubra",
           "Fagus sylvatica", "Acer platanoides",
           "Quercus petraea"),
  stringsAsFactors = FALSE
)
fs2 <- tempfile(fileext = ".vtr")
write_vtr(species, fs2)
```

Compute all three metrics against a reference string and filter by edit distance.

```{r similarity-metrics}
tbl(fs2) |>
  mutate(
    lev = levenshtein(name, "Quercus robur"),
    dl = dl_dist(name, "Quercus robur"),
    jw = jaro_winkler(name, "Quercus robur")
  ) |>
  filter(lev <= 5) |>
  arrange(lev) |>
  collect()
```

Normalized variants `levenshtein_norm()` and `dl_dist_norm()` divide the raw distance by the length of the longer string, returning a value between 0 and 1. This makes thresholds transferable across strings of different lengths.

```{r similarity-norm}
tbl(fs2) |>
  mutate(
    lev_norm = levenshtein_norm(name, "Quercus robur"),
    dl_norm = dl_dist_norm(name, "Quercus robur")
  ) |>
  collect()
```

Damerau-Levenshtein catches transposition errors that pure Levenshtein scores as two edits.

```{r dl-transposition}
tbl(fs2) |>
  mutate(
    lev = levenshtein(name, "Qurecus robur"),
    dl = dl_dist(name, "Qurecus robur")
  ) |>
  collect()
```

The "ue" / "re" swap in "Qurecus" counts as 2 edits for Levenshtein (delete + insert) but 1 for Damerau-Levenshtein (transposition). For full coverage of distance algorithms, threshold tuning, and blocking strategies, see `vignette("string-ops")`.

## Tree traversal

`resolve()` and `propagate()` are scalar self-join functions designed for hierarchical (parent-child) data stored in a single table. They operate within a batch by matching a foreign key column to a primary key column in the same batch, which means the entire hierarchy must fit within one row group (or the data must be collected first). These functions are particularly useful for taxonomic backbones, organizational charts, and bill-of-materials structures where each row has a `parent_id` pointing to another row's `id`.

`resolve(fk, pk, value)` returns the `value` column from the row where `pk == fk`. It is a single-level lookup, equivalent to a self left-join on `(fk, pk)`, but without the overhead of materializing a hash table. The lookup is O(n) per batch (linear scan), so it works well for hierarchies with up to a few million rows.

```{r resolve}
taxa <- data.frame(
  id        = c(1L, 2L, 3L, 4L),
  name      = c("Fagaceae", "Quercus", "Q. robur", "Q. petraea"),
  parent_id = c(NA, 1L, 2L, 2L),
  stringsAsFactors = FALSE
)
ft <- tempfile(fileext = ".vtr")
write_vtr(taxa, ft)

tbl(ft) |>
  mutate(parent_name = resolve(parent_id, id, name)) |>
  collect()
```

Row 1 (Fagaceae) has `parent_id = NA`, so `parent_name` is NA. Row 3 (Q. robur) has `parent_id = 2`, which matches `id = 2` (Quercus), so `parent_name` is "Quercus".

`propagate(parent_fk, pk, seed)` walks the tree upward from each row and fills every node with the nearest non-NA ancestor value of `seed`. Unlike `resolve()`, which does a single-level lookup, `propagate()` traverses the full ancestor chain until it finds a row where the seed expression is non-NA. This denormalizes a hierarchy in a single pass, which would otherwise require a recursive CTE in SQL or repeated self-joins in dplyr.

```{r propagate}
tbl(ft) |>
  mutate(family = propagate(
    parent_id, id,
    if_else(is.na(parent_id), name, NA_character_)
  )) |>
  collect()
```

The seed expression assigns "Fagaceae" to the root (the row with no parent) and NA everywhere else. `propagate()` then walks each row's ancestry until it finds "Fagaceae" and fills it in. All four rows now carry the family name.

## Format backends

vectra reads and writes several file formats beyond `.vtr`. Each backend produces a `vectra_node` that plugs into the same verb pipeline, so you can `filter()`, `mutate()`, `group_by()`, and `summarise()` on CSV files, SQLite databases, Excel workbooks, and GeoTIFF rasters using identical syntax. The key advantage is streaming: format conversion pipelines (e.g., CSV to `.vtr`) process data batch-by-batch without loading the entire file into memory.

`tbl_csv()` scans a CSV file lazily. Column types (int64, double, bool, string) are inferred from the first 1000 rows by probing each field for numeric, boolean, and date patterns. Gzip-compressed files (`.csv.gz`) are detected automatically by file extension and decompressed on the fly. The `batch_size` parameter controls how many rows are read per batch (default 65,536).

```{r csv-roundtrip}
csv_in <- tempfile(fileext = ".csv")
write.csv(mtcars, csv_in, row.names = FALSE)

tbl_csv(csv_in) |>
  filter(cyl == 6) |>
  select(mpg, cyl, hp) |>
  collect()
```

`tbl_sqlite()` scans a SQLite table. Column types are derived from the CREATE TABLE schema. All filtering and aggregation happens in vectra's C engine, not in SQLite.

```{r sqlite-roundtrip}
db <- tempfile(fileext = ".sqlite")
f_src <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f_src)
tbl(f_src) |> write_sqlite(db, "cars")

tbl_sqlite(db, "cars") |>
  filter(mpg > 25) |>
  collect()
```

`tbl_xlsx()` reads an Excel sheet via the openxlsx2 package (a Suggests dependency, not required for other backends). The sheet is read into memory as a data.frame, then converted to a vectra node. This is not a streaming operation (the full sheet is loaded), but it makes it possible to chain vectra verbs on Excel data without writing intermediate files. The `sheet` argument accepts a name or 1-based integer index.

`tbl_tiff()` reads GeoTIFF rasters. Each pixel becomes a row with `x`, `y`, and `band1`, `band2`, ... columns. Coordinates are pixel centers derived from the affine geotransform stored in the TIFF metadata. NoData values become NA. This enables spatial queries like `filter(band1 > 25, x >= xmin, x <= xmax)` on rasters that are too large to load as matrices. Results can be converted back to a raster with `terra::rast(df, type = "xyz")`. The corresponding `write_tiff()` function reconstructs the grid from x/y columns and writes a GeoTIFF with optional DEFLATE compression.

Streaming format conversion is a common pattern. The entire pipeline runs batch-by-batch, so converting a 20 GB CSV file to `.vtr` stays within a few hundred MB of memory. Write functions (`write_vtr()`, `write_csv()`, `write_sqlite()`, `write_tiff()`) all accept `vectra_node` inputs and consume batches from the upstream pipeline, writing each batch to the output format as it arrives.

```{r format-conversion}
csv_file <- tempfile(fileext = ".csv")
vtr_file <- tempfile(fileext = ".vtr")
csv_out <- tempfile(fileext = ".csv")

write.csv(mtcars, csv_file, row.names = FALSE)
tbl_csv(csv_file) |> write_vtr(vtr_file)

tbl(vtr_file) |>
  filter(cyl == 6) |>
  write_csv(csv_out)

read.csv(csv_out) |> head()
```

For format-specific options, batch sizes, and performance considerations, see `vignette("formats")`.

## Indexes

Hash indexes accelerate equality predicates on `.vtr` files. An index is stored as a `.vtri` sidecar file alongside the `.vtr` file (named `<path>.<column>.vtri` for single-column indexes). It maps key hashes to row group indices using FNV-1a hashing, enabling O(1) row group identification for `filter(col == value)` and `filter(col %in% values)`. When `tbl()` opens a `.vtr` file, it automatically detects and loads any `.vtri` sidecar files present in the same directory. The scan node then consults the index before reading data, skipping row groups that cannot contain matching keys. This composes with zone-map pruning: the index narrows down candidate row groups, and zone-map statistics further eliminate row groups where the column's min/max range excludes the filter value.

Create an index with `create_index()`. Check for an existing index with `has_index()`. Indexes persist on disk and survive R sessions.

```{r index-create}
f_idx <- tempfile(fileext = ".vtr")
write_vtr(
  data.frame(id = letters, val = 1:26, stringsAsFactors = FALSE),
  f_idx,
  batch_size = 5
)

has_index(f_idx, "id")  # FALSE
create_index(f_idx, "id")
has_index(f_idx, "id")  # TRUE
```

With 6 row groups of 5 rows each, a filter on `id` now skips row groups that cannot contain the target value.

```{r index-query}
tbl(f_idx) |>
  filter(id == "m") |>
  collect()
```

Composite indexes cover AND-combined equality predicates on multiple columns. Pass a character vector to `create_index()`.

```{r index-composite}
f_comp <- tempfile(fileext = ".vtr")
write_vtr(
  data.frame(
    region = rep(c("north", "south"), each = 13),
    id = letters,
    val = 1:26,
    stringsAsFactors = FALSE
  ),
  f_comp,
  batch_size = 5
)
create_index(f_comp, c("region", "id"))

tbl(f_comp) |>
  filter(region == "north", id == "c") |>
  collect()
```

The `ci` argument to `create_index()` builds a case-insensitive index by lowercasing all key values before hashing. This is useful for string columns with inconsistent casing (e.g., species names entered by different data providers). When a case-insensitive index is present, `filter(col == "Quercus")` will match "quercus", "QUERCUS", and any other casing variant.

Index pushdown composes with binary search on sorted columns and zone-map pruning. A well-indexed `.vtr` file with many small row groups can answer point queries over hundreds of millions of rows by reading a single row group. `explain()` shows whether an index was used and how many row groups were pruned. For details on index internals, composite index hash combining, and tuning batch size for optimal index selectivity, see `vignette("indexing")`.

## Incremental operations

Three functions support incremental workflows on `.vtr` files without full rewrites. These are designed for append-heavy pipelines where new data arrives in batches (daily sensor readings, transaction logs, biodiversity occurrence records) and rewriting the entire file on each update would be prohibitively slow.

`append_vtr()` adds new rows as additional row groups at the end of an existing file. The schema of the new data must match the existing file exactly: same column names, same types, same order. Existing row groups are untouched; the function writes the new row groups after the last existing one and patches the header to update the row group count. This is not fully atomic: if the process is interrupted between writing the row groups and patching the header, the file may be corrupted. For write-once workloads where atomicity matters, use `write_vtr()` (which writes to a temp file and renames).

```{r append}
fa <- tempfile(fileext = ".vtr")
write_vtr(mtcars[1:16, ], fa)
append_vtr(mtcars[17:32, ], fa)
tbl(fa) |> collect() |> nrow()
```

`delete_vtr()` marks rows as deleted by writing a tombstone sidecar file (`<path>.del`). The `.vtr` file itself is never modified. The next `tbl()` call automatically excludes tombstoned rows. Tombstones are cumulative across multiple `delete_vtr()` calls.

```{r delete}
delete_vtr(fa, c(0, 1, 2))  # 0-based row indices
tbl(fa) |> collect() |> nrow()
unlink(c(fa, paste0(fa, ".del")))
```

`diff_vtr()` computes a key-based logical diff between two snapshots of the same dataset. It returns a list with `$deleted` (a vector of key values present in the old file but not the new) and `$added` (a lazy `vectra_node` of rows present in the new file but not the old). This is a set-level difference based on key identity, not a binary file diff or row-level comparison. Rows with the same key that have changed values appear in both `$deleted` and `$added`, which makes it possible to detect updates by intersecting the two sets. The implementation streams both files: pass 1 builds a hash set from the old file's key column, pass 2 streams all columns from the new file and writes added rows to a temporary `.vtr` file. The temporary file is cleaned up automatically when the returned node is garbage-collected.

```{r diff}
fd1 <- tempfile(fileext = ".vtr")
fd2 <- tempfile(fileext = ".vtr")
old <- data.frame(id = 1:5, val = letters[1:5],
                  stringsAsFactors = FALSE)
new <- data.frame(id = c(3L, 4L, 5L, 6L, 7L),
                  val = c("C", "d", "e", "f", "g"),
                  stringsAsFactors = FALSE)
write_vtr(old, fd1)
write_vtr(new, fd2)

d <- diff_vtr(fd1, fd2, "id")
d$deleted
collect(d$added)
unlink(c(fd1, fd2))
```

Keys 1 and 2 were deleted. Keys 6 and 7 were added. Key 3 appears with a changed value ("c" to "C"), which shows up in both deleted and added. For incremental ETL patterns and large-data workflows, see `vignette("large-data")`.

## Materialized blocks

A `vectra_node` is single-use: once `collect()` or `write_vtr()` consumes it, the C-level external pointer is invalidated and the plan cannot be reused. This is a consequence of the pull-based execution model, where each node's `next_batch()` advances internal state that cannot be rewound. For repeated lookups against the same dataset (e.g., matching a stream of queries against a reference table), `materialize()` pulls all batches and stores the result as a persistent columnar block in memory. The block lives until it is garbage-collected or the R session ends.

```{r block-materialize}
blk_data <- data.frame(
  taxonID = c("T1", "T2", "T3", "T4", "T5"),
  name = c("Quercus robur", "Pinus sylvestris",
           "Fagus sylvatica", "Acer campestre",
           "Betula pendula"),
  stringsAsFactors = FALSE
)
f_blk <- tempfile(fileext = ".vtr")
write_vtr(blk_data, f_blk)

blk <- materialize(tbl(f_blk))
blk
```

`block_lookup()` performs a hash-based exact lookup on a string column of the materialized block. The hash index is built lazily on the first call to `block_lookup()` for a given column and cached inside the block for all subsequent calls. This means the first lookup pays a one-time O(n) indexing cost and all subsequent lookups are O(k) where k is the number of query keys. The `ci` argument enables case-insensitive matching by lowercasing keys before hashing.

```{r block-lookup}
block_lookup(blk, "name", c("Quercus robur", "Betula pendula"))
```

`block_fuzzy_lookup()` computes string distances between query keys and a column in the block, returning all pairs within the `max_dist` threshold. It supports Damerau-Levenshtein (`"dl"`), Levenshtein (`"levenshtein"`), and Jaro-Winkler (`"jw"`). The `block_col` and `block_keys` arguments enable exact-match blocking: only rows where the blocking column matches the corresponding blocking key are compared, which can reduce the number of distance computations by orders of magnitude. For example, blocking on genus before fuzzy-matching species names restricts comparisons to rows within the same genus. The distance computation is OpenMP-parallelized via the `n_threads` argument.

```{r block-fuzzy}
block_fuzzy_lookup(
  blk, "name",
  c("Qurecus robur", "Pinus silvestris"),
  method = "dl",
  max_dist = 0.2
)
```

The `query_idx` column (1-based) maps each result row back to the input query vector. The `fuzzy_dist` column gives the normalized distance for each match.

## Inspecting the plan

Every vectra pipeline builds a lazy execution plan as a tree of nodes. `explain()` prints this tree before any data moves, which makes it the primary debugging tool for understanding what the engine will do when you call `collect()`. The output shows node types (ScanNode, FilterNode, ProjectNode, SortNode, GroupAggNode, WindowNode, JoinNode, etc.), their parent-child relationships, and the output column schema at each level.

```{r explain-full}
tbl(f) |>
  filter(cyl > 4) |>
  select(mpg, cyl, hp) |>
  arrange(desc(mpg)) |>
  explain()
```

Several things to look for in the plan:

**Streaming vs. materializing.** FilterNode and ProjectNode are streaming: they process one batch at a time in constant memory. SortNode, GroupAggNode, and WindowNode are materializing: they consume all input before producing output. A plan with only streaming nodes above the scan will use memory proportional to one batch, regardless of file size.

**Column pruning.** ProjectNode appears when `select()` or `mutate()` is used. It lists only the columns that pass through. Columns not listed are never decoded from disk in the scan node, saving I/O and decompression work.

**Predicate pushdown.** When a hash index exists on the filter column, the ScanNode consults the `.vtri` index before reading row groups. `explain()` shows the index file path when pushdown is active. Zone-map pruning (based on per-row-group min/max statistics) applies to all columns automatically, with or without an explicit index. Combining both mechanisms gives the fastest possible scan: the index identifies candidate row groups by key hash, and zone-map statistics further narrow the set.

**Node order.** The plan is printed bottom-up (leaf to root). The ScanNode is always at the bottom. Reading the plan from bottom to top traces the path data takes through the pipeline: scan from disk, filter rows, project columns, sort, limit.

`glimpse()` provides a quick preview of column types and a sample of values without collecting the full result. It reads enough data to show one line per column.

```{r glimpse}
tbl(f) |> glimpse()
```

## Cleanup

```{r cleanup}
unlink(c(f, f_batched, f_na, fs, fs2, f_cyl, f_rat, f_ref, f_query, fd,
         ft, csv_in, csv_out, csv_file, vtr_file, db, f_src, f_idx,
         paste0(f_idx, ".id.vtri"), f_comp,
         paste0(f_comp, ".region_id.vtri"), f_blk))
```

---
title: "Format Backends"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Format Backends}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Introduction

vectra reads and writes five formats: its native `.vtr`, CSV, SQLite, Excel, and
GeoTIFF. Each reader returns the same `vectra_node` object. Once a node exists,
the engine treats it identically regardless of where the data came from. A filter
on a CSV scan produces the same plan tree as a filter on a `.vtr` scan. The same
verbs work, the same expressions evaluate, and `collect()` materializes an R
data.frame either way.

This matters because format choice becomes a deployment decision, not a code
decision. We can prototype a pipeline against a CSV export, convert the data to
`.vtr` for production, and the pipeline code stays the same apart from the
opening `tbl_*()` call.

Writers mirror the readers. `write_vtr()`, `write_csv()`, `write_sqlite()`, and
`write_tiff()` are all S3 generics dispatching on both `vectra_node` and
`data.frame`. When the input is a node, writes stream batch-by-batch through the
plan tree. The full dataset never needs to fit in memory. This is the core of
vectra's format conversion story: pipe a reader into a writer, and data flows
from source format to target format in bounded memory.

```{r intro-demo}
library(vectra)

# Write mtcars to .vtr, then read it back lazily
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
node <- tbl(f)
node
```

The node prints its schema but holds no data. Columns, types, and row count are
in the file header; the actual values stay on disk until `collect()`.

## The .vtr format

`.vtr` is vectra's native binary columnar format. It stores data in row groups,
where each row group contains all columns for a slice of rows. The current
version (v4) applies a two-stage encoding stack per column per row group:
first a logical encoding (dictionary for low-cardinality strings, delta for
monotonic integers, or plain pass-through), then byte-level compression via
Zstandard when it actually shrinks the data.

This layout gives the scan node several optimization paths that other formats
cannot support. Zone-map statistics (min/max per column per row group) let the
engine skip entire row groups that cannot match a filter predicate. Column
pruning avoids reading columns that the query never references. And hash indexes
(`.vtri` sidecar files) enable O(1) row group lookup on equality predicates.

```{r vtr-roundtrip}
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

tbl(f) |>
  filter(cyl == 6) |>
  select(mpg, cyl, hp) |>
  collect()
```

When we call `write_vtr()` with a `vectra_node` input, the writer streams
batches from the upstream plan and writes each batch as one row group. The full
result never materializes in memory. For `data.frame` inputs, the data is
written directly from R's memory in a single row group.

The `batch_size` parameter on `write_vtr()` controls how many rows accumulate
before flushing a row group. Smaller row groups mean more granular zone-map
statistics and finer predicate pruning. Larger row groups reduce per-group
overhead and compress better.

```{r vtr-batch-size}
f <- tempfile(fileext = ".vtr")
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)

# Convert CSV to .vtr with 10-row row groups
tbl_csv(csv) |> write_vtr(f, batch_size = 10)

# The file now has multiple row groups
tbl(f) |> collect() |> nrow()
```

`append_vtr()` adds new row groups to an existing file without rewriting it.
This is useful for incremental data ingestion where new batches arrive over
time.

## CSV

`tbl_csv()` opens a CSV file for streaming reads. Column types are inferred
from the first 1000 rows: numeric-looking columns become doubles, integer-
looking columns become integers, and everything else is a string. Gzip-
compressed files (`.csv.gz`) are detected and decompressed transparently.

```{r csv-read}
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)

tbl_csv(csv) |>
  filter(hp > 200) |>
  select(mpg, hp, wt) |>
  collect()
```

The `batch_size` parameter controls how many rows the CSV scanner reads per
internal batch. The default (65536) works well for most files. Smaller values
reduce peak memory at the cost of more read calls.

`write_csv()` streams from any node. It writes a standard comma-separated file
with a header row. There is no gzip output option; if we need compressed output,
we convert to `.vtr` instead, which applies Zstandard automatically.

```{r csv-write}
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

out_csv <- tempfile(fileext = ".csv")
tbl(f) |>
  filter(cyl == 4) |>
  write_csv(out_csv)

# Verify the output
read.csv(out_csv) |> head()
```

CSV has no predicate pushdown or column pruning. Every row and every column is
read and parsed, then the engine's filter and project nodes discard what the
query does not need. For one-off analyses this is fine. For repeated queries on
the same data, converting to `.vtr` once and querying many times is faster.

## SQLite

`tbl_sqlite()` connects to a SQLite database and streams a table through
vectra's engine. Column types are inferred from the declared types in the
`CREATE TABLE` statement. All filtering, grouping, and aggregation happen in
vectra's C engine, not via SQL queries. The SQLite library is used only as a
row source.

```{r sqlite-read}
db <- tempfile(fileext = ".sqlite")
write_sqlite(mtcars, db, "cars")

tbl_sqlite(db, "cars") |>
  filter(mpg > 25) |>
  select(mpg, cyl, wt) |>
  collect()
```

The `table` argument names which table to scan. A database can hold many tables;
each `tbl_sqlite()` call opens one.

SQLite is a natural choice when data already lives in a database, or when other
tools (Python, command-line utilities, web applications) need to read the same
file. The format is self-contained, widely supported, and handles concurrent
reads well.

`write_sqlite()` streams from any node into a SQLite table. If the table already
exists, rows are appended. This makes it straightforward to export vectra query
results for consumption by non-R tools.

```{r sqlite-write}
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

db <- tempfile(fileext = ".sqlite")
tbl(f) |>
  filter(cyl == 8) |>
  write_sqlite(db, "v8_cars")

# Read it back through vectra
tbl_sqlite(db, "v8_cars") |> collect()
```

## Excel

`tbl_xlsx()` reads a sheet from an Excel workbook. It requires the `openxlsx2`
package, which is listed in Suggests. The sheet is specified by name or by
1-based index (default: first sheet).

Unlike the other backends, Excel support is read-only. There is no `write_xlsx()`
function. Excel's format is complex and writing it adds little value when CSV
or SQLite serve the same interoperability purpose with less overhead.

Under the hood, `tbl_xlsx()` reads the sheet into a data.frame via `openxlsx2`
and then converts it to a vectra node. This means the sheet does load into
memory during the initial read, but all subsequent operations (filter, mutate,
join) are lazy and stream through the C engine.

```r
# Not run (requires openxlsx2)
node <- tbl_xlsx("survey_results.xlsx", sheet = "Q1_2025")
node |>
  filter(score >= 80) |>
  select(respondent, score, region) |>
  collect()
```

For large Excel files that do not fit in memory, export to CSV first and use
`tbl_csv()`.

## GeoTIFF

`tbl_tiff()` reads raster data. Each pixel becomes a row with `x` and `y`
columns (pixel center coordinates derived from the geotransform) and one column
per band (`band1`, `band2`, etc.). NoData values become `NA`. This tabular
representation lets us apply the full verb set to raster data: filter by
spatial extent, threshold band values, compute derived indices, join with
point observations.

The default `batch_size` is 256 raster rows, much smaller than the CSV/SQLite
default. Raster data is dense. A 10,000 x 10,000 single-band TIFF has 100
million pixels, so each raster row of 10,000 pixels already constitutes a
meaningful batch.

```r
# Not run (requires a GeoTIFF file)
tbl_tiff("temperature.tif") |>
  filter(band1 > 25, x >= 10, x <= 20) |>
  collect()
```

`write_tiff()` writes query results back to a GeoTIFF. The data must contain
`x` and `y` columns and one or more numeric band columns. Grid dimensions and
the geotransform are inferred from the coordinate arrays. The `compress`
parameter enables DEFLATE compression.

### Integer pixel types

By default, `write_tiff()` writes 64-bit float pixels. The `pixel_type`
parameter selects a smaller representation: `"int16"`, `"int32"`, `"uint8"`,
`"uint16"`, or `"float32"`. Smaller types compress better and produce much
smaller files — a global climate raster stored as `int16` + DEFLATE can be
5-10x smaller than the equivalent `float64`.

```r
# Scale temperature to integer: value * 100, stored as int16
df |>
  mutate(band1 = round(band1 * 100)) |>
  write_tiff("temperature_int16.tif", compress = TRUE, pixel_type = "int16")
```

`NA` values in the input are automatically mapped to the integer nodata value
(-32768 for `int16`, 65535 for `uint16`, etc.) and tagged in the
GDAL\_NODATA header. The reader converts them back to `NA` on read.

### Embedded metadata

The `metadata` parameter writes a GDAL\_METADATA XML string into TIFF tag
42112. This is useful for recording scale factors, offsets, variable names,
or provenance without a sidecar file.

```r
xml <- '<GDALMetadata>
  <Item name="scale">0.01</Item>
  <Item name="offset">0</Item>
  <Item name="variable">bio1</Item>
</GDALMetadata>'

write_tiff(df, "bio1.tif", pixel_type = "int16", metadata = xml)

# Read the metadata back
tiff_metadata("bio1.tif")
```

`tiff_metadata()` returns the XML string, or `NA` if the tag is absent.

### Point extraction

`tiff_extract_points()` samples band values at specific `(x, y)` coordinates
directly from the TIFF file. Only the strips containing query points are read,
making it efficient for sparse lookups on large rasters. No spatial package is
needed at runtime.

```r
pts <- data.frame(x = c(10.5, 11.2), y = c(47.1, 47.3))
tiff_extract_points("temperature.tif", pts)
```

This round-trip capability makes vectra useful for raster ETL: read a TIFF,
filter or transform it with standard verbs, write a new TIFF. No dependency
on spatial packages is needed for the read-process-write loop itself, though
`terra::rast(df, type = "xyz")` can convert results to `SpatRaster` objects
when spatial operations are needed.

## Streaming conversion pipelines

The practical payoff of uniform node types is format conversion with zero
full-dataset materialization. We pipe a reader into a writer, and data flows
from source to target one batch at a time. Memory use stays proportional to
one batch, not the full dataset.

The simplest case: convert CSV to `.vtr`.

```{r convert-csv-vtr}
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)

vtr <- tempfile(fileext = ".vtr")
tbl_csv(csv) |> write_vtr(vtr)

tbl(vtr) |> collect() |> head()
```

We can filter during conversion. Only rows passing the predicate are written.

```{r convert-filtered}
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)

vtr <- tempfile(fileext = ".vtr")
tbl_csv(csv) |>
  filter(mpg > 20) |>
  mutate(kpl = mpg * 0.425144) |>
  write_vtr(vtr)

tbl(vtr) |> collect()
```

Multi-step ETL works the same way. Read from one format, transform, write to
another.

```{r etl-pipeline}
# CSV -> filter + transform -> SQLite
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)

db <- tempfile(fileext = ".sqlite")
tbl_csv(csv) |>
  filter(cyl >= 6) |>
  select(mpg, cyl, hp, wt) |>
  mutate(power_weight = hp / wt) |>
  write_sqlite(db, "powerful_cars")

# SQLite -> VTR
vtr <- tempfile(fileext = ".vtr")
tbl_sqlite(db, "powerful_cars") |> write_vtr(vtr)

tbl(vtr) |> collect()
```

Each arrow in the pipeline is a streaming connection. The CSV scanner reads a
batch, the filter discards rows, the project computes new columns, and the
SQLite writer inserts the result. Then the next batch flows through. At no point
does the full dataset exist in memory. For datasets that dwarf available RAM,
this is the only way these conversions can work.

Conversion pipelines also compose with joins. We can read two sources in
different formats and join them.

```{r join-across-formats}
f1 <- tempfile(fileext = ".vtr")
f2 <- tempfile(fileext = ".csv")

cars_main <- mtcars[, c("mpg", "cyl", "hp")]
cars_extra <- data.frame(cyl = c(4, 6, 8), label = c("small", "mid", "big"))

write_vtr(cars_main, f1)
write.csv(cars_extra, f2, row.names = FALSE)

tbl(f1) |>
  left_join(tbl_csv(f2), by = "cyl") |>
  collect() |>
  head()
```

## Batch size

The `batch_size` parameter appears on readers and on `write_vtr()`. It controls
different things depending on context.

On **readers** (`tbl_csv`, `tbl_sqlite`), batch size sets how many rows the
scanner reads per pull from the source. The default 65536 balances memory use
against per-call overhead. Larger values read more rows per call, which can be
faster for simple scan-heavy queries. Smaller values keep peak memory lower,
which matters when individual rows are wide (many columns or long strings).

On **`write_vtr()`**, batch size controls how many rows accumulate before
flushing a row group to disk. This directly affects the structure of the
output file. Each row group carries its own zone-map statistics (min and max
per column), so more row groups mean finer-grained predicate pushdown. Fewer
row groups mean less metadata overhead and better compression ratios, because
the compressor sees more data per block.

```{r batch-size-effect}
csv <- tempfile(fileext = ".csv")
big <- data.frame(
  id = seq_len(1000),
  value = rnorm(1000)
)
write.csv(big, csv, row.names = FALSE)

# Small row groups: more granular zone maps
f_small <- tempfile(fileext = ".vtr")
tbl_csv(csv) |> write_vtr(f_small, batch_size = 100)

# Default: single row group for 1000 rows
f_default <- tempfile(fileext = ".vtr")
tbl_csv(csv) |> write_vtr(f_default)

cat("Small batches:", file.size(f_small), "bytes\n")
cat("Default:      ", file.size(f_default), "bytes\n")
```

For `tbl_tiff`, the default batch size is 256 raster rows, reflecting the
high per-row density of raster data. Increasing it speeds up reads on files
with few bands. Decreasing it helps when many bands produce very wide rows.

There is no batch size parameter on `tbl()` (the `.vtr` reader). Row groups
are defined by the file; the scanner reads one row group per `next_batch()`
call, respecting whatever batch size was chosen at write time.

## Format comparison

The table below summarizes what each format supports. The right format depends
on the workload: `.vtr` for repeated analytical queries, CSV for interchange,
SQLite for multi-tool access, Excel for one-off imports, GeoTIFF for raster
data.

| Feature | .vtr | CSV | SQLite | Excel | GeoTIFF |
|:--------|:-----|:----|:-------|:------|:--------|
| Streaming read | yes | yes | yes | no | yes |
| Streaming write | yes | yes | yes | -- | yes |
| Predicate pushdown | yes | no | no | no | no |
| Column pruning | yes | no | no | no | no |
| Zone-map skip | yes | no | no | no | no |
| Hash index support | yes | no | no | no | no |
| Compression | zstd | gzip (read) | -- | -- | deflate |
| Size on disk | small | large | medium | medium | variable |
| Random access | by row group | no | by rowid | no | by raster row |
| External tool support | vectra only | universal | wide | wide | GIS tools |

`.vtr` wins on query performance because it is the only format where the
scanner can skip data before reading it. Zone maps, column pruning, and hash
indexes all reduce I/O. For a dataset queried repeatedly, converting to `.vtr`
once pays for itself quickly.

CSV and SQLite have broad tool support. If the data needs to flow to Python,
a dashboard, or a command-line tool, these formats avoid lock-in. SQLite adds
transactional writes and concurrent read access that flat files lack.

GeoTIFF is specialized. It carries spatial metadata (coordinate reference
system, geotransform) that the other formats do not. When the output is a
raster, GeoTIFF is the right choice; when the output is a table derived from
raster data, `.vtr` or CSV may be more practical.

---
title: "Working with Large Data"
author: "Gilles Colling"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Working with Large Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Introduction

R holds data.frames entirely in memory. A 10 GB CSV loaded with `read.csv()`
becomes a 10 GB R object, sometimes larger: R's internal representation of
character vectors, factor levels, and ALTREP metadata can inflate the footprint
to 2 to 3x the raw data size. On a machine with 16 GB of RAM, that 10 GB file
will either fail to load or push the system into swap, grinding everything to a
halt. "Larger-than-RAM" is not a theoretical concern. It is the normal state of
affairs for anyone working with multi-year survey data, remote sensing extracts,
or event logs from sensor networks.

vectra sidesteps this by never loading the full dataset. When we call `tbl()` on
a `.vtr` file or `tbl_csv()` on a CSV, nothing is read yet. The call to
`collect()`, `write_vtr()`, or any other terminal verb pulls data through the
plan tree one row group at a time. Each row group is a self-contained slice of
the dataset: a few thousand to a few hundred thousand rows with all their
columns. The engine reads one group, applies every operation in the pipeline,
emits the result, then discards the group and moves to the next. The key
distinction from R's default approach: data moves through the pipeline in
fixed-size pieces, and each piece is freed before the next one arrives.

This batch-at-a-time design means that most vectra operations use constant
memory per batch, regardless of how many total rows exist. A `filter` node
reads a batch, applies a predicate, passes matching rows downstream, and
forgets about them. A `mutate` node reads a batch, computes new columns,
passes the result on. A `select` node simply drops columns from each batch.
None of these ever accumulate the full dataset in memory.

A few operations do need to see all the data before they can produce output.
Sorting requires a global ordering. Hash-based grouping needs to track every
distinct group key. Joins need a hash table of the build side. These
"materializing" operations have bounded memory budgets or known scaling
behaviour that we will discuss in detail.

The central implication is this: if we can express our workload as a chain of
streaming operations terminated by a streaming sink, the full dataset never
needs to fit in RAM. A 200 GB CSV can flow through `tbl_csv() |> filter() |>
mutate() |> write_vtr()` on a laptop with 8 GB of memory. The pipeline
processes each batch in isolation and writes the result directly to disk.

`collect()` is the one terminal verb that materializes everything into an R
data.frame. It breaks the streaming model: the entire result set crosses from C
into R and lives in R's heap until we remove it. For large data, we avoid it for
the final result and instead use `write_vtr()`, `write_csv()`, or
`write_sqlite()` as our streaming sinks. These functions pull batches through the
pipeline and write each one directly to disk, so the C engine's memory is
recycled on every iteration. The R process never holds more than one batch. We
still use `collect()` freely for small intermediate checks, filtered subsets,
and aggregation results that we know will fit in memory. The question to ask
before calling `collect()` is always: how many rows will this return? If the
answer is "roughly the same as the input," we want a streaming sink instead.

This vignette walks through the patterns and tools vectra provides for
working with datasets that exceed available RAM. We will build our examples
around a synthetic multi-year ecological survey dataset: species observations
recorded across monitoring sites over several years.

```{r synthetic-data}
library(vectra)

set.seed(42)

n <- 50000
species_pool <- c(
  "Quercus robur", "Fagus sylvatica", "Pinus sylvestris",
  "Betula pendula", "Acer pseudoplatanus", "Fraxinus excelsior",
  "Picea abies", "Tilia cordata", "Carpinus betulus",
  "Alnus glutinosa", "Sorbus aucuparia", "Ulmus glabra"
)
sites <- paste0("SITE_", sprintf("%03d", 1:50))
years <- 2015:2024

obs <- data.frame(
  obs_id    = seq_len(n),
  site      = sample(sites, n, replace = TRUE),
  species   = sample(species_pool, n, replace = TRUE),
  year      = sample(years, n, replace = TRUE),
  abundance = rpois(n, lambda = 15),
  cover_pct = round(runif(n, 0.1, 95.0), 1),
  quality   = sample(c("good", "moderate", "poor"), n,
                     replace = TRUE, prob = c(0.6, 0.3, 0.1)),
  stringsAsFactors = FALSE
)
```

We will write this as both CSV and `.vtr` to demonstrate different starting
points. In a real workflow the CSV might be a raw export from a database or
field data logger.

```{r write-sources}
csv_path <- tempfile(fileext = ".csv")
write.csv(obs, csv_path, row.names = FALSE)

vtr_path <- tempfile(fileext = ".vtr")
write_vtr(obs, vtr_path)
```

## Streaming pipelines

The core larger-than-RAM pattern in vectra is source, verbs, sink. We open a
lazy reference to the data, chain operations, and terminate with a write
function. No intermediate data.frame is ever created.

Suppose we want to extract only high-quality observations, compute a
derived column, and store the result as a new `.vtr` file. The entire pipeline
streams:

```{r streaming-basic}
clean_path <- tempfile(fileext = ".vtr")

tbl_csv(csv_path) |>
  filter(quality == "good") |>
  mutate(log_abundance = log(abundance + 1)) |>
  write_vtr(clean_path)
```

At no point did R hold 50,000 rows (or 50 million, if this were a larger
dataset) in memory. `tbl_csv()` created a lazy scan node. `filter()` and
`mutate()` each wrapped it in another lazy node. `write_vtr()` triggered
execution: the CSV scanner read a batch of rows, the filter discarded
non-matching rows, the mutate computed the log column, and the writer
serialized the result to disk. Then the next batch.

At the C level, each node in the pipeline exposes a `next_batch()` function. The
terminal node (the writer) calls `next_batch()` on its upstream node, which in
turn calls `next_batch()` on its own upstream, and so on down to the scan node
that actually reads from disk. Each `VecBatch` struct flows upward through the
chain, gets transformed at each stage, and arrives at the writer ready to be
serialized. Once the writer finishes with a batch, the memory is freed. The
pipeline then pulls the next batch. This pull-based design means memory
consumption stays flat regardless of how many millions of rows pass through. A
pipeline that processes 200 GB of CSV data uses the same peak memory as one that
processes 200 KB.

We can verify the result by reading back a small slice:

```{r verify-streaming}
tbl(clean_path) |>
  select(obs_id, site, species, abundance, log_abundance) |>
  slice_head(n = 5) |>
  collect()
```

The same pattern works for any source-sink combination. CSV to SQLite:

```{r csv-to-sqlite}
db_path <- tempfile(fileext = ".sqlite")

tbl_csv(csv_path) |>
  filter(year >= 2020) |>
  write_sqlite(db_path, "recent_obs")
```

Or `.vtr` to CSV, useful for handing data to colleagues who expect plain text:

```{r vtr-to-csv}
export_csv <- tempfile(fileext = ".csv")

tbl(vtr_path) |>
  filter(species == "Fagus sylvatica") |>
  select(obs_id, site, year, abundance) |>
  write_csv(export_csv)
```

Each pipeline reads only what it needs and writes directly to the target
format. The intermediate verbs never buffer more than one batch. This is
what makes vectra practical for datasets larger than RAM: the memory
footprint of a streaming pipeline is determined by the batch size, not the
dataset size.

When the result of a pipeline *is* small enough for R (aggregation results,
filtered subsets, summary statistics), `collect()` is still the right choice.
Aggregations in particular tend to produce far fewer rows than the input:

```{r collect-agg}
tbl(vtr_path) |>
  group_by(species, year) |>
  summarise(
    n_obs     = n(),
    mean_abun = mean(abundance),
    mean_cov  = mean(cover_pct)
  ) |>
  arrange(species, year) |>
  collect() |>
  head(10)
```

The rule of thumb: if the output might be large, write to disk. If it fits
in a data.frame, collect.

## Batch sizing

A `.vtr` file is organised into row groups. Each row group is a self-contained
unit: it has its own column data, validity bitmaps, dictionary tables, and
zone-map min/max statistics. When vectra scans a file, it reads one row group
at a time. The number of rows per group is the batch size.

Batch size affects two things: memory per batch and zone-map pruning
granularity. Larger batches mean fewer row groups and less per-group
overhead, but each batch consumes more memory during processing. Smaller
batches mean the zone-map statistics cover a narrower range of values, so
filter predicates can skip more groups entirely.

The default batch size is 65,536 rows. This is a reasonable middle ground for
most workloads: large enough that per-batch overhead (reading the row group
header, allocating column arrays, setting up validity bitmaps) is amortized
across many rows, but small enough that each batch stays in the low megabytes
for typical column counts. When choosing a different batch size, the decision
depends on the query pattern. Analytical workloads that scan most of the file
(aggregations, full-table joins) benefit from large batches because they reduce
the number of `next_batch()` calls and the associated per-group bookkeeping.
Point queries that target a few matching rows benefit from small batches,
especially when combined with indexes or sorted columns, because the engine can
skip entire row groups whose zone-map ranges do not overlap the predicate.

For analytical workloads with selective filters, smaller row groups often pay
off. If each row group spans a narrow range of `year` values, a filter like
`year == 2023` can skip most groups without reading them. With one giant row
group, the filter must scan every row.

We control batch size when writing:

```{r batch-size-write}
small_groups <- tempfile(fileext = ".vtr")
large_groups <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  arrange(year) |>
  write_vtr(small_groups, batch_size = 5000)

tbl(vtr_path) |>
  arrange(year) |>
  write_vtr(large_groups, batch_size = 50000)
```

We sorted by `year` before writing so that row groups have tight year ranges.
The small-batch file has about 10 row groups; the large-batch file has one.
When we filter on `year`, the small-batch file can prune most groups using
zone maps:

```{r batch-size-compare}
tbl(small_groups) |>
  filter(year == 2023) |>
  explain()
```

CSV and SQLite sources also have a `batch_size` parameter, but it controls
how many rows the scanner reads per pull rather than how the file is
structured on disk. The default of 65,536 works well for most cases. Reducing
it lowers peak memory; increasing it reduces per-batch overhead for simple
pass-through pipelines.

```{r csv-batch-size}
tbl_csv(csv_path, batch_size = 10000) |>
  filter(quality == "good") |>
  slice_head(n = 3) |>
  collect()
```

As a starting point: 50,000 to 100,000 rows per group works well for datasets
with a few dozen columns. If the data is sorted on a frequently-filtered
column, lean towards smaller groups (10,000 to 50,000) for better pruning. If
the workload is dominated by full scans or aggregations, larger groups
(100,000+) reduce overhead.

## Append workflows

Real datasets rarely arrive all at once. Monitoring data comes in daily,
monthly, or seasonally. `append_vtr()` adds new rows to an existing `.vtr`
file as a new row group, without touching existing data. The file header is
updated to reflect the additional group.

Here is a typical pattern: we start with one year of data and append
subsequent years as they arrive.

```{r append-init}
archive <- tempfile(fileext = ".vtr")
first_year <- obs[obs$year == 2015, ]
write_vtr(first_year, archive)

nrow(tbl(archive) |> collect())
```

Now we simulate receiving 2016 data:

```{r append-year}
year_2016 <- obs[obs$year == 2016, ]
append_vtr(year_2016, archive)

nrow(tbl(archive) |> collect())
```

`append_vtr()` also accepts a `vectra_node`, so we can stream data directly
from a CSV into an existing archive without loading it into R:

```{r append-streaming}
csv_2017 <- tempfile(fileext = ".csv")
write.csv(obs[obs$year == 2017, ], csv_2017, row.names = FALSE)

tbl_csv(csv_2017) |> append_vtr(archive)

nrow(tbl(archive) |> collect())
```

Each call to `append_vtr()` creates one new row group. Over time, the file
accumulates many groups of varying sizes. This is fine for append-heavy
workloads. If the number of groups becomes unwieldy (hundreds of small
appends, each adding a few rows), we can compact the file by rewriting it:

```{r compact}
compacted <- tempfile(fileext = ".vtr")
tbl(archive) |> write_vtr(compacted, batch_size = 50000)
```

This reads every row group from the archive, streams through the writer with
a fresh batch size, and produces a clean file. The old file can then be
replaced.

The schema of appended data must exactly match the target file: same column
names, same types, same order. If the schema drifts over time (a new column
added in 2018, say), we need to align it before appending. One approach is
to add the missing column with a default value in a `mutate()` step:

```{r append-schema}
new_data_node <- tbl_csv(csv_2017) |>
  mutate(observer = NA_character_)
```

This produces a node whose schema matches a hypothetical archive that includes
an `observer` column. The key point: because `append_vtr()` accepts nodes,
we can chain any transformation needed to match the target schema.

## Delete and tombstones

Sometimes we need to remove rows from an existing `.vtr` file. Perhaps a
batch of observations was flagged as erroneous, or privacy regulations
require deletion of certain records. `delete_vtr()` handles this without
rewriting the file.

Deletion works through a tombstone sidecar file. When we call
`delete_vtr(path, row_ids)`, vectra writes the specified row indices to a
`.del` file next to the `.vtr`. The original data file is never modified.
On the next `tbl()` call, the scan node reads the tombstone file and skips
those rows.

Row IDs are 0-based physical positions across the entire file. The first
row of the first row group is 0, the first row of the second group continues
from where the first left off.

```{r delete-basic}
del_demo <- tempfile(fileext = ".vtr")
write_vtr(obs[1:100, ], del_demo)

# Delete rows 0, 1, and 99 (first two and last)
delete_vtr(del_demo, c(0, 1, 99))

tbl(del_demo) |> collect() |> nrow()
```

Tombstone files are cumulative. Calling `delete_vtr()` again merges the new
indices with existing ones:

```{r delete-cumulative}
delete_vtr(del_demo, c(10, 11, 12))

tbl(del_demo) |> collect() |> nrow()
```

To undo all deletions, remove the `.del` file:

```{r delete-undo}
unlink(paste0(del_demo, ".del"))
tbl(del_demo) |> collect() |> nrow()
```

Tombstones are lightweight for sparse deletions. If we delete 1% of rows,
the overhead is negligible. But if deletions accumulate to a substantial
fraction of the file, scan performance degrades because the engine still
reads and discards the tombstoned rows. At that point, compaction makes
sense:

```{r delete-compact}
delete_vtr(del_demo, 0:49)
compacted_del <- tempfile(fileext = ".vtr")
tbl(del_demo) |> write_vtr(compacted_del)

nrow(tbl(compacted_del) |> collect())
```

The compacted file contains only the surviving rows, with no tombstone file.

## Diff between snapshots

Data pipelines often work with periodic snapshots. We receive yesterday's
extract and today's extract, and we need to know what changed. `diff_vtr()`
computes a key-based logical diff between two `.vtr` files.

The function takes the paths to the old and new files plus the name of a
key column. It streams both files and returns a list with two elements:
`$deleted` contains the key values that were present in the old file but
absent in the new one. `$added` is a lazy `vectra_node` of rows present
in the new file but not in the old.

```{r diff-setup}
snap_v1 <- tempfile(fileext = ".vtr")
snap_v2 <- tempfile(fileext = ".vtr")

# Version 1: observations 1-100
write_vtr(obs[1:100, ], snap_v1)

# Version 2: observations 51-150 (rows 1-50 removed, 101-150 added)
write_vtr(obs[51:150, ], snap_v2)
```

```{r diff-compute}
changes <- diff_vtr(snap_v1, snap_v2, "obs_id")

# Keys that disappeared
head(changes$deleted)
length(changes$deleted)
```

The deleted keys are the `obs_id` values 1 through 50. Meanwhile, the `$added`
element contains a lazy node pointing to observations with `obs_id` 101 through 150,
which we can collect or pipe into further processing:

```{r diff-added}
collect(changes$added) |> head()
```

Because `$added` is a lazy node, we can pipe it directly into further
processing. For instance, we might want to append only the truly new rows
to an archive:

```{r diff-append}
archive_diff <- tempfile(fileext = ".vtr")
write_vtr(obs[1:100, ], archive_diff)

changes$added |> append_vtr(archive_diff)

nrow(tbl(archive_diff) |> collect())
```

This is a common incremental-load pattern: diff today's snapshot against
yesterday's, then append only the additions. Combined with `delete_vtr()`
for removals, we can maintain an up-to-date archive without full rewrites.

Note that `diff_vtr()` performs a set-level diff on the key column. If a row
exists in both snapshots with the same key but different values in other
columns, it will not appear in either `$deleted` or `$added`. To detect
modifications, we would need to compare the full row content after
identifying shared keys.

## External sort

Sorting a dataset that does not fit in memory requires an external merge sort.
vectra's `arrange()` handles this automatically. The engine maintains a
1 GB memory budget. As data flows through, it accumulates rows in memory.
When the budget is reached, the accumulated rows are sorted and written to
a temporary `.vtr` file on disk (a "sorted run"). After all input has been
consumed, a k-way merge reads from all sorted runs simultaneously using a
min-heap, producing the final sorted output one batch at a time.

From the user's perspective, none of this is visible. We just call `arrange()`:

```{r external-sort}
sorted_path <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  arrange(species, desc(abundance)) |>
  write_vtr(sorted_path)

tbl(sorted_path) |>
  select(species, abundance, site) |>
  slice_head(n = 8) |>
  collect()
```

The sorted output streams to disk. Peak memory stays bounded at the spill
budget (1 GB) plus overhead for the merge heap, regardless of input size.
For our 50,000-row example the data fits entirely in memory and no spill
occurs. But the same code would work on a 500 million row file; the sort
would produce multiple temporary runs and merge them transparently.

The 1 GB budget governs how much unsorted data the engine accumulates before
flushing to a temporary file. Each flush produces one sorted run. A 10 GB
dataset with a 1 GB budget produces roughly 10 runs. The final merge opens all
runs simultaneously and maintains a min-heap with one entry per run. At each
step, the smallest element across all runs is popped from the heap and emitted
as the next output row; the run that contributed it advances by one row and
re-inserts into the heap. Because the heap has only as many entries as there are
runs (not as many as there are rows), the merge phase uses very little memory.

Data that is already partially sorted produces fewer runs. If the input is
sorted on a prefix of the sort key, long stretches of rows will already be in
order, so the engine can absorb more rows before hitting the budget and
flushing. In the best case (fully sorted input), no spill occurs at all and
the sort reduces to a streaming pass-through. There is nothing to tune here.
The engine detects the budget internally and manages spill files in the system
temp directory. They are cleaned up after the merge completes.

Writing the sorted data to a new `.vtr` file also improves query performance.
Once the file is sorted on `species`, zone-map statistics on that column
become tight, and filters on species can skip most row groups. Sorting on a
frequently-queried column is one of the most effective optimizations
available.

## Streaming joins

vectra's join engine uses a build-right, probe-left strategy. The right-side
table is fully materialized into a hash table in memory. Then the left-side
table streams through, probing the hash table for matches one batch at a time.
This means the left side can be arbitrarily large, because only the right side needs
to fit in memory.

The natural pattern for large-data joins is: huge fact table on the left,
small dimension table on the right.

```{r join-setup}
# Small reference table: site metadata
site_meta <- data.frame(
  site      = sites,
  region    = sample(c("North", "South", "East", "West"),
                     length(sites), replace = TRUE),
  elevation = round(runif(length(sites), 100, 2500)),
  stringsAsFactors = FALSE
)
site_path <- tempfile(fileext = ".vtr")
write_vtr(site_meta, site_path)
```

```{r join-streaming}
enriched <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  left_join(tbl(site_path), by = "site") |>
  write_vtr(enriched)

tbl(enriched) |>
  select(obs_id, site, region, elevation, species) |>
  slice_head(n = 5) |>
  collect()
```

The 50 site metadata rows become the hash table (right side). The observation
table streams through on the left, each batch probing the hash table for its
`site` key. Memory usage is proportional to the site metadata, not the
observation count.

What happens when both sides are large? The right side still must fit in
memory, so we need to reduce it. The most direct approach is to filter
the right side before the join:

```{r join-filter-right}
joined_path <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  filter(year >= 2022) |>
  inner_join(
    tbl(site_path) |> filter(region == "North"),
    by = "site"
  ) |>
  write_vtr(joined_path)

nrow(tbl(joined_path) |> collect())
```

The filtered right side is much smaller, so the hash table stays small. The
left-side filter also reduces the number of rows that flow through the probe,
but that affects throughput rather than memory.

For self-joins or joins between two truly large tables, the options are more
constrained. The core problem is that one side must fit in memory as a hash
table. When neither side is small, we need a strategy to make one of them
small.

The most general approach is split-apply-combine: partition the left side by a
key (say, `year`), and for each partition, join against only the matching rows
from the right side. If both tables span 10 years, each partition pair is
roughly 1/10th the size of the full join. We write each partial result to disk
and consolidate with `bind_rows()` afterward. This trades a single large join
for many small joins, each of which fits in memory.

Pre-aggregation is another powerful option. If we only need summary statistics
from the right table, `group_by() |> summarise()` can reduce millions of rows
to a handful of groups before the join. A right-side table with 50 million
observations across 10,000 sites collapses to 10,000 rows after a
site-level aggregation. That fits comfortably in a hash table.

```{r join-pre-agg}
# Right side: per-site-year summary from a second dataset
summary_path <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  group_by(site, year) |>
  summarise(site_year_avg = mean(abundance)) |>
  write_vtr(summary_path)

# Join the summary back to the detail table
tbl(vtr_path) |>
  left_join(tbl(summary_path), by = c("site", "year")) |>
  select(obs_id, site, year, abundance, site_year_avg) |>
  slice_head(n = 5) |>
  collect()
```

## Multi-file workflows

Partitioned data is common. Monthly exports, regional splits, sensor-specific
files. The partitioning dimension matters for query performance. Partitioning by
date works well when most queries filter on a time range, because we can skip
entire files that fall outside the range. Partitioning by category (region,
species, sensor type) works when queries target a specific category. The wrong
partitioning dimension forces us to read every file for every query, negating
the benefit.

`bind_rows()` combines multiple vectra nodes into a single streaming
pipeline. When all inputs share the same schema (column names and types),
vectra creates a C-level `ConcatNode` that reads from each child sequentially.
No data is copied; each source's batches flow through in order.

```{r multifile-setup}
monthly_paths <- character(6)
for (m in 1:6) {
  monthly_paths[m] <- tempfile(fileext = ".vtr")
  idx <- which(obs$year == 2024 & ((obs$obs_id %% 6) + 1) == m)
  month_data <- obs[idx[seq_len(min(200, length(idx)))], ]
  write_vtr(month_data, monthly_paths[m])
}
```

We can combine all monthly files and aggregate without loading them all:

```{r multifile-combine}
nodes <- lapply(monthly_paths, tbl)
combined <- do.call(bind_rows, nodes)

combined |>
  group_by(species) |>
  summarise(
    total_obs  = n(),
    mean_abun  = mean(abundance)
  ) |>
  arrange(desc(total_obs)) |>
  slice_head(n = 5) |>
  collect()
```

The `bind_rows()` call does not read any data. It returns a new node that,
when pulled, first exhausts all batches from the first child, then the
second, and so on. Downstream operations (the grouping, aggregation, sort)
see a single continuous stream.

We can also consolidate the partitioned files into one:

```{r multifile-consolidate}
consolidated <- tempfile(fileext = ".vtr")
nodes2 <- lapply(monthly_paths, tbl)
do.call(bind_rows, nodes2) |> write_vtr(consolidated)

nrow(tbl(consolidated) |> collect())
```

For partitioned files that arrive on a schedule, a common workflow combines
`bind_rows()` with `append_vtr()`. We process the latest batch of files
and append the result to a running archive:

```{r multifile-append}
running_archive <- tempfile(fileext = ".vtr")
initial_nodes <- lapply(monthly_paths[1:3], tbl)
do.call(bind_rows, initial_nodes) |> write_vtr(running_archive)

# Next month's files arrive
new_nodes <- lapply(monthly_paths[4:6], tbl)
do.call(bind_rows, new_nodes) |> append_vtr(running_archive)

nrow(tbl(running_archive) |> collect())
```

This gives us a single growing file that can be queried as a unit, while
each monthly append is a lightweight operation.

At some point we may want to consolidate partitioned files into a single sorted
`.vtr` for better query performance. `bind_rows()` piped into `arrange()` and
then `write_vtr()` does this in one streaming pass: the concat node reads from
each partition, the sort accumulates and spills as needed, and the writer
produces a single file with tight zone-map statistics. The original partitions
can then be archived or deleted.

## Format conversion ETL

A one-time conversion from CSV to `.vtr` pays off every time we query the
data afterwards. The `.vtr` format supports dictionary encoding, delta
encoding, zstd compression, and zone-map statistics. Repeated queries on a
`.vtr` file are faster because the engine can skip row groups and decompress
only the columns it needs. CSV requires a full parse every time.

A minimal conversion:

```{r etl-basic}
vtr_archive <- tempfile(fileext = ".vtr")
tbl_csv(csv_path) |> write_vtr(vtr_archive)
```

That single line streams the entire CSV through the vectra writer. The output
file is typed, compressed, and has zone-map statistics. For repeated queries,
the difference is substantial.

In practice, we rarely want a raw copy. The ingest step is a natural place
to clean, filter, and reshape:

```{r etl-clean}
clean_archive <- tempfile(fileext = ".vtr")

tbl_csv(csv_path) |>
  filter(quality != "poor") |>
  mutate(
    abundance_log = log(abundance + 1),
    cover_frac    = cover_pct / 100
  ) |>
  select(-quality) |>
  arrange(site, year) |>
  write_vtr(clean_archive, batch_size = 10000)
```

We filtered out poor-quality records, computed derived columns, dropped the
raw quality flag, sorted by site and year for better zone-map pruning, and
chose a 10,000-row batch size. The CSV was read once, streaming, and the
clean result was written directly to disk.

The inverse direction is also useful. We might need to export a subset
to a colleague who uses SQLite:

```{r etl-sqlite}
sqlite_export <- tempfile(fileext = ".sqlite")

tbl(clean_archive) |>
  filter(year >= 2020) |>
  write_sqlite(sqlite_export, "observations")
```

Or produce a CSV of summary statistics for a report:

```{r etl-summary-csv}
summary_csv <- tempfile(fileext = ".csv")

tbl(clean_archive) |>
  group_by(site, year) |>
  summarise(
    n_species = n_distinct(species),
    total_abundance = sum(abundance)
  ) |>
  write_csv(summary_csv)

read.csv(summary_csv) |> head()
```

Multi-source ETL pipelines combine several inputs. Suppose we have regional
CSV files that we want to merge and convert:

```{r etl-multi}
csv_north <- tempfile(fileext = ".csv")
csv_south <- tempfile(fileext = ".csv")
write.csv(obs[1:25000, ], csv_north, row.names = FALSE)
write.csv(obs[25001:50000, ], csv_south, row.names = FALSE)

merged_vtr <- tempfile(fileext = ".vtr")

bind_rows(
  tbl_csv(csv_north),
  tbl_csv(csv_south)
) |>
  filter(abundance > 0) |>
  write_vtr(merged_vtr, batch_size = 25000)

nrow(tbl(merged_vtr) |> collect())
```

Both CSVs streamed through the concat node, the filter, and the writer.
Neither was fully loaded into R at any point.

## Memory budget planning

Knowing which operations consume memory and how much lets us design pipelines
that stay within our system's limits. The central principle is "stream early,
materialize late." Every row that we can filter out before it reaches a
materializing operation (a sort, a join build, a grouping hash table) is a row
that never occupies memory. Pushing filters upstream is not just a performance
optimization; for larger-than-RAM data, it can be the difference between a
pipeline that completes and one that exhausts memory.

Here is a breakdown by operation type.

**Constant-memory (streaming) operations.** `filter`, `select`, `mutate`,
`limit`, `slice_head`, and `concat` (bind_rows) all process one batch at a
time. Their memory cost is proportional to the batch size and the number of
columns, not the total row count. For a typical batch of 50,000 rows with
10 columns, each batch might occupy a few megabytes. These operations are
safe for any dataset size.

**External sort (arrange).** vectra's sort node accumulates data in memory
up to a 1 GB budget. When the budget is exceeded, it flushes a sorted run
to a temporary file and continues. The final merge reads from all runs
simultaneously, using a heap that holds one row per run. Peak memory is
bounded at 1 GB plus the merge overhead. For datasets smaller than 1 GB
the sort completes entirely in memory.

**Hash aggregation (group_by + summarise).** The `group_agg` node maintains
one accumulator per distinct group key, so memory scales with the number of
distinct groups rather than the number of input rows. If we group by `species`
(12 values), the hash table is tiny. If we group by `obs_id` (50,000
distinct values), it is larger. Grouping by a high-cardinality column on
a billion-row dataset could create millions of accumulators, so it is worth
checking the expected group count.

**Hash join (build side).** The right-side table is fully materialized in a
hash table. Memory cost equals the right-side data size. A 1 million row
reference table with 5 columns might consume 50 to 100 MB. A 100 million row
table would require several gigabytes. The left side streams and adds no
persistent memory. String columns on the build side cost more than numeric
columns because each string value has variable length and requires its own
allocation. A build side with 1 million rows and a 200-character text column
will consume substantially more memory than one with only integer and double
columns.

**Window functions.** Window operations (`row_number`, `lag`, `lead`,
`cumsum`, etc.) operate on the current batch. Memory scales with batch size.
Partitioned windows (via `group_by`) hold data for the current partition.
If partitions are balanced and moderately sized, memory stays bounded.

To estimate memory for a pipeline, we identify the materializing operations
and estimate their footprints:

```{r memory-estimate}
# Example pipeline:
# tbl(huge.vtr) |>
#   filter(year == 2023) |>        -> streaming, ~5 MB per batch
#   left_join(sites, by = "site")  -> build side = 50 sites, ~1 KB
#   group_by(species) |>           -> 12 groups, ~1 KB
#   summarise(total = sum(abun))   -> 12 accumulators, ~1 KB
#   arrange(desc(total))           -> 12 rows, in-memory
#
# Total peak: ~5 MB (one batch) + negligible join/agg overhead

# Compare to:
# tbl(huge.vtr) |>
#   arrange(species)               -> external sort, up to 1 GB
#   left_join(big_ref, by = "id")  -> build side = big_ref size
#
# Total peak: 1 GB (sort) + big_ref size
```

The general strategy: stream everything we can, materialize only what we
must. Push filters as early as possible to reduce the volume flowing
through materializing operations. Put the smaller table on the right
side of joins. Pre-aggregate before joining when we only need summaries
from the right side.

To profile actual memory use, `explain()` prints the query plan tree before
execution. We can read the plan to count the materializing nodes and estimate
their size. For the build side of a join, we can `collect()` the right-side
node in isolation and check `object.size()` on the resulting data.frame. That
gives us the R-level footprint, which is a reasonable upper bound on the C-level
hash table size. For sort operations, the question is simpler: if the total
input is under 1 GB, the sort completes in memory; otherwise, it spills.

A practical example that puts these principles together:

```{r memory-pipeline}
final_path <- tempfile(fileext = ".vtr")

tbl(vtr_path) |>
  filter(year >= 2020, quality != "poor") |>
  left_join(tbl(site_path), by = "site") |>
  group_by(region, species) |>
  summarise(
    n_obs     = n(),
    mean_cov  = mean(cover_pct)
  ) |>
  arrange(region, desc(n_obs)) |>
  write_vtr(final_path)

tbl(final_path) |> collect() |> head(10)
```

The filter runs first, reducing the data volume before the join. The join's
build side is 50 rows of site metadata. The aggregation produces at most
4 regions times 12 species = 48 groups. The sort handles 48 rows in memory.
Total peak memory: one batch from the scan plus a few kilobytes. This
pipeline would work unchanged on a table with billions of rows.

## Cleanup

```{r cleanup}
all_files <- c(
  csv_path, vtr_path, clean_path, db_path, export_csv,
  del_demo, paste0(del_demo, ".del"),
  snap_v1, snap_v2,
  sorted_path, site_path, enriched, joined_path,
  summary_path, monthly_paths, consolidated, running_archive,
  vtr_archive, clean_archive, sqlite_export, summary_csv,
  csv_north, csv_south, merged_vtr,
  small_groups, large_groups,
  archive, csv_2017, archive_diff, compacted, compacted_del,
  final_path
)
unlink(all_files[file.exists(all_files)])
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_write_raster}
\alias{vec_write_raster}
\title{Write a raster matrix or 3D array to a .vec raster file}
\usage{
vec_write_raster(
  x,
  path,
  dtype = "f32",
  tile_size = 512L,
  extent = NULL,
  gt = NULL,
  epsg = 0L,
  nodata = NA_real_,
  band_names = NULL,
  compression = c("fast", "balanced", "max")
)
}
\arguments{
\item{x}{A numeric matrix \code{c(rows, cols)} for a single band, or a numeric
3D array \code{c(rows, cols, bands)} for multi-band.}

\item{path}{Output file path.}

\item{dtype}{Storage dtype, one of \code{"f64"}, \code{"f32"}, \code{"i8"}, \code{"u8"},
\code{"i16"}, \code{"u16"}, \code{"i32"}, \code{"u32"}, \code{"i64"}, \code{"u64"}. Defaults to \code{"f32"}
for floating-point input — \code{"f64"} doubles file size with no information
gain for typical climate rasters.}

\item{tile_size}{Square tile edge in pixels. Default 512.}

\item{extent}{Numeric vector \code{c(xmin, ymin, xmax, ymax)}. Used together with
the raster dimensions to derive the geotransform. Either \code{extent} or
\code{gt} must be supplied for georeferenced output.}

\item{gt}{Numeric(6) GDAL-style geotransform. Overrides \code{extent} if both
are given.}

\item{epsg}{EPSG code (integer) or 0L for none.}

\item{nodata}{Nodata value, or \code{NA_real_} to skip recording one.}

\item{band_names}{Optional character vector of length equal to the number
of bands.}

\item{compression}{Compression effort, one of \code{"fast"} (single spec, fast
encode), \code{"balanced"} (probe two entropy coders, ~2x encode time), or
\code{"max"} (probe six candidate specs per tile, slowest encode but
smallest file). Decode cost is unchanged across levels because each
tile records its own codec spec. Default \code{"fast"}.}
}
\value{
Invisible \code{NULL}.
}
\description{
Writes a row-major raster (one band) or a band-major 3D array (multi-band)
to the VECR raster format. Each tile is encoded as a self-describing tdc
block (PRED_2D + BYTE_SHUFFLE + LZ).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_more.R
\name{mask}
\alias{mask}
\title{Mask a streamed raster to a polygon layer}
\usage{
mask(
  x,
  mask,
  inverse = FALSE,
  band = NULL,
  background = NA_real_,
  path = NULL,
  dtype = "f32",
  compression = c("fast", "balanced", "max")
)
}
\arguments{
\item{x}{A \code{vectra_raster} (from \code{\link[=vec_open_raster]{vec_open_raster()}}) or a path to a \code{.vec}
raster.}

\item{mask}{An \code{sf} or \code{sfc} polygon layer to clip against. When it carries no
CRS it inherits the raster's EPSG.}

\item{inverse}{If \code{FALSE} (default) keep pixels inside \code{mask}; if \code{TRUE} keep
the pixels outside it.}

\item{band}{Band(s) to mask (1-based). Default \code{NULL} masks every band.}

\item{background}{Value written to cleared pixels. Default \code{NA_real_}.}

\item{path}{Optional output \code{.vec} path. When given the result is streamed to
disk and the opened \code{\link[=vec_open_raster]{vec_open_raster()}} handle is returned invisibly; when
\code{NULL} the result is returned in memory (a matrix for one band, a list of
matrices for several).}

\item{dtype}{Storage dtype for \code{.vec} output (see \code{\link[=vec_write_raster]{vec_write_raster()}}).
Default \code{"f32"}.}

\item{compression}{Compression effort for \code{.vec} output. Default \code{"fast"}.}
}
\value{
When \code{path} is \code{NULL}, a numeric matrix (one band) or a list of
matrices (several), each carrying \code{gt}, \code{extent}, and \code{crs} attributes
(row 1 northmost). When \code{path} is given, the written \code{vectra_raster} handle
(invisibly).
}
\description{
Keeps the pixels of a \code{.vec} raster whose cell centre falls inside a resident
polygon layer and sets the rest to \code{background}, reading the raster one
tile-row strip at a time so the whole grid is never resident. It is the raster
counterpart of \code{\link[=spatial_clip]{spatial_clip()}}: the streamed side is the (large) raster and
the small \code{mask} layer stays in memory. With \code{inverse = TRUE} the inside is
cleared and the outside kept.
}
\details{
This is the \emph{monoid fold} tier of the spatial toolbox: bounded to one strip
plus the resident mask, a single streaming pass, no spill. A pixel is tested
against the mask only when its centre falls in the mask bounding box, so the
point-in-polygon work stays proportional to the overlap rather than the whole
grid. Point-in-polygon is delegated to \pkg{sf} (an optional dependency);
topology and CRS handling are \pkg{sf}'s.
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
vals <- matrix(1:100, 10, 10, byrow = TRUE)
f <- tempfile(fileext = ".vec")
vec_write_raster(vals, f, dtype = "f64", extent = c(0, 0, 10, 10))

disc <- sf::st_buffer(sf::st_sfc(sf::st_point(c(5, 5))), 3)
inside <- mask(f, disc)
sum(!is.na(inside))
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_clip]{spatial_clip()}} for the vector analogue, \code{\link[=zonal]{zonal()}} for per-zone
summaries over the same pixel-in-polygon assignment.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_grouping.R
\name{summarise}
\alias{summarise}
\alias{summarize}
\title{Summarise grouped data}
\usage{
summarise(.data, ..., .groups = NULL)

summarize(.data, ..., .groups = NULL)
}
\arguments{
\item{.data}{A grouped \code{vectra_node} (from \code{\link[=group_by]{group_by()}}).}

\item{...}{Named aggregation expressions using \code{n()}, \code{sum()}, \code{mean()},
\code{min()}, \code{max()}, \code{sd()}, \code{var()}, \code{first()}, \code{last()}, \code{any()}, \code{all()},
\code{median()}, \code{n_distinct()}.}

\item{.groups}{How to handle groups in the result. One of \code{"drop_last"}
(default), \code{"drop"}, or \code{"keep"}.}
}
\value{
A \code{vectra_node} with one row per group.
}
\description{
Summarise grouped data
}
\details{
Aggregation is hash-based by default. When the engine detects it is
advantageous, it switches to a sort-based path that can spill to disk,
keeping memory bounded regardless of group count.

All aggregation functions accept \code{na.rm = TRUE} to skip NA values.
Without \code{na.rm}, any NA in a group poisons the result (returns NA).
R-matching edge cases: \code{sum(na.rm = TRUE)} on all-NA returns 0,
\code{mean(na.rm = TRUE)} on all-NA returns NaN, \code{min/max(na.rm = TRUE)} on
all-NA returns Inf/-Inf with a warning.

This is a materializing operation.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> group_by(cyl) |> summarise(avg_mpg = mean(mpg)) |> collect()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_grouping.R
\name{group_by}
\alias{group_by}
\title{Group a vectra query by columns}
\usage{
group_by(.data, ...)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Grouping column names (unquoted).}
}
\value{
A \code{vectra_node} with grouping information stored.
}
\description{
Group a vectra query by columns
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> group_by(cyl) |> summarise(avg = mean(mpg)) |> collect()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_join}
\alias{spatial_join}
\title{Spatial join a streamed query against a resident sf object}
\usage{
spatial_join(
  x,
  y,
  join = NULL,
  geom = "geometry",
  coords = NULL,
  crs = NA,
  left = TRUE,
  suffix = c(".x", ".y"),
  partition = NULL,
  y_geom = NULL,
  y_coords = NULL,
  out_geom = NULL,
  flush_rows = NULL,
  ...
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{y}{The right side of the join: an \code{sf} object held resident (the
default), or -- when \code{partition} is given -- a streamed \code{vectra_node}.}

\item{join}{An \pkg{sf} binary predicate function, e.g. \link[sf:st_intersects]{sf::st_intersects}
(default), \link[sf:st_within]{sf::st_within}, \link[sf:st_contains]{sf::st_contains}, \link[sf:st_nearest_feature]{sf::st_nearest_feature}.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{coords}{Optional length-2 character vector naming the x and y
coordinate columns to assemble point geometry from (e.g. \code{c("x", "y")}),
for inputs such as \code{\link[=tiff_extract_points]{tiff_extract_points()}} output. The coordinate columns
are retained.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{left}{If \code{TRUE} (default) keep every left row (left join); if \code{FALSE}
keep only matches (inner join).}

\item{suffix}{Length-2 character vector disambiguating columns present on
both sides. Default \code{c(".x", ".y")}.}

\item{partition}{Optional \code{\link[=grid]{grid()}} specification enabling the two-sided
streamed path, in which \code{y} is itself a \code{vectra_node}. Default \code{NULL} keeps
the resident-\code{y} path.}

\item{y_geom, y_coords}{Geometry transport for a streamed \code{y} under
\code{partition}: the name of \code{y}'s hex-WKB geometry column (\code{y_geom}, default
the left \code{geom}), or a length-2 character vector of \code{y}'s coordinate columns
(\code{y_coords}). Ignored without \code{partition}.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}

\item{...}{Further arguments passed to \code{\link[sf:st_join]{sf::st_join()}}.}
}
\value{
A \code{vectra_node} of the joined stream, backed by temporary \code{.vtr}
spills and carrying the left CRS.
}
\description{
Streams a large left side \code{x} through the engine and joins each batch against
a small right side \code{y} held resident in memory, using an \pkg{sf} binary
predicate (\code{st_intersects} by default). This is the spatial analogue of a
hash join with the small side on the build side: the billion-row left stream
never materializes, while \code{y} (admin polygons, habitat patches, ...) stays in
RAM. The dominant real workload it serves is tagging huge point sets with the
polygon they fall in.
}
\details{
For the recognised predicates -- the topological ones (intersects, within,
contains, overlaps, covers, covered by, touches, crosses), equals,
within-distance (\link[sf:st_is_within_distance]{sf::st_is_within_distance}, radius passed as \verb{dist =}), and
nearest feature (\link[sf:st_nearest_feature]{sf::st_nearest_feature}) -- on projected or unprojected
planar data, the match runs natively on the GEOS C API straight off the
hex-WKB column: \code{y} is parsed once into a spatial index, each batch's matches
come back from C, and \code{y}'s attributes are attached in R without decoding the
left side to \pkg{sf}. Coordinate-assembled (\code{coords}) point input runs
natively too, building each point in C (the emitted point geometry is built in
C as well). Geographic coordinates with spherical geometry on
(\code{sf::sf_use_s2()}), a disjoint join (whose matches are the bounding-box
complement an index cannot prune), and other extra \code{\link[sf:st_join]{sf::st_join()}} arguments
use \pkg{sf} instead, preserving its semantics.

When both sides are larger than RAM, pass \code{partition = grid(cellsize)} and a
streamed \code{vectra_node} as \code{y}: both inputs are binned to a uniform spatial
grid, then joined one shard at a time. Each left feature is assigned to the
single grid cell of its reference point while each right feature is
replicated to every cell its bounding box overlaps, so a left row is emitted
exactly once and the result equals the resident join. This is exact for point
left geometries (the dominant case -- tagging a huge point set with the
polygon it falls in) and finds, for an extended left feature, the matches
whose right bounding box overlaps the left reference cell; choose a \code{cellsize}
larger than the left features for an extended-on-extended join. The partition
path serves topological predicates (intersects, within, contains, overlaps,
covers, covered by). It also serves \link[sf:st_nearest_feature]{sf::st_nearest_feature}: because nearest
is not local to one cell, each left feature then searches its own cell and the
eight around it, so the true nearest is found when it lies within one cell of
the left reference cell (pick a \code{cellsize} at least the largest expected
nearest distance). Topology and CRS handling are \pkg{sf}'s; vectra supplies
the stream and the grid partition.
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)

# A stream of points, stored with x/y coordinate columns.
set.seed(1)
pts <- sf::st_coordinates(sf::st_sample(nc, 200))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), f)

# Tag each point with the county it falls in, streaming.
tagged <- tbl(f) |>
  spatial_join(nc["NAME"], join = sf::st_intersects,
               coords = c("x", "y"), crs = sf::st_crs(nc))
head(collect(tagged))

# Both sides streamed: bin to a grid and join per shard. Here y is a
# vectra_node rather than a resident sf object.
g <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME = nc$NAME,
  geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
), g)
tagged2 <- tbl(f) |>
  spatial_join(tbl(g), coords = c("x", "y"), crs = sf::st_crs(nc),
               partition = grid(0.5))
head(collect(tagged2))
unlink(c(f, g))
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_map]{spatial_map()}} for per-feature transforms, \code{\link[=collect_sf]{collect_sf()}} to
materialize as \code{sf}, \code{\link[=offload]{offload()}} to partition both-sides-huge joins.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explain.R
\name{explain}
\alias{explain}
\title{Print the execution plan for a vectra query}
\usage{
explain(x, ...)
}
\arguments{
\item{x}{A \code{vectra_node} object.}

\item{...}{Ignored.}
}
\value{
Invisible \code{x}.
}
\description{
Shows the node types, column schemas, and structure of the lazy query plan.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> filter(cyl > 4) |> select(mpg, cyl) |> explain()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete.R
\name{delete_vtr}
\alias{delete_vtr}
\title{Logically delete rows from a .vtr file}
\usage{
delete_vtr(path, row_ids)
}
\arguments{
\item{path}{File path of the \code{.vtr} file to delete rows from.}

\item{row_ids}{A numeric vector of \strong{0-based} physical row indices to
delete. Out-of-range indices are silently ignored on read (they will
never match a real row).}
}
\value{
Invisible \code{NULL}.
}
\description{
Marks the specified 0-based physical row indices as deleted by writing (or
updating) a tombstone side file (\verb{<path>.del}). The original \code{.vtr} file is
never modified. The next call to \code{\link[=tbl]{tbl()}} on the same path will automatically
exclude the deleted rows.
}
\details{
Tombstone files are cumulative: calling \code{delete_vtr()} multiple times on the
same file merges all deletions (union, deduplicated). To undo deletions,
remove the \code{.del} file manually with \code{unlink(paste0(path, ".del"))}.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

# Delete the first and third rows (0-based indices 0 and 2)
delete_vtr(f, c(0, 2))

result <- tbl(f) |> collect()
stopifnot(nrow(result) == nrow(mtcars) - 2L)

unlink(c(f, paste0(f, ".del")))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reframe.R
\name{reframe}
\alias{reframe}
\title{Summarise with variable-length output per group}
\usage{
reframe(.data, ...)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Named expressions.}
}
\value{
A data.frame (not a lazy node).
}
\description{
Like \code{\link[=summarise]{summarise()}} but allows expressions that return more than one row
per group. Currently implemented via \code{collect()} fallback.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(g = c("a", "a", "b"), x = c(1, 2, 3)), f)
tbl(f) |> group_by(g) |> reframe(range_x = range(x))
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_basic.R
\name{mutate}
\alias{mutate}
\title{Add or transform columns}
\usage{
mutate(.data, ...)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Named expressions for new or transformed columns.}
}
\value{
A new \code{vectra_node} with mutated columns.
}
\description{
Add or transform columns
}
\details{
Supported expression types: arithmetic (\code{+}, \code{-}, \code{*}, \code{/}, \code{\%\%}),
comparison, boolean, \code{is.na()}, \code{nchar()}, \code{substr()}, \code{grepl()} (fixed
match only). Window functions (\code{row_number()}, \code{rank()}, \code{dense_rank()},
\code{lag()}, \code{lead()}, \code{cumsum()}, \code{cummean()}, \code{cummin()}, \code{cummax()}) are
detected automatically and routed to a dedicated window node.

When grouped, window functions respect partition boundaries.

This is a streaming operation for regular expressions; window functions
materialize all rows within each partition.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> mutate(kpl = mpg * 0.425144) |> collect() |> head()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/offload.R
\name{offload}
\alias{offload}
\title{Spill a query to disk and stream it back (the offload functor)}
\usage{
offload(
  x,
  by = NULL,
  n = NULL,
  method = c("auto", "level", "range", "hash"),
  path = NULL,
  compress = c("fast", "small", "none")
)
}
\arguments{
\item{x}{A \code{vectra_node} to materialize.}

\item{by}{Optional name (string) of a partition key column. When supplied,
the result is a partition rather than a single node.}

\item{n}{Number of buckets for \code{method = "range"} or \code{"hash"}. Ignored for a
one-shard-per-value partition.}

\item{method}{Partition strategy: \code{"auto"} (default; one shard per value for
a discrete key, \code{n} ranges for a numeric key), \code{"level"} (one shard per
distinct value), \code{"range"} (\code{n} equal-width value ranges), or \code{"hash"}
(\code{n} buckets by a stable hash of the key, co-locating each key).}

\item{path}{Optional file path for a durable replay-cache spill (used only
when \code{by} is \code{NULL}). When \code{NULL} a temporary file is used and removed when
the returned node is garbage-collected.}

\item{compress}{Compression for spill files, passed to \code{\link[=write_vtr]{write_vtr()}}:
\code{"fast"} (default), \code{"small"}, or \code{"none"}.}
}
\value{
A \code{vectra_node} (no \code{by}) or a \code{vectra_partition} (with \code{by}), each
carrying a cost grade shown by \code{\link[=print]{print()}} and \code{\link[=explain]{explain()}}.
}
\description{
Materializes a query once to disk and returns a stream that holds the same
rows, so every later pass is a disk scan instead of a re-run of the upstream
pipeline. The materialization streams batch by batch, so peak memory stays at
one batch regardless of result size. This is the bridge from the bounded
single-pass world of \code{\link[=collect_chunked]{collect_chunked()}} to out-of-core fits.
}
\details{
With no \code{by}, \code{offload()} returns a \strong{replay cache}: a \code{vectra_node} backed
by one \code{.vtr} file. Feed it to a pull-based consumer such as
\code{biglm::bigglm()} through \code{\link[=chunk_feeder]{chunk_feeder()}}, which accepts an offloaded node
directly, so each iteratively reweighted pass reads the prepared columns from
disk rather than rebuilding them. Bake the selects and mutates into the query
you offload, and replay does no further work.

With \code{by}, \code{offload()} returns a \strong{partition}: the rows split into disjoint
shards, one per key value (discrete key) or per value range (\code{method = "range"}, or any numeric key), written in a single streaming pass. A
partition prints as a list of shards and behaves like one: \code{length()},
\code{names()} (the keys), \code{p[["key"]]} (a shard node), and \code{lapply(p, ...)} all
work. Fold it with \code{\link[=collect_chunked]{collect_chunked()}} (supplying \code{combine}). The union of
the shards reproduces the input; row totals are checked.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

# Replay cache: same rows, now on disk.
s <- offload(tbl(f) |> filter(cyl > 4) |> select(mpg, wt, hp))
nrow(collect(s))

# Partition by a key: a list of per-shard nodes.
p <- offload(tbl(f), by = "cyl")
names(p)
length(p)
nrow(collect(p[[1]]))
unlink(f)

}
\seealso{
\code{\link[=chunk_feeder]{chunk_feeder()}} (accepts an offloaded node), \code{\link[=collect_chunked]{collect_chunked()}}
for the partitioned monoidal reduce, and \code{\link[=arrange]{arrange()}} for the external-sort
instance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{write_tiff}
\alias{write_tiff}
\title{Write query results to a GeoTIFF file}
\usage{
write_tiff(
  x,
  path,
  compress = FALSE,
  pixel_type = "float64",
  metadata = NULL,
  crs = NULL,
  tiled = FALSE,
  tile_size = 256L,
  bigtiff = "auto",
  ...
)
}
\arguments{
\item{x}{A \code{vectra_node} (lazy query) or a \code{data.frame}.}

\item{path}{File path for the output GeoTIFF file.}

\item{compress}{Logical; use DEFLATE compression? Default \code{FALSE}.}

\item{pixel_type}{Character string specifying the output pixel type.
One of \code{"float64"} (default), \code{"float32"}, \code{"int16"}, \code{"int32"},
\code{"uint8"}, or \code{"uint16"}.}

\item{metadata}{Optional character string of GDAL_METADATA XML to embed
in the file (tag 42112). Use \code{\link[=tiff_metadata]{tiff_metadata()}} to read it back.}

\item{crs}{Optional CRS to embed as a GeoKey directory (TIFF tag 34735).
Accepts an integer EPSG code, an \code{"EPSG:xxxx"} string, or a list with
named fields \code{epsg}, \code{geographic} (\code{TRUE}/\code{FALSE}), and optionally
\code{citation}. Codes that are not auto-classified as projected/geographic
default to projected; pass \code{geographic = TRUE} to override.
Use \code{\link[=tiff_crs]{tiff_crs()}} to read it back.}

\item{tiled}{Logical; write a tiled GeoTIFF (TIFF tags 322/323/324/325)
instead of strips. Default \code{FALSE}. Tiled layout enables random-access
block reads and is required for Cloud-Optimized GeoTIFF (COG).}

\item{tile_size}{Integer; tile edge length in pixels. Must be a positive
multiple of 16 (TIFF spec). Either a single value (square tiles) or a
length-2 vector \code{c(width, height)}. Default \code{256}. Edge tiles at the
right and bottom of the image are padded to full tile size with the
NoData / NaN value.}

\item{bigtiff}{Controls BigTIFF dispatch. \code{"auto"} (default) emits BigTIFF
when the expected raw payload would exceed the classic-TIFF 4 GB
ceiling, otherwise emits classic TIFF. \code{TRUE} forces BigTIFF (magic
\code{0x002B}, 64-bit offsets), useful for round-trip tests on small data.
\code{FALSE} forces classic TIFF — beware that classic TIFF will silently
corrupt outputs larger than 4 GB. Tiled BigTIFF is not yet supported.}

\item{...}{Reserved for future use.}
}
\value{
Invisible \code{NULL}.
}
\description{
The data must contain \code{x} and \code{y} columns (pixel center coordinates) and
one or more numeric band columns. Grid dimensions and geotransform are
inferred from the x/y coordinate arrays. Missing pixels are written as NaN
(or the type-appropriate nodata value for integer pixel types).
}
\examples{
\donttest{
# Write as int16 with DEFLATE compression and an EPSG:4326 GeoKey
df <- data.frame(x = 1:4, y = rep(1:2, each = 2), band1 = c(100, 200, 300, 400))
f <- tempfile(fileext = ".tif")
write_tiff(df, f, compress = TRUE, pixel_type = "int16", crs = 4326L)
tiff_crs(f)
unlink(f)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{tiff_metadata}
\alias{tiff_metadata}
\title{Read GDAL_METADATA from a GeoTIFF}
\usage{
tiff_metadata(path)
}
\arguments{
\item{path}{Path to a GeoTIFF file.}
}
\value{
A single character string containing the XML, or \code{NA_character_}.
}
\description{
Returns the GDAL_METADATA XML string (TIFF tag 42112) embedded in a
GeoTIFF file. Returns \code{NA} if the tag is not present.
}
\examples{
\donttest{
f <- tempfile(fileext = ".tif")
df <- data.frame(x = 1:4, y = rep(1:2, each = 2), band1 = as.double(1:4))
write_tiff(df, f, metadata = "<GDALMetadata></GDALMetadata>")
tiff_metadata(f)
unlink(f)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_topology.R
\name{spatial_locate}
\alias{spatial_locate}
\title{Locate streamed points along a resident line layer}
\usage{
spatial_locate(
  x,
  line,
  geom = "geometry",
  coords = NULL,
  crs = NA,
  y_id = NULL,
  id_col = "line",
  measure_col = "measure",
  dist_col = "distance",
  snap = FALSE,
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{line}{An \code{sf} or \code{sfc} object of the reference lines (the resident
layer).}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{coords}{Optional length-2 character vector naming the x and y
coordinate columns to assemble point geometry from (e.g. \code{c("x", "y")}),
for inputs such as \code{\link[=tiff_extract_points]{tiff_extract_points()}} output. The coordinate columns
are retained.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{y_id}{Optional name of a column in \code{line} whose value identifies the
matched line in the output. Default \code{NULL} uses \code{line}'s 1-based row index.}

\item{id_col, measure_col, dist_col}{Names of the output columns holding the
matched-line identifier, the measure along the line, and the perpendicular
distance. Defaults \code{"line"}, \code{"measure"}, \code{"distance"}.}

\item{snap}{If \code{TRUE}, replace each point's geometry with its projection onto
the nearest line. Default \code{FALSE} keeps the original points.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of \code{x}'s rows -- geometry included (or snapped onto
the line) -- plus the matched-line identifier, the measure, and the
perpendicular distance, backed by temporary \code{.vtr} spills (removed when the
node is garbage-collected) and carrying the input CRS.
}
\description{
Streams a large point layer \code{x} through the engine and, for each point, finds
the nearest line of a small resident \code{line} layer and where the point falls
along it -- linear referencing (\code{sf::st_line_project}). Each point gets the
identifier of its nearest line, the \strong{measure} (distance along that line from
its start to the point's projection), and the perpendicular distance to the
line. With \code{snap = TRUE} the point geometry is moved onto the line at that
measure. This is the two-layer companion to a per-feature
\code{sf::st_line_interpolate}, which goes the other way (a measure back to a
point); the billion-row point stream never materializes, while \code{line} (the
reference network) stays resident.
}
\details{
Nearest line and distance are \pkg{sf}'s \link[sf:st_nearest_feature]{sf::st_nearest_feature} and
\link[sf:st_distance]{sf::st_distance}: planar (CRS units) on projected or unprojected planar data,
great-circle (metres) on geographic coordinates with spherical geometry on
(\code{sf::sf_use_s2()}). Points arrive either as a hex-WKB geometry column
(\code{geom}) or as two coordinate columns (\code{coords}). The \pkg{sf} package is an
optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
line <- sf::st_sfc(
  sf::st_linestring(rbind(c(0, 0), c(10, 0))),
  sf::st_linestring(rbind(c(0, 5), c(0, 15))))
line <- sf::st_sf(road = c("main", "side"), geometry = line)
pts  <- data.frame(id = 1:2, x = c(3, 1), y = c(1, 9))
f <- tempfile(fileext = ".vtr")
write_vtr(pts, f)

# Each point's position along its nearest road.
tbl(f) |>
  spatial_locate(line, coords = c("x", "y"), y_id = "road") |>
  collect()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_knn]{spatial_knn()}} for nearest neighbours with distances,
\code{\link[=spatial_join]{spatial_join()}} for a nearest-feature attribute join, \code{\link[=spatial_map]{spatial_map()}} with
\code{~ sf::st_line_interpolate(line, .x$m)} for the inverse, \code{\link[=collect_sf]{collect_sf()}} to
materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/across.R
\name{across}
\alias{across}
\title{Apply a function across multiple columns}
\usage{
across(.cols, .fns, ..., .names = NULL)
}
\arguments{
\item{.cols}{Column selection (tidyselect).}

\item{.fns}{A function, formula, or named list of functions.}

\item{...}{Additional arguments passed to \code{.fns}.}

\item{.names}{A glue-style naming pattern. Uses \code{{.col}} and \code{{.fn}}.
Default: \code{"{.col}"} if \code{.fns} is a single function,
\code{"{.col}_{.fn}"} if \code{.fns} is a named list.}
}
\value{
A named list used internally by mutate/summarise.
}
\description{
Used inside \code{\link[=mutate]{mutate()}} or \code{\link[=summarise]{summarise()}} to apply a function to multiple
columns selected with tidyselect. Returns a named list of expressions.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
# In summarise (conceptual; across is expanded to individual expressions)
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema.R
\name{vtr_schema}
\alias{vtr_schema}
\title{Create a star schema over linked vectra tables}
\usage{
vtr_schema(fact, ...)
}
\arguments{
\item{fact}{A \code{vectra_node} object (the central fact table). Must be
file-backed (created via \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_csv]{tbl_csv()}}, or \code{\link[=tbl_sqlite]{tbl_sqlite()}}).}

\item{...}{Named \code{vectra_link} objects created by \code{\link[=link]{link()}}. Names become
the dimension aliases used in \code{\link[=lookup]{lookup()}} (e.g., \code{species$name}).}
}
\value{
A \code{vectra_schema} object.
}
\description{
Registers a fact table with named dimension links. The schema enables
\code{\link[=lookup]{lookup()}} to resolve columns from dimension tables without writing
explicit joins.
}
\examples{
\donttest{
f_obs <- tempfile(fileext = ".vtr")
f_sp  <- tempfile(fileext = ".vtr")
f_ct  <- tempfile(fileext = ".vtr")
write_vtr(data.frame(sp_id = 1:3, ct_code = c("AT", "DE", "FR"),
                      value = 10:12), f_obs)
write_vtr(data.frame(sp_id = 1:3,
                      name = c("Oak", "Beech", "Pine")), f_sp)
write_vtr(data.frame(ct_code = c("AT", "DE", "FR"),
                      gdp = c(400, 3800, 2700)), f_ct)

s <- vtr_schema(
  fact    = tbl(f_obs),
  species = link("sp_id", tbl(f_sp)),
  country = link("ct_code", tbl(f_ct))
)
print(s)
unlink(c(f_obs, f_sp, f_ct))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_utility.R
\name{slice}
\alias{slice}
\title{Select rows by position}
\usage{
slice(.data, ...)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Integer row indices (positive or negative).}
}
\value{
A data.frame with the selected rows.
}
\description{
Select rows by position
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> slice(1, 3, 5)
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_more.R
\name{proximity}
\alias{proximity}
\title{Euclidean distance to the nearest feature (proximity)}
\usage{
proximity(
  x,
  target = NULL,
  band = 1L,
  path = NULL,
  dtype = "f32",
  compression = c("fast", "balanced", "max")
)
}
\arguments{
\item{x}{A \code{vectra_raster} (from \code{\link[=vec_open_raster]{vec_open_raster()}}) or a path to a \code{.vec}
raster.}

\item{target}{Optional numeric vector of feature values. When \code{NULL}
(default) every non-NA cell is a feature; otherwise a cell is a feature
when its value is in \code{target}.}

\item{band}{Band to read (1-based). Default 1.}

\item{path}{Optional output \code{.vec} path. When given the result is streamed to
disk and the opened \code{\link[=vec_open_raster]{vec_open_raster()}} handle is returned invisibly; when
\code{NULL} the result is returned as an in-memory matrix.}

\item{dtype}{Storage dtype for \code{.vec} output (see \code{\link[=vec_write_raster]{vec_write_raster()}}).
Default \code{"f32"}.}

\item{compression}{Compression effort for \code{.vec} output. Default \code{"fast"}.}
}
\value{
When \code{path} is \code{NULL}, a numeric matrix (row 1 northmost) carrying
\code{gt}, \code{extent}, and \code{crs} attributes, with distance in CRS units and \code{NA}
where the raster holds no feature anywhere. When \code{path} is given, the
written \code{vectra_raster} handle (invisibly).
}
\description{
Computes, for every cell of a \code{.vec} raster, the straight-line Euclidean
distance to the nearest feature cell, in CRS units. Feature cells are the
non-NA cells by default, or the cells whose value is in \code{target}. This is the
raster proximity / Euclidean-distance staple, the distance companion to
\code{\link[=rasterize]{rasterize()}}.
}
\details{
The exact Euclidean distance transform is separable (Felzenszwalb and
Huttenlocher 2012): a one-dimensional lower-envelope-of-parabolas transform
along the rows, then the same transform along the columns, each linear in the
line length and each line independent. vectra runs it as four streamed passes
over tile-row strips, with an out-of-core transpose between the row pass and
the column pass, so the whole grid is never resident. The row pass scales
squared distances by the x resolution and the column pass by the y
resolution, so the result is exact on anisotropic (non-square) cells. This
places proximity on the sort / partition tier of the spatial toolbox.

Distances are straight-line Euclidean in the raster CRS units. Cost-distance,
which accumulates a per-cell friction along the path, is a global
shortest-path problem and stays resident: \code{\link[=collect]{collect()}} the raster and run a
resident solver for that.
}
\examples{
m <- matrix(NA_real_, 12, 12)
m[3, 4] <- 1; m[9, 10] <- 1
f <- tempfile(fileext = ".vec")
vec_write_raster(m, f, dtype = "f64", extent = c(0, 0, 12, 12))

d <- proximity(f)
round(d[1:3, 1:3], 2)
unlink(f)

}
\seealso{
\code{\link[=rasterize]{rasterize()}} to build a raster from streamed points, \code{\link[=mask]{mask()}} to
clip a raster to a polygon layer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_utility.R
\name{head.vectra_node}
\alias{head.vectra_node}
\title{Limit results to first n rows}
\usage{
\method{head}{vectra_node}(x, n = 6L, ...)
}
\arguments{
\item{x}{A \code{vectra_node} object.}

\item{n}{Number of rows to return.}

\item{...}{Ignored.}
}
\value{
A data.frame with the first \code{n} rows.
}
\description{
Limit results to first n rows
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{zonal}
\alias{zonal}
\title{Summarise raster values within zones}
\usage{
zonal(
  raster,
  zones,
  fun = "mean",
  band = 1L,
  zone_band = 1L,
  zone_field = NULL,
  na.rm = TRUE
)
}
\arguments{
\item{raster}{A \code{vectra_raster} (from \code{\link[=vec_open_raster]{vec_open_raster()}}) or a path to a
\code{.vec} raster holding the values to summarise.}

\item{zones}{The zones to summarise within: a \code{vectra_raster} / \code{.vec} path
aligned to \code{raster} (zone id per pixel), or an \code{sf}/\code{sfc} polygon layer.}

\item{fun}{One or more of \code{"mean"}, \code{"sum"}, \code{"count"}, \code{"min"}, \code{"max"},
\code{"sd"}. Each becomes a column in the result. Default \code{"mean"}.}

\item{band}{Band of the value \code{raster} to summarise (1-based). Default 1.}

\item{zone_band}{Band of a raster \code{zones} holding the zone ids. Default 1.}

\item{zone_field}{For an \code{sf} \code{zones} layer, the column giving each polygon's
zone id. Default \code{NULL} uses the polygon row index \code{1:n}.}

\item{na.rm}{If \code{TRUE} (default) skip nodata pixels; if \code{FALSE} let a nodata
pixel propagate \code{NA} to its zone's statistics.}
}
\value{
A data.frame with a \code{zone} column (sorted) followed by one column per
\code{fun}, one row per zone.
}
\description{
Reduces a raster to one summary row per zone, streaming the raster one
tile-row strip at a time so the whole grid never has to be resident. Zones
come either from a second raster aligned to the value grid (each pixel's zone
is that raster's value, the \code{terra::zonal} pattern) or from an \pkg{sf}
polygon layer (each pixel is assigned the polygon its centre falls in). The
per-zone running moments (count, sum, sum of squares, min, max) are folded in
memory as strips arrive, so peak memory is one strip plus the small per-zone
table regardless of raster size. This is the \emph{monoid fold} tier of the
spatial toolbox: bounded memory, a single streaming pass, no spill.
}
\details{
\code{sd} is derived from the streamed moments
(\code{sqrt((sum2 - sum^2 / n) / (n - 1))}), so it needs no second pass. With
\code{na.rm = TRUE} (the default) nodata pixels are skipped; with \code{na.rm = FALSE}
any nodata pixel in a zone makes that zone's \code{sum}/\code{mean}/\code{min}/\code{max}/\code{sd}
\code{NA}, matching the resident behaviour. \code{count} always reports the number of
non-nodata cells in the zone.

Polygon zones assign each pixel centre to a polygon natively on the GEOS C
API (the polygons parsed once into a spatial index, every strip's centres
located in C), so \pkg{sf} is touched only to read the polygons in; geographic
polygons with spherical geometry on (\code{sf::sf_use_s2()}) keep the \pkg{sf}
point-in-polygon path. Raster zones are fully \pkg{sf}-free. The zone raster
must share the value raster's dimensions and geotransform.
}
\examples{
# A value raster and an aligned 2x2-block zone raster on a 4x4 grid.
vals <- matrix(1:16, 4, 4, byrow = TRUE)
zone <- matrix(c(1, 1, 2, 2, 1, 1, 2, 2,
                 3, 3, 4, 4, 3, 3, 4, 4), 4, 4, byrow = TRUE)
fv <- tempfile(fileext = ".vec"); fz <- tempfile(fileext = ".vec")
vec_write_raster(vals, fv, dtype = "f64", extent = c(0, 0, 4, 4))
vec_write_raster(zone, fz, dtype = "f64", extent = c(0, 0, 4, 4))

zonal(fv, fz, fun = c("mean", "sum", "count"))
unlink(c(fv, fz))

}
\seealso{
\code{\link[=rasterize]{rasterize()}} to build a value raster from streamed points,
\code{\link[=vec_open_raster]{vec_open_raster()}} to open the inputs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_basic.R
\name{desc}
\alias{desc}
\title{Mark a column for descending sort order}
\usage{
desc(x)
}
\arguments{
\item{x}{A column name.}
}
\value{
A marker used by \code{\link[=arrange]{arrange()}}.
}
\description{
Used inside \code{\link[=arrange]{arrange()}} to sort a column in descending order.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> arrange(desc(mpg)) |> collect() |> head()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_raster_times}
\alias{vec_raster_times}
\title{Distinct time stamps stored in a .vec time cube}
\usage{
vec_raster_times(r, band = 1L, level = 0L)
}
\arguments{
\item{r}{A \code{vectra_raster}.}

\item{band}{1-based band index.}

\item{level}{Overview level.}
}
\value{
Numeric vector of stamps (length 0 when the file has no time
information).
}
\description{
Returns the ascending vector of time stamps recorded for the given
(band, level). Pixel-major files store one consolidated table; image-
major files derive the list from the per-tile time field.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{tbl_csv}
\alias{tbl_csv}
\title{Create a lazy table reference from a CSV file}
\usage{
tbl_csv(path, batch_size = .DEFAULT_BATCH_SIZE)
}
\arguments{
\item{path}{Path to a \code{.csv} or \code{.csv.gz} file.}

\item{batch_size}{Number of rows per batch (default 65536).}
}
\value{
A \code{vectra_node} object representing a lazy scan of the CSV file.
}
\description{
Opens a CSV file for lazy, streaming query execution. Column types are
inferred from the first 1000 rows. No data is read until \code{\link[=collect]{collect()}} is
called. Gzip-compressed files (\code{.csv.gz}) are supported transparently.
}
\examples{
f <- tempfile(fileext = ".csv")
write.csv(mtcars, f, row.names = FALSE)
node <- tbl_csv(f)
print(node)
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{write_vtr}
\alias{write_vtr}
\title{Write data to a .vtr file}
\usage{
write_vtr(
  x,
  path,
  compress = c("fast", "small", "none"),
  batch_size = NULL,
  col_types = NULL,
  quantize = NULL,
  spatial = NULL,
  ...
)
}
\arguments{
\item{x}{A \code{vectra_node} (lazy query) or a \code{data.frame}.}

\item{path}{File path for the output .vtr file.}

\item{compress}{Compression level: \code{"fast"} (default, byte-shuffle + greedy
LZ), \code{"small"} (per-block adaptive — tries greedy LZ, separated-streams
LZ, and LZ + Huffman entropy coding, and writes whichever shrank the
block the most; never worse than \code{"fast"} on any block, typically
10-25 percent smaller files at the cost of slower encode), or \code{"none"}.}

\item{batch_size}{Target number of rows per row group in the output file.
Defaults to 131072 for data.frames (1 MB per double column, cache-friendly
for decompression). For nodes, defaults to \code{NULL} (one row group per
upstream batch).}

\item{col_types}{Optional named character vector specifying narrow integer
storage types. Names must match column names; values must be \code{"int8"},
\code{"int16"}, or \code{"int32"}. Only applies to integer columns.
Example: \code{col_types = c(age = "int8", year = "int16")}.}

\item{quantize}{Optional named list for lossy quantization of \code{double}
columns. Each element is named after a column and is itself a named list
with \code{scale} (or \code{precision = 1/scale}), \code{type} (\code{"int8"}, \code{"int16"},
\code{"int32"}; default \code{"int16"}), and optionally \code{offset} (default 0).
Example: \code{quantize = list(temp = list(precision = 0.001, type = "int16"))}.}

\item{spatial}{Optional list for 2D spatial predictor encoding. Either a
global spec applied to all numeric columns (\code{list(nx = 2000, ny = 2000)})
or per-column specs (\code{list(temp = list(nx = 2000, ny = 2000))}).
When provided, a spatial predictor removes smooth 2D trends before
compression, dramatically improving compression of raster data.
Combines with \code{quantize} for maximum effect.}

\item{...}{Additional arguments passed to methods.}
}
\value{
Invisible \code{NULL}.
}
\description{
For \code{vectra_node} inputs (lazy queries from any format: CSV, SQLite, TIFF,
or another .vtr), data is streamed batch-by-batch to disk without
materializing the full result in memory. Each batch becomes one row group.
The output file is written atomically (via temp file + rename) so readers
never see a partial file.
}
\details{
For \code{data.frame} inputs, the data is written directly from memory.
}
\examples{
# From a data.frame
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

# Streaming format conversion (CSV -> VTR)
csv <- tempfile(fileext = ".csv")
write.csv(mtcars, csv, row.names = FALSE)
f2 <- tempfile(fileext = ".vtr")
tbl_csv(csv) |> write_vtr(f2)

unlink(c(f, f2, csv))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_more.R
\name{rast_calc}
\alias{rast_calc}
\title{Cellwise calculation over aligned rasters (map algebra)}
\usage{
rast_calc(
  rasters,
  expr,
  band = 1L,
  path = NULL,
  dtype = "f32",
  compression = c("fast", "balanced", "max")
)
}
\arguments{
\item{rasters}{A named list of \code{vectra_raster} handles or \code{.vec} paths sharing
a grid. The names are the variables available inside \code{expr}.}

\item{expr}{An expression in those names producing one value per cell (or a
scalar, recycled). Evaluated against each strip with the caller's
environment as the enclosing scope.}

\item{band}{Band read from every input (1-based). Default 1.}

\item{path}{Optional output \code{.vec} path. When given the result is streamed to
disk and the opened \code{\link[=vec_open_raster]{vec_open_raster()}} handle is returned invisibly; when
\code{NULL} the result is returned as an in-memory matrix.}

\item{dtype}{Storage dtype for \code{.vec} output (see \code{\link[=vec_write_raster]{vec_write_raster()}}).
Default \code{"f32"}.}

\item{compression}{Compression effort for \code{.vec} output. Default \code{"fast"}.}
}
\value{
When \code{path} is \code{NULL}, a numeric matrix (row 1 northmost) carrying
\code{gt}, \code{extent}, and \code{crs} attributes. When \code{path} is given, the written
\code{vectra_raster} handle (invisibly).
}
\description{
Evaluates an expression cell by cell across one or more \code{.vec} rasters that
share a grid, reading every input one tile-row strip at a time so no whole
band is ever resident. Inside \code{expr} each name in \code{rasters} refers to that
raster's strip as a numeric vector, so ordinary vectorised R expresses the
calculation: a band index \code{(nir - red) / (nir + red)}, a reclassification
\code{cut(dem, breaks)}, a threshold \code{ifelse(slope > 30, 1L, 0L)}, or arithmetic
across layers. The result is written one strip at a time to a single-band
output.
}
\details{
This is the \emph{monoid fold} tier of the spatial toolbox: bounded to one strip
per input, a single streaming pass, no spill. The rasters must share
dimensions and geotransform; \code{\link[=warp]{warp()}} them onto a common grid first if they
do not. No \pkg{sf} is needed.
}
\examples{
nir <- matrix(c(40, 50, 60, 70), 2, 2)
red <- matrix(c(10, 20, 30, 40), 2, 2)
fn <- tempfile(fileext = ".vec"); fr <- tempfile(fileext = ".vec")
vec_write_raster(nir, fn, dtype = "f64", extent = c(0, 0, 2, 2))
vec_write_raster(red, fr, dtype = "f64", extent = c(0, 0, 2, 2))

ndvi <- rast_calc(list(nir = fn, red = fr), (nir - red) / (nir + red))
round(ndvi, 3)
unlink(c(fn, fr))

}
\seealso{
\code{\link[=warp]{warp()}} to align rasters onto a shared grid first, \code{\link[=focal]{focal()}} for
neighbourhood rather than cellwise calculation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{tiff_extract_points}
\alias{tiff_extract_points}
\title{Extract raster values at point coordinates}
\usage{
tiff_extract_points(path, x, y = NULL)
}
\arguments{
\item{path}{Path to a GeoTIFF file.}

\item{x}{Numeric vector of x coordinates, or a data.frame / matrix with
columns named \code{x} and \code{y}.}

\item{y}{Numeric vector of y coordinates (ignored if \code{x} is a data.frame).}
}
\value{
A data.frame with columns \code{x}, \code{y}, \code{band1}, \code{band2}, etc.
One row per input point, in the same order as the input.
}
\description{
Samples band values from a GeoTIFF at specific (x, y) locations using the
file's affine geotransform. Only the strips containing query points are read,
making this efficient for sparse point sets on large rasters.
}
\details{
Points that fall outside the raster extent return \code{NA} for all bands.
Pixel assignment uses nearest-pixel rounding (i.e., the point is assigned to
the pixel whose center is closest).
}
\examples{
\donttest{
f <- tempfile(fileext = ".tif")
df <- data.frame(x = as.double(rep(1:4, 3)),
                 y = as.double(rep(1:3, each = 4)),
                 band1 = as.double(1:12))
write_tiff(df, f)

# Sample at specific locations via data.frame
pts <- data.frame(x = c(2, 3), y = c(1, 2))
tiff_extract_points(f, pts)

# Or pass x and y separately
tiff_extract_points(f, x = c(2, 3), y = c(1, 2))
unlink(f)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom_expr.R
\name{geom_expressions}
\alias{geom_expressions}
\title{Geometry functions inside mutate(), filter(), and summarise()}
\description{
vectra recognizes a family of \verb{st_*} geometry functions directly inside
expression verbs. They run on the GEOS C library straight off the hex-WKB
geometry column, one row at a time, with no per-batch round-trip through
\pkg{sf}: \code{tbl(f) |> filter(st_area(geometry) > 1e6)} prunes the stream in C,
and \code{mutate(centroid = st_centroid(geometry))} adds a new hex-WKB geometry
column. The computation is planar, in the geometry's own coordinate units,
exactly as the streaming spatial verbs are.
}
\details{
These names are interpreted by the expression engine; they are not exported R
functions and are not called as such. They are available only inside
\code{\link[=mutate]{mutate()}}, \code{\link[=transmute]{transmute()}}, \code{\link[=filter]{filter()}}, and a grouped \code{\link[=summarise]{summarise()}} over a
\code{vectra_node}. The geometry argument is the hex-WKB column (named \code{geometry}
by convention). A function that returns a geometry produces another hex-WKB
column; materialize it with \code{\link[=collect_sf]{collect_sf()}} (point it at the column with
\verb{geom =}), or write it with \code{\link[=write_tiff]{write_tiff()}}/\code{sf::st_write()}.
}
\section{Measures (return a number)}{

\describe{
\item{\code{st_area(g)}}{Area of polygonal geometry (0 for lines and points).}
\item{\code{st_length(g)}, \code{st_perimeter(g)}}{Length of a line, or the perimeter
(boundary length) of a polygon. The two names are aliases.}
\item{\code{st_x(g)}, \code{st_y(g)}}{Coordinate of a point geometry; \code{NA} for a
non-point.}
\item{\code{st_npoints(g)}}{Number of coordinates in the geometry.}
\item{\code{st_ngeometries(g)}}{Number of sub-geometries in a collection or
multi-geometry (1 for a single geometry).}
\item{\code{st_distance(a, b)}}{Shortest planar distance between two geometries
(see the binary second argument below).}
}
}

\section{Predicates (return TRUE / FALSE)}{

Unary: \code{st_is_valid(g)}, \code{st_is_empty(g)}, \code{st_is_simple(g)}.

Binary topological relations, each taking a second geometry: \code{st_intersects},
\code{st_within}, \code{st_contains}, \code{st_overlaps}, \code{st_touches}, \code{st_crosses},
\code{st_equals}, \code{st_disjoint}, \code{st_covers}, \code{st_covered_by}. Used in \code{\link[=filter]{filter()}}
they keep the rows where the relation holds.
}

\section{Type (returns a string)}{

\code{st_geometry_type(g)} gives the GEOS geometry type name (\code{"Point"},
\code{"Polygon"}, \code{"MultiPolygon"}, ...).
}

\section{Transforms (return a geometry)}{

\describe{
\item{\code{st_centroid(g)}}{Area (or length, or vertex) centroid.}
\item{\code{st_point_on_surface(g)}}{A point guaranteed to lie on the geometry.}
\item{\code{st_boundary(g)}}{The topological boundary.}
\item{\code{st_envelope(g)}}{The axis-aligned bounding rectangle.}
\item{\code{st_convex_hull(g)}}{The convex hull.}
\item{\code{st_make_valid(g)}}{Repair an invalid geometry.}
\item{\code{st_buffer(g, dist)}}{Buffer by \code{dist} (round joins, 8 segments per
quadrant).}
\item{\code{st_simplify(g, tol)}}{Topology-preserving Douglas-Peucker
simplification with tolerance \code{tol}.}
}
}

\section{The second geometry of a binary op}{

For \code{st_distance} and the binary predicates, the second argument can be
another geometry column (compared row by row), a constant \code{sf}/\code{sfc} object
(a multi-feature object is unioned to one geometry), or a hex-WKB string. A
constant is parsed once and reused across every row, so testing a whole
stream against one area of interest stays cheap.
}

\section{Missing geometry}{

A missing (\code{NA}) or unparseable geometry, or an operation that has no answer
(a coordinate of a non-point, distance to a missing geometry), yields \code{NA}
for that row rather than an error.
}

\examples{
if (requireNamespace("sf", quietly = TRUE)) {
  nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
  f <- tempfile(fileext = ".vtr")
  write_vtr(data.frame(
    NAME     = nc$NAME,
    geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
  ), f)

  # measures: add area and perimeter columns
  tbl(f) |>
    mutate(area = st_area(geometry), perim = st_perimeter(geometry)) |>
    select(NAME, area, perim) |>
    collect() |>
    head()

  # predicate: keep counties intersecting an area of interest
  aoi <- sf::st_as_sfc(sf::st_bbox(
    c(xmin = -81.5, ymin = 36.2, xmax = -80.5, ymax = 36.6)))
  tbl(f) |> filter(st_intersects(geometry, aoi)) |> collect() |> nrow()

  # transform: replace each county with its centroid, materialize as sf
  tbl(f) |>
    mutate(geometry = st_centroid(geometry)) |>
    select(NAME, geometry) |>
    collect_sf()

  unlink(f)
}
}
\seealso{
\code{\link[=mutate]{mutate()}}, \code{\link[=filter]{filter()}}, \code{\link[=collect_sf]{collect_sf()}} to materialize a geometry
result as \code{sf}; \code{\link[=spatial_map]{spatial_map()}} for an arbitrary per-feature \pkg{sf}
transform; \code{\link[=spatial_filter]{spatial_filter()}} and \code{\link[=spatial_join]{spatial_join()}} for relating a stream to
a resident reference layer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_grouping.R
\name{count}
\alias{count}
\alias{tally}
\title{Count observations by group}
\usage{
count(x, ..., wt = NULL, sort = FALSE, name = NULL)

tally(x, wt = NULL, sort = FALSE, name = NULL)
}
\arguments{
\item{x}{A \code{vectra_node} object.}

\item{...}{Grouping columns (unquoted).}

\item{wt}{Column to weight by (unquoted). If \code{NULL}, counts rows.}

\item{sort}{If \code{TRUE}, sort output in descending order of \code{n}.}

\item{name}{Name of the count column (default \code{"n"}).}
}
\value{
A \code{vectra_node} with group columns and a count column.
}
\description{
Count observations by group
}
\details{
Equivalent to \code{group_by(...) |> summarise(n = n())}. When \code{wt} is
provided, uses \code{sum(wt)} instead of \code{n()}. When \code{sort = TRUE}, results
are sorted in descending order of the count column.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> count(cyl) |> collect()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_utility.R
\name{pull}
\alias{pull}
\title{Extract a single column as a vector}
\usage{
pull(.data, var = -1)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{var}{Column name (unquoted) or positive integer position.}
}
\value{
A vector.
}
\description{
Extract a single column as a vector
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> pull(mpg) |> head()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_knn}
\alias{spatial_knn}
\title{k nearest neighbours of a streamed layer, with distances}
\usage{
spatial_knn(
  x,
  y,
  k = 1L,
  geom = "geometry",
  coords = NULL,
  crs = NA,
  y_id = NULL,
  id_col = "neighbor",
  dist_col = "distance",
  rank_col = "rank",
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{y}{An \code{sf} or \code{sfc} object: the resident candidate-neighbour layer.}

\item{k}{Number of nearest neighbours to return per left feature (capped at
the number of \code{y} features). Default \code{1}.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{coords}{Optional length-2 character vector naming the x and y
coordinate columns to assemble point geometry from (e.g. \code{c("x", "y")}),
for inputs such as \code{\link[=tiff_extract_points]{tiff_extract_points()}} output. The coordinate columns
are retained.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{y_id}{Optional name of a column in \code{y} whose value identifies each
neighbour in the output. Default \code{NULL} uses \code{y}'s 1-based row index.}

\item{id_col, dist_col, rank_col}{Names of the output columns holding the
neighbour identifier, the distance, and the 1-based rank (1 = nearest).
Defaults \code{"neighbor"}, \code{"distance"}, \code{"rank"}.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of one row per (left, neighbour) pair -- \code{x}'s columns
(geometry included) plus the rank, neighbour identifier, and distance --
backed by temporary \code{.vtr} spills (removed when the node is garbage-
collected) and carrying the input CRS.
}
\description{
Streams a large left side \code{x} through the engine and, for each feature, finds
the \code{k} nearest features of a small resident layer \code{y}, returning one row per
(left, neighbour) pair with the neighbour's rank, identifier, and distance.
Where \code{\link[=spatial_join]{spatial_join()}} with \link[sf:st_nearest_feature]{sf::st_nearest_feature} attaches only the single
nearest match, this returns the top \code{k} and the distances themselves -- the
nearest-\code{k} query and the building block of a distance matrix. The billion-row
left stream never materializes; \code{y} (the candidate neighbours) stays resident.
}
\details{
Distances are \pkg{sf}'s \code{\link[sf:st_distance]{sf::st_distance()}}: planar (CRS units) on projected
or unprojected planar data, great-circle (metres) on geographic coordinates
with spherical geometry on (\code{sf::sf_use_s2()}). Each batch forms its
left-by-\code{y} distance matrix, so \code{y} should be the small side; when \code{y} carries
no CRS it inherits the stream's. The left geometry rides through unchanged
(replicated once per neighbour). The \pkg{sf} package is an optional
dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
towns <- sf::st_centroid(sf::st_geometry(nc))[1:5]
towns <- sf::st_sf(town = nc$NAME[1:5], geometry = towns)

set.seed(1)
pts <- sf::st_coordinates(sf::st_sample(nc, 100))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), f)

# The two nearest towns to each point, with distances.
tbl(f) |>
  spatial_knn(towns, k = 2, coords = c("x", "y"), crs = sf::st_crs(nc),
              y_id = "town") |>
  collect() |> head()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_join]{spatial_join()}} for a nearest-feature attribute join, \code{\link[=collect_sf]{collect_sf()}}
to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_utility.R
\name{relocate}
\alias{relocate}
\title{Relocate columns}
\usage{
relocate(.data, ..., .before = NULL, .after = NULL)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Column names to move.}

\item{.before}{Column name to place before (unquoted).}

\item{.after}{Column name to place after (unquoted).}
}
\value{
A new \code{vectra_node} with reordered columns.
}
\description{
Relocate columns
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> relocate(hp, wt, .before = cyl) |> collect() |> head()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/joins.R
\name{left_join}
\alias{left_join}
\alias{inner_join}
\alias{right_join}
\alias{full_join}
\alias{semi_join}
\alias{anti_join}
\title{Join two vectra tables}
\usage{
left_join(x, y, by = NULL, suffix = c(".x", ".y"), ...)

inner_join(x, y, by = NULL, suffix = c(".x", ".y"), ...)

right_join(x, y, by = NULL, suffix = c(".x", ".y"), ...)

full_join(x, y, by = NULL, suffix = c(".x", ".y"), ...)

semi_join(x, y, by = NULL, ...)

anti_join(x, y, by = NULL, ...)
}
\arguments{
\item{x}{A \code{vectra_node} object (left table).}

\item{y}{A \code{vectra_node} object (right table).}

\item{by}{A character vector of column names to join by, or a named vector
like \code{c("a" = "b")}. \code{NULL} for natural join (common columns).}

\item{suffix}{A character vector of length 2 for disambiguating non-key
columns with the same name (default \code{c(".x", ".y")}).}

\item{...}{Ignored.}
}
\value{
A \code{vectra_node} with the joined result.
}
\description{
Join two vectra tables
}
\details{
All joins use a build-right, probe-left hash join. The entire right-side
table is materialized into a hash table; left-side batches stream through.
Memory cost is proportional to the right-side table size.

NA keys never match (SQL NULL semantics). Key types are auto-coerced
following the \verb{bool < int64 < double} hierarchy. Joining string against
numeric keys is an error.
}
\examples{
f1 <- tempfile(fileext = ".vtr")
f2 <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = c(1, 2, 3), x = c(10, 20, 30)), f1)
write_vtr(data.frame(id = c(1, 2, 4), y = c(100, 200, 400)), f2)
left_join(tbl(f1), tbl(f2), by = "id") |> collect()
unlink(c(f1, f2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/offload.R
\name{group_map}
\alias{group_map}
\alias{group_map.vectra_partition}
\alias{group_modify}
\alias{group_modify.vectra_partition}
\title{Apply a function to each shard of a partition}
\usage{
group_map(.data, .f, ...)

\method{group_map}{vectra_partition}(.data, .f, ...)

group_modify(.data, .f, ...)

\method{group_modify}{vectra_partition}(.data, .f, ...)
}
\arguments{
\item{.data}{A \code{vectra_partition} from \code{\link[=offload]{offload()}} with a \code{by} key.}

\item{.f}{A function applied to each shard. It receives the shard as a
data.frame and the shard key (a string) as its first two arguments; any
further arguments in \code{...} follow. A purrr-style formula such as
\code{~ lm(y ~ x, .x)} also works, with \code{.x} the shard data and \code{.y} the key.
For \code{group_modify()}, \code{.f} must return a data.frame.}

\item{...}{Additional arguments passed on to \code{.f}.}
}
\value{
\code{group_map()} returns a named list with one element per shard.
\code{group_modify()} returns a single data.frame: the per-shard results
row-bound, with the shard key restored as a column when \code{.f} dropped it.
}
\description{
Run a function once per shard of a partition (\code{offload(x, by = ...)}) and
gather the results. Each shard is read into memory as a data.frame and passed
to \code{.f} together with its key, so a model that couples rows within a group
becomes a set of independent per-shard fits. This is the per-group
counterpart to \code{\link[=collect_chunked]{collect_chunked()}}, which instead merges every shard into a
single accumulator.
}
\details{
\code{group_map()} returns a named list, one element per shard keyed by the shard
key, and places no constraint on what \code{.f} returns. Use it for per-group
results that do not rebind into a table, such as fitted models.

\code{group_modify()} expects \code{.f} to return a data.frame for each shard and binds
those frames into one. When a shard's result does not already carry the
partition key column, the key is added as a leading column (named after the
partition's \code{by}), so every row records the shard it came from. Use it for
per-group summaries that recombine into a single table.

Each shard is materialized in full before \code{.f} sees it, so partition the
query on a key whose groups fit in memory. For a reduction that stays bounded
without ever holding a whole group, fold the partition with
\code{\link[=collect_chunked]{collect_chunked()}} instead.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
p <- offload(tbl(f), by = "cyl")

# One fit per shard, returned as a named list keyed by cyl.
fits <- group_map(p, function(d, cyl) coef(lm(mpg ~ wt, data = d)))
fits

# Per-shard summaries recombined into one table, key restored as a column.
group_modify(p, function(d, cyl)
  data.frame(n = nrow(d), mean_mpg = mean(d$mpg)))
unlink(f)

}
\seealso{
\code{\link[=offload]{offload()}} to build a partition, and \code{\link[=collect_chunked]{collect_chunked()}} for the
partitioned monoidal reduce.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block.R
\name{materialize}
\alias{materialize}
\title{Materialize a vectra node into a reusable in-memory block}
\usage{
materialize(.data)
}
\arguments{
\item{.data}{A \code{vectra_node} (consumed; cannot be used after this call).}
}
\value{
A \code{vectra_block} object (external pointer to C-level ColumnBlock).
}
\description{
Consumes a vectra node (pulling all batches) and stores the result as a
persistent columnar block in memory.  Unlike nodes, blocks can be probed
repeatedly via \code{\link[=block_lookup]{block_lookup()}} without re-scanning.
}
\examples{
\donttest{
f <- tempfile(fileext = ".vtr")
df <- data.frame(taxonID = 1:3,
                 canonicalName = c("Quercus robur", "Pinus sylvestris",
                                   "Fagus sylvatica"))
write_vtr(df, f)
blk <- materialize(tbl(f) |> select(taxonID, canonicalName))
hits <- block_lookup(blk, "canonicalName",
                     c("Quercus robur", "Pinus sylvestris"))
unlink(f)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index.R
\name{has_index}
\alias{has_index}
\title{Check if a hash index exists for a .vtr column}
\usage{
has_index(path, column)
}
\arguments{
\item{path}{Path to a \code{.vtr} file.}

\item{column}{Character vector. Name(s) of column(s).}
}
\value{
Logical scalar: \code{TRUE} if a \code{.vtri} index file exists.
}
\description{
Check if a hash index exists for a .vtr column
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = letters, val = 1:26, stringsAsFactors = FALSE), f)
has_index(f, "id")   # FALSE
create_index(f, "id")
has_index(f, "id")   # TRUE
unlink(c(f, paste0(f, ".id.vtri")))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{rasterize}
\alias{rasterize}
\title{Rasterize a streamed point layer onto a fixed grid}
\usage{
rasterize(
  x,
  template = NULL,
  field = NULL,
  fun = c("count", "sum", "mean", "min", "max"),
  extent = NULL,
  res = NULL,
  dims = NULL,
  coords = c("x", "y"),
  geom = NULL,
  crs = NA,
  background = NA_real_,
  path = NULL,
  dtype = "f32"
)
}
\arguments{
\item{x}{A \code{vectra_node} streaming the points (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_csv]{tbl_csv()}}, any
verb chain). It is consumed by the stream.}

\item{template}{Optional grid to borrow geometry and CRS from: a
\code{vectra_raster} (from \code{\link[=vec_open_raster]{vec_open_raster()}}), or a numeric
\code{c(xmin, ymin, xmax, ymax)} extent. When omitted, supply \code{extent} with
\code{res} or \code{dims}.}

\item{field}{Name of a numeric column to aggregate. Required for every \code{fun}
except \code{"count"} (which ignores it).}

\item{fun}{Reduction over the points in each cell: one of \code{"count"}, \code{"sum"},
\code{"mean"}, \code{"min"}, \code{"max"}. \code{NA} values in \code{field} are skipped.}

\item{extent}{Numeric \code{c(xmin, ymin, xmax, ymax)} defining the grid extent
when no \code{template} is given.}

\item{res}{Cell size: a single number for square cells, or \code{c(xres, yres)}.
The cell counts are rounded to fit \code{extent} exactly. Supply \code{res} or \code{dims}.}

\item{dims}{Grid shape \code{c(nrow, ncol)}, an alternative to \code{res}.}

\item{coords}{Length-2 character vector naming the x and y coordinate
columns. Default \code{c("x", "y")}. Ignored when \code{geom} is supplied.}

\item{geom}{Name of a hex-WKB point-geometry column to rasterize instead of
coordinate columns. Requires \pkg{sf}.}

\item{crs}{Coordinate reference system recorded on the output, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts or a bare EPSG integer. Defaults to the template's,
then the node's, else unknown.}

\item{background}{Value for cells that receive no point. Default \code{NA_real_}.}

\item{path}{Optional output path. When given, the grid is written to a \code{.vec}
raster via \code{\link[=vec_write_raster]{vec_write_raster()}} and the opened \code{\link[=vec_open_raster]{vec_open_raster()}} handle is
returned invisibly. When \code{NULL}, the grid is returned in memory.}

\item{dtype}{Storage dtype for the \code{.vec} output (see \code{\link[=vec_write_raster]{vec_write_raster()}}).
Default \code{"f32"}.}
}
\value{
When \code{path} is \code{NULL}, a numeric matrix with \code{nrow} grid rows
(row 1 northmost) and \code{ncol} grid columns, carrying \code{gt}, \code{extent}, \code{res},
\code{crs}, and \code{fun} attributes. When \code{path} is given, the written
\code{vectra_raster} handle (invisibly).
}
\description{
Folds a larger-than-RAM stream of points into a fixed raster grid one batch
at a time. The grid (\code{template}) is held resident in memory while the points
flow past the engine, so peak memory is the grid plus one batch regardless of
how many points there are -- the streaming counterpart to running
\code{terra::rasterize()} on a point set that has to fit in RAM. Each point's
coordinate is mapped to its grid cell through the raster geotransform and the
per-cell value is accumulated in C.
}
\details{
The reduction \code{fun} is a monoid over the points falling in each cell:
\code{"count"} tallies points (no \code{field} needed); \code{"sum"}, \code{"mean"}, \code{"min"},
\code{"max"} aggregate a numeric \code{field}. Cells that receive no point take the
\code{background} value (\code{NA} by default). This is the \emph{monoid fold} tier of the
spatial toolbox: bounded memory, a single streaming pass, no spill.

Points arrive either as two numeric coordinate columns (\code{coords}, the default
and fully \pkg{sf}-free path -- the headline larger-than-RAM case) or decoded
from a hex-WKB point-geometry column (\code{geom}, which needs \pkg{sf}). Geometry
input is expected to be points (one coordinate per row); line and polygon
coverage rasterization is out of scope here.
}
\examples{
set.seed(1)
n <- 1e4
pts <- data.frame(x = runif(n, 0, 10), y = runif(n, 0, 10), z = rnorm(n))
f <- tempfile(fileext = ".vtr")
write_vtr(pts, f)

# Point density on a 10x10 grid, streamed: the grid is resident, the
# points are not.
counts <- tbl(f) |> rasterize(extent = c(0, 0, 10, 10), dims = c(10, 10))
counts

# Mean of z per cell.
zmean <- tbl(f) |>
  rasterize(extent = c(0, 0, 10, 10), dims = c(10, 10),
            field = "z", fun = "mean")
unlink(f)

}
\seealso{
\code{\link[=vec_write_raster]{vec_write_raster()}} and \code{\link[=vec_to_tiff]{vec_to_tiff()}} for raster output,
\code{\link[=spatial_join]{spatial_join()}} to instead tag points with polygon attributes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_open_raster}
\alias{vec_open_raster}
\title{Open a .vec raster}
\usage{
vec_open_raster(path)
}
\arguments{
\item{path}{Path to a \code{.vec} raster file.}
}
\value{
A \code{vectra_raster} list with elements:
\code{ptr}, \code{width}, \code{height}, \code{n_bands}, \code{tile_size}, \code{dtype}, \code{gt},
\code{epsg}, \code{nodata}, \code{band_names}.
}
\description{
Lazy open: parses the header and tile index but does not decode any
tiles. Returns a list with metadata and an external pointer handle.
The pointer is auto-finalized when garbage collected; call
\code{vec_close_raster()} to release earlier.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/network.R
\name{spatial_service_area}
\alias{spatial_service_area}
\title{Service areas and isochrones over a network}
\usage{
spatial_service_area(
  x,
  network,
  cost,
  output = c("polygon", "lines", "nodes"),
  band_col = "band",
  geom = "geometry",
  coords = NULL,
  crs = NA,
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} of origin features (point geometry, or \code{coords}).}

\item{network}{A \code{vectra_network} from \code{\link[=spatial_network]{spatial_network()}}.}

\item{cost}{A cost budget (scalar), or several budgets for nested isochrone
bands (e.g. \code{c(5, 10, 15)}).}

\item{output}{\code{"polygon"} (default) for the convex hull of the reachable
nodes, \code{"lines"} for the reachable edges, or \code{"nodes"} for the reachable
nodes as a multipoint.}

\item{band_col}{Name of the output column holding each row's budget. Default
\code{"band"}.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{coords}{Optional length-2 character vector naming the x and y
coordinate columns to assemble point geometry from (e.g. \code{c("x", "y")}),
for inputs such as \code{\link[=tiff_extract_points]{tiff_extract_points()}} output. The coordinate columns
are retained.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node}: one row per (origin, band) carrying \code{x}'s attributes,
the band value, and the service-area geometry. Backed by temporary \code{.vtr}
spills removed when the node is garbage-collected, and carrying the network
CRS.
}
\description{
Streams a layer of origins \code{x} past a resident \code{\link[=spatial_network]{spatial_network()}} and, for
each origin, finds every part of the network reachable within a cost budget -
the service area, or, with several budgets, nested travel-cost isochrone bands.
Each origin is snapped to its nearest graph node and a budget-bounded Dijkstra
(native C) collects the reachable nodes; the reachable set is returned as the
convex hull (\code{output = "polygon"}, the generalised service area), the
reachable edges (\code{"lines"}), or the reachable nodes (\code{"nodes"}).
}
\details{
A vector \code{cost} returns one row per (origin, band), each band the area
reachable within that budget, so the rows nest from the smallest budget out.
Costs are in the network's \code{weight} units. The convex-hull polygon is a
generalisation; use \code{output = "lines"} for the exact reachable network. The
\pkg{sf} package is an optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mk <- function(x1, y1, x2, y2)
  sf::st_linestring(rbind(c(x1, y1), c(x2, y2)))
streets <- sf::st_sfc(
  mk(0, 0, 1, 0), mk(1, 0, 2, 0), mk(0, 0, 0, 1),
  mk(0, 1, 1, 1), mk(1, 0, 1, 1), mk(1, 1, 2, 1), mk(2, 0, 2, 1))
net <- spatial_network(streets)

f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = 1L, x = 0, y = 0), f)

tbl(f) |>
  spatial_service_area(net, cost = c(1, 2), output = "lines",
                       coords = c("x", "y")) |>
  collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_network]{spatial_network()}} to build the graph, \code{\link[=spatial_route]{spatial_route()}} for
shortest paths, \code{\link[=collect_sf]{collect_sf()}} to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block.R
\name{block_lookup}
\alias{block_lookup}
\title{Probe a materialized block by column value}
\usage{
block_lookup(block, column, keys, ci = FALSE)
}
\arguments{
\item{block}{A \code{vectra_block} from \code{\link[=materialize]{materialize()}}.}

\item{column}{Character scalar. Name of the string column to match against.}

\item{keys}{Character vector. Query values to look up.}

\item{ci}{Logical. Case-insensitive matching (default \code{FALSE}).}
}
\value{
A data.frame with column \code{query_idx} (1-based position in \code{keys})
plus all columns from the block, for each (query, block_row) match pair.
}
\description{
Performs a hash lookup on a string column of a materialized block.
Returns all rows where the column value matches one of the query keys.
Hash indices are built lazily on first use and cached for subsequent calls.
}
\examples{
\donttest{
f <- tempfile(fileext = ".vtr")
df <- data.frame(taxonID = 1:2,
                 canonicalName = c("Quercus robur", "Pinus sylvestris"))
write_vtr(df, f)
blk <- materialize(tbl(f))
hits <- block_lookup(blk, "canonicalName", c("Quercus robur"))
ci_hits <- block_lookup(blk, "canonicalName", c("quercus robur"), ci = TRUE)
unlink(f)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_utility.R
\name{transmute}
\alias{transmute}
\title{Keep only columns from mutate expressions}
\usage{
transmute(.data, ...)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Named expressions.}
}
\value{
A new \code{vectra_node} with only the computed columns.
}
\description{
Like \code{\link[=mutate]{mutate()}} but drops all other columns.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> transmute(kpl = mpg * 0.425) |> collect() |> head()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_basic.R
\name{select}
\alias{select}
\title{Select columns from a vectra query}
\usage{
select(.data, ...)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Column names (unquoted).}
}
\value{
A new \code{vectra_node} with only the selected columns.
}
\description{
Select columns from a vectra query
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> select(mpg, cyl) |> collect() |> head()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_read_pixel_series}
\alias{vec_read_pixel_series}
\title{Read the full time series at a single pixel from a .vec time cube}
\usage{
vec_read_pixel_series(
  r,
  x = NULL,
  y = NULL,
  col = NULL,
  row = NULL,
  band = 1L,
  level = 0L
)
}
\arguments{
\item{r}{A \code{vectra_raster} from \code{vec_open_raster()}.}

\item{x, y}{Pixel coordinates. Either both \code{x} and \code{y} (CRS units; the
geotransform is used to map to col/row) or both \code{col} and \code{row}
(1-based pixel indices).}

\item{col, row}{1-based pixel coordinates (alternative to x/y).}

\item{band}{Band index (1-based).}

\item{level}{Overview level. Default 0.}
}
\value{
A numeric vector of length \code{n_time}. NA marks pixels outside
the raster or matching nodata. The corresponding time stamps can
be obtained from \code{vec_raster_times(r, band, level)}.
}
\description{
Returns a numeric vector of length \code{n_time} — one value per time step
recorded in the file, in ascending time-stamp order.
}
\details{
For pixel-major files (written with
\code{vec_write_time_cube(layout = "pixel")}) this is the optimal access
pattern: a single tile decode yields all time values for the pixel.
For image-major files the reader scans the index for distinct time
stamps, decodes one spatial tile per stamp, and extracts the pixel
from each — correct but \code{n_time} slower than the optimal layout.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{tiff_band_names}
\alias{tiff_band_names}
\title{Read per-band names from a GeoTIFF}
\usage{
tiff_band_names(path)
}
\arguments{
\item{path}{Path to a GeoTIFF file.}
}
\value{
A character vector of length \code{nbands}. Element \code{i} is the name
of band \code{i} (or \code{NA_character_} if the file does not name it).
}
\description{
Returns the band names embedded in the file's GDAL_METADATA XML
(TIFF tag 42112). GDAL writes per-band names as
\verb{<Item name="DESCRIPTION" sample="N" role="description">...</Item>} entries,
where \code{sample} is the 0-based band index. Bands without a name in the XML
are reported as \code{NA}. Files with no GDAL_METADATA tag at all return a
length-\code{nbands} vector of \code{NA_character_}.
}
\details{
This is a small, dependency-free scanner intended for the common case
(\code{terra::names(r) <- ...} and similar). For arbitrary XML, parse the raw
string from \code{\link[=tiff_metadata]{tiff_metadata()}} yourself.
}
\examples{
\donttest{
f <- tempfile(fileext = ".tif")
df <- data.frame(x = rep(1:2, 2), y = rep(1:2, each = 2),
                 band1 = as.double(1:4), band2 = as.double(5:8))
xml <- paste0(
  "<GDALMetadata>",
  "<Item name=\"DESCRIPTION\" sample=\"0\" role=\"description\">temperature</Item>",
  "<Item name=\"DESCRIPTION\" sample=\"1\" role=\"description\">humidity</Item>",
  "</GDALMetadata>")
write_tiff(df, f, metadata = xml)
tiff_band_names(f)
unlink(f)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_topology.R
\name{spatial_simplify}
\alias{spatial_simplify}
\title{Simplify a polygon coverage without tearing shared edges}
\usage{
spatial_simplify(
  x,
  tolerance,
  by = NULL,
  geom = "geometry",
  crs = NA,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{tolerance}{Distance tolerance for the boundary simplification, in CRS
units: vertices that deviate less than this from the simplified line are
dropped. Larger values simplify more.}

\item{by}{Character vector of attribute columns whose groups are each
simplified as an independent coverage. \code{NULL} (default) treats the whole
layer as one coverage.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of the simplified polygons, each carrying its source
feature's attributes and the input CRS, backed by temporary \code{.vtr} spills
removed when the node is garbage-collected.
}
\description{
Simplifies polygon boundaries while keeping a shared border between two
polygons identical on both sides, so adjacent polygons stay edge-matched with
no slivers or gaps -- the topology-preserving simplification that a per-feature
\code{spatial_map(~ sf::st_simplify(.x))} cannot give, because it simplifies each
polygon's copy of a shared border independently. The boundaries of each group
are unioned so a shared border is one line, noded into arcs at every junction,
each arc simplified once (its junction endpoints pinned), and the arcs
re-polygonized; each resulting face inherits the attributes of the source
polygon containing it. Like \code{\link[=spatial_dissolve]{spatial_dissolve()}} it rides the \strong{partition
tier}: \code{x} is spilled once and routed into one disjoint shard per \code{by} group
in a single bounded pass, and each group is simplified as an independent
coverage. Peak memory is the routing budget during the pass, then one group's
geometry while it is simplified -- partition on a key whose groups fit in
memory. With no \code{by}, the whole layer is one coverage.
}
\details{
The simplification is topology-preserving Douglas-Peucker (\code{dTolerance = tolerance}) on the noded boundary arcs. Geometry travels through the engine as
hex-encoded WKB in a string column and the CRS is carried on the returned
node; the noding is \pkg{sf}/GEOS and expects projected or unprojected planar
data. The \pkg{sf} package is an optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
p1 <- sf::st_polygon(list(rbind(
  c(0, 0), c(1, 0), c(1, 0.5), c(1, 1), c(0, 1), c(0, 0))))
p2 <- sf::st_polygon(list(rbind(
  c(1, 0), c(2, 0), c(2, 1), c(1, 1), c(1, 0.5), c(1, 0))))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = c("a", "b"),
  geometry = sf::st_as_binary(sf::st_sfc(p1, p2), hex = TRUE)
), f)

# The shared edge is simplified once, so the two polygons stay edge-matched.
tbl(f) |> spatial_simplify(tolerance = 0.6) |> collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_map]{spatial_map()}} with \code{~ sf::st_simplify(.x)} for independent
per-feature simplification, \code{\link[=spatial_smooth]{spatial_smooth()}} for Chaikin corner-rounding,
\code{\link[=collect_sf]{collect_sf()}} to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_dissolve}
\alias{spatial_dissolve}
\title{Dissolve geometries by group}
\usage{
spatial_dissolve(
  x,
  by = NULL,
  ...,
  geom = "geometry",
  crs = NA,
  .fun = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{by}{Character vector of attribute columns to dissolve within: one
output feature per distinct combination of their values. \code{NULL} (default)
dissolves the entire layer into a single feature.}

\item{...}{Further arguments passed to \code{\link[sf:st_union]{sf::st_union()}} (e.g.
\code{is_coverage = TRUE}).}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{.fun}{Optional named list of attribute summaries. Each element is a
function taking the group's data.frame and returning a length-1 value; the
list name becomes the output column (e.g.
\code{.fun = list(total = function(d) sum(d$pop))}). Default \code{NULL} keeps only
the \code{by} columns and the dissolved geometry.}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of one row per group -- the \code{by} columns, any \code{.fun}
summaries, and the dissolved geometry -- backed by temporary \code{.vtr} spills
removed when the node is garbage-collected, carrying the input CRS for
\code{\link[=collect_sf]{collect_sf()}}.
}
\description{
Unions the geometries within each \code{by} group into a single feature (the GIS
"Dissolve" tool), optionally summarising attributes. Unlike the streamed
per-batch verbs, dissolve needs every geometry of a group together to union
them, so it rides the \strong{partition tier}: \code{x} is spilled once and routed into
one disjoint shard per group in a single bounded pass, then each shard is read
in and unioned with \pkg{sf}. Peak memory is the routing budget during the
pass, then one group's geometries while it is unioned -- partition the input
on a key whose groups fit in memory. With no \code{by}, the whole layer dissolves
into one feature.
}
\details{
Geometry travels through the engine as hex-encoded WKB in a string column and
the CRS is carried on the returned node; use \code{\link[=collect_sf]{collect_sf()}} to materialize.
On projected or unprojected planar data each group is unioned natively on the
GEOS C API straight off the hex-WKB column; geographic coordinates with
spherical geometry on (\code{sf::sf_use_s2()}), or any extra \code{\link[sf:st_union]{sf::st_union()}}
arguments (e.g. \code{is_coverage = TRUE}), union through \pkg{sf} instead. The
\pkg{sf} package is an optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
nc$band <- nc$SID74 > 5            # an attribute to dissolve within
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  band = nc$band, BIR74 = nc$BIR74,
  geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
), f)

# Merge the counties into two features by `band`, summing births.
merged <- tbl(f) |>
  spatial_dissolve(by = "band", crs = sf::st_crs(nc),
                   .fun = list(births = function(d) sum(d$BIR74)))
collect_sf(merged)
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_overlay]{spatial_overlay()}} to split overlaps apart rather than merge them,
\code{\link[=offload]{offload()}} for the partition tier this rides on, \code{\link[=collect_sf]{collect_sf()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema.R
\name{lookup}
\alias{lookup}
\title{Look up columns from linked dimension tables}
\usage{
lookup(.schema, ..., .join = "left", .report = TRUE)
}
\arguments{
\item{.schema}{A \code{vectra_schema} object.}

\item{...}{Column references: bare names for fact columns, or
\code{dimension$column} for dimension columns.}

\item{.join}{Join type: \code{"left"} (default, keeps all fact rows) or
\code{"inner"} (drops unmatched fact rows).}

\item{.report}{Logical. If \code{TRUE} (default), print a message with
the number of unmatched keys per dimension.}
}
\value{
A \code{vectra_node} with the selected columns.
}
\description{
Resolves columns from dimension tables registered in a \code{\link[=vtr_schema]{vtr_schema()}},
automatically building the necessary join tree. Reports unmatched keys
as a diagnostic message.
}
\details{
Column references use \code{dimension$column} syntax (e.g., \code{species$name}).
Columns from the fact table can be referenced by name directly.

When \code{.report = TRUE}, each needed dimension is checked for unmatched keys
by opening fresh scans of the fact and dimension tables. This adds one
extra read pass per dimension but does not affect the lazy result node.

Only dimensions referenced in \code{...} are joined. Unreferenced dimensions
are never scanned.
}
\examples{
\donttest{
f_obs <- tempfile(fileext = ".vtr")
f_sp  <- tempfile(fileext = ".vtr")
f_ct  <- tempfile(fileext = ".vtr")
write_vtr(data.frame(sp_id = 1:4, ct_code = c("AT", "DE", "FR", "XX"),
                      value = 10:13), f_obs)
write_vtr(data.frame(sp_id = 1:3,
                      name = c("Oak", "Beech", "Pine")), f_sp)
write_vtr(data.frame(ct_code = c("AT", "DE", "FR"),
                      gdp = c(400, 3800, 2700)), f_ct)

s <- vtr_schema(
  fact    = tbl(f_obs),
  species = link("sp_id", tbl(f_sp)),
  country = link("ct_code", tbl(f_ct))
)

# Pull columns from any linked dimension
result <- lookup(s, value, species$name, country$gdp)
collect(result)

unlink(c(f_obs, f_sp, f_ct))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{diff_vtr}
\alias{diff_vtr}
\title{Compute the logical diff between two .vtr files}
\usage{
diff_vtr(old_path, new_path, key_col)
}
\arguments{
\item{old_path}{Path to the older \code{.vtr} file.}

\item{new_path}{Path to the newer \code{.vtr} file.}

\item{key_col}{Name of the column to use as the row key (must exist in
both files with the same type).}
}
\value{
A named list with elements \code{added} (a \code{vectra_node}) and \code{deleted}
(a vector of key values).
}
\description{
Streams both files and computes a set-level diff keyed on \code{key_col}.
Returns a list with two elements:
}
\details{
\itemize{
\item \code{added}: a \code{vectra_node} (lazy \code{\link[=tbl]{tbl()}}) of rows present in \code{new_path}
but not \code{old_path} (matched on \code{key_col}).  Call \code{\link[=collect]{collect()}} to
materialise.  The underlying temp file is deleted when the node is
garbage-collected \strong{or} when the calling R session ends via
\code{on.exit()}.
\item \code{deleted}: a vector of key values present in \code{old_path} but not
\code{new_path}.
}

This is a \strong{logical diff} (key-based set difference), not a binary file
diff.  Rows with the same key that have changed values are not reported
as modified — use \code{added} and \code{deleted} together to detect updates (a key
that appears in both means a row was replaced).
}
\examples{
f1 <- tempfile(fileext = ".vtr")
f2 <- tempfile(fileext = ".vtr")
df1 <- data.frame(id = 1:5, val = letters[1:5], stringsAsFactors = FALSE)
df2 <- data.frame(id = c(3L, 4L, 5L, 6L, 7L),
                  val = c("C", "d", "e", "f", "g"),
                  stringsAsFactors = FALSE)
write_vtr(df1, f1)
write_vtr(df2, f2)

d <- diff_vtr(f1, f2, "id")
# Rows 1 and 2 deleted; rows 6 and 7 added
stopifnot(all(d$deleted \%in\% c(1, 2)))
stopifnot(all(collect(d$added)$id \%in\% c(6, 7)))

unlink(c(f1, f2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema.R
\name{link}
\alias{link}
\title{Define a link between a fact table and a dimension table}
\usage{
link(key, node)
}
\arguments{
\item{key}{A character vector or named character vector specifying join keys.
Unnamed: same column name in both tables. Named: \code{c("fact_col" = "dim_col")}.}

\item{node}{A \code{vectra_node} object (the dimension table). Must be file-backed
(created via \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_csv]{tbl_csv()}}, or \code{\link[=tbl_sqlite]{tbl_sqlite()}}).}
}
\value{
A \code{vectra_link} object.
}
\description{
Creates a link descriptor that specifies how to join a dimension table
to a fact table via one or more key columns.
}
\examples{
\donttest{
f_obs <- tempfile(fileext = ".vtr")
f_sp  <- tempfile(fileext = ".vtr")
write_vtr(data.frame(sp_id = 1:3, value = c(10, 20, 30)), f_obs)
write_vtr(data.frame(sp_id = 1:3, name = c("A", "B", "C")), f_sp)
lnk <- link("sp_id", tbl(f_sp))
unlink(c(f_obs, f_sp))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_more.R
\name{contours}
\alias{contours}
\title{Extract contour iso-lines from a streamed raster}
\usage{
contours(x, levels, band = 1L, merge = TRUE, crs = NA, flush_rows = NULL)
}
\arguments{
\item{x}{A \code{vectra_raster} (from \code{\link[=vec_open_raster]{vec_open_raster()}}) or a path to a \code{.vec}
raster.}

\item{levels}{Numeric vector of contour levels to trace.}

\item{band}{Band to contour (1-based). Default 1.}

\item{merge}{If \code{TRUE} (default) join each level's segments into continuous
lines with \code{\link[sf:st_line_merge]{sf::st_line_merge()}}; if \code{FALSE} return the raw per-cell
segments.}

\item{crs}{Coordinate reference system recorded on the node. Defaults to the
raster's EPSG, else unknown.}

\item{flush_rows}{Rows buffered before a spill flush. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} with a \code{level} column and a hex-WKB \code{geometry} column,
materialise it with \code{\link[=collect_sf]{collect_sf()}}.
}
\description{
Traces contour lines at one or more levels from a \code{.vec} raster with marching
squares, reading the raster one tile-row strip at a time (each strip expanded
by one row so a cell straddling the strip boundary is traced once). Each strip
contributes line segments, which are accumulated into a lazy \code{vectra_node}
carrying a \code{level} column and hex-WKB geometry. With \code{merge = TRUE} the
segments of each level are joined into continuous lines.
}
\details{
Extraction is the \emph{sort / partition} tier of the spatial toolbox: bounded to
one haloed strip at a time. The optional final merge collects the segment set,
which is small relative to the raster, and joins it per level; this is the
small all-to-all step on the output, not on the grid. Geometry assembly and
the merge are delegated to \pkg{sf} (an optional dependency).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
z <- outer(1:20, 1:20, function(r, c) r + c)
f <- tempfile(fileext = ".vec")
vec_write_raster(z, f, dtype = "f64", extent = c(0, 0, 20, 20))

iso <- contours(f, levels = c(15, 25, 35))
collect_sf(iso)
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=polygonize]{polygonize()}} for area features, \code{\link[=terrain]{terrain()}} for the DEM
derivatives contours often accompany, \code{\link[=collect_sf]{collect_sf()}} to materialise as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{tiff_crs}
\alias{tiff_crs}
\title{Read CRS metadata from a GeoTIFF}
\usage{
tiff_crs(path)
}
\arguments{
\item{path}{Path to a GeoTIFF file.}
}
\value{
A list with elements \code{epsg} (integer or \code{NA_integer_}) and
\code{citation} (character or \code{NA_character_}).
}
\description{
Returns the spatial reference system embedded in a GeoTIFF, parsed from
the GeoKey directory (TIFF tag 34735). The projected CRS EPSG
(\code{PCSTypeGeoKey} 3072) is preferred over the geographic CRS EPSG
(\code{GeographicTypeGeoKey} 2048). Citation strings are read from
\code{GeoAsciiParams} (tag 34737) with priority PCS > GeoTIFF > geographic.
}
\details{
Files written without a GeoKey directory return \code{NA} for both fields.
}
\examples{
\donttest{
f <- tempfile(fileext = ".tif")
df <- data.frame(x = 1:4, y = rep(1:2, each = 2), band1 = as.double(1:4))
write_tiff(df, f)
tiff_crs(f)  # epsg = NA, citation = NA — vectra writer omits GeoKeys
unlink(f)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_construct}
\alias{spatial_construct}
\title{Build a set-wise geometry construction, optionally per group}
\usage{
spatial_construct(
  x,
  kind = .CONSTRUCT_KINDS,
  by = NULL,
  geom = "geometry",
  crs = NA,
  ratio = 0.3,
  allow_holes = FALSE,
  tolerance = 0,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{kind}{The construction to build; one of the values above.}

\item{by}{Character vector of attribute columns to construct within: one
construction (or tessellation) per distinct combination of their values.
\code{NULL} (default) builds a single construction from the whole layer.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{ratio}{For \code{kind = "concave_hull"}, the concaveness in \verb{[0, 1]} (1 is
the convex hull). Default \code{0.3}.}

\item{allow_holes}{For \code{kind = "concave_hull"}, whether the hull may contain
holes. Default \code{FALSE}.}

\item{tolerance}{Distance tolerance for the kinds that take one
(\code{"inscribed_circle"}, \code{"pole"}, \code{"voronoi"}, \code{"delaunay"}). \code{0} (default)
lets the inscribed-circle kinds derive a tolerance from the extent and the
tessellations use the GEOS default.}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of the construction -- one row per group for the
enclosing kinds, one row per cell for the tessellations -- carrying the
\code{by} columns and the input CRS, backed by temporary \code{.vtr} spills removed
when the node is garbage-collected.
}
\description{
Constructs one geometry (or a tessellation) from a whole set of features --
the constructions a per-feature \code{\link[=spatial_map]{spatial_map()}} cannot express because they
need every feature in scope at once. Like \code{\link[=spatial_dissolve]{spatial_dissolve()}} it rides the
\strong{partition tier}: \code{x} is spilled once and routed into one disjoint shard per
\code{by} group in a single bounded pass, then each shard's geometry is combined
and the construction built with \pkg{sf}. With no \code{by}, the whole layer yields
one construction. Peak memory is the routing budget during the pass, then one
group's geometry while it is built -- partition on a key whose groups fit in
memory.
}
\details{
\code{kind} selects the construction:
\describe{
\item{\code{"convex_hull"}}{the convex hull of the set.}
\item{\code{"concave_hull"}}{the concave hull (\code{ratio}, \code{allow_holes}).}
\item{\code{"envelope"}}{the axis-aligned bounding rectangle.}
\item{\code{"oriented_box"}}{the minimum-area rotated bounding rectangle.}
\item{\code{"enclosing_circle"}}{the minimum bounding circle.}
\item{\code{"inscribed_circle"}}{the maximum inscribed circle (largest circle
that fits inside the set's union).}
\item{\code{"pole"}}{the pole of inaccessibility -- the centre of the maximum
inscribed circle, the point inside the shape farthest from its edges.}
\item{\code{"voronoi"}}{the Voronoi tessellation, one polygon per cell.}
\item{\code{"delaunay"}}{the Delaunay triangulation, one polygon per triangle.}
}
The enclosing kinds and \code{pole} emit one feature per group; \code{voronoi} and
\code{delaunay} emit one feature per cell, each carrying the group's \code{by} values.

Geometry travels through the engine as hex-encoded WKB in a string column and
the CRS is carried on the returned node; use \code{\link[=collect_sf]{collect_sf()}} to materialize.
Topology is \pkg{sf}/GEOS throughout (an optional dependency, Suggests); some
constructions need projected coordinates.
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
nc$band <- nc$SID74 > 5
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  band = nc$band,
  geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
), f)

# One convex hull per band.
tbl(f) |>
  spatial_construct("convex_hull", by = "band", crs = sf::st_crs(nc)) |>
  collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_dissolve]{spatial_dissolve()}} to merge a group into one feature,
\code{\link[=spatial_map]{spatial_map()}} for per-feature transforms, \code{\link[=collect_sf]{collect_sf()}} to materialize.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.vectra_node}
\alias{print.vectra_node}
\title{Print a vectra query node}
\usage{
\method{print}{vectra_node}(x, ...)
}
\arguments{
\item{x}{A \code{vectra_node} object.}

\item{...}{Ignored.}
}
\value{
Invisible \code{x}.
}
\description{
Print a vectra query node
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_filter}
\alias{spatial_filter}
\title{Keep streamed rows by their spatial relation to a resident layer}
\usage{
spatial_filter(
  x,
  y,
  predicate = NULL,
  negate = FALSE,
  geom = "geometry",
  coords = NULL,
  crs = NA,
  flush_rows = NULL,
  ...
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{y}{An \code{sf} or \code{sfc} object: the resident locator layer to test against.}

\item{predicate}{An \pkg{sf} binary predicate function, e.g.
\link[sf:st_intersects]{sf::st_intersects} (default), \link[sf:st_within]{sf::st_within}, \link[sf:st_covered_by]{sf::st_covered_by},
\link[sf:st_is_within_distance]{sf::st_is_within_distance}. A left row is kept when the predicate reports
at least one match against \code{y}.}

\item{negate}{If \code{TRUE}, keep the rows with no match instead (the inverted
select-by-location). Default \code{FALSE}.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{coords}{Optional length-2 character vector naming the x and y
coordinate columns to assemble point geometry from (e.g. \code{c("x", "y")}),
for inputs such as \code{\link[=tiff_extract_points]{tiff_extract_points()}} output. The coordinate columns
are retained.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}

\item{...}{Further arguments passed to \code{predicate}, e.g. \verb{dist =} for
\link[sf:st_is_within_distance]{sf::st_is_within_distance}.}
}
\value{
A \code{vectra_node} of the kept rows with \code{x}'s schema, backed by
temporary \code{.vtr} spills and carrying the input CRS.
}
\description{
Streams a large left side \code{x} through the engine and keeps each row whose
geometry satisfies an \pkg{sf} binary predicate against a small resident
layer \code{y} (select by location). This is the spatial counterpart to a
\code{\link[=semi_join]{semi_join()}}: rows are filtered, never duplicated, and no columns are added,
so the output carries \code{x}'s schema unchanged. With \code{negate = TRUE} it keeps
the rows that do \emph{not} match (select by location, inverted). The billion-row
left stream never materializes; \code{y} (a study region, habitat patches, a
coastline buffer, ...) stays resident.
}
\details{
For the recognised predicates -- the topological ones (intersects, within,
contains, overlaps, covers, covered by, touches, crosses), equals, disjoint,
and within-distance (\link[sf:st_is_within_distance]{sf::st_is_within_distance}, whose radius is passed as
\verb{dist =}) -- on projected or unprojected planar data, the test runs natively
on the GEOS C API straight off the hex-WKB column: \code{y} is parsed once into a
spatial index and each batch is tested in C, with no per-batch round-trip
through \pkg{sf}. Coordinate-assembled (\code{coords}) point input runs natively
too, building each point in C rather than through \pkg{sf}; disjoint is the
one exception there (its matches are the bounding boxes the index prunes
away) and keeps the \pkg{sf} loop, as it does for the join. Geographic
coordinates with spherical geometry on (\code{sf::sf_use_s2()}) and any other
predicate use \pkg{sf} instead, preserving its semantics. When \code{y} carries no
CRS it inherits the stream's so the predicate does not reject on a mismatch.
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
region <- nc[nc$NAME \%in\% c("Ashe", "Alleghany", "Surry"), "NAME"]

set.seed(1)
pts <- sf::st_coordinates(sf::st_sample(nc, 300))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = seq_len(nrow(pts)), x = pts[, 1], y = pts[, 2]), f)

# Keep only the points that fall inside the three-county region, streaming.
inside <- tbl(f) |>
  spatial_filter(region, coords = c("x", "y"), crs = sf::st_crs(nc))
nrow(collect(inside))
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_join]{spatial_join()}} to tag rows with \code{y}'s attributes, \code{\link[=spatial_clip]{spatial_clip()}}
to cut geometry against a mask, \code{\link[=filter]{filter()}} for attribute predicates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_utility.R
\name{distinct}
\alias{distinct}
\title{Keep distinct/unique rows}
\usage{
distinct(.data, ..., .keep_all = FALSE)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Column names (unquoted). If empty, uses all columns.}

\item{.keep_all}{If \code{TRUE}, keep all columns (not just those in \code{...}).}
}
\value{
A \code{vectra_node} with unique rows.
}
\description{
Keep distinct/unique rows
}
\details{
Uses hash-based grouping with zero aggregations. When \code{.keep_all = TRUE}
with a column subset, falls back to R's \code{duplicated()} with a message.

This is a materializing operation.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> distinct(cyl) |> collect()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/joins.R
\name{cross_join}
\alias{cross_join}
\title{Cross join two vectra tables}
\usage{
cross_join(x, y, suffix = c(".x", ".y"), ...)
}
\arguments{
\item{x}{A \code{vectra_node} object or data.frame.}

\item{y}{A \code{vectra_node} object or data.frame.}

\item{suffix}{Suffixes for disambiguating column names (default \code{c(".x", ".y")}).}

\item{...}{Ignored.}
}
\value{
A data.frame with \code{nrow(x) * nrow(y)} rows.
}
\description{
Returns every combination of rows from \code{x} and \code{y} (Cartesian product).
Both tables are collected before joining.
}
\examples{
f1 <- tempfile(fileext = ".vtr")
f2 <- tempfile(fileext = ".vtr")
write_vtr(data.frame(a = 1:2), f1)
write_vtr(data.frame(b = c("x", "y", "z"), stringsAsFactors = FALSE), f2)
cross_join(tbl(f1), tbl(f2))
unlink(c(f1, f2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_read_time_slice}
\alias{vec_read_time_slice}
\title{Read a single time slice from a .vec time cube}
\usage{
vec_read_time_slice(r, time, band = 1L, level = 0L, cols = NULL, rows = NULL)
}
\arguments{
\item{r}{A \code{vectra_raster} from \code{vec_open_raster()}.}

\item{time}{Time value to match (numeric/integer).}

\item{band}{Band index (1-based).}

\item{level}{Overview level. Default 0.}

\item{cols, rows}{1-based ranges, same as \code{vec_read_window}.}
}
\value{
A numeric matrix.
}
\description{
Performs a linear scan of the index for tiles with \code{time == time} and
decodes the matching window. The lookup is O(n_tiles) per call — Phase
6's optimized hash-map lookup is a follow-up.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_basic.R
\name{arrange}
\alias{arrange}
\title{Sort rows by column values}
\usage{
arrange(.data, ...)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Column names (unquoted). Wrap in \code{\link[=desc]{desc()}} for descending order.}
}
\value{
A new \code{vectra_node} with sorted rows.
}
\description{
Sort rows by column values
}
\details{
Uses an external merge sort with a 1 GB memory budget. When data exceeds
this limit, sorted runs are spilled to temporary \code{.vtr} files and merged
via a k-way min-heap. NAs sort last in ascending order.

This is a materializing operation.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> arrange(desc(mpg)) |> collect() |> head()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/network.R
\name{spatial_route}
\alias{spatial_route}
\title{Shortest paths and origin-destination costs over a network}
\usage{
spatial_route(
  x,
  network,
  to,
  to_id = NULL,
  geometry = TRUE,
  cost_col = "cost",
  dest_col = "destination",
  geom = "geometry",
  coords = NULL,
  crs = NA,
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} of origin features (point geometry, or \code{coords}).}

\item{network}{A \code{vectra_network} from \code{\link[=spatial_network]{spatial_network()}}.}

\item{to}{An \code{sf}/\code{sfc} of destination points. One destination routes every
origin to it; several produce one row per (origin, destination).}

\item{to_id}{Optional column in \code{to} identifying each destination in the
output. \code{NULL} (default) uses the 1-based destination index.}

\item{geometry}{If \code{TRUE} (default) each row carries the route line; if
\code{FALSE} only the cost column (the cost-matrix form, no geometry).}

\item{cost_col, dest_col}{Names of the output cost and destination-identifier
columns. Defaults \code{"cost"}, \code{"destination"}.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{coords}{Optional length-2 character vector naming the x and y
coordinate columns to assemble point geometry from (e.g. \code{c("x", "y")}),
for inputs such as \code{\link[=tiff_extract_points]{tiff_extract_points()}} output. The coordinate columns
are retained.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node}: one row per (origin, destination) carrying \code{x}'s
attributes, the destination id, the cost, and (when \code{geometry = TRUE}) the
route geometry. Backed by temporary \code{.vtr} spills removed when the node is
garbage-collected, and carrying the network CRS.
}
\description{
Streams a layer of origins \code{x} past a resident \code{\link[=spatial_network]{spatial_network()}} and, for
each origin, finds the shortest path to one or more destinations \code{to}. Each
origin and destination is snapped to its nearest graph node; the solver
(native-C Dijkstra) returns one row per (origin, destination) pair with the
total cost and, by default, the route geometry. With \code{geometry = FALSE} only
the cost is returned, so a destination set per origin yields the
origin-destination cost matrix in long form. The billion-origin stream never
materialises; the graph stays resident.
}
\details{
An unreachable destination returns an infinite cost and an empty geometry
rather than dropping the row, so a cost matrix stays rectangular. Snapping is
to the nearest node, so place origins and destinations on or near the network;
costs are in the units of the network's \code{weight} (or CRS length units when the
graph was built from geometry length). The \pkg{sf} package is an optional
dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mk <- function(x1, y1, x2, y2)
  sf::st_linestring(rbind(c(x1, y1), c(x2, y2)))
streets <- sf::st_sfc(
  mk(0, 0, 1, 0), mk(1, 0, 2, 0), mk(0, 0, 0, 1),
  mk(0, 1, 1, 1), mk(1, 0, 1, 1), mk(1, 1, 2, 1), mk(2, 0, 2, 1))
net <- spatial_network(streets)

f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = 1:2, x = c(0, 0), y = c(0, 1)), f)
dest <- sf::st_sfc(sf::st_point(c(2, 1)))

tbl(f) |>
  spatial_route(net, to = dest, coords = c("x", "y")) |>
  collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_network]{spatial_network()}} to build the graph, \code{\link[=spatial_service_area]{spatial_service_area()}} for
reachability, \code{\link[=collect_sf]{collect_sf()}} to materialize routes as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{glimpse}
\alias{glimpse}
\title{Get a glimpse of a vectra table}
\usage{
glimpse(x, width = 5L, ...)
}
\arguments{
\item{x}{A \code{vectra_node} object.}

\item{width}{Maximum number of preview rows to fetch (default 5).}

\item{...}{Ignored.}
}
\value{
Invisible \code{x}.
}
\description{
Shows column names, types, and a preview of the first few values without
collecting the full result.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> glimpse()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index.R
\name{create_index}
\alias{create_index}
\title{Create a hash index on a .vtr file column}
\usage{
create_index(path, column, ci = FALSE)
}
\arguments{
\item{path}{Path to a \code{.vtr} file.}

\item{column}{Character vector. Name(s) of column(s) to index.}

\item{ci}{Logical. Build a case-insensitive index? Default \code{FALSE}.}
}
\value{
Invisible \code{NULL}. The index is written as a \code{.vtri} sidecar file.
}
\description{
Builds a persistent hash index stored as a \code{.vtri} sidecar file alongside
the \code{.vtr} file. The index maps key hashes to row group indices, enabling
O(1) row group identification for equality predicates (\code{filter(col == value)}).
}
\details{
For composite indexes on multiple columns, pass a character vector.
Composite indexes accelerate AND-combined equality predicates
(e.g., \code{filter(col1 == "a", col2 == "b")}).

The index is automatically loaded by \code{\link[=tbl]{tbl()}} when present. It composes with
zone-map pruning and binary search on sorted columns.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(id = letters, val = 1:26, stringsAsFactors = FALSE), f)
create_index(f, "id")
tbl(f) |> filter(id == "m") |> collect()
unlink(c(f, paste0(f, ".id.vtri")))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_build_overviews}
\alias{vec_build_overviews}
\title{Build overview pyramids for a .vec raster}
\usage{
vec_build_overviews(
  path,
  levels,
  resampling = c("average", "nearest", "bilinear", "mode", "gauss"),
  compression = c("fast", "balanced", "max")
)
}
\arguments{
\item{path}{Path to a \code{.vec} raster file. The file is modified in place.}

\item{levels}{Total levels including level 0 (so \code{levels = 5} adds
four overviews: levels 1..4). Must be in \verb{[2, 16]}.}

\item{resampling}{One of \code{"nearest"}, \code{"average"}, \code{"bilinear"},
\code{"mode"}, \code{"gauss"}. \code{"average"} is the right choice for continuous
rasters; \code{"mode"} for categorical/land-cover.}

\item{compression}{Compression effort for the new tiles. Defaults to
\code{"fast"} because overview tiles are usually one-shot writes.}
}
\value{
Invisible \code{NULL}.
}
\description{
Appends \code{n_levels - 1} reduced-resolution copies of the raster to the
file. Each level is computed by 2x downsampling the previous level
with the chosen kernel. Reading via \code{vec_read_window(level = L)}
picks tiles at level L; the file's \code{n_levels} is updated in place.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{tbl_sqlite}
\alias{tbl_sqlite}
\title{Create a lazy table reference from a SQLite database}
\usage{
tbl_sqlite(path, table, batch_size = .DEFAULT_BATCH_SIZE)
}
\arguments{
\item{path}{Path to a SQLite database file.}

\item{table}{Name of the table to scan.}

\item{batch_size}{Number of rows per batch (default 65536).}
}
\value{
A \code{vectra_node} object representing a lazy scan of the table.
}
\description{
Opens a SQLite database and lazily scans a table. Column types are inferred
from declared types in the CREATE TABLE statement. All filtering, grouping,
and aggregation is handled by vectra's C engine --- no SQL parsing needed.
No data is read until \code{\link[=collect]{collect()}} is called.
}
\examples{
\donttest{
f <- tempfile(fileext = ".sqlite")
write_sqlite(mtcars, f, "cars")
node <- tbl_sqlite(f, "cars")
node |> filter(cyl == 6) |> collect()
unlink(f)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chunk.R
\name{chunk_feeder}
\alias{chunk_feeder}
\title{Turn a query into a resettable chunk generator}
\usage{
chunk_feeder(.source)
}
\arguments{
\item{.source}{Either a function of no arguments returning a fresh
\code{vectra_node} each time it is called (e.g. \code{function() tbl_csv("occ.csv") |> select(presence, bio1, bio12)}), or an offloaded node from \code{\link[=offload]{offload()}}.
Every chunk must contain all variables the consumer's formula references.}
}
\value{
A function \verb{function(reset = FALSE)}. With \code{reset = TRUE} it rewinds
and returns \code{invisible(NULL)}; with \code{reset = FALSE} it returns the next
chunk as a data.frame, or \code{NULL} at end of stream.
}
\description{
Wraps a query so a pull-based consumer can read it one chunk at a time and
re-read it from the start as many times as needed. The returned closure
follows the \code{data(reset)} protocol that \code{biglm::bigglm()} expects: called
with \code{reset = TRUE} it rewinds to the beginning of the data, and called with
\code{reset = FALSE} it returns the next chunk as a data.frame, or \code{NULL} once the
data is exhausted. This lets \code{bigglm()} fit a generalized linear model on a
dataset larger than RAM, streaming each iteratively reweighted pass through
the engine without ever holding the full design matrix.
}
\details{
Because a vectra node is consumed as it streams, re-reading requires a fresh
node on each pass. \code{chunk_feeder()} accepts either form: a \emph{factory}, a
function of no arguments that returns a new node each time it is called; or an
offloaded node from \code{\link[=offload]{offload()}}, which is backed by a file and replays from
disk directly. On every \code{reset = TRUE} a fresh stream is started, so the same
query is replayed on each pass.

Prefer feeding an \code{\link[=offload]{offload()}} of the prepared query: the pipeline (scan,
joins, mutate) runs once into the spill, and every reweighted pass is then a
disk scan of the prepared columns rather than a re-run of the pipeline.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

feed <- chunk_feeder(function() tbl(f) |> select(mpg, wt, hp))
feed(reset = TRUE)       # rewind to the start of the stream
first <- feed()          # first chunk as a data.frame
head(first)

\donttest{
# Out-of-core GLM: prepare once with offload(), then bigglm() replays it.
if (requireNamespace("biglm", quietly = TRUE)) {
  s <- offload(tbl(f) |> select(mpg, wt, hp))
  fit <- biglm::bigglm(mpg ~ wt + hp, data = chunk_feeder(s),
                       family = gaussian())
  coef(fit)
}
}
unlink(f)

}
\seealso{
\code{\link[=offload]{offload()}} for the replay cache, and \code{\link[=collect_chunked]{collect_chunked()}} for
single-pass reductions that vectra drives.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_grouping.R
\name{ungroup}
\alias{ungroup}
\title{Remove grouping from a vectra query}
\usage{
ungroup(x, ...)
}
\arguments{
\item{x}{A \code{vectra_node} object.}

\item{...}{Ignored.}
}
\value{
An ungrouped \code{vectra_node}.
}
\description{
Remove grouping from a vectra query
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> group_by(cyl) |> ungroup()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_write_time_cube}
\alias{vec_write_time_cube}
\title{Write a 4D time-cube raster to .vec}
\usage{
vec_write_time_cube(
  x,
  times,
  path,
  dtype = "f32",
  tile_size = 512L,
  layout = c("image", "pixel"),
  extent = NULL,
  gt = NULL,
  epsg = 0L,
  nodata = NA_real_,
  band_names = NULL,
  compression = c("fast", "balanced", "max")
)
}
\arguments{
\item{x}{Numeric 4D array \code{c(rows, cols, bands, time)}.}

\item{times}{Numeric/integer vector with \code{length(times) == dim(x)[4]},
in the unit of your choice (epoch ms, year, step index).}

\item{path}{Output \code{.vec} path.}

\item{dtype}{Storage dtype (see \code{vec_write_raster}).}

\item{tile_size}{Tile edge in pixels.}

\item{layout}{Tile layout — one of \code{"image"} (default; one tile per
\verb{(band, time, ty, tx)}, optimal for "give me one full image at time
T" reads) or \code{"pixel"} (Phase 6b; one tile per \verb{(band, ty, tx)}
holding the full time stack as \verb{[tw*th, n_time]}, optimal for "give
me the time series at pixel \verb{(x, y)}" reads).}

\item{extent, gt, epsg, nodata, band_names, compression}{Same semantics as
\code{vec_write_raster()}.}
}
\value{
Invisible \code{NULL}.
}
\description{
Each (band, time) combination becomes a stack of tiles tagged with the
chosen time stamp. Stamps are stored as int64 in the per-tile index
entry; a value of \code{0} is reserved for "untimed" so this writer remaps
any caller-supplied 0 to 1 internally.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{warp}
\alias{warp}
\title{Resample or reproject a streamed raster onto a target grid}
\usage{
warp(
  x,
  template,
  method = c("near", "bilinear", "cubic"),
  band = 1L,
  path = NULL,
  dtype = "f32",
  compression = c("fast", "balanced", "max")
)
}
\arguments{
\item{x}{A \code{vectra_raster} (from \code{\link[=vec_open_raster]{vec_open_raster()}}) or a path to a \code{.vec}
raster to warp.}

\item{template}{The target grid: a \code{vectra_raster} / \code{.vec} path whose grid
and CRS are borrowed, or a list \code{list(crs =, extent =, res =, dims =)}.
With \code{crs} and \code{res} but no \code{extent}, the target extent is the source's
corners projected into \code{crs}.}

\item{method}{Resampling method: \code{"near"}, \code{"bilinear"}, or \code{"cubic"}.
Default \code{"near"}.}

\item{band}{Band to warp (1-based). Default 1.}

\item{path}{Optional output \code{.vec} path. When given the result is streamed to
disk and the opened \code{\link[=vec_open_raster]{vec_open_raster()}} handle is returned invisibly; when
\code{NULL} the result is returned as an in-memory matrix.}

\item{dtype}{Storage dtype for \code{.vec} output (see \code{\link[=vec_write_raster]{vec_write_raster()}}).
Default \code{"f32"}.}

\item{compression}{Compression effort for \code{.vec} output. Default \code{"fast"}.}
}
\value{
When \code{path} is \code{NULL}, a numeric matrix on the target grid (row 1
northmost) carrying \code{gt}, \code{extent}, and \code{crs} attributes. When \code{path} is
given, the written \code{vectra_raster} handle (invisibly).
}
\description{
Warps a \code{.vec} raster onto a target grid, walking the \emph{output} one tile-row
strip at a time. For each strip the target pixel-centre coordinates are built,
projected into the source coordinate reference system when the two CRSs differ
(delegated to PROJ via \pkg{sf}), mapped through the source geotransform to
fractional source pixels, and sampled from the bounded source window those
coordinates fall in. The output is assembled in memory or streamed straight
back to a new \code{.vec}, so the whole output grid is never resident; the source
is read in bounded windows rather than held whole.
}
\details{
This is the \emph{sort / partition} tier of the spatial toolbox: each output strip
reads the source window it projects onto. For a mild reprojection or a plain
resample that window is a thin band; a strong reprojection can make it large,
but the output stays streamed throughout.

Sampling follows the GDAL / \pkg{terra} convention (pixel centres at
half-integer coordinates). \code{"near"} takes the nearest source cell;
\code{"bilinear"} the 2x2 weighted mean; \code{"cubic"} the 4x4 cubic convolution
(Catmull-Rom, a = -0.5). A target cell whose sampling kernel reaches outside
the source extent, or touches a nodata cell, comes back \code{NA}.

Reprojection happens only when both rasters carry a known EPSG code and the
codes differ; otherwise \code{warp()} resamples within a shared CRS and needs no
\pkg{sf}.
}
\examples{
z <- outer(1:8, 1:8, function(r, c) r + 2 * c)
f <- tempfile(fileext = ".vec")
vec_write_raster(z, f, dtype = "f64", extent = c(0, 0, 8, 8))

# Resample onto a finer grid over the same extent.
fine <- warp(f, list(extent = c(0, 0, 8, 8), res = 0.5), method = "bilinear")
dim(fine)
unlink(f)

}
\seealso{
\code{\link[=rasterize]{rasterize()}} to build a raster from streamed vector features,
\code{\link[=focal]{focal()}} for moving-window statistics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{tbl}
\alias{tbl}
\title{Create a lazy table reference from a .vtr file}
\usage{
tbl(path)
}
\arguments{
\item{path}{Path to a \code{.vtr} file.}
}
\value{
A \code{vectra_node} object representing a lazy scan of the file.
}
\description{
Opens a vectra1 file and returns a lazy query node. No data is read until
\code{\link[=collect]{collect()}} is called.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
node <- tbl(f)
print(node)
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{write_sqlite}
\alias{write_sqlite}
\title{Write query results or a data.frame to a SQLite table}
\usage{
write_sqlite(x, path, table, ...)
}
\arguments{
\item{x}{A \code{vectra_node} (lazy query) or a \code{data.frame}.}

\item{path}{File path for the SQLite database.}

\item{table}{Name of the table to create/write into.}

\item{...}{Reserved for future use.}
}
\value{
Invisible \code{NULL}.
}
\description{
For \code{vectra_node} inputs, data is streamed batch-by-batch to disk without
materializing the full result in memory. For \code{data.frame} inputs, the data
is written directly.
}
\examples{
db <- tempfile(fileext = ".sqlite")
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars[1:5, ], f)
tbl(f) |> write_sqlite(db, "cars")
unlink(c(f, db))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_split}
\alias{spatial_split}
\title{Split a streamed layer by a resident blade, or return its crossing points}
\usage{
spatial_split(
  x,
  blade,
  extract = c("pieces", "points"),
  geom = "geometry",
  crs = NA,
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{blade}{An \code{sf} or \code{sfc} object whose geometry cuts the stream (typically
lines, but any geometry whose boundary can node \code{x}).}

\item{extract}{\code{"pieces"} (default) to emit the split pieces, or \code{"points"}
to emit the intersection points of each feature with the blade.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node}: with \code{extract = "pieces"}, one row per piece carrying
\code{x}'s attributes; with \code{extract = "points"}, one row per crossing feature
carrying its intersection points. Backed by temporary \code{.vtr} spills (removed
when the node is garbage-collected) and carrying the input CRS.
}
\description{
Streams a large layer \code{x} through the engine and cuts each batch's geometry
against a small resident \code{blade} layer (the QGIS "split with lines"), one
batch at a time. With \code{extract = "pieces"} (the default) every feature is
divided where the blade crosses it -- a polygon into the faces the blade carves
out, a line into the arcs between crossings -- and each piece is emitted as its
own row with the source attributes copied; a feature the blade does not cross
passes through as a single piece. With \code{extract = "points"} the verb instead
returns, per feature, the points where it meets the blade (the "line
intersections" tool), dropping features that do not cross.
}
\details{
The split is built from \pkg{sf}/GEOS noding and polygonization, so it expects
projected or unprojected planar data; geographic coordinates are best
projected first. The blade is dissolved to one geometry once and held resident
while the left stream flows past. Geometry travels through the engine as
hex-encoded WKB in a string column and the CRS is carried on the returned
node; when \code{blade} carries no CRS it inherits the stream's. The \pkg{sf}
package is an optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
sq <- sf::st_polygon(list(rbind(c(0, 0), c(4, 0), c(4, 4), c(0, 4), c(0, 0))))
blade <- sf::st_sfc(sf::st_linestring(rbind(c(2, -1), c(2, 5))))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1L, geometry = sf::st_as_binary(sf::st_sfc(sq), hex = TRUE)
), f)

# Split the square into two halves along the blade.
tbl(f) |> spatial_split(blade) |> collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_clip]{spatial_clip()}} to cut against a mask without dividing into pieces,
\code{\link[=spatial_overlay]{spatial_overlay()}} to node two polygon layers into a partition,
\code{\link[=collect_sf]{collect_sf()}} to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_raster_layout}
\alias{vec_raster_layout}
\title{Tile layout of an open .vec raster}
\usage{
vec_raster_layout(r)
}
\arguments{
\item{r}{A \code{vectra_raster}.}
}
\value{
Character(1) \code{"image"} or \code{"pixel"}.
}
\description{
Returns \code{"image"} (default Phase 6 layout — one tile per
\verb{(band, time, ty, tx)}) or \code{"pixel"} (Phase 6b transpose layout — one
tile per \verb{(band, ty, tx)} holding the full time stack).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_utility.R
\name{rename}
\alias{rename}
\title{Rename columns}
\usage{
rename(.data, ...)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Rename pairs: \code{new_name = old_name}.}
}
\value{
A new \code{vectra_node} with renamed columns.
}
\description{
Rename columns
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> rename(miles_per_gallon = mpg) |> collect() |> head()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_close_raster}
\alias{vec_close_raster}
\title{Close a .vec raster handle}
\usage{
vec_close_raster(r)
}
\arguments{
\item{r}{A \code{vectra_raster} returned by \code{vec_open_raster()}.}
}
\value{
Invisible \code{NULL}.
}
\description{
Idempotent. The handle is also auto-released by R's garbage collector.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind.R
\name{bind_rows}
\alias{bind_rows}
\alias{bind_cols}
\title{Bind rows or columns from multiple vectra tables}
\usage{
bind_rows(..., .id = NULL)

bind_cols(...)
}
\arguments{
\item{...}{\code{vectra_node} objects or data.frames to combine.}

\item{.id}{Optional column name for a source identifier.}
}
\value{
A \code{vectra_node} (streaming) when all inputs are \code{vectra_node} with
identical schemas and \code{.id} is NULL. Otherwise a data.frame.
}
\description{
Bind rows or columns from multiple vectra tables
}
\details{
When all inputs are \code{vectra_node} objects with identical column names and
types and no \code{.id} is requested, \code{bind_rows} creates a streaming
\code{ConcatNode} that iterates children sequentially without materializing.

Otherwise, inputs are collected and combined in R. Missing columns are
filled with NA.

\code{bind_cols} requires the same number of rows in each input.
}
\examples{
f1 <- tempfile(fileext = ".vtr")
f2 <- tempfile(fileext = ".vtr")
write_vtr(data.frame(x = 1:3, y = 4:6), f1)
write_vtr(data.frame(x = 7:9, y = 10:12), f2)
bind_rows(tbl(f1), tbl(f2)) |> collect()
bind_cols(tbl(f1), tbl(f2))
unlink(c(f1, f2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{st_write.vectra_node}
\alias{st_write.vectra_node}
\title{Stream a vectra node's geometry to a vector file}
\usage{
\method{st_write}{vectra_node}(
  obj,
  dsn,
  layer = NULL,
  ...,
  geom = "geometry",
  crs = NULL,
  delete_dsn = FALSE,
  quiet = TRUE
)
}
\arguments{
\item{obj}{A \code{vectra_node} whose rows carry a hex-WKB geometry column (from
\code{\link[=spatial_overlay]{spatial_overlay()}}, a grouped \code{\link[=slice_min]{slice_min()}} / \code{\link[=slice_max]{slice_max()}} resolution, a
\code{.vtr} scan, ...). It is consumed by the stream.}

\item{dsn}{Destination data source name (file path).}

\item{layer}{Layer name. \code{NULL} lets \pkg{sf} derive it from \code{dsn}.}

\item{...}{Unused; for S3 generic compatibility.}

\item{geom}{Name of the hex-WKB geometry column. Default \code{"geometry"}.}

\item{crs}{CRS to tag the output with. \code{NULL} takes the CRS carried on the
node.}

\item{delete_dsn}{If \code{TRUE}, remove an existing \code{dsn} before writing.}

\item{quiet}{Passed to \code{\link[sf:st_write]{sf::st_write()}}.}
}
\value{
The \code{dsn}, invisibly.
}
\description{
An \code{\link[sf:st_write]{sf::st_write()}} method (also reached through \code{\link[sf:write_sf]{sf::write_sf()}}) for a
\code{vectra_node}: writes the result a batch at a time, appending each, so the
whole layer is never held in memory. This is the streaming counterpart to
\code{collect_sf(x) |> sf::st_write(...)} -- that route materializes every feature
as an \code{sf} object first, which for a multi-million-feature result dominates
memory; this route's peak is one batch.
}
\seealso{
\code{\link[=collect_sf]{collect_sf()}} to materialize the whole result as one \code{sf} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_to_tiff}
\alias{vec_to_tiff}
\title{Export a .vec raster to GeoTIFF}
\usage{
vec_to_tiff(r, path, compression = c("deflate", "lzw", "none"))
}
\arguments{
\item{r}{Either a path to a \code{.vec} raster or a \code{vectra_raster} returned
by \code{vec_open_raster()}. If a handle is passed it is left open.}

\item{path}{Output \code{.tif} path.}

\item{compression}{One of \code{"deflate"} (default), \code{"lzw"}, or \code{"none"}.}
}
\value{
Invisible \code{NULL}.
}
\description{
Writes the level-0 pixels of a \code{.vec} raster to a GeoTIFF file. The
TIFF inherits dtype, geotransform, EPSG, and nodata from the source.
Strip layout; the writer supports \code{"none"}, \code{"deflate"}, and \code{"lzw"}
compression. LZW also applies horizontal differencing (Predictor 2)
for integer pixel types, which dramatically improves compression on
smooth raster data and matches the layout most production GIS tools
produce by default. Tiled and BigTIFF output land in a follow-up.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/joins.R
\name{fuzzy_join}
\alias{fuzzy_join}
\title{Fuzzy join two vectra tables by string distance}
\usage{
fuzzy_join(
  x,
  y,
  by,
  method = "dl",
  max_dist = 0.2,
  block_by = NULL,
  n_threads = 4L,
  suffix = ".y"
)
}
\arguments{
\item{x}{A \code{vectra_node} object (probe / query side).}

\item{y}{A \code{vectra_node} object (build / reference side).}

\item{by}{A named character vector of length 1: \code{c("probe_col" = "build_col")}.
The columns to compute string distance on.}

\item{method}{Character. Distance algorithm: \code{"dl"} (Damerau-Levenshtein,
default), \code{"levenshtein"}, or \code{"jw"} (Jaro-Winkler).}

\item{max_dist}{Numeric. Maximum normalized distance (0-1) to keep a match.
Default \code{0.2}.}

\item{block_by}{Optional named character vector of length 1:
\code{c("probe_col" = "build_col")}. Rows must match exactly on these columns
before distance is computed. Dramatically reduces comparisons.}

\item{n_threads}{Integer. Number of OpenMP threads for parallel distance
computation over partitions. Default \code{4L}.}

\item{suffix}{Character. Suffix appended to build-side column names that
collide with probe-side names. Default \code{".y"}.}
}
\value{
A \code{vectra_node} with all probe columns, all build columns (suffixed
on collision), and a \code{fuzzy_dist} column (double).
}
\description{
Joins two tables using approximate string matching on key columns.
Optionally blocks by a second column (e.g., genus) for performance —
only rows sharing the same blocking key are compared.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{collect_sf}
\alias{collect_sf}
\title{Materialize a spatial query as an sf object}
\usage{
collect_sf(x, geom = "geometry", crs = NULL)
}
\arguments{
\item{x}{A \code{vectra_node} with a hex-WKB / WKT geometry column, or a
data.frame already collected from one.}

\item{geom}{Name of the geometry column. Default \code{"geometry"}.}

\item{crs}{Override the coordinate reference system. Defaults to the CRS the
node carries, or unknown.}
}
\value{
An \code{sf} object.
}
\description{
Collects a \code{vectra_node} (typically the result of \code{\link[=spatial_map]{spatial_map()}} or
\code{\link[=spatial_join]{spatial_join()}}) into memory and rebuilds an \code{sf} object from its hex-WKB
geometry column. The CRS defaults to the one carried on the node.
}
\details{
This is the spatial counterpart to \code{\link[=collect]{collect()}}: use it when the final result
fits in memory as \code{sf}. For a result still larger than RAM, keep it as a node
and write it out with \code{\link[=write_vtr]{write_vtr()}} (the geometry stays as a WKB string
column) or reduce it with \code{\link[=collect_chunked]{collect_chunked()}}.
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME = nc$NAME,
  geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
), f)
result <- tbl(f) |> spatial_map(~ sf::st_centroid(.x), crs = sf::st_crs(nc))
collect_sf(result)
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_map]{spatial_map()}}, \code{\link[=spatial_join]{spatial_join()}}, \code{\link[=collect]{collect()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{write_csv}
\alias{write_csv}
\title{Write query results or a data.frame to a CSV file}
\usage{
write_csv(x, path, ...)
}
\arguments{
\item{x}{A \code{vectra_node} (lazy query) or a \code{data.frame}.}

\item{path}{File path for the output CSV file.}

\item{...}{Reserved for future use.}
}
\value{
Invisible \code{NULL}.
}
\description{
For \code{vectra_node} inputs, data is streamed batch-by-batch to disk without
materializing the full result in memory. For \code{data.frame} inputs, the data
is written directly.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars[1:5, ], f)
csv <- tempfile(fileext = ".csv")
tbl(f) |> write_csv(csv)
unlink(c(f, csv))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_smooth}
\alias{spatial_smooth}
\title{Smooth streamed line and polygon geometry}
\usage{
spatial_smooth(
  x,
  iterations = 2L,
  keep_ends = TRUE,
  geom = "geometry",
  crs = NA,
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{iterations}{Number of corner-cutting passes (a positive integer). Each
pass roughly doubles the vertex count. Default \code{2}.}

\item{keep_ends}{If \code{TRUE} (default), pin the first and last vertex of an open
line so it is not shortened at its tips. Ignored for closed rings.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of the smoothed geometry with \code{x}'s attributes, backed
by temporary \code{.vtr} spills (removed when the node is garbage-collected) and
carrying the input CRS.
}
\description{
Rounds the corners of every line and polygon in a streamed layer by Chaikin
corner-cutting, one batch at a time. Each iteration replaces every vertex with
two points a quarter and three-quarters of the way along its adjacent edges,
so sharp angles become a sequence of short chamfers that read as a smooth
curve; more \code{iterations} give a smoother result with more vertices. Open lines
keep their endpoints (\code{keep_ends}); polygon rings are cut cyclically and shrink
slightly, as Chaikin smoothing does. Point geometry passes through unchanged.
}
\details{
The smoothing is computed directly on the coordinates (no GEOS call), so it is
dependency-light; \pkg{sf} is used only to decode and rebuild each batch.
Geometry travels through the engine as hex-encoded WKB in a string column and
the CRS is carried on the returned node; use \code{\link[=collect_sf]{collect_sf()}} to materialize.
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
zig <- sf::st_linestring(rbind(c(0, 0), c(1, 1), c(2, 0), c(3, 1), c(4, 0)))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1L, geometry = sf::st_as_binary(sf::st_sfc(zig), hex = TRUE)
), f)

# Smooth the zig-zag with three corner-cutting passes.
tbl(f) |> spatial_smooth(iterations = 3) |> collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_map]{spatial_map()}} for per-feature transforms such as densifying with
\code{~ sf::st_segmentize(.x, dfMaxLength)} or sampling points along a line with
\code{~ sf::st_line_sample(.x, n)}, \code{\link[=collect_sf]{collect_sf()}} to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_basic.R
\name{filter}
\alias{filter}
\title{Filter rows of a vectra query}
\usage{
filter(.data, ...)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{...}{Filter expressions (combined with \code{&}).}
}
\value{
A new \code{vectra_node} with the filter applied.
}
\description{
Filter rows of a vectra query
}
\details{
Filter uses zero-copy selection vectors: matching rows are indexed without
copying data. Multiple conditions are combined with \code{&}. Supported
expression types: arithmetic (\code{+}, \code{-}, \code{*}, \code{/}, \code{\%\%}), comparison
(\code{==}, \code{!=}, \code{<}, \code{<=}, \code{>}, \code{>=}), boolean (\code{&}, \code{|}, \code{!}), \code{is.na()},
and string functions (\code{nchar()}, \code{substr()}, \code{grepl()} with fixed patterns).

NA comparisons return NA (SQL semantics). Use \code{is.na()} to filter NAs
explicitly.

This is a streaming operation (constant memory per batch).
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> filter(cyl > 4) |> collect() |> head()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/network.R
\name{spatial_network}
\alias{spatial_network}
\title{Build a routable network graph from a line layer}
\usage{
spatial_network(
  lines,
  weight = NULL,
  directed = FALSE,
  direction = NULL,
  weight_to = NULL,
  tolerance = 0,
  geom = "geometry",
  crs = NA
)
}
\arguments{
\item{lines}{An \code{sf} or \code{sfc} object of (multi)linestrings, or a \code{vectra_node}
carrying a hex-WKB line geometry column (it is materialised).}

\item{weight}{Name of a column in \code{lines} holding each edge's traversal cost.
\code{NULL} (default) uses the geometry length (\code{\link[sf:st_length]{sf::st_length()}}).}

\item{directed}{If \code{TRUE}, edges are one-way (by \code{direction}, or the digitised
from->to direction). \code{FALSE} (default) makes every edge two-way.}

\item{direction}{Name of a column of one-way codes, used when \code{directed}:
\code{"B"} two-way, \code{"FT"} from->to only, \code{"TF"} to->from only, \code{"N"} closed (the
\code{+}/\code{-}/\code{0} sign convention is also accepted). \code{NULL} uses \code{"FT"} for every
line, or \code{"B"} when \code{weight_to} is given.}

\item{weight_to}{Name of a column holding the reverse-direction cost on a
two-way edge of a directed graph. \code{NULL} reuses \code{weight}.}

\item{tolerance}{Endpoints within this distance (CRS units) are snapped to one
node. \code{0} (default) joins only exactly-coincident endpoints.}

\item{geom, crs}{Geometry column name and CRS, as in \code{\link[=spatial_map]{spatial_map()}}. The CRS
defaults to the one \code{lines} carries.}
}
\value{
A \code{vectra_network} object: the resident graph (an external pointer
plus the node coordinates, edge table, and source-line geometry needed to
snap queries and rebuild routes). Print it to see the node, edge, and
connected-component counts.
}
\description{
Builds the node-edge graph a shortest-path query needs: nodes at line
endpoints, edges weighted by geometry length or a cost column. The graph is
held resident in an external pointer and passed to \code{\link[=spatial_route]{spatial_route()}} and
\code{\link[=spatial_service_area]{spatial_service_area()}}, which stream origin (and destination) batches past
it. This is the network counterpart of a resident \code{sf} \code{y} in \code{\link[=spatial_knn]{spatial_knn()}}:
the graph is the resident budget (bounded by the network size), while the
query side scales by streaming.
}
\details{
Endpoints within \code{tolerance} of each other are snapped to a single node, so a
layer whose touching lines do not share exactly-equal endpoint coordinates
still connects. The input must be (multi)linestrings; a \code{MULTILINESTRING} is
split into its parts, each becoming one edge with the source attributes. Lines
are treated as already split at their junctions (the usual road-network
convention); two lines that merely cross in their interiors are not connected
unless they share an endpoint. The graph and the Dijkstra solver are native C;
\pkg{sf} (an optional dependency, Suggests) supplies only the geometry.
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# A small grid of streets.
mk <- function(x1, y1, x2, y2)
  sf::st_linestring(rbind(c(x1, y1), c(x2, y2)))
streets <- sf::st_sfc(
  mk(0, 0, 1, 0), mk(1, 0, 2, 0), mk(0, 0, 0, 1),
  mk(0, 1, 1, 1), mk(1, 0, 1, 1), mk(1, 1, 2, 1), mk(2, 0, 2, 1))
net <- spatial_network(streets)
net
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_route]{spatial_route()}} for shortest paths and origin-destination costs,
\code{\link[=spatial_service_area]{spatial_service_area()}} for reachability and isochrones.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block.R
\name{block_fuzzy_lookup}
\alias{block_fuzzy_lookup}
\title{Fuzzy-match query keys against a materialized block}
\usage{
block_fuzzy_lookup(
  block,
  column,
  keys,
  method = "dl",
  max_dist = 0.2,
  block_col = NULL,
  block_keys = NULL,
  n_threads = 4L
)
}
\arguments{
\item{block}{A \code{vectra_block} from \code{\link[=materialize]{materialize()}}.}

\item{column}{Character scalar. Name of the string column to fuzzy-match against.}

\item{keys}{Character vector. Query strings to match.}

\item{method}{Character. Distance method: \code{"dl"} (Damerau-Levenshtein, default),
\code{"levenshtein"}, or \code{"jw"} (Jaro-Winkler).}

\item{max_dist}{Numeric. Maximum normalized distance (default 0.2).}

\item{block_col}{Optional character scalar. Column name for exact-match blocking
(e.g., genus). When provided, only rows where \code{block_col} matches the
corresponding \code{block_keys} value are compared.}

\item{block_keys}{Optional character vector (same length as \code{keys}). Exact-match
values for blocking. Required when \code{block_col} is provided.}

\item{n_threads}{Integer. Number of OpenMP threads (default 4L).}
}
\value{
A data.frame with columns \code{query_idx} (1-based position in \code{keys}),
\code{fuzzy_dist} (normalized distance), plus all columns from the block.
}
\description{
Computes string distances between query keys and a string column in a
materialized block. Optionally uses exact-match blocking on a second column
(e.g., genus) to reduce the search space.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{tbl_xlsx}
\alias{tbl_xlsx}
\title{Create a lazy table reference from an Excel (.xlsx) file}
\usage{
tbl_xlsx(path, sheet = 1L, batch_size = .DEFAULT_BATCH_SIZE)
}
\arguments{
\item{path}{Path to an \code{.xlsx} file.}

\item{sheet}{Sheet to read: either a name (character) or 1-based index
(integer). Default \code{1L} (first sheet).}

\item{batch_size}{Number of rows per batch (default 65536).}
}
\value{
A \code{vectra_node} object representing a lazy scan of the sheet.
}
\description{
Reads a sheet from an Excel workbook into a vectra node for lazy query
execution. The sheet is read into memory via
\code{\link[openxlsx2:read_xlsx]{openxlsx2::read_xlsx()}} and then converted
to vectra's internal format. Requires the \pkg{openxlsx2} package.
}
\examples{
\donttest{
if (requireNamespace("openxlsx2", quietly = TRUE)) {
  f <- tempfile(fileext = ".xlsx")
  openxlsx2::write_xlsx(mtcars, f)
  node <- tbl_xlsx(f)
  node |> filter(cyl == 6) |> collect()
  unlink(f)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_read_window}
\alias{vec_read_window}
\title{Read a window of pixels from a .vec raster}
\usage{
vec_read_window(r, band = 1L, level = 0L, cols = NULL, rows = NULL)
}
\arguments{
\item{r}{A \code{vectra_raster} from \code{vec_open_raster()}.}

\item{band}{Band index (1-based). Default 1.}

\item{level}{Overview level — 0 = full resolution, 1 = half, 2 =
quarter, etc. Must be < \code{r$n_levels} (which is 1 unless
\code{vec_build_overviews()} has been run on the file).}

\item{cols}{1-based column range \code{c(col_min, col_max)}. Inclusive.
Coordinates are in the chosen level's pixel grid (so at level 1 the
raster is half as wide). Default \code{c(1, level_width)}.}

\item{rows}{1-based row range \code{c(row_min, row_max)}. Inclusive.
Default \code{c(1, level_height)}.}
}
\value{
A numeric matrix with \code{nrow = row_max - row_min + 1} and
\code{ncol = col_max - col_min + 1}. Nodata pixels become \code{NA}.
}
\description{
Decodes only the tiles overlapping the requested window. Pixels outside
the raster extent come back as \code{NA}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_more.R
\name{mosaic}
\alias{mosaic}
\title{Merge aligned rasters onto a common grid}
\usage{
mosaic(
  rasters,
  fun = c("first", "last", "mean", "sum", "min", "max"),
  band = 1L,
  path = NULL,
  dtype = "f32",
  compression = c("fast", "balanced", "max")
)
}
\arguments{
\item{rasters}{A list of \code{vectra_raster} handles or \code{.vec} paths sharing
resolution and grid alignment.}

\item{fun}{Overlap rule where inputs cover the same cell: \code{"first"} (the
earliest input in \code{rasters}, default), \code{"last"}, \code{"mean"}, \code{"sum"}, \code{"min"},
or \code{"max"}. Cells covered by no input come back \code{NA}.}

\item{band}{Band read from every input (1-based). Default 1.}

\item{path}{Optional output \code{.vec} path. When given the result is streamed to
disk and the opened \code{\link[=vec_open_raster]{vec_open_raster()}} handle is returned invisibly; when
\code{NULL} the result is returned as an in-memory matrix.}

\item{dtype}{Storage dtype for \code{.vec} output (see \code{\link[=vec_write_raster]{vec_write_raster()}}).
Default \code{"f32"}.}

\item{compression}{Compression effort for \code{.vec} output. Default \code{"fast"}.}
}
\value{
When \code{path} is \code{NULL}, a numeric matrix on the union grid (row 1
northmost) carrying \code{gt}, \code{extent}, and \code{crs} attributes. When \code{path} is
given, the written \code{vectra_raster} handle (invisibly).
}
\description{
Combines several \code{.vec} rasters that share a resolution and cell grid into one
raster spanning their union, resolving overlap with \code{fun}. The output is
walked one tile-row strip at a time and each input contributes only the window
overlapping the current strip, so neither the inputs nor the output are held
whole in memory.
}
\details{
This is the \emph{monoid fold} tier of the spatial toolbox: each output strip folds
the overlapping input windows, bounded memory, no spill. The inputs must share
resolution and lie on a common cell grid; \code{\link[=warp]{warp()}} them onto a shared grid
first if they do not. No \pkg{sf} is needed.
}
\examples{
a <- matrix(1, 4, 4); b <- matrix(2, 4, 4)
fa <- tempfile(fileext = ".vec"); fb <- tempfile(fileext = ".vec")
vec_write_raster(a, fa, dtype = "f64", extent = c(0, 0, 4, 4))
vec_write_raster(b, fb, dtype = "f64", extent = c(2, 2, 6, 6))

m <- mosaic(list(fa, fb), fun = "mean")
dim(m)
unlink(c(fa, fb))

}
\seealso{
\code{\link[=warp]{warp()}} to bring rasters onto a shared grid, \code{\link[=rast_calc]{rast_calc()}} for
cellwise combination of already-aligned rasters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_explode}
\alias{spatial_explode}
\title{Explode multipart geometries into single-part features}
\usage{
spatial_explode(
  x,
  geom = "geometry",
  crs = NA,
  out_geom = NULL,
  part = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{part}{Optional name of an integer column numbering the parts within each
source feature, 1-based. Default \code{NULL} adds no such column.}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of single-part features, backed by temporary \code{.vtr}
spills (removed when the node is garbage-collected) and carrying the input
CRS.
}
\description{
Streams a lazy vectra query and splits every multipart geometry into its
component single-part geometries: a \code{MULTIPOLYGON} becomes one row per
polygon, a \code{MULTILINESTRING} one row per linestring, a \code{MULTIPOINT} one row
per point, and a \code{GEOMETRYCOLLECTION} one row per member (recursively). The
attributes of the source feature are copied onto each part. Already
single-part geometries pass through unchanged, as does an empty geometry
(kept as one row). This is the streaming counterpart of the QGIS
"multipart to singleparts" tool and of \code{\link[sf:st_cast]{sf::st_cast()}} to a single-part type.
}
\details{
One batch is decoded, exploded, and spilled at a time, so peak memory tracks
one batch and its parts, not the whole layer.
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mp <- sf::st_multipolygon(list(
  list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))),
  list(rbind(c(2, 2), c(3, 2), c(3, 3), c(2, 3), c(2, 2)))))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1L,
  geometry = sf::st_as_binary(sf::st_sfc(mp), hex = TRUE)
), f)

# One row per polygon, attributes copied, parts numbered.
tbl(f) |> spatial_explode(part = "part_id") |> collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_map]{spatial_map()}} for per-feature transforms, \code{\link[=spatial_dissolve]{spatial_dissolve()}} to
merge features the other way, \code{\link[=collect_sf]{collect_sf()}} to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tbl.R
\name{tbl_tiff}
\alias{tbl_tiff}
\title{Create a lazy table reference from a GeoTIFF raster}
\usage{
tbl_tiff(path, batch_size = .TIFF_BATCH_SIZE)
}
\arguments{
\item{path}{Path to a GeoTIFF file.}

\item{batch_size}{Number of raster rows per batch (default 256).}
}
\value{
A \code{vectra_node} object representing a lazy scan of the raster.
}
\description{
Opens a GeoTIFF file and returns a lazy query node. Each pixel becomes a row
with columns \code{x}, \code{y}, \code{band1}, \code{band2}, etc. Coordinates are pixel centers
derived from the affine geotransform. NoData values become \code{NA}.
}
\details{
Use \code{filter(x >= ..., y <= ...)} for extent-based cropping and
\code{filter(band1 > ...)} for value-based cropping. Results can be converted
back to a raster with \code{terra::rast(df, type = "xyz")}.
}
\examples{
\donttest{
f <- tempfile(fileext = ".tif")
df <- data.frame(x = as.double(rep(1:4, 3)),
                 y = as.double(rep(1:3, each = 4)),
                 band1 = as.double(1:12))
write_tiff(df, f)
node <- tbl_tiff(f)
node |> filter(band1 > 6) |> collect()
unlink(f)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collect.R
\name{collect}
\alias{collect}
\title{Execute a lazy query and return a data.frame}
\usage{
collect(x, ...)
}
\arguments{
\item{x}{A \code{vectra_node} object.}

\item{...}{Ignored.}
}
\value{
A data.frame with the query results.
}
\description{
Pulls all batches from the execution plan and materializes the result as an
R data.frame.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
result <- tbl(f) |> collect()
head(result)
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{terrain}
\alias{terrain}
\title{Terrain derivatives from a streamed elevation raster}
\usage{
terrain(
  x,
  v = c("slope", "aspect", "hillshade", "TPI", "roughness", "TRI"),
  unit = c("degrees", "radians"),
  azimuth = 315,
  altitude = 45,
  band = 1L,
  path = NULL,
  dtype = "f32",
  compression = c("fast", "balanced", "max")
)
}
\arguments{
\item{x}{A \code{vectra_raster} (from \code{\link[=vec_open_raster]{vec_open_raster()}}) or a path to a \code{.vec}
elevation raster.}

\item{v}{Derivatives to compute, any of \code{"slope"}, \code{"aspect"},
\code{"hillshade"}, \code{"TPI"} (topographic position index), \code{"roughness"},
\code{"TRI"} (terrain ruggedness index). The return follows the input: one
matrix for a single \code{v}, a named list for several.}

\item{unit}{Angular unit for \code{slope} and \code{aspect}: \code{"degrees"} (default) or
\code{"radians"}.}

\item{azimuth, altitude}{Sun position for \code{"hillshade"}, in degrees. Defaults
315 (NW) and 45.}

\item{band}{Band to read (1-based). Default 1.}

\item{path}{Optional output \code{.vec} path (one band per \code{v}, named after \code{v}).
When given the result is streamed to disk and the opened
\code{\link[=vec_open_raster]{vec_open_raster()}} handle is returned invisibly; when \code{NULL} the result is
returned in memory.}

\item{dtype}{Storage dtype for \code{.vec} output (see \code{\link[=vec_write_raster]{vec_write_raster()}}).
Default \code{"f32"}.}

\item{compression}{Compression effort for \code{.vec} output. Default \code{"fast"}.}
}
\value{
When \code{path} is \code{NULL}: a numeric matrix for a single \code{v}, or a named
list of matrices for several, each carrying \code{gt}, \code{extent}, and \code{crs}
attributes (row 1 northmost). When \code{path} is given, the written multi-band
\code{vectra_raster} handle (invisibly).
}
\description{
Computes DEM derivatives from a \code{.vec} elevation raster with Horn's 3x3
method, on the same haloed tile-row strip pass as \code{\link[=focal]{focal()}} -- the input is
read one strip at a time and, when \code{path} is given, the outputs are streamed
straight back to a multi-band \code{.vec}. Matches \pkg{terra}'s
\code{terrain()} / \code{shade()} conventions.
}
\details{
Slope and aspect use the Horn (1981) finite-difference gradient over
the 3x3 neighbourhood; \code{aspect} is degrees clockwise from north (flat cells
return 90). \code{hillshade} is the cosine of the incidence angle for the given
sun position, clamped at 0. \code{TPI} is the cell minus the mean of its eight
neighbours; \code{roughness} is the range over the 3x3; \code{TRI} is the mean
absolute difference to the eight neighbours. Cells whose 3x3 neighbourhood
touches a nodata value or the raster edge return \code{NA}.
}
\examples{
# A tilted surface so slope and aspect are well defined.
z <- outer(1:8, 1:8, function(r, c) 10 + 2 * c + r)
f <- tempfile(fileext = ".vec")
vec_write_raster(z, f, dtype = "f64", extent = c(0, 0, 8, 8))

slp <- terrain(f, v = "slope")
deriv <- terrain(f, v = c("slope", "aspect", "hillshade"))
names(deriv)
unlink(f)

}
\seealso{
\code{\link[=focal]{focal()}} for arbitrary moving windows.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write.R
\name{append_vtr}
\alias{append_vtr}
\title{Append rows to an existing .vtr file}
\usage{
append_vtr(x, path, ...)
}
\arguments{
\item{x}{A \code{vectra_node} (lazy query) or a \code{data.frame}.}

\item{path}{File path of an existing \code{.vtr} file to append to.}

\item{...}{Additional arguments passed to methods.}
}
\value{
Invisible \code{NULL}.
}
\description{
Appends one or more new row groups to the end of an existing \code{.vtr} file
without touching or recompressing existing row groups. The schema of \code{x}
must exactly match the schema of the target file (same column names and
types, in the same order).
}
\details{
The operation is not fully atomic: if the process is interrupted after
new row groups are written but before the header is patched, the file
will be in a corrupted state. Use \code{write_vtr()} for safety-critical
write-once workloads.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars[1:10, ], f)
append_vtr(mtcars[11:20, ], f)
result <- tbl(f) |> collect()
stopifnot(nrow(result) == 20L)
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/raster.R
\name{vec_extract_points}
\alias{vec_extract_points}
\title{Extract band values at (x, y) points from a .vec raster}
\usage{
vec_extract_points(r, x, y)
}
\arguments{
\item{r}{A \code{vectra_raster} from \code{vec_open_raster()}.}

\item{x}{Numeric vector of x coordinates in CRS units.}

\item{y}{Numeric vector of y coordinates, same length as \code{x}.}
}
\value{
A \code{data.frame} with columns \code{x}, \code{y}, then one column per band
(named after \code{r$band_names} if recorded, otherwise \code{band1}, \code{band2},
...). NA marks pixels outside the raster or matching nodata.
}
\description{
Extract band values at (x, y) points from a .vec raster
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_overlay}
\alias{spatial_overlay}
\title{Self-overlay a polygon layer into disjoint pieces (QGIS-style Union)}
\usage{
spatial_overlay(
  x,
  y = NULL,
  vars = NULL,
  vars_y = NULL,
  how = c("intersection", "union", "identity", "symdiff"),
  piece = "piece_id",
  geom = "geometry",
  grid = NULL,
  precision = NULL,
  dedup = TRUE,
  flush_rows = NULL,
  mem_limit = NULL,
  threads = NULL,
  quiet = TRUE,
  layer = NULL,
  query = NULL,
  layer_y = NULL,
  query_y = NULL,
  read_chunk = NULL
)
}
\arguments{
\item{x}{An \code{sf} object with polygon or multipolygon geometry, or a single
path to a vector file (e.g. a GeoPackage). A path is read in feature batches
via \code{layer} / \code{query}, so the whole layer is never held in memory at once --
peak memory then tracks the cleaned geometry, not the source size, which lets
a larger-than-RAM layer overlay on a modest machine.}

\item{y}{Optional second layer to overlay \code{x} against, in the same forms \code{x}
accepts (an \code{sf} object or a file path read via \code{layer_y} / \code{query_y}). It
must share the CRS of \code{x}. \code{NULL} (the default) self-unions \code{x}.}

\item{vars}{Character vector of attribute columns of \code{x} to carry onto each
piece. Default \code{NULL} keeps them all; name a subset to keep the streamed
output narrow.}

\item{vars_y}{Character vector of attribute columns of \code{y} to carry onto each
piece (two-layer overlay only). Default \code{NULL} keeps them all. A name shared
with an \code{x} column is disambiguated with a \code{.x} / \code{.y} suffix in the output.}

\item{how}{For a two-layer overlay, which pieces to keep: \code{"intersection"}
(covered by both layers; the default), \code{"union"} (every piece of either,
the absent side's attributes filled with \code{NA}), \code{"identity"} (all of \code{x},
split by \code{y}, with \code{y}'s attributes where it covers and \code{NA} elsewhere), or
\code{"symdiff"} (pieces in exactly one layer). Ignored when \code{y = NULL}.}

\item{piece}{Name of the integer piece-id column added to the output (the key
you group by to resolve overlaps). Default \code{"piece_id"}.}

\item{geom}{Name of the output hex-WKB geometry column. Default \code{"geometry"}.}

\item{grid}{Fixed-precision snapping grid size in CRS units. Coordinates are
snapped to this grid before noding so near-duplicate shared boundaries merge
into one. \code{NULL} (the default) derives it from coordinate magnitude
(\code{max(abs(st_bbox(x))) * 3e-8}), which suits projected layers. Pass a number
to override when that default is too coarse for fine geometry (or too coarse
because an outlier coordinate inflated the magnitude), or \code{0} to disable
snapping entirely.}

\item{precision}{Fixed-precision grid size, in CRS units, for noding the
boundary linework. Noding on a fixed grid is deterministic and avoids the
floating noder's repair-and-retry on dense overlapping linework, which is
what makes a large dense layer feasible to overlay. It is far finer than
\code{grid} so intersection points are not collapsed. \code{NULL} (the default)
derives it from coordinate magnitude (\code{max(abs(st_bbox(x))) * 1e-13}); pass
a number to override, or \code{0} to node in floating precision.}

\item{dedup}{Overlay one representative per group of byte-identical cleaned
geometries and fan the per-record attributes back onto its pieces afterwards.
Duplicates add no faces, so the result is identical; this only removes the
redundant noding when many records are stacked over one site (common in
WDPA-style data). \code{TRUE} by default; set \code{FALSE} to overlay every record.}

\item{flush_rows}{Exploded rows buffered before a spill flush. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}

\item{mem_limit}{Approximate peak working-set budget in bytes, bounding the
per-tile size (\code{tile_bytes = mem_limit / (threads * 24)}). It is a throughput
knob with an interior optimum, not "bigger is faster": too small replicates
features across many tiles, too large nodes too much linework per tile (a
superlinear cost), and a budget of tens of GB runs slower than the default on
a dense layer. Lower it for tighter memory. Defaults via
\code{getOption("vectra.overlay_mem_limit", ...)} to a value that scales with
\code{threads} to hold the per-tile size near its measured optimum.}

\item{threads}{Number of OpenMP threads for the per-component overlay within a
chunk. \code{0} (the default, via \code{getOption("vectra.overlay_threads", 0)}) uses
all available cores.}

\item{quiet}{If \code{FALSE}, show a text progress bar over the overlay chunks.}

\item{layer}{When \code{x} is a file path, the name of the layer to read. Ignored
for an \code{sf} \code{x}. Supply this or \code{query}.}

\item{query}{When \code{x} is a file path, a SQL statement selecting the features to
overlay (read in batches via \code{LIMIT}/\code{OFFSET}); use it instead of \code{layer}
for a subset or join. With \code{query} and no \code{layer}, pass \code{grid} explicitly,
since the layer extent cannot be read from the file metadata.}

\item{layer_y, query_y}{The \code{layer} / \code{query} equivalents for a file-path \code{y}.}

\item{read_chunk}{Features per read/parse batch. \code{NULL} (default) sizes it
from available RAM. Smaller batches lower peak memory; larger ones do fewer
round trips.}
}
\value{
A \code{vectra_node} over the exploded overlay, backed by temporary \code{.vtr}
spills removed when the node is garbage-collected, carrying the CRS of \code{x}
for \code{\link[=collect_sf]{collect_sf()}}. For a self-union it is one row per piece per covering
polygon; for a two-layer overlay one row per piece per covering
\code{x}-record / \code{y}-record pair, with the columns of both layers.
}
\description{
Splits a polygon layer along all its own overlaps into disjoint pieces and
returns a lazy node with one row per piece per covering polygon: where \code{k}
polygons overlap, that piece appears \code{k} times, each row carrying one source
polygon's attributes. This is the union overlay GIS tools expose as
"Union (single layer)", with the overlap retained once per contributing
feature rather than dissolved. Resolve the duplicates with a grouped
\code{\link[=slice_min]{slice_min()}} / \code{\link[=slice_max]{slice_max()}} -- for example earliest designation year wins:
\code{group_by(piece_id) |> slice_min(year)}.
}
\details{
The topology is done once with \pkg{sf}/GEOS and tiled over connected overlap
clusters (disjoint clusters never share a piece, so the tiling is exact and
bounded in memory), then the exploded pieces are streamed to a \code{.vtr} and
handed back as a lazy node. Geometry rides through the engine as hex-encoded
WKB in a string column; the CRS is carried on the node for \code{\link[=collect_sf]{collect_sf()}}.

The overlay runs on a fixed-precision model: coordinates are snapped to a
grid derived from their own magnitude so the pieces come out disjoint and
their areas reconstruct the union of the inputs, instead of drifting by the
fraction of a percent that floating-point sliver artefacts on invalid input
otherwise introduce. Inputs are also passed through \code{\link[sf:st_make_valid]{sf::st_make_valid()}}.

With a second layer \code{y}, the same machinery overlays two layers instead of
self-unioning one: both layers are noded together into one planar partition,
and each piece carries the attributes of the \code{x} record and the \code{y} record
that cover it. \code{how} selects which pieces to keep -- the intersection (pieces
covered by both), the union (every piece of either), \code{x} split by \code{y}
(\code{"identity"}), or the parts in exactly one layer (\code{"symdiff"}). With
\code{y = NULL} (the default) the function self-unions \code{x} and \code{how} is ignored.
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Two overlapping squares designated in different years.
sq <- function(a, b) sf::st_polygon(list(rbind(
  c(a, 0), c(b, 0), c(b, 1), c(a, 1), c(a, 0))))
polys <- sf::st_sf(year = c(1990L, 2010L),
                   geometry = sf::st_sfc(sq(0, 2), sq(1, 3)))

# Split into disjoint pieces; earliest year wins where they overlap.
first <- spatial_overlay(polys) |>
  group_by(piece_id) |>
  slice_min(year, n = 1, with_ties = FALSE) |>
  collect_sf()
first

# Two-layer overlay: intersect the squares with a zone layer, keeping both
# sets of attributes on each overlapping piece.
zones <- sf::st_sf(zone = c("A", "B"),
                   geometry = sf::st_sfc(sq(0, 1.5), sq(1.5, 3)))
inter <- spatial_overlay(polys, zones, how = "intersection") |> collect_sf()
inter
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=slice_min]{slice_min()}} / \code{\link[=slice_max]{slice_max()}} to resolve each piece to one winner,
\code{\link[=collect_sf]{collect_sf()}} to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_more.R
\name{polygonize}
\alias{polygonize}
\title{Vectorise a raster into polygons}
\usage{
polygonize(
  x,
  band = 1L,
  dissolve = TRUE,
  na_rm = TRUE,
  values = "value",
  crs = NA,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_raster} (from \code{\link[=vec_open_raster]{vec_open_raster()}}) or a path to a \code{.vec}
raster.}

\item{band}{Band to vectorise (1-based). Default 1.}

\item{dissolve}{If \code{TRUE} (default) merge cells of equal value into one polygon
per value; if \code{FALSE} emit one square polygon per cell.}

\item{na_rm}{Drop nodata cells (\code{TRUE}, default) or vectorise them as a value.}

\item{values}{Name of the output value column. Default \code{"value"}.}

\item{crs}{Coordinate reference system recorded on the node. Defaults to the
raster's EPSG, else unknown.}

\item{flush_rows}{Rows buffered before a spill flush. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} with the value column and a hex-WKB \code{geometry} column,
materialise it with \code{\link[=collect_sf]{collect_sf()}}.
}
\description{
Converts a \code{.vec} raster into polygon features, the inverse of \code{\link[=rasterize]{rasterize()}}.
The raster is read one tile-row strip at a time; within each strip equal-valued
cells along a row collapse to a rectangle, and (with \code{dissolve = TRUE}) the
rectangles are then merged by value through \code{\link[=spatial_dissolve]{spatial_dissolve()}} so each
distinct value becomes a single polygon spanning the whole raster. The result
is a lazy \code{vectra_node} carrying a value column and hex-WKB geometry.
}
\details{
Extraction is the \emph{monoid fold} tier (one strip at a time); the by-value
dissolve rides the \emph{sort / partition} tier of \code{\link[=spatial_dissolve]{spatial_dissolve()}}, localising
each value to a shard before the union. Geometry assembly is delegated to
\pkg{sf} (an optional dependency).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
m <- matrix(c(1, 1, 2, 2, 1, 1, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3), 4, 4,
            byrow = TRUE)
f <- tempfile(fileext = ".vec")
vec_write_raster(m, f, dtype = "f64", extent = c(0, 0, 4, 4))

polys <- polygonize(f)
collect_sf(polys)
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=rasterize]{rasterize()}} for the inverse, \code{\link[=contours]{contours()}} for iso-lines,
\code{\link[=collect_sf]{collect_sf()}} to materialise as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_snap_grid}
\alias{spatial_snap_grid}
\title{Snap a streamed layer's coordinates to a fixed grid}
\usage{
spatial_snap_grid(
  x,
  size,
  geom = "geometry",
  crs = NA,
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{size}{Grid spacing in CRS units (a positive number). Coordinates are
rounded to the nearest multiple; a larger \code{size} snaps more aggressively.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of the snapped geometry with \code{x}'s attributes, backed
by temporary \code{.vtr} spills (removed when the node is garbage-collected) and
carrying the input CRS.
}
\description{
Rounds every coordinate of a streamed layer to a regular grid of spacing
\code{size} (in CRS units) and repairs the result, one batch at a time. This is the
fixed-precision snap-rounding the overlay noder
(\code{\link[=spatial_overlay]{spatial_overlay()}}) applies internally, exposed as a standalone verb: it
merges near-coincident vertices and removes the slivers that floating-point
coordinates leave between shared boundaries, so a layer can be cleaned (or
pre-noded to a common precision) without running a full overlay. Snapping is
done in C straight off the hex-WKB column; one cleaned geometry comes back per
input feature, so attributes ride through untouched.
}
\details{
Geometry travels through the engine as hex-encoded WKB in a string column and
the CRS is carried on the returned node; use \code{\link[=collect_sf]{collect_sf()}} to materialize.
The \pkg{sf} package is an optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
p <- sf::st_polygon(list(rbind(c(0.04, 0.03), c(1.02, 0.01),
                               c(0.98, 1.03), c(0.01, 0.97), c(0.04, 0.03))))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1L, geometry = sf::st_as_binary(sf::st_sfc(p), hex = TRUE)
), f)

# Snap the jittered corners back onto a 0.1 grid.
tbl(f) |> spatial_snap_grid(0.1) |> collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_snap]{spatial_snap()}} to snap toward another layer instead of a grid,
\code{\link[=spatial_overlay]{spatial_overlay()}} whose noding uses the same snap-rounding, \code{\link[=collect_sf]{collect_sf()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_map}
\alias{spatial_map}
\title{Stream a query through an sf transform}
\usage{
spatial_map(
  x,
  fn,
  geom = "geometry",
  coords = NULL,
  crs = NA,
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{fn}{A function (or purrr-style formula such as \code{~ sf::st_buffer(.x, 1000)}) taking one \code{sf} batch and returning an \code{sf} object, \code{sfc}, or plain
data.frame. The active geometry of the return becomes the output geometry.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{coords}{Optional length-2 character vector naming the x and y
coordinate columns to assemble point geometry from (e.g. \code{c("x", "y")}),
for inputs such as \code{\link[=tiff_extract_points]{tiff_extract_points()}} output. The coordinate columns
are retained.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} backed by temporary \code{.vtr} spills (removed when the
node is garbage-collected), carrying the output CRS for \code{\link[=collect_sf]{collect_sf()}}.
}
\description{
Applies a per-feature \pkg{sf} operation (buffer, centroid, area, CRS
transform, simplify, ...) to a lazy vectra query one batch at a time and
returns a new lazy node. The engine pulls one batch, hands it to \code{fn} as an
\code{sf} object, encodes the result back into the stream, and spills to disk, so
peak memory is one batch regardless of result size. This is the streaming,
larger-than-RAM counterpart to running the same \code{sf} call on a whole
in-memory table.
}
\details{
Geometry travels through the engine as hex-encoded WKB in an ordinary string
column (vectra has no native geometry type), and the coordinate reference
system is carried on the returned node rather than in the \code{.vtr} file. Use
\code{\link[=collect_sf]{collect_sf()}} to materialize the result as an \code{sf} object, or \code{\link[=collect]{collect()}}
to get the underlying data.frame with the WKB string column.

Topology is delegated entirely to \pkg{sf}/GEOS; vectra only supplies the
streaming. The \code{sf} package is an optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME = nc$NAME,
  geometry = sf::st_as_binary(sf::st_centroid(sf::st_geometry(nc)),
                              hex = TRUE)
), f)

# Buffer every county centroid by 0.1 degree, streaming.
buffered <- tbl(f) |>
  spatial_map(~ sf::st_buffer(.x, 0.1), crs = sf::st_crs(nc))
collect_sf(buffered)
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_join]{spatial_join()}} to join a streamed side against a resident \code{sf}
object, \code{\link[=collect_sf]{collect_sf()}} to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_topology.R
\name{spatial_polygonize}
\alias{spatial_polygonize}
\title{Build polygonal faces from a line network}
\usage{
spatial_polygonize(
  x,
  by = NULL,
  geom = "geometry",
  crs = NA,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{by}{Character vector of attribute columns to polygonize within: one set
of faces per distinct combination of their values. \code{NULL} (default)
polygonizes the whole layer at once.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of one row per face, carrying the \code{by} columns and the
input CRS, backed by temporary \code{.vtr} spills removed when the node is
garbage-collected.
}
\description{
Forms the polygons enclosed by a set of lines (the QGIS "Polygonize", GEOS
\code{Polygonize}): the inverse of taking polygon boundaries. The lines of each
group are unioned and noded so every crossing becomes a shared vertex, then
the faces of that planar arrangement are returned, one per row. A pile of
lines that does not close any area yields no faces. Like \code{\link[=spatial_dissolve]{spatial_dissolve()}}
and \code{\link[=spatial_construct]{spatial_construct()}} it rides the \strong{partition tier}: \code{x} is spilled once
and routed into one disjoint shard per \code{by} group in a single bounded pass,
then each group's lines are polygonized together. Peak memory is the routing
budget during the pass, then one group's geometry while its faces are built --
partition on a key whose groups fit in memory. With no \code{by}, the whole layer
yields one set of faces.
}
\details{
Each face is new geometry built from the whole group, so it carries the \code{by}
columns only, not the attributes of any single source line. Geometry travels
through the engine as hex-encoded WKB in a string column and the CRS is
carried on the returned node; the noding is \pkg{sf}/GEOS and expects
projected or unprojected planar data. The \pkg{sf} package is an optional
dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
grid <- sf::st_sfc(
  sf::st_linestring(rbind(c(0, 0), c(2, 0))),
  sf::st_linestring(rbind(c(0, 1), c(2, 1))),
  sf::st_linestring(rbind(c(0, 2), c(2, 2))),
  sf::st_linestring(rbind(c(0, 0), c(0, 2))),
  sf::st_linestring(rbind(c(1, 0), c(1, 2))),
  sf::st_linestring(rbind(c(2, 0), c(2, 2))))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  geometry = sf::st_as_binary(grid, hex = TRUE)
), f)

# The four unit cells enclosed by the grid of lines.
tbl(f) |> spatial_polygonize() |> collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_split]{spatial_split()}} to cut existing polygons by a blade,
\code{\link[=spatial_construct]{spatial_construct()}} for hulls and tessellations, \code{\link[=spatial_dissolve]{spatial_dissolve()}} to
merge geometries by group, \code{\link[=collect_sf]{collect_sf()}} to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_topology.R
\name{spatial_eliminate}
\alias{spatial_eliminate}
\title{Merge sliver polygons into a neighbour}
\usage{
spatial_eliminate(
  x,
  max_area,
  by = NULL,
  into = c("longest_border", "largest_area"),
  geom = "geometry",
  crs = NA,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{max_area}{Area threshold in CRS units squared: a feature smaller than
this is a sliver and is merged into a neighbour. Larger values absorb more.}

\item{by}{Character vector of attribute columns whose groups are each cleaned
as an independent coverage. \code{NULL} (default) treats the whole layer as one
coverage.}

\item{into}{How to pick the neighbour a sliver merges into: \code{"longest_border"}
(default) the neighbour sharing the longest boundary, or \code{"largest_area"}
the neighbour with the greatest area.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of the cleaned coverage -- one row per surviving
feature, each carrying its (largest member's) attributes and the input CRS,
backed by temporary \code{.vtr} spills removed when the node is garbage-collected.
}
\description{
Cleans a polygon coverage by absorbing every feature whose area is below
\code{max_area} into an adjacent feature (the QGIS "Eliminate Selected Polygons"):
the sliver removal a per-feature transform cannot do, because the target a
sliver merges into is one of its neighbours, not the sliver itself. Each small
feature is joined to the neighbour it shares the longest border with (or the
largest-area neighbour, with \code{into = "largest_area"}); chains of slivers
collapse so a connected run of small features flows to its single largest
member, whose attribute row survives. A small feature with no neighbour is
kept unchanged, so nothing vanishes. Like \code{\link[=spatial_dissolve]{spatial_dissolve()}} it rides the
\strong{partition tier}: \code{x} is spilled once and routed into one disjoint shard per
\code{by} group in a single bounded pass, and each group is cleaned as an
independent coverage. Peak memory is the routing budget during the pass, then
one group's geometry while its slivers are merged -- partition on a key whose
groups fit in memory. With no \code{by}, the whole layer is one coverage.
}
\details{
Adjacency and shared-border length are \pkg{sf}/GEOS (\link[sf:st_intersects]{sf::st_intersects},
\link[sf:st_boundary]{sf::st_boundary}, \link[sf:st_intersection]{sf::st_intersection}) and expect projected or unprojected
planar data; \code{max_area} is in CRS units squared. Geometry travels through the
engine as hex-encoded WKB in a string column and the CRS is carried on the
returned node. The \pkg{sf} package is an optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
big    <- sf::st_polygon(list(rbind(
  c(0, 0), c(10, 0), c(10, 10), c(0, 10), c(0, 0))))
sliver <- sf::st_polygon(list(rbind(
  c(10, 0), c(10.3, 0), c(10.3, 10), c(10, 10), c(10, 0))))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = c("keep", "sliver"),
  geometry = sf::st_as_binary(sf::st_sfc(big, sliver), hex = TRUE)
), f)

# The thin sliver is absorbed into the square it borders.
tbl(f) |> spatial_eliminate(max_area = 5) |> collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_dissolve]{spatial_dissolve()}} to merge geometries by attribute,
\code{\link[=spatial_simplify]{spatial_simplify()}} for coverage-preserving simplification,
\code{\link[=spatial_topology]{spatial_topology()}} for the shared-edge adjacency, \code{\link[=collect_sf]{collect_sf()}} to
materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_clip}
\alias{spatial_clip}
\title{Clip or erase a streamed layer against a resident mask}
\usage{
spatial_clip(
  x,
  mask,
  erase = FALSE,
  geom = "geometry",
  coords = NULL,
  crs = NA,
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{mask}{An \code{sf} or \code{sfc} object whose dissolved geometry clips (or, with
\code{erase = TRUE}, erases) the stream.}

\item{erase}{If \code{TRUE}, keep the parts of \code{x} \emph{outside} \code{mask} (difference)
rather than inside (intersection). Default \code{FALSE}.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{coords}{Optional length-2 character vector naming the x and y
coordinate columns to assemble point geometry from (e.g. \code{c("x", "y")}),
for inputs such as \code{\link[=tiff_extract_points]{tiff_extract_points()}} output. The coordinate columns
are retained.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of the cut geometry with \code{x}'s attributes, backed by
temporary \code{.vtr} spills and carrying the input CRS.
}
\description{
Streams a large layer \code{x} through the engine and cuts each batch's geometry
against a small resident \code{mask} (a study boundary, a buffer, a set of
patches). By default this clips -- the intersection with the mask, the GIS
"Clip" tool -- keeping only the parts of \code{x} that fall inside \code{mask}. With
\code{erase = TRUE} it instead erases -- the difference, the "Erase"/"Difference"
tool -- keeping the parts of \code{x} outside \code{mask}. The mask is dissolved to a
single geometry once and held resident while the billion-row left stream
flows past one batch at a time.
}
\details{
Geometry travels through the engine as hex-encoded WKB in a string column and
the CRS is carried on the returned node; use \code{\link[=collect_sf]{collect_sf()}} to materialize.
On projected or unprojected planar data the cut runs natively on the GEOS C
API straight off the hex-WKB column (the mask parsed once); geographic
coordinates with spherical geometry on (\code{sf::sf_use_s2()}) and
coordinate-assembled (\code{coords}) input cut through \pkg{sf} instead. When
\code{mask} carries no CRS it inherits the stream's.
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
nc <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
mask <- sf::st_union(nc[nc$NAME \%in\% c("Ashe", "Alleghany"), ])

f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  NAME = nc$NAME,
  geometry = sf::st_as_binary(sf::st_geometry(nc), hex = TRUE)
), f)

# Clip every county polygon to the two-county mask, streaming.
clipped <- tbl(f) |> spatial_clip(mask, crs = sf::st_crs(nc))
collect_sf(clipped)
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_filter]{spatial_filter()}} to keep whole features by location without
cutting them, \code{\link[=spatial_map]{spatial_map()}} for per-feature transforms, \code{\link[=collect_sf]{collect_sf()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_topology.R
\name{spatial_line_merge}
\alias{spatial_line_merge}
\title{Merge contiguous line segments into maximal lines}
\usage{
spatial_line_merge(
  x,
  by = NULL,
  geom = "geometry",
  crs = NA,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{by}{Character vector of attribute columns to merge within: one set of
maximal lines per distinct combination of their values. \code{NULL} (default)
merges the whole layer at once.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of one row per maximal merged line, carrying the \code{by}
columns and the input CRS, backed by temporary \code{.vtr} spills removed when
the node is garbage-collected.
}
\description{
Sews the line segments of each group into the longest possible linestrings
(\code{sf::st_line_merge}, the line counterpart of a dissolve): segments that meet
end to end become one chain, and each chain is emitted as its own row. Where a
plain union of lines returns a single multilinestring of all the parts, this
joins the parts through their shared endpoints; at a crossing where more than
two segments meet the merge is ambiguous and the segments stay separate. Like
\code{\link[=spatial_dissolve]{spatial_dissolve()}} it rides the \strong{partition tier}: \code{x} is spilled once and
routed into one disjoint shard per \code{by} group in a single bounded pass, then
each group's segments are merged together. Peak memory is the routing budget
during the pass, then one group's geometry while it is merged. With no \code{by},
the whole layer is merged at once.
}
\details{
Each merged line is new geometry built from the whole group, so it carries the
\code{by} columns only, not the attributes of any single source segment. Geometry
travels through the engine as hex-encoded WKB in a string column and the CRS
is carried on the returned node. The \pkg{sf} package is an optional
dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
seg <- sf::st_sfc(
  sf::st_linestring(rbind(c(0, 0), c(1, 0))),
  sf::st_linestring(rbind(c(1, 0), c(2, 0))),
  sf::st_linestring(rbind(c(2, 0), c(3, 0))))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  geometry = sf::st_as_binary(seg, hex = TRUE)
), f)

# The three end-to-end segments become one line.
tbl(f) |> spatial_line_merge() |> collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_dissolve]{spatial_dissolve()}} to union geometries by group,
\code{\link[=spatial_explode]{spatial_explode()}} for the opposite direction (multipart to single part),
\code{\link[=collect_sf]{collect_sf()}} to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{spatial_snap}
\alias{spatial_snap}
\title{Snap a streamed layer toward a resident reference layer}
\usage{
spatial_snap(
  x,
  y,
  tolerance,
  geom = "geometry",
  coords = NULL,
  crs = NA,
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{y}{An \code{sf} or \code{sfc} object: the resident reference layer to snap toward.}

\item{tolerance}{Snapping distance in CRS units (a positive number). Vertices
and edges of \code{x} within this distance of \code{y} are moved onto \code{y}.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{coords}{Optional length-2 character vector naming the x and y
coordinate columns to assemble point geometry from (e.g. \code{c("x", "y")}),
for inputs such as \code{\link[=tiff_extract_points]{tiff_extract_points()}} output. The coordinate columns
are retained.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of the snapped geometry with \code{x}'s attributes, backed
by temporary \code{.vtr} spills (removed when the node is garbage-collected) and
carrying the input CRS.
}
\description{
Streams a large layer \code{x} through the engine and snaps each batch's vertices
toward a small resident reference layer \code{y} when they lie within \code{tolerance}
(in CRS units), one batch at a time (the QGIS "snap geometries to layer").
Vertices and edges of \code{x} closer than \code{tolerance} to \code{y} are pulled onto \code{y},
which closes the small gaps and overshoots between two layers that should
share a boundary. The reference layer stays resident while the billion-row
left stream flows past; the snap itself is \pkg{sf}'s \code{\link[sf:st_snap]{sf::st_snap()}}.
}
\details{
Geometry travels through the engine as hex-encoded WKB in a string column and
the CRS is carried on the returned node; use \code{\link[=collect_sf]{collect_sf()}} to materialize.
When \code{y} carries no CRS it inherits the stream's. The \pkg{sf} package is an
optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
ref <- sf::st_sfc(sf::st_linestring(rbind(c(0, 0), c(10, 0))))
line <- sf::st_linestring(rbind(c(0, 0.2), c(5, 0.1), c(10, 0.2)))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = 1L, geometry = sf::st_as_binary(sf::st_sfc(line), hex = TRUE)
), f)

# Pull the near-zero vertices down onto the reference line.
tbl(f) |> spatial_snap(ref, tolerance = 0.5) |> collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_snap_grid]{spatial_snap_grid()}} to snap to a grid instead of a layer,
\code{\link[=spatial_clip]{spatial_clip()}} for the resident-mask streaming pattern, \code{\link[=collect_sf]{collect_sf()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_topology.R
\name{spatial_centerline}
\alias{spatial_centerline}
\title{Trace the centerline (medial axis) of streamed polygons}
\usage{
spatial_centerline(
  x,
  density = NULL,
  prune = 0,
  geom = "geometry",
  crs = NA,
  out_geom = NULL,
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{density}{Boundary sampling spacing in CRS units: the polygon outline is
densified to at most this vertex spacing before the Voronoi diagram is
built. \code{NULL} (default) uses one-hundredth of each polygon's bounding-box
diagonal.}

\item{prune}{Drop centerline branches shorter than this length (CRS units),
removing the short spurs the skeleton grows toward convex corners. \code{0}
(default) keeps every branch.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{out_geom}{Name of the output geometry column. Defaults to \code{geom}
(or \code{"geometry"} when \code{coords} is used).}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of the centerlines, each carrying its source polygon's
attributes (replicated if the centerline is several lines) and the input
CRS, backed by temporary \code{.vtr} spills removed when the node is
garbage-collected.
}
\description{
Approximates the centerline of every polygon in a streamed layer, one batch
at a time -- the medial axis a per-feature transform such as a buffer cannot
produce. Each polygon's boundary is densified and its Voronoi diagram taken;
the Voronoi edges that fall inside the polygon trace the points equidistant
from two stretches of boundary, which is its skeleton, and they are merged
into maximal lines. This is the usual approximation for river or road
centerlines from a filled shape; \code{prune} drops the short branches that the
skeleton grows toward convex corners. A non-polygon geometry passes through
unchanged.
}
\details{
The centerline is an approximation whose detail is set by \code{density} (the
boundary sampling spacing): finer sampling traces the axis more closely at
more cost. Geometry travels through the engine as hex-encoded WKB in a string
column and the CRS is carried on the returned node; the Voronoi construction
is \pkg{sf}/GEOS and expects projected or unprojected planar data. The
\pkg{sf} package is an optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
road <- sf::st_polygon(list(rbind(
  c(0, 0), c(10, 0), c(10, 2), c(0, 2), c(0, 0))))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  geometry = sf::st_as_binary(sf::st_sfc(road), hex = TRUE)
), f)

# The centerline runs down the middle of the strip.
tbl(f) |> spatial_centerline(density = 0.25, prune = 0.5) |> collect_sf()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_construct]{spatial_construct()}} with \code{kind = "pole"} for the single deepest
interior point, \code{\link[=spatial_simplify]{spatial_simplify()}} to simplify a coverage, \code{\link[=collect_sf]{collect_sf()}}
to materialize as \code{sf}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{focal}
\alias{focal}
\title{Moving-window (focal) statistics over a streamed raster}
\usage{
focal(
  x,
  w = matrix(1, 3, 3),
  fun = c("mean", "sum", "min", "max", "sd", "median"),
  na.rm = TRUE,
  band = 1L,
  path = NULL,
  dtype = "f32",
  compression = c("fast", "balanced", "max")
)
}
\arguments{
\item{x}{A \code{vectra_raster} (from \code{\link[=vec_open_raster]{vec_open_raster()}}) or a path to a \code{.vec}
raster.}

\item{w}{A numeric weight matrix with odd dimensions, or a single positive
odd integer \code{k} for a \verb{k x k} window of ones. Default \code{matrix(1, 3, 3)}.}

\item{fun}{Window statistic: one of \code{"sum"}, \code{"mean"}, \code{"min"}, \code{"max"},
\code{"sd"}, \code{"median"}. Default \code{"mean"}.}

\item{na.rm}{Skip nodata cells inside the window (\code{TRUE}, default) or let
them propagate \code{NA} (\code{FALSE}).}

\item{band}{Band to read (1-based). Default 1.}

\item{path}{Optional output \code{.vec} path. When given the result is streamed to
disk and the opened \code{\link[=vec_open_raster]{vec_open_raster()}} handle is returned invisibly; when
\code{NULL} the result is returned as an in-memory matrix.}

\item{dtype}{Storage dtype for \code{.vec} output (see \code{\link[=vec_write_raster]{vec_write_raster()}}).
Default \code{"f32"}.}

\item{compression}{Compression effort for \code{.vec} output. Default \code{"fast"}.}
}
\value{
When \code{path} is \code{NULL}, a numeric matrix (row 1 northmost) carrying
\code{gt}, \code{extent}, \code{crs}, and \code{fun} attributes. When \code{path} is given, the
written \code{vectra_raster} handle (invisibly).
}
\description{
Applies a moving window to a \code{.vec} raster, reading the input one tile-row
strip at a time -- each strip expanded by the kernel radius (a halo read) so
window neighbours are available without ever holding the whole grid resident.
The per-window statistic is computed in C. When \code{path} is given the output is
streamed straight back to a new \code{.vec} one tile-row at a time, so neither the
input nor the output band is ever fully in memory; this is the raster op that
runs out of core where an in-memory engine needs the whole raster at once.
}
\details{
This is the \emph{sort / partition} tier of the spatial toolbox: bounded to one
haloed strip at a time, exploiting tile locality.

The window \code{w} is a numeric weight matrix with odd dimensions (or a single
odd integer \code{k}, shorthand for a \verb{k x k} matrix of ones). \code{NA} weights mark
cells outside the window. For \code{fun = "sum"}/\code{"mean"} the weights scale the
values (sum is \code{sum(w * x)}, mean is \code{sum(w * x) / sum(w)}); for the other
statistics a finite weight only marks membership. With \code{na.rm = TRUE} (the
default) nodata cells inside the window are skipped; with \code{na.rm = FALSE} any
nodata cell -- including a window that runs off the raster edge -- makes the
result \code{NA}, matching the resident behaviour.
}
\examples{
m <- matrix(1:36, 6, 6, byrow = TRUE)
f <- tempfile(fileext = ".vec")
vec_write_raster(m, f, dtype = "f64", extent = c(0, 0, 6, 6))

# 3x3 mean smoother; edge cells see off-raster neighbours.
focal(f, w = matrix(1, 3, 3), fun = "mean")
unlink(f)

}
\seealso{
\code{\link[=terrain]{terrain()}} for DEM derivatives built on the same strip pass,
\code{\link[=zonal]{zonal()}} for per-zone summaries.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chunk.R, R/offload.R
\name{collect_chunked}
\alias{collect_chunked}
\alias{collect_chunked.default}
\alias{collect_chunked.vectra_node}
\alias{collect_chunked.vectra_partition}
\title{Fold a function over a query, one batch at a time}
\usage{
collect_chunked(x, f, .init = NULL, combine = NULL, commutative = FALSE)

\method{collect_chunked}{default}(x, f, .init = NULL, combine = NULL, commutative = FALSE)

\method{collect_chunked}{vectra_node}(x, f, .init = NULL, combine = NULL, commutative = FALSE)

\method{collect_chunked}{vectra_partition}(x, f, .init = NULL, combine = NULL, commutative = FALSE)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_csv]{tbl_csv()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, ... and
any chain of verbs). It is consumed: after \code{collect_chunked()} returns, the
stream is drained and \code{x} cannot be collected again.}

\item{f}{A function of two arguments \verb{function(acc, chunk)} returning the
updated accumulator. \code{chunk} is a data.frame holding the next batch of
rows.}

\item{.init}{Initial accumulator value. Passed to \code{f} with the first batch
and returned unchanged if the query yields no rows. When \code{combine} is
supplied this is also the monoid identity (the value \code{combine} leaves
unchanged).}

\item{combine}{Optional function \verb{function(acc, acc)} that merges two
accumulators. Supplying it declares the reduction a monoid with \code{.init} as
identity, which is what lets the fold run over the independent shards of a
partition (\code{offload(x, by = ...)}) and have the partial results merged
correctly. For a plain node the stream is a single sequence, so \code{combine}
is not needed and is ignored.}

\item{commutative}{Logical; declare that \code{combine} does not depend on the
order of its arguments. Lets a partitioned fold merge shards in any order
(no stable sort required). Default \code{FALSE}.}
}
\value{
The final accumulator. For a node: \code{f} applied left-to-right across
every batch, seeded with \code{.init}. For a partition: each shard folded with
\code{f}/\code{.init}, then those per-shard accumulators merged with \code{combine}.
}
\description{
Streams a lazy query through R in bounded pieces and reduces them with \code{f},
instead of materializing the whole result the way \code{\link[=collect]{collect()}} does. The
engine pulls one batch (a data.frame of up to a few hundred thousand rows)
at a time; \code{f} is called as \code{f(acc, chunk)} and its return value becomes the
accumulator for the next batch. Peak memory is one batch plus whatever the
accumulator holds, so a result far larger than RAM can be reduced to a small
summary in a single pass.
}
\details{
This is the streaming counterpart to a fold (\code{Reduce()}): use it when the
query returns more rows than fit in memory but the \emph{reduction} is small. A
running count, per-group sufficient statistics, the cross-products \verb{X'X} and
\verb{X'y} behind a linear fit, an online mean or histogram - all accumulate in
bounded space across the stream. When you instead need the model-fitting
consumer to drive the iteration (and to re-read the data on each pass, as an
iteratively reweighted GLM does), use \code{\link[=chunk_feeder]{chunk_feeder()}}.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)

# Row count without materializing the result.
collect_chunked(tbl(f), function(acc, chunk) acc + nrow(chunk), .init = 0L)

# Accumulate the normal-equation pieces X'X and X'y for an exact OLS fit
# of mpg ~ wt + hp, in one streaming pass.
acc <- collect_chunked(
  tbl(f) |> select(mpg, wt, hp),
  function(acc, chunk) {
    X <- cbind(1, chunk$wt, chunk$hp)
    y <- chunk$mpg
    list(XtX = acc$XtX + crossprod(X), Xty = acc$Xty + crossprod(X, y))
  },
  .init = list(XtX = matrix(0, 3, 3), Xty = matrix(0, 3, 1))
)
solve(acc$XtX, acc$Xty)            # same as coef(lm(mpg ~ wt + hp, mtcars))
unlink(f)

}
\seealso{
\code{\link[=chunk_feeder]{chunk_feeder()}} for pull-based consumers such as \code{biglm::bigglm()},
\code{\link[=offload]{offload()}} for the replay cache and the partitioned monoidal reduce,
\code{\link[=group_map]{group_map()}} and \code{\link[=group_modify]{group_modify()}} for per-shard application, and
\code{\link[=collect]{collect()}} to materialize the full result.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verbs_utility.R
\name{slice_head}
\alias{slice_head}
\alias{slice_tail}
\alias{slice_min}
\alias{slice_max}
\title{Select first or last rows}
\usage{
slice_head(.data, n = 1L)

slice_tail(.data, n = 1L)

slice_min(.data, order_by, n = 1L, with_ties = TRUE)

slice_max(.data, order_by, n = 1L, with_ties = TRUE)
}
\arguments{
\item{.data}{A \code{vectra_node} object.}

\item{n}{Number of rows to select.}

\item{order_by}{Column to order by (for \code{slice_min}/\code{slice_max}).}

\item{with_ties}{If \code{TRUE} (default), includes all rows that tie with the
nth value. If \code{FALSE}, returns exactly \code{n} rows.}
}
\value{
A \code{vectra_node} for \code{slice_head()}, for grouped
\code{slice_min()}/\code{slice_max()}, and for ungrouped \code{slice_min/max(..., with_ties = FALSE)}. A data.frame for \code{slice_tail()} and ungrouped
\code{slice_min/max(..., with_ties = TRUE)} (the default), since these must
materialize all rows.
}
\description{
Select first or last rows
}
\details{
When \code{slice_min()}/\code{slice_max()} follow \code{\link[=group_by]{group_by()}}, the n smallest/largest
rows are taken within each group and the whole winning row is kept (every
column, including geometry carried as a string). \code{with_ties = FALSE} returns
exactly \code{n} rows per group; \code{with_ties = TRUE} keeps rows tied at the nth
value via min-rank. The \code{n = 1}, \code{with_ties = FALSE} case streams: it holds
only the running winner per group, so memory scales with the number of groups
(the result size), not the input. Other grouped cases buffer their input.
}
\examples{
f <- tempfile(fileext = ".vtr")
write_vtr(mtcars, f)
tbl(f) |> slice_head(n = 3) |> collect()
tbl(f) |> slice_min(order_by = mpg, n = 3) |> collect()
tbl(f) |> slice_max(order_by = mpg, n = 3) |> collect()
# earliest row per group, geometry/attrs preserved:
tbl(f) |> group_by(cyl) |> slice_min(mpg, n = 1, with_ties = FALSE) |> collect()
unlink(f)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial.R
\name{grid}
\alias{grid}
\title{Define a uniform grid for a partitioned spatial join}
\usage{
grid(cellsize, origin = c(0, 0))
}
\arguments{
\item{cellsize}{Cell size: a single number for square cells, or
\code{c(cellsize_x, cellsize_y)}.}

\item{origin}{Grid origin \code{c(x0, y0)} (a cell corner). Default \code{c(0, 0)}.}
}
\value{
A \code{vectra_grid} specification to pass as \code{spatial_join(partition =)}.
}
\description{
Describes the regular grid that \code{\link[=spatial_join]{spatial_join()}} uses to partition two
streamed layers for the both-sides-larger-than-RAM case. Cell \verb{(cx, cy)}
covers \verb{[origin_x + cx * cellsize_x, origin_x + (cx + 1) * cellsize_x)} and
likewise in y, so a coordinate maps to the cell \code{floor((coord - origin) / cellsize)}. Pick a \code{cellsize} comparable to the scale of the join (large
enough that most cells hold a workable shard, small enough that one cell's
features fit in memory); for an extended-on-extended join choose it larger
than the left features.
}
\examples{
grid(1000)
grid(c(0.5, 0.25), origin = c(-180, -90))

}
\seealso{
\code{\link[=spatial_join]{spatial_join()}} for the join it partitions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial_topology.R
\name{spatial_topology}
\alias{spatial_topology}
\title{Build the shared-edge topology of a polygon coverage}
\usage{
spatial_topology(
  x,
  id = NULL,
  by = NULL,
  geom = "geometry",
  crs = NA,
  face_cols = c("face1", "face2"),
  flush_rows = NULL
)
}
\arguments{
\item{x}{A \code{vectra_node} (from \code{\link[=tbl]{tbl()}}, \code{\link[=tbl_tiff]{tbl_tiff()}}, any verb chain, ...).
It is consumed by the stream.}

\item{id}{Optional name of a column of \code{x} whose value identifies each polygon
in the face columns. \code{NULL} (default) uses the 1-based feature order within
the group.}

\item{by}{Character vector of attribute columns whose groups are each built as
an independent coverage. \code{NULL} (default) treats the whole layer as one
coverage.}

\item{geom}{Name of the input geometry column holding hex-WKB or WKT strings.
Default \code{"geometry"}. Ignored when \code{coords} is given.}

\item{crs}{Coordinate reference system of the input geometry, in any form
\code{\link[sf:st_crs]{sf::st_crs()}} accepts (EPSG integer, WKT, proj string). Defaults to the
CRS the upstream node carries, or unknown.}

\item{face_cols}{Length-2 character vector naming the two output columns that
hold the identifiers of the polygons on either side of each arc. Default
\code{c("face1", "face2")}.}

\item{flush_rows}{Transformed rows buffered before a spill flush. Larger
values mean fewer, bigger temporary files. Defaults to
\code{getOption("vectra.spatial_flush", 5e5)}.}
}
\value{
A \code{vectra_node} of one row per arc -- the arc geometry plus the two
face-identifier columns (and any \code{by} columns) -- carrying the input CRS and
backed by temporary \code{.vtr} spills removed when the node is garbage-collected.
}
\description{
Decomposes a polygon coverage into the arcs of its planar topology: each
border is returned once, tagged with the polygons on either side. Where the
raw boundaries of adjacent polygons each carry their own copy of a shared
edge, this nodes the unioned boundaries so a shared border is a single arc
carrying the identifiers of both neighbours -- the "build topology" of a GIS,
and the adjacency a dissolve or a coverage edit needs. An internal arc names
two faces; an outer arc names one and leaves the other side \code{NA}. Like
\code{\link[=spatial_dissolve]{spatial_dissolve()}} it rides the \strong{partition tier}: \code{x} is spilled once and
routed into one disjoint shard per \code{by} group in a single bounded pass, and
each group is treated as an independent coverage. Peak memory is the routing
budget during the pass, then one group's geometry while its arcs are built.
}
\details{
Geometry travels through the engine as hex-encoded WKB in a string column and
the CRS is carried on the returned node; the noding is \pkg{sf}/GEOS and
expects projected or unprojected planar data. The \pkg{sf} package is an
optional dependency (Suggests).
}
\examples{
\dontshow{if (requireNamespace("sf", quietly = TRUE)) withAutoprint(\{ # examplesIf}
p1 <- sf::st_polygon(list(rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))))
p2 <- sf::st_polygon(list(rbind(c(1, 0), c(2, 0), c(2, 1), c(1, 1), c(1, 0))))
f <- tempfile(fileext = ".vtr")
write_vtr(data.frame(
  id = c("a", "b"),
  geometry = sf::st_as_binary(sf::st_sfc(p1, p2), hex = TRUE)
), f)

# The shared edge appears once, tagged with both neighbours.
tbl(f) |> spatial_topology(id = "id") |> collect()
unlink(f)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=spatial_polygonize]{spatial_polygonize()}} to rebuild faces from arcs,
\code{\link[=spatial_dissolve]{spatial_dissolve()}} to merge geometries by group, \code{\link[=collect_sf]{collect_sf()}} to
materialize as \code{sf}.
}
